/*
 * Ubitrack - Library for Ubiquitous Tracking
 * Copyright 2006, Technische Universitaet Muenchen, and individual
 * contributors as indicated by the @authors tag. See the
 * copyright.txt in the distribution for a full listing of individual
 * contributors.
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA, or see the FSF site: http://www.fsf.org.
 */

/**
 * @ingroup Database client library
 * @file
 * Implements interface for communication with the trackframe database server
 *
 * @author Peter Keitler <keitler@in.tum.de>
 */

#include <stdlib.h>
#include <ios>
#include <iostream>
#include <fstream>
#include <curl/curl.h>
#include <log4cpp/Category.hh>

#include "DatabaseClient.h"
#include "Exception.h"
#include "Logging.h"


// get a logger
static log4cpp::Category& logger( log4cpp::Category::getInstance( "DataflowClient" ) );

namespace DbClient {
using namespace std;

static char * tmp_file_name;
static int streamCount;



bool init( void )
{
    LOG4CPP_INFO( logger, "DbClient::init()" );
    streamCount = 0;

    try {
	// initialize log4cpp
	initLogging();
    }
    catch( DbClient::Exception& e )
    {
	LOG4CPP_ERROR( logger, "DbClient::init(): " << e );
	return false;
    }
    return true;
}


istream * getInputStream( char * url ) 
{
    CURLcode res;
    CURL * curl;
    FILE * tmp_file;
    istream * srg = NULL;

    LOG4CPP_INFO( logger, "DbClient::getInputStream()" );

    try {
	curl = curl_easy_init();

	if ( curl == NULL ) 
	    DBCLIENT_THROW("Error initializing cURL interface");

	if ( streamCount > 0 ) 
	    DBCLIENT_THROW("Database library only supports  o n e  open stream at the moment");

	// CURL wants to have a fileptr. Thus, retrieve unique name for temporary file.
	tmp_file_name = tmpnam( NULL );
	if ( tmp_file_name == NULL ) 
	    DBCLIENT_THROW("Error retrieving temporary file name");

	tmp_file = fopen( tmp_file_name, "w+" );
	if ( tmp_file == NULL ) 
	    DBCLIENT_THROW("Error opening temporary file");

	curl_easy_setopt(curl, CURLOPT_URL, url);
	curl_easy_setopt(curl, CURLOPT_WRITEDATA, tmp_file);
	res = curl_easy_perform(curl);
	curl_easy_cleanup(curl);
	std::fclose( tmp_file );
	if ( res != CURLE_OK )
	    DBCLIENT_THROW("Error retrieving data from URL");
    
	srg = new ifstream( tmp_file_name );
	if ( srg == NULL )
	    DBCLIENT_THROW("Error creating file input stream");

	streamCount ++;
    }
    catch( DbClient::Exception& e )
    {
	LOG4CPP_ERROR( logger, "DbClient::getInputStream(): " << e );
	return NULL;
    }

    return srg;
}


bool closeStream( ios * stream ) 
{
    LOG4CPP_INFO( logger, "DbClient::closeStream()" );

    streamCount --;
    if ( stream != NULL )
	delete stream;
    std::remove( tmp_file_name );

    return true;
}


bool cleanup( void ) 
{
    LOG4CPP_INFO( logger, "DbClient::cleanup()" );

    // Do nothing
    return true;
}


} // namespace DbClient
