/*
 * Ubitrack - Library for Ubiquitous Tracking
 * Copyright 2006, Technische Universitaet Muenchen, and individual
 * contributors as indicated by the @authors tag. See the
 * copyright.txt in the distribution for a full listing of individual
 * contributors.
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA, or see the FSF site: http://www.fsf.org.
 */


#include <stdlib.h>
#include <iostream>
#include <boost/thread.hpp>
#include <boost/program_options.hpp>

#include <Ubitrack/Facade/AdvancedFacade.h>
#include <Ubitrack/Util/Exception.h>
#include <Ubitrack/Util/Logging.h>

using namespace Ubitrack;


int main( int ac, char** av )
{
 	try
	{
		// initialize logging
		Util::initLogging();

		// program options
		std::string sServerAddress;
		std::string sUtqlFile;
		std::string sExtraUtqlFile;
		std::string sComponentsPath;

		try
		{
			// describe program options
			namespace po = boost::program_options;
			po::options_description poDesc( "Allowed options", 80 );
			poDesc.add_options()
				( "help", "print this help message" )
				( "server,s", po::value< std::string >( &sServerAddress ), "Ubitrack server address <host>[:<port>] to connect to" )
				( "components_path", po::value< std::string >( &sComponentsPath ), "Directory from which to load components" )
				( "utql", po::value< std::string >( &sUtqlFile ), "UTQL request or response file, depending on whether a server is specified. "
					"Without specifying this option, the UTQL file can also be given directly on the command line." )
				( "extra-dataflow", po::value< std::string >( &sExtraUtqlFile ), "Additional UTQL response file to be loaded directly without using the server" )
			;
			
			// specify default options
			po::positional_options_description inputOptions;
			inputOptions.add( "utql", 1 );		
			
			// parse options from command line and environment
			po::variables_map poOptions;
			po::store( po::command_line_parser( ac, av ).options( poDesc ).positional( inputOptions ).run(), poOptions );
			po::store( po::parse_environment( poDesc, "UBITRACK_" ), poOptions );
			po::notify( poOptions );
			
			// print help message if nothing specified
			if ( poOptions.count( "help" ) || sUtqlFile.empty() )
			{
				std::cout << "Syntax: utConsole [options] [--utql] <UTQL file>" << std::endl << std::endl;
				std::cout << poDesc << std::endl;
				return 1;
			}
		}
		catch( std::exception& e )
		{
			std::cerr << "Error parsing command line parameters : " << e.what() << std::endl;
			std::cerr << "Try utConsole --help for help" << std::endl;
			return 1;
		}		
		
		// configure ubitrack
		std::cout << "Loading components..." << std::endl << std::flush;
		Facade::AdvancedFacade utFacade( sComponentsPath );

		if ( sServerAddress.empty() )
		{
			std::cout << "Instantiating dataflow network from " << sUtqlFile << "..." << std::endl << std::flush;
			utFacade.loadDataflow( sUtqlFile );
		}
		else
		{
			if ( !sExtraUtqlFile.empty() )
				utFacade.loadDataflow( sExtraUtqlFile, false );
				
			std::cout << "Connecting to server " << sServerAddress << "..." << std::endl << std::flush;
			utFacade.connectToServer( sServerAddress );
			
			std::cout << "Sending UTQL to to server " << sUtqlFile << "..." << std::endl << std::flush;
			utFacade.sendUtqlToServer( sUtqlFile );
		}

		std::cout << "Starting dataflow" << std::endl;
		utFacade.startDataflow();

		std::cout << "Press RETURN to stop..." << std::endl << std::flush;
		std::cin.get();

		utFacade.stopDataflow();

		std::cout << "Finished, cleaning up..." << std::endl << std::flush;
	}
	catch( Util::Exception& e )
	{
		std::cerr << e << std::endl;
	}

}
