/*
 * Ubitrack - Library for Ubiquitous Tracking
 * Copyright 2006, Technische Universitaet Muenchen, and individual
 * contributors as indicated by the @authors tag. See the
 * copyright.txt in the distribution for a full listing of individual
 * contributors.
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA, or see the FSF site: http://www.fsf.org.
 */


#include <boost/thread.hpp>
#include <boost/program_options.hpp>

#include <log4cpp/Category.hh>

#include <Ubitrack/Facade/AdvancedFacade.h>
#include <Ubitrack/Facade/DataflowObserver.h>
#include <Ubitrack/Measurement/Measurement.h>
#include <Ubitrack/Util/Exception.h>
#include <Ubitrack/Util/Logging.h>
#include <Ubitrack/Util/OS.h>

static log4cpp::Category& logger( log4cpp::Category::getInstance( "IdentityMapper" ) );

using namespace Ubitrack;

double g_fUpperDistanceThreshold = 1.0;
double g_fLowerDistanceThreshold = 0.5;

Measurement::Timestamp absTimeDiff( Measurement::Timestamp a, Measurement::Timestamp b )
{ return a > b ? a - b : b - a; }

class IdentityMapper
	: protected Facade::DataflowObserver
{
public:
	/** constructor */
	IdentityMapper( Facade::AdvancedFacade& facade );
	
	/** needs to be called regularly */
	void check();
	
protected:
	
	/** stores information about a tracked object */
	struct TrackedObject
	{
		std::string sName;
		std::string sMatchedObject;
		std::string sIdentityPatternName;
		Math::Vector< 2 > lastPosition;
		Measurement::Timestamp lastTime;
		int nUsageCounter;
	
		TrackedObject()
			: lastTime( 0 )
			, nUsageCounter( 0 )
		{}
	
		bool isMatched()
		{ return !sMatchedObject.empty(); }
		
		void updatePosition( const Measurement::Position& pos )
		{ 
			lastTime = pos.time(); 
			lastPosition[ 0 ] = (*pos)( 0 );
			lastPosition[ 1 ] = (*pos)( 1 );
			LOG4CPP_DEBUG( logger, "Position update " << sName << ": " << lastPosition );
		}
		
	};
	
	// reference to facade
	Facade::AdvancedFacade& m_rFacade;
	
	typedef std::map< std::string, TrackedObject > TrackedObjectsMap;
	
	// anonymous objects
	TrackedObjectsMap m_anonymousObjects;
	
	// candidate objects
	TrackedObjectsMap m_candidateObjects;

	// mapping pattern to object ids
	typedef std::map< std::string, std::string > PatternToObjMap;
	PatternToObjMap m_patternToObjMap;
	
	// thread synchronization
	boost::mutex m_objectMutex;

	// implementation of DataflowObserver interface
	virtual void notifyAddComponent( const std::string& sPatternName, const std::string& sComponentName, const Graph::UTQLSubgraph& pattern ) ;
		
	virtual void notifyDeleteComponent( const std::string& sPatternName, const std::string& sComponentName );
};


IdentityMapper::IdentityMapper( Facade::AdvancedFacade& facade )
	: m_rFacade( facade )
{
	facade.addDataflowObserver( this );
}


void IdentityMapper::check()
{
	LOG4CPP_DEBUG( logger, "running check()" );
	
	boost::mutex::scoped_lock l( m_objectMutex );
	
	// really cheap solution, better use Munkres

	// remove untracked candidate objects
	LOG4CPP_TRACE( logger, "Checking for deleted candidate objects" );
	for ( TrackedObjectsMap::iterator candIt = m_candidateObjects.begin(); candIt != m_candidateObjects.end(); )
		if ( !candIt->second.nUsageCounter )
		{
			LOG4CPP_DEBUG( logger, "Removing candidate object " << candIt->first );
			m_candidateObjects.erase( candIt++ );
		}
		else
			candIt++;
	
	// iterate all anonymous objects
	LOG4CPP_TRACE( logger, "Checking anonymous objects" );
	for ( TrackedObjectsMap::iterator anonIt = m_anonymousObjects.begin(); anonIt != m_anonymousObjects.end(); )
		if ( !anonIt->second.nUsageCounter )
		{
			// object no longer tracked
			if ( anonIt->second.isMatched() )
			{
				// matched anonymous object removed -> remove identity edge
				LOG4CPP_INFO( logger, "Removing matching of " << anonIt->first );
				m_rFacade.sendUtqlToServerString( "<UTQLRequest><Pattern name=\"" + anonIt->second.sIdentityPatternName + "\"/></UTQLRequest>" );
			}
			
			LOG4CPP_DEBUG( logger, "Removing anonymous object " << anonIt->first );
			m_anonymousObjects.erase( anonIt++ );
		}
		else 
		{
			// do we have a recent measurement?
			if ( anonIt->second.lastTime > Measurement::now() - 500000000 )
			{
				if ( anonIt->second.isMatched() )
				{
					LOG4CPP_TRACE( logger, "Checking matched object " << anonIt->first );
					TrackedObjectsMap::iterator matchedCandIt = m_candidateObjects.find( anonIt->second.sMatchedObject );
					if ( matchedCandIt == m_candidateObjects.end() )
						// matched candidate object has disappeared -- maintain matching
						;
					else
						// measurements approximately synchronous?
						if ( absTimeDiff( anonIt->second.lastTime, matchedCandIt->second.lastTime ) < 200000000 )
						{
							double matchedDist = boost::numeric::ublas::norm_2( anonIt->second.lastPosition - matchedCandIt->second.lastPosition );;
							if ( matchedDist > g_fUpperDistanceThreshold )
							{
								// wrong matching -> remove
								LOG4CPP_INFO( logger, "Removing object matching " << anonIt->first << "->" << matchedCandIt->first );
								
								m_rFacade.sendUtqlToServerString( "<UTQLRequest><Pattern name=\"" + anonIt->second.sIdentityPatternName + "\"/></UTQLRequest>" );
								matchedCandIt->second.sMatchedObject.erase();
								anonIt->second.sMatchedObject.erase();
							}
						}
				}
				else
				{
					// not matched -> find closest candidate object
					LOG4CPP_TRACE( logger, "Checking unmatched object " << anonIt->first );
					
					double bestDist = 1e10;
					TrackedObjectsMap::iterator bestCandIt = m_candidateObjects.end();
					for ( TrackedObjectsMap::iterator candIt = m_candidateObjects.begin(); candIt != m_candidateObjects.end(); candIt++ )
						// approximately same-time measurements?
						if ( absTimeDiff( anonIt->second.lastTime, candIt->second.lastTime ) < 200000000 )
						{
							double dist = boost::numeric::ublas::norm_2( anonIt->second.lastPosition - candIt->second.lastPosition );
							
							// check for best distance
							if ( dist < bestDist )
							{
								bestDist = dist;
								bestCandIt = candIt;
							}
						}
					
					if ( bestCandIt != m_candidateObjects.end() && bestDist < g_fLowerDistanceThreshold && !bestCandIt->second.isMatched() )
					{
						// establish matching
						LOG4CPP_INFO( logger, "Establishing object matching " << anonIt->first << "->" << bestCandIt->first );
						
						anonIt->second.sMatchedObject = bestCandIt->first;
						bestCandIt->second.sMatchedObject = anonIt->first;
						anonIt->second.sIdentityPatternName = "Identity" + anonIt->first + "Is" + bestCandIt->first;
						
						// send to server
						std::string sPattern = 
							"<UTQLRequest>\n"
							"<Pattern name=\"" + anonIt->second.sIdentityPatternName + "\">\n"
							"   <Output>\n"
							"      <Node name=\"A\" id=\"" + anonIt->first + "\"/>\n"
							"      <Node name=\"B\" id=\"" + bestCandIt->first + "\"/>\n"
							"      <Edge name=\"Identity\" source=\"A\" destination=\"B\">\n"
							"         <Attribute name=\"type\" value=\"identity\"/>\n"
							"      </Edge>\n"
							"   </Output>\n"
							"</Pattern>\n"
							"</UTQLRequest>";
							
						LOG4CPP_TRACE( logger, "Sending: " << sPattern );

						m_rFacade.sendUtqlToServerString( sPattern );
					}
				}
			}
			
			anonIt++;
		}
}


void IdentityMapper::notifyAddComponent( const std::string& sPatternName, const std::string& sComponentName, 
	const Graph::UTQLSubgraph& pattern )
{
	// called whenever a new object is detected
	boost::mutex::scoped_lock l( m_objectMutex );
	
	LOG4CPP_TRACE( logger, "notifyAddComponent( " << sPatternName << ", " << sComponentName << " )" );
	
	// is this an anonymous or a candidate object?
	if ( sPatternName == "AnonymousObjectQuery" )
	{
		LOG4CPP_TRACE( logger, "Adding anonymous object, pattern " << sComponentName );

		// anonymous object -> get ID
		std::string sObjectID = pattern.getNode( "Object" )->m_QualifiedName;
		
		// add to list
		m_anonymousObjects[ sObjectID ].nUsageCounter++;
		m_anonymousObjects[ sObjectID ].sName = sObjectID;
		m_patternToObjMap[ sComponentName ] = sObjectID;

		// set callback
		m_rFacade.setCallback< Measurement::Position >( sComponentName, 
			boost::bind( &TrackedObject::updatePosition, &m_anonymousObjects[ sObjectID ], _1 ) );
			
		LOG4CPP_DEBUG( logger, "Added anonymous object " << sObjectID << ", pattern " << sComponentName );
	}
	else if ( sPatternName == "CandidateObjectQuery" )
	{
		LOG4CPP_TRACE( logger, "Adding candidate object, pattern " << sComponentName );
		
		// candidate object -> get ID
		std::string sObjectID = pattern.getNode( "Object" )->m_QualifiedName;

		// add to list
		m_candidateObjects[ sObjectID ].nUsageCounter++;
		m_candidateObjects[ sObjectID ].sName = sObjectID;
		m_patternToObjMap[ sComponentName ] = sObjectID;
		
		// set callback
		m_rFacade.setCallback< Measurement::Position >( sComponentName, 
			boost::bind( &TrackedObject::updatePosition, &m_candidateObjects[ sObjectID ], _1 ) );
			
		LOG4CPP_DEBUG( logger, "Added candidate object " << sObjectID << ", pattern " << sComponentName );
	}
	else;
		// other dataflow component
}


void IdentityMapper::notifyDeleteComponent( const std::string& sPatternName, const std::string& sComponentName )
{
	boost::mutex::scoped_lock l( m_objectMutex );

	LOG4CPP_TRACE( logger, "notifyDeeteComponent( " << sComponentName << " )" );

	PatternToObjMap::iterator itP2M = m_patternToObjMap.find( sComponentName );
	if ( itP2M != m_patternToObjMap.end() )
	{
		LOG4CPP_TRACE( logger, "Deleting pattern " << sComponentName );

		std::string sObjectID = itP2M->second;
		m_patternToObjMap.erase( itP2M );
		
		TrackedObjectsMap::iterator it;

		it = m_anonymousObjects.find( sObjectID );
		if ( it != m_anonymousObjects.end() )
		{
			// anonymous object removed
			LOG4CPP_DEBUG( logger, "Deleting anonymous object " << sObjectID << ", pattern " << sComponentName );

			// remove object from list
			it->second.nUsageCounter--;
			return;
		}
		
		it = m_candidateObjects.find( sObjectID );
		if ( it != m_candidateObjects.end() )
		{
			// candidate object removed - >remove from list
			LOG4CPP_DEBUG( logger, "Deleting candidate object " << sObjectID << ", pattern " << sComponentName );

			it->second.nUsageCounter--;
		}
	}
}


int main( int ac, char** av )
{
	std::string sServerAddress( "127.0.0.1" );
	std::string sComponentsPath;

	try
	{
		// describe program options
		namespace po = boost::program_options;
		po::options_description poDesc( "Allowed options", 80 );
		poDesc.add_options()
			( "help", "print this help message" )
			( "server,s", po::value< std::string >( &sServerAddress ), 
				"Ubitrack server address <host>[:<port>] to connect to. Defaults to 127.0.0.1." )
			( "components_path", po::value< std::string >( &sComponentsPath ), "Directory from which to load components" )
		;
		
		// parse options from command line and environment
		po::variables_map poOptions;
		po::store( po::command_line_parser( ac, av ).options( poDesc ).run(), poOptions );
		po::store( po::parse_environment( poDesc, "UBITRACK_" ), poOptions );
		po::notify( poOptions );
		
		// print help message if nothing specified
		if ( poOptions.count( "help" ) )
		{
			std::cout << "Syntax: utIdentityMapper [options]" << std::endl << std::endl;
			std::cout << poDesc << std::endl;
			return 1;
		}
	}
	catch( std::exception& e )
	{
		std::cerr << "Error parsing command line parameters : " << e.what() << std::endl;
		std::cerr << "Try utIdentityMapper --help for help" << std::endl;
		return 1;
	}		

	try
	{
		// initialize logging
		Util::initLogging();
		
		std::cout << "Loading components..." << std::endl << std::flush;
		Facade::AdvancedFacade utFacade( sComponentsPath );

		IdentityMapper theBeast( utFacade );

		std::cout << "Connecting to server at " << sServerAddress << "..." << std::endl << std::flush;
		utFacade.connectToServer( sServerAddress );

		std::cout << "Sending pattern file \"IMPatterns.utql\"..." << std::endl << std::flush;
		utFacade.sendUtqlToServer( "IMPatterns.utql" );
		
		std::cout << "Sending query file \"IMQuery.utql\"..." << std::endl << std::flush;
		utFacade.sendUtqlToServer( "IMQuery.utql" );

		std::cout << "Starting dataflow" << std::endl;
		utFacade.startDataflow();

		std::cout << "Press CTRL-C to stop..." << std::endl << std::flush;
		
		// we don't care much about clean stopping...
		while ( true )
		{
			Util::sleep( 1000 );
			theBeast.check();
		}

		utFacade.stopDataflow();

		std::cout << "Finished, cleaning up..." << std::endl << std::flush;
	}
	catch( Util::Exception& e )
	{
		std::cerr << e << std::endl;
	}
}
