/*
 * Ubitrack - Library for Ubiquitous Tracking
 * Copyright 2006, Technische Universitaet Muenchen, and individual
 * contributors as indicated by the @authors tag. See the
 * copyright.txt in the distribution for a full listing of individual
 * contributors.
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA, or see the FSF site: http://www.fsf.org.
 */


#include <assert.h>
#include <Ubitrack/Util/Exception.h>
#include <Ubitrack/Util/Logging.h>

#include "ConvexHull.h"

using namespace Ubitrack;
namespace ublas = boost::numeric::ublas;

ConvexHull::Plane::Plane( const Math::Vector< 3 >& p1, const Math::Vector< 3 >& p2, const Math::Vector< 3 >& p3 )
{
	m_point = p1;
	m_normal = Math::cross_prod( p2 - p1, p3 - p1 );
	
	// normalize normal
	double norm = ublas::norm_2( m_normal );
	if ( fabs( norm ) >= 1e-10 )
		m_normal /= norm;
}


ConvexHull::ConvexHull( const TiXmlNode* pXml )
{
	// 1. read points
	std::vector< Math::Vector< 3 > > points;
	
	const TiXmlElement* pXmlPoint = pXml->FirstChildElement( "Point" );
	while ( pXmlPoint )
	{
		Math::Vector< 3 > pt;
		std::istringstream stream( pXmlPoint->Attribute( "position" ) );
		for ( unsigned i = 0; i < 3; i++ )
			stream >> pt[ i ]; 
		points.push_back( pt );
		pXmlPoint = pXmlPoint->NextSiblingElement( "Point" );
	}
	
	// construct convex hull
	if ( points.size() < 4 )
		UBITRACK_THROW( "More than four points needed in ConvexHull" );
		
	// try all possible planes through three points and check if they form one of the outside planes of the ch
	for ( unsigned i1 = 0; i1 < points.size() - 2; i1++ )
		for ( unsigned i2 = i1 + 1; i2 < points.size() - 1; i2++ )
			for ( unsigned i3 = i2 + 1; i3 < points.size(); i3++ )
			{
				Plane plane( points[ i1 ], points[ i2 ], points[ i3 ] );
				
				// for each plane, check if all other points are on one side.
				// if yes, this is a boundary plane of the convex hull
				
				// find first other point
				unsigned iTest = 0;
				while ( iTest == i1 || iTest == i2 || iTest == i3 )
					iTest++;
					
				// make sure normal points to the outside
				if ( plane.distance( points[ iTest ] ) > 0 )
					plane.swapNormal();
					
				iTest++;

				// check other points
				for ( ; iTest < points.size(); iTest++ )
					if ( iTest != i1 && iTest != i2 && iTest != i3 && plane.distance( points[ iTest ] ) > 0 )
						break;
						
				// all points on inside
				if ( iTest == points.size() )
					m_planes.push_back( plane );
			}
			
	assert( m_planes.size() >= 4 );
}


double ConvexHull::distance( const Math::Vector< 3 >& point ) const
{
	double d = m_planes[ 0 ].distance( point );
	for ( unsigned i = 1; i < m_planes.size(); i++ )
		d = std::max( d, m_planes[ i ].distance( point ) );
	return d;
}
