/*
 * Ubitrack - Library for Ubiquitous Tracking
 * Copyright 2006, Technische Universitaet Muenchen, and individual
 * contributors as indicated by the @authors tag. See the
 * copyright.txt in the distribution for a full listing of individual
 * contributors.
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA, or see the FSF site: http://www.fsf.org.
 */


#include <Ubitrack/Facade/AdvancedFacade.h>
#include <Ubitrack/Util/Exception.h>
#include <Ubitrack/Util/Logging.h>

#include "ReconfigurationRegion.h"

using namespace Ubitrack;

static log4cpp::Category& logger( log4cpp::Category::getInstance( "Reconfigurator.ReconfigurationRegion" ) );



ReconfigurationRegion::ReconfigurationRegion( const TiXmlElement* pXml, Facade::AdvancedFacade& facade )
	: m_state( undecided )
	, m_lastChangeTime( Measurement::now() )
	, m_rFacade( facade )
{
	// parse timeout and hysteresis
	if ( !pXml->Attribute( "timeout", &m_timeout ) )
		m_timeout = 1000;
	if ( !pXml->Attribute( "hysteresis", &m_hysteresis ) )
		m_hysteresis = 0.25;
	
	LOG4CPP_FATAL( logger, "timeout=" << m_timeout );
	LOG4CPP_FATAL( logger, "hysteresis=" << m_hysteresis );
	
	// parse Receiver
	const TiXmlElement* pReceiverXml = pXml->FirstChildElement( "Receiver" );
	if ( !pReceiverXml )
		UBITRACK_THROW( "No Receiver element specified" );
	const TiXmlElement* pReceiverPatternXml = pReceiverXml->FirstChildElement( "Pattern" );
	if ( !pReceiverPatternXml )
		UBITRACK_THROW( "Receiver must contain a Pattern specification" );
	const char* sReceiverName = pReceiverPatternXml->Attribute( "name" );
	if ( !sReceiverName )
		UBITRACK_THROW( "Receiver Pattern must have a name attribute" );
	m_sReceiverName = sReceiverName;

	{
	std::ostringstream patternStream;
	patternStream << *pReceiverPatternXml;
	m_sReceiverPattern = patternStream.str();
	}
	
	// parse ConvexHull
	const TiXmlElement* pConvexHullXml = pXml->FirstChildElement( "ConvexHull" );
	if ( pConvexHullXml )
		m_pConvexHull.reset( new ConvexHull( pConvexHullXml ) );
		
	// parse EnterAction
	const TiXmlElement* pEnterActionXml = pXml->FirstChildElement( "EnterAction" );
	if ( !pEnterActionXml )
		UBITRACK_THROW( "No EnterAction specified" );

	{
	std::ostringstream patternStream;
	for ( const TiXmlElement* pNode = pEnterActionXml->FirstChildElement( "Pattern" ); pNode; 
		pNode = pNode->NextSiblingElement( "Pattern" ) )
		patternStream << *pNode;
	m_sEnterAction = "<UTQLRequest>" + patternStream.str() + "</UTQLRequest>";
	}
	
	// parse LeaveAction
	const TiXmlElement* pLeaveActionXml = pXml->FirstChildElement( "LeaveAction" );
	if ( !pLeaveActionXml )
		UBITRACK_THROW( "No LeaveAction specified" );

	{
	std::ostringstream patternStream;
	for ( const TiXmlElement* pNode = pLeaveActionXml->FirstChildElement( "Pattern" ); pNode; 
		pNode = pNode->NextSiblingElement( "Pattern" ) )
		patternStream << *pNode;
	m_sLeaveAction = "<UTQLRequest>" + patternStream.str() + "</UTQLRequest>";
	}
	
	// add as observer
	m_rFacade.addDataflowObserver( this );
}


void ReconfigurationRegion::receiveTracking( const Measurement::Position& m )
{
	switch ( m_state )
	{
		case undecided:
			LOG4CPP_INFO( logger, m_sReceiverName << " received first measurement" );
			if ( !m_pConvexHull || m_pConvexHull->distance( *m ) < 0 )
				changeState( inside );
			else
				changeState( outside );
			break;
		
		case inside:
			if ( m_pConvexHull && m_pConvexHull->distance( *m ) > m_hysteresis )
				changeState( outside );
			break;
			
		case outside:
			if ( !m_pConvexHull || m_pConvexHull->distance( *m ) < -m_hysteresis )
				changeState( inside );
			break;
	}
	
	m_lastChangeTime = Measurement::now();
}


void ReconfigurationRegion::check()
{
	if ( ( m_state == inside || m_state == undecided ) && m_timeout && m_lastChangeTime + 1000000L * m_timeout < Measurement::now() )
	{
		LOG4CPP_INFO( logger, m_sReceiverName << ": timeout" );
		changeState( outside );
	}
}


void ReconfigurationRegion::changeState( ReconfigurationRegion::RegionState newState )
{
	if ( newState == m_state )
		return;

	LOG4CPP_INFO( logger, m_sReceiverName << " changing state to " << (newState == inside ? "inside" : "outside" ) );

	m_state = newState;
	switch ( m_state )
	{
		case inside:
			m_rFacade.sendUtqlToServerString( m_sEnterAction );
			break;

		case outside:
			m_rFacade.sendUtqlToServerString( m_sLeaveAction );
			break;
		
		default:
			// nothing
			;
	}
}


void ReconfigurationRegion::notifyAddComponent( const std::string& sPatternName, const std::string& sPatternId, 
	const Graph::UTQLSubgraph& )
{
	// set callback
	if ( sPatternName == m_sReceiverName )
	{
		LOG4CPP_INFO( logger, "Received answer for query " << m_sReceiverName << ", setting callback" );
		m_rFacade.setCallback< Measurement::Position >( sPatternId, boost::bind( &ReconfigurationRegion::receiveTracking, this, _1 ) );
	}
}

void ReconfigurationRegion::notifyDeleteComponent( const std::string&, const std::string& )
{
	// do nothing
}
