/*
 * Ubitrack - Library for Ubiquitous Tracking
 * Copyright 2006, Technische Universitaet Muenchen, and individual
 * contributors as indicated by the @authors tag. See the
 * copyright.txt in the distribution for a full listing of individual
 * contributors.
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA, or see the FSF site: http://www.fsf.org.
 */


#include <stdlib.h>
#include <iostream>

#include <Ubitrack/Facade/AdvancedFacade.h>
#include <Ubitrack/Util/Exception.h>
#include <Ubitrack/Util/Logging.h>
#include <Ubitrack/Util/OS.h>

#include "ReconfigurationRegion.h"

using namespace Ubitrack;


class Reconfigurator
{
public:
	/** constructor */
	Reconfigurator( Ubitrack::Facade::AdvancedFacade& facade, const TiXmlElement* pXml );
	
	/** needs to be called regularly */
	void check();
	
protected:
	/** typedef for list of regions */
	typedef std::list< boost::shared_ptr< ReconfigurationRegion > > RegionList;

	/** list of regions for reconfiguration */
	RegionList m_regions;
};


Reconfigurator::Reconfigurator( Ubitrack::Facade::AdvancedFacade& facade, const TiXmlElement* pXml )
{
	std::string sInitString( "<UTQLRequest>" );
	
	// find server address
	const char* sServerName = pXml->Attribute( "server" );
	if ( !sServerName )
		UBITRACK_THROW( "no server name specified" );
	facade.connectToServer( sServerName );

	// load init part
	const TiXmlElement* pInitXml = pXml->FirstChildElement( "Init" );
	if ( pInitXml )
	{
		std::ostringstream patternStream;
		for ( const TiXmlElement* pNode = pInitXml->FirstChildElement( "Pattern" ); pNode; 
			pNode = pNode->NextSiblingElement( "Pattern" ) )
			patternStream << *pNode;
		sInitString += patternStream.str();
	}
	
	// parse all regions
	for ( const TiXmlElement* pXmlRegion = pXml->FirstChildElement( "Region" );
		pXmlRegion; pXmlRegion = pXmlRegion->NextSiblingElement( "Region" ) )
	{
		m_regions.push_back( boost::shared_ptr< ReconfigurationRegion >( new ReconfigurationRegion( pXmlRegion, facade ) ) );
		sInitString += m_regions.back()->getReceiverPattern();
	}
	
	// send initialization to server
	sInitString += "</UTQLRequest>";
	facade.sendUtqlToServerString( sInitString );
}


void Reconfigurator::check()
{
	for ( RegionList::iterator it = m_regions.begin(); it != m_regions.end(); it++ )
		(*it)->check();
}


int main( int ac, char** av )
{
 	try
	{
		if (ac < 2)
		{
			std::cerr << "Usage: " << av[0] << " filename" << std::endl;
			exit (1);
		}

		// initialize logging
		Util::initLogging();
		
		std::cout << "Loading components..." << std::endl << std::flush;
		const char* sComponentsPath = getenv( "UBITRACK_COMPONENTS_PATH" );
		Facade::AdvancedFacade utFacade( sComponentsPath ? sComponentsPath : std::string() );

		std::cout << "Parsing XML..." << std::endl << std::flush;
		TiXmlDocument doc( av[ 1 ] );
		if ( !doc.LoadFile() )
		{
			std::cout << "Error parsing " << av[ 1 ] << " in line " << doc.ErrorRow() << "/" << doc.ErrorCol() << ": " << std::endl;
			std::cout << doc.ErrorDesc();
			return 0;
		}
		
		std::cout << "Sending requests..." << std::endl << std::flush;
		Reconfigurator theBeast( utFacade, doc.RootElement() );
		
		std::cout << "Starting dataflow" << std::endl;
		utFacade.startDataflow();

		std::cout << "Press CTRL-C to stop..." << std::endl << std::flush;
		
		// we don't care much about clean stopping...
		while ( true )
		{
			Util::sleep( 25 );
			theBeast.check();
		}

		utFacade.stopDataflow();

		std::cout << "Finished, cleaning up..." << std::endl << std::flush;
	}
	catch( Util::Exception& e )
	{
		std::cerr << e << std::endl;
	}
}
