/*
 * Ubitrack - Library for Ubiquitous Tracking
 * Copyright 2006, Technische Universitaet Muenchen, and individual
 * contributors as indicated by the @authors tag. See the 
 * copyright.txt in the distribution for a full listing of individual
 * contributors.
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA, or see the FSF site: http://www.fsf.org.
 */

/**
 * @file
 * Implementation of a TCP socket acceptor for the ubitrack server
 *
 * @author Daniel Pustka <daniel.pustka@in.tum.de>
 */ 

#include <boost/bind.hpp>
 
#include <log4cpp/Category.hh>

#include "TcpAcceptor.h"
#include "UtqlServer.h"

// logger for logging...
static log4cpp::Category& logger( log4cpp::Category::getInstance( "TcpAcceptor" ) );

using Ubitrack::ClientServer::TcpConnection;
using asio::ip::tcp;

TcpAcceptor::TcpAcceptor( int nPort, UtqlServer& rServer )
	: m_ioService()
	, m_acceptor( m_ioService, tcp::endpoint( tcp::v4(), nPort ) )
	, m_rServer( rServer )
{
	// create the initial socket
	handleAccept( asio::error_code() );
	
	// create a new thread for dispatching events
	m_pThread.reset( new boost::thread( boost::bind( &asio::io_service::run, &m_ioService ) ) );
}


TcpAcceptor::~TcpAcceptor()
{
	// TODO: look into the Asio examples on how to do a clean shutdown
	
	// stop Asio acceptor
	m_acceptor.close();
	m_pSpareSocket.reset();
	
	// delete all connections
	for ( ConnectionList::iterator it = m_connections.begin(); it != m_connections.end(); it++ )
		m_rServer.removeClient( *it );
	m_connections.clear();
	
	// wait for dispatch thread
	m_pThread->join();
}


void TcpAcceptor::checkDeadClients()
{
	// test all connections if they still are up
	for ( ConnectionList::iterator it = m_connections.begin(); it != m_connections.end(); )
		if ( (*it)->badConnection() )
		{
			LOG4CPP_INFO( logger, "Removing connection " );
			
			m_rServer.removeClient( *it );
			it = m_connections.erase( it );
		}
		else
			it++;
}


void TcpAcceptor::handleAccept( const asio::error_code error )
{
	if ( error )
	{
		// should not really happen, except maybe when shutdown
		LOG4CPP_NOTICE( logger, "error in handleAccept" );
		return;
	}
	
	if ( m_pSpareSocket )
	{
		LOG4CPP_INFO( logger, "Accepting connection from " << m_pSpareSocket->remote_endpoint().address().to_string() );
	
		// create a new ClientConnection object
		boost::shared_ptr< TcpConnection > pConnection( new TcpConnection( m_pSpareSocket ) );
		m_connections.push_back( pConnection );
		
		// add client to server
		m_rServer.addClient( pConnection );
	}
	
	// create a new spare socket
	m_pSpareSocket.reset( new tcp::socket( m_ioService ) );
	m_acceptor.async_accept( *m_pSpareSocket, boost::bind( &TcpAcceptor::handleAccept, this, _1 ) );
}
