/*
 * Ubitrack - Library for Ubiquitous Tracking
 * Copyright 2006, Technische Universitaet Muenchen, and individual
 * contributors as indicated by the @authors tag. See the
 * copyright.txt in the distribution for a full listing of individual
 * contributors.
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA, or see the FSF site: http://www.fsf.org.
 */

/**
 * @file
 * Implementation of a the UTQL server
 *
 * @author Daniel Pustka <daniel.pustka@in.tum.de>
 */

#include <string>
#include <iostream>
#include <sstream>
#include <boost/bind.hpp>

#include <log4cpp/Category.hh>

#include "UtqlServer.h"

#include <Ubitrack/Graph/UTQLDocument.h>
#include <Ubitrack/Graph/UTQLReader.h>

// logger for logging...
static log4cpp::Category& logger( log4cpp::Category::getInstance( "UtqlServer" ) );


UtqlServer::UtqlServer()
{
	// nothing -- so far
}


void UtqlServer::addClient( UtqlServer::ClientPtr pConnection )
{
	LOG4CPP_INFO( logger, "Adding client " << pConnection->getName() );

	// add a ClientInfo to the client list
	ClientInfo client;
	client.pClient = pConnection;

	m_clients.push_front( client );

	// register callback for queries
	pConnection->setReceiver( boost::bind( &UtqlServer::handleQuery, this, m_clients.begin(), _1 ) );
}


void UtqlServer::removeClient( UtqlServer::ClientPtr pConnection )
{
	LOG4CPP_INFO( logger, "Removing client " << pConnection->getName() );

	LOG4CPP_INFO( logger, "Unregistering announcements form client" );
	m_Server.deregisterClient( pConnection->getName() );

	// find client in list
	for ( ClientIterator itC = m_clients.begin(); itC != m_clients.end(); itC++ )
		if ( itC->pClient == pConnection )
		{
			m_clients.erase( itC );
			break;
		}


	// need to notify all clients again

	boost::shared_ptr< std::map< std::string, std::string > > responses = m_Server.generateResponses();

	std::cerr << "After deregister size: " << responses->size() << std::endl;

	for ( std::map< std::string, std::string >::iterator it = responses->begin();
		  it != responses->end(); ++it )
	{
		LOG4CPP_DEBUG( logger, "------------------------" );
		LOG4CPP_DEBUG( logger, it->first << ": " );
		LOG4CPP_DEBUG( logger, it->second );
	}

	for ( ClientIterator itC = m_clients.begin(); itC != m_clients.end(); itC++ )
	{
		std::string clientName = itC->pClient->getName();

		if ( responses->find( clientName ) != responses->end() )
		{
			itC->pClient->send( (*responses)[ clientName ] );
		}
	}

}


void UtqlServer::handleQuery( UtqlServer::ClientIterator itClient,
	boost::shared_ptr< Ubitrack::ClientServer::ClientServerConnection::BufferType > pData )
{
	LOG4CPP_INFO( logger, "Announcement from " << itClient->pClient->getName() );

	std::string query( pData->begin(), pData->end() );
	std::istringstream iss( query );

	boost::shared_ptr< Ubitrack::Graph::UTQLDocument > doc = Ubitrack::Graph::UTQLReader::processInput(iss);
	for ( Ubitrack::Graph::UTQLDocument::SubgraphList::iterator it = doc->m_Subgraphs.begin();
		  it != doc->m_Subgraphs.end(); ++it )
	{
		LOG4CPP_INFO( logger, "Received: " << (*it)->m_Name );
		m_Server.processAnnouncement( *it, itClient->pClient->getName() );
		LOG4CPP_DEBUG( logger, "Done" );
	}

	boost::shared_ptr< std::map< std::string, std::string > > responses = m_Server.generateResponses();
	for ( std::map< std::string, std::string >::iterator it = responses->begin();
		  it != responses->end(); ++it )
	{
		LOG4CPP_DEBUG( logger, "------------------------" );
		LOG4CPP_DEBUG( logger, it->first << ": " );
		LOG4CPP_DEBUG( logger, it->second );
	}

	for ( ClientIterator itC = m_clients.begin(); itC != m_clients.end(); itC++ )
	{
		std::string clientName = itC->pClient->getName();

		if ( responses->find( clientName ) != responses->end() )
		{
			itC->pClient->send( (*responses)[ clientName ] );
		}
	}

	// std::cout << "Received query: " << std::string( pData->begin(), pData->end() ) << std::endl;


}
