/**
 * Totally undocumented utility to compute the intrinsic and extrinsic camera parameters of multiple cameras
 * given recordings of (more or less) corresponding 2d and 3d measurements.
 */
 
#include <fstream>
#include <string>
#include <boost/numeric/ublas/io.hpp>
#include <boost/archive/text_iarchive.hpp>
#include <Ubitrack/Util/Exception.h>
#include <Ubitrack/Util/Logging.h>
#include <Ubitrack/Calibration/Projection.h>
#include <Ubitrack/Calibration/2D3DPoseEstimation.h>
#include <Ubitrack/Calibration/BundleAdjustment.h>
#include <Ubitrack/Measurement/Measurement.h>
#include <Ubitrack/Calibration/Function/MultiplePointProjection.h>
#include <Ubitrack/Calibration/LensDistortion.h>
#include <Ubitrack/Util/CalibFile.h>

using namespace Ubitrack;
using namespace Ubitrack::Calibration;
namespace ublas = boost::numeric::ublas;
typedef BundleAdjustmentNetwork< double > BAN;

template< class EventType >
void readRecording( std::vector< EventType >& v, const char* filename )
{
	// open file
	std::ifstream file( filename );
	if ( !file.good() )
		UBITRACK_THROW( "Could not open file " + std::string( filename ) );

	boost::archive::text_iarchive archive( file );

	// read contents until end-of-file exception
	try
	{
		while ( true )
		{
			EventType e( boost::shared_ptr< typename EventType::value_type >( new typename EventType::value_type() ) );
			std::string dummy; // for newline character in archive
			archive >> dummy >> e;
			v.push_back( e );
		}
	}
	catch( ... )
	{}
}

void removeOutliers( BAN& baNet, double maxDist )
{
	BAN::PoseList& baImages( baNet.images );
	BAN::IntrinsicList& baIntrinsics( baNet.intrinsics );
	BAN::DistortionList& baDistortions( baNet.distortions );
	BAN::BodyList& baBodies( baNet.bodies );
	//BAN::PoseList baBodyPoses;
		
	// robustify... (ISMAR on-site-hack!)
	unsigned nOrigMeasurements = baNet.bodyPointMeasurements.size();
	for ( std::vector< BAN::BodyPointMeasurement >::iterator it = baNet.bodyPointMeasurements.begin(); 
		it != baNet.bodyPointMeasurements.end(); )
	{
		std::vector< BAN::BodyPointMeasurement >::iterator itNext = it;
		itNext++;
		
		Math::Vector< 3 > camPoint( baImages[ it->iImage ] * baBodies[ 0 ][ it->iPoint ] );
		Math::Vector< 2 > imgPoint = Calibration::projectWithDistortion( camPoint, baDistortions[ it->iCamera ], baIntrinsics[ it->iCamera ] );

		double err = ublas::norm_2( imgPoint - it->measurement );
		if ( err > maxDist )
			baNet.bodyPointMeasurements.erase( it );
		it = itNext;
	}
	std::cout << "Removed " << nOrigMeasurements - baNet.bodyPointMeasurements.size() << " of " << nOrigMeasurements << " measurements" << std::endl;
}

int main( int argc, char* argv[] )
{
	Util::initLogging();
	try
	{
		// configuration
		//#define USE_BUNDLEADJUSTMENT
		// one argument for the program name and one  for the ART3D file -> argc - 2
		const unsigned nCams = argc - 2;
		const unsigned nDivisor = 1;
		
		// read recorded data
		std::vector< Measurement::Position2D >* recording2d = new std::vector< Measurement::Position2D >[ nCams ];
		std::vector< Measurement::Position > recording3d;

		for ( unsigned i = 0; i < nCams; i++ )
			readRecording( recording2d[i], ( argv[ i + 1] ) );
		readRecording( recording3d, argv[ nCams + 1 ] );

		// dlt structures
		std::vector< Math::Vector< 2 > >* dlt2d = new std::vector< Math::Vector< 2 > >[ nCams ];
		std::vector< Math::Vector< 3 > >* dlt3d = new std::vector< Math::Vector< 3 > >[ nCams ];

		// bundle adjustment structure
		BAN::PointsList baPoints;
		BAN::PoseList baImages;
		BAN::IntrinsicList baIntrinsics;
		BAN::DistortionList baDistortions;
		BAN::BodyList baBodies;
		BAN::PoseList baBodyPoses;

		baBodies.push_back( BAN::BodyList::value_type() );
		baBodyPoses.push_back( BAN::PoseList::value_type() );

		BAN baNet( baPoints, baImages, baIntrinsics, baDistortions, baBodies, baBodyPoses );

		// synchronize 2d and 3d measurements
		const unsigned eps = 9000000;

		std::vector< Measurement::Position2D >::iterator* it2d = new std::vector< Measurement::Position2D >::iterator[nCams];
		for ( unsigned i = 0; i < nCams; i++ )
			it2d[ i ] = recording2d[ i ].begin();
		std::vector< Measurement::Position >::iterator it3d = recording3d.begin();

		unsigned iIt = 0;
		
		while ( it3d != recording3d.end() )
		{
			bool* bCam = new bool[ nCams ];
			int nMeasurements = 0;
			for ( unsigned i = 0; i < nCams; i++ )
			{
				while ( it2d[ i ] != recording2d[ i ].end() && it2d[ i ]->time() + eps < it3d->time() )
					it2d[ i ]++;
					
				if ( it2d[ i ] == recording2d[ i ].end() )
					bCam[ i ] = false;
				else
					bCam[ i ] = it2d[ i ]->time() < it3d->time() + eps;
					
				if ( bCam[ i ] )
					nMeasurements++;
			}

#ifdef USE_BUNDLEADJUSTMENT
			if ( nMeasurements >= 2 && !( iIt++ % nDivisor ) )
#else
			if ( !( iIt++ % nDivisor ) )
#endif
			{
				for ( unsigned i = 0; i < nCams; i++ )
					if ( bCam[ i ] )
					{
						dlt2d[i].push_back( **it2d[ i ] );
						dlt3d[i].push_back( **it3d );

#ifdef USE_BUNDLEADJUSTMENT
						baNet.freePointMeasurements.push_back( BAN::FreePointMeasurement( baPoints.size(), i, i, **it2d[ i ] ) );
#else
						baNet.bodyPointMeasurements.push_back( BAN::BodyPointMeasurement( 0, baBodies[ 0 ].size(), 0, i, i, **it2d[ i ] ) );
#endif
					}
					
#ifdef USE_BUNDLEADJUSTMENT
				baPoints.push_back( **it3d );
#else
				baBodies[ 0 ].push_back( **it3d );
#endif
			}

			it3d++;
		}

		// compute projection matrices via dlt
		for ( unsigned i = 0; i < nCams; i++ )
		{
			std::cout << "Calculating matrix " << i + 1 << "...." << std::endl;
			Math::Matrix< 3, 4 > P = Calibration::projectionDLT( dlt3d[ i ], dlt2d[ i ] );
			std::cout << P << std::endl;

			std::cout << "Decomposing matrix...." << std::endl;
			Math::Matrix< 3, 3 > K;
			Math::Matrix< 3, 3 > R;
			Math::Vector< 3 > t;
			Calibration::decomposeProjection( K, R, t, P );
			std::cout << K << std::endl << R << std::endl << t << std::endl << std::endl;

			baImages.push_back( Math::Pose( Math::Quaternion( R ), t ) );
			baIntrinsics.push_back( K );
			baDistortions.push_back( Math::Vector< 4 >( 0, 0, 0, 0 ) );
		}

		removeOutliers( baNet, 20 );

		std::cout << std::endl;
		std::cout << "Bundle adjustment 1..." << std::endl;
		
		// run bundle adjustment
		baNet.bEstimateIntrinsics = false;
		bundleAdjustment( baNet );

		removeOutliers( baNet, 4 );
		std::cout << "Bundle adjustment 2..." << std::endl;
		bundleAdjustment( baNet );
		
		
		// print result
		for ( unsigned i = 0; i < nCams; i++ )
		{
			std::cout << "--- Camera " << i + 1 << " ---" << std::endl;
			std::cout << baImages[ i ] << std::endl << std::endl << baIntrinsics[ i ] << std::endl << baDistortions[ i ] << std::endl;
		}

		// write result to file
		std::cout << "Writing results to file..." << std::endl;
		Measurement::Timestamp calibTime = recording3d.back().time();
		for ( unsigned i = 0; i < nCams; i++ )
		{
			char camNum = '1' + i;
			Util::writeCalibFile( std::string( "ArtCamera" ) + camNum + "Pose.calib", Measurement::Pose( calibTime, baImages[ i ] ) );
			Util::writeCalibFile( std::string( "ArtCamera" ) + camNum + "Intrinsics.calib", Measurement::Matrix3x3( calibTime, baIntrinsics[ i ] ) );
			Util::writeCalibFile( std::string( "ArtCamera" ) + camNum + "Distortion.calib", Measurement::Vector4D( calibTime, baDistortions[ i ] ) );
		}

		// check image distance of some points
		std::cout << std::endl << "Randomly selected points and their reprojection" << std::endl;
		for ( unsigned i = 0; i < 30; i++ )
		{
			unsigned iCam = i % nCams;
			unsigned iPt = ( i * 173 ) % dlt3d[ iCam ].size();
			Math::Vector< 3 > camPoint( baImages[ iCam ] * dlt3d[ iCam ][ iPt ] );
			Math::Vector< 2 > imgPoint = Calibration::projectWithDistortion( camPoint, baDistortions[ iCam ], baIntrinsics[ iCam ] );

			double err = ublas::norm_2( imgPoint - dlt2d[ iCam ][ iPt ] );
			std::cout << dlt3d[ iCam ][ iPt ] << " -> " << imgPoint << ", measured: " << dlt2d[ iCam ][ iPt ] 
				<< ", error: " << err << std::endl;
		}
	}
	catch( std::runtime_error& e )
	{
		std::cerr << e.what() << std::endl;
		return 1;
	}
	return 0;
}
