/*
 * Ubitrack - Library for Ubiquitous Tracking
 * Copyright 2006, Technische Universitaet Muenchen, and individual
 * contributors as indicated by the @authors tag. See the
 * copyright.txt in the distribution for a full listing of individual
 * contributors.
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA, or see the FSF site: http://www.fsf.org.
 */

/**
 * @file
 * Takes a sequence of measurements without motion and computes the real 2D covariance.
 *
 * The input is a sequence of 2D measurements as created by the RecordArt2D3D recorder.
 * @author Tobias Reichl <reichl@in.tum.de>
 */

#include <stdlib.h>
#include <iostream>
#include <fstream>
#include <vector>
#include <boost/program_options.hpp>
#include <boost/archive/text_iarchive.hpp>
#include <boost/numeric/ublas/vector_proxy.hpp>

#include <Ubitrack/Measurement/Measurement.h>
#include <Ubitrack/Util/Exception.h>

using namespace Ubitrack;
namespace ublas = boost::numeric::ublas;


int main( int ac, char** av )
{
 	try
	{
		// program options
		std::string sFile;
		
		try
		{
			// describe program options
			namespace po = boost::program_options;
			po::options_description poDesc( "Allowed options", 80 );
			poDesc.add_options()
				( "help", "print this help message" )
				( "file", po::value< std::string >( &sFile ), "File with a sequence of 2D measurements" )
			;
			
			// specify default options
			po::positional_options_description inputOptions;
			inputOptions.add( "file", 1 );		
			
			// parse options from command line
			po::variables_map poOptions;
			po::store( po::command_line_parser( ac, av ).options( poDesc ).positional( inputOptions ).run(), poOptions );
			po::notify( poOptions );

			// print help message if nothing specified
			if ( poOptions.count( "help" ) || sFile.empty() )
			{
				std::cout << "Syntax: scaleCovariance [--file] <logfile>" << std::endl << std::endl;
				std::cout << poDesc << std::endl;
				return 1;
			}
		}
		catch( std::exception& e )
		{
			std::cerr << "Error parsing command line parameters : " << e.what() << std::endl;
			std::cerr << "Try \"scaleCovariance --help\" for help" << std::endl;
			return 1;
		}		

		// load poses from file
		std::ifstream inFile( sFile.c_str() );
		if ( !inFile.good() )
			UBITRACK_THROW( "Error opening " + sFile );

		boost::archive::text_iarchive archive( inFile );
		
		// read contents until end-of-file exception
		std::vector< Measurement::Position2D > positions;
		try
		{
			while ( true )
			{
				Measurement::Position2D e( boost::shared_ptr< Math::Vector< 2 > >( new Math::Vector< 2 >() ) );
				std::string dummy; // for newline character in archive
				archive >> dummy >> e;
				positions.push_back( e );
			}
		}
		catch( ... )
		{}
		
		std::cout << "Loaded " << positions.size() << " measurements" << std::endl;
		
		// compute average measurement
		Math::Vector< 2 > avgPosition( Math::Vector< 2 >( 0, 0 ) );
		for ( std::vector< Measurement::Position2D >::iterator it = positions.begin(); it != positions.end(); it++ )
		{
			avgPosition += *(*it);
		}
		
		avgPosition /= positions.size();
		
		std::cout << "Average 2D position: " << std::endl;
		std::cout << avgPosition << std::endl;

		// compute real covariance
		Math::Matrix< 2, 2 > realCovariance( ublas::zero_matrix< double >( 2, 2 ) );
		
		for ( std::vector< Measurement::Position2D >::iterator it = positions.begin(); it != positions.end(); it++ )
		{
			Math::Vector< 2 > localError;
			localError = *(*it) - avgPosition;
			realCovariance += ublas::outer_prod( localError, localError );
		}
		
		realCovariance /= positions.size();
		
		std::cout << "Real 2D covariance: " << std::endl;
		std::cout << realCovariance << std::endl;
	}
	catch( Util::Exception& e )
	{
		std::cerr << e << std::endl;
	}
}
