/*
 * Ubitrack - Library for Ubiquitous Tracking
 * Copyright 2006, Technische Universitaet Muenchen, and individual
 * contributors as indicated by the @authors tag. See the
 * copyright.txt in the distribution for a full listing of individual
 * contributors.
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA, or see the FSF site: http://www.fsf.org.
 */

/**
 * @file
 * Tunes a kalman filter for best prediction.
 * The input is a list of ErrorPose measurements as created by the Recorder component.
 * @author Daniel Pustka <daniel.pustka@in.tum.de>
 */

#include <stdlib.h>
#include <iostream>
#include <fstream>
#include <vector>
#include <boost/program_options.hpp>
#include <boost/archive/text_iarchive.hpp>
#include <boost/numeric/ublas/vector_proxy.hpp>
#include <boost/numeric/ublas/io.hpp>

#include <Ubitrack/Measurement/Measurement.h>
#include <Ubitrack/Tracking/PoseKalmanFilter.h>
//#include <Ubitrack/Math/LevenbergMarquardt.h>
#include <Ubitrack/Math/DownhillSimplex.h>
#include <Ubitrack/Math/Function/DiscreteJacobianApproximation.h>
#include <Ubitrack/Util/Exception.h>
#include <Ubitrack/Util/Logging.h>

#include <log4cpp/PropertyConfigurator.hh>

using namespace Ubitrack;
namespace ublas = boost::numeric::ublas;

/**
 * a list of measurements
 */
class MeasurementListBase
{
public:
	/** always have a virtual destructor */
	virtual ~MeasurementListBase()
	{}

	/** reset the iterators */
	virtual void reset() = 0;

	/** add the next measurement to a kalman filter, returns the prediction error */
	virtual double updateKF( Tracking::PoseKalmanFilter& kf ) = 0;

	/** returns the time of the next measurement, 0 if no more exist */
	Measurement::Timestamp nextTime() const
	{ return m_nextTime; }

protected:
	Measurement::Timestamp m_nextTime;
};


template< class Type >
class MeasurementList
	: public MeasurementListBase
{
public:
	MeasurementList( const char* filename );

	void reset()
	{ 
		m_it = m_measurements.begin();
		m_nextTime = m_it->time(); 
	}

	virtual double updateKF( Tracking::PoseKalmanFilter& kf );

	std::vector< Type > m_measurements;
	typename std::vector< Type >::const_iterator m_it;
};


template< class Type > 
MeasurementList< Type >::MeasurementList( const char* filename )
{
	// load poses from file
	std::cout << "Loading measurements from file " << std::string( filename ) << "..." << std::endl << std::flush;
	
	std::ifstream inFile( filename );
	if ( !inFile.good() )
		UBITRACK_THROW( "Error opening " + std::string( filename ) );

	boost::archive::text_iarchive archive( inFile );
	
	// read contents until end-of-file exception
	try
	{
		while ( true )
		{
			Type e( boost::shared_ptr< typename Type::value_type >( new typename Type::value_type ) );
			std::string dummy; // for newline character in archive
			archive >> dummy >> e;
			m_measurements.push_back( e );
		}
	}
	catch( ... )
	{}

	std::cout << "Loaded " << m_measurements.size() << " measurements" << std::endl;

	if ( m_measurements.size() < 10 )
		UBITRACK_THROW( "...which is not enough!" );

	reset();
}


template<>
double MeasurementList< Measurement::ErrorPose >::updateKF( Tracking::PoseKalmanFilter& kf )
{
	// compute error
	double posError = 0;
	double rotError = 0;

	if ( m_it != m_measurements.begin() )
	{
		kf.timeUpdate( m_it->time() );

		Math::Vector< 3 > diffPos( (*m_it)->translation() - ublas::subrange( kf.getState(), 0, 3 ) );
		diffPos( 2 ) *= 0.1; // give z less weight (for inside-out)
		posError = ublas::norm_2( diffPos );

		Math::Vector< 4 > rotVec;
		(*m_it)->rotation().toVector( rotVec );
		unsigned rs = ( kf.getMotionModel().posOrder() + 1 ) * 3;
		if ( ublas::inner_prod( rotVec, ublas::subrange( kf.getState(), rs, rs + 4 ) ) < 0 )
			rotError = ublas::norm_2( rotVec + ublas::subrange( kf.getState(), rs, rs + 4 ) );
		else
			rotError = ublas::norm_2( rotVec - ublas::subrange( kf.getState(), rs, rs + 4 ) );
	}

	// integrate measurement
	kf.addPoseMeasurement( *m_it );

	// go to next measurement
	m_it++;
	m_nextTime = m_it == m_measurements.end() ? 0 : m_it->time();

	return 10 * posError + rotError;
}


template<>
double MeasurementList< Measurement::RotationVelocity >::updateKF( Tracking::PoseKalmanFilter& kf )
{
	// integrate measurement
	kf.addInverseRotationVelocityMeasurement( *m_it );

	// go to next measurement
	m_it++;
	m_nextTime = m_it == m_measurements.end() ? 0 : m_it->time();

	return 0; // rotation velocity measurement errors do not count
}


typedef std::vector< boost::shared_ptr< MeasurementListBase > > ManyMeasurements;


/** required to make sure the motion model values don't become negative */
struct MotionModelNormalize
{
	template< class VT1, class VT2 >
	void evaluate( VT1& result, const VT2& ) const
	{
		for ( typename VT1::iterator it = result.begin(); it != result.end(); it++ )
			if ( *it < 0 )
				*it = 1e-9;
	}
};


/**
 * function that computes the prediction error of the kalman filter
 */
class KalmanPredictionErrorFunction
{
public:
	KalmanPredictionErrorFunction( ManyMeasurements& measurements, unsigned posParams, unsigned rotParams, bool bInsideOut )
		: m_measurements( measurements )
		, m_posParams( posParams )
		, m_rotParams( rotParams )
		, m_bInsideOut( bInsideOut )
	{}
	
	unsigned size() const
	{ return 1; }

	template< class VT1, class VT2 >
	void evaluate( VT1& result, const VT2& input ) const
	{
		std::cout << "Evaluating prediction error for " << input << std::endl;
		
		// construct a motion model from the input vector
		Tracking::LinearPoseMotionModel mm( m_posParams - 1, m_rotParams - 1 );
		for ( unsigned i = 0; i < m_posParams; i++ )
			mm.setPosPN( i, input( i ) );
		for ( unsigned i = 0; i < m_rotParams; i++ )
			mm.setOriPN( i, input( m_posParams + i ) );
		Tracking::PoseKalmanFilter kf( mm, m_bInsideOut );

		// reset measurement lists
		for ( ManyMeasurements::iterator it = m_measurements.begin(); it != m_measurements.end(); it++ )
			(*it)->reset();

		// apply measurements in right order
		double totalError = 0;
		while ( true )
		{
			Measurement::Timestamp nextTime = 0;
			MeasurementListBase* pNext = 0;

			for ( ManyMeasurements::iterator it = m_measurements.begin(); it != m_measurements.end(); it++ )
				if ( (*it)->nextTime() && ( !nextTime || (*it)->nextTime() < nextTime ) )
				{
					pNext = it->get();
					nextTime = pNext->nextTime();
				}

			if ( pNext )
				totalError += pNext->updateKF( kf );
			else
				break;
		}

		result( 0 ) = totalError;

		std::cout << "Error = " << totalError << std::endl;
	}
	
protected:
	ManyMeasurements& m_measurements;
	unsigned m_posParams;
	unsigned m_rotParams;
	bool m_bInsideOut;
};


void tuneKalmanFilter( ManyMeasurements& measurements, const ublas::vector< double >& posPN, const ublas::vector< double >& oriPN,
	unsigned nIterations, bool bInsideOut )
{
	ublas::vector< double > motionModel( posPN.size() + oriPN.size() );
	
	// set initial parameters
	ublas::subrange( motionModel, 0, posPN.size() ) = posPN;
	ublas::subrange( motionModel, posPN.size(), posPN.size() + oriPN.size() ) = oriPN;
	
	double res = Math::downhillSimplex( KalmanPredictionErrorFunction( measurements, posPN.size(), oriPN.size(), bInsideOut ), 
		motionModel, ublas::zero_vector< double >( 1 ), Math::OptTerminate( nIterations ), MotionModelNormalize() );
		
	std::cout << "Resulting parameters: " << motionModel << std::endl;
	std::cout << "Residual: " << res << std::endl;
}


int main( int ac, char** av )
{
 	try
	{
		// initialize logging
		Util::initLogging();

		// program options
		std::vector< std::string > poseFiles;
		std::vector< std::string > rvFiles;
		std::string sPosPN( "0.1 1.0" );
		std::string sOriPN( "0.1 1.0" );
		unsigned nIterations( 100 );
		bool bInsideOut;
		bool bInvertPose;

		try
		{
			// describe program options
			namespace po = boost::program_options;
			po::options_description poDesc( "Allowed options", 80 );
			poDesc.add_options()
				( "help", "print this help message" )
				( "iso", "Inside-Out mode" )
				( "invertpose", "Inverts pose measurements" )
				( "pospn", po::value< std::string >( &sPosPN ), "Initial process noise in position" )
				( "oripn", po::value< std::string >( &sOriPN ), "Initial process noise in orientation" )
				( "it", po::value< unsigned >( &nIterations ), "Number of iterations" )
				( "posefile", po::value< std::vector< std::string > >( &poseFiles ), "Recorded ErrorPose sequence." )
				( "rvfile", po::value< std::vector< std::string > >( &rvFiles ), "Recorded RotationVelocity sequence." )
			;
			
			// specify default options
			po::positional_options_description inputOptions;
			inputOptions.add( "posefile", 1 );		
			
			// parse options from command line and environment
			po::variables_map poOptions;
			po::store( po::command_line_parser( ac, av ).options( poDesc ).positional( inputOptions ).run(), poOptions );
			po::notify( poOptions );

			// print help message if nothing specified
			if ( poOptions.count( "help" ) || poseFiles.empty() )
			{
				std::cout << "Syntax: kalmanTuner [options] [--posefile] <pose file>" << std::endl << std::endl;
				std::cout << poDesc << std::endl;
				return 1;
			}

			// boolean options
			bInsideOut = poOptions.count( "iso" ) != 0;
			bInvertPose = poOptions.count( "invertpose" ) != 0;
		}
		catch( std::exception& e )
		{
			std::cerr << "Error parsing command line parameters : " << e.what() << std::endl;
			std::cerr << "Try \"kalmanTuner --help\" for help" << std::endl;
			return 1;
		}		

		// load motion model
		ublas::vector< double > posPN( 20 );
		{
			std::istringstream inStream( sPosPN );
			unsigned nElems = 0;
			double t;
			while ( inStream >> t )
				posPN( nElems++ ) = t;
			posPN.resize( nElems );
		}

		ublas::vector< double > oriPN( 20 );
		{
			std::istringstream inStream( sOriPN );
			unsigned nElems = 0;
			double t;
			while ( inStream >> t )
				oriPN( nElems++ ) = t;
			oriPN.resize( nElems );
		}

		// load poses from file
		ManyMeasurements measurements;

		for ( unsigned i = 0; i < poseFiles.size(); i++ )
		{
			MeasurementList< Measurement::ErrorPose >* list = new MeasurementList< Measurement::ErrorPose >( poseFiles[ i ].c_str() );
			if ( bInvertPose )
				for ( std::vector< Measurement::ErrorPose >::iterator it = list->m_measurements.begin(); 
					it != list->m_measurements.end(); it++ )
					**it = ~**it;
			measurements.push_back( boost::shared_ptr< MeasurementListBase >( list ) );
		}

		for ( unsigned i = 0; i < rvFiles.size(); i++ )
			measurements.push_back( boost::shared_ptr< MeasurementListBase >( 
				new MeasurementList< Measurement::RotationVelocity >( rvFiles[ i ].c_str() ) ) );

		// perform tuning
		tuneKalmanFilter( measurements, posPN, oriPN, nIterations, bInsideOut );
	}
	catch( Util::Exception& e )
	{
		std::cerr << e << std::endl;
	}
}
