/*
 * Ubitrack - Library for Ubiquitous Tracking
 * Copyright 2006, Technische Universitaet Muenchen, and individual
 * contributors as indicated by the @authors tag. See the
 * copyright.txt in the distribution for a full listing of individual
 * contributors.
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA, or see the FSF site: http://www.fsf.org.
 */

/**
 * @file
 * Takes a sequence of measurements without motion, computes the real covariance and
 * compares it with the computed covariance of the measurements. This is used to
 * generate a scale factor for input covariances.
 *
 * The input is a list of ErrorPose measurements as created by the Recorder component.
 * @author Daniel Pustka <daniel.pustka@in.tum.de>
 */

#include <stdlib.h>
#include <iostream>
#include <fstream>
#include <vector>
#include <boost/program_options.hpp>
#include <boost/archive/text_iarchive.hpp>
#include <boost/numeric/ublas/vector_proxy.hpp>

#include <Ubitrack/Measurement/Measurement.h>
#include <Ubitrack/Util/Exception.h>

using namespace Ubitrack;
namespace ublas = boost::numeric::ublas;


int main( int ac, char** av )
{
 	try
	{
		// program options
		std::string sFile;
		
		try
		{
			// describe program options
			namespace po = boost::program_options;
			po::options_description poDesc( "Allowed options", 80 );
			poDesc.add_options()
				( "help", "print this help message" )
				( "file", po::value< std::string >( &sFile ), "File with a sequence of ErrorPose measurements" )
			;
			
			// specify default options
			po::positional_options_description inputOptions;
			inputOptions.add( "file", 1 );		
			
			// parse options from command line
			po::variables_map poOptions;
			po::store( po::command_line_parser( ac, av ).options( poDesc ).positional( inputOptions ).run(), poOptions );
			po::notify( poOptions );

			// print help message if nothing specified
			if ( poOptions.count( "help" ) || sFile.empty() )
			{
				std::cout << "Syntax: scaleCovariance [--file] <logfile>" << std::endl << std::endl;
				std::cout << poDesc << std::endl;
				return 1;
			}
		}
		catch( std::exception& e )
		{
			std::cerr << "Error parsing command line parameters : " << e.what() << std::endl;
			std::cerr << "Try \"scaleCovariance --help\" for help" << std::endl;
			return 1;
		}		

		// load poses from file
		std::ifstream inFile( sFile.c_str() );
		if ( !inFile.good() )
			UBITRACK_THROW( "Error opening " + sFile );

		boost::archive::text_iarchive archive( inFile );
		
		// read contents until end-of-file exception
		std::vector< Measurement::ErrorPose > poses;
		try
		{
			while ( true )
			{
				Measurement::ErrorPose e( boost::shared_ptr< Math::ErrorPose >( new Math::ErrorPose ) );
				std::string dummy; // for newline character in archive
				archive >> dummy >> e;
				poses.push_back( e );
			}
		}
		catch( ... )
		{}

		std::cout << "Loaded " << poses.size() << " measurements" << std::endl;
		
		// compute average measurement and computed covariance
		Math::Vector< 7 > avgPose( ublas::zero_vector< double >( 7 ) );
		Math::Matrix< 6, 6 > avgCovariance( ublas::zero_matrix< double >( 6, 6 ) );
		for ( std::vector< Measurement::ErrorPose >::iterator it = poses.begin(); it != poses.end(); it++ )
		{
			Math::Vector< 7 > vecPose;
			(*it)->toVector( vecPose );
			if ( it != poses.begin() && 
				ublas::inner_prod( ublas::subrange( vecPose, 3, 7 ), ublas::subrange( avgPose, 3, 7 ) ) < 0 )
				ublas::subrange( vecPose, 3, 7 ) *= -1;
			avgPose += vecPose;
			avgCovariance += (*it)->covariance();
		}
		
		avgPose /= poses.size();
		avgCovariance /= poses.size();
		
		std::cout << "Average pose:" << std::endl;
		std::cout << avgPose << std::endl;

		Math::Quaternion avgQuat( Math::Quaternion::fromVector( ublas::subrange( avgPose, 3, 7 ) ) );
		avgQuat.normalize();
		
		// compute real covariance
		Math::Matrix< 6, 6 > realCovariance( ublas::zero_matrix< double >( 6, 6 ) );
		
		for ( std::vector< Measurement::ErrorPose >::iterator it = poses.begin(); it != poses.end(); it++ )
		{
			Math::Vector< 6 > localError;
			ublas::subrange( localError, 0, 3 ) = (*it)->translation() - ublas::subrange( avgPose, 0, 3 );;
			
			Math::Quaternion qDiff( avgQuat * ~(*it)->rotation() );
			localError( 3 ) = qDiff.x();
			localError( 4 ) = qDiff.y();
			localError( 5 ) = qDiff.z();
			if ( qDiff.w() < 0 )
				ublas::subrange( localError, 3, 6 ) *= -1;
				
			realCovariance += ublas::outer_prod( localError, localError );
		}
		
		realCovariance /= poses.size();
		
		// compute scaling
		double scale = ublas::norm_frobenius( realCovariance ) / ublas::norm_frobenius( avgCovariance );
		
		// compute jitter
		Math::Vector< 6 > jitter;
		for ( unsigned i = 0; i < 6; i++ )
			jitter( i ) = sqrt( realCovariance( i, i ) );
		
		// print results
		std::cout << std::endl;
		std::cout << "Average predicted covariance:" << std::endl;
		std::cout << avgCovariance << std::endl;
		std::cout << "Real covariance:" << std::endl;
		std::cout << realCovariance << std::endl;
		std::cout << "Recommended input covariance scaling: " << scale << ", sigma: " << sqrt( scale ) << std::endl;
		std::cout << "Measured jitter: " << jitter << std::endl;
	}
	catch( Util::Exception& e )
	{
		std::cerr << e << std::endl;
	}
}
