/*
 * Ubitrack - Library for Ubiquitous Tracking
 * Copyright 2006, Technische Universitaet Muenchen, and individual
 * contributors as indicated by the @authors tag. See the 
 * copyright.txt in the distribution for a full listing of individual
 * contributors.
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA, or see the FSF site: http://www.fsf.org.
 */

/**
 * @ingroup tracking_algorithms
 * @file
 * Implementats functions for 2D-3D pose estimation.
 *
 * @author Daniel Pustka <daniel.pustka@in.tum.de>
 */
 
#include <iostream>

#include <log4cpp/Category.hh>
#include <boost/numeric/ublas/matrix.hpp>
#include <boost/numeric/ublas/matrix_proxy.hpp>
#include <boost/numeric/ublas/vector_proxy.hpp>

#include <Ubitrack/Math/MatrixOperations.h>
#include <Ubitrack/Math/BackwardPropagation.h>

#include "2D3DPoseEstimation.h"
#include "Function/MultiplePointProjection.h"
#include "Function/MultiplePointProjectionError.h"
#include "Function/MultipleCameraProjectionError.h"
#include "Function/ProjectivePoseNormalize.h"

#ifdef HAVE_LAPACK
#include <boost/numeric/bindings/lapack/gesvd.hpp>
#endif

#define OPTIMIZATION_LOGGING
static log4cpp::Category& optLogger( log4cpp::Category::getInstance( "Ubitrack.Calibration.2D3DPoseEstimation" ) );
#include <Ubitrack/Math/LevenbergMarquardt.h>


// shortcuts to namespaces
namespace ublas = boost::numeric::ublas;
using namespace Ubitrack::Math;

namespace Ubitrack { namespace Calibration {

/** \internal */
template< typename T >
Math::Pose poseFromHomographyImpl( const Matrix< 3, 3, T > H, const Matrix< 3, 3, T > invK )
{
	// compute R = K^-1 H
	Matrix< 3, 3 > R( ublas::prod( invK, H ) );
	
	// make sure the z-coordinate is negative
	if ( R( 2, 2 ) > 0 )
		R *= -1.0;

	// compute length of the first two colums
	double fXLen = ublas::norm_2( ublas::column( R, 0 ) );
	double fYLen = ublas::norm_2( ublas::column( R, 1 ) );
	
	// copy & normalize translation
	double fTransScale = 2.0f / ( fXLen + fYLen );
	Vector< 3 > t( ublas::column( R, 2 ) * fTransScale );
		
#if defined( HAVE_LAPACK ) && !defined( __APPLE__ ) // APPLE vecLib sucks

	// perform svd-based orthogonalization
	Math::Matrix< 3, 3 > u;
	Math::Matrix< 3, 3 > right;
	Math::Vector< 2 > s;
	ublas::matrix_range< Matrix< 3, 3 > > Rleft( R, ublas::range( 0, 3 ), ublas::range( 0, 2 ) );
	ublas::matrix_range< Matrix< 3, 3 > > vt( right, ublas::range( 0, 2 ), ublas::range( 0, 2 ) );
	boost::numeric::bindings::lapack::gesvd( 'A', 'A', Rleft, s, u, vt );

	right( 0, 2 ) = right( 1, 2 ) = 0;
	right( 2, 0 ) = right( 2, 1 ) = 0;
	right( 2, 2 ) = Math::determinant( vt ) * Math::determinant( u ); // should be -1 or +1
	R = ublas::prod( u, right );

#else

	// normalize first two colums
	ublas::column( R, 0 ) /= fXLen;
	ublas::column( R, 1 ) /= fYLen;

	// compute third row as vector product
	ublas::column( R, 2 ) = cross_prod( ublas::column( R, 0 ), ublas::column( R, 1 ) );
	
	// normalize cross product
	double fZLen = ublas::norm_2( ublas::column( R, 2 ) );
	ublas::column( R, 2 ) /= fZLen;
	
	// recompute y vector from x and z
	ublas::column( R, 1 ) = cross_prod( ublas::column( R, 2 ), ublas::column( R, 0 ) );

#endif
	// compute rotation quaternion from matrix
	return Pose( Quaternion( R ), t );
}

Math::Pose poseFromHomography( const Math::Matrix< 3, 3, float > H, const Math::Matrix< 3, 3, float > invK )
{
	return poseFromHomographyImpl( H, invK );
}

Math::Pose poseFromHomography( const Math::Matrix< 3, 3, double > H, const Math::Matrix< 3, 3, double > invK )
{
	return poseFromHomographyImpl( H, invK );
}


#ifdef HAVE_LAPACK

/** \internal */	
template< typename T > 
T optimizePoseImpl( Pose& p, const std::vector< Vector< 2, T > >& p2D, const std::vector< Vector< 3, T > >& p3D, 
	const Matrix< 3, 3, T >& cam, unsigned nIterations  )
{
	// copy rot & trans to parameter vector
	Vector< 7, T > params;
	p.toVector( params );

	// copy 2D points to measurement vector
	ublas::vector< T >  measurements( 2 * p2D.size() );
	for ( unsigned i = 0; i < p2D.size(); i++ )
		ublas::subrange( measurements, 2*i, (i+1)*2 ) = p2D[ i ];

	// perform optimization
	Function::MultiplePointProjection< T > projection( p3D, cam );
	T fRes = levenbergMarquardt( projection, params, measurements, 
		OptTerminate( nIterations, 1e-6 ), Function::ProjectivePoseNormalize() );

	// copy back rot & trans from vector
	p = Pose::fromVector( params );
		
	return fRes;
}

float optimizePose( Math::Pose& p, const std::vector< Math::Vector< 2, float > >& p2D, 
	const std::vector< Math::Vector< 3, float > >& p3D, const Math::Matrix< 3, 3, float >& cam,
	unsigned nIterations )
{
	return optimizePoseImpl( p, p2D, p3D, cam, nIterations );
}

double optimizePose( Math::Pose& p, const std::vector< Math::Vector< 2, double > >& p2D, 
	const std::vector< Math::Vector< 3, double > >& p3D, const Math::Matrix< 3, 3, double >& cam,
	unsigned nIterations )
{
	return optimizePoseImpl( p, p2D, p3D, cam, nIterations );
}


/** \internal */
template< typename T >
Matrix< 6, 6, T > singleCameraPoseErrorImpl( const Pose& p, const std::vector< Vector< 3, T > >& p3D, 
	const Matrix< 3, 3, T >& cam, T imageError )
{
	// copy rot & trans to parameter vector
	Vector< 7, T > params;
	p.toVector( params );

	// calculate error
	Matrix< 6, 6, T > result;
	Function::MultiplePointProjectionError< T > projection( p3D, cam );

	backwardPropagationIdentity( result, imageError, projection, params );
	
	return result;
}

Matrix< 6, 6, float > singleCameraPoseError( const Math::Pose& p, const std::vector< Math::Vector< 3, float > >& p3D, 
	const Math::Matrix< 3, 3, float >& cam, float imageError )
{
	return singleCameraPoseErrorImpl( p, p3D, cam, imageError );
}

Matrix< 6, 6, double > singleCameraPoseError( const Math::Pose& p, const std::vector< Math::Vector< 3, double > >& p3D, 
	const Math::Matrix< 3, 3, double >& cam, double imageError )
{
	return singleCameraPoseErrorImpl( p, p3D, cam, imageError );
}


/** \internal */
template< typename T >
Math::Matrix< 6, 6, T > multipleCameraPoseErrorImpl( const Math::Pose& p, 
	const std::vector< Math::Vector< 3, T > >& p3D, 
	const std::vector< Math::Matrix< 3, 4, T > >& cameras, 
	const std::vector< std::pair< unsigned, unsigned > > observations, 
	T imageError )
{
	// copy rot & trans to parameter vector
	Vector< 7, T > params;
	p.toVector( params );

	// calculate error
	Matrix< 6, 6, T > result;
	Function::MultipleCameraProjectionError< T > projection( p3D, cameras, observations );

	backwardPropagationIdentity( result, imageError, projection, params );
	
	return result;
}
	
Math::Matrix< 6, 6, float > multipleCameraPoseError( const Math::Pose& p, 
	const std::vector< Math::Vector< 3, float > >& p3D, 
	const std::vector< Math::Matrix< 3, 4, float > >& cameras, 
	const std::vector< std::pair< unsigned, unsigned > > observations, 
	float imageError )
{
	return multipleCameraPoseErrorImpl( p, p3D, cameras, observations, imageError );
}
	
UBITRACK_EXPORT Math::Matrix< 6, 6, double > multipleCameraPoseError( const Math::Pose& p, 
	const std::vector< Math::Vector< 3, double > >& p3D, 
	const std::vector< Math::Matrix< 3, 4, double > >& cameras, 
	const std::vector< std::pair< unsigned, unsigned > > observations, 
	double imageError )
{
	return multipleCameraPoseErrorImpl( p, p3D, cameras, observations, imageError );
}

#endif // HAVE_LAPACK

} } // namespace Ubitrack::Calibration
