/*
 * Ubitrack - Library for Ubiquitous Tracking
 * Copyright 2006, Technische Universitaet Muenchen, and individual
 * contributors as indicated by the @authors tag. See the 
 * copyright.txt in the distribution for a full listing of individual
 * contributors.
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA, or see the FSF site: http://www.fsf.org.
 */

/**
 * @ingroup tracking_algorithms
 * @file
 * Implements  functions for 3D Point Reconstruction
 * @author Daniel Muhra <muhra@in.tum.de>
 */ 

#include "3DPointReconstruction.h"

#include <log4cpp/Category.hh>
#include <Ubitrack/Util/Exception.h>
#include <Ubitrack/Util/Logging.h>
#include <iostream>

namespace ublas = boost::numeric::ublas;

#ifdef HAVE_LAPACK
#include <boost/numeric/bindings/lapack/gesvd.hpp>
#include <Ubitrack/Calibration/Munkres.h>
namespace lapack = boost::numeric::bindings::lapack;
#endif

namespace Ubitrack { namespace Calibration {

/** internal of pointToPointDist function */
template< typename T >
T pointToPointDistImp( const Math::Vector< 3, T > & from, const Math::Vector< 3, T > & to, const Math::Matrix< 3, 3, T > & fM  )
{
	Math::Vector< 3, T > from_  = ublas::prod( fM, from );

	T term = from_( 0 ) * to( 0 ) + from_( 1 )* to( 1 ) + from_( 2 );
	return ( term * term ) / ( from_( 0 ) * from_( 0 ) + from_( 1 )* from_( 1 ) );
}

float pointToPointDist( const Math::Vector< 2, float > & from, const Math::Vector< 2, float > & to, const Math::Matrix< 3, 3, float > & fM  )
{
	const Math::Vector< 3, float > from_( from( 0 ), from( 1 ), 1.0f );
	const Math::Vector< 3, float > to_( to( 0 ), to( 1 ), 1.0f );

	return pointToPointDistImp( from_, to_, fM );
}

double pointToPointDist( const Math::Vector< 2, double > & from, const Math::Vector< 2, double > & to, const Math::Matrix< 3, 3, double > & fM  )
{
	const Math::Vector< 3, double > from_( from( 0 ), from( 1 ), 1.0 );
	const Math::Vector< 3, double > to_( to( 0 ), to( 1 ), 1.0 );

	return pointToPointDistImp( from_, to_, fM );
}

float pointToPointDist( const Math::Vector< 3, float > & from, const Math::Vector< 3, float > & to, const Math::Matrix< 3, 3, float > & fM  )
{
	return pointToPointDistImp( from, to, fM );
}

double pointToPointDist( const Math::Vector< 3, double > & from, const Math::Vector< 3, double > & to, const Math::Matrix< 3, 3, double > & fM  )
{
	return pointToPointDistImp( from, to, fM );
}

/** internal of get3DPostion function */
#ifdef HAVE_LAPACK
template< typename T >
Math::Vector< 3, T > get3DPositionImp( const Math::Matrix< 3, 4, T > & P1, const Math::Matrix< 3, 4, T > & P2, const Math::Vector< 2, T > & x, const Math::Vector< 2, T > & x_ )
{
	Math::Matrix< 4, 4, T > A;
	ublas::row( A, 0 ) = ( x( 0 ) * ublas::row( P1, 2 ) ) - ublas::row( P1, 0 );
	ublas::row( A, 1 ) = ( x( 1 ) * ublas::row( P1, 2 ) ) - ublas::row( P1, 1 );
	ublas::row( A, 2 ) = ( x_( 0 ) * ublas::row( P2, 2 ) ) - ublas::row( P2, 0 );
	ublas::row( A, 3 ) = ( x_( 1 ) * ublas::row( P2, 2 ) ) - ublas::row( P2, 1 );

	//solving using svd
	ublas::vector< T > s1( 4 );
	Math::Matrix< 4, 4, T > Vt;
	Math::Matrix< 4, 4, T > U;
	lapack::gesvd( 'N', 'A', A, s1, U, Vt );

	return ( Math::Vector< 3, T >( Vt( 3, 0 ) / Vt( 3, 3), Vt( 3, 1 ) / Vt( 3, 3), Vt( 3, 2 ) / Vt( 3, 3) ) );
}

Math::Vector< 3, float > get3DPosition( const Math::Matrix< 3, 4, float > & P1, const Math::Matrix< 3, 4, float > & P2, const Math::Vector< 2, float > & x, const Math::Vector< 2, float > & x_ )
{
	return get3DPositionImp( P1, P2, x, x_ );
}

Math::Vector< 3, double > get3DPosition( const Math::Matrix< 3, 4, double > & P1, const Math::Matrix< 3, 4, double > & P2, const Math::Vector< 2, double > & x, const Math::Vector< 2, double > & x_ )
{
	return get3DPositionImp( P1, P2, x, x_ );
}

/** insternal of reconstruct3DPoints function */
template< typename T >
std::vector< Math::Vector< 3, T > > reconstruct3DPointsImp( const std::vector< Math::Vector< 2, T > > & p1, const std::vector< Math::Vector< 2, T > > & p2,
																			const Math::Matrix< 3, 4, T > & P1, const Math::Matrix< 3, 4, T > & P2, const Math::Matrix< 3, 3, T > & fM )
{
	
	unsigned int p1Size = p1.size();
	unsigned int p2Size = p2.size();

	//create match matrix
	ublas::matrix< T > matrix( p1Size, p2Size );

	for( unsigned int row=0; row < p1Size; row++ )
	{
		for( unsigned int col=0; col < p2Size; col++ )
		{
			matrix( row, col ) = pointToPointDist( p1.at( row ), p2.at( col ), fM );
		}
	}

	Munkres< T > m( matrix );
	m.solve();
	std::vector< unsigned int > matchList = m.getRowMatchList();

	std::vector< Math::Vector< 3, T > > list;

	for(unsigned int i=0; i < p1Size; i++ )
	{
		if( matchList.at( i ) < p2Size )
		{
			list.push_back( get3DPosition( P1, P2, p1.at( i ), p2.at( matchList.at( i ) ) ) );
		}
	}

	return list;
}

std::vector< Math::Vector< 3, float > > reconstruct3DPoints( const std::vector< Math::Vector< 2, float > > & p1, const std::vector< Math::Vector< 2, float > > & p2,
																			const Math::Matrix< 3, 4, float > & P1, const Math::Matrix< 3, 4, float > & P2, const Math::Matrix< 3, 3, float > & fM )
{
	return reconstruct3DPointsImp( p1, p2, P1, P2, fM );
}

std::vector< Math::Vector< 3, double > > reconstruct3DPoints( const std::vector< Math::Vector< 2, double > > & p1, const std::vector< Math::Vector< 2, double > > & p2,
																			const Math::Matrix< 3, 4, double > & P1, const Math::Matrix< 3, 4, double > & P2, const Math::Matrix< 3, 3, double > & fM )
{
	return reconstruct3DPointsImp( p1, p2, P1, P2, fM );
}
#endif

} } // namespace Ubitrack::Calibration
