/*
 * Ubitrack - Library for Ubiquitous Tracking
 * Copyright 2006, Technische Universitaet Muenchen, and individual
 * contributors as indicated by the @authors tag. See the 
 * copyright.txt in the distribution for a full listing of individual
 * contributors.
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA, or see the FSF site: http://www.fsf.org.
 */

/**
 * @ingroup tracking_algorithms
 * @file
 * Functions for homography estimation.
 *
 * @author Daniel Pustka <daniel.pustka@in.tum.de>
 */ 

#ifndef __UBITRACK_CALIBRATION_HOMOGRAPHY_H_INCLUDED__
#define __UBITRACK_CALIBRATION_HOMOGRAPHY_H_INCLUDED__

#include <Ubitrack/config.h>

#ifdef HAVE_LAPACK

#include <Ubitrack/Ubitrack.h>
#include <Ubitrack/Math/Matrix.h>
#include <Ubitrack/Math/Vector.h>
#include <vector>

namespace Ubitrack { namespace Calibration {


/**
 * @ingroup tracking_algorithms
 * Computes a general homography using a linear DLT method.
 *
 * The result is a homography H that maps points x to points x' via x'=Hx.
 * See Hartley&Zisserman for details.
 *
 * Note: also exists with \c double parameters.
 *
 * @param fromPoints Points x as inhomogeneous 2-vectors
 * @param toPoints Points x' as inhomogeneous 2-vectors
 * @return calculated homography
 */
UBITRACK_EXPORT Math::Matrix< 3, 3, float > homographyDLT( const std::vector< Math::Vector< 2, float > >& fromPoints, 
	const std::vector< Math::Vector< 2, float > >& toPoints );

UBITRACK_EXPORT Math::Matrix< 3, 3, double > homographyDLT( const std::vector< Math::Vector< 2, double > >& fromPoints, 
	const std::vector< Math::Vector< 2, double > >& toPoints );
	

/**
 * @ingroup tracking_algorithms
 * Computes a homography for a square.
 *
 * The result is a homography H that maps vectors x = [(-0.5, +0.5), (-0.5, -0.5), (+0.5, -0.5), (+0.5, +0.5)] 
 * to the points x' given in the corners argument via x' = H x. The homography is computed
 * using the Harker & O'Leary method.
 * 
 * Note: also exists with \c double parameters.
 *
 * @param corners list of four 2D points x'
 * @return the homography
 */
UBITRACK_EXPORT Math::Matrix< 3, 3, float > squareHomography( const std::vector< Math::Vector< 2, float > >& corners );

UBITRACK_EXPORT Math::Matrix< 3, 3, double > squareHomography( const std::vector< Math::Vector< 2, double > >& corners );


/**
 * Computes the normalization of a set of vectors required for DLT.
 * Mainly to be used internally by homography and projection matrix estimation.
 * The points have to be transformed according to p'=(p-shift)/scale
 *
 * @param shift returns the mean value of the vectors.
 * @param scale returns the non-isotropic extension of the vectors around the mean
 * @param modMatrix returns a matrix used to modify the computed transformation to correct the normalization
 * @param modInverse if true, the inverse matrix is returned
 * @param p the set of vectors to normalize
 */
template< int N, typename T > void dltNormalizeVectors( Math::Vector< N, T >& shift, Math::Vector< N, T >& scale,
	Math::Matrix< N+1, N+1, T >& modMatrix, bool modInverse, const std::vector< Math::Vector< N, T > >& p )
{
	namespace ublas = boost::numeric::ublas;

	// compute mean and mean of square
	shift = ublas::zero_vector< T >( N );
	scale = ublas::zero_vector< T >( N );
	for ( unsigned i = 0; i < p.size(); i++ )
	{
		shift = shift + p[ i ];
		scale = scale + ublas::element_prod( p[ i ], p[ i ] );
	}
	shift *= T( 1 ) / p.size();
	scale *= T( 1 ) / p.size();
	
	// compute standard deviation
	for ( unsigned i = 0; i < N; i++ )
		scale( i ) = std::sqrt( scale( i ) - ( shift( i ) * shift( i ) ) );
		
	// compute correction matrix
	modMatrix = ublas::zero_matrix< T >( N+1, N+1 );
	modMatrix( N, N ) = T( 1 );
	if ( modInverse )
		for ( unsigned i = 0; i < N; i++ )
		{
			modMatrix( i, i ) = scale( i );
			modMatrix( i, N ) = shift( i );
		}
	else
		for ( unsigned i = 0; i < N; i++ )
		{
			modMatrix( i, i ) = T( 1 ) / scale( i );
			modMatrix( i, N ) = -modMatrix( i, i ) * shift( i );
		}
}

} } // namespace Ubitrack::Calibration

#endif // HAVE_LAPACK

#endif
