/*
 * Ubitrack - Library for Ubiquitous Tracking
 * Copyright 2006, Technische Universitaet Muenchen, and individual
 * contributors as indicated by the @authors tag. See the
 * copyright.txt in the distribution for a full listing of individual
 * contributors.
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA, or see the FSF site: http://www.fsf.org.
 */


/**
 * @ingroup dataflow_components
 * @file
 * 2D-3D pose estimation component.
 *
 * @author Daniel Pustka <daniel.pustka@in.tum.de>
 */
#include <log4cpp/Category.hh>

#include <Ubitrack/Calibration/Homography.h>
#include <Ubitrack/Calibration/Projection.h>
#include <Ubitrack/Math/MatrixOperations.h>
#include <Ubitrack/Calibration/2D3DPoseEstimation.h>
#include <Ubitrack/Dataflow/TriggerComponent.h>
#include <Ubitrack/Dataflow/TriggerInPort.h>
#include <Ubitrack/Dataflow/ExpansionInPort.h>
#include <Ubitrack/Dataflow/TriggerOutPort.h>
#include <Ubitrack/Dataflow/ComponentFactory.h>
#include <Ubitrack/Measurement/Measurement.h>

// get a logger
static log4cpp::Category& logger( log4cpp::Category::getInstance( "Ubitrack.Events.Components.2D3DPoseEstimation" ) );

namespace Ubitrack { namespace Components {

/**
 * @ingroup dataflow_components
 * 2D-3D pose estimation component.
 * Given pairs of corresponding 2D and 3D points, the component computes the pose.
 *
 * @par Operation
 * The component computes the 6D pose from corresponding 2d and 3d points. It first computes
 * a rough initial pose which is refined by nonlinear optimization.
 *
 * Currently, the component has one drawback: for the first step (the initialization), the 
 * component assumes that the first 4 3D points lie on a plane, which in general is NOT the 
 * case! It is however for square markers...
 */
class PoseEstimation2D3D
	: public Dataflow::TriggerComponent
{
public:
	/**
	 * UTQL component constructor.
	 *
	 * @param sName Unique name of the component.
	 * @param subgraph UTQL subgraph
	 */
	PoseEstimation2D3D( const std::string& sName, boost::shared_ptr< Graph::UTQLSubgraph > config )
		: Dataflow::TriggerComponent( sName, config )
		, m_in2d( "Input2d", *this )
		, m_in3d( "Input3d", *this )
		, m_inCam( "Intrinsics", *this )
		, m_outPort( "Output", *this )
    {
		generateSpaceExpansionPorts( config );
    }

	/** Method that computes the result. */
	void compute( Measurement::Timestamp ts )
	{
		namespace ublas = boost::numeric::ublas;

		const std::vector< Math::Vector< 2 > >& p2d( *m_in2d.get() );
		const std::vector< Math::Vector< 3 > >& p3d( *m_in3d.get() );

		if ( p2d.size() < 4 )
			UBITRACK_THROW( "2D3D pose estimation requires at least 4 points" );

		LOG4CPP_DEBUG( logger, "Performing pose estimation using " << p2d.size() << " points" );

		// invert camera matrix
		Math::Matrix< 3, 3 > invK( Math::invert_matrix( *m_inCam.get() ) );

		// compute initial pose
		Math::Pose initialPose;

		bool bInitialized = false;
		if ( p2d.size() >= 6 )
		{
			// initialize from 3x4 projection matrix
			Math::Matrix< 3, 4 > P( Calibration::projectionDLT( p3d, p2d ) );
			Math::Matrix< 3, 4 > Rt( ublas::prod( invK, P ) );
			LOG4CPP_TRACE( logger, "inital [R|t]: " << std::endl << Rt );

			// perform svd decomposition to get a pure rotation matrix
			Math::Matrix< 3, 3 > u;
			Math::Matrix< 3, 3 > vt;
			Math::Vector< 3 > s;
			ublas::matrix_range< Math::Matrix< 3, 4 > > R( Rt, ublas::range( 0, 3 ), ublas::range( 0, 3 ) );
			ublas::matrix_column< Math::Matrix< 3, 4 > > t( Rt, 3 );

			if ( Math::determinant( R ) < 0 )
				Rt *= -1;

			boost::numeric::bindings::lapack::gesvd( 'A', 'A', R, s, u, vt );
			LOG4CPP_TRACE( logger, "s: " << s );
			LOG4CPP_TRACE( logger, "U: " << std::endl << u );
			LOG4CPP_TRACE( logger, "V^T: " << std::endl << vt );

			if ( s( 0 ) > 2 * s( 3 ) ) // compute condition number
			{
				R = ublas::prod( u, vt );
				t /= s( 0 ) * s( 1 ) * s( 2 ); // det( original-R )

				initialPose = Math::Pose( Math::Quaternion( R ), t );
				bInitialized = true;

				LOG4CPP_TRACE( logger, "Pose from projection matrix: " << initialPose );
			}
			else
				LOG4CPP_DEBUG( logger, "3x4 DLT was unstable (planar target?)" );
		}

		if ( !bInitialized )
			if ( p3d[ 0 ]( 2 ) == 0 && p3d[ 1 ]( 2 ) == 0 && p3d[ 2 ]( 2 ) == 0 && p3d[ 3 ]( 2 ) == 0 )
			{
				// markers already have z=0
				std::vector< Math::Vector< 2 > > p3dAs2d;
				for ( unsigned i = 0; i < 4; i++ )
					p3dAs2d.push_back( Math::Vector< 2 >( p3d[ i ]( 0 ), p3d[ i ]( 1 ) ) );

				// compute homography
				Math::Matrix< 3, 3 > H;
				if ( p2d.size() > 4 )
					// copy first four elements to new vector
					H =  Calibration::homographyDLT( p3dAs2d, std::vector< Math::Vector< 2 > >( p2d.begin(), p2d.begin() + 4 ) );
				else
					H =  Calibration::homographyDLT( p3dAs2d, p2d );
				LOG4CPP_TRACE( logger, "Homography: " << H );

				// compute initial pose from homography
				initialPose = Calibration::poseFromHomography( H, invK );
				LOG4CPP_TRACE( logger, "Pose from homography: " << initialPose );
			}
			else
			{
				// compute a rotation matrix that will bring the points into a plane with equal z
				Math::Vector< 3 > vX( p3d[ 1 ] - p3d[ 0 ] );
				double f = ublas::norm_2( vX );
				vX /= f;

				Math::Vector< 3 > vZ( Math::cross_prod( vX, p3d[ 2 ] - p3d[ 0 ] ) );
				f = ublas::norm_2( vZ );
				vZ /= f;

				Math::Matrix< 3, 3 > P;
				ublas::row( P, 0 ) = vX;
				ublas::row( P, 2 ) = vZ;
				ublas::row( P, 1 ) = Math::cross_prod( vZ, vX );

				// compute a translation
				Math::Vector< 3 > t( -ublas::prod( P, p3d[ 0 ] ) );

				std::vector< Math::Vector< 2 > > p3dAs2d;
				for ( unsigned i = 0; i < 4; i++ )
				{
					Math::Vector< 3 > p3dtrans = ublas::prod( P, p3d[ i ] ) + t;
					if ( fabs( p3dtrans( 2 ) ) > 1e-5 )
						UBITRACK_THROW( "Pose estimation requires four coplanar points" ); 
					p3dAs2d.push_back( Math::Vector< 2 >( p3dtrans( 0 ), p3dtrans( 1 ) ) );
				}

				// compute homography
				Math::Matrix< 3, 3 > H;
				if ( p2d.size() > 4 )
					// copy first four elements to new vector
					H =  Calibration::homographyDLT( p3dAs2d, std::vector< Math::Vector< 2 > >( p2d.begin(), p2d.begin() + 4 ) );
				else
					H =  Calibration::homographyDLT( p3dAs2d, p2d );

				// compute initial pose from homography
				initialPose = Calibration::poseFromHomography( H, invK ) * Math::Pose( Math::Quaternion( P ), t );
				LOG4CPP_TRACE( logger, "Pose from homography (rotated): " << initialPose );
			}
		
		// non-linear minimization
		Calibration::optimizePose( initialPose, p2d, p3d, *m_inCam.get() );
		LOG4CPP_DEBUG( logger, "Refined pose: " << initialPose );

		m_outPort.send( Measurement::Pose( ts, initialPose ) );
		
		m_lastPoseTime = ts;
		m_lastPose = initialPose;
    }

protected:
	/** time of last pose computation */
	Measurement::Timestamp m_lastPoseTime;
	
	/** previous pose */
	Math::Pose m_lastPose;
	
	/** Input port Input2d of the component. */
	Dataflow::ExpansionInPort< Math::Vector< 2 > > m_in2d;

	/** Input port Input3d of the component. */
	Dataflow::ExpansionInPort< Math::Vector< 3 > > m_in3d;

	/** Input port Intrinsics of the component. */
	Dataflow::TriggerInPort< Measurement::Matrix3x3 > m_inCam;

	/** Output port of the component. */
	Dataflow::TriggerOutPort< Measurement::Pose > m_outPort;
};


UBITRACK_REGISTER_COMPONENT( Dataflow::ComponentFactory* const cf ) {
	cf->registerComponent< PoseEstimation2D3D > ( "2D3DPoseEstimation" );
}

} } // namespace Ubitrack::Components
