/*
 * Ubitrack - Library for Ubiquitous Tracking
 * Copyright 2006, Technische Universitaet Muenchen, and individual
 * contributors as indicated by the @authors tag. See the
 * copyright.txt in the distribution for a full listing of individual
 * contributors.
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA, or see the FSF site: http://www.fsf.org.
 */


/**
 * @ingroup dataflow_components
 *
 * @file
 * Components that gets a button pushed. The timestampe is then used
 * for a pull-push conversion
 *
 * @author Michael Schlegel <schlegem@in.tum.de>
 */

#include <log4cpp/Category.hh>

#include <Ubitrack/Dataflow/Component.h>
#include <Ubitrack/Dataflow/PushConsumer.h>
#include <Ubitrack/Dataflow/PullConsumer.h>
#include <Ubitrack/Dataflow/PushSupplier.h>
#include <Ubitrack/Dataflow/ComponentFactory.h>
#include <Ubitrack/Measurement/Measurement.h>


namespace Ubitrack { namespace Components {

  /**
   * @ingroup dataflow_components This component does a push pull
   * conversion. If a butten event is arriving a measurement will be
   * pulled. The measurement is then sent via push
   *
   * @par Input Ports
   * PullConsumer<Measurement> with name "Input".
   * PushConsumer<Measurement::Button> with name "TriggerButton".
   *
   * @par Output Ports
   * PushSupplier<Measurement> with name "Output".
   *
   * @par Configuration
   * none
   *
   * @par Operation
   *
   * @par Instances
   */
  template< class EventType >
  class ButtonTrigger
    : public Dataflow::Component
  {
  public:
    /**
     * UTQL component constructor.
     *
     * @param sName Unique name of the component.
     * @param subgraph UTQL subgraph
     */
    ButtonTrigger( const std::string& sName, boost::shared_ptr< Graph::UTQLSubgraph > subgraph  )
      : Dataflow::Component( sName )
      , m_inPort("Input", *this)
      , m_inTriggerPort( "Trigger", *this, boost::bind( &ButtonTrigger::receiveButton, this, _1 ) )
      , m_outPort( "Output", *this )
      , m_logger( log4cpp::Category::getInstance( "Ubitrack.Components.ButtonTrigger" ) )
    {
    }

    /** Method that computes the result. */
    void receiveButton( const Measurement::Button& event )
    {
      m_outPort.send(m_inPort.get( event.time() ));
    }

  protected:
     
    /** Input port of the component. */
    Dataflow::PullConsumer< EventType > m_inPort;
    Dataflow::PushConsumer< Measurement::Button > m_inTriggerPort;

    // Output port of the component
    Dataflow::PushSupplier< EventType > m_outPort;

    /** log4cpp logger reference */
    log4cpp::Category& m_logger;	
  };


  UBITRACK_REGISTER_COMPONENT( Dataflow::ComponentFactory* const cf ) {
    cf->registerComponent< ButtonTrigger< Measurement::Rotation > > ( "ButtonTriggerRotation" );
    cf->registerComponent< ButtonTrigger< Measurement::RotationVelocity > > ( "ButtonTriggerRotationVelocity" );
    cf->registerComponent< ButtonTrigger< Measurement::Position > > ( "ButtonTriggerPosition" );
    cf->registerComponent< ButtonTrigger< Measurement::Pose > > ( "ButtonTriggerPose" );
  }

} } // namespace Ubitrack::Components
