/*
 * Ubitrack - Library for Ubiquitous Tracking
 * Copyright 2006, Technische Universitaet Muenchen, and individual
 * contributors as indicated by the @authors tag. See the
 * copyright.txt in the distribution for a full listing of individual
 * contributors.
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA, or see the FSF site: http://www.fsf.org.
 */


/**
 * @ingroup dataflow_components
 * @file
 * Components that pulls its input at a given frequency and pushes the result
 *
 * @author Peter Keitler <keitler@in.tum.de>
 */

#include <string>

#include <boost/scoped_ptr.hpp>
#include <boost/bind.hpp>
#include <boost/thread.hpp>
#include <log4cpp/Category.hh>
#include <boost/numeric/ublas/vector.hpp>
#include <boost/numeric/ublas/vector_proxy.hpp>

#include <Ubitrack/Dataflow/Component.h>
#include <Ubitrack/Dataflow/PullConsumer.h>
#include <Ubitrack/Dataflow/PushSupplier.h>
#include <Ubitrack/Dataflow/ComponentFactory.h>
#include <Ubitrack/Measurement/Measurement.h>
#include <Ubitrack/Util/Exception.h>
#include <Ubitrack/Util/OS.h>

using namespace Ubitrack;
using namespace boost::numeric;

//namespace ublas = boost::numeric::ublas;


// get a logger
static log4cpp::Category& logger( log4cpp::Category::getInstance( "Ubitrack.Components.CovarianceEstimation" ) );
static log4cpp::Category& eventLogger( log4cpp::Category::getInstance( "Ubitrack.Events.Components.CovarianceEstimation" ) );

namespace Ubitrack { namespace Components {

/**
 * @ingroup dataflow_components
 * Covariance estimation component
 *
 * Components that pulls single measurements and pushes the resulting (updated) distribution on its output.
 *
 * @par Input Ports
 * PullConsumer< EventType > with name "PerturbedInput".
 * PullConsumer< EventType > with name "ReferenceInput".
 *
 * @par Output Ports
 * PushSupplier< EventType > with name "Distribution".
 *
 * @par Configuration
 *
 * @par Operation
 * A task pulls measurements as often as possible so that the distribution of the input signal can be computed as fast as possible.
 *
 * @par Instances
 *
 */
template< class EventType, class ResultType >
class CovarianceEstimation
	: public Dataflow::Component
{
public:
	/**
	 * UTQL component constructor.
	 *
	 * @param sName Unique name of the component.
	 * @param subgraph UTQL subgraph
	 */
	CovarianceEstimation( const std::string& sName, boost::shared_ptr< Graph::UTQLSubgraph > subgraph )
		: Dataflow::Component( sName )
		, m_inPortPerturbed( "PerturbedInput", *this )
		, m_inPortReference( "ReferenceInput", *this )
		, m_outPort( "Distribution", *this )
		, m_bStop( true )
		, counter ( 0 )
    {
		stop();

		LOG4CPP_DEBUG( logger, "Setup CovarianceEstimation component" );

		// Initialize position and orientation mean vectors and covariance matrix
		posMean = ublas::zero_vector< double >( 3 );
		rotMean = Math::Quaternion();
		covariance = ublas::zero_matrix< double >( 6, 6 );
		partialCovariance = ublas::zero_matrix< double >( 3, 3 );
    }

	~CovarianceEstimation()
	{
		stop();
	}

	/**
	 * component start method, starts thread
	 */
	virtual void start()
	{
		if ( !m_running )
		{
			m_running = true;
			m_bStop = false;
			m_pThread.reset( new boost::thread( boost::bind( &CovarianceEstimation< EventType, ResultType >::threadMethod, this ) ) );
		}
	}

	/**
	 * component stop method, stops thread
	 */
	virtual void stop()
	{
		// stop thread
		// mutex?
		if ( m_running )
		{
			m_running = false;
			m_bStop = true;
			if ( m_pThread )
			{
				m_pThread->join();
			}
		}
	}


protected:
	/** Method that computes the result. */
	void threadMethod();

	typename ResultType::value_type incrementalEstimate( typename EventType::value_type& perturbed, typename EventType::value_type& ref );

	/** Input port of the component. */
	Dataflow::PullConsumer< EventType > m_inPortPerturbed;
	Dataflow::PullConsumer< EventType > m_inPortReference;

	// Output ports of the component
	Dataflow::PushSupplier< ResultType > m_outPort;

	// stop
	bool m_bStop;

	// pointer to the thread
	boost::scoped_ptr< boost::thread > m_pThread;

	Math::Vector< 3 > posMean;
	Math::Quaternion rotMean;
	Math::Matrix< 6, 6 > covariance;
	Math::Matrix< 3, 3 > partialCovariance;
	
	int counter;
};


template<>
Math::ErrorVector< 3 > CovarianceEstimation< Measurement::Position, Measurement::ErrorPosition >::incrementalEstimate( Math::Vector< 3 >& posNew, Math::Vector< 3 >& posRef )
{
	LOG4CPP_TRACE ( logger, "Assumed position mean:" << posRef );
	LOG4CPP_TRACE ( logger, "update position event: " << posNew );

	counter ++;
	LOG4CPP_TRACE( logger, "Counter: " << counter );

	// Maybe information about the empirical mean is useful somehow? It's NOT needed later on for covariance estimation.
	// TODO However, we publish it in the resulting ErrorVector. Is this statistically correct?
	posMean = ( ( ((double)counter - 1) / (double)counter ) * posMean ) + ( ( 1 / (double)counter ) * posNew );
	LOG4CPP_TRACE ( logger, "Empirical position mean: " << posMean );

	// Positional deviation of measurement
	Math::Vector< 3 > errVec = posNew - posRef;
	LOG4CPP_TRACE( logger, "deviation vector update: " << errVec );
	
	// Incremental covariance matrix
	Math::Matrix< 3, 3 > covarianceUpdate = ublas::outer_prod( errVec, errVec );
	LOG4CPP_TRACE( logger, "position covariance update: " << std::endl << covarianceUpdate );
	// Updated (running) covariance matrix (not yet normlized by number of measurements)
	partialCovariance = partialCovariance + covarianceUpdate;
	LOG4CPP_TRACE( logger, "new running position covariance: " << std::endl << partialCovariance );

	if ( counter == 1 ) 
	{
		LOG4CPP_TRACE( logger, "not enough data to compute covariance matrix" );
		return Math::ErrorVector< 3 >();
	}
	else 
	{
		Math::ErrorVector< 3 > result( posMean, partialCovariance / ((double)counter - 1) );
		return result;
	}
}


/**
 * TODO A similar implementation can be found in Ubitrack/Math/UnscentedTransform.h, though that implementation is not incremental.
 * Maybe, we should provide covariance estimation in a more general way...
 */
template<>
Math::ErrorPose CovarianceEstimation< Measurement::Pose, Measurement::ErrorPose >::incrementalEstimate( Math::Pose& poseNew, Math::Pose& poseRef )
{
	LOG4CPP_TRACE ( logger, "Assumed pose mean:" << poseRef );
	LOG4CPP_TRACE ( logger, "update pose event: " << poseNew );

	counter ++;
	LOG4CPP_TRACE( logger, "Counter: " << counter );

	// Additional vectors for computational purposes
	ublas::vector< double > rotMeanVec( 4 );
	// The order is qx, qy, qz, qw.
	rotMean.toVector( rotMeanVec );

	ublas::vector< double > poseNewVec( 7 );
	// The order is tx, ty, tz, qx, qy, qz, qw.
	poseNew.toVector( poseNewVec );
	ublas::vector_range< ublas::vector<double> > posNew( poseNewVec, ublas::range( 0, 3 ) );
	ublas::vector_range< ublas::vector<double> > rotNew( poseNewVec, ublas::range( 3, 7 ) );

	ublas::vector< double > poseRefVec( 7 );
	// The order is tx, ty, tz, qx, qy, qz, qw.
	poseRef.toVector( poseRefVec );
	ublas::vector_range< ublas::vector<double> > posRef( poseRefVec, ublas::range( 0, 3 ) );
	ublas::vector_range< ublas::vector<double> > rotRef( poseRefVec, ublas::range( 3, 7 ) );

	// Take care of quaternion ambiguity
	if ( ublas::inner_prod( rotNew, rotMeanVec ) < 0 )
		rotNew *= -1;

	// Maybe information about the empirical mean is useful somehow? It's NOT needed later on for covariance estimation.
	// TODO However, we publish it in the resulting ErrorVector. Is this statistically correct?
	posMean = ( ( ((double)counter - 1) / (double)counter ) * posMean ) + ( ( 1 / (double)counter ) * posNew );
	rotMeanVec = ( ( ((double)counter - 1) / (double)counter ) * rotMeanVec ) + ( ( 1 / (double)counter ) * rotNew );
	rotMean = Math::Quaternion::fromVector( rotMeanVec );
	rotMean.normalize();
	LOG4CPP_TRACE( logger, "Empirical pose mean: position: " << posMean << ", orientation: " << rotMean );

	// Compute incremental covariance matrix
	Math::Vector< 6 > errVec;
	// Incremental positional deviation
	boost::numeric::ublas::subrange( errVec, 0, 3 ) = posNew - posRef;
	// Incremental rotational deviation
	Math::Quaternion rotNewQuat = Math::Quaternion::fromVector( rotNew );
	rotNewQuat.normalize();
	// rotMean is normalized already
	Math::Quaternion diffQuat( Math::Quaternion::fromVector( rotRef ) * ~rotNewQuat );
	errVec( 3 ) = diffQuat.x();
	errVec( 4 ) = diffQuat.y();
	errVec( 5 ) = diffQuat.z();
	if ( diffQuat.w() < 0 )
		ublas::subrange( errVec, 3, 6 ) *= -1;
	LOG4CPP_TRACE( logger, "deviation vector update: " << errVec );

	// Incremental covariance matrix
	Math::Matrix< 6, 6 > covarianceUpdate = ublas::outer_prod( errVec, errVec );
	LOG4CPP_TRACE( logger, "pose covariance update: " << std::endl << covarianceUpdate );
	// Updated (running) covariance matrix (not yet normlized by number of measurements)
	covariance = covariance + covarianceUpdate;
	LOG4CPP_TRACE( logger, "new running pose covariance: " << std::endl << covariance );
	if ( counter == 1 ) 
	{
		LOG4CPP_TRACE( logger, "not enough data to compute covariance matrix" );
		return Math::ErrorPose();
	}
	else 
	{
		Math::ErrorPose result( rotMean, posMean, covariance / ((double)counter - 1) );
		return result;
	}
}


template< class EventType, class ResultType >
void CovarianceEstimation< EventType, ResultType >::threadMethod()
{
	LOG4CPP_TRACE( logger, "thread started" );

	Measurement::Timestamp now( Measurement::now() );

	while ( !m_bStop )
	{
		LOG4CPP_TRACE( logger, "yield" );

		// Be nice to other threads
		boost::thread::yield();

		LOG4CPP_TRACE( logger, "pull next event" );

		// Pull next measurement. Don't sleep, timestamps are irrelevant. This is not a tracking dataflow!
		now = Measurement::now();
		try
		{
			EventType measurement = m_inPortPerturbed.get( now );
			if ( measurement != NULL )
				m_outPort.send( ResultType ( now, incrementalEstimate ( *measurement, *m_inPortReference.get( now ) ) ) );
			else
				LOG4CPP_INFO( logger, "no more items in list" );
		}
		catch ( const Util::Exception& e )
		{
			LOG4CPP_WARN( eventLogger, "Got exception: " << e );
			Util::sleep ( 100 );
			continue;
		}
		catch ( const std::exception& e )
		{
			LOG4CPP_WARN( eventLogger, "Got unknown exception: " << e.what() );
			Util::sleep ( 100 );
			continue;
		}
	}
}



UBITRACK_REGISTER_COMPONENT( Dataflow::ComponentFactory* const cf ) {
	cf->registerComponent< CovarianceEstimation< Measurement::Pose, Measurement::ErrorPose > > ( "PoseCovarianceEstimation" );
	cf->registerComponent< CovarianceEstimation< Measurement::Position, Measurement::ErrorPosition > > ( "3DPositionCovarianceEstimation" );
}

} } // Namespace Ubitrack::Components
