/*
 * Ubitrack - Library for Ubiquitous Tracking
 * Copyright 2006, Technische Universitaet Muenchen, and individual
 * contributors as indicated by the @authors tag. See the
 * copyright.txt in the distribution for a full listing of individual
 * contributors.
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA, or see the FSF site: http://www.fsf.org.
 */


/**
 * @ingroup dataflow_components
 * @file
 * Direction component.
 * This file contains a direction calculation between two poses implemented as a \c TriggerComponent.
 *
 * @author Florian Echtler <echtler@in.tum.de>
 */
#include <Ubitrack/Dataflow/TriggerComponent.h>
#include <Ubitrack/Dataflow/TriggerInPort.h>
#include <Ubitrack/Dataflow/TriggerOutPort.h>
#include <Ubitrack/Dataflow/ComponentFactory.h>
#include <Ubitrack/Measurement/Measurement.h>

namespace ublas = boost::numeric::ublas;

namespace Ubitrack { namespace Components {

/**
 * @ingroup dataflow_components
 * Direction component.
 * This class contains a direction calculation between two poses implemented as a \c TriggerComponent.
 *
 * @par Input Ports
 * TriggerInPort<Pose> with name "InputA".
 * TriggerInPort<EventType> with name "InputB".
 *
 * @par Output Ports
 * TriggerOutPort<Rotation> with name "Output".
 *
 * @par The Pattern
 * This describes just one of several push/pull combinations
 * @verbatim
<Pattern name="Direction">
	<Input>
		<Node name="A">
			<Attribute name="sourceVector" value="0.0 1.0 0.0"/>
		</Node>
		<Node name="B"/>
		<Node name="C"/>
		<Edge name="AB" source="A" destination="B">
			<Predicate>type=="6D"&amp;&amp;mode=="push"</Predicate>
		</Edge>
		<Edge name="BC" source="B" destination="C">
			<Predicate>type=="3DPos"&amp;&amp;mode=="pull"</Predicate>
		</Edge>
	</Input>
	<Output>
		<Edge name="AC" source="A" destination="C">
			<Attribute name="type" value="3DRot"/>
			<Attribute name="mode" value="push"/>
		</Edge>
	</Output>
	<DataflowConfiguration>
		<UbitrackLib class="PositionPushPullDirection"/>
	</DataflowConfiguration>
</Pattern>
@endverbatim
 *
 * @par Operation
 * The component calculates the direction from pose A to pose/position B, relative to the sourceVector attribute.
 */

template < class EventType > 
class DirectionComponent
	: public Dataflow::TriggerComponent
{
	public:
		/**
		 * UTQL component constructor.
		 *
		 * @param sName Unique name of the component.
		 * @param subgraph UTQL subgraph
		 */
		DirectionComponent( const std::string& sName, boost::shared_ptr< Graph::UTQLSubgraph > cfg )
			: Dataflow::TriggerComponent( sName, cfg )
			, m_inPortA( "AB", *this )
			, m_inPortB( "BC", *this )
			, m_outPort( "AC", *this )
			, m_orig(1, 0, 0)
		{
			Graph::UTQLSubgraph::NodePtr node = cfg->getNode( "A" );
			std::string vector;
			if (node) vector = node->getAttribute( "sourceVector" ).getText();
			if (vector.length() != 0) {
				double v[3] = { 1, 0, 0 };
				std::istringstream vecstream( vector );
				for (int i=0; i < 3; ++i) vecstream >> v[i];
				m_orig = Math::Vector<3>(v);
				m_orig = m_orig / ublas::norm_2( m_orig );
			}
		}

		/** Method that computes the result. */
		void compute( Measurement::Timestamp t )
		{
			m_outPort.send( Measurement::Rotation( t, calculate( m_orig, *m_inPortA.get(), *m_inPortB.get() ) ) );
		}

	protected:

		/** Input port A of the component. */
		Dataflow::TriggerInPort< Measurement::Pose > m_inPortA;

		/** Input port B of the component. */
		Dataflow::TriggerInPort< EventType > m_inPortB;

		/** Output port of the component. */
		Dataflow::TriggerOutPort< Measurement::Rotation > m_outPort;

		/** Unit Vector to be rotated towards target */
		Math::Vector<3> m_orig;

};


/** Method to calculate rotation from unit vector to other pose */
template< class Type > Math::Quaternion calculate( Math::Vector<3>& orig, Math::Pose& pose1, Type& t2 )
{
	return Math::Quaternion();
}

template<> Math::Quaternion calculate< Math::Vector<3> >( Math::Vector<3>& orig, Math::Pose& pose1, Math::Vector<3>& pos2 )
{
	Math::Vector<3> dir = (~pose1) * pos2;
	dir = dir / ublas::norm_2( dir );

	Math::Vector<3> axis = cross_prod( orig, dir );
	
	Math::Quaternion q( axis[0], axis[1], axis[2], 1 + ublas::inner_prod( orig, dir ) );
	q.normalize();

	return pose1.rotation() * q;
}

template<> Math::Quaternion calculate< Math::Pose >( Math::Vector<3>& orig, Math::Pose& pose1, Math::Pose& pose2 )
{
	Math::Vector<3> pos2 = pose2.translation();
	return calculate( orig, pose1, pos2 );
}


UBITRACK_REGISTER_COMPONENT( Dataflow::ComponentFactory* const cf ) {

	cf->registerComponent< DirectionComponent< Measurement::Pose > > ( "PoseDirection" );
	cf->registerComponent< DirectionComponent< Measurement::Position > > ( "PositionDirection" );

}

} } // namespace Ubitrack::Components

