/*
 * Ubitrack - Library for Ubiquitous Tracking
 * Copyright 2006, Technische Universitaet Muenchen, and individual
 * contributors as indicated by the @authors tag. See the 
 * copyright.txt in the distribution for a full listing of individual
 * contributors.
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA, or see the FSF site: http://www.fsf.org.
 */


/**
 * @ingroup dataflow_components
 * @file
 * Matrix multiplication component.
 * This file contains a generic 4x4 matrix transformation implemented as a \c TriggerComponent.
 *
 * @author Florian Echtler <echtler@in.tum.de>
 */
#include <Ubitrack/Dataflow/TriggerComponent.h>
#include <Ubitrack/Dataflow/TriggerInPort.h>
#include <Ubitrack/Dataflow/TriggerOutPort.h>
#include <Ubitrack/Dataflow/ComponentFactory.h>
#include <Ubitrack/Measurement/Measurement.h>
#include <Ubitrack/Math/Matrix.h>

// get a logger
static log4cpp::Category& logger( log4cpp::Category::getInstance( "Ubitrack.Components.Multiplication" ) );

namespace Ubitrack { namespace Components {

/**
 * @ingroup dataflow_components
 * Matrix multiplication component.
 * This file contains a generic 4x4 matrix transformation implemented as a \c TriggerComponent.
 *
 * The component multiplies requested/incoming events with the transformation matrix
 * and returns the homogenized result.
 */
template< class EventType >
class MatrixMultiplicationComponent
	: public Dataflow::TriggerComponent
{
public:
	/**
	 * Standard component constructor.
	 *
	 * @param sName Unique name of the component.
	 * @param cfg ComponentConfiguration containing all configuration.
	 */
	MatrixMultiplicationComponent( const std::string& sName, boost::shared_ptr< Graph::UTQLSubgraph > pConfig )
		: Dataflow::TriggerComponent( sName, pConfig )
		, m_inPortA( "InputA", *this )
		, m_inPortB( "InputB", *this )
		, m_outPort( "Output", *this )
   {}

	/** Method that computes the result. */
	void compute( Measurement::Timestamp t )
	{
		Math::Matrix< 4, 4 > trans = *(m_inPortB.get());
		Math::Vector< 3 > res = *(m_inPortA.get());
		Math::Vector< 4 > hom( res[0], res[1], res[2], 1.0 );
		Math::Vector< 4 > tmp;

		LOG4CPP_TRACE( logger, "received: " << res );
		LOG4CPP_TRACE( logger, "Matrix: " << trans );

		try {
			tmp = boost::numeric::ublas::prod( trans, hom );
			double w = tmp[3];
			tmp = tmp / w;
		}
		catch (void*) {
			return;
		}

		res[0] = tmp[0];
		res[1] = tmp[1];
		res[2] = tmp[2];

		LOG4CPP_TRACE( logger, "transmitting: " << res );

		m_outPort.send( EventType( t, res ) );
	}

protected:
	/** Input ports of the component. */
	Dataflow::TriggerInPort< EventType > m_inPortA;
	Dataflow::TriggerInPort< Measurement::Matrix4x4 > m_inPortB;

	/** Output port of the component. */
	Dataflow::TriggerOutPort< EventType > m_outPort;
};


UBITRACK_REGISTER_COMPONENT( Dataflow::ComponentFactory* const cf ) {
	cf->registerComponent< MatrixMultiplicationComponent< Measurement::Position > > ( "MatrixMultiplication" );
}

} } // namespace Ubitrack::Components

