/*
 * Ubitrack - Library for Ubiquitous Tracking
 * Copyright 2006, Technische Universitaet Muenchen, and individual
 * contributors as indicated by the @authors tag. See the
 * copyright.txt in the distribution for a full listing of individual
 * contributors.
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA, or see the FSF site: http://www.fsf.org.
 */


/**
 * @ingroup dataflow_components
 * @file
 * Perturbation component.
 * This file contains perturbation functionality for common measurement types. It is implemented as a \c TriggerComponent.
 *
 * @author Peter Keitler <keitler@in.tum.de>
 */
#include <string>

#include <Ubitrack/Dataflow/TriggerComponent.h>
#include <Ubitrack/Dataflow/TriggerInPort.h>
#include <Ubitrack/Dataflow/TriggerOutPort.h>
#include <Ubitrack/Dataflow/ComponentFactory.h>
#include <Ubitrack/Measurement/Measurement.h>

#include <log4cpp/Category.hh>
#include <boost/scoped_ptr.hpp>
#include <boost/random.hpp>

using namespace Ubitrack;

// get a logger
static log4cpp::Category& logger( log4cpp::Category::getInstance( "Ubitrack.Components.PerturbationComponent" ) );
static log4cpp::Category& eventLogger( log4cpp::Category::getInstance( "Ubitrack.Events.Components.PerturbationComponent" ) );

namespace Ubitrack { namespace Components {

/**
 * @ingroup dataflow_components
 * Perturbation component.
 *
 * This file contains perturbation functionality for common measurement types. It is implemented as a \c TriggerComponent.
 *
 * @par Input Ports
 * TriggerInPort< EventType > with name "AB" providing correct data.
 *
 * @par Output Ports
 * TriggerOutPort< EventType > with name "AB-Perturbed" providing the input data perturbed by gaussian noise.
 *
 * @par Configuration
 * @verbatim
 * <Configuration posMean="..." posStdDev="..."/>
 * @endverbatim
 *   - \c posMean : float giving the mean of isotropic positional error in m (defaults to "0.0")
 *   - \c posStdDev : float giving the standard deviation of isotropic positional error in m (defaults to "0.001")
 *
 * @par Operation
 *
 * @par Instances
 */
template< class EventType >
class PerturbationComponent
	: public Dataflow::TriggerComponent
{
public:
	/**
	 * UTQL component constructor.
	 *
	 * @param sName Unique name of the component.
	 * @param subgraph UTQL subgraph
	 */
    PerturbationComponent( const std::string& sName, boost::shared_ptr< Graph::UTQLSubgraph > subgraph )
		: Dataflow::TriggerComponent( sName, subgraph )
		, m_inPort( "AB", *this )
		, m_outPort( "AB-Perturbed", *this )
	{
		double posMean( 0.0 );
		double posStdDev( 0.01 );
		subgraph->m_DataflowAttributes.getAttributeData( "posMean", posMean );
		subgraph->m_DataflowAttributes.getAttributeData( "posStdDev", posStdDev );
		double rotMean( 0.0 );
		double rotStdDev( 0.0 );
		subgraph->m_DataflowAttributes.getAttributeData( "rotMean", rotMean );
		subgraph->m_DataflowAttributes.getAttributeData( "rotStdDev", rotStdDev );
		LOG4CPP_DEBUG( logger, "Setup perturbation component " << sName << ". posMean: " << posMean << ", posStdDev: " << posStdDev << ", rotMean: " << rotMean << ", rotStdDev: " << rotStdDev );

		/** Position random number generator. This one is simple (without seed) and allows for reproducible results.*/
		rngPos.reset ( new boost::mt19937( Measurement::now() ) );
		/** Position error distribution. */
		distPos.reset ( new boost::normal_distribution<double>( posMean, posStdDev ) );
		/** Glues randomness with mapping. */
		distSamplerPos.reset ( new boost::variate_generator<boost::mt19937&, boost::normal_distribution<double> >( *rngPos, *distPos ) );

		/** Orientation random number generator. This one is simple (without seed) and allows for reproducible results.*/
		rngRot.reset ( new boost::mt19937( Measurement::now() ) );
		/** Orientation error distribution. */
		distRot.reset ( new boost::normal_distribution<double>( rotMean, rotStdDev ) );
		/** Glues randomness with mapping. */
		distSamplerRot.reset ( new boost::variate_generator<boost::mt19937&, boost::normal_distribution<double> >( *rngRot, *distRot ) );
	}

	/** Method that computes the result. */
	void compute( Measurement::Timestamp t )
	{
		EventType event( t, perturb ( *(m_inPort.get()) ) );
	    m_outPort.send( event );
	}	

protected:
	Math::Vector< 3 > perturbPosition( Math::Vector<3> pos ) 
	{
		return pos + Math::Vector< 3 >( (*distSamplerPos)(), (*distSamplerPos)(), (*distSamplerPos)() );
	}
	
	Math::Quaternion perturbOrientation( Math::Quaternion rot ) 
	{
		Math::Quaternion errQuat( (*distSamplerRot)(), (*distSamplerRot)(), (*distSamplerRot)(), 1 );
		return rot * errQuat.normalize();
	}

	typename EventType::value_type perturb( const typename EventType::value_type& ref );
	
	/** Input port of the component. */
	Dataflow::TriggerInPort< EventType > m_inPort;
	/** Output port of the component. */
	Dataflow::TriggerOutPort< EventType > m_outPort;

	/** Position random number generator. */
	boost::scoped_ptr< boost::mt19937 > rngPos;
	/** Position error distribution. */
	boost::scoped_ptr< boost::normal_distribution<double> > distPos;
	/** Glues randomness with mapping. */
	boost::scoped_ptr< boost::variate_generator<boost::mt19937&, boost::normal_distribution<double> > > distSamplerPos;
	/** Orientation random number generator. */
	boost::scoped_ptr< boost::mt19937 > rngRot;
	/** Orientation error distribution. */
	boost::scoped_ptr< boost::normal_distribution<double> > distRot;
	/** Glues randomness with mapping. */
	boost::scoped_ptr< boost::variate_generator<boost::mt19937&, boost::normal_distribution<double> > > distSamplerRot;
};


template<>
Math::Vector< 3 > PerturbationComponent< Measurement::Position >::perturb(
	const Math::Vector< 3 >& ref )
{
	return perturbPosition ( ref );
}


template<>
std::vector< Math::Vector< 3 > > PerturbationComponent< Measurement::PositionList >::perturb( const std::vector< Math::Vector< 3 > >& refList )
{
	std::vector< Math::Vector< 3 > > result( refList.size() );
	for ( unsigned i = 0; i < refList.size(); i++ ) 
	{
		result[ i ] = perturbPosition ( refList[ i ] );
		LOG4CPP_TRACE( logger, "Perturbed list point: " << result[i] );
	}
	
	return result;
}


template<>
std::vector< Math::Pose > PerturbationComponent< Measurement::PoseList >::perturb( const std::vector< Math::Pose >& refList )
{
	std::vector< Math::Pose > result( refList.size() );
	for ( unsigned i = 0; i < refList.size(); i++ ) 
	{
		result[ i ] = Math::Pose( perturbOrientation ( refList[ i ].rotation() ), perturbPosition ( refList[ i ].translation() ) );
		LOG4CPP_TRACE( logger, "Perturbed list pose: " << result[i] );
	}
	
	return result;
}


template<>
Math::Quaternion PerturbationComponent< Measurement::Rotation >::perturb(
	const Math::Quaternion& ref )
{
	return perturbOrientation ( ref );
}


template<>
Math::Pose PerturbationComponent< Measurement::Pose >::perturb(
	const Math::Pose& ref )
{
	return Math::Pose( perturbOrientation ( ref.rotation() ), perturbPosition ( ref.translation() )  );
}


UBITRACK_REGISTER_COMPONENT( Dataflow::ComponentFactory* const cf ) {
	cf->registerComponent< PerturbationComponent< Measurement::Position > > ( "3DPositionPerturbation" );
	cf->registerComponent< PerturbationComponent< Measurement::PositionList > > ( "3DPositionListPerturbation" );
	cf->registerComponent< PerturbationComponent< Measurement::Rotation > > ( "RotationPerturbation" );
	cf->registerComponent< PerturbationComponent< Measurement::Pose > > ( "PosePerturbation" );
	cf->registerComponent< PerturbationComponent< Measurement::PoseList > > ( "PoseListPerturbation" );
}

} } // namespace Ubitrack::Components
