/*
 * Ubitrack - Library for Ubiquitous Tracking
 * Copyright 2006, Technische Universitaet Muenchen, and individual
 * contributors as indicated by the @authors tag. See the
 * copyright.txt in the distribution for a full listing of individual
 * contributors.
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA, or see the FSF site: http://www.fsf.org.
 */


/**
 * @ingroup dataflow_components
 * @file
 * Residual error component.
 * This file contains a component to compute the residual error between two lists of points.
 *
 * @author Peter Keitler <keitler@in.tum.de>
 */
#include <Ubitrack/Dataflow/TriggerComponent.h>
#include <Ubitrack/Dataflow/ExpansionInPort.h>
#include <Ubitrack/Dataflow/TriggerOutPort.h>
#include <Ubitrack/Dataflow/ComponentFactory.h>
#include <Ubitrack/Measurement/Measurement.h>

namespace Ubitrack { namespace Components {

/**
 * @ingroup dataflow_components
 * Residual error component.
 * This file contains a component to compute the residual error between two lists of points.
 *
 * @par Input Ports
 * ExpansionInPort<Position> with name "InputA".
 * ExpansionInPort<Position> with name "InputB".
 *
 * @par Output Ports
 * TriggerOutPort<Distance> with name "Residual".
 *
 * @par Configuration
 * DataflowConfiguration: expansion="space" or "time" for time/space expansion
 *
 * @par Operation
 * The component computes the residual error between two lists of points having equal length.
 * The computed value is the sum of the sqaured euclidean distances of the point correspondences, normalized (divided) by the amount of correspondences.
 */
class ResidualErrorComponent
	: public Dataflow::TriggerComponent
{
public:
	/**
	 * UTQL component constructor.
	 *
	 * @param sName Unique name of the component.
	 * @param subgraph UTQL subgraph
	 */
	ResidualErrorComponent( const std::string& sName, boost::shared_ptr< Graph::UTQLSubgraph > pConfig )
		: Dataflow::TriggerComponent( sName, pConfig )
		, m_inPortA( "InputA", *this )
		, m_inPortB( "InputB", *this )
		, m_outPort( "Residual", *this )
    {
    }

	/** Method that computes the result. */
	void compute( Measurement::Timestamp t )
	{
		if ( m_inPortA.get()->size() != m_inPortB.get()->size() || m_inPortA.get()->size() < 1 )
			UBITRACK_THROW( "Illegal number of correspondences" );
		
		// Get measurement payload (*) for both lists
		std::vector<Ubitrack::Math::Vector<3, double> >& left = *m_inPortA.get();
		std::vector<Ubitrack::Math::Vector<3, double> >& right = *m_inPortB.get();

		// Sum squared euclidean distances over all point correspondences
		double residual = 0;
		for ( unsigned int i = 0; i < left.size(); i++ ) {
		    Ubitrack::Math::Vector<3, double> l = left[i];
		    Ubitrack::Math::Vector<3, double> r = right[i];
		    
		    residual += (l[0]-r[0])*(l[0]-r[0]) + (l[1]-r[1])*(l[1]-r[1]) + (l[2]-r[2])*(l[2]-r[2]);
		}

 		boost::shared_ptr< Math::Scalar<double> > res( new Math::Scalar<double>( residual ) );

		m_outPort.send( Measurement::Distance( t, res ) );
    }

protected:
	/** Input port A of the component. */
	Dataflow::ExpansionInPort< Math::Vector< 3 > > m_inPortA;

	/** Input port B of the component. */
	Dataflow::ExpansionInPort< Math::Vector< 3 > > m_inPortB;

	/** Output port of the component. */
	Dataflow::TriggerOutPort< Measurement::Distance > m_outPort;
};


UBITRACK_REGISTER_COMPONENT( Dataflow::ComponentFactory* const cf ) {
	cf->registerComponent< ResidualErrorComponent > ( "ResidualError" );
}

} } // namespace Ubitrack::Components
