/*
 * Ubitrack - Library for Ubiquitous Tracking
 * Copyright 2006, Technische Universitaet Muenchen, and individual
 * contributors as indicated by the @authors tag. See the
 * copyright.txt in the distribution for a full listing of individual
 * contributors.
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA, or see the FSF site: http://www.fsf.org.
 */

#ifndef __UBITRACK_COMPONENTS_TRIGGER_H_INCLUDED__
#define __UBITRACK_COMPONENTS_TRIGGER_H_INCLUDED__

/**
 * @ingroup dataflow_components
 * @file
 * Trigger component
 * This class pulles measurements and pushes it, if a trigger measurement is received
 *
 * @author Michael Schlegel <schlegem@in.tum.de>
 */

#include <boost/bind.hpp>
#include <log4cpp/Category.hh>

#include <Ubitrack/Dataflow/PushConsumer.h>
#include <Ubitrack/Dataflow/PushSupplier.h>
#include <Ubitrack/Dataflow/PullConsumer.h>
#include <Ubitrack/Dataflow/Component.h>
#include <Ubitrack/Measurement/Measurement.h>
#include <Ubitrack/Graph/UTQLSubgraph.h>

namespace Ubitrack { namespace Components {

    using namespace Dataflow;

    /**
     * @ingroup dataflow_components
     * Trigger component
     * This class pulles measurements and pushes it, if a trigger measurement is received
     *
     * @par Operation
     * Whenever a trigger measurement  is received, an pulled measurement is pushed incoming Event is allowed to pass
     *
     */		

    template< class EventType > 
      class Trigger : public Component
    {
    public:

      /**
       * Standard component constructor.
       *
       * @param nm Unique name of the component.
       * @param pCfg ComponentConfiguration containing all configuration.
       */
      Trigger( const std::string& nm, boost::shared_ptr< Graph::UTQLSubgraph > pCfg )
        : Ubitrack::Dataflow::Component( nm )
	, m_inPort( "Input", *this )
	, m_SignalPort( "TriggerSignal", *this, boost::bind( &Trigger::trigger, this, _1 ) )
	, m_outPort( "Output", *this )
	, logger( log4cpp::Category::getInstance( "Ubitrack.Components.Trigger" ) )
	{
	  //nop
	}

    protected:

      void trigger( const EventType& triggerData )
      {
	// pull data with time stamp from triggerData and push the result
	m_outPort.send( m_inPort.get( triggerData.time() ) );
	LOG4CPP_DEBUG( logger, "Trigger sending measurement" );
      }
      
      
      
      /** properties of the trigger*/
      
      /**last received event*/
      EventType data;
      
      /** Ports of the component. */
      PullConsumer< EventType > m_inPort;
      PushConsumer< EventType > m_SignalPort;
      PushSupplier< EventType > m_outPort;
      
      /** log4cpp logger reference */
      log4cpp::Category& logger;	
    };
    
  }
} // namespace Ubitrack::Components

#endif
