/*
 * Ubitrack - Library for Ubiquitous Tracking
 * Copyright 2006, Technische Universitaet Muenchen, and individual
 * contributors as indicated by the @authors tag. See the
 * copyright.txt in the distribution for a full listing of individual
 * contributors.
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA, or see the FSF site: http://www.fsf.org.
 */

/**
 * @ingroup api
 * @file
 * Simple datatypes for the \c SimpleFacade
 *
 * @author Daniel Pustka <daniel.pustka@in.tum.de>
 */
#ifndef __UBITRACK_FACADE_SIMPLEDATATYPES_H_INCLUDED__
#define __UBITRACK_FACADE_SIMPLEDATATYPES_H_INCLUDED__

namespace Ubitrack { namespace Facade {

/**
 * A simplified version of the Measurement::Pose type without boost.
 */
struct SimplePose
{
	/** position as a 3-vector */
	double tx, ty, tz;

	/** orientation as a (x, y, z, w) quaternion */
	double rx, ry, rz, rw;

	/** timestamp in nanoseconds since unix epoch */
	unsigned long long int timestamp;
};


/**
 * A simplified version of the Measurement::Distance type without boost.
 */
struct SimpleDistance
{
	/** distance */
	double d;

	/** timestamp in nanoseconds since unix epoch */
	unsigned long long int timestamp;
};


/**
 * A simple callback interface to transport SimplePoses
 */
class SimplePoseReceiver
{
public:
	/** receives a SimplePose */
	virtual void receivePose( const SimplePose& pose ) throw() = 0;
	
	/** virtual destructor */
	virtual ~SimplePoseReceiver()
	{}
};


/**
 * A simple callback interface to transport SimpleInts
 */
class SimpleDistanceReceiver
{
public:
	/** receives a SimpleDistance */
	virtual void receiveDistance( const SimpleDistance& dist ) throw() = 0;
	
	/** virtual destructor */
	virtual ~SimpleDistanceReceiver()
	{}
};


/**
 * A simplified version of the Measurement::Button type without boost.
 */
struct SimpleButton
{
	/** button event as scalar */
	int event;

	/** timestamp in nanoseconds since unix epoch */
	unsigned long long int timestamp;
};


/**
 * A simple callback interface to transport SimpleButton
 */
class SimpleButtonReceiver
{
public:
	/** receives a SimpleButton */
	virtual void receiveButton( const SimpleButton& button ) throw() = 0;
	
	/** virtual destructor */
	virtual ~SimpleButtonReceiver()
	{}
};


/**
 * A simplified version of the Vision::ImageMeasurement type without boost.
 */
struct SimpleImage
{
	/** image raw data */
	int width;
	int height;
	int imageSize;
	int widthStep;
	int depth;
	int nChannels;
	signed char * imageData;

	/** timestamp in nanoseconds since unix epoch */
	unsigned long long int timestamp;
};


/**
 * A simple callback interface to transport SimpleImage
 */
class SimpleImageReceiver
{
public:
	/** receives a SimpleImage */
	virtual void receiveImage( const SimpleImage& image ) throw() = 0;
	
	/** virtual destructor */
	virtual ~SimpleImageReceiver()
	{}
};


/**
 * A simple callback interface to transport strings.
 * Can be used for language-independent transport of more complex data types.
 */
class SimpleStringReceiver
{
public:
	/** receives a SimpleButton */
	virtual void receiveString( const char* ) throw() = 0;
	
	/** virtual destructor */
	virtual ~SimpleStringReceiver()
	{}
};


/**
 * A simple data flow observer
 */
class SimpleDataflowObserver
{
public:
	/** called when a component is added */
	virtual void notifyAddComponent( const char* sPatternName, const char* sComponentName ) throw() = 0;

	/** called when a component is removed */
	virtual void notifyDeleteComponent( const char* sPatternName, const char* sComponentName ) throw() = 0;
	
	/** virtual destructor */
	virtual ~SimpleDataflowObserver()
	{}
};


/**
 * A simple proxy object for an ApplicationPullSinkPose component
 */
class SimpleApplicationPullSinkPose {
protected:
	SimpleApplicationPullSinkPose()
	{}

public:
	/** virtual destructor */
	virtual ~SimpleApplicationPullSinkPose()
	{}

	/**
	 * Retrieves current pose data with the given timestamp
	 *
	 * @param pose Pose data is returned in this object on success
	 * @param timestamp Timestamp for which measurement shall be retrieved
	 * @return true if successfull
	 */
	virtual bool getPose( SimplePose & pose, unsigned long long int timestamp ) = 0;
};


} } // namespace Ubitrack::Facade

#endif
