/*
 * Ubitrack - Library for Ubiquitous Tracking
 * Copyright 2006, Technische Universitaet Muenchen, and individual
 * contributors as indicated by the @authors tag. See the
 * copyright.txt in the distribution for a full listing of individual
 * contributors.
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA, or see the FSF site: http://www.fsf.org.
 */

/**
 * @ingroup api
 * @file
 * Implements a simple interface to the ubitrack dataflow network.
 *
 * @author Daniel Pustka <daniel.pustka@in.tum.de>
 */

#include <stdlib.h>
#include <string.h>
#include <boost/bind.hpp>
#include <log4cpp/Category.hh>

#include <Ubitrack/Measurement/Measurement.h>
#include <Ubitrack/Components/ApplicationPullSink.h>

#ifdef HAVE_OPENCV
#include <Vision/Image.h>
#endif

#include "SimpleFacade.h"
#include "AdvancedFacade.h"
#include "DataflowObserver.h"

// get a logger
static log4cpp::Category& logger( log4cpp::Category::getInstance( "Ubitrack.Facade.SimpleFacade" ) );


namespace {

// this function converts Measurement::Pose to SimplePose in a callback
void convertPoseCallback( Ubitrack::Facade::SimplePoseReceiver* receiver, const Ubitrack::Measurement::Pose& measurement )
{
	using namespace Ubitrack;
	Facade::SimplePose p;
	p.tx = measurement->translation()( 0 );
	p.ty = measurement->translation()( 1 );
	p.tz = measurement->translation()( 2 );
	p.rx = measurement->rotation().x();
	p.ry = measurement->rotation().y();
	p.rz = measurement->rotation().z();
	p.rw = measurement->rotation().w();
	p.timestamp = measurement.time();
	receiver->receivePose( p );
}

// this function converts Measurement::Distance to SimpleDistance in a callback
void convertDistanceCallback( Ubitrack::Facade::SimpleDistanceReceiver* receiver, const Ubitrack::Measurement::Distance& measurement )
{
	using namespace Ubitrack;
	Facade::SimpleDistance d;
	d.d = *measurement;
	d.timestamp = measurement.time();
	receiver->receiveDistance( d );
}

#ifdef HAVE_OPENCV
// this function converts Vision::ImageMeasurements to SimpleImage in a callback
void convertImageCallback( Ubitrack::Facade::SimpleImageReceiver* receiver, const Ubitrack::Vision::ImageMeasurement& measurement )
{
	using namespace Ubitrack;
	Facade::SimpleImage i;

	i.width = measurement->width;
	i.height = measurement->height;
	i.imageData = (signed char *) measurement->imageData;
	i.widthStep = measurement->widthStep;
	i.depth = measurement->depth;
	i.nChannels = measurement->nChannels;

	i.timestamp = measurement.time();

	receiver->receiveImage( i );
}
#endif

}


namespace Ubitrack { namespace Facade {

/** private things */
class SimpleFacadePrivate
	: public AdvancedFacade
	, public DataflowObserver
{
public:
	SimpleFacadePrivate( const char* sComponentPath )
		: AdvancedFacade( sComponentPath ? sComponentPath : std::string() )
		, m_pSimpleObserver( 0 )
	{}

	// translate from DataflowObserver to SimpleDataflowObserver
	void notifyAddComponent( const std::string& sPatternName, const std::string& sComponentName, const Graph::UTQLSubgraph& )
	{
		if ( m_pSimpleObserver )
			m_pSimpleObserver->notifyAddComponent( sPatternName.c_str(), sComponentName.c_str() );
	}

	void notifyDeleteComponent( const std::string& sPatternName, const std::string& sComponentName )
	{
		if ( m_pSimpleObserver )
			m_pSimpleObserver->notifyDeleteComponent( sPatternName.c_str(), sComponentName.c_str() );
	}
	
	SimpleDataflowObserver* m_pSimpleObserver;
};



/**
 * Override SimpleApplicationPullSinkPose to provide proxy functionality for the
 * actual dataflow component
 */
class SimpleApplicationPullSinkPosePrivate 
	: public SimpleApplicationPullSinkPose {

private:
	/** dataflow component this proxy object represents */
	Ubitrack::Components::ApplicationPullSinkPose * pSink;

	/** content of the last error message */
	char* m_sError;

	/** sets the internal error string */
	void setError( const char* sMsg ) throw();


public:
	SimpleApplicationPullSinkPosePrivate( Ubitrack::Components::ApplicationPullSinkPose * pSink );
	
	~SimpleApplicationPullSinkPosePrivate();
	
	virtual bool getPose( SimplePose & pose, unsigned long long int timestamp );
};



SimpleApplicationPullSinkPosePrivate::SimpleApplicationPullSinkPosePrivate( Ubitrack::Components::ApplicationPullSinkPose * pSink )
	: m_sError( 0 )
{
	this->pSink = pSink;
}


SimpleApplicationPullSinkPosePrivate::~SimpleApplicationPullSinkPosePrivate()
{
	if ( m_sError )
		free( m_sError );
}


/** retrieves a pose for the given timestamp */
bool SimpleApplicationPullSinkPosePrivate::getPose( SimplePose & pose, unsigned long long int timestamp )
{
	if ( pSink == NULL )
		return false;
	
	try {
		LOG4CPP_INFO( logger, "Trying to pull pose in SimpleApplicationPullSinkPosePrivate::getPose() for timestamp: " << timestamp );
		// Retrieve measurement for current timestamp
		Ubitrack::Measurement::Pose p = pSink->get(timestamp);
		LOG4CPP_INFO( logger, "Sucessfully pulled pose in SimpleApplicationPullSinkPosePrivate::getPose(): " << p );
		// Convert measurement
		pose.tx = p->translation()( 0 );
		pose.ty = p->translation()( 1 );
		pose.tz = p->translation()( 2 );
		pose.rx = p->rotation().x();
		pose.ry = p->rotation().y();
		pose.rz = p->rotation().z();
		pose.rw = p->rotation().w();
		pose.timestamp = p.time();
	}
	catch ( const Ubitrack::Util::Exception& e )
	{
		LOG4CPP_ERROR( logger, "Caught exception in SimpleApplicationPullSinkPosePrivate::getPose(): " << e );
		setError( e.what() );
		return false;
	}

	return true;
}


void SimpleApplicationPullSinkPosePrivate::setError( const char* sMsg ) throw()
{
	if ( m_sError )
		free( m_sError );
	m_sError = strdup( sMsg );
}


SimpleFacade::SimpleFacade( const char* sComponentPath ) throw()
	: m_pPrivate( 0 )
	, m_sError( 0 )
{
	try
	{
		m_pPrivate = new SimpleFacadePrivate( sComponentPath );
		LOG4CPP_TRACE( logger, "SimpleFacadePrivate created successfully" );
	}
	catch ( const Ubitrack::Util::Exception& e )
	{
		LOG4CPP_ERROR( logger, "Caught exception constructing SimpleFacade: " << e );
		setError( e.what() );
	}
}


SimpleFacade::~SimpleFacade()
{
	delete m_pPrivate;
	if ( m_sError )
		free( m_sError );
}


bool SimpleFacade::loadDataflow( const char* sDfSrg ) throw()
{
	try
	{
		m_pPrivate->loadDataflow( sDfSrg );
	}
	catch ( const Ubitrack::Util::Exception& e )
	{
		LOG4CPP_ERROR( logger, "Caught exception in SimpleFacade::loadDataflow: " << e );
		setError( e.what() );
		return false;
	}

	return true;
}


void SimpleFacade::startDataflow() throw()
{
	try
	{
		// start the event queue
		m_pPrivate->startDataflow();
	}
	catch ( const Ubitrack::Util::Exception& e )
	{
		LOG4CPP_ERROR( logger, "Caught exception in SimpleFacade::startDataflow: " << e );
		setError( e.what() );
	}
}


void SimpleFacade::stopDataflow() throw()
{
	try
	{
		// stop the event queue
		m_pPrivate->stopDataflow();
	}
	catch ( const Ubitrack::Util::Exception& e )
	{
		LOG4CPP_ERROR( logger, "Caught exception in SimpleFacade::stopDataflow: " << e );
		setError( e.what() );
	}
}


void SimpleFacade::connectToServer( const char* sAddress ) throw()
{
	try
	{
		m_pPrivate->connectToServer( sAddress );
	}
	catch ( const Ubitrack::Util::Exception& e )
	{
		LOG4CPP_ERROR( logger, "Caught exception in SimpleFacade::connectToServer: " << e );
		setError( e.what() );
	}
}


void SimpleFacade::sendUtqlToServer( const char* sUtqlFile ) throw()
{
	try
	{
		m_pPrivate->sendUtqlToServer( sUtqlFile );
	}
	catch ( const Ubitrack::Util::Exception& e )
	{
		LOG4CPP_ERROR( logger, "Caught exception in SimpleFacade::sendUtqlToServer: " << e );
		setError( e.what() );
	}
}


void SimpleFacade::sendUtqlToServerString( const char* buffer ) throw()
{
	try
	{
		m_pPrivate->sendUtqlToServerString( buffer );
	}
	catch ( const Ubitrack::Util::Exception& e )
	{
		LOG4CPP_ERROR( logger, "Caught exception in SimpleFacade::sendUtqlToServerString: " << e );
		setError( e.what() );
	}
}


bool SimpleFacade::setPoseCallback( const char* sCallbackName, SimplePoseReceiver* pCallback ) throw()
{
	try
	{
		m_pPrivate->setCallback< Measurement::Pose >( sCallbackName, boost::bind( &convertPoseCallback, pCallback, _1 ) );
	}
	catch ( const Ubitrack::Util::Exception& e )
	{
		LOG4CPP_ERROR( logger, "Caught exception in SimpleFacade::setPoseCallback( " << sCallbackName <<" ): " << e );
		setError( e.what() );
		return false;
	}

	return true;
}


bool SimpleFacade::setDistanceCallback( const char* sCallbackName, SimpleDistanceReceiver* pCallback ) throw()
{
	try
	{
		m_pPrivate->setCallback< Measurement::Distance >( sCallbackName, boost::bind( &convertDistanceCallback, pCallback, _1 ) );
	}
	catch ( const Ubitrack::Util::Exception& e )
	{
		LOG4CPP_ERROR( logger, "Caught exception in SimpleFacade::setDistanceCallback( " << sCallbackName <<" ): " << e );
		setError( e.what() );
		return false;
	}

	return true;
}


bool SimpleFacade::setStringCallback( const char* sCallbackName, SimpleStringReceiver* pCallback ) throw()
{
	try
	{
		m_pPrivate->componentByName< Components::ApplicationPushSinkBase >( sCallbackName )->setStringCallback( pCallback );
	}
	catch ( const Ubitrack::Util::Exception& e )
	{
		LOG4CPP_ERROR( logger, "Caught exception in SimpleFacade::setStringCallback( " << sCallbackName <<" ): " << e );
		setError( e.what() );
		return false;
	}

	return true;
}


#ifdef HAVE_OPENCV
bool SimpleFacade::setImageCallback( const char* sCallbackName, SimpleImageReceiver* pCallback ) throw()
{
	try
	{
		m_pPrivate->setCallback< Vision::ImageMeasurement >( sCallbackName, boost::bind( &convertImageCallback, pCallback, _1 ) );
	}
	catch ( const Ubitrack::Util::Exception& e )
	{
		LOG4CPP_ERROR( logger, "Caught exception in SimpleFacade::setImageCallback( " << sCallbackName <<" ): " << e );
		setError( e.what() );
		return false;
	}

	return true;
}
#endif


SimplePoseReceiver* SimpleFacade::getPushSourcePose( const char* sComponentName ) throw()
{
	try
	{
		return m_pPrivate->componentByName< Components::ApplicationPushSourcePose >( sComponentName ).get();
	}
	catch ( const Ubitrack::Util::Exception& e )
	{
		LOG4CPP_ERROR( logger, "Caught exception in SimpleFacade::getPushSourcePose( " << sComponentName <<" ): " << e );
		setError( e.what() );
		return 0;
	}
}


SimpleButtonReceiver* SimpleFacade::getPushSourceButton( const char* sComponentName ) throw()
{
	try
	{
	  SimpleButtonReceiver * pr = m_pPrivate->componentByName< Components::ApplicationPushSourceButton >( sComponentName ).get();
	  LOG4CPP_DEBUG( logger, "Successfully retrieved component " << sComponentName << " in SimpleFacade::getPushSourceButton " );
	  return pr;
	}
	catch ( const Ubitrack::Util::Exception& e )
	{
		LOG4CPP_ERROR( logger, "Caught exception in SimpleFacade::getPushSourceButton( " << sComponentName <<" ): " << e );
		setError( e.what() );
		return 0;
	}
}


SimpleStringReceiver* SimpleFacade::getPushSourceString( const char* sComponentName ) throw()
{
	try
	{
	  SimpleStringReceiver * pr = m_pPrivate->componentByName< SimpleStringReceiver >( sComponentName ).get();
	  LOG4CPP_DEBUG( logger, "Successfully retrieved component " << sComponentName << " in SimpleFacade::getPushSourceString " );
	  return pr;
	}
	catch ( const Ubitrack::Util::Exception& e )
	{
		LOG4CPP_ERROR( logger, "Caught exception in SimpleFacade::getPushSourceString( " << sComponentName <<" ): " << e );
		setError( e.what() );
		return 0;
	}
}


SimpleApplicationPullSinkPose* SimpleFacade::getSimplePullSinkPose( const char* sComponentName ) throw()
{
	Ubitrack::Components::ApplicationPullSinkPose * pSink;

	try
	{
		 pSink = m_pPrivate->componentByName< Components::ApplicationPullSinkPose >( sComponentName ).get();
	}
	catch ( const Ubitrack::Util::Exception& e )
	{
		LOG4CPP_ERROR( logger, "Caught exception in SimpleFacade::getSimplePullSinkPose( " << sComponentName <<" ): " << e );
		setError( e.what() );
		return 0;
	}

	// Create proxy object and return it
	SimpleApplicationPullSinkPose * pSinkPriv = new SimpleApplicationPullSinkPosePrivate( pSink );

	return pSinkPriv;
}


void SimpleFacade::setDataflowObserver( SimpleDataflowObserver* pObserver ) throw()
{
	if ( !m_pPrivate->m_pSimpleObserver )
		m_pPrivate->addDataflowObserver( m_pPrivate );
	m_pPrivate->m_pSimpleObserver = pObserver;
}

	
void SimpleFacade::removeDataflowObserver() throw()
{
	m_pPrivate->removeDataflowObserver( m_pPrivate );
	m_pPrivate->m_pSimpleObserver = 0;
}


const char* SimpleFacade::getLastError() throw()
{
	if ( m_pPrivate )
		return m_sError;
	else
		return "Error constructing SimpleFacade -- see logging for details";
}


void SimpleFacade::setError( const char* sMsg ) throw()
{
	if ( m_sError )
		free( m_sError );
	m_sError = strdup( sMsg );
}

} } // namespace Ubitrack::Facade
