/*
 * Ubitrack - Library for Ubiquitous Tracking
 * Copyright 2006, Technische Universitaet Muenchen, and individual
 * contributors as indicated by the @authors tag. See the
 * copyright.txt in the distribution for a full listing of individual
 * contributors.
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
5~ * 02110-1301 USA, or see the FSF site: http://www.fsf.org.
 */


/**
 * @ingroup srg_algorithms
 *
 * @author Manuel Huber <huberma@in.tum.de>
 */

#ifndef __Ubitrack_Graph_Pattern_INCLUDED__
#define __Ubitrack_Graph_Pattern_INCLUDED__ __Ubitrack_Graph_Pattern_INCLUDED__

#include <Ubitrack/Ubitrack.h>
#include <Ubitrack/Graph/EdgeMatching.h>
#include <boost/shared_ptr.hpp>

#include <Ubitrack/Graph/SRGraph.h>
#include <Ubitrack/Graph/UTQLSubgraph.h>
#include <Ubitrack/Graph/EdgeMatching.h>
#include <Ubitrack/Graph/Predicate.h>
#include <Ubitrack/Graph/EvaluationContext.h>

#include <stack>

namespace Ubitrack { namespace Graph {

	/** internal information about a pattern */
	class Pattern
    {
    public:
		/** constructor */
		Pattern( boost::shared_ptr< UTQLSubgraph > graph, const std::string& clientId )
			: m_Name( graph->m_Name )
			, m_clientId( clientId )
			, m_Graph( graph )
		{}

		/** name of the pattern */
		std::string m_Name;

		/** client id of the pattern */
		std::string m_clientId;

		/** the utql pattern description */
		boost::shared_ptr< UTQLSubgraph > m_Graph;

		// code the world is not yet ready for..
		// bool m_hasCorrespondence;
		// std::set< SrGraph::Node > m_FreeNodes;
		// std::string m_Expansion = "S";
		// Node u1, u2, v1, v2
		// Multi ?
		//
    };

	class PatternFunc
	{
	public:

		static bool isVertexCompatible( const UTQLSubgraph::NodePtr patternNode, const SRGraph::NodePtr srgNode )
		{
			for ( std::list< boost::shared_ptr< Predicate > >::const_iterator predicate = patternNode->m_predicateList.begin();
				  predicate != patternNode->m_predicateList.end(); ++predicate )
			{
				try 
				{ 
					if ( !(*predicate)->evaluate( EvaluationContext( *srgNode ) ) )
						return false;
				}
				catch( ... )
				{ return false; }

			}
			return true;
		}

		static bool isEdgeCompatible( UTQLSubgraph::EdgePtr patternEdge, SRGraph::EdgePtr srgEdge )
		{
			const UTQLSubgraph::NodePtr patternSource = patternEdge->m_Source.lock();
			const UTQLSubgraph::NodePtr patternTarget = patternEdge->m_Target.lock();

			const SRGraph::NodePtr srgSource = srgEdge->m_Source.lock();
			const SRGraph::NodePtr srgTarget = srgEdge->m_Target.lock();

			if ( ! ( isVertexCompatible( patternSource, srgSource ) && isVertexCompatible( patternTarget, srgTarget ) ) )
			{
				return false;
			}

			for ( std::list< boost::shared_ptr< Predicate > >::const_iterator predicate = patternEdge->m_predicateList.begin();
				  predicate != patternEdge->m_predicateList.end(); ++predicate )
			{
				try 
				{ 
					if ( !(*predicate)->evaluate( EvaluationContext( *srgEdge ) ) )
						return false;
				}
				catch( ... )
				{ return false; }

			}
			return true;
		}


		static boost::shared_ptr< std::vector< EdgeMatching< UTQLSubgraph, SRGraph > > > checkPattern ( boost::shared_ptr< Pattern > p, SRGraph& srg )
		{
			boost::shared_ptr< std::vector< EdgeMatching< UTQLSubgraph, SRGraph > > > detectedMatches ( new std::vector< EdgeMatching< UTQLSubgraph, SRGraph > > );

			std::stack < EdgeMatching< UTQLSubgraph, SRGraph > > stack;

			// start with a completely unmatched pattern
			stack.push( EdgeMatching< UTQLSubgraph, SRGraph >() );

			// while stack is not empty
			while ( !stack.empty() )
			{
				EdgeMatching< UTQLSubgraph, SRGraph > state = stack.top();
				stack.pop();

				// some general purpose edge iterators
				// SrGraph::EdgeIterator edgeBegin, edgeEnd;

				// find next unmacthed edge in pattern
				// bool newEdgeFound = false;

				bool newEdgeFound = false;
				UTQLSubgraph::EdgePtr newEdge;

				bool newVertexFound = false;
				UTQLSubgraph::NodePtr newVertex;

				// iterate over all edges in the pattern
				for (std::map< std::string, UTQLSubgraph::EdgePtr >::iterator it = p->m_Graph->m_Edges.begin();
					 it != p->m_Graph->m_Edges.end(); ++it )
				{

					// XXX: TODO: sucks. this should be handeled by the InOutAttributeIterator
					// is unknown edge? TODO: should go somewhere else
					UTQLSubgraph::EdgePtr e = it->second;
					if ( !e->isInput() )
					{
						continue;
					}

					if ( state.isPatternEdgeMatched ( e ) )
					{
						continue;
					}

					newEdge = e;
					newEdgeFound = true;
					break;

					// TODO: potential: prefer edges which have already one macthed vertex
				}

				if ( !newEdgeFound )
				{
					for (std::map< std::string, UTQLSubgraph::NodePtr >::iterator it = p->m_Graph->m_Nodes.begin();
						 it != p->m_Graph->m_Nodes.end(); ++it )
					{
						UTQLSubgraph::NodePtr u = it->second;
						if ( !u->isInput() )
						{
							continue;
						}

						if ( state.isPatternVertexMatched( u ) )
						{
							continue;
						}

						newVertex = u;
						newVertexFound = true;
					}
				}


				if ( !newEdgeFound && !newVertexFound )
				{
					// no unmatched edges in pattern -> pattern instance found
					// ...
					detectedMatches->push_back ( state );
					continue;
				}

				// const SRGEdgeProperties& pData = p.m_Graph.m_SRG[ newEdge ];

				// find suitable edges in srg

				// iterate over all edges in srg
				if ( newEdgeFound )
				{
					for (std::map< std::string, SRGraph::EdgePtr >::iterator it = srg.m_Edges.begin();
						 it != srg.m_Edges.end(); ++it )
					{
						const SRGraph::EdgePtr e = it->second;

						if ( state.isSrgEdgeMatched( e ) )
							continue;

						if ( ! state.isEdgeMatchingCompatible ( newEdge, e ) )
							continue;

						if ( ! isEdgeCompatible ( newEdge, e ) )
						{
							continue;
						}

						// we found a new matching edge
						// push refined patternInstance onto stack
						EdgeMatching< UTQLSubgraph, SRGraph > newState ( state );
						newState.addMatchedEdge ( newEdge, e );
						stack.push ( newState );
					}
				}

				if ( newVertexFound )
				{
					for (std::map< std::string, SRGraph::NodePtr >::iterator it = srg.m_Nodes.begin();
						 it != srg.m_Nodes.end(); ++it )
					{
						const SRGraph::NodePtr v = it->second;

						// since u(=newVertex) is not matched via any edge in the pattern, it has to be
						// an isolated vertex in the pattern.
						// since u is isolated it should also only be matched to an vertex v
						// in the srg that is not incident to any edge matched by the pattern
						if ( state.isSrgVertexMatched( v ) )
							continue;

						if ( ! state.isVertexMatchingCompatible ( newVertex, v ) )
							continue;

						if ( ! isVertexCompatible ( newVertex, v ) )
						{
							continue;
						}

						// we found a new matching vertex
						// push refined patternInstance onto stack
						EdgeMatching< UTQLSubgraph, SRGraph > newState ( state );
						newState.addMatchedVertex ( newVertex, v );
						stack.push ( newState );
					}
				}
			}
			return detectedMatches;
		}

	};

}}

#endif // __Ubitrack_Graph_Pattern_INCLUDED__
