/*
 * Ubitrack - Library for Ubiquitous Tracking
 * Copyright 2006, Technische Universitaet Muenchen, and individual
 * contributors as indicated by the @authors tag. See the 
 * copyright.txt in the distribution for a full listing of individual
 * contributors.
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA, or see the FSF site: http://www.fsf.org.
 */

/**
 * @file
 * A vector with an associated covariance matrix.
 *
 * @author Daniel Pustka <daniel.pustka@in.tum.de>
 */ 

#ifndef __UBITRACK_ERROR_ERRORVECTOR_H_INCLUDED__
#define __UBITRACK_ERROR_ERRORVECTOR_H_INCLUDED__
 
#include <boost/numeric/ublas/vector.hpp>
#include <boost/numeric/ublas/matrix.hpp>
#include <boost/numeric/ublas/vector_proxy.hpp>
#include <boost/numeric/ublas/matrix_proxy.hpp>
#include <boost/numeric/bindings/blas/blas3.hpp>
#include <boost/numeric/bindings/traits/ublas_matrix.hpp>

#include <Ubitrack/Math/Vector.h>
#include <Ubitrack/Math/Matrix.h>

namespace Ubitrack { namespace Math {

/**
 * This class stores an N-vector with an associated covariance matrix 
 * and provides methods for transformations, etc.
 *
 * @param N size of the vector
 * @param VType type of vector/matrix elements
 */
template< int N, class VType = double >
struct ErrorVector
{
	/** default constructor */
	ErrorVector()
	{}

	/** construct from value and covariance */
	template< class VT, class MT >
	ErrorVector( const VT& _value, const MT& _covariance )
		: value( _value )
		, covariance( _covariance )
	{}

	/** vector contents */
	Math::Vector< N, VType > value;

	/** covariance matrix of \c value */
	Math::Matrix< N, N, VType > covariance;
	
	/**
	 * (un-)serialization helper function
	 */
	template< class Archive > 
	void serialize( Archive& ar, const unsigned int version )
	{
		ar & value;
		ar & covariance;
	}
};

} } // namespace Ubitrack::Error

#endif
