/*
 * Ubitrack - Library for Ubiquitous Tracking
 * Copyright 2006, Technische Universitaet Muenchen, and individual
 * contributors as indicated by the @authors tag. See the 
 * copyright.txt in the distribution for a full listing of individual
 * contributors.
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA, or see the FSF site: http://www.fsf.org.
 */

/**
 * @file
 * Provides functions to read and write calibration files.
 *
 * @author Daniel Pustka <daniel.pustka@in.tum.de>
 */
 
#ifndef __UBITRACK_UTIL_CALIBFILE_H_INCLUDED__
#define __UBITRACK_UTIL_CALIBFILE_H_INCLUDED__

#include <string>
#include <fstream>
#include <boost/archive/text_iarchive.hpp>
#include <boost/archive/text_oarchive.hpp>
#include <Ubitrack/Util/Exception.h>
#include <Ubitrack/Measurement/Measurement.h>

namespace Ubitrack { namespace Util {

/** 
 * read data from a calibration file.
 * @param sFile filename of file to open
 * @param result where the result will be stored
 */
template< typename T >
void readCalibFile( const std::string& sFile, T& result )
{
	// create ifstream
	std::ifstream stream( sFile.c_str() );
	if ( !stream.good() )
		UBITRACK_THROW( "Could not open file " + sFile + " for reading" );

	// create iarchive
	boost::archive::text_iarchive archive( stream );

	// read data
	archive >> result;
}

/** 
 * read data from a calibration file (specialized for Measurements).
 * @param sFile filename of file to open
 * @param result where the result will be stored
 */
template< typename T >
void readCalibFile( const std::string& sFile, Measurement::Measurement< T >& result )
{
	// initialize measurement
	result = Measurement::Measurement< T >( 0, boost::shared_ptr< T >( new T() ) );

	// create ifstream
	std::ifstream stream( sFile.c_str() );
	if ( !stream.good() )
		UBITRACK_THROW( "Could not open file " + sFile + " for reading" );

	// create iarchive
	boost::archive::text_iarchive archive( stream );

	// read data
	archive >> result;
}

/** 
 * write data to a calibration file.
 * @param sFile filename of file to open
 * @param data data to save
 */
template< typename T >
void writeCalibFile( const std::string& sFile, const T& data )
{
	// create ofstream
	std::ofstream stream( sFile.c_str() );
	if ( !stream.good() )
		UBITRACK_THROW( "Could not open file " + sFile + " for writing" );

	// create iarchive
	boost::archive::text_oarchive archive( stream );

	// read data
	archive << data;
}

} } // namespace Ubitrack::Util

#endif

