/*
 * Ubitrack - Library for Ubiquitous Tracking
 * Copyright 2006, Technische Universitaet Muenchen, and individual
 * contributors as indicated by the @authors tag. See the 
 * copyright.txt in the distribution for a full listing of individual
 * contributors.
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA, or see the FSF site: http://www.fsf.org.
 */

/**
 * @ingroup vision_components
 * @file
 * A component that reads images from a sequence of files.
 *
 * @author Benjamin Becker <benjamin.becker@eads.net>
 */

#include <string>
#include <list>
#include <iostream>

#include <boost/scoped_ptr.hpp>
#include <log4cpp/Category.hh>

#include <Ubitrack/Dataflow/PushSupplier.h>
#include <Ubitrack/Dataflow/Component.h>
#include <Ubitrack/Dataflow/ComponentFactory.h>
#include <Ubitrack/Measurement/Measurement.h>
#include <Ubitrack/Measurement/TimestampSync.h>
#include <Ubitrack/Util/OS.h>
#include <Ubitrack/Util/Exception.h>
#include <Vision/Image.h>

#include <FGCamera.h>
#include <convert.h>
#include <cv.h>
#include <Ubitrack/Util/CalibFile.h>



//<--------------------------------------------->//

using namespace Ubitrack;
using namespace Ubitrack::Vision;

static log4cpp::Category& logger( log4cpp::Category::getInstance( "Ubitrack.Vision.AVTFrameGrabber" ) );

namespace Ubitrack { namespace Drivers {

/**
 * @ingroup vision_components
 * Opens an AVT Camera for capturing.
 *
 * @par Input Ports
 * None.
 * 
 * @par Output Ports
 * PushSupplier< Ubitrack::Vision::ImageMeasurement > port with name "Output"
 *
 * @par Configuration
 * The configuration looks like this:
 * \verbatim 
 * <Configuration obsolete right now! />
 * \endverbatim
 */




namespace AVTCallback {

typedef struct
{
	CFGCamera* pcamera;
	unsigned long Result;
	FGFRAME* Frame;
	int width;
	int height;
	Dataflow::PushSupplier< ImageMeasurement >* pm_colorPort;
	Dataflow::PushSupplier< ImageMeasurement >* pm_outPort;
	Measurement::TimestampSync* syncer;
//	Image* tempGrayScale;
//	Image* m_pMapX;
//	Image* m_pMapY;
} scontext;

void ConvertRawY8RGGB(UINT32 XSize,UINT32 YSize,UINT8 *pBuf,UINT8 *pBGR)
{
  UINT8 *pR,*pB,*pG0,*pG1;
  int i,j;
  struct pdst
  {
    UINT8 B;
    UINT8 G;
    UINT8 R;
  }*pDst;

  pDst=(struct pdst*)pBGR;

  for(i=YSize-2;i>=0;i--) // not flip y : for(i=0;i<YSize-1;i++)
  {
    if(i&1)
    {
      pG1=pBuf+i*XSize;
      pB=pG1+1;
      pR=pG1+XSize;
      pG0=pR+1;
    }
    else
    {
      pR=pBuf+i*XSize;
      pG0=pR+1;
      pG1=pR+XSize;
      pB=pG1+1;
    }

    // Go through all pixels
    for(j=0;j<XSize-1;j++)
    {
      pDst->B=*pB;
      pDst->G=(UINT8)((*pG0+*pG1)/2);
      pDst->R=*pR;

      pDst++;

      if(j&1)
      {
        pB+=2;
        pG0+=2;
      }
      else
      {
        pR+=2;
        pG1+=2;
      }
    }
    pDst->B=0;
    pDst->G=0;
    pDst->R=0;

    pDst++;
  }
  memset(pBGR+(XSize*(YSize-1))*3,0,XSize*3);
}


void ConvertRawY8InterpolateGreen(UINT32 XSize,UINT32 YSize,UINT8 *pBuf,UINT8 *pBGR)
{
	memset(pBGR,0,XSize); // bottom elements are black (not handled yet)
	pBGR += XSize;


	for(int i=YSize-2;i>0;i--) // not flip y : for(int i=1;i<YSize-1;i++)
	{
		UINT8* r_this = pBuf + i * XSize;
		UINT8* r_low  = r_this - XSize;
		UINT8* r_high = r_this + XSize;

		memset(pBGR,0,1); // first elements are black (not handled yet)
		pBGR++;

		for(int j=1;j<XSize-1;j++)
		{
			if( (i^j)&1 )	// green ccd sensor
			{
				pBGR[0] = r_this[1];
			}
			else			// interpolate red+blue ones
			{
				unsigned short horr = abs(short(r_low[1])-short(r_high[1]));
				unsigned short vert = abs(short(r_this[0])-short(r_this[2]));
/*				if( abs(horr-vert) < 8)
				{
					unsigned short temp = r_low[1];
					temp += r_high[1];
					temp += r_this[0];
					temp += r_this[2];
					pBGR[0] = UINT8(temp/4);
					if( temp & 2 )
						pBGR[0]++;
				}
				else
*/				if( horr < vert)
				{
					pBGR[0] = r_low[1]/2+r_high[1]/2 + (r_low[1]&1 + r_high[1]&1)/2;
				}
				else
				{
					pBGR[0] = r_this[0]/2+r_this[2]/2 + (r_this[0]&1 + r_this[2]&1)/2;
				}
			}
			pBGR++;
			r_this++;
			r_low++;
			r_high++;
		}
		memset(pBGR,0,1); // last elements are black (not handled yet)
		pBGR++;
	}
	memset(pBGR,0,XSize); // bottom elements are black (not handled yet)
}

void Raw2Undistorted(UINT32 XSize,UINT32 YSize,UINT8 *pBuf,UINT8 *pBGR)
{
	memset(pBGR,0,XSize); // bottom elements are black (not handled yet)
	pBGR += XSize;


	for(int i=YSize-2;i>0;i--) // not flip y : for(int i=1;i<YSize-1;i++)
	{
		UINT8* r_this = pBuf + i * XSize;
		UINT8* r_low  = r_this - XSize;
		UINT8* r_high = r_this + XSize;

		memset(pBGR,0,1); // first elements are black (not handled yet)
		pBGR++;

		for(int j=1;j<XSize-1;j++)
		{
			if( (i^j)&1 )	// green ccd sensor
			{
				pBGR[0] = r_this[1];
			}
			else			// interpolate red+blue ones
			{
				unsigned short horr = abs(short(r_low[1])-short(r_high[1]));
				unsigned short vert = abs(short(r_this[0])-short(r_this[2]));
/*				if( abs(horr-vert) < 8)
				{
					unsigned short temp = r_low[1];
					temp += r_high[1];
					temp += r_this[0];
					temp += r_this[2];
					pBGR[0] = UINT8(temp/4);
					if( temp & 2 )
						pBGR[0]++;
				}
				else
*/				if( horr < vert)
				{
					pBGR[0] = r_low[1]/2+r_high[1]/2 + (r_low[1]&1 + r_high[1]&1)/2;
				}
				else
				{
					pBGR[0] = r_this[0]/2+r_this[2]/2 + (r_this[0]&1 + r_this[2]&1)/2;
				}
			}
			pBGR++;
			r_this++;
			r_low++;
			r_high++;
		}
		memset(pBGR,0,1); // last elements are black (not handled yet)
		pBGR++;
	}
	memset(pBGR,0,XSize); // bottom elements are black (not handled yet)
} 


void LINKCONVENTION callback(void* vcontext, UINT32 wParam, void* lParam)
{
	Measurement::Timestamp time = Measurement::now();
	scontext* context = (scontext*)vcontext;

	if(wParam == WPARAM_FRAMESREADY )
	{
		do
		{
			context->Result=context->pcamera->GetFrame(context->Frame,0);
			unsigned long int frametime = context->Frame->RxTime.High * 4294967295 + context->Frame->RxTime.Low;
//			std::cerr << "FT: "<<frametime<<std::endl;
			time = context->syncer->convertNativeToLocal(double(frametime), time );
//			std::cerr << "TT: "<<Measurement::timestampToShortString(time)<<std::endl;

			// ... processs image data in Frame.pData
			if(context->Result==FCE_NOERROR)
			{
				if(context->Frame->pData)
				{
					if(context->pm_colorPort->isConnected())
					{
						boost::shared_ptr< Image > pImage( new Image( context->width, context->height, 3 ) );
						ConvertRawY8RGGB(context->width,context->height,context->Frame->pData,(unsigned char*)pImage->imageData);
						pImage->origin=1;
						context->pm_colorPort->send( ImageMeasurement( time - 20000000LL, pImage ) );	// 20ms later
					}
					if(context->pm_outPort->isConnected())
					{
										
/*						ConvertRawY8InterpolateGreen(context->width,context->height,context->Frame->pData,(unsigned char*)context->tempGrayScale->imageData);
						boost::shared_ptr< Image > pUndistorted( new Image( context->width, context->height, 1 ) );	
						pUndistorted->origin=context->tempGrayScale->origin;
						cvRemap( context->tempGrayScale, *pUndistorted, context->m_pMapX, context->m_pMapY);
						context->pm_outPort->send( ImageMeasurement( time - 20000000LL, pUndistorted ) ); // 20ms later
*/
						boost::shared_ptr< Image > pUndistorted( new Image( context->width, context->height, 1 ) );	
						ConvertRawY8InterpolateGreen(context->width,context->height,context->Frame->pData,(unsigned char*)pUndistorted->imageData);
						pUndistorted->origin = 1;
						context->pm_outPort->send( ImageMeasurement( time - 20000000LL, pUndistorted ) ); // 20ms later
					}
				}
			}
			// Return frame to module
			if(context->Result==FCE_NOERROR)
			{
				context->Result=context->pcamera->PutFrame(context->Frame);
			}
		} while(!(context->Frame->Flags&FGF_LAST));
	}
} } // namespace Ubitrack::Driver::AVTCallback


class AVTFrameGrabber
	: public Dataflow::Component
{
public:

	/** constructor */
	AVTFrameGrabber( const std::string& sName, boost::shared_ptr< Graph::UTQLSubgraph > pCfg  );

	/** destructor, waits until thread stops */
	~AVTFrameGrabber();

	/** Component start method. starts the thread */
	virtual void start();

	/** Component stop method, stops thread */
	virtual void stop();
	
protected:
	// callback data
	CFGCamera* Camera;
	AVTCallback::scontext* callbackcontext;

//	Math::Matrix< 3, 3 > m_intrinsicMatrix;
//	Math::Vector< 4 > m_distortionCoefficients;

	// undistortion map
//	Image* m_pMapX;
//	Image* m_pMapY;

//	void initMap( bool bOrigin, int width, int height, const Math::Vector< 4 >& coeffs, const Math::Matrix< 3, 3 >& intrinsics );

	int m_Speed;
	int m_PacketSize;

	// stop the thread?
	volatile bool m_bStop;

	// the ports
	Dataflow::PushSupplier< ImageMeasurement > m_outPort;
	Dataflow::PushSupplier< ImageMeasurement > m_colorPort;
};

AVTFrameGrabber::AVTFrameGrabber( const std::string& sName, boost::shared_ptr< Graph::UTQLSubgraph > pCfg )
	: Dataflow::Component( sName )
	, m_bStop( true )
	, m_outPort( "Output", *this )
	, m_colorPort( "ColorOutput", *this )
{
//TODO: define xml config for the component
	// load intrinsic matrix

/*	if ( pCfg->m_DataflowAttributes.hasAttribute( "intrinsicMatrixFile" ) )
	{
		Measurement::Matrix3x3 measMat;
		measMat.reset( new Math::Matrix< 3, 3 >() );
		Util::readCalibFile( pCfg->m_DataflowAttributes.getAttributeString( "intrinsicMatrixFile" ), measMat );
		m_intrinsicMatrix = *measMat;
	}
	else
		UBITRACK_THROW( "AVT Frame Grabber has no 'intrinsicMatrixFile' defined." );


	// load distortion file
	if ( pCfg->m_DataflowAttributes.hasAttribute( "distortionFile" ) )
	{
		Measurement::Vector4D measVec;
		measVec.reset( new Math::Vector< 4 >() );
		Util::readCalibFile( pCfg->m_DataflowAttributes.getAttributeString( "distortionFile" ), measVec );
		m_distortionCoefficients = *measVec;
	}
	else
		UBITRACK_THROW( "AVT Frame Grabber has no 'distortionFile' defined." );
*/
	int m_width = 780;
	int m_height = 582;

//	this->initMap(1,m_width,m_height,m_distortionCoefficients,m_intrinsicMatrix);


	m_Speed = m_PacketSize = 0;

	int p_Speed;
	pCfg->m_DataflowAttributes.getAttributeData( "BusSpeed", p_Speed );
	m_Speed = p_Speed / 100;
	
	if(m_Speed == 0)
	{
		m_Speed = 1;
		m_PacketSize = pow(2.0,p_Speed/10);
		std::cout<<"Psize: " <<m_PacketSize<<std::endl;
	}
	else if(m_Speed == 1)
	{
		m_PacketSize = 1024;
	}
	else if(m_Speed == 2)
	{
		m_PacketSize = 2048;
	}
	else if(m_Speed == 4)
	{
		m_PacketSize = 2936;
	}
	if(m_PacketSize == 0)
	{
		m_Speed = 2;
		m_PacketSize = 2048;
	}

	// using callback instead:
	Camera = new CFGCamera();
	callbackcontext = new AVTCallback::scontext();
	callbackcontext->pcamera = Camera;
	callbackcontext->Frame = new FGFRAME();
	callbackcontext->pm_colorPort = &m_colorPort;
	callbackcontext->pm_outPort = &m_outPort;
	callbackcontext->width = m_width;
	callbackcontext->height = m_height;
	callbackcontext->syncer = new Measurement::TimestampSync( 1e7 ); // timestamps are in 100ns-units
//	callbackcontext->tempGrayScale = new Image( m_width, m_height, 1 );
//	callbackcontext->m_pMapX = m_pMapX;
//	callbackcontext->m_pMapY = m_pMapY;
//	callbackcontext->frameblocker = 0;

//	framewidth=780;
//	frameheight=582;

	FGNODEINFO NodeInfo[3];
	UINT32 NodeCnt;


	FGINIT moduleparam;
	moduleparam.hWnd=NULL;
	moduleparam.Msg=NULL;
	moduleparam.pCallback = &AVTCallback::callback;

	moduleparam.Context = callbackcontext;
	moduleparam.Flags = 0;
	

	// Init module
	callbackcontext->Result=FGInitModule(&moduleparam);

	// Get list of connected nodes
	if(callbackcontext->Result==FCE_NOERROR)
		callbackcontext->Result=FGGetNodeList(NodeInfo,3,&NodeCnt);

	// Connect with node
	if(callbackcontext->Result==FCE_NOERROR && NodeCnt)
		callbackcontext->Result=Camera->Connect(&NodeInfo[0].Guid);

	// Set camera parameter
	if(callbackcontext->Result==FCE_NOERROR)
		callbackcontext->Result=callbackcontext->pcamera->SetParameter(FGP_IMAGEFORMAT,461056);
	if(callbackcontext->Result==FCE_NOERROR)
		callbackcontext->Result=callbackcontext->pcamera->SetParameter(FGP_XSIZE,callbackcontext->width);
    if(callbackcontext->Result==FCE_NOERROR)
		callbackcontext->Result=callbackcontext->pcamera->SetParameter(FGP_YSIZE,callbackcontext->height);
    if(callbackcontext->Result==FCE_NOERROR)
		callbackcontext->Result=callbackcontext->pcamera->SetParameter(FGP_XPOSITION,0);
    if(callbackcontext->Result==FCE_NOERROR)
		callbackcontext->Result=callbackcontext->pcamera->SetParameter(FGP_YPOSITION,0);
    if(callbackcontext->Result==FCE_NOERROR)
		callbackcontext->Result=callbackcontext->pcamera->SetParameter(FGP_PACKETSIZE,m_PacketSize);
    if(callbackcontext->Result==FCE_NOERROR)
		callbackcontext->Result=callbackcontext->pcamera->SetParameter(FGP_PHYSPEED,m_Speed);
	if(callbackcontext->Result==FCE_NOERROR)
		callbackcontext->Result=callbackcontext->pcamera->SetParameter(FGP_TRIGGER,0);
	if(callbackcontext->Result==FCE_NOERROR)
		callbackcontext->Result=callbackcontext->pcamera->SetParameter(FGP_FRAMEBUFFERCOUNT,16);
	if(callbackcontext->Result==FCE_NOERROR)
		callbackcontext->Result=callbackcontext->pcamera->SetParameter(FGP_DMAFLAGS,0);
	if(callbackcontext->Result==FCE_NOERROR)
		callbackcontext->Result=callbackcontext->pcamera->SetParameter(FGP_WHITEBALCB,PVAL_AUTO); // Blue
	if(callbackcontext->Result==FCE_NOERROR)
		callbackcontext->Result=callbackcontext->pcamera->SetParameter(FGP_WHITEBALCR,PVAL_AUTO); // Red
	if(callbackcontext->Result==FCE_NOERROR)
		callbackcontext->Result=callbackcontext->pcamera->SetParameter(FGP_SHUTTER,PVAL_AUTO); // Red
}

void AVTFrameGrabber::stop()
{
	if ( m_running )
	{
		// Stop the device
		Camera->StopDevice();

		// wait some time for processes to finish up.
		Util::sleep( 500 );

		// Close capture
		Camera->CloseCapture();

		m_running = false;
		m_bStop = true;
	}
}

void AVTFrameGrabber::start()
{
	if ( !m_running )
	{
		// Start DMA logic
		if(callbackcontext->Result==FCE_NOERROR)
			callbackcontext->Result=Camera->OpenCapture();

		// Start image device
		if(callbackcontext->Result==FCE_NOERROR)
			callbackcontext->Result=Camera->StartDevice();

		m_running = true;
		m_bStop = false;
	}
}



AVTFrameGrabber::~AVTFrameGrabber()
{
	//cleanup: callbackcontext TODO!
	delete callbackcontext;

	// Exit module
	FGExitModule();
}

/*
void AVTFrameGrabber::initMap( bool bOrigin, int width, int height, const Math::Vector< 4 >& coeffs, const Math::Matrix< 3, 3 >& intrinsics )
{
	// create map images
	m_pMapX = new Image( width, height, 1, IPL_DEPTH_32F );
	m_pMapY = new Image( width, height, 1, IPL_DEPTH_32F );
		
	LOG4CPP_INFO( logger, "Creating undistortion map" );
	LOG4CPP_DEBUG( logger, "coeffs=" << coeffs );
	LOG4CPP_DEBUG( logger, "intrinsic=" << intrinsics );

	// copy ublas to OpenCV parameters
	CvMat* pCvCoeffs = cvCreateMat( 1, 4, CV_32FC1 );
	for ( unsigned i = 0; i < 4; i++ )
		pCvCoeffs->data.fl[ i ] = static_cast< float >( coeffs( i ) );
		
	CvMat* pCvIntrinsics = cvCreateMat( 3, 3, CV_32FC1 );
	for ( unsigned i = 0; i < 3; i++ )
		for ( unsigned j = 0; j < 3; j++ )
			reinterpret_cast< float* >( pCvIntrinsics->data.ptr + i * pCvIntrinsics->step)[ j ] 
				= static_cast< float >( intrinsics( i, j ) ) * ( j == 2 ? -1 : 1 ); // compensate for left-handed OpenCV coordinate frame
	
	// compensate if origin==0
	if ( !bOrigin )
	{
		reinterpret_cast< float* >( pCvIntrinsics->data.ptr + 1 * pCvIntrinsics->step )[ 2 ] = 
			height - 1 - reinterpret_cast< float* >( pCvIntrinsics->data.ptr + 1 * pCvIntrinsics->step )[ 2 ];
		pCvCoeffs->data.fl[ 2 ] *= -1;
	}


	
	cvInitUndistortMap( pCvIntrinsics, pCvCoeffs, *m_pMapX, *m_pMapY );
	
	LOG4CPP_DEBUG( logger, "first pixel mapped from " << 
		*reinterpret_cast< float* >( m_pMapX->imageData ) << ", " <<
		*reinterpret_cast< float* >( m_pMapY->imageData ) );
		
	// release data
	cvReleaseMat( &pCvCoeffs );
	cvReleaseMat( &pCvIntrinsics );
}
*/

} } // namespace Ubitrack::Driver

UBITRACK_REGISTER_COMPONENT( Dataflow::ComponentFactory* const cf ) {
	cf->registerComponent< Ubitrack::Drivers::AVTFrameGrabber > ( "AVTFrameGrabber" );
}


/*

void ConvertRawY8InterpolateRGB(UINT32 XSize,UINT32 YSize,UINT8 *pBuf,UINT8 *pBGR)
{
	memset(pBGR,0,XSize*3); // bottom elements are black (not handled yet)
	pBGR += XSize*3;
	
	for(int i=YSize-1;i>1;i--)
	{
		UINT8* r_this = pBuf + i * XSize;
		UINT8* r_low  = r_this - XSize;
		UINT8* r_high = r_this + XSize;
		
		memset(pBGR,0,3); // first elements are black (not handled yet)
		pBGR+=3;

		for(int j=1;j<XSize-1;j++)
		{
			if( i&1 )
			{
				if( j&1 )	// blue row & blue center
				{
					pBGR[0] = r_this[1];	// blue center
					pBGR[1] = r_this[0]/4 + r_this[2]/4 + r_low[1]/4 + r_high[1]/4; // green up left right down
					pBGR[2] = r_low[0]/4 + r_low[2]/4 + r_high[0]/4 + r_high[2]/4; // red up/left down/left up/right down/right
				}
				else		// blue row & green center	
				{
					pBGR[0] = r_this[0]/2 + r_this[2]/2 ; // blue left right
					pBGR[1] = r_this[1];	// green center
					pBGR[2] = r_low[1]/2 + r_high[1]/2; // red up down
				}
			}
			else
			{
				if( j&1 )	// red row green center
				{
					pBGR[0] = r_low[1]/2 + r_high[1]/2; // blue up down	
					pBGR[1] = r_this[1];	// green center
					pBGR[2] = r_this[0]/2 + r_this[2]/2 ; // red left right
				}
				else		// red row red center	
				{
					pBGR[0] = r_low[0]/4 + r_low[2]/4 + r_high[0]/4 + r_high[2]/4; // blue up/left down/left up/right down/right
					pBGR[1] = r_this[0]/4 + r_this[2]/4 + r_low[1]/4 + r_high[1]/4; // green up left right down
					pBGR[2] = r_this[1];	// red center
				}
			}
			pBGR+=3;
			r_this+=1;
			r_low +=1;
			r_high+=1;
		}
		memset(pBGR,0,3); // last elements are black (not handled yet)
		pBGR+=3;
	}
	memset(pBGR,0,XSize*3); // bottom elements are black (not handled yet)
}

void ConvertRawY8InterpolateMono(UINT32 XSize,UINT32 YSize,UINT8 *pBuf,UINT8 *pBGR)
{
	memset(pBGR,0,XSize); // bottom elements are black (not handled yet)
	pBGR += XSize;
	
	for(int i=1;i<YSize-1;i++)
	{
		UINT8* r_this = pBuf + i * XSize;
		UINT8* r_low  = r_this - XSize;
		UINT8* r_high = r_this + XSize;
		
		memset(pBGR,0,1); // first elements are black (not handled yet)
		pBGR++;

		for(int j=1;j<XSize-1;j++)
		{
			if( i&1 )
			{
				if( j&1 )	// blue row & blue center
				{
					pBGR[0] = (r_this[1])/3 
							+ (r_this[0]/4 + r_this[2]/4 + r_low[1]/4 + r_high[1]/4) / 3
							+ (r_low[0]/4 + r_low[2]/4 + r_high[0]/4 + r_high[2]/4)/3;
				}
				else		// blue row & green center	
				{
					pBGR[0] = (r_this[0]/2 + r_this[2]/2)/3
							+ (r_this[1])/3
							+ (r_low[1]/2 + r_high[1]/2)/3;
				}
			}
			else
			{
				if( j&1 )	// red row green center
				{
					pBGR[0] = (r_low[1]/2 + r_high[1]/2)/3
							+ (r_this[1])/3
							+ (r_this[0]/2 + r_this[2]/2)/3;
				}
				else		// red row red center	
				{
					pBGR[0] = (r_low[0]/4 + r_low[2]/4 + r_high[0]/4 + r_high[2]/4)/3
							+ (r_this[0]/4 + r_this[2]/4 + r_low[1]/4 + r_high[1]/4)/3
							+ (r_this[1])/3;
				}
			}
			pBGR++;
			r_this++;
			r_low++;
			r_high++;
		}
		memset(pBGR,0,1); // last elements are black (not handled yet)
		pBGR++;
	}
	memset(pBGR,0,XSize); // bottom elements are black (not handled yet)
}
*/