/*
 * Ubitrack - Library for Ubiquitous Tracking
 * Copyright 2006, Technische Universitaet Muenchen, and individual
 * contributors as indicated by the @authors tag. See the 
 * copyright.txt in the distribution for a full listing of individual
 * contributors.
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA, or see the FSF site: http://www.fsf.org.
 */


/**
 * @ingroup vision_components
 * @file
 * Finding the edges of a chessboard in an image
 *
 * @author Daniel Muhra <muhra@in.tum.de>
 */
#include <Ubitrack/Dataflow/Component.h>
#include <Ubitrack/Dataflow/PushConsumer.h>
#include <Ubitrack/Dataflow/PushSupplier.h>
#include <Ubitrack/Dataflow/ComponentFactory.h>
#include <Ubitrack/Measurement/Measurement.h>
#include <Ubitrack/Math/Vector.h>
#include <fstream>
#include <sstream>
#include <boost/scoped_array.hpp>

#include <Vision/Image.h>
#include <cv.h>

namespace Ubitrack { namespace Vision {

/**
 * @ingroup vision_components
 * Chessboard detection component.
 * This class contains a chessboard detection implemented as a \c TriggerComponent.
 *
 * @par Input Ports
 * PushConsumer< Ubitrack::Vision::ImageMeasurement > of name "Input", accepts only greyscale images
 *
 * @par Output Ports
 * PushSupplier<Measurement::PositionList2> with name "Output".
 * PushSupplier<ImageMeasurement> with name "DebugImage".
 *
 * @par Configuration
 * \c chessBoardHeight: number of edges the chessboard has (height). 
 * \c chessBoardWidth: number of edges the chessboard has (width).
 *
 * @par Operation
 * The component returns an vector of 2D coordinates which represents the edges of an given chessboard
 *
 */
class ChessBoardDetectionComponent
	: public Dataflow::Component
{
public:
	/**
	 * Standard component constructor.
	 *
	 * @param sName Unique name of the component.
	 * @param cfg ComponentConfiguration containing all configuration.
	 */
	ChessBoardDetectionComponent( const std::string& sName, boost::shared_ptr< Graph::UTQLSubgraph > pCfg )
		: Dataflow::Component( sName )
		, m_height( -1 )
		, m_width( -1 )
		, m_inPort( "Image", *this, boost::bind( &ChessBoardDetectionComponent::pushImage, this, _1 ) )
		, m_outPort( "Output", *this )
		, m_debugPort( "DebugImage", *this )
    {
		// get height and width attributes from ChessBoard node
		Graph::UTQLSubgraph::NodePtr pCbNode = pCfg->getNode( "ChessBoard" );
		pCbNode->getAttributeData( "chessBoardHeight", m_height );
		pCbNode->getAttributeData( "chessBoardWidth", m_width );
		
		if ( m_height < 0 || m_width < 0 )
			UBITRACK_THROW( "ChessBoard nodes has no chessBoardHeight or chessBoardWidth attribute" );
		
		m_edges = m_height * m_width;
    }

	/** Method that computes the result. */
	void pushImage( const ImageMeasurement& img )
	{
		int pattern_was_found;
		boost::scoped_array< CvPoint2D32f > corners( new CvPoint2D32f[ m_edges ] );
		int found;

		pattern_was_found = cvFindChessboardCorners( *img, cvSize(m_width, m_height), corners.get(), &found, CV_CALIB_CB_ADAPTIVE_THRESH );

		if(pattern_was_found && found == m_edges)
		{
			cvFindCornerSubPix( *img, corners.get(), m_edges, cvSize(5,5), cvSize(-1,-1), cvTermCriteria( CV_TERMCRIT_ITER, 10, 0.1f ));

			Measurement::PositionList2 out( img.time(), Measurement::PositionList2::value_type() );
			Ubitrack::Math::Vector< 2 > temp;

			for(int i=0; i < m_edges; i++)
			{
				temp(0) = corners[i].x ;
				temp(1) = img->origin ? corners[i].y : ( img->height - 1 - corners[ i ].y );
				out->push_back(temp);
			}

			m_outPort.send( out );			
		}		

		if( m_debugPort.isConnected() )
		{
			// debug image
			boost::shared_ptr< Image > debugImg;
			debugImg = img->CvtColor( CV_GRAY2RGB, 3 );
			debugImg->origin = img->origin; // it should not flip.....

			const CvScalar color_red = CV_RGB( 255, 0, 0 );
			const CvScalar color_green = CV_RGB( 0, 255, 0 );

				for(int i = 0; i < found; i++) 
				{
					// draw corner point
					cvLine( *debugImg
						, cvPoint((int)corners[i].x - 2, (int)corners[i].y - 2)
						, cvPoint((int)corners[i].x + 2, (int)corners[i].y + 2)
						, color_red, 1 );
					cvLine( *debugImg
						, cvPoint((int)corners[i].x + 2, (int)corners[i].y - 2)
						, cvPoint((int)corners[i].x - 2, (int)corners[i].y + 2)
						, color_red, 1 );

				}

				cvCircle( *debugImg, cvPoint((int)corners[0].x, (int)corners[0].y), 3, color_green, -1 );
				
				for(int i = 1; i < found; i++) 
				{
					cvLine( *debugImg
							, cvPoint((int)corners[i-1].x, (int)corners[i-1].y)
							, cvPoint((int)corners[i].x, (int)corners[i].y)
							, color_green, 1 );
				}

			m_debugPort.send( ImageMeasurement( img.time(), debugImg ) );
		}
    }

protected:
	/** defines the chessboard to be tracked */
	int m_height;
	int m_width;
	int m_edges;
	/** Input port of the component. */
	Dataflow::PushConsumer< Ubitrack::Vision::ImageMeasurement > m_inPort;
	/** Output ports of the component. */
	Dataflow::PushSupplier< Measurement::PositionList2 > m_outPort;
	Dataflow::PushSupplier< ImageMeasurement > m_debugPort;
};


UBITRACK_REGISTER_COMPONENT( Dataflow::ComponentFactory* const cf ) {
	cf->registerComponent< ChessBoardDetectionComponent > ( "ChessBoardDetection" );
}

} } // namespace Ubitrack::Components

