/*
 * Ubitrack - Library for Ubiquitous Tracking
 * Copyright 2006, Technische Universitaet Muenchen, and individual
 * contributors as indicated by the @authors tag. See the 
 * copyright.txt in the distribution for a full listing of individual
 * contributors.
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA, or see the FSF site: http://www.fsf.org.
 */


/**
 * @ingroup vision_components
 * @file
 * Finding the edges of a chessboard in an image
 *
 * @author Daniel Muhra <muhra@in.tum.de>
 */
#include <Ubitrack/Dataflow/Component.h>
#include <Ubitrack/Dataflow/PushConsumer.h>
#include <Ubitrack/Dataflow/PullConsumer.h>
#include <Ubitrack/Dataflow/PushSupplier.h>
#include <Ubitrack/Dataflow/ComponentFactory.h>
#include <Ubitrack/Measurement/Measurement.h>
#include <Ubitrack/Math/Vector.h>
#include <fstream>
#include <sstream>

#include <Vision/Image.h>
#include <cv.h>

namespace Ubitrack { namespace Vision {

/**
 * @ingroup vision_components
 * Chessboard detection component.
 * This class contains a chessboard detection implemented as a \c TriggerComponent.
 *
 * @par Input Ports
 * PushConsumer< Ubitrack::Vision::ImageMeasurement > of name "Input", accepts only greyscale images
 *
 * @par Output Ports
 * PushSupplier<Measurement::PositionList2> with name "Output".
 * PushSupplier<ImageMeasurement> with name "DebugImage".
 *
 * @par Configuration
 * \c chessBoardHeight: number of edges the chessboard has (height). 
 * \c chessBoardWidth: number of edges the chessboard has (width).
 * \c Debug: if set to 1, image output is enabled.
 *
 * @par Operation
 * The component returns an vector of 2D coordinates which represents the edges of an given chessboard
 *
 * @par Example Configuration
 * \verbatim 
<Pattern name="ChessBoardTracking" id="someId">
	<Input>
		<Node name="ChessBoard" id="someId">
			<Attribute name="chessBoardHeight" value="6"/>
			<Attribute name="chessBoardWidth" value="7"/>
			<Attribute name="chessBoardWidth" value="0.025"/>
		</Node>
		<Node name="ImagePlane" id="someId2"/>
		<Node name="Intrinsic" id="someId3"/>
		<Node name="Distortion" id="someId4"/>
		<Edge name="Image" source="ChessBoard" destination="ImagePlane" pattern-ref="..." edge-ref="..."/>
		<Edge name="Intrinsic" source="ImagePlane" destination="Intrinsic" pattern-ref="..." edge-ref="..."/>
		<Edge name="Distortion" source="ImagePlane" destination="Distortion" pattern-ref="..." edge-ref="..."/>
	</Input>
	<Output>
		<Node name="Values" id="someId5"/>
		<Node name="Consumer" id="someId6"/>
		<Edge name="Output" source="Values" destination="Consumer"/>
	</Output>
	<DataflowConfiguration>
		<UbitrackLib class="ChessBoardTracking"/>
	</DataflowConfiguration>
</Pattern> 
\endverbatim
 */
class ChessBoardTrackingComponent
	: public Dataflow::Component
{
public:
	/**
	 * Standard component constructor.
	 *
	 * @param sName Unique name of the component.
	 * @param cfg ComponentConfiguration containing all configuration.
	 */
	ChessBoardTrackingComponent( const std::string& sName, boost::shared_ptr< Graph::UTQLSubgraph > pCfg )
		: Dataflow::Component( sName )
		, m_height( -1 )
		, m_width( -1 )
		, m_inPort( "Image", *this, boost::bind( &ChessBoardTrackingComponent::pushImage, this, _1 ) )
		, m_intrinsicPort( "Intrinsic", *this )
		, m_distortionPort( "Distortion", *this )
		, m_outPort( "Output", *this )
    {
		// get height and width attributes from ChessBoard node
		Graph::UTQLSubgraph::NodePtr pCbNode = pCfg->getNode( "ChessBoard" );
		pCbNode->getAttributeData( "chessBoardHeight", m_height );
		pCbNode->getAttributeData( "chessBoardWidth", m_width );
		pCbNode->getAttributeData( "chessBoardSize", m_size );
		
		if ( m_height < 0 || m_width < 0 )
			UBITRACK_THROW( "ChessBoard nodes has no chessBoardHeight or chessBoardWidth attribute" );
		
		m_edges = m_height * m_width;

		objPoints = new float[3*m_edges];

		for(int i=0; i < m_edges ;i++)
		{
			objPoints[3*i] = float(i % m_height)*m_size;
			objPoints[3*i+1] = float(i / m_height)*m_size;
			objPoints[3*i+2] = 0.0;
		}
    }

	/** returns the camera intrinsics */
	Math::Matrix< 3, 3 > intrinsic( Measurement::Timestamp t )
	{ return *( m_intrinsicPort.get( t ) ); }

	/** returns the camera distortion */
	Math::Vector< 4 > distortion( Measurement::Timestamp t )
	{ return *( m_distortionPort.get( t ) ); }

	/** Method that computes the result. */
	void pushImage( const ImageMeasurement& img )
	{
		int pattern_was_found;
		CvPoint2D32f* corners = new CvPoint2D32f[m_edges];
		int found;

		pattern_was_found = cvFindChessboardCorners( *img, cvSize(m_height, m_width) , corners, &found, CV_CALIB_CB_ADAPTIVE_THRESH );
		cvFindCornerSubPix( *img, corners, m_edges, cvSize(5,5), cvSize(-1,-1), cvTermCriteria( CV_TERMCRIT_ITER, 10, 0.1f ));

		if(pattern_was_found && found == m_edges)
		{
			Measurement::PositionList2 out( img.time(), Measurement::PositionList2::value_type() );
			Ubitrack::Math::Vector< 2 > temp;

			/** copies all values into OpenCV compatible formats */
			float* imgPoints = new float[2*m_edges];
			CvMat image_points = cvMat ( m_edges, 2, CV_32F, imgPoints);
			CvMat object_points = cvMat ( m_edges, 3, CV_32F, objPoints);

			for(int i=0; i < m_edges ;i++)
			{
				imgPoints[2*i] = corners[i].x;
				imgPoints[2*i+1] = img->origin ? corners[i].y : ( img->height - 1 - corners[ i ].y );
			}

			float intrVal[9];
			CvMat intrinsic_matrix = cvMat ( 3, 3, CV_32FC1, intrVal);

			float disVal[4];
			CvMat distortion_coeffs = cvMat( 4, 1, CV_32FC1, disVal);

			disVal[0] = ( float )( distortion( img.time() ))(0);
			disVal[1] = ( float )( distortion( img.time() ))(1);
			disVal[2] = ( float )( distortion( img.time() ))(2);
			disVal[3] = ( float )( distortion( img.time() ))(3);
			
			// compensate for left-handed OpenCV coordinate frame
			intrVal[0] = ( float )( intrinsic( img.time() ))(0, 0);
			intrVal[1] = ( float )( intrinsic( img.time() ))(0, 1);
			intrVal[2] = ( float )( -1.0 )*( intrinsic( img.time() ))(0, 2);
			intrVal[3] = ( float )( intrinsic( img.time() ))(1, 0);
			intrVal[4] = ( float )( intrinsic( img.time() ))(1, 1);
			intrVal[5] = ( float )( -1.0 )*( intrinsic( img.time() ))(1, 2);
			intrVal[6] = ( float )( intrinsic( img.time() ))(2, 0);
			intrVal[7] = ( float )( intrinsic( img.time() ))(2, 1);
			intrVal[8] = ( float )( -1.0 )*( intrinsic( img.time() ))(2, 2);

			// compensate if origin==0
			if ( !(*img).origin )
			{
				intrVal[5] = (*img).height - 1 - intrVal[5];
				disVal[2] *= -1.0;
			}


			/** rotation and translation vectors for OpenCV  output */
			CvMat* current_rot_vec = cvCreateMat( 1, 3, CV_32FC1 );
			
			float m_current_translation_vector[3];	
			CvMat current_trans_vec = cvMat (3, 1, CV_32FC1, m_current_translation_vector);

			/** actual computation */
			cvFindExtrinsicCameraParams2(
						&object_points,
						&image_points,
						&intrinsic_matrix,
						&distortion_coeffs,
						current_rot_vec,
						&current_trans_vec);


			/** convert rotation vector into a matrix */
			float m_current_rotation_matrix[9];
			CvMat current_rot_mat = cvMat (3, 3, CV_32FC1, m_current_rotation_matrix);
			
			cvRodrigues2(current_rot_vec , &current_rot_mat);

			/** copying into Ubitrack formats */
			double rot_mat[9];

			for( int i=0; i<9; i++)
			{
				rot_mat[i] = double( m_current_rotation_matrix[i] ); 
			}

			/** conversion into Pose and right-hand-coordinate-system*/
			Math::Vector< 3 > trans( m_current_translation_vector[0], m_current_translation_vector[1], -m_current_translation_vector[2] );
			Math::Matrix< 3, 3 > rot( rot_mat );
			
			rot( 2, 0 ) = - rot( 2, 0 );
			rot( 2, 1 ) = - rot( 2, 1 );
			rot( 0, 2 ) = - rot( 0, 2 );
			rot( 1, 2 ) = - rot( 1, 2 );

			// release data
			cvReleaseMat( &current_rot_vec );

			m_outPort.send( Measurement::Pose( img.time(), Math::Pose( Math::Quaternion(rot), trans ) ) );			
		}		
    }

protected:
	/** defines the chessboard to be tracked */
	int m_height;
	int m_width;
	float m_size;
	int m_edges;
	float* objPoints;
	/** Input port of the component. */
	Dataflow::PushConsumer< Ubitrack::Vision::ImageMeasurement > m_inPort;
	Dataflow::PullConsumer< Measurement::Matrix3x3 > m_intrinsicPort;
	Dataflow::PullConsumer< Measurement::Vector4D > m_distortionPort;
	/** Output ports of the component. */
	Dataflow::PushSupplier< Measurement::Pose > m_outPort;
};


UBITRACK_REGISTER_COMPONENT( Dataflow::ComponentFactory* const cf ) {
	cf->registerComponent< ChessBoardTrackingComponent > ( "ChessBoardTracking" );
}

} } // namespace Ubitrack::Components

