/*
 * Ubitrack - Library for Ubiquitous Tracking
 * Copyright 2006, Technische Universitaet Muenchen, and individual
 * contributors as indicated by the @authors tag. See the
 * copyright.txt in the distribution for a full listing of individual
 * contributors.
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA, or see the FSF site: http://www.fsf.org.
 */

/**
 * @ingroup vision_components
 * @file
 * Reads camera images using Thomas Pintaric's DSVideoLib library.
 * http://sourceforge.net/projects/dsvideolib
 *
 * @author Daniel Pustka <daniel.pustka@in.tum.de>
 */

#include <string>
#include <list>
#include <iostream>
#include <strstream>
#include <log4cpp/Category.hh>

#include <boost/thread.hpp>
#include <boost/bind.hpp>
#include <boost/scoped_ptr.hpp>

#include <Ubitrack/Dataflow/PushSupplier.h>
#include <Ubitrack/Dataflow/Component.h>
#include <Ubitrack/Dataflow/ComponentFactory.h>
#include <Ubitrack/Measurement/Measurement.h>
#include <Ubitrack/Measurement/TimestampSync.h>
#include <Vision/Image.h>

#include <Ubitrack/Util/CleanWindows.h>
#ifndef _WIN32_WINNT
	#define _WIN32_WINNT WINVER
#endif
#include <objbase.h>
#include <DSVL.h>
#include <highgui.h>

// get a logger
static log4cpp::Category& logger( log4cpp::Category::getInstance( "Ubitrack.Vision.DSVideoLibFrameGrabber" ) );

using namespace Ubitrack;
using namespace Ubitrack::Vision;

namespace Ubitrack { namespace Drivers {

/**
 * @ingroup vision_components
 * Reads camera images using Thomas Pintaric's DSVideoLib library.
 * http://sourceforge.net/projects/dsvideolib
 *
 * @par Input Ports
 * None.
 *
 * @par Output Ports
 * \c Output push port of type Ubitrack::Vision::ImageMeasurement.
 *
 * @par Configuration
 * The configuration tag contains a \c <dsvl_input> configuration.
 * For details, see the DSVideoLib documentation...
 *
 * Examples:
\verbatim
<DataflowConfiguration>
  <Attribute name="DSVL">
  <dsvl_input>
	<camera show_format_dialog="true">
		<pixel_format>
			<RGB24/>
		</pixel_format>
	</camera>
  </dsvl_input>
  </Attribute>
</DataflowConfiguration>
\endverbatim
 *
\verbatim
<DataflowConfiguration>
  <Attribute name="DSVL">
  <dsvl_input>
	<camera frame_width="320" frame_height="200" frame_rate="15">
		<pixel_format>
			<RGB24/>
		</pixel_format>
	</camera>
  </dsvl_input>
  </Attribute>
</DataflowConfiguration>
\endverbatim
 *
 * The DSVideoLib can also read video files:
<DataflowConfiguration>
  <Attribute name="DSVL">
  <dsvl_input>
	<avi_file use_reference_clock="true" file_name="..\media\cube_left.avi" loop_avi="true" render_secondary="true">
		<pixel_format>
			<RGB24/>
		</pixel_format>
	</avi_file>
  </dsvl_input>
  </Attribute>
</DataflowConfiguration>
 */
class DSVideoLibFrameGrabber
	: public Dataflow::Component
{
public:

	/** constructor */
	DSVideoLibFrameGrabber( const std::string& sName, boost::shared_ptr< Graph::UTQLSubgraph >  );

	/** destructor, waits until thread stops */
	~DSVideoLibFrameGrabber();

protected:
	// width of resulting image
	LONG m_width;

	// height of resulting image
	LONG m_height;

	// shift timestamps (ms)
	int m_timeOffset;

	// only send every nth image
	int m_divisor;

	// pointer to DSVideoLib VideoSource
	boost::scoped_ptr< DSVL_VideoSource > m_pVideoSource;

	// thread main loop
	void ThreadProc();

	// the thread
	boost::scoped_ptr< boost::thread > m_Thread;

	// stop the thread?
	volatile bool m_bStop;

	// the ports
	Dataflow::PushSupplier< ImageMeasurement > m_outPort;
	Dataflow::PushSupplier< ImageMeasurement > m_colorOutPort;
};


DSVideoLibFrameGrabber::DSVideoLibFrameGrabber( const std::string& sName, boost::shared_ptr< Graph::UTQLSubgraph > subgraph )
	: Dataflow::Component( sName )
	, m_timeOffset( 0 )
	, m_divisor( 1 )
	, m_bStop( false )
	, m_outPort( "Output", *this )
	, m_colorOutPort( "ColorOutput", *this )
{
	subgraph->m_DataflowAttributes.getAttributeData( "timeOffset", m_timeOffset );
	subgraph->m_DataflowAttributes.getAttributeData( "divisor", m_divisor );

	const std::string& sConfig( subgraph->m_DataflowAttributes.getAttribute( "DSVL" ).getText() );
	LOG4CPP_DEBUG( logger, "DSVideoLibConfiguration: >>" << sConfig << "<<" );

	if ( FAILED( CoInitializeEx( NULL, COINIT_MULTITHREADED ) ) )
		UBITRACK_THROW( "Error in CoInitializeEx" );
	m_pVideoSource.reset( new DSVL_VideoSource() );

	if ( FAILED( m_pVideoSource->BuildGraphFromXMLString( const_cast< char* >( sConfig.c_str() ) ) ) )
		UBITRACK_THROW( "Invalid DSVideoLib configuration string: " + sConfig );

	PIXELFORMAT pxf;
	double cap_fps;
	if ( FAILED( m_pVideoSource->GetCurrentMediaFormat( &m_width, &m_height, &cap_fps, &pxf ) ) )
		UBITRACK_THROW( "DSVideoLib initialization problem" );
	LOG4CPP_DEBUG( logger, "Frame dimensions: " << m_width << "x" << m_height )

	if ( pxf != PIXELFORMAT_RGB24 )
		UBITRACK_THROW( "DSVideoLib driver only supports RGB24 format" );

	if ( FAILED( m_pVideoSource->EnableMemoryBuffer( 3 ) ) )
		UBITRACK_THROW( "DSVideoLib EnableMemoryBuffer failed" );

	// start thread
	m_Thread.reset( new boost::thread( boost::bind( &DSVideoLibFrameGrabber::ThreadProc, this ) ) );
}


DSVideoLibFrameGrabber::~DSVideoLibFrameGrabber()
{
	if ( m_Thread )
	{
		m_bStop = true;
		m_Thread->join();
	}
}


void DSVideoLibFrameGrabber::ThreadProc()
{
	LOG4CPP_DEBUG( logger, "Thread started" );

	if ( FAILED( m_pVideoSource->Run() ) )
	{
		LOG4CPP_ERROR( logger, "DSVideoLib failed to Run()" );
		return;
	}

	Measurement::TimestampSync syncer( 1e7 ); // timestamps are in 100ns-units
	bool bPrintedTimestampWarning = false;
	LONGLONG lastTimestamp = 13;
	unsigned lastFrame = 29645234;
	unsigned nFrames = 0;

	while ( !m_bStop )
	{
		MemoryBufferHandle l_mbHandle;
		BYTE* l_pPixelBuffer = NULL;

		m_pVideoSource->WaitForNextSample( 250 );
		Measurement::Timestamp time = Measurement::now();

		if ( SUCCEEDED( m_pVideoSource->CheckoutMemoryBuffer( &l_mbHandle, &l_pPixelBuffer ) ) )
		{
			LOG4CPP_DEBUG( logger, "CheckoutMemoryBuffer successful: (" << l_mbHandle.n << ", " << l_mbHandle.t << ")" );

			if ( l_mbHandle.n == lastFrame || !m_running )
			{
				// DSVideoLib sometimes sends frames multiple times, especially, when using more than one camera...
				LOG4CPP_DEBUG( logger, "Not sending repeated frame or not yet started" );
			}
			else
			{
				lastFrame = l_mbHandle.n;

				// create timestamp
				if ( l_mbHandle.t <= lastTimestamp )
				{
					if ( !bPrintedTimestampWarning )
					{
						LOG4CPP_WARN( logger, "DSVideoLib timestamps do not work. Try a patched version." );
						bPrintedTimestampWarning = true;
					}
				}
				else
				{
					time = syncer.convertNativeToLocal( double( l_mbHandle.t ), time );
					lastTimestamp = l_mbHandle.t;
				}

				// create an IplImage buffer for the frame
				Vision::Image camBuffer( m_width, m_height, 3, static_cast< void* >( l_pPixelBuffer ), IPL_DEPTH_8U, 1 );
				
				// send color image if anyone is interested
				if ( !( nFrames % m_divisor ) && m_colorOutPort.isConnected() )
					m_colorOutPort.send( ImageMeasurement( time + m_timeOffset * 1000000LL, camBuffer.Clone() ) );

				// convert to greyscale
				boost::shared_ptr< Image > pImage( new Image( camBuffer.width, camBuffer.height, 1 ) );
				cvConvertImage( camBuffer, *pImage, camBuffer.origin ? CV_CVTIMG_FLIP : 0 );

				// send image
				if ( !( nFrames++ % m_divisor ) )
					m_outPort.send( ImageMeasurement( time + m_timeOffset * 1000000LL, pImage ) );
			}

			// release buffer
			m_pVideoSource->CheckinMemoryBuffer( l_mbHandle );
		}
		else
			LOG4CPP_DEBUG( logger, "CheckoutMemoryBuffer failed" );
	}

	m_pVideoSource->Stop();

	LOG4CPP_DEBUG( logger, "Thread stopped" );
}

} } // namespace Ubitrack::Driver

UBITRACK_REGISTER_COMPONENT( Dataflow::ComponentFactory* const cf ) {
	cf->registerComponent< Ubitrack::Drivers::DSVideoLibFrameGrabber > ( "DSVideoLibFrameGrabber" );
}

