/*
 * Ubitrack - Library for Ubiquitous Tracking
 * Copyright 2006, Technische Universitaet Muenchen, and individual
 * contributors as indicated by the @authors tag. See the
 * copyright.txt in the distribution for a full listing of individual
 * contributors.
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA, or see the FSF site: http://www.fsf.org.
 */

// TODO: implement module start/stop mechanics

/**
 * @ingroup vision_components
 * @file
 * Implements a window to display images, based on HighGUI.
 *
 * @author Daniel Pustka <daniel.pustka@in.tum.de>
 */

#include <string>
#include <iostream>
#include <deque>

#include <boost/thread.hpp>
#include <boost/thread/xtime.hpp>
#include <boost/function.hpp>
#include <boost/bind.hpp>
#include <boost/scoped_ptr.hpp>
#include <log4cpp/Category.hh>

#include <Ubitrack/Dataflow/PushConsumer.h>
#include <Ubitrack/Dataflow/PushSupplier.h>
#include <Ubitrack/Dataflow/Module.h>
#include <Ubitrack/Dataflow/ComponentFactory.h>
#include <Ubitrack/Measurement/Measurement.h>
#include <Ubitrack/Util/OS.h>
#include <Vision/Image.h>

#include <highgui.h>

static log4cpp::Category& logger( log4cpp::Category::getInstance( "Ubitrack.Vision.HighguiWindow" ) );

using namespace Ubitrack;
using namespace Ubitrack::Vision;
using namespace Ubitrack::Dataflow;

namespace Ubitrack { namespace Drivers {

/** the component key -- reads the window title */
class HWCKey
	: public std::string
{
public:
	/** extract window title from configuration */
	HWCKey( boost::shared_ptr< Graph::UTQLSubgraph > subgraph )
	{
		std::string sTitle = subgraph->m_DataflowAttributes.getAttribute( "windowTitle" ).getText();
		if ( sTitle.length() == 0 )
			UBITRACK_THROW( "Configuration for HighguiWindow component has empty 'title' attribute" );
		*static_cast< std::string* >( this ) = sTitle;
	}
};

// forward declarations
class HighguiWindow;
class HighguiWindowModule;

typedef Module< SingleModuleKey, HWCKey, HighguiWindowModule, HighguiWindow > ModuleBase;

/**
 * The module -- only responsible for a thread that calls cvGetKey every 100ms
 */
class HighguiWindowModule
	: public ModuleBase
{
public:
	/** constructor, creates thread */
	HighguiWindowModule( const SingleModuleKey& key, boost::shared_ptr< Graph::UTQLSubgraph >, FactoryHelper* pFactory )
		: ModuleBase( key, pFactory )
		, m_bStop( false )
		, m_nWindows( 0 )
	{
		m_Thread.reset( new boost::thread( boost::bind( &HighguiWindowModule::threadProc, this ) ) );
	}

	/** destructor, stops thread */
	~HighguiWindowModule()
	{
		m_bStop = true;
		m_Thread->join();
	}

	/** creates a new window */
	void createWindow( const std::string& name )
	{
		boost::mutex::scoped_lock l( m_queueMutex );
		m_queue.push_back( boost::bind( &HighguiWindowModule::myCreateWindow, this, name ) );
		m_queueCondition.notify_all();
	}

	/** destroys a window */
	void destroyWindow( const std::string& name )
	{
		boost::mutex::scoped_lock l( m_queueMutex );
		m_queue.push_back( boost::bind( &HighguiWindowModule::myDestroyWindow, this, name ) );
		m_queueCondition.notify_all();
	}

	/** shows an image */
	void showImage( const std::string& name, const boost::shared_ptr< Image > pImage )
	{
		boost::mutex::scoped_lock l( m_queueMutex );
		m_queue.push_back( boost::bind( &HighguiWindowModule::myShowImage, this, name, pImage ) );
		m_queueCondition.notify_all();
	}


protected:
	// thread main loop
	void threadProc();

	void myCreateWindow( const std::string& name )
	{
		cvNamedWindow( name.c_str(), 0 );//CV_WINDOW_AUTOSIZE );
		m_nWindows++;
	}

	void myDestroyWindow( const std::string& name )
	{
		cvDestroyWindow( name.c_str() );
	}

	void myShowImage( const std::string& name, const boost::shared_ptr< Image > pImage )
	{ cvShowImage( name.c_str(), *pImage ); }

	// the thread
	boost::scoped_ptr< boost::thread > m_Thread;

	// stop the thread?
	volatile bool m_bStop;

	// number of windows open
	int m_nWindows;

	// a mutex object to synchronize highgui calls
	boost::mutex m_queueMutex;

	// notification when the queue has changed
	boost::condition m_queueCondition;

	// A queue to store procedure calls. Why? Because in Win32, all procedure calls that
	// modify a window should come from the same thread. Big debugging otherwise. So we
	// store the calls that the window thread should make in a single event queue.
	typedef std::deque< boost::function< void() > > QueueType;
	QueueType m_queue;
};


/**
 * @ingroup vision_components
 * A Component that displays images in a OpenCV HighGUI window.
 *
 * @par Input Ports
 * PushConsumer< Ubitrack::Vision::ImageMeasurement > port with name "Input".
 *
 * @par Output Ports
 * PushSupplier< Ubitrack::Measurement::Button > port with name "Button"
 *
 * @par Example Configuration
 * \verbatim 
<Pattern name="HighguiWindow" id="someId">
	<Input>
		<Node name="Camera" id="someId"/>
		<Node name="ImagePlane" id="someId2"/>
		<Edge name="Input" source="Camera" destination="ImagePlane" pattern-ref="..." edge-ref="..."/>
	</Input>
	<Output>
		<Node name="ImagePlane" id="someId3"/>
		<Node name="PButton" id="someId4"/>
		<Edge name="Button" source="ImagePlane" destination="PButton"/>
	</Output>
	<DataflowConfiguration>
		<UbitrackLib class="HighguiWindow"/>
		<Attribute name="windowTitle" value="My Window"/>
	</DataflowConfiguration>
</Pattern> 
\endverbatim
 */
class HighguiWindow
	: public HighguiWindowModule::Component
{
public:
	/** constructor, creates window */
	HighguiWindow( const std::string& sName, boost::shared_ptr< Graph::UTQLSubgraph >, const HWCKey& componentKey, HighguiWindowModule* pModule )
		: HighguiWindowModule::Component( sName, componentKey, pModule )
		, m_inPort( "Input", *this, boost::bind( &HighguiWindow::pushImage, this, _1 ) )
		, m_buttonPort( "Button", *this )
	{ getModule().createWindow( getKey() ); }

	/** destructor, destroys window */
	~HighguiWindow()
	{ getModule().destroyWindow( getKey() ); }

	/* signals a button event */
	void signalButton(char nKey)
	{ 
		Measurement::Timestamp t = Ubitrack::Measurement::now();
		m_buttonPort.send( Measurement::Button(t, int(nKey)) ); 
	}
	
protected:
	/** method that receives events and displays the image */
	void pushImage( const ImageMeasurement& m )
	{ getModule().showImage( getKey(), m ); }

	// the ports
	Dataflow::PushConsumer< ImageMeasurement > m_inPort;

	Dataflow::PushSupplier< Measurement::Button > m_buttonPort;
};


void HighguiWindowModule::threadProc()
{
	while ( !m_bStop )
	{
		// get next event from queue
		boost::function< void() > nextCall;
		{
			boost::mutex::scoped_lock l( m_queueMutex );
			if ( m_queue.empty() )
			{
				// wait for event or message dispatching timeout
				boost::xtime xt;
				boost::xtime_get( &xt, boost::TIME_UTC );
				xt.nsec += 100000000;
				xt.sec += xt.nsec / 1000000000;
				xt.nsec %= 1000000000;
				m_queueCondition.timed_wait( l, xt );
			}

			if ( !m_queue.empty() )
			{
				nextCall = m_queue.front();
				m_queue.pop_front();
			}
		}

		// execute the queued event
		if ( nextCall )
			nextCall();

		// let the opencv message dispatching do something
		int nKey = -1;
		if ( m_nWindows )
			nKey = cvWaitKey( 1 );
		
		// when a button is pushed, signal a button event
		if ( nKey != -1 )
		{
			LOG4CPP_DEBUG( logger, "Received Key" );
			ComponentList comps = getAllComponents();
			for ( ComponentList::iterator it = comps.begin(); it != comps.end(); it++ )
				(*it)->signalButton(nKey);
		}
	}
}

	
} } // namespace Ubitrack::Driver


UBITRACK_REGISTER_COMPONENT( Dataflow::ComponentFactory* const cf ) {
	cf->registerModule< Ubitrack::Drivers::HighguiWindowModule > ( "HighguiWindow" );
}
