/*
 * Ubitrack - Library for Ubiquitous Tracking
 * Copyright 2006, Technische Universitaet Muenchen, and individual
 * contributors as indicated by the @authors tag. See the
 * copyright.txt in the distribution for a full listing of individual
 * contributors.
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA, or see the FSF site: http://www.fsf.org.
 */

/**
 * @ingroup vision_components
 * @file
 * A compontent that tracks square markers.
 *
 * @author Daniel Pustka <daniel.pustka@in.tum.de>
 */

#include <stdlib.h>
#include <string>
#include <sstream>
#include <log4cpp/Category.hh>

#include <Ubitrack/Dataflow/PushConsumer.h>
#include <Ubitrack/Dataflow/PullConsumer.h>
#include <Ubitrack/Dataflow/PushSupplier.h>
#include <Ubitrack/Dataflow/Module.h>
#include <Ubitrack/Dataflow/ComponentFactory.h>
#include <Ubitrack/Measurement/Measurement.h>
#include <Ubitrack/Math/Scalar.h>
#include <Ubitrack/Math/cast_assign.h>

#include <Vision/Image.h>
#include <Vision/MarkerDetection.h>

#include <cv.h>

// get a logger
static log4cpp::Category& logger( log4cpp::Category::getInstance( "Ubitrack.Vision.MarkerTracker" ) );

// namespace shortcuts
using namespace Ubitrack;
using namespace Ubitrack::Dataflow;
using namespace Ubitrack::Vision;
using namespace Ubitrack::Vision::Markers;
using namespace Ubitrack::Math;
namespace ublas = boost::numeric::ublas;


namespace Ubitrack { namespace Drivers {


/** the component key -- reads the marker ID */
class IdKey
	: public Math::Scalar< unsigned int >
{
public:
	/** extract id from configuration */
	IdKey( boost::shared_ptr< Graph::UTQLSubgraph > subgraph )
	{
		std::string sId = subgraph->getNode( "Marker" )->getAttributeString( "markerId" );
		if ( sId.empty() )
			UBITRACK_THROW( "Missing markerId attribute on node Marker" );
		std::istringstream idStream( sId );
		idStream >> std::hex >> m_value;
	}

	// create from int
	IdKey( int v )
	{ m_value = v; }
};

/** camera key just takes the node id */
MAKE_NODEIDKEY( CameraIdKey, "Camera" );

// forward declarations
class MarkerTracker;
class MarkerTrackerModule;

//typedef Measurement::Measurement< boost::shared_ptr< Image > > ImageMeasurement;
typedef Module< CameraIdKey, IdKey, MarkerTrackerModule, MarkerTracker > MarkerTrackerModuleBase;

/**
 * Tracks markers and send the results to the components.
 * This is done only once for each timestamp.
 */
class MarkerTrackerModule
	: public MarkerTrackerModuleBase
{
public:
	MarkerTrackerModule( const CameraIdKey& key, boost::shared_ptr< Graph::UTQLSubgraph >, FactoryHelper* pFactory )
		: MarkerTrackerModuleBase( key, pFactory )
		, m_lastTimestamp( 0 )
	{}

	/** find markers in the image and send results to the components */
	void trackMarkers( const ImageMeasurement& );

protected:

	// timestamp of the last measurement
	Measurement::Timestamp m_lastTimestamp;
};


/**
 * @ingroup vision_components
 * Tracks markers in greyscale images.
 *
 * @par input ports
 * PushConsumer< Ubitrack::Vision::ImageMeasurement > of name "Input", accepts
 * only greyscale images
 *
 * @par Output Ports
 * PushSupplier< Measurement::Pose > Output -- sends the pose of tracked marker \n
 * PushSupplier< ImageMeasurement > DebugOut -- sends a debug image if connected
 *
 * @par The Pattern
 * @verbatim
<Pattern name="MarkerTracker">
	<Input>
		<Node name="Camera"/>
		<Node name="ImagePlane"/>
		<Node name="Marker" id="someOtherId">
			<Predicate>markerType=="UbitrackSquareMarker"&amp;&amp;markerId!=""</Predicate> 
		</Node>
		<Edge name="Image" source="Camera" destination="ImagePlane">
			<Predicate>type=="image"&amp;&amp;mode="push"</Predicate>
		</Edge>
		<Edge name="Config" source="Camera" destination="Marker">
			<Predicate>trackable=="UbitrackSquareMarker"</Predicate>
		</Edge>
	</Input>
	<Output>
		<Edge name="Output" source="Camera" destination="Marker">
			<Attribute name="type" value="6D"/>
			<Attribute name="mode" value="push"/>
		</Edge>
	</Output>
	<DataflowConfiguration>
		<UbitrackLib class="MarkerTracker"/>
	</DataflowConfiguration>
</Pattern>
@endverbatim
 * Note that the "Config" edge is optional and can be used to restrict marker tracking to only "visible"
 * markers, where "visibility" is determined by some other process.
 *
 * @par Dataflow Configuration
 * @verbatim
<Pattern name="MarkerTracker">
	<Input>
		<Node name="Camera" id="someId1"/>
		<Node name="ImagePlane" id="someId2"/>
		<Edge name="Image" source="Camera" destination="ImagePlane" pattern-ref="..." edge-ref="..."/>
	</Input>
	<Output>
		<Node name="Marker" id="someOtherId">
			<Attribute name="markerId" value="0xID"/> 
			<Attribute name="markerSize" value="0.06"/>
		</Node>
		<Edge name="Output" source="Camera" destination="Marker">
			<Attribute name="type" value="6D"/>
			<Attribute name="mode" value="push"/>
		</Edge>
	</Output>
	<DataflowConfiguration>
		<UbitrackLib class="MarkerTracker"/>
	</DataflowConfiguration>
</Pattern>
@endverbatim
 *
 * \p id is the normalized hexadecimal ID of a marker. If marker codes are interpreted as
 * binary numbers read row-wise starting top-left, the normalized ID is the one which represents
 * the smallest number. The pose is also computed w.r.t the normal marker orientation.
 * \p markerSize is in meters
 *
 * There is also a \c DebugImage output port, which - when connected - creates debug images.
 * An additional \c CameraIntrinsics input port can be used to supply a 3x3 camera intrinsics
 * matrix as pull.
 *
 * The \c ErrorPose output port gives the resulting marker pose with a covariance matrix, 
 * if connected.
 *
 * The component also has an output port named \c Corners that only give the 2D corner positions
 * of the marker in counter-clockwise order, starting top-left.
 */
class MarkerTracker
	: public MarkerTrackerModule::Component
{
public:
	MarkerTracker( const std::string& sName, boost::shared_ptr< Graph::UTQLSubgraph > subgraph, const IdKey& componentKey, MarkerTrackerModule* pModule )
		: MarkerTrackerModule::Component( sName, componentKey, pModule )
		, m_inPort( "Image", *this, boost::bind( &MarkerTracker::pushImage, this, _1 ) )
		, m_inIntrinsics( "CameraIntrinsics", *this )
		, m_outCorners( "Corners", *this )
		, m_outPort( "Output", *this )
		, m_outErrorPose( "ErrorPose", *this )
		, m_debugPort( "DebugImage", *this )
		, m_markerSize( 0.06f )
		, m_bEdgeTracking( true )
		, m_lastTime( 0 )
	{
		// get configuration
		subgraph->getNode( "Marker" )->getAttributeData( "markerSize", m_markerSize );

		if ( subgraph->m_DataflowAttributes.hasAttribute( "edgeTracking" ) )
			m_bEdgeTracking = subgraph->m_DataflowAttributes.getAttributeString( "edgeTracking" ) == "true";
	}

	/** is the debug port connected? */
	bool debug()
	{ return m_debugPort.isConnected(); }

	/** returns the camera intrinsics if connected. throws otherwise */
	Measurement::Matrix3x3 intrinsics( Measurement::Timestamp t )
	{ return m_inIntrinsics.get( t ); }
	
	/** returns the marker size in meters */
	float markerSize() const
	{ return m_markerSize; }
	
	/** use edge tracking? */
	bool useEdgeTracking() const
	{ return m_bEdgeTracking; }

protected:
	/** method that receives events and displays the image */
	void pushImage( const ImageMeasurement& m )
	{ getModule().trackMarkers( m ); }

	// ports
	Dataflow::PushConsumer< ImageMeasurement > m_inPort;
	Dataflow::PullConsumer< Measurement::Matrix3x3 > m_inIntrinsics;
	Dataflow::PushSupplier< Measurement::PositionList2 > m_outCorners;
	Dataflow::PushSupplier< Measurement::Pose > m_outPort;
	Dataflow::PushSupplier< Measurement::ErrorPose > m_outErrorPose;
	Dataflow::PushSupplier< ImageMeasurement > m_debugPort;

	// config
	float m_markerSize;
	bool m_bEdgeTracking;

	// some variables where the module stores information
	Measurement::Timestamp m_lastTime;
	Math::Pose m_lastPose;

	friend class MarkerTrackerModule;
};


void MarkerTrackerModule::trackMarkers( const ImageMeasurement& m )
{
	// check if image was already analyzed
	if ( m.time() == m_lastTimestamp )
		return;
	m_lastTimestamp = m.time();

	// debug image (if anybody is interested )
	boost::shared_ptr< Image > pDebugImg;
	{
		// check if debug image needs to be created
		ComponentList allComponents( getAllComponents() );
		for ( ComponentList::iterator it = allComponents.begin(); it != allComponents.end(); it++ )
			if ( (*it)->debug() )
			{
				pDebugImg = m->CvtColor( CV_GRAY2RGB, 3 );
				break;
			}
	}

	// get all components
	ComponentList components = getAllComponents();

	// get intrinsics matrix from first component
	Math::Matrix< 3, 3, float > K;
	try
	{
		Math::matrix_cast_assign( K, *components.front()->intrinsics( m.time() ) );
	}
	catch ( ... )
	{
		LOG4CPP_DEBUG( logger, "No intrinsics matrix given" );
		
		// compute cheap camera matrix if none given
		float fTx = static_cast< float >( m->width / 2 );
		float fTy = static_cast< float >( m->height / 2 );
		float f = 1.25f * m->width;
		K( 0, 0 ) = f;
		K( 0, 1 ) = 0.0f;
		K( 0, 2 ) = -fTx;
		K( 1, 1 ) = f;
		K( 1, 2 ) = -fTy;
		K( 2, 2 ) = -1.0f;
		K( 1, 0 ) = K( 2, 0 ) = K( 2, 1 ) = 0.0f;
	}
	
	// create marker map
	std::map< unsigned, MarkerInfo > markerMap;
	for ( ComponentList::iterator it = components.begin(); it != components.end(); it++ )
	{
		MarkerInfo& info( markerMap[ (*it)->getKey() ] );
		info.fSize = (*it)->markerSize();

		// set refinement
		if ( (*it)->m_outPort.isConnected() || (*it)->m_outErrorPose.isConnected() || pDebugImg )
			if ( (*it)->useEdgeTracking() )
				info.refinement = MarkerInfo::EEdgeRefinedPose;
			else
				info.refinement = MarkerInfo::ERefinedPose;
		else
			info.refinement = MarkerInfo::ECorners;

		// use last marker pose?
		if ( m.time() < (*it)->m_lastTime + 100000000 )
		{
			info.bUseInitialPose = true;
			info.pose = (*it)->m_lastPose;
		}

		// covariance?
		info.bCalculateCovariance = (*it)->m_outErrorPose.isConnected();
	}

	// run marker tracker
	detectMarkers( *m, markerMap, K, pDebugImg.get() );

	// check which markers were found
	for ( ComponentList::iterator it = components.begin(); it != components.end(); it++ )
	{
		MarkerInfo& info( markerMap[ (*it)->getKey() ] );
		if ( info.bFound )
		{
			// remember pose
			(*it)->m_lastTime = m.time();
			(*it)->m_lastPose = info.pose;

			// send corners
			if ( (*it)->m_outCorners.isConnected() )
			{
				LOG4CPP_DEBUG( logger, "Marker " << std::hex << (*it)->getKey() << ": sending corners [" 
					<< info.corners[ 0 ] << ", " << info.corners[ 1 ] << ", " << info.corners[ 2 ] << ", " << info.corners[ 3 ] << "]" );

				// convert from float to double
				boost::shared_ptr< std::vector< Math::Vector< 2 > > > pCorners( new std::vector< Math::Vector< 2 > > );
				for ( unsigned i = 0; i < 4; i++ )
					pCorners->push_back( Math::Vector< 2 >( info.corners[ i ] ) );

				(*it)->m_outCorners.send( Measurement::PositionList2( m.time(), pCorners ) );
			}

			// send to component
			try
			{
				LOG4CPP_DEBUG( logger, "Marker " << std::hex << (*it)->getKey() << ": sending pose " << info.pose );
				(*it)->m_outPort.send( Measurement::Pose( m.time(), info.pose ) );
			}
			catch ( ... )
			{}

			// also send ErrorPose if anybody is connected
			if ( info.bCalculateCovariance )
				(*it)->m_outErrorPose.send( Measurement::ErrorPose( m.time(), ErrorPose( info.pose, info.covariance ) ) );
		}
	}

	// push debug image
	if ( pDebugImg )
	{
		ComponentList components = getAllComponents();
		for ( ComponentList::iterator it = components.begin(); it != components.end(); it++ )
			(*it)->m_debugPort.send( ImageMeasurement( m.time(), pDebugImg ) );
	}
}


} } // namespace Ubitrack::Driver


UBITRACK_REGISTER_COMPONENT( Dataflow::ComponentFactory* const cf ) {
	cf->registerModule< Ubitrack::Drivers::MarkerTrackerModule > ( "MarkerTracker" );
}
