/*
 * Ubitrack - Library for Ubiquitous Tracking
 * Copyright 2006, Technische Universitaet Muenchen, and individual
 * contributors as indicated by the @authors tag. See the
 * copyright.txt in the distribution for a full listing of individual
 * contributors.
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA, or see the FSF site: http://www.fsf.org.
 */

/**
 * @ingroup vision_components
 * @file
 * Acquires stereo images from Firewire cameras using libdc1394.
 *
 * @author Florian Echtler <echtler@in.tum.de>
 */

#include <string>
#include <list>
#include <iostream>
#include <algorithm>
#include <log4cpp/Category.hh>

#include <boost/thread.hpp>
#include <boost/bind.hpp>
#include <boost/scoped_ptr.hpp>

#include <Ubitrack/Dataflow/PushSupplier.h>
#include <Ubitrack/Dataflow/Component.h>
#include <Ubitrack/Dataflow/ComponentFactory.h>
#include <Ubitrack/Measurement/Measurement.h>
#include <Ubitrack/Util/OS.h>
#include <Vision/Image.h>
#include <cv.h>

#include <dc1394/control.h>
#include <dc1394/conversions.h>


#define NUM_DMA_BUFFERS 4
#define MAX_PORTS 1


// get a logger
static log4cpp::Category& logger( log4cpp::Category::getInstance( "Ubitrack.Vision.Multi1394FrameGrabber" ) );

using namespace Ubitrack;
using namespace Ubitrack::Vision;


namespace Ubitrack { namespace Drivers {

/**
 * @ingroup vision_components
 * Pushes images from multiple cameras using libdc1394.
 *
 * @par Input Ports
 * None.
 *
 * @par Output Ports
 * \c ColorOutputX push port(s) of type Ubitrack::Vision::ImageMeasurement.
 * \c GreyOutputX  push port(s) of type Ubitrack::Vision::ImageMeasurement.
 */
class Multi1394FrameGrabber
	: public Dataflow::Component
{
public:

	/** constructor */
	Multi1394FrameGrabber( const std::string& sName, boost::shared_ptr< Graph::UTQLSubgraph >  );

	/** destructor, waits until thread stops */
	~Multi1394FrameGrabber();

	/** Component start method. starts the thread */
	virtual void start();

	/** Component stop method, stops thread */
	virtual void stop();

protected:
	// thread main loop
	void ThreadProc();

	// helper functions
	int  camera_setup( int num );
	void camera_cleanup( int num );
	void process_frame( int num );

	// the thread
	boost::scoped_ptr< boost::thread > m_Thread;

	// stop the thread?
	volatile bool m_bStop;

	// the ports
	Dataflow::PushSupplier< ImageMeasurement >* m_colorPorts[MAX_PORTS];
	Dataflow::PushSupplier< ImageMeasurement >* m_greyPorts [MAX_PORTS];

	// camera data structures
	dc1394camera_t* m_camera[MAX_PORTS];
	unsigned int m_numCameras;
	unsigned int m_width, m_height;
	unsigned int m_portCount;
};


void Multi1394FrameGrabber::stop()
{
	if ( m_running )
	{
		if ( m_Thread )
		{
			m_bStop = true;
			m_Thread->join();
		}
		m_running = false;
	}
}

void Multi1394FrameGrabber::start()
{
	if ( !m_running )
	{
		m_bStop = false;
		m_Thread.reset( new boost::thread( boost::bind ( &Multi1394FrameGrabber::ThreadProc, this ) ) );
		m_running = true;
	}
}



Multi1394FrameGrabber::Multi1394FrameGrabber( const std::string& sName, boost::shared_ptr< Graph::UTQLSubgraph > subgraph )
	: Dataflow::Component( sName )
	, m_bStop( true )
	, m_numCameras( 0 )
	, m_width( 640 )
	, m_height( 480 )
	, m_portCount( 2 )
{
	if (m_portCount > MAX_PORTS) m_portCount = MAX_PORTS;
	for (unsigned int i = 0; i < m_portCount; i++)
	{
		std::ostringstream cname; cname << "ColorOutput" << i+1;
		std::ostringstream gname; gname << "GreyOutput"  << i+1;
		m_colorPorts[i] = new Dataflow::PushSupplier< ImageMeasurement > ( cname.str(), *this );
		m_greyPorts [i] = new Dataflow::PushSupplier< ImageMeasurement > ( gname.str(), *this );
	}
	std::cout << "ports created" << std::endl;

	stop();
}


Multi1394FrameGrabber::~Multi1394FrameGrabber()
{
	stop();

	for (unsigned int i = 0; i < m_portCount; i++)
	{
		delete m_colorPorts[i];
		delete m_greyPorts [i];
	}
}


void Multi1394FrameGrabber::camera_cleanup( int num )
{
	dc1394_video_set_transmission( m_camera[num], DC1394_OFF );
	dc1394_capture_stop( m_camera[num] );
	//Bjrn  dc1394_free_camera( m_camera[num] );
	dc1394_camera_free( m_camera[num] );

}


int Multi1394FrameGrabber::camera_setup( int num )
{
 	LOG4CPP_INFO( logger, "Setting up camera #" << num << ".." );

	//dc1394_reset_camera( m_camera[num] );
	//dc1394_cleanup_iso_channels_and_bandwidth( m_camera[num] );
	dc1394_video_set_transmission( m_camera[num], DC1394_OFF );

	dc1394_video_set_iso_speed( m_camera[num], DC1394_ISO_SPEED_400 );
	dc1394_video_set_mode( m_camera[num], DC1394_VIDEO_MODE_640x480_MONO8 );
	dc1394_video_set_framerate( m_camera[num], DC1394_FRAMERATE_15 );

	if (dc1394_capture_setup( m_camera[num], NUM_DMA_BUFFERS, DC1394_CAPTURE_FLAGS_DEFAULT ) != DC1394_SUCCESS)
	{
		camera_cleanup( num );
		LOG4CPP_ERROR( logger, "Error during camera initialization. Check that 640x480_YUV411@30FPS is supported." );
		return 0;
	}
 
	if (dc1394_video_set_transmission( m_camera[num], DC1394_ON ) != DC1394_SUCCESS)
	{
		camera_cleanup( num );
		LOG4CPP_ERROR( logger, "Unable to start ISO transmission." );
		return 0;
	}

	LOG4CPP_INFO( logger, "Camera #" << num << " setup successful." );
	return 1;
}


/** convert YUV411 image data to greyscale - this function is missing in libdc1394 */
void dc1394_YUV411_to_MONO8( uint8_t* src, uint8_t* dest, uint32_t width, uint32_t height )
{
  register int i = (width*height) + ( (width*height) >> 1 ) -1;
  register int j = (width*height)                           -1;
  register int y0, y1, y2, y3;
  
  while (i >= 0) {
    y3 = (uint8_t) src[i--];
    y2 = (uint8_t) src[i--]; i--;
    y1 = (uint8_t) src[i--];
    y0 = (uint8_t) src[i--]; i--;
    dest[j--] = y3;
    dest[j--] = y2;
    dest[j--] = y1;
    dest[j--] = y0;
  }
}


/** retrieve frame from camera, convert to RGB or greyscale, send through port */
void Multi1394FrameGrabber::process_frame( int num )
{
	dc1394video_frame_t* frame;

	/* retrieve raw YUV411 data */
	dc1394_capture_dequeue( m_camera[num], DC1394_CAPTURE_POLICY_WAIT, &frame );
	
	/* if we are lagging behind, dump a frame to catch up */
	if (frame->frames_behind > 0) {
		LOG4CPP_INFO( logger, "dumping frame (lag: " << frame->frames_behind << " frame(s))" );
		dc1394_capture_enqueue( m_camera[num], frame );
		dc1394_capture_dequeue( m_camera[num], DC1394_CAPTURE_POLICY_WAIT, &frame );
	}

	/* store the last frame's timestamp */
	Measurement::Timestamp time( frame->timestamp * 1000 );

	bool bDeBayer = true;
	boost::shared_ptr< Image > pColorImage;
	if ( bDeBayer )
	{
		pColorImage.reset( new Image( m_width, m_height, 3 ) );
		dc1394_bayer_decoding_8bit( frame->image, reinterpret_cast< unsigned char* >( pColorImage->imageData ), m_width, m_height,
			DC1394_COLOR_FILTER_GBRG, DC1394_BAYER_METHOD_HQLINEAR);
	}

	/* convert to greyscale and RGB if a consumer is connected */
	if (m_colorPorts[num]->isConnected() && pColorImage ) {
		/*boost::shared_ptr< Image > pColorImage( new Image( m_width, m_height, 3 ) );
		pColorImage->origin = 0;
		dc1394_convert_to_RGB8( rawyuv, (unsigned char*)(pColorImage->imageData), m_width, m_height, 0, DC1394_COLOR_CODING_YUV411, 0 );*/
		m_colorPorts[num]->send( ImageMeasurement( time, pColorImage ) );
	}

	if (m_greyPorts[num]->isConnected()) {
		boost::shared_ptr< Image > pGreyImage;
		if ( bDeBayer )
		{
			pGreyImage = pColorImage->CvtColor( CV_RGB2GRAY, 1 );
			//dc1394_convert_to_MONO8( (unsigned char*)(pColorImage->imageData), (unsigned char*)(pGreyImage->imageData),
			//	m_width, m_height, 0,    DC1394_COLOR_CODING_, 16 );
		}
		else
		{
			pGreyImage.reset( new Image( m_width, m_height, 1 ) );
			pGreyImage->origin = 0;
			memcpy( pGreyImage->imageData, frame->image, m_width * m_height );
		}
		m_greyPorts[num]->send( ImageMeasurement( time, pGreyImage ) );
	}

	dc1394_capture_enqueue( m_camera[num], frame );
}


/** helper function to sort the camera array */
bool cmp( dc1394camera_t* a, dc1394camera_t* b )
{
	//Bjrn return a->euid_64 > b->euid_64;
	return a->guid > b->guid;
}


void Multi1394FrameGrabber::ThreadProc()
{
	std::cout << "thread starting" << std::endl;
	
	/*******************************///
	//// Bjrn
	
	dc1394_t * d;
    dc1394camera_list_t * list;
	dc1394error_t err;

    d = dc1394_new ();
	err=dc1394_camera_enumerate (d, &list);
    //Bjrn Print ToDo Error
	//DC1394_ERR_RTN(err,"Failed to enumerate cameras");

    if (list->num == 0) {
        LOG4CPP_ERROR( logger, "No cameras found" );       return;
    }
	
	m_numCameras = list->num;
	LOG4CPP_INFO( logger, m_numCameras << " cameras found" );

    m_camera[0] = dc1394_camera_new (d, list->ids[0].guid);
    if (!m_camera[0]) {
        dc1394_log_error("Failed to initialize camera with guid %llx", list->ids[0].guid);
       // return 1;
    }
    dc1394_camera_free_list (list);

    printf("Using camera with GUID %llx\n", m_camera[0]->guid);

    //dc1394video_modes_t modes;
	
	/********************************///
	
	/* Find cameras */
	
	/* Bjoern
	if (dc1394_find_cameras( &m_camera, &m_numCameras ) != DC1394_SUCCESS)
	{
		LOG4CPP_ERROR( logger, "Unable to scan for IIDC cameras. Check your Firewire bus." );
		return;
	}
	*/
	if (m_numCameras < m_portCount) {
		LOG4CPP_ERROR( logger, "This module requires at least " << m_portCount << " IIDC cameras, but only " << m_numCameras << " were found." );
		return;
	}

	/* Sort the cameras according to their GUID for a consistent ordering. */
	std::sort( m_camera, m_camera + m_numCameras, cmp );

	/* setup capture parameters */
	for (unsigned int i = 0; i < m_portCount; i++)
	  if (!camera_setup( i )) return;

	while ( !m_bStop )
	{
		/* acquire frames from all cameras and wait for completion */
		for (unsigned int i = 0; i < m_portCount; i++)
			process_frame( i );
	}

	/* housekeeping time.. */
	for (unsigned int i = 0; i < m_portCount; i++)
		camera_cleanup( i );
	free( m_camera );
}

} } // namespace Ubitrack::Driver


UBITRACK_REGISTER_COMPONENT( Dataflow::ComponentFactory* const cf ) {
	cf->registerComponent< Ubitrack::Drivers::Multi1394FrameGrabber > ( "Multi1394FrameGrabber" );
}

