/*
 * Ubitrack - Library for Ubiquitous Tracking
 * Copyright 2006, Technische Universitaet Muenchen, and individual
 * contributors as indicated by the @authors tag. See the 
 * copyright.txt in the distribution for a full listing of individual
 * contributors.
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA, or see the FSF site: http://www.fsf.org.
 */

/**
 * @ingroup vision
 * @file
 * Implements basic routines for handling OpenCV images in a "Resource Acquisition Is Initialization" (RAII) fashion
 *
 * @author Daniel Pustka <daniel.pustka@in.tum.de>
 */

#include <cv.h>
#include "Image.h"

namespace Ubitrack { namespace Vision {


Image::Image( int nWidth, int nHeight, int nChannels, void* pImageData, int nDepth, int nOrigin, int nAlign )
	: m_bOwned( false )
{
	cvInitImageHeader( this, cvSize( nWidth, nHeight ), nDepth, nChannels, nOrigin, nAlign );
	imageData = static_cast< char* >( pImageData );
}


Image::Image( int nWidth, int nHeight, int nChannels, int nDepth )
	: m_bOwned( true )
{
	cvInitImageHeader( this, cvSize( nWidth, nHeight ), nDepth, nChannels );
	imageDataOrigin = imageData = static_cast< char* >( cvAlloc( imageSize ) );
}


Image::Image( IplImage* pIplImage, bool bDestroy )
	: m_bOwned( bDestroy )
{
	memcpy( static_cast< IplImage* >( this ), pIplImage, sizeof( IplImage ) );
	if ( bDestroy )
		cvReleaseImageHeader( &pIplImage );
}


Image::~Image()
{
	if ( m_bOwned )
		cvFree( reinterpret_cast< void** >( &imageDataOrigin ) );
}


boost::shared_ptr< Image > Image::CvtColor( int nCode, int nChannels, int nDepth ) const
{
	boost::shared_ptr< Image > r( new Image( width, height, nChannels, nDepth ) );
	cvCvtColor( this, *r, nCode ); 
	r->origin = origin; 
	return r;
}


boost::shared_ptr< Image > Image::Clone() const
{
	return boost::shared_ptr< Image >( new Image( cvCloneImage( *this ) ) );
}

} } // namespace Ubitrack::Vision
