/*
 * Ubitrack - Library for Ubiquitous Tracking
 * Copyright 2006, Technische Universitaet Muenchen, and individual
 * contributors as indicated by the @authors tag. See the 
 * copyright.txt in the distribution for a full listing of individual
 * contributors.
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA, or see the FSF site: http://www.fsf.org.
 */

/**
 * @ingroup vision
 * @file
 * A class for handling OpenCV images in a "Resource Acquisition Is Initialization" (RAII) fashion
 *
 * @author Daniel Pustka <daniel.pustka@in.tum.de>
 */

#ifndef __UBITRACK_VISION_IMAGE_H_INCLUDED__
#define __UBITRACK_VISION_IMAGE_H_INCLUDED__

#include <boost/shared_ptr.hpp>
#include <boost/utility.hpp>
#include <boost/serialization/access.hpp>
#include <cxcore.h>
#include <Vision/Vision.h>
#include <Ubitrack/Measurement/Measurement.h>
#include <Ubitrack/Dataflow/EventTypeTraits.h>

namespace Ubitrack { namespace Vision {

/**
 * @ingroup vision
 * A wrapper object for IplImages in a "Resource Acquisition Is Initialization" (RAII) fashion.
 *
 * We don't want to change the OpenCV feeling more than necessary, but we need images that
 * destroy themselves when they run out of scope!
 */
class UTVISION_EXPORT Image
	: public IplImage
	, private boost::noncopyable
{
public:
	/**
	 * Creates a reference to an existing image array.
    * The resulting \c Image object does not own the data.
    *
    * @param nWidth width of image in pixels
    * @param nHeight height of image in pixels
    * @param nChannels number of channels (1=grey, 3=rgb)
    * @param pImageData pointer to raw image data
    * @param nDepth information about pixel representation (see OpenCV docs)
    * @param nOrigin 0 if pixels start in top-left corner, 1 if in bottom-left
    * @param nAlign alignment of rows in bytes
	 */
	Image( int nWidth, int nHeight, int nChannels, void* pImageData,
		int nDepth = IPL_DEPTH_8U, int nOrigin = 0, int nAlign = 4 );

	/**
	 * Creates a new image and allocates memory for it.
	 *
    * @param nWidth width of image in pixels
    * @param nHeight height of image in pixels
    * @param nChannels number of channels (1=grey, 3=rgb)
    * @param nDepth information about pixel representation (see OpenCV docs)

	 */
	Image( int nWidth, int nHeight, int nChannels, int nDepth = IPL_DEPTH_8U );

	/**
	 * Create from pointer to IplImage.
	 *
	 * @param pIplImage pointer to existing IplImage
	 * @param bDestroy if true, the IplImage is destroyed (using cvReleaseImageHeader) and
	 * ownership of the data is taken. Otherwise the data is not owned.
	 */
	explicit Image( IplImage* pIplImage, bool bDestroy = true );

	/** releases the image data if the data block is owned by this object. */
	~Image();

	/** convert to CvArr* for usage in some OpenCV functions */
	operator CvArr*()
		{ return static_cast< IplImage* >( this ); }

	/** convert to CvArr* for usage in some OpenCV functions */
	operator const CvArr*() const
		{ return static_cast< const IplImage* >( this ); }

	/** also convert to IplImage* for more consistent interface */
	operator IplImage*()
		{ return static_cast< IplImage* >( this ); }

	/** also convert to IplImage* for more consistent interface */
	operator const IplImage*() const
		{ return static_cast< const IplImage* >( this ); }

	/**
    * Convert color space.
    * Wraps \c cvCvtColor, but keeps the origin flag intact.
    *
    * @param nCode conversion Code (see OpenCV docs of \c cvCvtColor, e.g. CV_RGB2GRAY)
    * @param nChannels number of channels (1=grey, 3=rgb)
    * @param nDepth information about pixel representation (see OpenCV docs)
    */
	boost::shared_ptr< Image > CvtColor( int nCode, int nChannels, int nDepth = IPL_DEPTH_8U ) const;

	/** Creates a copy of this image */
	boost::shared_ptr< Image > Clone() const;

private:
	// does this object own the data imageData points to?
	bool m_bOwned;
	
	friend class ::boost::serialization::access;
	/** boost serialization helper. does not do any useful serialization... */
	template< class Archive >
	void serialize( Archive& ar, const unsigned int )
	{ char X( 'X' ); ar & X; }
};

/** define a shortcut for image measurements */
typedef Measurement::Measurement< Image > ImageMeasurement;

} } // namespace Ubitrack::Vision


namespace Ubitrack { namespace Dataflow {

/**
 * \internal
 * Defines how to extract the priority out of a data type. 
 * Specialized for image measurements to reduce the maximum queue length.
 */
template<>
struct EventTypeTraits< Vision::ImageMeasurement >
{
	unsigned long long getPriority( const Vision::ImageMeasurement& m ) const
	{ return m.time(); }

	/** the maximum queue length for images is 1! */
	int getMaxQueueLength() const
	{ return 1; }
};

} } // namespace Ubitrack::Dataflow

#endif
