/*
 * Ubitrack - Library for Ubiquitous Tracking
 * Copyright 2006, Technische Universitaet Muenchen, and individual
 * contributors as indicated by the @authors tag. See the 
 * copyright.txt in the distribution for a full listing of individual
 * contributors.
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA, or see the FSF site: http://www.fsf.org.
 */

/**
 * @ingroup vision
 * @file
 * Implementation of routines to detect square markers in images.
 *
 * @author Daniel Pustka <daniel.pustka@in.tum.de>
 */

#define _USE_MATH_DEFINES
#include <math.h>
#include <algorithm>
#include <boost/scoped_array.hpp>
#include <cv.h>

#include <iostream>

#include <Ubitrack/Math/MatrixOperations.h>
#include <Ubitrack/Math/cast_assign.h>
#include <Ubitrack/Calibration/Homography.h>
#include <Ubitrack/Calibration/2D3DPoseEstimation.h>
#include <boost/numeric/ublas/matrix_proxy.hpp>

#include "MarkerDetection.h"
#include "EdgeExtraction.h"

// get a logger
#include <log4cpp/Category.hh>
static log4cpp::Category& logger( log4cpp::Category::getInstance( "Ubitrack.Vision.MarkerTracker" ) );

// for edge tracking
#define OPTIMIZATION_LOGGING
static log4cpp::Category& optLogger( log4cpp::Category::getInstance( "Ubitrack.Vision.MarkerTracker" ) );
#include <Ubitrack/Math/LevenbergMarquardt.h>
#include <Ubitrack/Math/GaussNewton.h>
#include <Vision/EdgeMeasurement.h>
#include <Ubitrack/Calibration/Function/ProjectivePoseNormalize.h>


namespace ublas = boost::numeric::ublas;

namespace Ubitrack { namespace Vision { namespace Markers {

// some constants for tuning

/** width of area in which to search for edges, as percentage of marker extension */
const float g_fEdgeSearchArea = 0.12f;

/** number of tracked points on an edge */
const int g_nEdgePoints = 8;

/** number of tracked points on an edge necessary for line fit */
const int g_nFitMinEdgePoints = 4;

/** minimal gradient response for edges */
int g_nMinimalEdgeStrength = 40;

/** maximal number of non-black pixels on marker border */
const int g_nMaxWhiteBorderPixels = 4;


// forward declaration of private functions
static void drawCube( Vision::Image& img, const Math::Pose& pose, const Math::Matrix< 3, 3, float >& K, double markerSize, CvScalar color );


#ifdef HAVE_LAPACK
void detectMarkers( const Image& img, std::map< unsigned, MarkerInfo >& markerInfos, 
	const Math::Matrix< 3, 3, float >& _K, Image* pDebugImg )
{
	// threshold image
	Image thresholded( img.width, img.height, 1 );

	// The source image is copied as OpenCV from beta 5 on destroys it
	cvAdaptiveThreshold( *img.Clone(), thresholded, 255, CV_ADAPTIVE_THRESH_MEAN_C, CV_THRESH_BINARY,
		( img.height / 16 ) | 1 , 0 );

	Math::Matrix< 3, 3, float > K( _K );
		
	// flip image coordinates if origin==0 
	LOG4CPP_DEBUG( logger, "image origin flag = " << img.origin );
	if ( !img.origin )
	{
		K( 1, 1 ) *= -1.0f;
		K( 1, 2 ) = -K( 1, 2 ) + K( 2, 2 ) * static_cast< float >( img.height - 1 );
	}

	// compute inverse camera matrix
	Math::Matrix< 3, 3, float > invK( invert_matrix( K ) );
	
	// find markers in the image
	MarkerList markers = findQuadrangles( thresholded );

	int iShownMarker = 0;
	for ( MarkerList::iterator it = markers.begin(); it != markers.end(); it++ )
	{
		// refine corner positions
		if ( refineCorners( img, *it/*, pDebugImg */) )
		{
			// if image is top-down, exchange point 2 and 4 to assure counter-clock-wise order for squareHomography
			if ( !img.origin )
			{
				Math::Vector< 2, float > temp( (*it)[ 3 ] );
				(*it)[ 3 ] = (*it)[ 1 ];
				(*it)[ 1 ] = temp;
			}

			// compute homography
			Math::Matrix< 3, 3, float > H( Calibration::squareHomography( *it ) );

			// get marker image & decode
			boost::shared_ptr< Image > pMarker( getMarkerImage( img, H, 6 ) );
			unsigned nCode = readCode( *pMarker );

			// show marker image in debug mode
			if ( pDebugImg )
			{
				int markersPerCol = img.height / ( pMarker->height * 5 );
				int xStart = ( iShownMarker / markersPerCol ) * ( pMarker->height * 5 );
				int yStart = ( iShownMarker % markersPerCol ) * ( pMarker->height * 5 );
				iShownMarker++;
				for ( int x = 0; x < 6; x++ )
					for ( int y = 0; y < 6; y++ )
					{
						unsigned c = ( (unsigned char*)pMarker->imageData )[ y * pMarker->widthStep + x ];
						cvRectangle( *pDebugImg, cvPoint( xStart + 5 * x, yStart + 5 * y ), 
							cvPoint( xStart + 5 * (x+1), yStart + 5 * (y+1) ),
							nCode ? CV_RGB( c, c, c ) : CV_RGB( c, (c*c)>>8, (c*c)>>8 ), CV_FILLED );
					}
			}

			// normalize marker code
			int nRotate = 0;
			nCode = normalizeCode( nCode, 4, nRotate );
			
			if ( nCode )
			{ LOG4CPP_TRACE( logger, "found marker with code=" << std::hex << nCode ); }

			if ( nCode && ( markerInfos.find( nCode ) != markerInfos.end() || markerInfos.find( 0 ) != markerInfos.end() ) )
			{
				// find the info struct for this marker
				if ( markerInfos.find( nCode ) == markerInfos.end() )
					markerInfos[ nCode ] = markerInfos[ 0 ];

				MarkerInfo& info( markerInfos[ nCode ] );
				info.bFound = true;

				// create corner list
				info.corners.resize( 4 );
				for ( unsigned i = 0; i < 4; i++ )
					Math::vector_cast_assign( info.corners[ i ], Math::Vector< 2 >( (*it)[ ( i + nRotate ) % 4 ]( 0 ), (*it)[ ( i + nRotate ) % 4 ]( 1 ) ) );

				// origin correction
				if ( !img.origin )
					for ( unsigned i = 0; i < 4; i++ )
						info.corners[ i ]( 1 ) = img.height - 1 - info.corners[ i ]( 1 );


				if ( info.refinement >= MarkerInfo::EInitialPose )
				{
					// scale homography with marker size
					ublas::column( H, 2 ) *= info.fSize;
					LOG4CPP_DEBUG( logger, "Homography: " << H );

					// compute pose
					Math::Pose initialPose( Calibration::poseFromHomography( H, invK ) );

					// rotate pose to account for different order of corner points
					const double fSqrtHalf = 0.70710678118654752440084436210485;
					static const double csMap[ 4 ][ 2 ] =
						{ { 1.0f, 0.0f }, { fSqrtHalf, fSqrtHalf }, { 0.0f, 1.0f }, { fSqrtHalf, -fSqrtHalf } };
					initialPose = Math::Pose( initialPose.rotation() * Math::Quaternion( 0.0, 0.0, csMap[ nRotate ][ 1 ], csMap[ nRotate ][ 0 ] ), initialPose.translation() );
					LOG4CPP_DEBUG( logger, "initial pose: " << initialPose );

					// create list of 3D marker points
					static float std2dPoints[ 4 ][ 2 ] =
						{ { -1.0f, 1.0f }, { -1.0f, -1.0f }, { 1.0f, -1.0f }, { 1.0f, 1.0f } };
					std::vector< Math::Vector< 3, float > > p3D;
					const float fF = 0.5f * info.fSize;
					for ( unsigned i = 0; i < 4; i++ )
					{
						unsigned j = ( i + 4 - nRotate ) % 4;
						p3D.push_back( Math::Vector< 3, float >( fF * std2dPoints[ j ][ 0 ], fF * std2dPoints[ j ][ 1 ], 0.0f ) );
					}

					// in debug mode, draw a cube onto the marker to visualize the initial pose
					if ( pDebugImg )
						drawCube( *pDebugImg, initialPose, K, info.fSize, CV_RGB( 0, 127, 127 ) );

					Math::Pose pose( initialPose );

					// perform some LM iterations on the edge points to get a more stable initialization
					double optRes = Calibration::optimizePose( pose, *it, p3D, K, 4 );

					// try optimization with the rotation from the last pose
					if ( info.bUseInitialPose )
					{
						Math::Pose testPose = Math::Pose( info.pose.rotation(), initialPose.translation() );
						double testOptRes = Calibration::optimizePose( testPose, *it, p3D, K, 4 );
						if ( testOptRes < optRes * 9 ) // prefer last pose rotation
							pose = testPose;
					}

					// refine the pose
					if ( info.refinement >= MarkerInfo::EEdgeRefinedPose )
					{

						// create edgel lists
						std::vector< Math::Vector< 3, float > > edgels1;
						std::vector< Math::Vector< 3, float > > edgels2;
						computeMarkerEdgels( nCode, info.fSize, edgels1, edgels2 );

						// optimize pose
						Math::Vector< 7, float > poseVector;
						pose.toVector( poseVector );

						// compute scaling of normals as covariance of the marker corners
						Math::Matrix< 2, 2, float > cornerCov( ublas::zero_matrix< float >( 2, 2 ) );
						Math::Vector< 2, float > cornerAvg( ublas::zero_vector< float >( 2 ) );
						for ( unsigned int i = 0; i < 4; i++ )
						{
							cornerCov += ublas::outer_prod( (*it)[ i ], (*it)[ i ] );
							cornerAvg += (*it)[ i ];
						}
						cornerCov /= 4;
						cornerAvg /= 4;
						cornerCov -= ublas::outer_prod( cornerAvg, cornerAvg );

						// scale covariance matrix
						cornerCov *=  ( 1.5f * 1.5f ) / ( 6.0f * 6.0f );

						LOG4CPP_TRACE( logger, "cornerCov = " << cornerCov );

						// Optimization iterations on edgels
						EdgeListMeasurementFunction< float, FindEdgePositiveMaximum > edgeMF( 
							edgels1, edgels2, K, cornerCov, img, pDebugImg );
						Math::levenbergMarquardt( 
							edgeMF,
							poseVector,
							boost::numeric::ublas::zero_vector< float >( edgels1.size() ),
							Math::OptTerminate( 6, 1e-4 ),
							Calibration::Function::ProjectivePoseNormalize()
						);

						pose = Math::Pose::fromVector( poseVector );
					}
					else
					{
						// more iterations on the edge points
						Calibration::optimizePose( pose, *it, p3D, K, 7 );
					}

					// remember pose
					LOG4CPP_DEBUG( logger, "optimized pose: " << pose );
					info.pose = pose;
					
					// also send ErrorPose if anybody is connected
					if ( info.bCalculateCovariance )
						// FIXME: make pixel variance configurable
						info.covariance = Calibration::singleCameraPoseError( pose, p3D, K, 0.04f * 0.04f );
					
					// in debug mode, draw a nice cube onto the marker
					if ( pDebugImg )
						drawCube( *pDebugImg, pose, K, info.fSize, CV_RGB( 255, 255, 0 ) );
				}
			}
			else
			{ LOG4CPP_TRACE( logger, "no marker info found" ); }
		}
	}
}
#endif // HAVE_LAPACK


MarkerList findQuadrangles( Image& bwImage ) 
{
	// initialize return value
	MarkerList ret;

	// create memory storage that will contain all the dynamic data
	CvMemStorage* pStorage = cvCreateMemStorage( 0 );

	// Create dynamic structure and sequence.
	CvSeq* pContours = cvCreateSeq( CV_SEQ_ELTYPE_POINT, sizeof( CvSeq ), sizeof( CvPoint ), pStorage );

	// Find all contours.
	cvFindContours( bwImage, pStorage, &pContours, sizeof( CvContour ), 
		CV_RETR_LIST, CV_CHAIN_APPROX_SIMPLE, cvPoint( 0, 0 ) );

	// approximate all contours by polygons and check size and corner number
	for ( CvSeq* pContour = pContours; pContour; pContour = pContour->h_next ) 
	{
		CvRect boundingRect = cvBoundingRect( pContour );
		float fDiameter = sqrtf( static_cast< float >( boundingRect.height * boundingRect.height + boundingRect.width * boundingRect.width ) );

		// discard too small contours
		if ( fDiameter < 16 )
			continue;
	
		// approximate contour by few corners
		CvSeq* pApproxChain = cvApproxPoly( pContour, sizeof( CvContour ), pStorage,
			CV_POLY_APPROX_DP, fDiameter * 0.03f + 2.0f, 0 );

		int nCorners = pApproxChain->total; // This is number point in contour

		// discard non-quadrangles
		if ( nCorners != 4 )
			continue;

		// Alloc memory for contour point set. 
		boost::scoped_array< CvPoint > pPointArray( new CvPoint[ nCorners ] );

		// Get contour point set and append to list
		cvCvtSeqToArray( pApproxChain, pPointArray.get(), CV_WHOLE_SEQ );

		CornerList rect( 4 );
		for ( int i = 0; i < 4; i++ ) 
		{
			rect[ i ][ 0 ] = static_cast< float >( pPointArray[ i ].x );
			rect[ i ][ 1 ] = static_cast< float >( pPointArray[ i ].y );
		}

		ret.push_back( rect );
	}

	// cleanup	     
	cvReleaseMemStorage( &pStorage );

	return ret;
}


/**
 * \internal
 * Given two points, find some accurate points on the line between the points with subpixel precision.
 *
 * Note: integer coordinates lie at pixel centers.
 * This function does not respect the origin flag and assumes that a y-coordinate of 0 is on top.
 *
 * @param pPoints: array where the result will be stored
 * @param pStrengths: edge strength at the point
 * @param nPoints: maximum number of points to find
 * @param pGreyImage: greyscale image in which to find points
 * @param point1, point2: two points describing the line
 * @param nSearchPixels: number of pixels around the line to search for edges
 * @param pDebugImg if not NULL, we will draw some debugging info into this image
 * @return number of detected points
 */
int refineLinePoints( CvPoint2D32f* pPoints, int* pStrengths, int nPoints, const Image& greyImage,
	const Math::Vector< 2, float >& point1, const Math::Vector< 2, float >& point2, int nSearchPixels, 
	Image* pDebugImg = NULL )
{
	int iPoint = 0;

	// compute normalized direction vector of input line
	Math::Vector< 2, float > direction( point2 - point1 );
	float fLength = boost::numeric::ublas::norm_2( direction );
	direction /= fLength;

	// compute normal of input line
	Math::Vector< 2, float > normal( -direction( 1 ), direction( 0 ) );

	// create buffer
	boost::scoped_array< int > pBuffer( new int[ nSearchPixels ] );

	// refine at some points on input line
	for ( int i = 0; i < nPoints; i++ )
	{
		// sample image perpendicular to input line with sobel operator
		float fPos = ( i + 2 ) * fLength / ( nPoints + 3 );
		Math::Vector< 2, float > startPoint( point1 + fPos * direction );

		int nMax;
		float fMaxSubPix = findEdge< FindEdgePositiveMaximum, ExtractLineSobel>
		//float fMaxSubPix = findEdge< FindEdgePositiveMaximum, ExtractLineSimpleGradient>
			( greyImage, startPoint, normal, nSearchPixels >> 1, nMax );

		if ( nMax >= g_nMinimalEdgeStrength )
		{
			// compute subpixel edge point
			pPoints[ iPoint ].x = startPoint( 0 ) + fMaxSubPix * normal( 0 );
			pPoints[ iPoint ].y = startPoint( 1 ) + fMaxSubPix * normal( 1 );
			pStrengths[ iPoint ] = nMax;

			iPoint++;
		}

		// debugging
		if ( pDebugImg )
			cvCircle( *pDebugImg, cvPoint( cvRound( pPoints[ iPoint ].x * 16 ), cvRound( pPoints[ iPoint ].y * 16 ) ),
				cvRound( 0.4 * 16 ), nMax < g_nMinimalEdgeStrength ? CV_RGB( 255, 0, 0 ) : CV_RGB( 0, 255, 0 ), -1, CV_AA, 4 );
	}

	return iPoint;
}


bool refineCorners( const Image& img, CornerList& corners, Image* pDebugImg )
{
	int nAvgEdgeStrength = 0;

	// compute bounding box
	CvPoint2D32f boundMin = cvPoint2D32f( corners[ 0 ][ 0 ], corners[ 0 ][ 1 ] );
	CvPoint2D32f boundMax = cvPoint2D32f( corners[ 0 ][ 0 ], corners[ 0 ][ 1 ] );
	for ( unsigned i = 1; i < corners.size(); i++ )
	{
		boundMin.x = std::min( boundMin.x, corners[ i ][ 0 ] );
		boundMin.y = std::min( boundMin.y, corners[ i ][ 1 ] );
		boundMax.x = std::max( boundMax.x, corners[ i ][ 0 ] );
		boundMax.y = std::max( boundMax.y, corners[ i ][ 1 ] );
	}

	// localize edges
	boost::scoped_array< float > lineParams( new float[ 4 * corners.size() ] );
	for ( unsigned i = 0; i < corners.size(); i++ )
	{
		unsigned iNext = ( i + 1 ) % corners.size();
		int nSearchPixels;
		if ( fabsf( corners[ i ][ 0 ] - corners[ iNext ][ 0 ] ) >
			fabsf( corners[ i ][ 1 ] - corners[ iNext ][ 1 ] ) )
			nSearchPixels = static_cast< int >( ( boundMax.y - boundMin.y ) * g_fEdgeSearchArea + 2 );
		else
			nSearchPixels = static_cast< int >( ( boundMax.x - boundMin.x ) * g_fEdgeSearchArea + 2 );

		// make sure this is an odd number
		nSearchPixels |= 1;

		// get some subpixel points along edge
		CvPoint2D32f points[ g_nEdgePoints ];
		int strengths[ g_nEdgePoints ];
		int nPoints = refineLinePoints( points, strengths, g_nEdgePoints, img,
			corners[ i ], corners[ iNext ],	nSearchPixels, pDebugImg );

		// update average edge strength (not detected edges get 0 weight)
		int nStrengthSum = 0;
		for ( int j = 0; j < nPoints; j++ )
			nStrengthSum += strengths[ j ];
		nAvgEdgeStrength += nStrengthSum / g_nEdgePoints;

		// fit line trough these points
		if ( nPoints >= g_nFitMinEdgePoints )
		{
			// TODO: here one could experiment with different fitting distances to make this more robust
			CvMat mat = cvMat( 1, nPoints, CV_32FC2, points );
			cvFitLine( &mat, CV_DIST_L2, 0, 0.01, 0.01, &lineParams[ 4 * i ] );
		}
		else
		{
			// not enough edge points detected -> do not fit to points
			lineParams[ 4 * i + 0 ] = corners[ i ][ 0 ] - corners[ iNext ][ 0 ];
			lineParams[ 4 * i + 1 ] = corners[ i ][ 1 ] - corners[ iNext ][ 1 ];
			lineParams[ 4 * i + 2 ] = corners[ i ][ 0 ];
			lineParams[ 4 * i + 3 ] = corners[ i ][ 1 ];
		}

		// draw refined lines and corners into the image -- is rather annoying
		if ( pDebugImg && false )
			cvLine( *pDebugImg, 
				cvPoint( cvRound( lineParams[ 4 * i + 2 ] + 100 * lineParams[ 4 * i + 0 ] ), cvRound( lineParams[ 4 * i + 3 ] + 100 * lineParams[ 4 * i + 1 ] ) ), 
				cvPoint( cvRound( lineParams[ 4 * i + 2 ] - 100 * lineParams[ 4 * i + 0 ] ), cvRound( lineParams[ 4 * i + 3 ] - 100 * lineParams[ 4 * i + 1 ] ) ), 
				CV_RGB( 0, 255, 255 ), 1, 4, 0 );
	}
		
	// now recompute corners by intersecting the edges
	for ( unsigned i = 0; i < corners.size(); i++ )
	{
		unsigned iNext = ( i + 1 ) % corners.size();
		float* pLine1 = &lineParams[ 4 * i ];
		float* pLine2 = &lineParams[ 4 * iNext ];

		float a = pLine1[ 0 ] * ( pLine2[ 3 ] - pLine1[ 3 ] ) - 
			pLine1[ 1 ] * ( pLine2[ 2 ] - pLine1[ 2 ] );
		float b = pLine2[ 0 ] * pLine1[ 1 ] - pLine1[ 0 ] * pLine2[ 1 ];
		
		// check if parallel
		if ( fabsf( a ) < fabsf( b * 10e6f ) )
		{
			corners[ iNext ][ 0 ] = pLine2[ 0 ] * a / b + pLine2[ 2 ];
			corners[ iNext ][ 1 ] = pLine2[ 1 ] * a / b + pLine2[ 3 ];
		}
	}
	
	if ( pDebugImg )
	{
		// draw corners as circles
		for ( unsigned i = 0; i < corners.size(); i++ ) 
			cvCircle( *pDebugImg, cvPoint( cvRound( corners[ i ][ 0 ] * 16 ), cvRound( corners[ i ][ 1 ] * 16 ) ),
				cvRound( 1.8f * 16 ), CV_RGB( i*80, 0, 0 ), -1, CV_AA, 4 );
	}

	return ( nAvgEdgeStrength / 4 ) > g_nMinimalEdgeStrength;
}
 

boost::shared_ptr< Image > getMarkerImage( const Image& img, const Math::Matrix< 3, 3, float > homography, int nSize )
	{
	// compute a suitable scaling matrix that maps 
	// for x: -0.5 -> -1 and (nSize-0.5) -> +1
	// for y: -0.5 -> +1 and (nSize-0.5) -> -1
	Math::Matrix< 3, 3, float > hScale;
	
	float m = 1.0f / nSize;
	hScale( 0, 0 ) = m;
	hScale( 0, 1 ) = 0.0f;
	hScale( 0, 2 ) = 0.5f * m - 0.5f;
	hScale( 1, 0 ) = 0.0f;
	hScale( 1, 1 ) = -hScale( 0, 0 );
	hScale( 1, 2 ) = -hScale( 0, 2 );
	hScale( 2, 0 ) = 0.0f;
	hScale( 2, 1 ) = 0.0f;
	hScale( 2, 2 ) = 1.0f;

	// transpose because of column major representation in Matrix
	Math::Matrix< 3, 3, float > Htrans = ublas::trans( ublas::prod( homography, hScale ) );	
	CvMat cvH = cvMat( 3, 3, CV_32FC1, Htrans.content() );
	
	// create target image
	boost::shared_ptr< Image > r( new Image( nSize, nSize, 1 ) );
	
	// apply homography
	cvWarpPerspective( img, *r, &cvH,
		CV_INTER_LINEAR + CV_WARP_FILL_OUTLIERS + CV_WARP_INVERSE_MAP, cvScalar( 128 ) );

	return r;
	}


unsigned readCode( const Image& markerImage )
	{
	// find threshold
	int nAvg = 0;
	const unsigned char* pData = reinterpret_cast< const unsigned char* >( markerImage.imageData ); 
	for ( int y = 0; y < markerImage.width; y++ )	
		{
		for ( int x = 0; x < markerImage.width; x++ )	
			nAvg += *pData++;
		pData += markerImage.widthStep - markerImage.width;
		}
		
	nAvg /= markerImage.width * markerImage.height;
	
	// check if border is black
	int nBlackBorderPixels = 0;
	pData = reinterpret_cast< const unsigned char* >( markerImage.imageData ); 
	for ( int i = 0; i < markerImage.width; i++ )
		{
		if ( pData[ i ] <= nAvg )
			nBlackBorderPixels++;
		if ( pData[ i + markerImage.widthStep * ( markerImage.height - 1 ) ] <= nAvg )
			nBlackBorderPixels++;
		}
	for ( int i = 1; i < markerImage.height - 1; i++ )
		{
		if ( pData[ i * markerImage.widthStep ] <= nAvg )
			nBlackBorderPixels++;
		if ( pData[ i * markerImage.widthStep + markerImage.width - 1 ] <= nAvg )
			nBlackBorderPixels++;
		}
		
	if ( nBlackBorderPixels < 2 * markerImage.width + 2 * ( markerImage.height - 2 ) - g_nMaxWhiteBorderPixels )
		return 0;
		
	// compute marker code
	unsigned nMarkerCode = 0;
	for ( int y = 1; y < markerImage.width - 1; y++ )	
		for ( int x = 1; x < markerImage.width - 1; x++ )
			if ( pData[ y * markerImage.widthStep + x ] <= nAvg )
				nMarkerCode |= 1 << ( ( markerImage.width - 2 - y ) * ( markerImage.width - 2 ) + ( markerImage.width - 2 - x ) );
				
	return nMarkerCode;
	}


/** 
 * \internal
 * rotates a marker code 90 degrees counter-clockwise.
 *
 * @param nCode code to rotate
 * @param codeLen side-length of marker code
 * @return rotated marker code
 */
unsigned rotateMarkerCode( unsigned nCode, int codeLen )
{
	unsigned nRotCode = 0;
	for ( int j = 0; j < codeLen * codeLen; j++ )
		if ( ( nCode & ( 1 << j ) ) != 0 )
			nRotCode |= 1 << ( codeLen * ( j % codeLen ) + codeLen - 1 - ( j / codeLen ) );
	return nRotCode;
}


unsigned normalizeCode( unsigned nCode, int codeSize, int& nRotations )
{
	// return the code with the smallest numeric representation
	unsigned nMinCode = nCode;
	unsigned rotCode = nCode;
	nRotations = 0;
	for ( int i = 0; i < 3; i++ )
	{
		rotCode = rotateMarkerCode( rotCode, codeSize );
		if ( rotCode < nMinCode )
		{
			nMinCode = rotCode;
			nRotations = i + 1;
		}
	}

	return nMinCode;
}


void computeMarkerEdgels( unsigned markerCode, float markerSize, 
	std::vector< Math::Vector< 3, float > >& points1, std::vector< Math::Vector< 3, float > >& points2 )
{
	float scale = markerSize / 6.0f;

	// start with outer border pixels
	for ( int i = 0; i < 6; i++ )
	{
		float p1 = ( -2.5f + i ) * scale;
		float p2 = ( -3.2f + i ) * scale;

		// left
		points1.push_back( Math::Vector< 3, float >( -3.0f * scale, p1, 0 ) );
		points2.push_back( Math::Vector< 3, float >( -3.0f * scale, p2, 0 ) );

		// right
		points1.push_back( Math::Vector< 3, float >( 3.0f * scale, -p1, 0 ) );
		points2.push_back( Math::Vector< 3, float >( 3.0f * scale, -p2, 0 ) );

		// top
		points1.push_back( Math::Vector< 3, float >( p1, 3.0f * scale, 0 ) );
		points2.push_back( Math::Vector< 3, float >( p2, 3.0f * scale, 0 ) );

		// bottom
		points1.push_back( Math::Vector< 3, float >( -p1, -3.0f * scale, 0 ) );
		points2.push_back( Math::Vector< 3, float >( -p2, -3.0f * scale, 0 ) );
	}

	// inner border pixels
	for ( int i = 0; i < 4; i++ )
	{
		float p = ( -1.5f + i ) * scale;
		float p1 = ( -2.2f + i ) * scale;
		float p2 = ( -0.8f + i ) * scale;

		// left
		if ( ! ( markerCode & ( 1 << ( i * 4 + 3 ) ) ) )
		{
			points1.push_back( Math::Vector< 3, float >( -2.0f * scale, p, 0 ) );
			points2.push_back( Math::Vector< 3, float >( -2.0f * scale, p2, 0 ) );
		}

		// right
		if ( ! ( markerCode & ( 1 << ( i * 4 ) ) ) )
		{
			points1.push_back( Math::Vector< 3, float >( 2.0f * scale, p, 0 ) );
			points2.push_back( Math::Vector< 3, float >( 2.0f * scale, p1, 0 ) );
		}

		// top
		if ( ! ( markerCode & ( 1 << ( i + 12 ) ) ) )
		{
			points1.push_back( Math::Vector< 3, float >( -p, 2.0f * scale, 0 ) );
			points2.push_back( Math::Vector< 3, float >( -p1, 2.0f * scale, 0 ) );
		}

		// bottom
		if ( ! ( markerCode & ( 1 << i ) ) )
		{
			points1.push_back( Math::Vector< 3, float >( -p, -2.0f * scale, 0 ) );
			points2.push_back( Math::Vector< 3, float >( -p2, -2.0f * scale, 0 ) );
		}
	}

	// inner edges
	for ( int y = 0; y < 3; y++ )
		for ( int x = 0; x < 3; x++ )
		{
			bool bSet = ( markerCode & ( 1 << ( y * 4 + x ) ) ) != 0;

			// left edge
			bool bLeftSet = ( markerCode & ( 1 << ( y * 4 + x + 1 ) ) ) != 0;
			if ( !bSet && bLeftSet )
			{
				points1.push_back( Math::Vector< 3, float >( ( 1.0f - x ) * scale, ( -1.5f + y ) * scale, 0 ) );
				points2.push_back( Math::Vector< 3, float >( ( 1.0f - x ) * scale, ( -0.8f + y ) * scale, 0 ) );
			}
			else if ( bSet && !bLeftSet )
			{
				points1.push_back( Math::Vector< 3, float >( ( 1.0f - x ) * scale, ( -1.5f + y ) * scale, 0 ) );
				points2.push_back( Math::Vector< 3, float >( ( 1.0f - x ) * scale, ( -2.2f + y ) * scale, 0 ) );
			}

			// down edge
			bool bDownSet = ( markerCode & ( 1 << ( y * 4 + x + 4 ) ) ) != 0;
			if ( !bSet && bDownSet )
			{
				points1.push_back( Math::Vector< 3, float >( ( 1.5f - x ) * scale, ( -1.0f + y ) * scale, 0 ) );
				points2.push_back( Math::Vector< 3, float >( ( 2.2f - x ) * scale, ( -1.0f + y ) * scale, 0 ) );
			}
			else if ( bSet && !bDownSet )
			{
				points1.push_back( Math::Vector< 3, float >( ( 1.5f - x ) * scale, ( -1.0f + y ) * scale, 0 ) );
				points2.push_back( Math::Vector< 3, float >( ( 0.8f - x ) * scale, ( -1.0f + y ) * scale, 0 ) );
			}
		}
}


Math::Pose alternateMarkerPose( const Math::Pose& p )
{
	namespace ublas = boost::numeric::ublas;

	// compute normal and viewing direction
	Math::Vector< 3 > normal( p.rotation() * Math::Vector< 3 >( 0, 0, 1 ) );
	Math::Vector< 3 > viewdir( -p.translation() / ublas::norm_2( p.translation() ) );
	
	// compute rotation
	Math::Vector< 3 > rotAxis( Math::cross_prod( normal, viewdir ) );
	double angle = asin( ublas::norm_2( rotAxis ) );
	Math::Quaternion q( rotAxis, -2 * angle );

	// add rotation
	return Math::Pose( q * p.rotation(), p.translation() );
}


void drawCube( Vision::Image& img, const Math::Pose& pose, const Math::Matrix< 3, 3, float >& K, double markerSize, CvScalar color )
{
	// the corner points
	static float points3D[ 8 ][ 3 ] = {
		{ 0.5f, 0.5f, 0.0f }, { -0.5f, 0.5f, 0.0f }, { -0.5f, -0.5f, 0.0f }, { 0.5f, -0.5f, 0.0f },
		{ 0.5f, 0.5f, 1.0f }, { -0.5f, 0.5f, 1.0f }, { -0.5f, -0.5f, 1.0f }, { 0.5f, -0.5f, 1.0f }
	};

	// project points
	Math::Vector< 3, double > p2D[ 8 ];
	for ( int i = 0; i < 8; i++ )
	{
		p2D[ i ] = ublas::prod( K, pose * Math::Vector< 3 >( markerSize * Math::Vector< 3, float >( points3D[ i ] ) ) );
		p2D[ i ] /= p2D[ i ][ 2 ];
	}

	// draw some lines
	for ( int c = 0; c < 8; c += 4 )
		for ( int i = 0; i < 4; i++ )
			cvLine( img, cvPoint( cvRound( p2D[ c + i ][ 0 ] * 16 ), cvRound( p2D[ c + i ][ 1 ] * 16 ) ),
				cvPoint( cvRound( p2D[ c + ( i + 1 ) % 4 ][ 0 ] * 16 ), cvRound( p2D[ c + ( i + 1 ) % 4 ][ 1 ] * 16 ) ),
				color, 1, CV_AA, 4 );
	for ( int i = 0; i < 4; i++ )
		cvLine( img, cvPoint( cvRound( p2D[ i ][ 0 ] * 16 ), cvRound( p2D[ i ][ 1 ] * 16 ) ),
			cvPoint( cvRound( p2D[ i + 4 ][ 0 ] * 16 ), cvRound( p2D[ i + 4 ][ 1 ] * 16 ) ),
			i == 0 ? CV_RGB( 0, 0, 0 ) : color, 1, CV_AA, 4 );
}


} } } // namespace Ubitrack::Vision::Markers
