/*
 * Ubitrack - Library for Ubiquitous Tracking
 * Copyright 2006, Technische Universitaet Muenchen, and individual
 * contributors as indicated by the @authors tag. See the 
 * copyright.txt in the distribution for a full listing of individual
 * contributors.
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA, or see the FSF site: http://www.fsf.org.
 */

/**
 * @ingroup vision
 * @file
 * Routines to detect square markers in images.
 *
 * @author Daniel Pustka <daniel.pustka@in.tum.de>
 */

#ifndef __UBITRACK_VISION_MARKERDETECTION_H_INCLUDED__
#define __UBITRACK_VISION_MARKERDETECTION_H_INCLUDED__

#include <list>
#include <vector>
#include <map>
#include "../Ubitrack/config.h"
#include <Vision/Vision.h>
#include <Ubitrack/Math/Vector.h>
#include <Ubitrack/Math/Matrix.h>
#include <Ubitrack/Math/Pose.h>
#include "Image.h"

namespace Ubitrack { namespace Vision { namespace Markers {

/** a list of corner points */
typedef std::vector< Ubitrack::Math::Vector< 2, float > > CornerList;

/** a list of polygons */
typedef std::list< CornerList > MarkerList;

/** information about a marker */
struct MarkerInfo
{
	/** constructor */
	MarkerInfo( float _fSize = 0 )
		: fSize( _fSize )
		, refinement( EEdgeRefinedPose )
		, bFound( false )
		, bUseInitialPose( false )
		, bCalculateCovariance( false )
	{}

	/** marker size in meters */
	float fSize;

	/** desired level of refinement */
	enum 
	{ 
		ECorners, // only corner location
		EInitialPose, // also perform initial linear pose estimation
		ERefinedPose, // refined non-linear pose estimation
		EEdgeRefinedPose // even more refined pose estimation that takes inner-marker edges into account
	}
	refinement;

	/** was the marker found in this image */
	bool bFound;

	/** is the pose initialized for a previous frame? */
	bool bUseInitialPose;

	/** calculate covariance? */
	bool bCalculateCovariance;

	/** found marker corners */
	CornerList corners;

	/** pose of the marker */
	Math::Pose pose;

	/** covariance matrix */
	Math::Matrix< 6, 6 > covariance;
};

#ifdef HAVE_LAPACK
/**
 * @ingroup vision
 * Detects markers in the image.
 *
 * @param img Greyscale image
 * @param markers Set of markers to detect. The map key is the normalized code of each marker. If the marker is detected,
 *    the method will set the bFound flag and, depending on the refinement value, also the corners and pose value.
 *    If a marker of code "0000" is present in the map, the method will detect ALL markers in the image and add them to the map.
 * @param K camera intrinsics matrix
 * @param pDebugImg creates a debug image if non-null
 */
UTVISION_EXPORT void detectMarkers( const Image& img, std::map< unsigned, MarkerInfo >& markers, 
	const Math::Matrix< 3, 3, float >& K, Image* pDebugImg = 0 );
#endif

/**
 * @ingroup vision
 * Returns a list of all quadrangles found in the image.
 *
 * For better accuracy, the corner coordinates should be refined using refineCorners afterwards.
 * This function does not respect the origin flag and assumes a y-coordinate of 0 is top.
 *
 * @param img thresholded grey-scale input image. Note: The image will be destructively modified by this routine.
 * @return list of found markers in image
 */
UTVISION_EXPORT MarkerList findQuadrangles( Image& img );

/**
 * @ingroup vision
 * Refines marker corners with sub-pixel accuracy.
 *
 * @param img grey-scale camera image
 * @param list of corners to be refined in clock-wise order
 * @param pDebugImg. If not NULL, some information will be drawn into this image for debugging
 * @return true if edges were considered strong enough for this to be a real marker
 */
UTVISION_EXPORT bool refineCorners( const Image& img, CornerList& corners, Image* pDebugImg = NULL );

/**
 * @ingroup vision
 * Returns a rectified image of a marker.
 *
 * @param img the grey-scale camera image
 * @param homography a homography as computed by Ubitrack::Calibration::squareHomograph
 * @param nSize sidelength of the resulting image in pixels
 */
UTVISION_EXPORT boost::shared_ptr< Image > getMarkerImage( const Image& img, const Math::Matrix< 3, 3, float > homography, int nSize );

/**
 * @ingroup vision
 * Reads the marker code.
 *
 * @param img image of the marker
 * @return unnormalized marker code, 0 if marker code could not be read
 */
UTVISION_EXPORT unsigned readCode( const Image& img );

/**
 * @ingroup vision
 * Normalizes a marker code.
 *
 * @param nCode the unnormalized code
 * @param codeSize side-length of code pattern
 * @param nRotations returns the number of counter-clockwise marker rotations used
 * @return normalized marker code
 */
UTVISION_EXPORT unsigned normalizeCode( unsigned nCode, int codeSize, int& nRotations );

/**
 * @ingroup vision
 * Computes edgels for a marker consisting of inner and outer edges
 *
 * @param markerCode the code of the marker
 * @param markerSize size of the marker
 * @param points1 a list of 3-vectors that will be filled with the center-points of each edgel
 * @param points2 a list of 3-vectors that will be filled with another point on each edgel
 */
UTVISION_EXPORT void computeMarkerEdgels( unsigned markerCode, float markerSize, 
	std::vector< Math::Vector< 3, float > >& points1, std::vector< Math::Vector< 3, float > >& points2 );

/**
 * @ingroup vision
 * Computes an alternate marker pose.
 *
 * Sometimes the pose estimation process runs into a wrong local minimum. This function 
 * computes the pose that roughly corresponds to the other minimum. The resulting pose is
 * meant to be refined by the iterative pose estimation again. If the residual is smaller 
 * than for the original pose, use this one.
 *
 * @param p pose computed by the first try
 * @return alternate marker pose
 */
UTVISION_EXPORT Math::Pose alternateMarkerPose( const Math::Pose& p );

} } } // namespace Ubitrack::Vision::Markers

#endif

