/*
 * Ubitrack - Library for Ubiquitous Tracking
 * Copyright 2006, Technische Universitaet Muenchen, and individual
 * contributors as indicated by the @authors tag. See the
 * copyright.txt in the distribution for a full listing of individual
 * contributors.
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA, or see the FSF site: http://www.fsf.org.
 */

/**
 * @ingroup vision_components
 * @file
 * All-in-one component for ISMAR '07 demo
 *   - Frame grabbing using DsVideoLib
 *   - Reading intrinsic camera parameters
 *   - Undistortion
 *
 * @author Daniel Pustka <daniel.pustka@in.tum.de>
 */

#include <string>
#include <list>
#include <iostream>
#include <strstream>
#include <log4cpp/Category.hh>
#include <fstream>
#include <cv.h>

#include <boost/thread.hpp>
#include <boost/bind.hpp>
#include <boost/scoped_ptr.hpp>
#include <boost/archive/text_iarchive.hpp>

#include <Ubitrack/Measurement/Measurement.h>
#include <Ubitrack/Dataflow/PushSupplier.h>
#include <Ubitrack/Dataflow/PullSupplier.h>
#include <Ubitrack/Dataflow/Component.h>
#include <Ubitrack/Dataflow/ComponentFactory.h>
#include <Ubitrack/Measurement/TimestampSync.h>
#include <Vision/Image.h>

#include <Ubitrack/Util/CleanWindows.h>
#ifndef _WIN32_WINNT
	#define _WIN32_WINNT WINVER
#endif
#include <objbase.h>
#include <DSVL.h>
#include <highgui.h>

// get a logger
static log4cpp::Category& logger( log4cpp::Category::getInstance( "Drivers.DsvlCamera" ) );

using namespace Ubitrack;
using namespace Ubitrack::Vision;

namespace Ubitrack { namespace Drivers {

/**
 * All-in-one component for ISMAR '07 demo
 *   - Frame grabbing using DsVideoLib
 *   - Reading intrinsic camera parameters
 *   - Undistortion
 *
 * @par Input Ports
 * None.
 *
 * @par Output Ports
 * \c Output push port of type Ubitrack::Vision::ImageMeasurement.
 * \c Intrinsics 3x3 camera matrix as pull
 *
 * @par Configuration
 *
 * Examples:
\verbatim
<DataflowConfiguration>
  <Attribute name="CameraName" value="Logitech"/>
  <Attribute name="IntrinsicMatrixFile" value="matrix.calib"/>
  <Attribute name="DistortionFile" value="coeffs.calib"/>
  <Attribute name="timeOffset" value="0"/>
  <Attribute name="divisor" value="1"/>
</DataflowConfiguration>
\endverbatim
 *
 */
class DsvlCamera
	: public Dataflow::Component
{
public:

	/** constructor */
	DsvlCamera( const std::string& sName, boost::shared_ptr< Graph::UTQLSubgraph >  );

	/** destructor, waits until thread stops */
	~DsvlCamera();

protected:
	/** reads from a calibration file */
	template< class MT >
	void readCalibData( MT& data, const std::string& sFile )
	{
		// create ifstream
		std::ifstream stream( sFile.c_str() );
		if ( !stream.good() )
			UBITRACK_THROW( "Could not open file " + sFile + " for reading" );

		// create iarchive
		boost::archive::text_iarchive archive( stream );
		
		// read data
		archive >> data;
	}
	
	/** handler method for incoming pull requests */
	Measurement::Matrix3x3 getIntrinsic( Measurement::Timestamp t )
	{ return Measurement::Matrix3x3( t, m_intrinsicMatrix ); }

	// undistortion
	void initMap( bool bOrigin, int width, int height, const Math::Vector< 4 >& coeffs, const Math::Matrix< 3, 3 >& intrinsics );
	boost::shared_ptr< Vision::Image > undistort( boost::shared_ptr< Vision::Image > pImage );

	Math::Matrix< 3, 3 > m_intrinsicMatrix;
	Math::Vector< 4 > m_distortionCoefficients;
	std::string m_cameraName;
	
	// width of resulting image
	LONG m_width;

	// height of resulting image
	LONG m_height;

	// shift timestamps (ms)
	int m_timeOffset;

	// only send every nth image
	int m_divisor;

	// pointer to DSVideoLib VideoSource
	boost::scoped_ptr< DSVL_VideoSource > m_pVideoSource;

	// undistortion map
	boost::scoped_ptr< Image > m_pMapX;
	boost::scoped_ptr< Image > m_pMapY;

	// thread main loop
	void ThreadProc();

	// the thread
	boost::scoped_ptr< boost::thread > m_Thread;

	// stop the thread?
	volatile bool m_bStop;

	// the ports
	Dataflow::PushSupplier< ImageMeasurement > m_outPort;
	Dataflow::PushSupplier< ImageMeasurement > m_colorOutPort;
	Dataflow::PullSupplier< Measurement::Matrix3x3 > m_outIntrinsic;

};


DsvlCamera::DsvlCamera( const std::string& sName, boost::shared_ptr< Graph::UTQLSubgraph > subgraph )
	: Dataflow::Component( sName )
	, m_timeOffset( 0 )
	, m_divisor( 1 )
	, m_bStop( false )
	, m_outPort( "Output", *this )
	, m_colorOutPort( "ColorOutput", *this )
	, m_outIntrinsic( "Intrinsics", *this, boost::bind( &DsvlCamera::getIntrinsic, this, _1 ) )
{
	// read configuration
	subgraph->m_DataflowAttributes.getAttributeData( "timeOffset", m_timeOffset );
	subgraph->m_DataflowAttributes.getAttributeData( "divisor", m_divisor );

	// camera name
	if ( subgraph->m_DataflowAttributes.hasAttribute( "cameraName" ) )
		m_cameraName = subgraph->m_DataflowAttributes.getAttributeString( "cameraName" );

	// load intrinsic matrix
	if ( subgraph->m_DataflowAttributes.hasAttribute( "intrinsicMatrixFile" ) )
	{
		Measurement::Matrix3x3 measMat;
		measMat.reset( new Math::Matrix< 3, 3 >() );
		readCalibData( measMat, subgraph->m_DataflowAttributes.getAttributeString( "intrinsicMatrixFile" ) );
		m_intrinsicMatrix = *measMat;
	}
	else
	{
		m_intrinsicMatrix( 0, 0 ) = 400;
		m_intrinsicMatrix( 0, 1 ) = 0;
		m_intrinsicMatrix( 0, 2 ) = -160;
		m_intrinsicMatrix( 1, 0 ) = 0;
		m_intrinsicMatrix( 1, 1 ) = 400;
		m_intrinsicMatrix( 1, 2 ) = -120;
		m_intrinsicMatrix( 2, 0 ) = 0;
		m_intrinsicMatrix( 2, 1 ) = 0;
		m_intrinsicMatrix( 2, 2 ) = -1;
	}

	// load distortion file
	if ( subgraph->m_DataflowAttributes.hasAttribute( "distortionFile" ) )
	{
		Measurement::Vector4D measVec;
		measVec.reset( new Math::Vector< 4 >() );
		readCalibData( measVec, subgraph->m_DataflowAttributes.getAttributeString( "distortionFile" ) );
		m_distortionCoefficients = *measVec;
	}
	else
		m_distortionCoefficients = boost::numeric::ublas::zero_vector< double >( 4 );

	// build DSVL config string
	std::string sConfig = "<dsvl_input><camera frame_width=\"320\" frame_height=\"240\" frame_rate=\"30\" ";
	if ( !m_cameraName.empty() )
		sConfig += "friendly_name=\"" + m_cameraName + "\"";
	sConfig += "><pixel_format><RGB24/></pixel_format></camera></dsvl_input>";
	LOG4CPP_DEBUG( logger, "DSVideoLibConfiguration: >>" << sConfig << "<<" );

	if ( FAILED( CoInitializeEx( NULL, COINIT_MULTITHREADED ) ) )
		UBITRACK_THROW( "Error in CoInitializeEx" );
	m_pVideoSource.reset( new DSVL_VideoSource() );

	if ( FAILED( m_pVideoSource->BuildGraphFromXMLString( const_cast< char* >( sConfig.c_str() ) ) ) )
		UBITRACK_THROW( "Invalid DSVideoLib configuration string: " + sConfig );

	PIXELFORMAT pxf;
	double cap_fps;
	if ( FAILED( m_pVideoSource->GetCurrentMediaFormat( &m_width, &m_height, &cap_fps, &pxf ) ) )
		UBITRACK_THROW( "DSVideoLib initialization problem" );
	LOG4CPP_DEBUG( logger, "Frame dimensions: " << m_width << "x" << m_height )

	if ( pxf != PIXELFORMAT_RGB24 )
		UBITRACK_THROW( "DSVideoLib driver only supports RGB24 format" );

	if ( FAILED( m_pVideoSource->EnableMemoryBuffer( 3 ) ) )
		UBITRACK_THROW( "DSVideoLib EnableMemoryBuffer failed" );

	// start thread
	m_Thread.reset( new boost::thread( boost::bind( &DsvlCamera::ThreadProc, this ) ) );
}


DsvlCamera::~DsvlCamera()
{
	if ( m_Thread )
	{
		m_bStop = true;
		m_Thread->join();
	}
}


void DsvlCamera::ThreadProc()
{
	LOG4CPP_DEBUG( logger, "Thread started" );

	if ( FAILED( m_pVideoSource->Run() ) )
	{
		LOG4CPP_ERROR( logger, "DSVideoLib failed to Run()" );
		return;
	}

	Measurement::TimestampSync syncer( 1e7 ); // timestamps are in 100ns-units
	bool bPrintedTimestampWarning = false;
	LONGLONG lastTimestamp = 13;
	unsigned lastFrame = 29645234;
	unsigned nFrames = 0;

	while ( !m_bStop )
	{
		MemoryBufferHandle l_mbHandle;
		BYTE* l_pPixelBuffer = NULL;

		m_pVideoSource->WaitForNextSample( 250 );
		Measurement::Timestamp time = Measurement::now();

		if ( SUCCEEDED( m_pVideoSource->CheckoutMemoryBuffer( &l_mbHandle, &l_pPixelBuffer ) ) )
		{
			LOG4CPP_DEBUG( logger, "CheckoutMemoryBuffer successful: (" << l_mbHandle.n << ", " << l_mbHandle.t << ")" );

			if ( l_mbHandle.n == lastFrame )
			{
				// DSVideoLib sometimes sends frames multiple times, especially, when using more than one camera...
				LOG4CPP_DEBUG( logger, "Not sending repeated frame" );
			}
			else
			{
				lastFrame = l_mbHandle.n;

				// create timestamp
				if ( l_mbHandle.t <= lastTimestamp )
				{
					if ( !bPrintedTimestampWarning )
					{
						LOG4CPP_WARN( logger, "DSVideoLib timestamps do not work. Try a patched version." );
						bPrintedTimestampWarning = true;
					}
				}
				else
				{
					time = syncer.convertNativeToLocal( double( l_mbHandle.t ), time );
					lastTimestamp = l_mbHandle.t;
				}

				// create an IplImage buffer for the frame
				Vision::Image camBuffer( m_width, m_height, 3, static_cast< void* >( l_pPixelBuffer ), IPL_DEPTH_8U, 1 );

				// send color image if anyone is interested
				if ( m_running && !( nFrames % m_divisor ) && m_colorOutPort.isConnected() )
					m_colorOutPort.send( ImageMeasurement( time + m_timeOffset * 1000000LL, camBuffer.Clone() ) );

				// send image
				if ( m_running && !( nFrames++ % m_divisor ) )
				{
					// convert to greyscale
					boost::shared_ptr< Image > pImage( new Image( camBuffer.width, camBuffer.height, 1 ) );
					cvConvertImage( camBuffer, *pImage, camBuffer.origin ? CV_CVTIMG_FLIP : 0 );

					// undistort
					pImage = undistort( pImage );

					m_outPort.send( ImageMeasurement( time + m_timeOffset * 1000000LL, pImage ) );
				}
			}

			// release buffer
			m_pVideoSource->CheckinMemoryBuffer( l_mbHandle );
		}
		else
			LOG4CPP_DEBUG( logger, "CheckoutMemoryBuffer failed" );
	}

	m_pVideoSource->Stop();

	LOG4CPP_DEBUG( logger, "Thread stopped" );
}


void DsvlCamera::initMap( bool bOrigin, int width, int height, const Math::Vector< 4 >& coeffs, const Math::Matrix< 3, 3 >& intrinsics )
{
	// skip if already initialized with same values
	if ( m_pMapX && m_pMapX->width == width && m_pMapX->height == height )
		return;
		
	LOG4CPP_INFO( logger, "Creating undistortion map" );
	LOG4CPP_DEBUG( logger, "coeffs=" << coeffs );
	LOG4CPP_DEBUG( logger, "intrinsic=" << intrinsics );

	// copy ublas to OpenCV parameters
	CvMat* pCvCoeffs = cvCreateMat( 1, 4, CV_32FC1 );
	for ( unsigned i = 0; i < 4; i++ )
		pCvCoeffs->data.fl[ i ] = static_cast< float >( coeffs( i ) );
		
	CvMat* pCvIntrinsics = cvCreateMat( 3, 3, CV_32FC1 );
	for ( unsigned i = 0; i < 3; i++ )
		for ( unsigned j = 0; j < 3; j++ )
			reinterpret_cast< float* >( pCvIntrinsics->data.ptr + i * pCvIntrinsics->step)[ j ] 
				= static_cast< float >( intrinsics( i, j ) ) * ( j == 2 ? -1 : 1 ); // compensate for left-handed OpenCV coordinate frame
	
	// compensate if origin==0
	if ( !bOrigin )
	{
		reinterpret_cast< float* >( pCvIntrinsics->data.ptr + 1 * pCvIntrinsics->step )[ 2 ] = 
			height - 1 - reinterpret_cast< float* >( pCvIntrinsics->data.ptr + 1 * pCvIntrinsics->step )[ 2 ];
		pCvCoeffs->data.fl[ 2 ] *= -1;
	}

	// create map images
	m_pMapX.reset( new Image( width, height, 1, IPL_DEPTH_32F ) );
	m_pMapY.reset( new Image( width, height, 1, IPL_DEPTH_32F ) );
	
	cvInitUndistortMap( pCvIntrinsics, pCvCoeffs, *m_pMapX, *m_pMapY );
	
	LOG4CPP_DEBUG( logger, "first pixel mapped from " << 
		*reinterpret_cast< float* >( m_pMapX->imageData ) << ", " <<
		*reinterpret_cast< float* >( m_pMapY->imageData ) );
		
	// release data
	cvReleaseMat( &pCvCoeffs );
	cvReleaseMat( &pCvIntrinsics );
}


boost::shared_ptr< Vision::Image > DsvlCamera::undistort( boost::shared_ptr< Vision::Image > pImage )
{
	if ( m_distortionCoefficients( 0 ) == 0 )
		return pImage;

	// initialize the distortion map
	initMap( pImage->origin != 0, pImage->width, pImage->height, m_distortionCoefficients, m_intrinsicMatrix );
	
	// undistort
	boost::shared_ptr< Vision::Image > pUndistorted( 
		new Vision::Image( pImage->width, pImage->height, pImage->nChannels, pImage->depth ) );
	pUndistorted->origin = pImage->origin;

	cvRemap( *pImage, *pUndistorted, *m_pMapX, *m_pMapY );
	
	// send result
	return pUndistorted;
}


} } // namespace Ubitrack::Driver

UBITRACK_REGISTER_COMPONENT( Dataflow::ComponentFactory* const cf ) {
	cf->registerComponent< Ubitrack::Drivers::DsvlCamera > ( "DsvlCamera" );
}

