/*
 * Ubitrack - Library for Ubiquitous Tracking
 * Copyright 2006, Technische Universitaet Muenchen, and individual
 * contributors as indicated by the @authors tag. See the
 * copyright.txt in the distribution for a full listing of individual
 * contributors.
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA, or see the FSF site: http://www.fsf.org.
 */


/**
 * @ingroup driver_components
 * @file
 * Driver for the EasyLase USB Laser Scanner
 *
 * Example Node Configuration:
 * @code
 * <Node name="VirtualSquare"/>
 * <Node name="LaserProjector">
 *   <DataflowConfiguration class="EasyLaseModule">
 *     <LaserConfiguration num="0" pps="2000"/>
 *   </DataflowConfiguration>
 * </Node>
 * @endcode
 * Example Edge Configuration:
 * @code
 * <Edge name="VirtualSquareTransform" source="VirtualSquare" destination="LaserProjector">
 *   <Attribute name="type" value="3D"/>
 *   <Dataflow class="EasyLaseModule">
 *
 *     <!-- transformation between virtual object and laser projector -->
 *     <Input port="Input" edge="SquarePosition"/>
 *
 *     <!-- pattern to display. currently, we only support triangle, square, cross and hexagon -->
 *     <Configuration>
 *       <Pattern name="square"/>
 *     </Configuration>
 *
 *   </Dataflow>
 * </Edge>
 * @endcode
 * @author Florian Echtler <echtler@in.tum.de>
 */


#ifndef __EasyLaseModule_h_INCLUDED__
#define __EasyLaseModule_h_INCLUDED__

#include <string>
#include <cstdlib>

#include <boost/bind.hpp>

#include <Ubitrack/Dataflow/PushConsumer.h>
#include <Ubitrack/Dataflow/Component.h>
#include <Ubitrack/Dataflow/Module.h>
#include <Ubitrack/Measurement/Measurement.h>
#include <Ubitrack/Math/Vector.h>

#include "easylase.h"


using namespace Ubitrack::Dataflow;


namespace Ubitrack { namespace Drivers {

// forward declaration
class LaserObject;


/**
 * @ingroup driver_components
 * Module key for viewer component.
 * Represents the used laser projector.
 */
class LaserProjectorKey
{
public:

	LaserProjectorKey( boost::shared_ptr< Graph::UTQLSubgraph > subgraph )
	{
		m_num = 0;
		m_pps = 20000;

		Graph::UTQLSubgraph::NodePtr artNode = subgraph->getNode( "Laser" );
		const TiXmlElement* config = artNode->getAttribute( "LaserConfiguration" ).getXML();
		if ( config == NULL ) return;

		const TiXmlElement* elem = config->ToElement();
		if ( elem == NULL ) return;

		const char* sNum = elem->Attribute( "num" );
		if (sNum) m_num = atoi( sNum );

		const char* sPPS = elem->Attribute( "pps" );
		if (sPPS) m_pps = atoi( sPPS );
	}

	/** less than operator for std::map */
	bool operator<( const LaserProjectorKey& b ) const
	{
		return m_num < b.m_num;
	}

	int m_num, m_pps;
};


/**
 * @ingroup driver_components
 * Component key for laser objects.
 * Represents the pointset.
 */
class LaserObjectKey
{
public:

	LaserObjectKey( boost::shared_ptr< Graph::UTQLSubgraph > subgraph )
		: m_name( "" )
	{
		const TiXmlElement* conf = subgraph->m_DataflowConfiguration.getXML();
		if (conf == NULL)
			UBITRACK_THROW( "LaserObject component without configuration." );

		const TiXmlNode* node = conf->FirstChild( "Pattern" );
		if (node == NULL)
			UBITRACK_THROW( "LaserObject component without XML element \"Pattern\"." );

		const TiXmlElement* elem = node->ToElement();
		if (elem == NULL)
			UBITRACK_THROW( "LaserObject component: can't convert node to element??" );

		const char* name = elem->Attribute( "name" );
		if (name == NULL)
			UBITRACK_THROW( "LaserObject component without name attribute in Pattern element." );

		m_name = name;
	}

	/** less than operator for std::map */
	bool operator<( const LaserObjectKey& b ) const
	{
		return m_name < b.m_name;
	}

	std::string m_name;

};


/**
 * @ingroup driver_components
 * Module for laser projector.
 * Does hardware abstraction and combines objects into single laser frame.
 */
class LaserProjector
	: public Module< LaserProjectorKey, LaserObjectKey, LaserProjector, LaserObject >
{
public:

	/**
	 * Constructor
	 * @param key which LaserProjector am I?
	 * @param pFactory internal parameter from ComponentFactory
	 */
	LaserProjector( const LaserProjectorKey& key, boost::shared_ptr< Graph::UTQLSubgraph > subgraph , FactoryHelper* pFactory );

	/** Destructor */
	~LaserProjector();

	/** callback from the models if position has changed, causes retransmit of frame */
	void redraw();

protected:

	int m_num, m_pps;
	std::vector<LaserPoint> m_frame;
};

inline unsigned short clamp( unsigned short val, unsigned short min, unsigned short max ) { return (val < min ? min : (val > max ? max : val)); }

/**
 * @ingroup driver_components
 * Component for laser-projected objects.
 * Provides a push-in port for positions.
 */
class LaserObject
	: public LaserProjector::Component
{
public:

	/**
	 * Constructor
	 * @param name edge name
	 * @param subgraph UTQL subgraph
	 * @param componentKey the unique identifier for this component
	 * @param pModule parent object
	 */
	LaserObject( const std::string& name, boost::shared_ptr< Graph::UTQLSubgraph > subgraph, const LaserObjectKey& componentKey, LaserProjector* pModule )
		: LaserProjector::Component( name, componentKey, pModule )
		, m_push ( "Input", *this, boost::bind( &LaserObject::positionIn, this, _1 ))
		, m_changed(1)
		, m_pattern()
		, m_output()
	{
		if (!componentKey.m_name.compare(0,8,"triangle")) {
			m_pattern.push_back(LaserPoint(2000,2000));
			m_pattern.push_back(LaserPoint(2096,2000));
			m_pattern.push_back(LaserPoint(2048,2096));
			m_pattern.push_back(LaserPoint(2000,2000));
		}
		if (!componentKey.m_name.compare(0,6,"square")) {
			m_pattern.push_back(LaserPoint(2000,2000));
			m_pattern.push_back(LaserPoint(2000,2096));
			m_pattern.push_back(LaserPoint(2096,2096));
			m_pattern.push_back(LaserPoint(2096,2000));
			m_pattern.push_back(LaserPoint(2000,2000));
		}
		if (!componentKey.m_name.compare(0,7,"hexagon")) {
			m_pattern.push_back(LaserPoint(2072,2096));
			m_pattern.push_back(LaserPoint(2096,2048));
			m_pattern.push_back(LaserPoint(2072,2000));
			m_pattern.push_back(LaserPoint(2024,2000));
			m_pattern.push_back(LaserPoint(2000,2048));
			m_pattern.push_back(LaserPoint(2024,2096));
			m_pattern.push_back(LaserPoint(2072,2096));
		}
		if (!componentKey.m_name.compare(0,5,"cross")) {
			m_pattern.push_back(LaserPoint(2048,2096));
			m_pattern.push_back(LaserPoint(2048,2000));
			m_pattern.push_back(LaserPoint(2000,2048));
			m_pattern.push_back(LaserPoint(2096,2048));
		}
	}

	std::vector<LaserPoint>* getPattern() {
		if (m_changed) {
			m_changed = 0;
			m_output.clear();
			for (std::vector<LaserPoint>::iterator it = m_pattern.begin(); it != m_pattern.end(); it++) {
				LaserPoint tmp = *it;
				// TODO: add depth handling
				tmp.x = clamp( (unsigned short)(tmp.x + m_pos[0]), 200, 4095 );
				tmp.y = clamp( (unsigned short)(tmp.y + m_pos[1]), 200, 4095 );
				m_output.push_back(tmp);
			}
		}
		return &m_output;
	}

protected:

	/**
	 * callback from Pose port
	 * @param pose current transformation
	 */
	void positionIn( const Ubitrack::Measurement::Position& pos )
	{
		m_pos = *pos;
		m_changed = 1;
		// the position has changed, so redraw the frame
		m_pModule->redraw();
	}

	// position input
	PushConsumer< Measurement::Position > m_push;
	Math::Vector<3> m_pos;
	int m_changed;

	// the pattern to display
	std::vector<LaserPoint> m_pattern;
	std::vector<LaserPoint> m_output;
};

} } // namespace Ubitrack::Drivers

#endif
