/*
* Ubitrack - Library for Ubiquitous Tracking
* Copyright 2006, Technische Universitaet Muenchen, and individual
* contributors as indicated by the @authors tag. See the
* copyright.txt in the distribution for a full listing of individual
* contributors.
*
* This is free software; you can redistribute it and/or modify it
* under the terms of the GNU Lesser General Public License as
* published by the Free Software Foundation; either version 2.1 of
* the License, or (at your option) any later version.
*
* This software is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
* Lesser General Public License for more details.
*
* You should have received a copy of the GNU Lesser General Public
* License along with this software; if not, write to the Free
* Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
* 02110-1301 USA, or see the FSF site: http://www.fsf.org.
*/


/**
* @ingroup dataflow_components
* @file
* Components that pulls its input at a given frequency and pushes the result
*
* @author Bjoern Schwerdtfeger <schwerdt@in.tum.de>
*/

#include "HALCONComponent.h"

namespace Ubitrack { namespace Drivers {

	/**
	* @ingroup dataflow_components
	* Components that pulls its input at a given frequency and pushes the result
	*
	* @par Input Ports
	* PullConsumer< EventType > with name "Input".
	*
	* @par Output Ports
	* PushSupplier< EventType > with name "Output".
	*
	* @par Configuration
	* @verbatim <Configuration frequency="100"/> @endverbatim
	* \c frequency: floating point number giving the sampling frequency in Hz
	*
	* @par Operation
	*
	* @par Instances
	*
	*/
	
	HALCONComponent::~HALCONComponent()
	{
		stop();
	}

	/**
	* component start method, starts thread
	*/
	void HALCONComponent::start()
	{
		if ( !m_running )
		{
			m_running = true;
			m_bStop = false;
			m_pThread.reset( new boost::thread( boost::bind( &HALCONComponent::threadMethod, this ) ) );
		}
	}

	/**
	* component stop method, stops thread
	*/
	void HALCONComponent::stop()
	{
		// stop thread
		// mutex?

		//close_all_framegrabbers();
		if ( m_running )
		{
			m_running = false;
			m_bStop = true;
			if ( m_pThread )
			{
				m_pThread->join();
			}
		}
	}

	void HALCONComponent::readDetectionAndTrackingParameters (Halcon::HTuple Filename, Halcon::HTuple *detectionParameters, 
	Halcon::HTuple *trackingParameters)
	{
		using namespace Halcon;

		// Local control variables 
		HTuple  FileHandle, Index, tempString, Real, justAComment;
		HTuple  IsEOF;

		//read tracking and detection parameters from file
		open_file(Filename, "input", &FileHandle);
		//detectionParameters := []
		for (Index=0; Index<=10; Index+=1)
		{
			fread_string(FileHandle, &justAComment, &IsEOF);
			fread_string(FileHandle, &tempString, &IsEOF);
			(*detectionParameters)[Index] = tempString;
		}
		tuple_number((*detectionParameters), &(*detectionParameters));
		tuple_real((*detectionParameters), &Real);
		//trackingParameters := []
		for (Index=0; Index<=10; Index+=1)
		{
			fread_string(FileHandle, &justAComment, &IsEOF);
			fread_string(FileHandle, &tempString, &IsEOF);
			(*trackingParameters)[Index] = tempString;
		}
		tuple_number((*trackingParameters), &(*trackingParameters));
		tuple_real((*trackingParameters), &Real);
		close_file(FileHandle);
		return;
	}
	
	void HALCONComponent::threadMethod()
	{
		Measurement::Timestamp step = Measurement::Timestamp( 1 / m_fFrequency * 1e9 );
		Measurement::Timestamp now( Measurement::now() );
		Measurement::Timestamp nextTime( now + step );

		using namespace Halcon;

		// Local iconic variables 
		Hobject  WithCalTabImage, WithOutCalTabImage;
		Hobject  Map, ModelContours, Caltab, ContoursProjTrans, CrossModel;
		Hobject  ImageWorld, Cross2, Cross5, Cross3, Image, ImageMapped;
		Hobject  Cross4, Cross;
		Hobject  RegionTracking, TransImage, ImageReduced1;


		// Local control variables 
		HTuple  CamParam, CamParOut, ModelID, CaltabDescrName;
		HTuple  WindowHandle, RCoordTmp, CCoordTmp, StartPoseTmp;
		HTuple  X, Y, Z, CamParam1, NFinalPose, Errors, HomMat2D;
		HTuple  Score, HomMat2DToModel, Index1, Selected, Qx, Qy;
		HTuple  QxModel, QyModel, QxDummy, QyDummy, Pointer, Type;
		HTuple  Width, Height, PoseNewOrigin, Y1, X1, Y1Dummy, X1Dummy;
		HTuple  Yfour, Xfour, HomMat3D1, Qx2, Qy2, Qz2, Row2, Column2;
		HTuple  XOrig, YOrig, HomMat3DIdentity, HomMat3DTranslate;
		HTuple  HomMat3DCompose, HomMat3DCompose1, PoseModel, AcqHandle;
		HTuple  CameraMatrix, ImageWidth, ImageHeight, HomMat2DInvert;
		HTuple  HomMat2DCompose, RowTrans, ColTrans, RowTrans1;
		HTuple  ColTrans1, HomMat2DWorldImage, Covariance, Pose1;
		HTuple  CamParam2, NFinalPose1, Errors1, HomMat3DPose, HomMat3DCompose2;
		HTuple  XModel, YModel, ZModel, AlphaModel, BetaModel, GammaModel;
		HTuple   HomMat2DFirst;
		HTuple   HomMat2DFirstInvert;
		HTuple  HomMat2DTrack, HomMat2DInvert2, HomMat2DCompose2;
		HTuple  HomMat2DCompose3;
		HTuple  detectionParameters, trackingParameters, TryToTrackCounter, BNeedForDetection;
		HTuple  FinalPoseMatrix;

		// HDevelop default window handle 
		HTuple CPPExpDefaultWindow;
		// Open default window 
		open_window(0,0,640,480,0,"","",&CPPExpDefaultWindow);
		//Code generated by Image Acquisition 01
		
		std::cerr <<("# The current model to track: ./HALCONApplikations/Models/"+m_TrackingObjectName+".shm").c_str()<< std::endl;

		Herror err = read_image(&WithCalTabImage,("./HALCONApplikations/Models/"+m_TrackingObjectName+".shm.jpg").c_str());
		if (err != H_MSG_TRUE) 
		{
			std::cerr <<("could not find image: ./HALCONApplikations/Models/"+m_TrackingObjectName+".shm.jpg").c_str()<< std::endl;
			exit(1); 
		}
		err = read_image(&WithOutCalTabImage,("./HALCONApplikations/Models/"+m_TrackingObjectName+".shm.jpg").c_str());
		if (err != H_MSG_TRUE) 
		{ 
			std::cerr <<("could not find image: ./HALCONApplikations/Models/"+m_TrackingObjectName+".shm.jpg").c_str()<< std::endl;
			exit(1); 
		}

		read_region(&RegionTracking, ("./HALCONApplikations/Models/"+m_TrackingObjectName+"_tracking_region.reg").c_str());

		read_cam_par("./HALCONApplikations/ConfigFiles/gridspace_campar.cal", &CamParam);
		change_radial_distortion_cam_par("adaptive", CamParam, 0, &CamParOut);
		gen_radial_distortion_map(&Map, CamParam, CamParOut, "bilinear");
		read_deformable_model(("./HALCONApplikations/Models/"+m_TrackingObjectName+".shm").c_str(), &ModelID);
		get_deformable_model_contours(&ModelContours, ModelID, 1);
		// dev_update_window(...); only in hdevelop
		//CaltabDescrName = "./HALCONApplikations/ConfigFiles/caltab_200mm.descr";
		WindowHandle = 3600;
		
		read_pose(("./HALCONApplikations/Models/"+m_TrackingObjectName+"_Pose.dat").c_str(), &NFinalPose);


		readDetectionAndTrackingParameters("./HALCONApplikations/ConfigFiles/TrackingConfig.txt", &detectionParameters, &trackingParameters);

		// stop(); only in hdevelop
		//std::cerr << "11" << std::endl;
		disp_obj(WithOutCalTabImage, HTuple(CPPExpDefaultWindow));
		//find_planar_uncalib_deformable_model(WithOutCalTabImage, ModelID, -0.1, 0.39, 1.0, 
		//	2.0, 1, 1, 0.6, 1, 1, 0, 0.8, "Subpixel", 6, &HomMat2D, &Score);

		find_planar_uncalib_deformable_model(WithOutCalTabImage, ModelID, HTuple(detectionParameters[0]), 
			HTuple(detectionParameters[1]), HTuple(detectionParameters[2]), HTuple(detectionParameters[3]), 
			HTuple(detectionParameters[4]), HTuple(detectionParameters[5]), HTuple(detectionParameters[6]), 
			HTuple(detectionParameters[7]), HTuple(detectionParameters[8]), HTuple(detectionParameters[9]), 
			HTuple(detectionParameters[10]), "TopLevelMatches", 10, &HomMat2D, &Score);

		HomMat2DToModel = HomMat2D;
		HomMat2DFirst  = HomMat2D;
		//std::cerr << "1" << std::endl;
		for (Index1=0; Index1<=(Score.Num())-1; Index1+=1)
		{	
			std::cerr << "12" << std::endl;
			tuple_select_range(HomMat2D, Index1*9, ((Index1+1)*9)-1, &Selected);
			projective_trans_contour_xld(ModelContours, &ContoursProjTrans, Selected);
			projective_trans_pixel(Selected, 0, 0, &Qx, &Qy);
			QxModel = Qx;
			QyModel = Qy;
			gen_cross_contour_xld(&CrossModel, Qx, Qy, 10, 0.785398);
			projective_trans_pixel(Selected, HTuple(-1).Concat(HTuple(-1).Concat(HTuple(1).Concat(1))), 
				HTuple(-1).Concat(HTuple(1).Concat(HTuple(-1).Concat(1))), &QxDummy, &QyDummy);
			//disp_obj(ContoursProjTrans, HTuple(CPPExpDefaultWindow));
			//disp_obj(CrossModel, HTuple(CPPExpDefaultWindow));
			//stop ()
		}
		if (Score.Num() ==0)
		{
			std::cerr << "model not found in initialisation proces" << std::endl;

		}
		// stop(); only in hdevelop
		get_image_pointer1(WithOutCalTabImage, &Pointer, &Type, &Width, &Height);
		close_window(HTuple(CPPExpDefaultWindow));
		//set_window_attr("background_color","black");
		//set_system ("backing_store","false");  
		//set_system ("flush_graphic","false");  
		open_window(0,0,640,480,"root","","",&WindowHandle);

		
		/*HWindow hWindow("");
		hWindow.OpenWindow(0,0,Width[0].I(),Height[0].I());
		hWindow.SetPart(0,0,Height[0].I()-1,Width[0].I()-1);
		hWindow.Display(WithOutCalTabImage);*/

		//std::cerr << "14" << std::endl;

		set_origin_pose(NFinalPose, 0.0, 0.0, 0, &PoseNewOrigin);
		image_to_world_plane(WithCalTabImage, &ImageWorld, CamParOut, PoseNewOrigin, Width, 
			Height, "mm", "bilinear");
		image_points_to_world_plane(CamParOut, PoseNewOrigin, Qx, Qy, "mm", &Y1, &X1);
		gen_cross_contour_xld(&Cross2, X1, Y1, 6, 0.785398);
		image_points_to_world_plane(CamParOut, PoseNewOrigin, QxDummy, QyDummy, "mm", &Y1Dummy, 
			&X1Dummy);
		gen_cross_contour_xld(&Cross5, X1Dummy, Y1Dummy, 6, 0.785398);
		// stop(); only in hdevelop
		image_points_to_world_plane(CamParOut, PoseNewOrigin, HTuple(-1.0).Concat(HTuple(-1.0).Concat(HTuple(1.0).Concat(1.0))), 
			HTuple(-1.0).Concat(HTuple(1.0).Concat(HTuple(1.0).Concat(-1.0))), "mm", &Yfour, 
			&Xfour);
		pose_to_hom_mat3d(NFinalPose, &HomMat3D1);

		affine_trans_point_3d(HomMat3D1, Y1*0.001, X1*0.001, 0, &Qx2, &Qy2, &Qz2);

		project_3d_point(Qx2, Qy2, Qz2, CamParOut, &Row2, &Column2);
		gen_cross_contour_xld(&Cross3, Row2, Column2, 6, 0.785398);
		//image_points_to_world_plane (CamParOut, PoseNewOrigin, 0, 0, 'mm', XOrig, YOrig)
		hom_mat3d_identity(&HomMat3DIdentity);
		//hom_mat3d_translate (HomMat3DIdentity, Y1*0.001, X1*0.001, 0, HomMat3DTranslate)
		hom_mat3d_translate_local(HomMat3DIdentity, Y1*0.001, X1*0.001, 0.0, &HomMat3DTranslate);
		//hom_mat3d_compose (HomMat3D1, HomMat3DTranslate, HomMat3DCompose)
		//disp_obj(ImageWorld, HTuple(WindowHandle));
		// stop(); only in hdevelop
		hom_mat3d_compose(HomMat3D1, HomMat3DTranslate, &HomMat3DCompose1);
		hom_mat3d_to_pose(HomMat3DCompose1, &PoseModel);
		affine_trans_point_3d(HomMat3DCompose1, 0.0, 0.0, 0, &Qx2, &Qy2, &Qz2);
		project_3d_point(Qx2, Qy2, Qz2, CamParOut, &Row2, &Column2);
		gen_cross_contour_xld(&Cross3, Row2, Column2, 6, 0.785398);
		// stop(); only in hdevelop
		std::cerr << "try to open framegrabber" << std::endl;
		
		close_all_framegrabbers();
		open_framegrabber("DirectShow", 1, 1, 0, 0, 0, 0, "default", 8, "rgb", -1, "false", 
		  "default", "DMx 31AU03", 0, -1, &AcqHandle);
		std::cerr << "opend framegrabber" << std::endl;
		//set_framegrabber_param(AcqHandle, "camera_type", "7:0:9");
		
		cam_par_to_cam_mat(CamParOut, &CameraMatrix, &ImageWidth, &ImageHeight);
		hom_mat2d_invert(HomMat2DToModel, &HomMat2DInvert);
		//std::cerr << "15" << std::endl;

		Score = HTuple();

		int cnt = 0;
		while ( !m_bStop )
		{

			/*if (cnt%10==0){
			close_window(HTuple(WindowHandle));
			open_window(0,0,Width,Height,"root","","",&WindowHandle);
			}*/
			//cnt++;

			// compute sleep time and sleep
			now = Measurement::now();
			//if ( now < nextTime )
			//{
			//	// sampling took less than the sleep duration
			//	long long int sleepdur = nextTime - now;
			//	nextTime += step;
			//	Util::sleep( int( sleepdur / 1000000 ), int( sleepdur % 1000000 ) );
			//}
			//else
			//{
			//	// sampling took more than the sleep duration -> yield some processing time to other threads
			//	nextTime = now + step;
			//	boost::thread::yield();
			//}

			//##############################################

			grab_image(&Image, AcqHandle);
			map_image(Image, Map, &ImageMapped);

			if ((Score.Num())==0)
			{
				std::cout<<"Detection"<<std::endl;


				/*find_planar_uncalib_deformable_model(ImageMapped, ModelID, -0.79, 1.39,0.5, 1.5, 
				1, 1, 0.4, 1, 1, 0, 0.9, "TopLevel", 6, &HomMat2D, &Score);*/
				//find_planar_uncalib_deformable_model (ImageMapped, ModelID, -0.6, 1.2, 0.2, 1.3, 1, 1, 0.5, 1, 1, 0, 0.1, "TopLevelMatches", 300, &HomMat2D, &Score);

				find_planar_uncalib_deformable_model(ImageMapped, ModelID, HTuple(detectionParameters[0]), 
					HTuple(detectionParameters[1]), HTuple(detectionParameters[2]), HTuple(detectionParameters[3]), 
					HTuple(detectionParameters[4]), HTuple(detectionParameters[5]), HTuple(detectionParameters[6]), 
					HTuple(detectionParameters[7]), HTuple(detectionParameters[8]), HTuple(detectionParameters[9]), 
					HTuple(detectionParameters[10]), "TopLevelMatches", 10, &HomMat2D, &Score);

				//disp_image(ImageMapped, HTuple(WindowHandle));
				//disp_image(ImageMapped, HTuple(CPPExpDefaultWindow));

				/* hWindow.Display(ImageMapped);*/
				//close_window(HTuple(CPPExpDefaultWindow));

				//disp_obj(ImageMapped, HTuple(WindowHandle));
				for (Index1=0; Index1<=(Score.Num())-1; Index1+=1)
				{
					tuple_select_range(HomMat2D, Index1*9, ((Index1+1)*9)-1, &Selected);
					//hom_mat2d_compose (Selected, HomMat2DInvert, HomMat2DCompose)
					//projective_trans_pixel (HomMat2DCompose, QxModel, QyModel, RowTrans, ColTrans)
					//gen_cross_contour_xld (Cross4, RowTrans, ColTrans, 6, 0.785398)
					projective_trans_pixel(HomMat2D, HTuple(-1).Concat(HTuple(-1).Concat(HTuple(1).Concat(1))), 
						HTuple(-1).Concat(HTuple(1).Concat(HTuple(-1).Concat(1))), &RowTrans1, 
						&ColTrans1);
					gen_cross_contour_xld(&Cross, RowTrans1, ColTrans1, 10, 0.785398);
					vector_to_proj_hom_mat2d(Y1Dummy*0.001, X1Dummy*0.001, ColTrans1, RowTrans1, 
						"normalized_dlt", HTuple(), HTuple(), HTuple(), HTuple(), HTuple(), HTuple(), 
						&HomMat2DWorldImage, &Covariance);
					proj_hom_mat2d_to_pose(HomMat2DWorldImage, CameraMatrix, "zhang", &Pose1);



					camera_calibration(Y1Dummy*0.001, X1Dummy*0.001, HTuple(0.0).Concat(HTuple(0.0).Concat(HTuple(0.0).Concat(0.0))), 
						RowTrans1, ColTrans1, CamParOut, Pose1, "pose", &CamParam2, &NFinalPose1, 
						&Errors1);
					projective_trans_contour_xld(ModelContours, &ContoursProjTrans, Selected);

					//hWindow.Display(ContoursProjTrans);


					projective_trans_pixel(Selected, 0, 0, &Qx, &Qy);
					gen_cross_contour_xld(&Cross, Qx, Qy, 10, 0.785398);
					//disp_obj(ContoursProjTrans, HTuple(WindowHandle));
					//dev_display (Cross)
					pose_to_hom_mat3d(Pose1, &HomMat3DPose);
					hom_mat3d_compose(HomMat3DPose, HomMat3DTranslate, &HomMat3DCompose2);
					affine_trans_point_3d(HomMat3DCompose2, 0, 0, 0, &Qx2, &Qy2, &Qz2);
					project_3d_point(Qx2, Qy2, Qz2, CamParOut, &Row2, &Column2);
					gen_cross_contour_xld(&Cross3, Row2, Column2, 6, 0.785398);
					//disp_obj(Cross3, HTuple(WindowHandle));

					//stop ()
				}
			}


			else
			{
				std::cout<<"Tracking"<<std::endl;
				hom_mat2d_invert(HomMat2D, &HomMat2DInvert);
				hom_mat2d_invert(HomMat2DFirst, &HomMat2DFirstInvert);
				hom_mat2d_compose(HomMat2DFirst, HomMat2DInvert, &HomMat2DCompose);
				projective_trans_image(ImageMapped, &TransImage, HomMat2DCompose, "bilinear", 
					"false", "false");
				reduce_domain(TransImage, RegionTracking, &ImageReduced1);
				//find_planar_uncalib_deformable_model(ImageReduced1, ModelID, -0.2, 0.39, 1.0, 
				//	1.0, 1, 1, 0.6, 1, 1, 0, 0.95, HTuple(), HTuple(), &HomMat2DTrack, &Score);
				find_planar_uncalib_deformable_model(ImageReduced1, ModelID, HTuple(trackingParameters[0]), 
					HTuple(trackingParameters[1]), HTuple(trackingParameters[2]), HTuple(trackingParameters[3]), 
					HTuple(trackingParameters[4]), HTuple(trackingParameters[5]), HTuple(trackingParameters[6]), 
					HTuple(trackingParameters[7]), HTuple(trackingParameters[8]), HTuple(trackingParameters[9]), 
					HTuple(trackingParameters[10]), HTuple(), HTuple(), &HomMat2DTrack, &Score);
				//disp_image(ImageReduced1, HTuple(WindowHandle));
				std::cout<<"Score: "<< Score.Num()<<std::endl;
				//disp_obj(ImageMapped, HTuple(CPPExpDefaultWindow));
				//disp_image(ImageMapped, HTuple(WindowHandle));
				for (Index1=0; Index1<=(Score.Num())-1; Index1+=1)
				{
					hom_mat2d_invert(HomMat2DCompose, &HomMat2DInvert2);
					tuple_select_range(HomMat2DTrack, Index1*9, ((Index1+1)*9)-1, &Selected);
					projective_trans_contour_xld(ModelContours, &ContoursProjTrans, Selected);
					hom_mat2d_compose(HomMat2DFirstInvert, Selected, &HomMat2DCompose2);
					hom_mat2d_compose(HomMat2D, HomMat2DCompose2, &HomMat2DCompose3);
					projective_trans_contour_xld(ModelContours, &ContoursProjTrans, HomMat2DCompose3);
					proj_hom_mat2d_to_pose(HomMat2DCompose3, CameraMatrix, "faugeras", &Pose1);
					projective_trans_pixel(Selected, 0, 0, &Qx, &Qy);
					gen_cross_contour_xld(&Cross, Qx, Qy, 10, 0.785398);
					//disp_obj(ContoursProjTrans, HTuple(CPPExpDefaultWindow));
					//dev_display (Cross)
					//stop ()
					HomMat2D = HomMat2DCompose3;
				}

				for (Index1=0; Index1<=(Score.Num())-1; Index1+=1)
				{
					tuple_select_range(HomMat2D, Index1*9, ((Index1+1)*9)-1, &Selected);
					//hom_mat2d_compose (Selected, HomMat2DInvert, HomMat2DCompose)
					//projective_trans_pixel (HomMat2DCompose, QxModel, QyModel, RowTrans, ColTrans)
					//gen_cross_contour_xld (Cross4, RowTrans, ColTrans, 6, 0.785398)
					projective_trans_pixel(HomMat2D, HTuple(-1).Concat(HTuple(-1).Concat(HTuple(1).Concat(1))), 
						HTuple(-1).Concat(HTuple(1).Concat(HTuple(-1).Concat(1))), &RowTrans1, 
						&ColTrans1);
					gen_cross_contour_xld(&Cross, RowTrans1, ColTrans1, 10, 0.785398);
					vector_to_proj_hom_mat2d(Y1Dummy*0.001, X1Dummy*0.001, ColTrans1, RowTrans1, 
						"normalized_dlt", HTuple(), HTuple(), HTuple(), HTuple(), HTuple(), HTuple(), 
						&HomMat2DWorldImage, &Covariance);
					proj_hom_mat2d_to_pose(HomMat2DWorldImage, CameraMatrix, "zhang", &Pose1);



					camera_calibration(Y1Dummy*0.001, X1Dummy*0.001, HTuple(0.0).Concat(HTuple(0.0).Concat(HTuple(0.0).Concat(0.0))), 
						RowTrans1, ColTrans1, CamParOut, Pose1, "pose", &CamParam2, &NFinalPose1, 
						&Errors1);
					projective_trans_contour_xld(ModelContours, &ContoursProjTrans, Selected);

					//hWindow.Display(ContoursProjTrans);


					projective_trans_pixel(Selected, 0, 0, &Qx, &Qy);
					gen_cross_contour_xld(&Cross, Qx, Qy, 10, 0.785398);
					//disp_obj(ContoursProjTrans, HTuple(WindowHandle));
					//dev_display (Cross)
					pose_to_hom_mat3d(Pose1, &HomMat3DPose);
					hom_mat3d_compose(HomMat3DPose, HomMat3DTranslate, &HomMat3DCompose2);
					affine_trans_point_3d(HomMat3DCompose2, 0, 0, 0, &Qx2, &Qy2, &Qz2);
					project_3d_point(Qx2, Qy2, Qz2, CamParOut, &Row2, &Column2);
					gen_cross_contour_xld(&Cross3, Row2, Column2, 6, 0.785398);
					//disp_obj(Cross3, HTuple(WindowHandle));

					//stop ()
				}



			}





			//##############################################


			// do sampling
			now = Measurement::now() + m_nOffset;
			try
			{	
				disp_obj(ImageMapped, HTuple(WindowHandle));
				if (Score.Num()>0){

				double tempFinalMatrix [16];

				pose_to_hom_mat3d(NFinalPose1, &FinalPoseMatrix);
				
				tempFinalMatrix[0] = (double) FinalPoseMatrix[0];
				tempFinalMatrix[1] = (double) FinalPoseMatrix[1];
				tempFinalMatrix[2] = (double) FinalPoseMatrix[2];
				tempFinalMatrix[3] = (double) FinalPoseMatrix[3];
				tempFinalMatrix[4] = (double) FinalPoseMatrix[4];
				tempFinalMatrix[5] = (double) FinalPoseMatrix[5];
				tempFinalMatrix[6] = (double) FinalPoseMatrix[6];
				tempFinalMatrix[7] = (double) FinalPoseMatrix[7];
				tempFinalMatrix[8] = (double) FinalPoseMatrix[8];
				tempFinalMatrix[9] = (double) FinalPoseMatrix[9];
				tempFinalMatrix[10] = (double) FinalPoseMatrix[10];
				tempFinalMatrix[11] = (double) FinalPoseMatrix[11];
				tempFinalMatrix[12] = (double) FinalPoseMatrix[8];
				tempFinalMatrix[13] = (double) FinalPoseMatrix[9];
				tempFinalMatrix[14] = (double) FinalPoseMatrix[10];
				tempFinalMatrix[15] = (double) FinalPoseMatrix[11];

				double tx = (double)NFinalPose1[0];
				double ty = (double)NFinalPose1[1];
				double tz = (double)NFinalPose1[2];

				double alpha = (double)NFinalPose1[3];
				double beta =(double)NFinalPose1[4];
				double gamma = (double)NFinalPose1[5];

				std::cout<<"Last door Pose: x:		   " <<tx << " y: " << ty<< " z: " << tz<< " alpha: " <<alpha << " beta: " << beta<< " gamma: " << gamma<<std::endl;
				std::cout<<"Last door Pose (via matrix):  x: " <<tempFinalMatrix[3] << " y: " << tempFinalMatrix[7]<< " z: " << tempFinalMatrix[11]<<std::endl;
				
				Math::Quaternion qx = Math::Quaternion( sin(alpha*(M_PI/90)), 0.0, 0.0, cos(alpha*(M_PI/90)));
				Math::Quaternion qy = Math::Quaternion( 0.0, sin(beta*(M_PI/90)), 0.0, cos(beta*(M_PI/90)));
				Math::Quaternion qz = Math::Quaternion( 0.0, 0.0, sin(gamma*(M_PI/90)),cos(gamma*(M_PI/90)));
				Math::Quaternion qTemp, quat;
				qTemp = qy*qz;

				qTemp.normalize();
				quat	= qTemp*qx;
				quat.normalize();
				
				
				//Math::Quaternion qx = Math::Quaternion( cos(alpha*(M_PI/90)),sin(alpha*(M_PI/90)), 0.0, 0.0 );
				//Math::Quaternion qy = Math::Quaternion( cos(beta*(M_PI/90)), 0.0, sin(beta*(M_PI/90)), 0.0 );
				//Math::Quaternion qz = Math::Quaternion( cos(gamma*(M_PI/90)), 0.0, 0.0, sin(gamma*(M_PI/90)));
				//Math::Quaternion qTemp, quat;
				//qTemp = qy*qx;
				//qTemp.normalize();
				//quat	= qz*qTemp;
				//quat.normalize();

				//double c1 = cos(alpha*(M_PI/90));
				//double s1 = sin(alpha*(M_PI/90));
				//double c2 = cos(beta*(M_PI/90));
				//double s2 = sin(beta*(M_PI/90));
				//double c3 = cos(gamma*(M_PI/90));
				//double s3 = sin(gamma*(M_PI/90));

				//double w = c1*c2*c3 - s1*s2*s3;
				//double x = c1*c2*s3 + s1*s2*c3;
				//double y = s1*c2*c3 + c1*s2*s3;
				//double z = c1*s2*c3 - s1*c2*s3; 
				
				//Math::Quaternion quat = Math::Quaternion(0.0,0.0,0.0,1.0);
				
				//Math::Pose halconPose = Math::Pose(quat,Math::Vector< 3 >( tx, ty, tz ));
				Math::Pose halconPose = Math::Pose(Math::Matrix<4,4>(tempFinalMatrix)); 


				Math::Quaternion offset = Math::Quaternion( 1.0, 0.0, 0.0, 0.0);
				offset.normalize();
				Math::Pose offsetHalconUbitrack = Math::Pose(offset,Math::Vector< 3 >( 0.0, 0.0, 0.0 ));

				Math::Pose finalPose = offsetHalconUbitrack*halconPose;

				//provide image

				//end provide image

				m_outPort.send( Measurement::Pose( now,	finalPose ) );
				//disp_obj(ImageMapped, HTuple(WindowHandle));
				disp_obj(ContoursProjTrans, HTuple(WindowHandle));
			}
			else
			{
				std::cerr << "did not find the model";
			}

			m_imgPort.send( Vision::ImageMeasurement( now, copyImage( ImageMapped ) ) );
			std::cout << "sent. " << std::endl;

		}
		catch ( const Util::Exception& e )
		{
			LOG4CPP_WARN( eventLogger, "Got exception: " << e );
		}
		catch ( const std::exception& e )
		{
			LOG4CPP_WARN( eventLogger, "Got unknown exception: " << e.what() );
		}
	}
}

boost::shared_ptr< Vision::Image > HALCONComponent::copyImage(const Halcon::Hobject& hImage)
{
	char	type[128];
	long width, height;
	unsigned char* ptr = NULL;
	Halcon::HTuple a,b,c;

	Herror err = get_image_pointer1(hImage, (long*) &ptr, type, &width, &height);

	if( err != H_MSG_TRUE )
		std::cout << "something went wrong: " << err << std::endl;

	Vision::Image pImage( (int) width, ( int ) height, 1, ptr);
	return pImage.Clone();
}

} } // namespace Ubitrack::Components

UBITRACK_REGISTER_COMPONENT( Ubitrack::Dataflow::ComponentFactory* const cf ) {
	cf->registerComponent< Ubitrack::Drivers::HALCONComponent > ( "HALCONComponent" );
}
