/*
* Ubitrack - Library for Ubiquitous Tracking
* Copyright 2006, Technische Universitaet Muenchen, and individual
* contributors as indicated by the @authors tag. See the
* copyright.txt in the distribution for a full listing of individual
* contributors.
*
* This is free software; you can redistribute it and/or modify it
* under the terms of the GNU Lesser General Public License as
* published by the Free Software Foundation; either version 2.1 of
* the License, or (at your option) any later version.
*
* This software is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
* Lesser General Public License for more details.
*
* You should have received a copy of the GNU Lesser General Public
* License along with this software; if not, write to the Free
* Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
* 02110-1301 USA, or see the FSF site: http://www.fsf.org.
*/


/**
* @ingroup dataflow_components
* @file
* Components that pulls its input at a given frequency and pushes the result
*
* @author Bjoern Schwerdtfeger <schwerdt@in.tum.de>
*/

#include <boost/scoped_ptr.hpp>
#include <boost/bind.hpp>
#include <boost/thread.hpp>
#include <log4cpp/Category.hh>

#include <Ubitrack/Dataflow/Component.h>
#include <Ubitrack/Dataflow/PullConsumer.h>
#include <Ubitrack/Dataflow/PushSupplier.h>
#include <Ubitrack/Dataflow/ComponentFactory.h>
#include <Ubitrack/Measurement/Measurement.h>
#include <Ubitrack/Util/Exception.h>
#include <Ubitrack/Util/OS.h>
#include <string>

#include <Vision/Image.h>
#include <cv.h>

#include "HalconCpp.h"
//#include "HCPPdeform_shape.h"
//#include "HCPPplanar_pose_estimation.h"

#include "Math.h"


// get a logger
static log4cpp::Category& logger( log4cpp::Category::getInstance( "Drivers.HALCON" ) );
static log4cpp::Category& eventLogger( log4cpp::Category::getInstance( "Drivers.Events.HALCON" ) );

namespace Ubitrack { namespace Drivers {

	/**
	* @ingroup dataflow_components
	* Components that pulls its input at a given frequency and pushes the result
	*
	* @par Input Ports
	* PullConsumer< EventType > with name "Input".
	*
	* @par Output Ports
	* PushSupplier< EventType > with name "Output".
	*
	* @par Configuration
	* @verbatim <Configuration frequency="100"/> @endverbatim
	* \c frequency: floating point number giving the sampling frequency in Hz
	*
	* @par Operation
	*
	* @par Instances
	*
	*/
	class HALCONComponent
		: public Dataflow::Component
	{
	public:
		/**
		* UTQL component constructor.
		*
		* @param sName Unique name of the component.
		* @param subgraph UTQL subgraph
		*/
		HALCONComponent( const std::string& sName, boost::shared_ptr< Graph::UTQLSubgraph > subgraph )
			: Dataflow::Component( sName )
			//, m_inPort( "Input", *this )
			, m_outPort( "Output", *this )
			, m_imgPort( "Image", *this )
			, m_fFrequency( 0.5 )
			, m_nOffset( 0 )
			, m_bStop( true )
			//, m_TrackingObjectName("dummy")
		{
			std::cerr << "HALCON Component: Init" << std::endl;
			m_TrackingObjectName = subgraph->m_DataflowAttributes.getAttributeString("trackingObjectName");
			std::cerr << ("HALCON Component: Model Name: "+m_TrackingObjectName).c_str() << std::endl;
			double offset( 0 );
			//subgraph->m_DataflowAttributes.getAttributeData( "offset", offset );
			//m_nOffset = (long long int)( 1e9 * offset );

			stop();
		}

		~HALCONComponent();

		virtual void start();
		virtual void stop();

	protected:
		/** Method that computes the result. */
		void threadMethod();
		void provideImage(/*const HObject& hImage,*/ Vision::ImageMeasurement& image );


		// Output ports of the component
		Dataflow::PushSupplier< Measurement::Pose > m_outPort;
		Dataflow::PushSupplier< Vision::ImageMeasurement > m_imgPort;

		// sampling frequency
		double m_fFrequency;// = 1.0;

		// offset to add to sampled timestamps
		long long int m_nOffset;

		// stop?
		bool m_bStop;

		std::string m_TrackingObjectName;



		// pointer to the thread
		boost::scoped_ptr< boost::thread > m_pThread;

	private:

	void readDetectionAndTrackingParameters (Halcon::HTuple Filename, Halcon::HTuple *detectionParameters, 
		Halcon::HTuple *trackingParameters);
	
	boost::shared_ptr< Vision::Image > copyImage(const Halcon::Hobject& hImage);
};} } // namespace Ubitrack::Components
