/*
 * Ubitrack - Library for Ubiquitous Tracking
 * Copyright 2006, Technische Universitaet Muenchen, and individual
 * contributors as indicated by the @authors tag. See the 
 * copyright.txt in the distribution for a full listing of individual
 * contributors.
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA, or see the FSF site: http://www.fsf.org.
 */
 
 /**
 * @ingroup drivers
 * @file
 * Implementation of the InterSense Driver.
 *
 * @author Christian Waechter <christian.waechter@in.tum.de>
 */
 
#include "intersense.h"

#include <stdio.h> 
#include <stdlib.h> 
#include <memory.h>
#include <string.h>

/*#if defined(_WIN32) || defined(WIN32) || defined(__WIN32__)
#include <conio.h>
#endif
*/

#include <log4cpp/Category.hh>
static log4cpp::Category& logger( log4cpp::Category::getInstance( "Drivers.InterSense" ) );

namespace Ubitrack { namespace Driver { 

InterSenseModule::InterSenseModule( const InterSenseModuleKey& moduleKey, boost::shared_ptr< Graph::UTQLSubgraph >, FactoryHelper* pFactory )
	: Module< InterSenseModuleKey, InterSenseComponentKey, InterSenseModule, InterSenseComponent >( moduleKey, pFactory )
		, m_bStop( true )
		{
			stopModule( );
			LOG4CPP_DEBUG ( logger, "InterSense driver created" );
		};

	
int InterSenseModule::threadMethod()
{
	LOG4CPP_INFO( logger, "Starting InterSense thread. This takes some time" );

	ISD_TRACKER_HANDLE              handle[ISD_MAX_TRACKERS];
    ISD_TRACKING_DATA_TYPE          data;
    ISD_STATION_INFO_TYPE           Station[ISD_MAX_TRACKERS][ISD_MAX_STATIONS];
    ISD_TRACKER_INFO_TYPE           Tracker[ISD_MAX_TRACKERS];
    ISD_HARDWARE_INFO_TYPE          hwInfo[ISD_MAX_TRACKERS];
    ISD_STATION_HARDWARE_INFO_TYPE  stationHwInfo[ISD_MAX_TRACKERS][ISD_MAX_STATIONS];

   
    WORD done = FALSE, station = 1;
    DWORD maxStations[ISD_MAX_TRACKERS];
    float lastTime;

    // Detect first tracker. If you have more than one InterSense device and
    // would like to have a specific tracker, connected to a known port, 
    // initialized first, then enter the port number instead of 0. Otherwise, 
    // tracker connected to the rs232 port with lower number is found first 
	//handle[tracker-1] = ISD_OpenTracker( (Hwnd) NULL, 0, FALSE, FALSE );
	

	int count_comp = 0;
	ComponentList components = getAllComponents();
	for( ComponentList::iterator it = components.begin(); it != components.end(); it++ )
	{
		if( count_comp == (ISD_MAX_TRACKERS-1) )
		{
			LOG4CPP_DEBUG( logger, "Number of components is greater than number of trackers." );
			break;
		}	
		LOG4CPP_DEBUG( logger, "Starting to detect InterSense tracker : " << (*it)->getKey() );

		//int port = (*it)->getKey() > 1 ? (*it)->getKey() : 0;
		int port = getModuleKey() > 1 ? getModuleKey() : 0;
		
		handle[count_comp] = ISD_OpenTracker( (Hwnd) NULL, port, FALSE, TRUE );
		if( handle[count_comp] < 1 )
			break;
		
		ISD_GetTrackerConfig( handle[count_comp], &Tracker[count_comp], FALSE );
                
        memset((void *) hwInfo, 0, sizeof(hwInfo));
		
		maxStations[count_comp] = 4;

        if( ISD_GetSystemHardwareInfo( handle[count_comp], &hwInfo[count_comp] ) )
        {
            if( hwInfo[count_comp].Valid )
            {
                maxStations[count_comp] = hwInfo[count_comp].Capability.MaxStations;
            }
        }

        // Clear station configuration info to make sure GetAnalogData and other flags are FALSE 
        memset( (void *) Station, 0, sizeof(Station) );

        // General procedure for changing any setting is to first retrieve current 
        // configuration, make the change, and then apply it. Calling 
        // ISD_GetStationConfig is important because you only want to change 
        // some of the settings, leaving the rest unchanged. 
        
        if( Tracker[count_comp].TrackerType == ISD_PRECISION_SERIES )
        {
            for( station = 1; station <= maxStations[count_comp]; station++ )
            {    
				LOG4CPP_DEBUG( logger, "Number of found stations :" << station );			
                // fill ISD_STATION_INFO_TYPE structure with current station configuration 
                if( !ISD_GetStationConfig( handle[count_comp], 
                    &Station[count_comp][station-1], station, TRUE ) ) break;
                
                if( !ISD_GetStationHardwareInfo( handle[count_comp], 
                    &stationHwInfo[count_comp][station-1], station ) ) break;
            }
        }
		count_comp++;
		LOG4CPP_DEBUG( logger, "Detected InterSense tracking device " << count_comp );
	}
	LOG4CPP_INFO( logger, "Finished sensor detection, devices found : " << count_comp );
    
	//TODO: make something with nthe station, not figured out why necessary!
    station = 1;
    lastTime = ISD_GetTime();

    while( !done && m_running )
    {
		Measurement::Timestamp timestamp = Measurement::now();
		if( ISD_GetTime() - lastTime > 0.01f )
		{
			lastTime = ISD_GetTime();
			count_comp = 0;
			for( ComponentList::iterator it = components.begin(); it != components.end(); it++ )
			{
				ISD_GetTrackingData( handle[count_comp], &data );
				
				LOG4CPP_TRACE( logger, "device handle : " << handle[count_comp] );
			
				(*it)->sendPose( timestamp, handle[count_comp], &Tracker[count_comp], &Station[count_comp][station-1], &data.Station[station-1] );
				
				count_comp++;
			}
        }
#ifdef _WIN32
        Sleep( 1 );
#endif
    }
	count_comp = 0;
	for( ComponentList::iterator it = components.begin(); it != components.end(); it++ )
		ISD_CloseTracker( handle[count_comp++] );
		
    return 1;
		
}

void InterSenseComponent::sendPose( const Measurement::Timestamp timestamp
			, ISD_TRACKER_HANDLE handle
			, ISD_TRACKER_INFO_TYPE *Tracker
			, ISD_STATION_INFO_TYPE *Station
			, ISD_STATION_DATA_TYPE *data )
{
    // Get comm port statistics for display with tracker data 
    if( ISD_GetCommInfo( handle, Tracker ) )
	{
		//Kbps: Tracker->KBitsPerSec
		//R/s: Tracker->RecordsPerSec
		//Euler-angles: data->Euler[0], data->Euler[1], data->Euler[2]
		//Timestamp: data->TimeStamp
		//???: data->TrackingStatus
		LOG4CPP_DEBUG( logger, "InterSense data received" );
		//LOG4CPP_INFO( logger, "Number of Cubes : " << Station->InertiaCube );
		
		
		
		// all products can return orientation // if( Station->AngleFormat == ISD_QUATERNION )
		Math::Quaternion quat ( data->Quaternion[1], data->Quaternion[2], data->Quaternion[3], data->Quaternion[0] );
		LOG4CPP_TRACE( logger, "InterSense quaternion : " << quat );
		if( m_port3DRotation.isConnected() )
			m_port3DRotation.send( Measurement::Rotation( timestamp, quat ) );	
	
		//position only if system supports it 
		if( Tracker->TrackerModel == ISD_IS600 || 
			Tracker->TrackerModel == ISD_IS900 ||
			Tracker->TrackerModel == ISD_IS1200 )
		{
			//TODO: making correct transformation that the unit is meters. 
			Math::Vector< 3 > pos ( data->Position[0], data->Position[1], data->Position[2] );
			
			if( m_port3DPosition.isConnected() )
				m_port3DPosition.send( Measurement::Position( timestamp, pos ) );
				
			if( m_port6DOF.isConnected() )//&& ( Station->AngleFormat == ISD_QUATERNION ) )
				m_port6DOF.send( Measurement::Pose( timestamp, Math::Pose( quat, pos ) ) );
		}
		
        // if( Station->GetAuxInputs ) 
        // {
            // printf("%d %d %d %d ", 
                // (int) data->AuxInputs[0], 
                // (int) data->AuxInputs[1], 
                // (int) data->AuxInputs[2], 
                // (int) data->AuxInputs[3]);
        // }
        
        // if system is configured to read stylus or wand buttons 
        // if( Station->GetInputs ) 
        // {
            // // Currently available products have at most 5 buttons,
            // // stylus has 2, wand has 5
            // printf("%d%d%d%d%d ", 
                // (int) data->ButtonState[0], 
                // (int) data->ButtonState[1], 
                // (int) data->ButtonState[2], 
                // (int) data->ButtonState[3], 
                // (int) data->ButtonState[4]);

            // printf( "%d %d ", data->AnalogData[0], data->AnalogData[1] ); 
        // }

    }
}

	
void InterSenseModule::stopModule()
{
	if ( m_running )
	{
		m_running = false;
		m_bStop = true;
		if ( m_pThread )
		{
			m_pThread->join();
		}
	}
}
	
void InterSenseModule::startModule()
{
	if ( !m_running )
	{
		m_running = true;
		m_bStop = false;
		m_pThread.reset( new boost::thread( boost::bind( &InterSenseModule::threadMethod, this ) ) );
	}
}

UBITRACK_REGISTER_COMPONENT( Ubitrack::Dataflow::ComponentFactory* const cf )
{
	cf->registerModule< InterSenseModule > ( "InterSense" );
}

} } // namespace Ubitrack::Driver
