/*
 * Ubitrack - Library for Ubiquitous Tracking
 * Copyright 2006, Technische Universitaet Muenchen, and individual
 * contributors as indicated by the @authors tag. See the 
 * copyright.txt in the distribution for a full listing of individual
 * contributors.
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA, or see the FSF site: http://www.fsf.org.
 */
 
 /**
 * @ingroup drivers
 * @file
 * Definition of the InterSense Driver class.
 * The class contains the driver to access measurments from all available InterSense products. 
 * Tested with Wireless InertialCube3, but should also work with other trackers.
 *
 * @author Christian Waechter <christian.waechter@in.tum.de>
 */

#ifndef __Ubitrack_Drivers_InterSense_InterSense_h_included__
#define __Ubitrack_Drivers_InterSense_InterSense_h_included__

#include <Ubitrack/Dataflow/Module.h>
#include <Ubitrack/Dataflow/Component.h>
#include <Ubitrack/Dataflow/ComponentFactory.h>
#include <Ubitrack/Dataflow/PushSupplier.h>
#include <Ubitrack/Measurement/Measurement.h>
#include <Ubitrack/Util/Exception.h>

#include <boost/thread.hpp>

#include "isense.h"
#include "isense.c"

using namespace Ubitrack;
using namespace Ubitrack::Dataflow;

namespace Ubitrack { namespace Driver {

MAKE_DATAFLOWCONFIGURATIONATTRIBUTEKEY_DEFAULT( InterSenseModuleKey, int, "port", 0 );
MAKE_NODEATTRIBUTEKEY_DEFAULT( InterSenseComponentKey, int, "InterSense", "tracker", 0 );

class InterSenseComponent;

class InterSenseModule :  public Module< InterSenseModuleKey, InterSenseComponentKey, InterSenseModule, InterSenseComponent >
{
	public:

		/** constructor */
		InterSenseModule( const InterSenseModuleKey& moduleKey, boost::shared_ptr< Graph::UTQLSubgraph >, FactoryHelper* pFactory );
	
		/** destructor */
		~InterSenseModule( ) { stopModule(); };
	
		/** start thread*/
		virtual void startModule( );
		
		/** stop thread*/
		virtual void stopModule( );
	
	protected:
	
		/** thread method */
		int threadMethod( );
	
		// pointer to the thread
		boost::shared_ptr< boost::thread > m_pThread;
		
		/** thread is running?*/
		bool m_bStop;
};

class InterSenseComponent
	: public InterSenseModule::Component
{
	public:
		InterSenseComponent( const std::string& name, boost::shared_ptr< Graph::UTQLSubgraph > subgraph
			, const InterSenseComponentKey& componentKey, InterSenseModule* pModule )
			: InterSenseModule::Component( name, componentKey, pModule )
			, m_port6DOF( "Output", *this )
			, m_port3DRotation( "Rotation", *this )
			, m_port3DPosition( "Position", *this )
			{ };

	protected:
	
		/** method to fetch the data and send the measurements */
		void sendPose ( const Measurement::Timestamp
				, ISD_TRACKER_HANDLE handle
				, ISD_TRACKER_INFO_TYPE *Tracker
				, ISD_STATION_INFO_TYPE *Station
				, ISD_STATION_DATA_TYPE *data
		);
		
		// Output ports of the component
		Dataflow::PushSupplier< Measurement::Pose > m_port6DOF;
		Dataflow::PushSupplier< Measurement::Rotation > m_port3DRotation;
		Dataflow::PushSupplier< Measurement::Position > m_port3DPosition;
		
		friend class InterSenseModule;
	
};

} } // namespace Ubitrack::Drivers

#endif