/*
 * Ubitrack - Library for Ubiquitous Tracking
 * Copyright 2006, Technische Universitaet Muenchen, and individual
 * contributors as indicated by the @authors tag. See the
 * copyright.txt in the distribution for a full listing of individual
 * contributors.
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA, or see the FSF site: http://www.fsf.org.
 */


/**
 * @ingroup dataflow_components
 * @file
 * Intersects a 2D point with a plane and returns the 3D camera coordinates of the point.
 *
 * @author Daniel Pustka <daniel.pustka@in.tum.de>
 */
#include <Ubitrack/Dataflow/TriggerComponent.h>
#include <Ubitrack/Dataflow/TriggerInPort.h>
#include <Ubitrack/Dataflow/TriggerOutPort.h>
#include <Ubitrack/Dataflow/ComponentFactory.h>
#include <Ubitrack/Measurement/Measurement.h>
#include <Ubitrack/Math/MatrixOperations.h>

namespace ublas = boost::numeric::ublas;

static log4cpp::Category& logger( log4cpp::Category::getInstance( "Drivers.ImagePointPlaneIntersection" ) );

namespace Ubitrack { namespace Components {

/**
 * @ingroup dataflow_components
 * Intersects a 2D point on screen with a 3D plane and returns the 3D camera coordinates of the point.
 *
 * @par Input Ports
 * TriggerInPort<Matrix3x3> with name "Intrinsics".
 * TriggerInPort<Position2> with name "ImagePoint".
 * TriggerInPort<Pose> with name "PlanePose".
 *
 * @par Output Ports
 * TriggerOutPort<Position> with name "Output".
 *
 * @par The Pattern
 */
class ImagePointPlaneIntersection
	: public Dataflow::TriggerComponent
{
public:
	/**
	 * UTQL component constructor.
	 *
	 * @param sName Unique name of the component.
	 * @param subgraph UTQL subgraph
	 */
	ImagePointPlaneIntersection( const std::string& sName, boost::shared_ptr< Graph::UTQLSubgraph > pConfig )
		: Dataflow::TriggerComponent( sName, pConfig )
		, m_inIntrinsics( "Intrinsics", *this, false )
		, m_inImagePoint( "ImagePoint", *this )
		, m_inPlanePose( "PlanePose", *this )
		, m_outPort( "Output", *this )
    {
    }

	/** Method that computes the result. */
	void compute( Measurement::Timestamp t )
	{
		// plane equation
		Math::Vector< 3 > planeNormal = m_inPlanePose.get()->rotation() * Math::Vector< 3 >( 0, 0, 1 );
		double planeD = - ublas::inner_prod( planeNormal, m_inPlanePose.get()->translation() );

		// compute point in front of camera
		Math::Matrix< 3, 3 > Kinv = Math::invert_matrix( *m_inIntrinsics.get() );
		Math::Vector< 2 > imgPoint = *m_inImagePoint.get();
		Math::Vector< 3 > worldPoint = ublas::prod( Kinv, Math::Vector< 3 >( imgPoint( 0 ), imgPoint( 1 ), 1 ) );
		
		// intersect screen point with plane
		double lambda = - planeD / ublas::inner_prod( planeNormal, worldPoint );
		Math::Vector< 3 > planePoint = lambda * worldPoint;
		
		m_outPort.send( Measurement::Position( t, planePoint ) );
		
		LOG4CPP_DEBUG( logger, "plane pose: " << m_inPlanePose.get() );
		LOG4CPP_DEBUG( logger, "plane normal: " << planeNormal );
		LOG4CPP_DEBUG( logger, "plane distance: " << planeD );
		LOG4CPP_DEBUG( logger, "image point: " << imgPoint );
		LOG4CPP_DEBUG( logger, "worldPoint: " << worldPoint );
		LOG4CPP_DEBUG( logger, "planePoint: " << planePoint );
    }

protected:
	/** Input port A of the component. */
	Dataflow::TriggerInPort< Measurement::Matrix3x3 > m_inIntrinsics;

	/** Input port B of the component. */
	Dataflow::TriggerInPort< Measurement::Position2D > m_inImagePoint;

	/** Input port B of the component. */
	Dataflow::TriggerInPort< Measurement::Pose > m_inPlanePose;
	
	/** Output port of the component. */
	Dataflow::TriggerOutPort< Measurement::Position > m_outPort;
};


UBITRACK_REGISTER_COMPONENT( Dataflow::ComponentFactory* const cf ) {
	cf->registerComponent< ImagePointPlaneIntersection> ( "ImagePointPlaneIntersection" );
}

} } // namespace Ubitrack::Components
