/*
 * Ubitrack - Library for Ubiquitous Tracking
 * Copyright 2006, Technische Universitaet Muenchen, and individual
 * contributors as indicated by the @authors tag. See the
 * copyright.txt in the distribution for a full listing of individual
 * contributors.
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA, or see the FSF site: http://www.fsf.org.
 */


/**
 * @ingroup dataflow_components
 * @file
 * Computes image difference to a reference image and extracts a single point with high difference.
 *
 * @author Daniel Pustka <daniel.pustka@in.tum.de>
 */
#include <Ubitrack/Dataflow/TriggerComponent.h>
#include <Ubitrack/Dataflow/TriggerInPort.h>
#include <Ubitrack/Dataflow/TriggerOutPort.h>
#include <Ubitrack/Dataflow/ComponentFactory.h>
#include <Ubitrack/Measurement/Measurement.h>
#include <Vision/Image.h>
#include <cv.h>

static log4cpp::Category& logger( log4cpp::Category::getInstance( "Drivers.LightPointDetection" ) );

namespace Ubitrack { namespace Components {

/**
 * @ingroup dataflow_components
 * Computes image difference to a reference image and extracts a single point with high difference.
 *
 * @par Input Ports
 * TriggerInPort<ImageMeasurement> with name "Input".
 * TriggerInPort<ImageMeasurement> with name "Reference".
 *
 * @par Output Ports
 * TriggerOutPort<Position2> with name "Output".
 *
 * @par Configuration
 * None.
 *
 * @par Operation
 *
 */
class LightPointDetection
	: public Dataflow::TriggerComponent
{
public:
	/**
	 * UTQL component constructor.
	 *
	 * @param sName Unique name of the component.
	 * @param subgraph UTQL subgraph
	 */
	LightPointDetection( const std::string& sName, boost::shared_ptr< Graph::UTQLSubgraph > subgraph )
		: Dataflow::TriggerComponent( sName, subgraph )
		, m_inPort( "Input", *this )
		, m_inPortReference( "Background", *this )
		, m_outPort( "Output", *this )
		, m_threshold( 50 )
		, m_minSize( 2 )
		, m_maxSize( 10 )
	{
		subgraph->m_DataflowAttributes.getAttributeData( "threshold", m_threshold );
		subgraph->m_DataflowAttributes.getAttributeData( "minSize", m_minSize );
		subgraph->m_DataflowAttributes.getAttributeData( "maxSize", m_maxSize );
	}

	/** Method that computes the result. */
	void compute( Measurement::Timestamp t )
	{
		LOG4CPP_TRACE( logger, "got images" );
		
		// create new image
		Vision::Image differenceImage( m_inPort.get()->width, m_inPort.get()->height, m_inPort.get()->nChannels );
		Vision::Image thresholdImage( m_inPort.get()->width, m_inPort.get()->height, 1 );
			
		// compute difference
		cvAbsDiff( *m_inPort.get(), *m_inPortReference.get(), differenceImage );
		
		// threshold
		cvThreshold( differenceImage, thresholdImage, m_threshold, 255, CV_THRESH_BINARY );
		
		// find points
		// create memory storage that will contain all the dynamic data
		CvMemStorage* pStorage = cvCreateMemStorage( 0 );

		// Create dynamic structure and sequence.
		CvSeq* pContours = cvCreateSeq( CV_SEQ_ELTYPE_POINT, sizeof( CvSeq ), sizeof( CvPoint ), pStorage );

		// Find all contours.
		cvFindContours( thresholdImage, pStorage, &pContours, sizeof( CvContour ), 
			CV_RETR_LIST, CV_CHAIN_APPROX_SIMPLE, cvPoint( 0, 0 ) );

		// approximate all contours by polygons and check size and corner number
		int nPoints = 0;
		Math::Vector< 2 > point;
		for ( CvSeq* pContour = pContours; pContour; pContour = pContour->h_next ) 
		{
			CvRect boundingRect = cvBoundingRect( pContour );
			
			float fDiameter = sqrtf( static_cast< float >( boundingRect.height * boundingRect.height + boundingRect.width * boundingRect.width ) );
			LOG4CPP_TRACE( logger, "Difference region of size " << fDiameter );
			
			if ( fDiameter >= m_minSize && fDiameter <= m_maxSize )
			{
				nPoints++;
				point = Math::Vector< 2 >( boundingRect.x + boundingRect.width / 2.0, 
					boundingRect.y + boundingRect.height / 2.0 );
				LOG4CPP_DEBUG( logger, "Point Center " << point[0]<<" "<<point[1] << " Difference region of size  "<< fDiameter );
		
				
				// TODO: set ROI to 5x5-area around center and compute grey value moment for subpixel precision
				// cvSetImageROI( differenceImage, boundingRect );
				// cvMoments( ... )
			}
		}
		
		// cleanup
		cvReleaseMemStorage( &pStorage );

		if ( nPoints == 1 )
		{
			if ( !m_inPort.get()->origin )
				point( 1 ) = m_inPort.get()->height - 1 - point( 1 );
			m_outPort.send( Measurement::Position2D( t, point ) );
		}
	}

protected:
	/** threshold for laser point */
	int m_threshold;

	/** minimal size of laser point */
	int m_minSize;
	
	/** maximal size of laser point */
	int m_maxSize;
	
	/** Input port A of the component. */
	Dataflow::TriggerInPort< Vision::ImageMeasurement > m_inPort;

	/** Input port B of the component. */
	Dataflow::TriggerInPort< Vision::ImageMeasurement > m_inPortReference;

	/** Output port of the component. */
	Dataflow::TriggerOutPort< Measurement::Position2D > m_outPort;
};


UBITRACK_REGISTER_COMPONENT( Dataflow::ComponentFactory* const cf ) {
	cf->registerComponent< LightPointDetection > ( "LightPointDetection" );
}

} } // namespace Ubitrack::Components
