/*
 * Ubitrack - Library for Ubiquitous Tracking
 * Copyright 2006, Technische Universitaet Muenchen, and individual
 * contributors as indicated by the @authors tag. See the
 * copyright.txt in the distribution for a full listing of individual
 * contributors.
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA, or see the FSF site: http://www.fsf.org.
 */


/**
 * @ingroup driver_components
 * @file
 * GPS driver
 * This file contains the driver component to
 * get NMEA location data from a GPS receiver.
 *
 * The received data is sent via a push interface.
 *
 * @author Florian Echtler <echtler@in.tum.de>
 */

#include <map>
#include <string>
#include <cstdlib>
#include <log4cpp/Category.hh>

#include <asio.hpp>

#include <Ubitrack/Dataflow/ComponentFactory.h>
#include <Ubitrack/Dataflow/Component.h>
#include <Ubitrack/Dataflow/PushConsumer.h>
#include <Ubitrack/Dataflow/PushSupplier.h>
#include <Ubitrack/Measurement/Measurement.h>
#include <Ubitrack/Util/Exception.h>
#include <Ubitrack/Util/OS.h>

namespace Ubitrack { namespace Drivers {

template< class EventType >
class Aggregator
	: public Dataflow::Component
{

public:

	/** constructor */
	Aggregator( const std::string& sName, boost::shared_ptr< Graph::UTQLSubgraph > pConfig )
		: Dataflow::Component( sName )
		, m_outPort( "Output", *this )
		, m_absPort( "Absolute", *this, boost::bind( &Aggregator::absFunc, this, _1 ) )
		, m_incPort( "Increment", *this, boost::bind( &Aggregator::incFunc, this, _1 ) )
		, m_valid( false )
	{
	}

	/** destructor */
	~Aggregator()
	{}


	void absFunc ( const EventType& data )
	{
		m_accu = data;
		m_outPort.send( m_accu );
		if ( !m_valid )
			m_valid = true;
	}

	void incFunc ( const EventType& data )
	{
		if ( !m_valid )
		{
			UBITRACK_THROW( "no initialization value present" );
		}
		else
		{
			Measurement::Timestamp t = data.time();
			m_accu = EventType( t, *m_accu * *data );
			m_outPort.send( m_accu );
		}
	}


protected:
	Dataflow::PushSupplier< EventType > m_outPort;

	Dataflow::PushConsumer< EventType > m_absPort;
	Dataflow::PushConsumer< EventType > m_incPort;

	EventType m_accu;

	bool m_valid;
};


UBITRACK_REGISTER_COMPONENT( Dataflow::ComponentFactory* const cf ) {
	cf->registerComponent< Aggregator< Measurement::Pose > > ( "MausometerPoseAggregator" );
}


} } // namespace Ubitrack::Drivers
