/*
 * Ubitrack - Library for Ubiquitous Tracking
 * Copyright 2006, Technische Universitaet Muenchen, and individual
 * contributors as indicated by the @authors tag. See the
 * copyright.txt in the distribution for a full listing of individual
 * contributors.
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA, or see the FSF site: http://www.fsf.org.
 */


/**
 * @ingroup driver_components
 * @file
 * Mausometer driver
 * This file contains the driver component for
 * the mausometer optical odometric tracker
 * interfaced via ethernet.
 *
 * The received data is sent via a push interface.
 *
 * @author Manuel Huber <huberma@in.tum.de>
 */

#include "MausometerDriver.h"
#include <Ubitrack/Util/Exception.h>
#include <sstream>
#include <algorithm>
#include <iomanip>

static log4cpp::Category& logger( log4cpp::Category::getInstance( "Drivers.MausometerDriver" ) );

namespace Ubitrack { namespace Drivers {


using namespace Dataflow;


/** constructor */
MausometerDriver::MausometerDriver( const std::string& sName, boost::shared_ptr< Graph::UTQLSubgraph > subgraph )
	: Component( sName )
	, m_keepRunning(1)
	, m_synchronizer(1e3)
{
	m_sensorPorts[0].reset ( new Dataflow::PushSupplier< Measurement::Position2D > ( "Movement1", *this ) );
	m_sensorPorts[1].reset ( new Dataflow::PushSupplier< Measurement::Position2D > ( "Movement2", *this ) );
	m_sensorPorts[2].reset ( new Dataflow::PushSupplier< Measurement::Position2D > ( "Movement3", *this ) );

	m_buttonPorts[0].reset ( new Dataflow::PushSupplier< Measurement::Button > ( "Button1", *this ) );
	m_buttonPorts[1].reset ( new Dataflow::PushSupplier< Measurement::Button > ( "Button2", *this ) );
	m_buttonPorts[2].reset ( new Dataflow::PushSupplier< Measurement::Button > ( "Button3", *this ) );


	std::string mausIP("131.159.10.60");
	std::string mausPort("10001");

	LOG4CPP_CRIT( logger, "entering MausometerDriver" );

	if ( subgraph->m_DataflowAttributes.hasAttribute( "mausIP" ) )
	{
		mausIP = subgraph->m_DataflowAttributes.getAttributeString( "mausIP" );
	}

	if ( subgraph->m_DataflowAttributes.hasAttribute( "mausPort" ) )
	{
		mausPort = subgraph->m_DataflowAttributes.getAttributeString( "mausPort" );
	}

	LOG4CPP_INFO( logger, "accessing " << mausIP << ":" << mausPort );

	using asio::ip::tcp;

	m_IoService.reset( new asio::io_service );

	tcp::resolver resolver(*m_IoService);
	tcp::resolver::query query(tcp::v4(), mausIP, mausPort);
	tcp::resolver::iterator iterator = resolver.resolve(query);

	m_Socket.reset( new tcp::socket(*m_IoService) );
	m_Socket->connect(*iterator);

	if ( !m_Socket->is_open() )
	{
		UBITRACK_THROW( "Cannot open socket" );
	}

	receive_line_ptr = &receive_line[0];
	receive_line_end_ptr = &receive_line[max_receive_length];

	// start scanning on board
	asio::write(*m_Socket, asio::buffer("\nB\nY\n", 5));

	m_Socket->async_read_some(asio::buffer( receive_data, max_receive_length ),
							  boost::bind (&MausometerDriver::HandleReceive, this, asio::placeholders::error, asio::placeholders::bytes_transferred ) );

	m_thread.reset( new boost::thread( boost::bind( &asio::io_service::run, m_IoService.get() ) ) );
}


/** destructor */
MausometerDriver::~MausometerDriver()
{
	m_keepRunning = 0;
	m_IoService->stop();

	if ( m_thread )
		m_thread->join();

	m_thread.reset();

	if ( m_Socket )
	{
		m_Socket->close();
		m_Socket.reset();
	}

	m_IoService.reset();
}


void MausometerDriver::HandleReceive(const asio::error_code err, size_t length)
{
	Ubitrack::Measurement::Timestamp timestamp = Ubitrack::Measurement::now();

	if ( err && err != asio::error::message_size )
	{
		std::ostringstream msg;
		msg << "Error receiving from socket: \"" << err << "\"";
		LOG4CPP_CRIT( logger, msg.str() );
		goto restart;
	}

	if ( length >= max_receive_length )
	{
		LOG4CPP_CRIT( logger, "received more than max_receive_length bytes." );
		goto restart;
	}

	for (size_t p = 0; p < length; ++p)
	{
		switch ( receive_data[p] )
		{
		case '\r':
			break;
		case '\n':
			*receive_line_ptr = 0;
			processLine(timestamp);
			receive_line_ptr = &receive_line[0];
			break;
		default:
			if ( receive_line_ptr >= receive_line_end_ptr )
			{
				LOG4CPP_CRIT( logger, "incomplete line with more than max_recieve_length chars. flushing" );
				receive_line_ptr = &receive_line[0];
			}
			else
			{
				*receive_line_ptr = receive_data[p];
				receive_line_ptr++;
			}
			break;
		}
	}

restart:
	if ( m_keepRunning )
	{
		m_Socket->async_read_some(asio::buffer( receive_data, max_receive_length ),
								  boost::bind (&MausometerDriver::HandleReceive, this, asio::placeholders::error, asio::placeholders::bytes_transferred ) );
	}
}

void MausometerDriver::processLine(Ubitrack::Measurement::Timestamp time)
{
	LOG4CPP_DEBUG( logger, "received line" );
	LOG4CPP_TRACE( logger, receive_line );

	std::istringstream in( receive_line );

	unsigned long ts;
	Ubitrack::Math::Vector< 2 > movements[3];

	in >> ts;
	time = m_synchronizer.convertNativeToLocal( ts, time );

	for (int i=0; i<3; ++i)
	{
		in >> movements[i]( 0 );
		in >> movements[i]( 1 );
		if ( movements[i]( 0 ) > 127 )
		{
			movements[i]( 0 ) -= 256;
		}
		if ( movements[i]( 1 ) > 127 )
		{
			movements[i]( 1 ) -= 256;
		}

		Ubitrack::Measurement::Position2D measurement( time, movements[i] );
		Ubitrack::Measurement::Button trigger( time, 1 );
		m_sensorPorts[i]->send ( measurement );
		m_buttonPorts[i]->send ( trigger );

		/*
		if ( ( movements[i]( 0 ) != 0 ) || ( movements[i]( 1 ) != 0 ) )
		{
		}
		*/
	}

// 	std::ostringstream log;
// 	log << "Recived movement: ";
// 	for (int i=0; i<3; ++i)
// 	{
// 		log << "[" << movements[i]( 0 ) << " / " << movements[i]( 1 ) << "] ";
// 	}
// 	LOG4CPP_INFO( logger, log.str() );

}



UBITRACK_REGISTER_COMPONENT( ComponentFactory* const cf ) {
	cf->registerComponent< MausometerDriver > ( "MausometerDriver" );
}

} } // namespace Ubitrack::Drivers
