/*
 * Ubitrack - Library for Ubiquitous Tracking
 * Copyright 2006, Technische Universitaet Muenchen, and individual
 * contributors as indicated by the @authors tag. See the
 * copyright.txt in the distribution for a full listing of individual
 * contributors.
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA, or see the FSF site: http://www.fsf.org.
 */

/**
 * @ingroup calibration
 * @file
 * TODO: Documentation
 *
 * @author Manuel Huber <huberma@in.tum.de>
 * @author Michael Schlegel <schlegem@in.tum.de>
 */

#ifndef __UBITRACK_CALIBRATION_FUNCTION_MAUSOMETERTRACKER_H_INCLUDED__
#define  __UBITRACK_CALIBRATION_FUNCTION_MAUSOMETERTRACKER_H_INCLUDED__

namespace Ubitrack { namespace Drivers { namespace Function {

#include <Ubitrack/Math/Pose.h>

  template< typename T >
  struct MausCalibration
  {
	  T x;
	  T y;
	  T gamma;
	  T xScale;
	  T yScale;
  };

  /**
   * Function that calculates 2D movements of a mouse tracker wrt given pose differences of a reference target frame.
   * The reference target frame has 3DoF (dx, dy, dGamma)
   *
   */
  template< typename T >
  class TrackerProblem
  {
  public:
    /**
     * Constructor.
     * @param (poses, scales)
     *  < <x1, y1, g1, s1>, <x2, y2, g2, s2> , <x3, y3, g3, s3>>
     */
	TrackerProblem( const std::vector< MausCalibration<T> >& c )
      : m_calibData( c )
    {}

    /**
     * return the size of the result vector
     * in this case
     */
    unsigned int size() const
    { return 6; }

    /*
     * @param result 2n vector (messwert)
     * @param input containing the parameters for the 3D pose of the mouse, the scale and the movement of the reference object (mx, my, mGamma, sx, sy)
     */
    template< class VT1, class VT2 >
    void evaluate( VT1& result, const VT2& input ) const
    {

		T x = input( 0 );
		T y = input( 1 );
		T gamma = input( 2 );

		for (int maus=0; maus < 3; ++maus )
		{

			T temp1=-m_calibData[maus].gamma;
			T temp2=gamma+temp1;
			T temp3=sin(m_calibData[maus].gamma);
			T temp4=cos(temp2);
			T temp5=cos(m_calibData[maus].gamma);
			T temp6=sin(temp2);

			result( 2*maus + 0 ) = (m_calibData[maus].x*temp4+(x-m_calibData[maus].x)*temp5-m_calibData[maus].y*temp6+y*temp3-m_calibData[maus].y*temp3)/m_calibData[maus].xScale;
			result( 2*maus + 1 ) = (m_calibData[maus].y*temp4+(y-m_calibData[maus].y)*temp5+m_calibData[maus].x*temp6-x*temp3+m_calibData[maus].x*temp3)/m_calibData[maus].yScale;
		}
    }

    /*
     * @param result the n calculated mouse movements
     * @param input the 5-vector of the mouse pose ( mx, my, mGamma, sx, sy)
     * @param J jacobian (4nx5-matrix)
     */
    template< class VT1, class VT2, class MT >
    void evaluateWithJacobian( VT1& result, const VT2& input, MT& J ) const
    {
      evaluate( result, input );
      jacobian( input, J );
    }

    /*
     * @param input the 3-vector of the target pose mouse pose ( mx, my, mGamma, sx, sy)
     * @param J jacobian (2nx5-matrix)
     */
    template< class VT2, class MT >
    void jacobian( const VT2& input, MT& J ) const
    {
		// J is independent of x and y..
		// T x = input( 0 );
		// T y = input( 1 );
		T gamma = input( 2 );

		for (int maus=0; maus<3; ++maus)
		{
			T temp46=1/m_calibData[maus].xScale;
			T temp51=-m_calibData[maus].gamma;
			T temp52=gamma+temp51;
			T temp49=sin(m_calibData[maus].gamma);
			T temp59=1/m_calibData[maus].yScale;
			T temp47=cos(m_calibData[maus].gamma);
			T temp53=cos(temp52);
			T temp55=sin(temp52);

			J( maus*2 + 0 , 0)=temp46*temp47;
			J( maus*2 + 0 , 1)=temp46*temp49;
			J( maus*2 + 0 , 2)=temp46*(-(m_calibData[maus].y*temp53)-m_calibData[maus].x*temp55);
			J( maus*2 + 1 , 0)=-(temp49*temp59);
			J( maus*2 + 1 , 1)=temp47*temp59;
			J( maus*2 + 1 , 2)=(m_calibData[maus].x*temp53-m_calibData[maus].y*temp55)*temp59;
		}
	}

  protected:
	  const std::vector< MausCalibration< double > >& m_calibData;
  };

} } } // namespace Ubitrack::Calibration::Function

#endif
