/*
 * Ubitrack - Library for Ubiquitous Tracking
 * Copyright 2006, Technische Universitaet Muenchen, and individual
 * contributors as indicated by the @authors tag. See the
 * copyright.txt in the distribution for a full listing of individual
 * contributors.
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA, or see the FSF site: http://www.fsf.org.
 */


/**
 * @ingroup driver_components
 * @file
 * GPS driver
 * This file contains the driver component to
 * get NMEA location data from a GPS receiver.
 *
 * The received data is sent via a push interface.
 *
 * @author Florian Echtler <echtler@in.tum.de>
 */

#include <map>
#include <string>
#include <cstdlib>
#include <log4cpp/Category.hh>

#include <asio.hpp>

#include <Ubitrack/Dataflow/ComponentFactory.h>
#include <Ubitrack/Dataflow/TriggerComponent.h>
#include <Ubitrack/Dataflow/TriggerInPort.h>
#include <Ubitrack/Dataflow/TriggerOutPort.h>
#include <Ubitrack/Measurement/Measurement.h>
#include <Ubitrack/Util/Exception.h>
#include <Ubitrack/Util/OS.h>


#include <Ubitrack/Math/LevenbergMarquardt.h>

#include "Function/TrackerProblem.h"

namespace Ubitrack { namespace Drivers {

namespace ublas = boost::numeric::ublas;

class MausometerTracker
	: public Dataflow::TriggerComponent
{

public:

	/** constructor */
	MausometerTracker( const std::string& sName, boost::shared_ptr< Graph::UTQLSubgraph > pConfig )
		: Dataflow::TriggerComponent( sName, pConfig )
		, m_poseDiff( "WagenPoseDelta", *this )
		, m_mausIn1( "Movement1", *this )
		, m_mausIn2( "Movement2", *this )
		, m_mausIn3( "Movement3", *this )
		, m_mausPose1( "MausPose1", *this )
		, m_mausScale1( "MausScale1", *this )
		, m_mausPose2( "MausPose2", *this )
		, m_mausScale2( "MausScale2", *this )
		, m_mausPose3( "MausPose3", *this )
		, m_mausScale3( "MausScale3", *this )
	{ }


	/** destructor */
	~MausometerTracker()
	{}

	void compute( Measurement::Timestamp t )
	{
		if (m_calibData.size() == 0)
		{
			Math::Vector< 3 > scales[3];
			scales[0] = *m_mausScale1.get();
			scales[1] = *m_mausScale2.get();
			scales[2] = *m_mausScale3.get();

			Math::Vector< 3 > poses[3];
			poses[0] = *m_mausPose1.get();
			poses[1] = *m_mausPose2.get();
			poses[2] = *m_mausPose3.get();

			for (int i=0; i<3; ++i)
			{
				Function::MausCalibration< double > calib;
				calib.xScale = scales[i](0);
				calib.yScale = scales[i](1);
				calib.x = poses[i](0);
				calib.y = poses[i](1);
				calib.gamma = poses[i](2);

				m_calibData.push_back( calib );
			}
		}



		// get Sensordata
		Math::Vector< 2 > m[3];
		m[0] = *m_mausIn1.get();
		m[1] = *m_mausIn2.get();
		m[2] = *m_mausIn3.get();

		ublas::vector< double > params( 3 );
		params( 0 ) = 0;
		params( 1 ) = 0;
		params( 2 ) = 0;

		ublas::vector< double > measurements( 2*3 );
		for (int maus=0; maus<3; ++maus)
		{
			measurements( maus*2 + 0 ) = m[maus](0);
			measurements( maus*2 + 1 ) = m[maus](1);
		}

		Function::TrackerProblem< double > problem( m_calibData );

		// residuum is not used anywhere, but generates compiler warnings..
		// double residuum =
		Math::levenbergMarquardt( problem, params, measurements,
								  Math::OptTerminate( 100, 1e-6 ), Math::OptNoNormalize() );


		Math::Pose p( Math::Quaternion(Math::Vector< 3 >(0,0,1), params(2)),
					  Math::Vector< 3 >(params(0), params(1), 0) );

		m_poseDiff.send( Measurement::Pose( t, p ) );
	}


protected:
	Dataflow::TriggerOutPort< Measurement::Pose > m_poseDiff ;

	Dataflow::TriggerInPort< Measurement::Position2D > m_mausIn1;
	Dataflow::TriggerInPort< Measurement::Position2D > m_mausIn2;
	Dataflow::TriggerInPort< Measurement::Position2D > m_mausIn3;
	Dataflow::TriggerInPort< Measurement::Position > m_mausPose1;
	Dataflow::TriggerInPort< Measurement::Position > m_mausScale1;
	Dataflow::TriggerInPort< Measurement::Position > m_mausPose2;
	Dataflow::TriggerInPort< Measurement::Position > m_mausScale2;
	Dataflow::TriggerInPort< Measurement::Position > m_mausPose3;
	Dataflow::TriggerInPort< Measurement::Position > m_mausScale3;

	std::vector< Function::MausCalibration< double > > m_calibData;
};


UBITRACK_REGISTER_COMPONENT( Dataflow::ComponentFactory* const cf ) {
	cf->registerComponent< MausometerTracker > ( "MausometerTracker" );
}


} } // namespace Ubitrack::Drivers
