/*
 * Ubitrack - Library for Ubiquitous Tracking
 * Copyright 2006, Technische Universitaet Muenchen, and individual
 * contributors as indicated by the @authors tag. See the
 * copyright.txt in the distribution for a full listing of individual
 * contributors.
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA, or see the FSF site: http://www.fsf.org.
 */


#include "Network2Module.h"

#include <iostream>
#include <sstream>

#include <boost/array.hpp>
#include <boost/bind.hpp>

#include <log4cpp/Category.hh>

#include <Ubitrack/Dataflow/ComponentFactory.h>
#include <Ubitrack/Measurement/Measurement.h>
#include <Ubitrack/Util/OS.h>


namespace Ubitrack { namespace Drivers {

static log4cpp::Category& logger( log4cpp::Category::getInstance( "Drivers.Network2.Module" ) );

Network2Module::Network2Module( const Network2ModuleKey& moduleKey, boost::shared_ptr< Graph::UTQLSubgraph >, FactoryHelper* pFactory )
	: Dataflow::Module< Network2ModuleKey, Dataflow::SubgraphIdKey, Network2Module, Network2ComponentBase >( moduleKey, pFactory )
	, m_bStopModule( false )
{
	m_pSocketSingleton = Network2SocketSingleton::getSingleton( moduleKey );
	
	m_pSocketSingleton->async_receive_from(
		asio::buffer( m_receiveBuffer, max_receive_length ),
		m_senderEndpoint,
		boost::bind( &Network2Module::HandleReceive, this, _1, _2 )
	);

	// start keepalive thread
	m_pKeepAliveThread = boost::shared_ptr< boost::thread >( new boost::thread( boost::bind( &Network2Module::runKeepAlive, this ) ) );
}


Network2Module::~Network2Module()
{
	m_bStopModule = true;

	// stop keepalive thread
	m_pKeepAliveThread->join();
	
	m_pSocketSingleton->removeReceiver();
}


void Network2Module::HandleReceive( const asio::error_code err, size_t length )
{
	Measurement::Timestamp recvtime = Measurement::now();

	LOG4CPP_DEBUG( logger, "Received " << length << " bytes from " << m_senderEndpoint );
	
	if ( length >= max_receive_length )
	{
		LOG4CPP_ERROR( logger, "Too many bytes received" );
		return; // abort
	}

	// some error checking
	if ( err )
	{
		switch ( err.value() )
		{
		case asio::error::message_size:
		case asio::error::connection_refused:
		case asio::error::connection_reset:
			// only log, but ignore otherwise
			LOG4CPP_INFO( logger, "Socket error (continuing): " << err.message() );
			break;

		default:
			// log as error and abort
			LOG4CPP_ERROR( logger, "Socket error (aborting): " << err.message() );
			return;
		}
	}
	else
	{
		try
		{
			// make receive data null terminated and create a string stream
			m_receiveBuffer[ length ] = 0;
			std::string data( m_receiveBuffer );
			LOG4CPP_TRACE( logger, "data: " << data );
			std::istringstream stream( data );
			boost::archive::text_iarchive message( stream );

			// parse packet
			std::string name;
			message >> name;
			LOG4CPP_DEBUG( logger, "Message for component " << name );
			
			ComponentKey key( name );

			if ( hasComponent( key ) ) {
				boost::shared_ptr< Network2ComponentBase > pComp = getComponent( key );
				pComp->parse( message, recvtime, m_senderEndpoint );
			}
			else
				LOG4CPP_WARN( logger, "No such component " << name );
		}
		catch ( const std::exception& e )
		{
			LOG4CPP_ERROR( logger, "Caught exception " << e.what() );
		}
	}

	// restart receiving new packet
	m_pSocketSingleton->async_receive_from(
		asio::buffer( m_receiveBuffer, max_receive_length ),
		m_senderEndpoint,
		boost::bind( &Network2Module::HandleReceive, this, _1, _2 )
	);
}


void Network2Module::runKeepAlive()
{
	while ( !m_bStopModule )
	{
		// sleep 2 second
		Util::sleep( 2000 );

		// call keepalive method on all components
		ComponentList l( getAllComponents() );
		for ( ComponentList::iterator it = l.begin(); it != l.end(); it++ )
			(*it)->sendKeepAlive();
	}
}


} } // namespace Ubitrack::Drivers

