/*
 * Ubitrack - Library for Ubiquitous Tracking
 * Copyright 2006, Technische Universitaet Muenchen, and individual
 * contributors as indicated by the @authors tag. See the 
 * copyright.txt in the distribution for a full listing of individual
 * contributors.
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA, or see the FSF site: http://www.fsf.org.
 */


/**
 * @ingroup driver_components
 * @file
 * Network2Module
 * 
 * A common module for both the network2source and the network2sink components.
 * The module  provides a  network socket on which to receive both packets and 
 * subscription requests.
 *
 * @author Daniel Pustka <pustka@in.tum.de>
 */

#ifndef _NETWORK2MODULE_H_
#define _NETWORK2MODULE_H_

#include "Network2SocketSingleton.h"

// WARNING: all boost/serialization headers should be
//          included AFTER all boost/archive headers
#include <boost/archive/text_iarchive.hpp>
#include <boost/serialization/vector.hpp>
#include <boost/serialization/utility.hpp>

#include <Ubitrack/Dataflow/Component.h>
#include <Ubitrack/Dataflow/Module.h>
#include <Ubitrack/Measurement/Timestamp.h>


namespace Ubitrack { namespace Drivers {

// forward declaration
class Network2ComponentBase;

/**
 * Module key for network source.
 * Represents the port number on which to listen.
 */
MAKE_DATAFLOWCONFIGURATIONATTRIBUTEKEY_DEFAULT( Network2ModuleKey, int, "networkPort", 0x5554 );


/**
 * Module for network source.
 * Does all the dirty asio work.
 */
class Network2Module
	: public Dataflow::Module< Network2ModuleKey, Dataflow::SubgraphIdKey, Network2Module, Network2ComponentBase >
{
public:
	/** constructor */
	Network2Module( const Network2ModuleKey& key, boost::shared_ptr< Graph::UTQLSubgraph > subgraph, FactoryHelper* pFactory );

	/** destruktor */
	~Network2Module();

	/** thread method */
	void HandleReceive( const asio::error_code err, size_t length );

	boost::shared_ptr< Network2ComponentBase > createComponent( const std::string& type, const std::string& name, 
		boost::shared_ptr< Graph::UTQLSubgraph > subgraph, const ComponentKey& key, ModuleClass* pModule );	

	/** accessor method for components */
	asio::io_service& getIoService()
	{ return m_pSocketSingleton->getIoService(); }

	/** send something to somebody */
	void send_to( const std::string& data, const asio::ip::udp::endpoint& to )
	{ m_pSocketSingleton->send_to( data, to ); }

protected:

	/** thread method for keepalive-checking */
	void runKeepAlive();

	// Receive data. Do not touch from outside of async network thread
	enum { max_receive_length = 10240, receive_buffer_size = 10242 };
	char m_receiveBuffer[ receive_buffer_size ];
	asio::ip::udp::endpoint m_senderEndpoint;

	boost::shared_ptr< boost::thread > m_pKeepAliveThread;

	/** the socket singleton */
	boost::shared_ptr< Network2SocketSingleton > m_pSocketSingleton;
	
	// stop the module
	bool m_bStopModule;
};


/**
 * Virtual base class for all other components
 */
class Network2ComponentBase
	: public Network2Module::Component
{
public:

	/** constructor */
	Network2ComponentBase( const std::string& name, boost::shared_ptr< Graph::UTQLSubgraph > subgraph, const Dataflow::SubgraphIdKey& componentKey, Network2Module* pModule )
		: Network2Module::Component( name, componentKey, pModule )
	{}

	/** virtual destructor */
	virtual ~Network2ComponentBase()
	{}

	/** parse a received network packet */
	virtual void parse( boost::archive::text_iarchive& ar, Measurement::Timestamp recvtime, const asio::ip::udp::endpoint& sender )
	{}

	/** called by the module every 5 seconds to keep event subscriptions alive */
	virtual void sendKeepAlive()
	{}

};

} } // namespace Ubitrack::Drivers

#endif

