/*
 * Ubitrack - Library for Ubiquitous Tracking
 * Copyright 2006, Technische Universitaet Muenchen, and individual
 * contributors as indicated by the @authors tag. See the 
 * copyright.txt in the distribution for a full listing of individual
 * contributors.
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA, or see the FSF site: http://www.fsf.org.
 */

/**
 * @ingroup driver_components
 * @file
 * Network2Sink component.
 *
 * @author Daniel Pustka <pustka@in.tum.de>
 */

#include "Network2Module.h"

// WARNING: all boost/serialization headers should be
//          included AFTER all boost/archive headers
#include <boost/archive/text_oarchive.hpp>

#include <sstream>

#include <boost/bind.hpp>
#include <boost/thread.hpp>

#include <Ubitrack/Dataflow/PushConsumer.h>

#include "RequestPacket.h"

using asio::ip::udp;

namespace Ubitrack { namespace Drivers {

/**
 * @ingroup dataflow_components
 * Transmits measurements over the network.
 */
template< class EventType > 
class Network2Sink 
	: public Network2ComponentBase
{
public:

	/** constructor */
	Network2Sink( const std::string& name, boost::shared_ptr< Graph::UTQLSubgraph > subgraph, const Dataflow::SubgraphIdKey& componentKey, Network2Module* pModule )
		: Network2ComponentBase( name, subgraph, componentKey, pModule )
		, m_inPort( "Input", *this, boost::bind( &Network2Sink::eventIn, this, _1 ) )
		, m_logger( log4cpp::Category::getInstance( "Drivers.Network2.Sink" ) )
	{
	}

protected:

	/** Proxy class for the receiving side */
	class ReceiverProxy
	{
	public:
		// just for map
		ReceiverProxy()
		{}

		ReceiverProxy( boost::shared_ptr< asio::ip::udp::endpoint > pEndpoint, const std::string& receiverName )
			: m_pEndpoint( pEndpoint )
			, m_receiverName( receiverName )
			, m_lastAliveTime( Measurement::now() )
		{}

		/** send an event to the receiver */
		void send( Network2Module& module, const EventType& m )
		{
			std::ostringstream stream;
			boost::archive::text_oarchive packet( stream );

			std::string suffix( "\n" );
			Measurement::Timestamp timeNow( Measurement::now() );

			// serialize the measurement, component name and current local time
			packet << m_receiverName;
			packet << m;
			packet << timeNow;
			packet << suffix;

			module.send_to( stream.str(), *m_pEndpoint );
		}

		/** is the receiver still alive? */
		bool alive( Measurement::Timestamp t ) const
		{ return t <= m_lastAliveTime + 6000000000LL; }

		/** receive keep-alive message */
		void keepAlive( Measurement::Timestamp t )
		{ m_lastAliveTime = t; }

	protected:
		boost::shared_ptr< udp::endpoint > m_pEndpoint;
		std::string m_receiverName;
		Measurement::Timestamp m_lastAliveTime;
	};

	// shortcut for maps of receiverproxys
	typedef std::map< std::string, ReceiverProxy > ProxyList;

	// receive a new pose from the dataflow
	void eventIn( const EventType& m )
	{
		boost::mutex::scoped_lock l( m_mutex );
		Measurement::Timestamp t( Measurement::now() );
		for ( typename ProxyList::iterator it = m_receivers.begin(); it != m_receivers.end(); )
			if ( !it->second.alive( t ) )
			{
				LOG4CPP_INFO( m_logger, getName() << ": receiver " << it->first << " timed out, removing" );
				m_receivers.erase( it++ );
			}
			else
			{
				it->second.send( *m_pModule, m );
				it++;
			}
	}


	void parse( boost::archive::text_iarchive& ar, Measurement::Timestamp recvtime, const asio::ip::udp::endpoint& sender )
	{
		boost::mutex::scoped_lock l( m_mutex );

		// parse request packet
		RequestPacket rp;
		ar >>  rp;

		typename ProxyList::iterator it = m_receivers.find( rp.sComponentName );
		if ( it != m_receivers.end() )
		{
			// already in map, update time
			it->second.keepAlive( recvtime );
		}
		else
		{
			boost::shared_ptr< udp::endpoint > pEndpoint;
			if ( !rp.sHost.empty() )
			{
				// resolve destination pair and store the remote endpoint
				udp::resolver resolver( m_pModule->getIoService() );

				std::ostringstream portString;
				portString << rp.nPort;

				udp::resolver::query query( udp::v4(), rp.sHost, portString.str() );
				pEndpoint.reset( new udp::endpoint( *resolver.resolve( query ) ) );
			}
			else
				// take sender of packet as receiver endpoint
				pEndpoint.reset( new udp::endpoint( sender ) );

			// add receiver to map
			m_receivers[ rp.sComponentName ] = ReceiverProxy( pEndpoint, rp.sComponentName );

			LOG4CPP_INFO( m_logger, getName() << ": Network data request from " << *pEndpoint << ":" << rp.sComponentName );
		}
	}

	// thread synchronization
	boost::mutex m_mutex;

	// consumer port
	Dataflow::PushConsumer< EventType > m_inPort;

	// map of receivers
	ProxyList m_receivers;

	// logging
	log4cpp::Category& m_logger;
};


} } // namespace Ubitrack::Drivers

