/*
 * Ubitrack - Library for Ubiquitous Tracking
 * Copyright 2006, Technische Universitaet Muenchen, and individual
 * contributors as indicated by the @authors tag. See the
 * copyright.txt in the distribution for a full listing of individual
 * contributors.
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA, or see the FSF site: http://www.fsf.org.
 */


#include "Network2SocketSingleton.h"
#include <log4cpp/Category.hh>

namespace Ubitrack { namespace Drivers {

static log4cpp::Category& logger( log4cpp::Category::getInstance( "Drivers.Network2.SocketSingleton" ) );

// declare static member
Network2SocketSingleton::SingletonMap Network2SocketSingleton::s_singletonMap;


Network2SocketSingleton::Network2SocketSingleton( int port )
{
	using asio::ip::udp;

	// create io service
	m_pIoService.reset( new asio::io_service );
	
	udp::endpoint endpoint( udp::v4(), port );
	
	// open receive socket
	LOG4CPP_DEBUG( logger, "Creating receiver on " << endpoint );
	m_pReceiveSocket.reset( new udp::socket( *m_pIoService, endpoint ) );

	// network thread runs until io_service is interrupted
	LOG4CPP_DEBUG( logger, "Starting network receiver thread" );
	m_pNetworkThread = boost::shared_ptr< boost::thread >( new boost::thread( boost::bind( &asio::io_service::run, m_pIoService.get() ) ) );
}


Network2SocketSingleton::~Network2SocketSingleton()
{
	// close socket
	boost::mutex::scoped_lock l( m_socketMutex );
	m_pReceiveSocket->close();
	m_pReceiveSocket.reset();

	// stop the network thread
	m_pIoService->stop();
	m_pIoService.reset();

	m_pNetworkThread->join();
	m_pNetworkThread.reset();
}


void Network2SocketSingleton::send_to( const std::string& data, const asio::ip::udp::endpoint& to )
{
	LOG4CPP_TRACE( logger, "Sending to " << to << ": " << data );
	boost::mutex::scoped_lock l( m_socketMutex );
	m_pReceiveSocket->send_to( asio::buffer( data.c_str(), data.size() ), to );

	// Note: If there are strange threading bugs, try to use async_send_to!
}


void Network2SocketSingleton::ReceiveProxy( const asio::error_code err, size_t length )
{
	boost::recursive_mutex::scoped_lock l( m_receiverMutex );
	if ( m_receiverFunction )
		m_receiverFunction( err, length );
	else
		// drop packet
		LOG4CPP_DEBUG( logger, "silently dropping received UDP packet (no receiver)" );
}


boost::shared_ptr< Network2SocketSingleton > Network2SocketSingleton::getSingleton( int port )
{
	boost::shared_ptr< Network2SocketSingleton >& p( s_singletonMap[ port ] );
	if ( !p )
		p.reset( new Network2SocketSingleton( port ) );
	return p;
}

} } // namespace Ubitrack::Drivers

