/*
 * Ubitrack - Library for Ubiquitous Tracking
 * Copyright 2006, Technische Universitaet Muenchen, and individual
 * contributors as indicated by the @authors tag. See the 
 * copyright.txt in the distribution for a full listing of individual
 * contributors.
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA, or see the FSF site: http://www.fsf.org.
 */


/**
 * @ingroup driver_components
 * @file
 * Network2Singleton
 * 
 * A singleton object that holds different important networking resources that should not be destroyed
 * when modules are deleted and re-instantiated.
 *
 * @author Daniel Pustka <pustka@in.tum.de>
 */

#ifndef _NETWORK2SOCKETSINGLETON_H_
#define _NETWORK2SOCKETSINGLETON_H_

// on windows, asio must be included before anything that possible includes windows.h
// don't ask why.
#include <asio.hpp>

#include <map>

#include <boost/utility.hpp>
#include <boost/shared_ptr.hpp>
#include <boost/function.hpp>
#include <boost/bind.hpp>
#include <boost/thread.hpp>

namespace Ubitrack { namespace Drivers {

/**
 * Singleton class for network resources required by a single socket.
 */
class Network2SocketSingleton
	: protected boost::noncopyable
{
protected:
	/** constructor */
	Network2SocketSingleton( int port );

public:
	/** destruktor */
	~Network2SocketSingleton();

	/** get the singleton */
	static boost::shared_ptr< Network2SocketSingleton > getSingleton( int port );
	
	/** accessor method for components */
	asio::io_service& getIoService()
	{ return *m_pIoService; }

	/** sets the receiver */
	template < class Buffer >
	void async_receive_from( const Buffer& buffer, asio::ip::udp::endpoint& to, boost::function< void ( asio::error_code, size_t ) > function )
	{
		{
			boost::recursive_mutex::scoped_lock l( m_receiverMutex );
			m_receiverFunction = function;
		}

		boost::mutex::scoped_lock l( m_socketMutex );
		m_pReceiveSocket->async_receive_from( buffer, to, 
			boost::bind( &Network2SocketSingleton::ReceiveProxy, this, asio::placeholders::error, asio::placeholders::bytes_transferred ) );
	}

	/** removes the receiver */
	void removeReceiver()
	{
		boost::recursive_mutex::scoped_lock l( m_receiverMutex );
		m_receiverFunction = 0;
	}
	
	/** sends udp packets */
	void send_to( const std::string& data, const asio::ip::udp::endpoint& to );
	
protected:

	/** proxies receive requests */
	void ReceiveProxy( const asio::error_code err, size_t length );

	/** singleton map type definition */
	typedef std::map< int, boost::shared_ptr< Network2SocketSingleton > > SingletonMap;
	
	/** static map of singletons by endpoint */
	static SingletonMap s_singletonMap;

	/** Boost::ASIO IO Service */
	boost::shared_ptr< asio::io_service > m_pIoService;

	/** thread running the IO service */
	boost::shared_ptr< boost::thread > m_pNetworkThread;

	/** socket for receiving data */
	boost::shared_ptr< asio::ip::udp::socket > m_pReceiveSocket;
	
	/** mutex for socket access */
	boost::mutex m_socketMutex;

	/** functor to call when data is received */
	boost::function< void ( asio::error_code, size_t ) > m_receiverFunction;

	/** mutex for receiver functor */
	boost::recursive_mutex m_receiverMutex;

};


} } // namespace Ubitrack::Drivers

#endif

