/*
 * Ubitrack - Library for Ubiquitous Tracking
 * Copyright 2006, Technische Universitaet Muenchen, and individual
 * contributors as indicated by the @authors tag. See the 
 * copyright.txt in the distribution for a full listing of individual
 * contributors.
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA, or see the FSF site: http://www.fsf.org.
 */

/**
 * @ingroup driver_components
 * @file
 * Network2Source component.
 *
 * @author Daniel Pustka <pustka@in.tum.de>
 */

#include "Network2Module.h"

#include <sstream>

#include <log4cpp/Category.hh>

#include <Ubitrack/Dataflow/PushSupplier.h>
#include <Ubitrack/Measurement/TimestampSync.h>

#include "RequestPacket.h"

using asio::ip::udp;

namespace Ubitrack { namespace Drivers {

/**
 * @ingroup dataflow_components
 * Receives measurements from the network.
 */
template< class EventType > 
class Network2Source 
	: public Network2ComponentBase
{
public:

	/** constructor */
	Network2Source( const std::string& name, boost::shared_ptr< Graph::UTQLSubgraph > subgraph, const Dataflow::SubgraphIdKey& componentKey, Network2Module* pModule )
		: Network2ComponentBase( name, subgraph, componentKey, pModule )
		, m_synchronizer( 1e9 )
		, m_outPort( "Output", *this )
		, m_logger( log4cpp::Category::getInstance( "Drivers.Network2.Source" ) )
	{
		// check for configuration
		std::string senderIP = subgraph->getEdge( "NetworkInput" )->getAttributeString( "senderIP" );
		std::string senderPort = subgraph->getEdge( "NetworkInput" )->getAttributeString( "senderPort" );
		m_senderName = subgraph->getEdge( "NetworkInput" )->getAttributeString( "remotePatternID" );

		// resolve destination pair and store the remote endpoint
		udp::resolver resolver( m_pModule->getIoService() );
		udp::resolver::query query( udp::v4(), senderIP, senderPort );
		m_pSenderEndpoint.reset( new udp::endpoint( *resolver.resolve( query ) ) );

		// notify sender
		sendRequest();
	}

	void sendKeepAlive()
	{ sendRequest(); }

protected:

	void parse( boost::archive::text_iarchive& ar, Measurement::Timestamp recvtime, const asio::ip::udp::endpoint& )
	{
		EventType mm( boost::shared_ptr< typename EventType::value_type >( new typename EventType::value_type() ) );
		Measurement::Timestamp sendtime;
		ar >> mm;
		ar >> sendtime;

		LOG4CPP_DEBUG( m_logger, "perceived host clock offset: " << static_cast< long long >( recvtime - sendtime ) * 1e-6 << "ms" );

		// subtract first timestamp to avoid losing timing precision
		if ( !m_firstTimestamp )
			m_firstTimestamp = sendtime;

		// synchronize sender time with receiver time
		Measurement::Timestamp correctedTime = m_synchronizer.convertNativeToLocal( sendtime - m_firstTimestamp, recvtime );

		// add offset of individual measurements
		correctedTime -= static_cast< long long >( sendtime - mm.time() );

		LOG4CPP_TRACE( m_logger, "Timestamps measurement: " << Measurement::timestampToShortString( mm.time() )
			<< ", sent: " << Measurement::timestampToShortString( sendtime )
			<< ", arrival: " << Measurement::timestampToShortString( recvtime )
			<< ", corrected: " << Measurement::timestampToShortString( correctedTime ) );

		m_outPort.send( EventType( correctedTime, mm ) );
	}


	void sendRequest()
	{
		if ( !m_pSenderEndpoint )
			return;

		// construct request packet
		// allow overriding of host and port parameters (by default, sender of UDP packet is taken)
		RequestPacket rp;
		rp.sHost = std::string();
		rp.nPort = 0;
		rp.sComponentName = getName();

		// serialize in string
		std::ostringstream stream;
		boost::archive::text_oarchive packet( stream );

		std::string suffix( "\n" );
		Measurement::Timestamp timeNow( Measurement::now() );

		// serialize the request packet, component name and current local time
		packet << m_senderName;
		packet << const_cast< const RequestPacket& >( rp );
		packet << timeNow;
		packet << suffix;

		m_pModule->send_to( stream.str(), *m_pSenderEndpoint );
	}

	// timestamp synchronization
	Measurement::TimestampSync m_synchronizer;
	Measurement::Timestamp m_firstTimestamp;

	// consumer port
	Dataflow::PushSupplier< EventType > m_outPort;

	// data of sender
	boost::shared_ptr< udp::endpoint > m_pSenderEndpoint;
	std::string m_senderName;

	// logging
	log4cpp::Category& m_logger;
};


} } // namespace Ubitrack::Drivers

