/*
 * Ubitrack - Library for Ubiquitous Tracking
 * Copyright 2006, Technische Universitaet Muenchen, and individual
 * contributors as indicated by the @authors tag. See the
 * copyright.txt in the distribution for a full listing of individual
 * contributors.
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA, or see the FSF site: http://www.fsf.org.
 */

/**
 * @ingroup driver_components
 * @file OffaxisProjection.cpp
 * Calculate an offaxis projection matrix based on eye and screen position.
 *
 * @author Nicolas Heuser (heuser@in.tum.de)
 */

#include <Ubitrack/Dataflow/TriggerComponent.h>
#include <Ubitrack/Dataflow/TriggerInPort.h>
#include <Ubitrack/Dataflow/TriggerOutPort.h>
#include <Ubitrack/Dataflow/ComponentFactory.h>
#include <Ubitrack/Math/Matrix.h>
#include <Ubitrack/Measurement/Measurement.h>
#include <Ubitrack/Calibration/Projection.h>

static log4cpp::Category& logger( log4cpp::Category::getInstance( "Ubitrack.Drivers.OffaxisProjection" ) );

using namespace Ubitrack::Dataflow;

namespace Ubitrack { namespace Drivers {

class OffaxisProjectionComponent
	: public Dataflow::TriggerComponent
{
	/** Convert string into vector. For reading the XML attributes. */
	Math::Vector<3> stringToVector(const std::string& s) {
		double p[3];
		std::istringstream vecString( s );
		for (int i=0; i < 3; ++i)
		{
			vecString >> p[i];
		}
		Math::Vector<3> trans( p );
		return trans;
	} 

public:

	OffaxisProjectionComponent( const std::string& sName, boost::shared_ptr< Graph::UTQLSubgraph > subgraph )
		: Dataflow::TriggerComponent( sName, subgraph )
		, m_inPortEyePose( "EyePose", *this )
		, m_inPortScreenPose( "ScreenPose", *this )
		, m_outPort( "ProjectionMatrix", *this )
		, m_near(0.01)
		, m_far(10.0)
	{
		//m_screen_pose = Ubitrack::Math::Pose(Ubitrack::Math::Quaternion(0, 0, 0, 1), Ubitrack::Math::Vector<3>(0,0,0));
		Graph::UTQLSubgraph::NodePtr objectNode = subgraph->getNode( "OffaxisProjection" );
		if ( !objectNode )
			UBITRACK_THROW( "No offaxis projection node in configuration" );

		if ( !objectNode->hasAttribute( "ScreenOffsetPosition" ) ||
			 !objectNode->hasAttribute( "ScreenOffsetRotation" ) )
		{
			UBITRACK_THROW( "Missing screen calibration values for offaxis projection." );
		}
		Math::Vector<3> trans = stringToVector( objectNode->getAttribute( "ScreenOffsetPosition" ).getText());

		std::string rotationChars = objectNode->getAttribute( "ScreenOffsetRotation" ).getText();
		double r[4];
		std::istringstream rotationString( rotationChars );
		for (int i=0; i < 4; ++i)
		{
			rotationString >> r[i];
		}
		Math::Quaternion q( r[0], r[1], r[2], r[3] );

		// Pose of the screen in world coordinates at calibration time
		Math::Pose screen_calibration_pose = Math::Pose( q.normalize(), trans );

		// Screen corners in world coordinates at calibration time
		Math::Vector<3> conf_ll = stringToVector( objectNode->getAttribute( "ScreenLowerLeft" ).getText());
		Math::Vector<3> conf_ul = stringToVector( objectNode->getAttribute( "ScreenUpperLeft" ).getText());
		Math::Vector<3> conf_lr = stringToVector( objectNode->getAttribute( "ScreenLowerRight" ).getText());

		// Calculate the offset from the screen pose to the screen corners
		m_screen_ll = ~screen_calibration_pose * conf_ll;
		m_screen_ul =    ~screen_calibration_pose * conf_ul;
		m_screen_lr = ~screen_calibration_pose * conf_lr;

		// Normalize the resulting vvectors. Screen width and height are a byproduct.
		Ubitrack::Math::Vector<3> Xs = m_screen_lr - m_screen_ll;
		m_screen_width = sqrt( Xs(0)*Xs(0) + Xs(1)*Xs(1) + Xs(2)*Xs(2) );	
		Ubitrack::Math::Vector<3> Ys = m_screen_ul - m_screen_ll;
		m_screen_height = sqrt( Ys(0)*Ys(0) + Ys(1)*Ys(1) + Ys(2)*Ys(2) );

		// There should be a better than have the clipping planes twice in the utql workflow.
		objectNode->getAttributeData( "NearClippingPlane", m_near );
		objectNode->getAttributeData( "FarClippingPlane", m_far );
	}

	/** Method that computes the result. */
	void compute( Measurement::Timestamp t ) {

		Math::Pose eye_pose = *(m_inPortEyePose.get());
		Math::Pose screen_pose = *(m_inPortScreenPose.get());

		LOG4CPP_TRACE(logger, "Computing projection matrix for view at " << eye_pose << " to screen at " << screen_pose);

		Math::Vector<3> ll = screen_pose * m_screen_ll;
		Math::Vector<3> ul = screen_pose * m_screen_ul;
		Math::Vector<3> lr = screen_pose * m_screen_lr;

		Math::Vector<3> eye = eye_pose.translation();

		Math::Matrix<4, 4> projMatrix = Calibration::offAxisProjectionMatrix( eye, ll, ul, lr, m_near, m_far, m_screen_width, m_screen_height );
		m_outPort.send( Measurement::Matrix4x4(t, projMatrix) );
	}

protected:
	/** Eye pose port */
	Dataflow::TriggerInPort< Measurement::Pose >		m_inPortEyePose;

	/** Screen pose port */
	Dataflow::TriggerInPort< Measurement::Pose > 		m_inPortScreenPose;

	/** Output port of the component. */
	Dataflow::TriggerOutPort< Measurement::Matrix4x4 >	m_outPort;

	/** Pose of the screen during calibration */
	Math::Pose						m_screen_calibration_pose;

	/** Offset of 3 screen corners to the screen pose */
	Math::Vector<3>						m_screen_ll;
	Math::Vector<3>						m_screen_lr;
	Math::Vector<3>						m_screen_ul;

	/** Width and height of the screen */
	double							m_screen_width;
	double							m_screen_height;

	/** Near and far clipping plane dinstances */
	double							m_near;
	double							m_far;
};

} } // namespace Ubitrack::Drivers

UBITRACK_REGISTER_COMPONENT( Ubitrack::Dataflow::ComponentFactory* const cf )
{
	cf->registerComponent< Ubitrack::Drivers::OffaxisProjectionComponent >( "OffaxisProjection" );
}
