/*
 * Ubitrack - Library for Ubiquitous Tracking
 * Copyright 2006, Technische Universitaet Muenchen, and individual
 * contributors as indicated by the @authors tag. See the
 * copyright.txt in the distribution for a full listing of individual
 * contributors.
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA, or see the FSF site: http://www.fsf.org.
 */


/**
 * @ingroup driver_components
 * @file
 * OptiTrack driver
 * This file contains the driver component to
 * get tracking data from an OptiTracker.
 *
 * The received data is sent via a push interface.
 *
 * @author Daniel Muhra <muhra@in.tum.de>
 */

#include <Ubitrack/Util/CleanWindows.h>
#ifndef _WIN32_WINNT
	#define _WIN32_WINNT WINVER
#endif

#include <objbase.h>
#include <atlbase.h>

#include <string>
#include <deque>
#include <cstdlib>
#include <boost/thread.hpp>
#include <boost/thread/xtime.hpp>
#include <boost/function.hpp>
#include <boost/bind.hpp>
#include <boost/scoped_ptr.hpp>
#include <log4cpp/Category.hh>

#include <Ubitrack/Dataflow/PushSupplier.h>
#include <Ubitrack/Dataflow/ComponentFactory.h>
#include <Ubitrack/Dataflow/Module.h>
#include <Ubitrack/Dataflow/Component.h>
#include <Ubitrack/Measurement/Measurement.h>
#include <Ubitrack/Util/Exception.h>
#include <Ubitrack/Util/OS.h>

#include "optitrack.h"
#import  "optitrack.tlb"

static log4cpp::Category& logger( log4cpp::Category::getInstance( "Drivers.OptiTrack" ) );

using namespace Ubitrack;
using namespace Ubitrack::Dataflow;

namespace Ubitrack { namespace Drivers {

// forward declarations
class OptiTrack;
class OptiTrackModule;

typedef Module< SingleModuleKey, SubgraphIdKey, OptiTrackModule, OptiTrack > ModuleBase;

class OptiTrackModule
	: public ModuleBase
{
public:
	/** constructor, creates thread */
	OptiTrackModule( const SingleModuleKey& key, boost::shared_ptr< Graph::UTQLSubgraph >, FactoryHelper* pFactory )
		: ModuleBase( key, pFactory )
		, m_bStop( false )
		, m_nCams( 0 )
	{
		comps = getAllComponents();
		m_Thread.reset( new boost::thread( boost::bind( &OptiTrackModule::threadProc, this ) ) );
	}

	/** destructor, stops thread */
	~OptiTrackModule()
	{
		m_bStop = true;
		m_Thread->join();
		m_cameraCollection.Release();
		CoUninitialize();
	}

	/** creates a cam */
	void createCam( OptiTrack * component )
	{
		boost::mutex::scoped_lock l( m_queueMutex );
		m_queue.push_back( boost::bind( &OptiTrackModule::myCreateCam, this, component ) );
		m_queueCondition.notify_all();
	}

	/** destroys a cam */
	void destroyCam( OptiTrack * component )
	{
		boost::mutex::scoped_lock l( m_queueMutex );
		m_queue.push_back( boost::bind( &OptiTrackModule::myDestroyCam, this, component ) );
		m_queueCondition.notify_all();
	}

protected:
	// thread main loop
	void threadProc();

	void myCreateCam( OptiTrack * component );	

	void myDestroyCam( OptiTrack * component );
	

	// the thread
	boost::scoped_ptr< boost::thread > m_Thread;

	// stop the thread?
	volatile bool m_bStop;

	// number of cams open
	int m_nCams;

	//number of available cams
	long m_camCount;

	// a mutex object to synchronize cam calls
	boost::mutex m_queueMutex;

	// notification when the queue has changed
	boost::condition m_queueCondition;

	//list of Componenst
	CComPtr<INPCameraCollection> m_cameraCollection;
	ComponentList comps;

	typedef std::deque< boost::function< void() > > QueueType;
	QueueType m_queue;
};

/**
 * @ingroup driver_components
 * A Component that returns a list of 2D points found by the OptiTrack camera.
 *
 * @par Output Ports
 * PushSupplier< Ubitrack::Measurement::PositionList2 > port with name "Output"
 *
 * @par Configuration
 * \c serial: serial number of the OptiTrack camera
 * \c fps: fps, should 120 or lower
 * \c threshold: the value must be between 1 and 253, default value is 130
 *
 * @par Example Configuration
 * \verbatim 
<Pattern name="OptiTrack" id="OptiTrack1">
	<Output>
		<Node name="Camera" id="Camera"/>
		<Node name="ImagePlane" id="ImagePlane"/>
		<Edge name="Output" source="Camera" destination="ImagePlane"/>
	</Output>
	
	<DataflowConfiguration>
		<UbitrackLib class="OptiTrack"/>
		<Attribute name="serial" value="35185"/>
		<Attribute name="fps" value="60"/>
		<Attribute name="threshold" value="130"/>
	</DataflowConfiguration>
</Pattern>
\endverbatim
 */

class OptiTrack
	: public OptiTrackModule::Component
{
public:
	/** constructor, creates camera */
	OptiTrack( const std::string& sName, boost::shared_ptr< Graph::UTQLSubgraph > pCfg, const SubgraphIdKey& componentKey, OptiTrackModule* pModule )
		: OptiTrackModule::Component( sName, componentKey, pModule )
		, m_outPort( "Output", *this )
		, m_serial ( 0 )
		, m_count ( 0 )
		, m_thres ( 130 )
		, m_fps ( 120 )
	{ 
		pCfg->m_DataflowAttributes.getAttributeData( "serial", m_serial );
		pCfg->m_DataflowAttributes.getAttributeData( "fps", m_fps );
		pCfg->m_DataflowAttributes.getAttributeData( "threshold", m_thres );
		getModule().createCam( this );
	}

	/** destructor, destroys window */
	~OptiTrack()
	{ getModule().destroyCam( this ); }

	bool init( CComPtr<INPCameraCollection> & cameraCollection )
	{
		bool found = false;
		LONG s, camCount;
		VARIANT thres;
		int i=0;

		cameraCollection->get_Count( &camCount );

		while( i<camCount && !found )
		{
			cameraCollection->Item( i, &m_camera );
			m_camera->get_SerialNumber( &s );
			i++;

			if( s == m_serial )
				found = true;
			else m_camera.Release();
		}

		if( found )
		{
			setFrameRate();
			m_camera->SetLED( NP_LED_ONE, VARIANT_TRUE );
			
			thres.ulVal = m_thres;
			m_camera->SetOption( NP_OPTION_THRESHOLD, thres );
			
			m_camera->Open();
			m_camera->Start();
			return true;
		}
		else
		{
			UBITRACK_THROW( "no camera found with current serial for construction" );
			return false;
		}
		
	}

	bool exit( CComPtr<INPCameraCollection> & cameraCollection )
	{
		bool found = false;
		LONG s, camCount;
		int i=0;

		cameraCollection->get_Count( &camCount );

		while( i<camCount && !found )
		{
			cameraCollection->Item( i, &m_camera );
			m_camera->get_SerialNumber( &s );
			i++;

			if( s == m_serial )
				found = true;
		}

		if( found )
		{
			m_camera->Stop();
			m_camera->Close();
			m_camera->SetLED( NP_LED_ONE, VARIANT_FALSE );
			m_camera.Release();
			return true;
		}
		else
		{
			UBITRACK_THROW( "no camera found with current serial for destruction" );
			return false;
		}
	}
	
	/** method that sends coordinates of 2D Points found in the image */
	void poll()
	{ 
		HRESULT hrFrame, hrObject;
		CComPtr<INPCameraFrame> frame;

		//get new frame
		hrFrame = m_camera->GetFrame(0, &frame);
		Ubitrack::Measurement::Timestamp time = Ubitrack::Measurement::now();

		//how many objects were found
		LONG c = 0;
		if( hrFrame == S_OK && frame )
			hrFrame = frame->get_Count( &c );
		
		if( c > 0 && hrFrame == S_OK && frame && m_count == 0 )
		{
			Math::Vector< 2 > temp;
			std::vector< Math::Vector< 2 > > list; 
			VARIANT v;

			for( int i=0; i<c; i++ )
			{
				CComPtr<INPObject> object;
				hrObject = frame->Item( i, &object );
		
				if( hrObject == S_OK && object )
				{
					hrObject = object->get_X( &v );
					temp( 0 ) = v.dblVal;
				}
				
				if( hrObject == S_OK && object )
				{
					hrObject = object->get_Y( &v );
					temp( 1 ) = 290.0 - v.dblVal;
				}
				
				if( hrObject == S_OK && object )
					list.push_back( temp );
		
				object.Release();
			}

			m_outPort.send( Measurement::PositionList2( time, list ) ); 
		}
		
		
		if ( hrFrame == S_OK && frame )
		{
			frame->Free();			
			if( m_count < m_skip )
				m_count++;
			else m_count = 0;
		}
		
		frame.Release();
	}

	void setFrameRate()
	{
		LONG r;
		m_camera->get_FrameRate( &r);

		if ( m_fps >= r )
			m_skip = 0;
		else
			m_skip = (r / m_fps) - 1;
	}

protected:
	//options
	CComPtr<INPCamera> m_camera;
	int m_serial;
	int m_skip;
	int m_fps;
	int m_thres;
	int m_count;

	// the port
	Dataflow::PushSupplier< Measurement::PositionList2 > m_outPort;
};

void OptiTrackModule::threadProc()
{
	HRESULT hRes;

	//intitializing microsoft COM
	hRes = CoInitializeEx(NULL, COINIT_APARTMENTTHREADED);
	if ( FAILED( hRes ) )
		UBITRACK_THROW( "CoInitialize() failed" );

	hRes = m_cameraCollection.CoCreateInstance(CLSID_NPCameraCollection);
	

	/** enumerate (identify) cameras */
	hRes = m_cameraCollection->Enum();
	if ( FAILED( hRes ) )
		UBITRACK_THROW( "Could not identify cameras" );
	
	/** get the number of available cameras */
	hRes = m_cameraCollection->get_Count( &m_camCount );
	if ( FAILED( hRes ) )
				UBITRACK_THROW( "Could not get the number of cameras" );

	while ( !m_bStop || m_nCams > 0 )
	{
		// get next event from queue
		boost::function< void() > nextCall;
		{
			boost::mutex::scoped_lock l( m_queueMutex );
			if ( m_queue.empty() && m_nCams )
			{	
				// let the message dispatching do something
				MSG msg;
				if( PeekMessage( &msg, NULL, 0, 0, PM_NOREMOVE ) )
				{
					if(GetMessage( &msg, NULL, 0, 0 ) )
					{
						TranslateMessage(&msg);
						DispatchMessage(&msg);
					}
				}
				//no new events, now handle cameras
				Sleep(5);
				
				for ( ComponentList::iterator it = comps.begin(); it != comps.end(); it++ )
				(*it)->poll();
			}

			if ( !m_queue.empty() )
			{
				nextCall = m_queue.front();
				m_queue.pop_front();
			}
		}

		// execute the queued event
		if ( nextCall )
			nextCall();		
		
		// what components do we now have
		comps = getAllComponents();
	}
}

void OptiTrackModule::myCreateCam( OptiTrack * component )
{
	bool b;
	b = component->init( m_cameraCollection );
	if( b )
		m_nCams++;
}

void OptiTrackModule::myDestroyCam( OptiTrack * component )
{
	bool b;
	b = component->exit( m_cameraCollection );
	if( b )
		m_nCams--;
}
	
} } // namespace Ubitrack::Driver

UBITRACK_REGISTER_COMPONENT( Dataflow::ComponentFactory* const cf ) {
	cf->registerModule< Ubitrack::Drivers::OptiTrackModule > ( "OptiTrack" );
}
