/*
 * Ubitrack - Library for Ubiquitous Tracking
 * Copyright 2006, Technische Universitaet Muenchen, and individual
 * contributors as indicated by the @authors tag. See the
 * copyright.txt in the distribution for a full listing of individual
 * contributors.
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA, or see the FSF site: http://www.fsf.org.
 */

// TODO: implement module start/stop mechanics

/**
 * @ingroup vision_components
 * @file
 * Implements a component for the people tracker.
 *
 * @author Daniel Pustka <daniel.pustka@in.tum.de>
 */

#include <string>

#include <boost/scoped_ptr.hpp>
#include <log4cpp/Category.hh>

#include <Ubitrack/Dataflow/Component.h>
#include <Ubitrack/Dataflow/PushConsumer.h>
#include <Ubitrack/Dataflow/PullConsumer.h>
#include <Ubitrack/Dataflow/PushSupplier.h>
#include <Ubitrack/Dataflow/ComponentFactory.h>
#include <Vision/Image.h>

#include "PeopleTracker.h"

static log4cpp::Category& logger( log4cpp::Category::getInstance( "Ubitrack.Drivers.PeopleTracking" ) );

namespace ublas = boost::numeric::ublas;

namespace Ubitrack { namespace Drivers {

class PeopleTrackingComponent
	: public Dataflow::Component
{
public:
	PeopleTrackingComponent( const std::string& sName, boost::shared_ptr< Graph::UTQLSubgraph > pConfig )
		: Dataflow::Component( sName )
		, m_imageIn( "Image", *this, boost::bind( &PeopleTrackingComponent::pushImage, this, _1 ) )
		, m_debugImageOut( "DebugImage", *this )
		, m_intrinsicsIn( "Intrinsics", *this )
		, m_floorIn( "Floor", *this )
	{
	}

	void pushImage( const Vision::ImageMeasurement& m )
	{
		// create 3x4 projection matrix
		Measurement::Matrix3x3 pIntrinsics( m_intrinsicsIn.get( m.time() ) );
		Math::Matrix< 3, 4 > intrinsics( ublas::zero_matrix< double >( 3, 4 ) );
		ublas::subrange( intrinsics, 0, 3, 0, 3 ) = *pIntrinsics;

		if ( !m->origin )
		{
			intrinsics( 1, 1 ) *= -1.0f;
			intrinsics( 1, 2 ) = -intrinsics( 1, 2 ) + intrinsics( 2, 2 ) * static_cast< float >( m->height - 1 );
		}
		
		Measurement::Pose pFloor( m_floorIn.get( m.time() ) );
		Math::Matrix< 4, 4 > floorMatrix( *pFloor );
		
		Math::Matrix< 3, 4 > P( ublas::prod( intrinsics, floorMatrix ) );
		m_tracker.setProjection( P );
	
		// create a debug image
		boost::shared_ptr< Vision::Image > pDebugImg( new Vision::Image( m->width, m->height, 3 ) );
		pDebugImg->origin = m->origin;

		// run the people tracker
		m_tracker.addImage( m, pDebugImg.get() );
		
		// send result
		m_debugImageOut.send( Vision::ImageMeasurement( m.time(), pDebugImg ) );
	}

protected:
	Dataflow::PushConsumer< Vision::ImageMeasurement > m_imageIn;
	Dataflow::PushSupplier< Vision::ImageMeasurement > m_debugImageOut;
	Dataflow::PullConsumer< Measurement::Matrix3x3 > m_intrinsicsIn;
	Dataflow::PullConsumer< Measurement::Pose > m_floorIn;
	
	PeopleTracking::PeopleTracker m_tracker;
};

} } // namespace Ubitrack::Driver


UBITRACK_REGISTER_COMPONENT( Ubitrack::Dataflow::ComponentFactory* const cf ) 
{
	cf->registerComponent< Ubitrack::Drivers::PeopleTrackingComponent > ( "PeopleTracking" );
}
