/*
 * Ubitrack - Library for Ubiquitous Tracking
 * Copyright 2006, Technische Universitaet Muenchen, and individual
 * contributors as indicated by the @authors tag. See the 
 * copyright.txt in the distribution for a full listing of individual
 * contributors.
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA, or see the FSF site: http://www.fsf.org.
 */

/**
 * @ingroup drivers
 * @file
 * Declaration of a general conic class.
 *
 * @author Daniel Pustka <daniel.pustka@in.tum.de>
 */


#ifndef __Ubitrack_Drivers_PeopleTracker_Conic_h_included__
#define __Ubitrack_Drivers_PeopleTracker_Conic_h_included__

#include <math.h>
#include <algorithm>
#include <Ubitrack/Math/Matrix.h>

namespace Ubitrack { namespace Drivers { namespace PeopleTracking {

/** a conic represented by its 3x3 matrix */
class Conic
{
public:
	/** dummy default constructor */
	Conic()
	{}

	/** construct by projecting an ellipsoid onto the screen */
	Conic( const Math::Vector< 3 >& pos, const Math::Vector< 3 >& size, const Math::Matrix< 3, 4 >& P );

	/** return 3x3 matrix */
	const Math::Matrix< 3,3 >& getMatrix() const
	{ return m_matrix; }

	/** compute top and bottom boundaries */
	std::pair< double, double > getTopBottom() const
	{
 		// compute begin and end in y direction
		double a = 4 * ( m_matrix( 0, 1 ) * m_matrix( 0, 1 ) - m_matrix( 0, 0 ) * m_matrix( 1, 1 ) );
		double b = 8 * ( m_matrix( 0, 1 ) * m_matrix( 0, 2 ) - m_matrix( 0, 0 ) * m_matrix( 1, 2 ) );
		double c = 4 * ( m_matrix( 0, 2 ) * m_matrix( 0, 2 ) - m_matrix( 0, 0 ) * m_matrix( 2, 2 ) );
		double d = sqrt( b * b - ( 4 * a * c ) );
		a *= 2;
		double y1 = -( (b/a) + (d/a) );
		double y2 = -( (b/a) - (d/a) );

		if ( y1 < y2 )
			return std::pair< double, double >( y1, y2 );
		else
			return std::pair< double, double >( y2, y1 );
	}

	/** compute intersections with a horizontal line */
	std::pair< double, double > getHLineIntersections( double y ) const
	{
	    double b = m_matrix( 0, 1 ) * y + m_matrix( 0, 2 ) ;
	    double c = ( m_matrix( 1, 1 ) * y + 2 * m_matrix( 1, 2 ) ) * y + m_matrix( 2, 2 );
	    double d = sqrt( b * b - m_matrix( 0, 0 ) * c );
	    int x1 = (int)( (-b + d) / m_matrix( 0, 0 ) );
	    int x2 = (int)( (-b - d) / m_matrix( 0, 0 ) );
		if ( x1 < x2 )
			return std::pair< double, double >( x1, x2 );
		else
			return std::pair< double, double >( x2, x1 );
	}

	/** is a point inside the conic? */
	bool pointInside( const Math::Vector< 2 >& p ) const
	{
		double t = p( 0 ) * ( p( 0 ) * m_matrix( 0, 0 ) + p( 1 ) * m_matrix( 0, 1 ) + m_matrix( 0, 2 ) );
		t       += p( 1 ) * ( p( 0 ) * m_matrix( 1, 0 ) + p( 1 ) * m_matrix( 1, 1 ) + m_matrix( 1, 2 ) );
		t       +=            p( 0 ) * m_matrix( 2, 0 ) + p( 1 ) * m_matrix( 2, 1 ) + m_matrix( 2, 2 );
		return t < 0;
	}

protected:
	/** the matrix representation of the conic */
	Math::Matrix< 3, 3 > m_matrix;
};

} } } // namepsace Ubitrack::Drivers::PeopleTracking

#endif