/*
 * Ubitrack - Library for Ubiquitous Tracking
 * Copyright 2006, Technische Universitaet Muenchen, and individual
 * contributors as indicated by the @authors tag. See the 
 * copyright.txt in the distribution for a full listing of individual
 * contributors.
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA, or see the FSF site: http://www.fsf.org.
 */

/**
 * @ingroup drivers
 * @file
 * Implementation of the main people tracker class.
 *
 * @author Daniel Pustka <daniel.pustka@in.tum.de>
 */

#include <iostream>
#include <boost/scoped_array.hpp>
#include <boost/scoped_ptr.hpp>
#include <cv.h>
#include <log4cpp/Category.hh>

#include <Ubitrack/Math/MatrixOperations.h>

#include "PeopleTracker.h"
#include "VirtualScene.h"
#include "PersonSystemPdf.h"
#include "PersonMeasurementPdf.h"

#include <Ubitrack/Util/BlockTimer.h>

// logger for all parts of the people tracking
static log4cpp::Category& logger( log4cpp::Category::getInstance( "Ubitrack.Drivers.PeopleTracking" ) );

namespace Ubitrack { namespace Drivers { namespace PeopleTracking {

#if 1
	static Util::BlockTimer g_totalTimer( "total", logger );
	static Util::BlockTimer g_imageGenerationTimer( "image generation", logger );
	static Util::BlockTimer g_sceneGenerationTimer( "scene generation", logger );
	static Util::BlockTimer g_debugImageGenerationTimer( "debug image generation", logger );
	static Util::BlockTimer g_newObjectDetectionTimer( "new object detection", logger );
	#define DO_TIME( name, x ) UBITRACK_TIME( x )
#else
	#define DO_TIME( name, x )
#endif

PeopleTracker::PeopleTracker( )
	: m_minIntensity( 20 )
	, m_maxIntensity( 60 )
	, m_lastTime( 0 )
	, m_lastId( 1 )
	, m_imageCounter( 0 )
{
	// initialization: add one new object at the origin
//	m_states.push_back( PersonState( m_lastId++, Math::Vector< 2 >( 0, 0 ) ) );
//	m_states.push_back( PersonState( m_lastId++, Math::Vector< 2 >( -1.5, 0 ) ) );
}

void PeopleTracker::addImage( const Vision::ImageMeasurement& image, Vision::Image* pDebugImg )
{
	const int nWaitImages = 50;
	m_imageCounter++;
	if ( m_imageCounter < nWaitImages )
		return;
	
	// store background image if necessary
	if ( !m_pBackgroundImage )
	{
		LOG4CPP_INFO( logger, "Storing background image" );
		m_pBackgroundImage = image;
		return;
	}
		
	if ( m_imageCounter < nWaitImages + 5 )
		return;

	//DO_TIME( _totalTimer_, g_totalTimer );
	UBITRACK_TIME( g_totalTimer );
		
	// create sum image
	boost::scoped_array< int > pSumImage( new int[ image->width * image->height + 1 ] );
	Vision::Image subtractedImage( image->width, image->height, 1 );
	makeImages( *image, *m_pBackgroundImage, pSumImage.get(), subtractedImage, pDebugImg );
	
	for ( std::list< PersonState >::iterator it = m_states.begin(); it != m_states.end(); )
	{
		// advance the state of all persons
		if ( m_lastTime )
			it->advance( double( image.time() - m_lastTime ) * 1e-9 );
			
		// check if person is still valid and remove if not
		if ( !it->isValid() )
		{
			it = m_states.erase( it );
			//m_states.push_back( PersonState( m_lastId++, Math::Vector< 2 >( 0, 0 ) ) );
		}
		else
			it++;
	}
	
	m_lastTime = image.time();
	
	// search for new objects (only every n-th frame)
	if ( m_imageCounter % 10 == 0 )
		addNewObjects( subtractedImage );
	//if ( pDebugImg ) cvCvtColor( subtractedImage, *pDebugImg, CV_GRAY2RGB );


	// create virtual scene
	VirtualScene scene( image->width, image->height, m_states.size() );
	{
	DO_TIME( _timer_, g_sceneGenerationTimer );
	for ( std::list< PersonState >::iterator it = m_states.begin(); it != m_states.end(); it++ )
		scene.addObject( it->getID(), OnScreenConic( it->getOcclusionConic( m_P ), image->width, image->height ) );
		
	scene.normalize();
	}

	// evaluate all particles
	for ( std::list< PersonState >::iterator it = m_states.begin(); it != m_states.end(); it++ )
		it->measurement( scene, pSumImage.get(), m_states, m_P );
		
	if ( pDebugImg )
	{
		DO_TIME( _dbgImgTimer_, g_debugImageGenerationTimer );
		for ( std::list< PersonState >::iterator it = m_states.begin(); it != m_states.end(); it++ )
			it->drawDebugImage( *pDebugImg, scene, pSumImage.get(), m_P );
	}
}


void PeopleTracker::makeImages( const Vision::Image& image, const Vision::Image& backgroundImage, 
	int* pSumImage, Vision::Image& subtractedImage, Vision::Image* pDebugImg )
{
	{
	DO_TIME( _timer_, g_imageGenerationTimer );
	
	unsigned char* pLineStart = reinterpret_cast< unsigned char* >( image.imageData );
	unsigned char* pBgLineStart = reinterpret_cast< unsigned char* >( backgroundImage.imageData );
	unsigned char* pSubtractedLineStart = reinterpret_cast< unsigned char* >( subtractedImage.imageData );
	int sum = 0;
	// walk over all lines
	for ( unsigned i = image.height; i; i-- )
	{
		unsigned char* p = pLineStart;
		unsigned char* pBg = pBgLineStart;
		unsigned char* pSubtracted = pSubtractedLineStart;

		if ( image.nChannels == 1 )
		{
			// greyscale image

			// walk over all pixels in line
			for ( unsigned x = image.width; x; x-- )
			{
				// compute absolute difference and scale selected intensity region to 0-255
				int v = *p++ - *pBg++;
				if ( v < 0 )
					v = -v;
				
				if ( v >= m_maxIntensity )
				{
					v = 255;
					*pSubtracted++ = 255;
				}
				else
				{
					if ( v <= m_minIntensity )
						v = 0;
					else
						v = ( v - m_minIntensity ) * 255 / ( m_maxIntensity - m_minIntensity );
					*pSubtracted++ = 0;
				}

				// compute sum
				*pSumImage++ = sum;
				sum += v;
			}
		}
		else
		{
			// color image

			// walk over all pixels in line
			for ( unsigned x = image.width; x; x-- )
			{
				// compute absolute difference and scale selected intensity region to 0-255
				int v, vc;
				vc = *p++ - *pBg++;
				v  = ( vc < 0 ? -vc : vc ) * 77;
				vc = *p++ - *pBg++;
				v += ( vc < 0 ? -vc : vc ) * 150;
				vc = *p++ - *pBg++;
				v += ( vc < 0 ? -vc : vc ) * 29;
				v >>= 8;
				
				if ( v >= m_maxIntensity )
				{
					v = 255;
					*pSubtracted++ = 255;
				}
				else
				{
					if ( v <= m_minIntensity )
						v = 0;
					else
						v = ( v - m_minIntensity ) * 255 / ( m_maxIntensity - m_minIntensity );
					*pSubtracted++ = 0;
				}

				// compute sum
				*pSumImage++ = sum;
				sum += v;
			}
		}
		
		pLineStart += image.widthStep;
		pBgLineStart += backgroundImage.widthStep;
		pSubtractedLineStart += subtractedImage.widthStep;
	}
	
	// set value after the last pixel of the image
	*pSumImage = sum;
	}
	
	// create debug image
	if ( pDebugImg )
	{
		DO_TIME( _dbgImgTimer_, g_debugImageGenerationTimer );

		boost::scoped_ptr< Vision::Image > pDbgSubtracted( new Vision::Image( image.width, image.height, image.nChannels ) );
		cvAbsDiff( image, backgroundImage, *pDbgSubtracted );

		if ( image.nChannels == 3 )
		{
			// convert to greyscale
			boost::scoped_ptr< Vision::Image > pDbgSubtractedGrey( new Vision::Image( image.width, image.height, 1 ) );
			cvCvtColor( *pDbgSubtracted, *pDbgSubtractedGrey, CV_RGB2GRAY );
			pDbgSubtracted.swap( pDbgSubtractedGrey );

		}

		double scale = 255.0 / ( m_maxIntensity - m_minIntensity );
		if ( pDebugImg->nChannels == 3 )
		{
			cvConvertScale( *pDbgSubtracted, *pDbgSubtracted, scale, -m_minIntensity * scale );
			cvCvtColor( *pDbgSubtracted, *pDebugImg, CV_GRAY2RGB );
		}
		else
			cvConvertScale( *pDbgSubtracted, *pDebugImg, scale, -m_minIntensity * scale );
	}
}


void PeopleTracker::setProjection( const Math::Matrix< 3, 4 >& P )
{ 
	// store projection matrix
	m_P = P; 

	// compute projection matrix relative to plane at half-height of persons
	Math::Matrix< 3, 4 > PPersonToCam( P );
	ublas::column( PPersonToCam, 3 ) += ublas::column( PPersonToCam, 2 ) * g_personZ;

	// compute homography mapping half-height plane to screen
	Math::Matrix< 3, 3 > HPersonToCam;
	ublas::subrange( HPersonToCam, 0, 3, 0, 2 ) = ublas::subrange( PPersonToCam, 0, 3, 0, 2 );
	ublas::column( HPersonToCam, 2 ) = ublas::column( PPersonToCam, 3 );

	// compute homography from image plane to half-height plane
	m_HImagePlaneToPerson = Math::invert_matrix( HPersonToCam );
}


void PeopleTracker::addNewObjects( Vision::Image& image )
{
	DO_TIME( _timer_, g_newObjectDetectionTimer );

	// first, do some morphology to eliminate noise
	IplConvKernel* pKernel = cvCreateStructuringElementEx( 10, 10, 5, 5, CV_SHAPE_ELLIPSE );
	cvMorphologyEx( image, image, 0, pKernel, CV_MOP_CLOSE );
	cvReleaseStructuringElement( &pKernel );

	// create memory storage that will contain all the dynamic data
	CvMemStorage* pStorage = cvCreateMemStorage( 0 );

	// Create dynamic structure and sequence.
	CvSeq* pContours = cvCreateSeq( CV_SEQ_ELTYPE_POINT, sizeof( CvSeq ), sizeof( CvPoint ), pStorage );

	// Find all contours.
	cvFindContours( image, pStorage, &pContours, sizeof( CvContour ), 
		CV_RETR_LIST, CV_CHAIN_APPROX_SIMPLE, cvPoint( 0, 0 ) );

	// approximate all contours by polygons and check size and corner number
	for ( CvSeq* pContour = pContours; pContour; pContour = pContour->h_next ) 
	{
		CvRect boundingRect = cvBoundingRect( pContour );
		float fDiameter = sqrtf( float( boundingRect.height * boundingRect.height + boundingRect.width * boundingRect.width ) );
		Math::Vector< 2 > center( boundingRect.x + boundingRect.width / 2, boundingRect.y + boundingRect.height / 2 );

		// discard too small contours
		if ( fDiameter < 100 )
			continue;

		// TODO: check if object is inside one of the known visitor areas
		// if not, compute world-position of center of object (set z=0.5*h), project conic and determine its dimensions.
		// if the dimensions roughly agree with that of the detected object, accept as new person.

		// check if this is an already existing object, by checking the center and some outline points
		std::vector< Math::Vector< 2 > > objectPoints( 5 );
		objectPoints[ 0 ] = Math::Vector< 2 >( boundingRect.x + boundingRect.width / 2, boundingRect.y + boundingRect.height / 2 );
		objectPoints[ 1 ] = Math::Vector< 2 >( boundingRect.x                         , boundingRect.y + boundingRect.height / 2 );
		objectPoints[ 2 ] = Math::Vector< 2 >( boundingRect.x + boundingRect.width    , boundingRect.y + boundingRect.height / 2 );
		objectPoints[ 3 ] = Math::Vector< 2 >( boundingRect.x + boundingRect.width / 2, boundingRect.y                           );
		objectPoints[ 4 ] = Math::Vector< 2 >( boundingRect.x + boundingRect.width / 2, boundingRect.y + boundingRect.height     );

		if ( pointsInKnownObject( objectPoints ) )
			continue;

		// compute world coordinates of person
		Math::Vector< 3 > posH( ublas::prod( ublas::subrange( m_HImagePlaneToPerson, 0, 3, 0, 2 ), center ) );
		posH += ublas::column( m_HImagePlaneToPerson, 2 );

		// create a new object
		m_states.push_back( PersonState( m_lastId++, ublas::subrange( posH, 0, 2 ) / posH( 2 ) ) );
	}

	// cleanup	     
	cvReleaseMemStorage( &pStorage );
}


bool PeopleTracker::pointsInKnownObject( const std::vector< Math::Vector< 2 > >& objectPoints )
{
	for ( std::list< PersonState >::iterator it = m_states.begin(); it != m_states.end(); it++ )
	{
		const Conic& C( it->getOcclusionConic( m_P ) );

		for ( std::vector< Math::Vector< 2 > >::const_iterator itP = objectPoints.begin(); itP != objectPoints.end(); itP++ )
			if ( C.pointInside( *itP ) )
				return true;
	}

	return false;
}
	
} } } // namespace Ubitrack::Drivers::PeopleTracking
