/*
 * Ubitrack - Library for Ubiquitous Tracking
 * Copyright 2006, Technische Universitaet Muenchen, and individual
 * contributors as indicated by the @authors tag. See the 
 * copyright.txt in the distribution for a full listing of individual
 * contributors.
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA, or see the FSF site: http://www.fsf.org.
 */

/**
 * @ingroup drivers
 * @file
 * Definition of the main people tracker class.
 *
 * @author Daniel Pustka <daniel.pustka@in.tum.de>
 */

#ifndef __Ubitrack_Drivers_PeopleTracker_PeopleTracker_h_included__
#define __Ubitrack_Drivers_PeopleTracker_PeopleTracker_h_included__

#include <list>

#include <Ubitrack/Math/Matrix.h>
#include <Vision/Image.h>

#include "PersonState.h"

namespace Ubitrack { namespace Drivers { namespace PeopleTracking {

/**
 * The main people tracker class that encapsulates the behaviour of the whole thing.
 */
class PeopleTracker
{
public:
	/** constructor */
	PeopleTracker();
	
	/** 
	 * Update the people tracker by an image.
	 * @param image the image measurement
	 * @param pDebugImg draw a nice picture into this image if pDebugImg != 0. The image must be color and 
	 *    have the same dimensions as the measurement image.
	 */
	void addImage( const Vision::ImageMeasurement& image, Vision::Image* pDebugImg );

	/** set the 3x4 projection matrix that describes both internal and external camera parameters */
	void setProjection( const Math::Matrix< 3, 4 >& P );

	// returns the current list of persons detected by the tracker */
	const std::list< PersonState >& getPersons() const
	{ return m_states; }
	
protected:
	/** maximum intensity in the difference image */
	int m_maxIntensity;

	/** minimum intensity in the difference image */
	int m_minIntensity;

	/** time of the last image measurement */
	Measurement::Timestamp m_lastTime;

	/** id of the last detected person */
	unsigned m_lastId;

	/** the current background image */
	boost::shared_ptr< Vision::Image > m_pBackgroundImage;

	/** currently set projection matrix */
	Math::Matrix< 3, 4 > m_P;

	/** homography mapping image points to person positions */
	Math::Matrix< 3, 3 > m_HImagePlaneToPerson;

	/** list of detected persons */
	std::list< PersonState > m_states;

	/** number of images added so far */
	unsigned m_imageCounter;
	
	/** 
	 * Creates the images required for the tracker.
	 * @param image the image measurement (input)
	 * @param backgroundImage background image (input)
	 * @param pSumImage summed up pixel image (output). The array must at least have size (image.width * image.height + 1).
	 * @param subtractedImage background subtracted image for blob detection of new persons
	 * @param pDebugImg if non-null, the background subtracted image will be copied here.
	 */
	void makeImages( const Vision::Image& image, const Vision::Image& backgroundImage, 
		int* pSumImage, Vision::Image& subtractedImage, Vision::Image* pDebugImg );
	
	/** 
	 * Scan the image for new objects and add them to the list.
	 * @param image the thresholded background-subtracted image (may be modified by the routine)
	 */
	void addNewObjects( Vision::Image& image );

	/** returns true if one of the points given lies within an already known object */
	bool pointsInKnownObject( const std::vector< Math::Vector< 2 > >& objectPoints );
};

} } } // namespace Ubitrack::Drivers::PeopleTracking

#endif
