/*
 * Ubitrack - Library for Ubiquitous Tracking
 * Copyright 2006, Technische Universitaet Muenchen, and individual
 * contributors as indicated by the @authors tag. See the 
 * copyright.txt in the distribution for a full listing of individual
 * contributors.
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA, or see the FSF site: http://www.fsf.org.
 */

/**
 * @ingroup drivers
 * @file
 * Implementation of a measurement function for the orocos BFL library.
 *
 * @author Daniel Pustka <daniel.pustka@in.tum.de>
 */

#include <iostream>

#include <Ubitrack/Util/BlockTimer.h>

#include "Conic.h"
#include "PersonMeasurementPdf.h"

#include <log4cpp/Category.hh>
static log4cpp::Category& logger( log4cpp::Category::getInstance( "Ubitrack.Drivers.PeopleTracking" ) );

namespace ublas = boost::numeric::ublas;

namespace Ubitrack { namespace Drivers { namespace PeopleTracking {

#if 1
	static Util::BlockTimer g_conicGenerationTimer( "conic generation", "Ubitrack.Drivers.PeopleTracking" );
	static Util::BlockTimer g_imageEvaluationTimer( "conic evaluation", "Ubitrack.Drivers.PeopleTracking" );
	#define DO_TIME( name, x ) UBITRACK_TIME( x )
#else
	#define DO_TIME( name, x )
#endif


BFL::Probability PersonMeasurementPdf::ProbabilityGet( const MatrixWrapper::ColumnVector& ) const
{
	MatrixWrapper::ColumnVector fullState = ConditionalArgumentGet(0);
	Math::Vector< 2 > state( ublas::subrange( fullState, 0, 2 ) );
	
	// check for overlap with other persons
	for ( std::list< PersonState >::const_iterator it = m_persons.begin(); it != m_persons.end(); it++ )
		if ( it->getID() != m_id && ublas::norm_2( state - it->getPos() ) < g_personSize( 0 ) * 2.5 )
			return 0.0;
			
	OnScreenConic roi;
	OnScreenConic conic;
	
	{
	DO_TIME( _timer1_, g_conicGenerationTimer );
	
	// compute ROI
	roi = OnScreenConic( Conic( Math::Vector< 3 >( state( 0 ), state( 1 ), g_personZ ),	g_personSize * 2.0, m_P ), 
		m_scene.getWidth(), m_scene.getHeight() );
		
	// compute person conic
	conic = OnScreenConic( Conic( Math::Vector< 3 >( state( 0 ), state( 1 ), g_personZ ), g_personSize, m_P ), 
		m_scene.getWidth(), m_scene.getHeight() );
	}
	
	// evaluate
	double weight;
	{
	DO_TIME( _timer2_, g_imageEvaluationTimer );
	weight = m_scene.evaluateParticle( conic, m_id, roi, m_pSumImage );
	}

	LOG4CPP_TRACE( logger, "Evaluated measurement at " << state << ": " << weight );

	m_maxWeight = std::max( m_maxWeight, weight );
	
	return weight;
}

} } } // namespace Ubitrack::Drivers::PeopleTracking
