/*
 * Ubitrack - Library for Ubiquitous Tracking
 * Copyright 2006, Technische Universitaet Muenchen, and individual
 * contributors as indicated by the @authors tag. See the 
 * copyright.txt in the distribution for a full listing of individual
 * contributors.
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA, or see the FSF site: http://www.fsf.org.
 */

/**
 * @ingroup drivers
 * @file
 * Declaration of a measurement function for the orocos BFL library.
 *
 * @author Daniel Pustka <daniel.pustka@in.tum.de>
 */

#ifndef __Ubitrack_Drivers_PeopleTracker_PersonMeasurementPdf_h_included__
#define __Ubitrack_Drivers_PeopleTracker_PersonMeasurementPdf_h_included__

#include <list>

#include <bfl/pdf/conditionalpdf.h>
#include <bfl/wrappers/matrix/matrix_wrapper.h>
#include <bfl/wrappers/matrix/vector_wrapper.h>

#include "PersonState.h"
#include "VirtualScene.h"

#include <log4cpp/Category.hh>
extern log4cpp::Category& logger;

namespace ublas = boost::numeric::ublas;

namespace Ubitrack { namespace Drivers { namespace PeopleTracking {

/** Defines the image measurement function for BFL */
class PersonMeasurementPdf
	: public BFL::ConditionalPdf< MatrixWrapper::ColumnVector, MatrixWrapper::ColumnVector >
{
public:
	/**
	 * Constructor.
	 * @param id id of the person to measure
	 * @param scene the virtual scene
	 * @param pSumImage the image of summed difference pixels
	 * @param persons list of all persons
	 * @param P 3x4 projection matrix
	 */
	PersonMeasurementPdf( unsigned id, const VirtualScene& scene, int* pSumImage, const std::list< PersonState >& persons,
		const Math::Matrix< 3, 4 >& P )
		: BFL::ConditionalPdf< MatrixWrapper::ColumnVector, MatrixWrapper::ColumnVector >( 4, 1 )
		, m_id( id )
		, m_scene( scene )
		, m_pSumImage( pSumImage )
		, m_persons( persons )
		, m_P( P )
		, m_maxWeight( 0 )
	{}
	
	/** evaluate a given particle */
	BFL::Probability ProbabilityGet( const MatrixWrapper::ColumnVector& ) const;

	/** returns the maximum weight given to a single particle */
	double getMaxWeight() const
	{ return m_maxWeight; }

protected:
	unsigned m_id;
	const VirtualScene& m_scene;
	int* m_pSumImage;
	const std::list< PersonState >& m_persons;
	const Math::Matrix< 3, 4 > m_P;
	mutable double m_maxWeight;
};

} } } // namespace Ubitrack::Drivers::PeopleTracking

#endif
