/*
 * Ubitrack - Library for Ubiquitous Tracking
 * Copyright 2006, Technische Universitaet Muenchen, and individual
 * contributors as indicated by the @authors tag. See the 
 * copyright.txt in the distribution for a full listing of individual
 * contributors.
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA, or see the FSF site: http://www.fsf.org.
 */

/**
 * @ingroup drivers
 * @file
 * Implementation of the person state that contains all information about a person.
 *
 * @author Daniel Pustka <daniel.pustka@in.tum.de>
 */

#include <algorithm>
#include <iostream>

#include <Ubitrack/Math/Vector.h>
#include <Ubitrack/Math/MatrixOperations.h>
#include <Ubitrack/Util/BlockTimer.h>

#include <bfl/pdf/gaussian.h>
#include <bfl/pdf/mcpdf.h>

#include "OnScreenConic.h"
#include "PersonState.h"
#include "PersonSystemPdf.h"
#include "PersonMeasurementPdf.h"

#include <log4cpp/Category.hh>
static log4cpp::Category& logger( log4cpp::Category::getInstance( "Ubitrack.Drivers.PeopleTracking" ) );

namespace ublas = boost::numeric::ublas;

namespace Ubitrack { namespace Drivers { namespace PeopleTracking {

#if 1
	static Util::BlockTimer g_timeUpdateTimer( "time update", "Ubitrack.Drivers.PeopleTracking" );
	static Util::BlockTimer g_measurementUpdateTimer( "measurement update", "Ubitrack.Drivers.PeopleTracking" );
	#define DO_TIME( name, x ) UBITRACK_TIME( x )
#else
	#define DO_TIME( name, x )
#endif

// global constants
const Math::Vector< 3 > g_personSize( 0.25, 0.25, 0.9 );
const double g_personZ( 0.9 );
const unsigned g_samples( 200 );
const double g_personQ( 0.4 );


PersonState::PersonState( unsigned id, const Math::Vector< 2 >& startPos )
	: m_id( id )
	, m_pos( startPos )
	, m_bUpdateOcclusionConic( true )
	, m_maxParticleWeight( 1 )
{
	LOG4CPP_DEBUG( logger, "New object created: id=" << id );

	// define starting distribution
	MatrixWrapper::ColumnVector prior_Mu( 4 );
	prior_Mu( 1 ) = startPos( 0 );
	prior_Mu( 2 ) = startPos( 1 );
	prior_Mu( 3 ) = 0;
	prior_Mu( 4 ) = 0;
	MatrixWrapper::SymmetricMatrix prior_Cov( 4 );
	static_cast< BoostSymmetricMatrix >( prior_Cov ) = ublas::identity_matrix< double >( 4 ) * 0.5;
	prior_Cov( 3, 3 ) = prior_Cov( 4, 4 ) = 0.05;
	BFL::Gaussian prior_cont( prior_Mu,prior_Cov ); 

	// sample starting distribution
	std::vector< BFL::Sample< MatrixWrapper::ColumnVector > > prior_samples( g_samples );
	BFL::MCPdf< MatrixWrapper::ColumnVector > prior_discr( g_samples, 4 );
	prior_cont.SampleFrom( prior_samples, g_samples, CHOLESKY, NULL );
	prior_discr.ListOfSamplesSet( prior_samples );

	// init filter
	m_pFilter.reset( new BFL::BootstrapFilter< MatrixWrapper::ColumnVector, MatrixWrapper::ColumnVector >( &prior_discr, 0, g_samples / 4.0 ) );
}


PersonState::~PersonState()
{
	LOG4CPP_DEBUG( logger, "Object deleted: id=" << m_id );
}


void PersonState::advance( double deltaT )
{
	DO_TIME( _timer_, g_timeUpdateTimer );
	
	PersonSystemPdf sysPdf( g_personQ, deltaT );
	BFL::SystemModel< MatrixWrapper::ColumnVector > sysModel( &sysPdf );
	LOG4CPP_TRACE( logger, "time update id=" << m_id );
	m_pFilter->Update( &sysModel );
	
	// update local state
	LOG4CPP_TRACE( logger, "local state update id=" << m_id << ", pos=" << m_pos );
	BFL::Pdf< MatrixWrapper::ColumnVector >* prop = m_pFilter->PostGet();
	MatrixWrapper::ColumnVector expect = prop->ExpectedValueGet();
	m_pos = ublas::subrange( expect, 0, 2 );
	LOG4CPP_DEBUG( logger, "local state update after time update id=" << m_id << ", pos=" << m_pos );

	m_bUpdateOcclusionConic = true;
}
	

void PersonState::measurement( const VirtualScene& scene, int* pSumImage, const std::list< PersonState >& persons, const Math::Matrix< 3, 4 >& P )
{
	DO_TIME( _timer_, g_measurementUpdateTimer );
	
	// construct measurement
	PersonMeasurementPdf mPdf( m_id, scene, pSumImage, persons, P );
	BFL::MeasurementModel< MatrixWrapper::ColumnVector, MatrixWrapper::ColumnVector > mModel( &mPdf );
	MatrixWrapper::ColumnVector dummy( 1 );
	dummy( 1 ) = 0.0;
	
	// update filter
	LOG4CPP_TRACE( logger, "Measurement update id=" << m_id );
	m_pFilter->Update( &mModel, dummy );
	m_maxParticleWeight = mPdf.getMaxWeight();
	
	// update local state
	BFL::Pdf< MatrixWrapper::ColumnVector >* posterior = m_pFilter->PostGet();
	MatrixWrapper::ColumnVector expect = posterior->ExpectedValueGet();
	m_pos = ublas::subrange( expect, 0, 2 );
	LOG4CPP_DEBUG( logger, "local state update after measurement id=" << m_id << ", pos=" << m_pos );

	m_bUpdateOcclusionConic = true;
}


const Conic& PersonState::getOcclusionConic( const Math::Matrix< 3, 4 >& P )
{
	if ( m_bUpdateOcclusionConic )
	{
		m_occlusionConic = Conic( Math::Vector< 3 >( m_pos( 0 ), m_pos( 1 ), g_personZ ), g_personSize * 1.2, P );
		m_bUpdateOcclusionConic = false;
	}

	return m_occlusionConic;
}


bool PersonState::isValid() const
{
	// based on maximum particle weight
	if ( m_maxParticleWeight < 0.01 )
		return false;

	// perform self-validation
	BFL::MCPdf< MatrixWrapper::ColumnVector >* posterior = m_pFilter->PostGet();
	if ( !posterior )
		return true;
		
	MatrixWrapper::SymmetricMatrix cov = posterior->CovarianceGet();
	if ( sqrt( cov( 1, 1 ) + cov( 2, 2 ) ) > 1.0 ) 
		return false;
		
	// check if all particles have zero weight
	for ( std::vector< BFL::WeightedSample< MatrixWrapper::ColumnVector > >::const_iterator it = posterior->ListOfSamplesGet().begin();
		it != posterior->ListOfSamplesGet().end(); it++ )
		if ( it->WeightGet() > 0.0 )
			return true;
			
	return false;
}


void PersonState::drawDebugImage( Vision::Image& img, const VirtualScene& scene, int* pSumImage, const Math::Matrix< 3, 4 >& P )
{
	BFL::MCPdf< MatrixWrapper::ColumnVector >* posterior = m_pFilter->PostGet();
	
	// evaluate current position and draw scene into image
	OnScreenConic mainConic = OnScreenConic( Conic( Math::Vector< 3 >( m_pos( 0 ), m_pos( 1 ), g_personZ ), g_personSize, P ), 
		img.width, img.height );
	OnScreenConic roiConic = OnScreenConic( Conic( Math::Vector< 3 >( m_pos( 0 ), m_pos( 1 ), g_personZ ), g_personSize * 2, P ), 
		img.width, img.height );
	scene.evaluateParticleDebug( mainConic, m_id, roiConic, pSumImage, img );

	// draw expected value
	MatrixWrapper::ColumnVector expect = posterior->ExpectedValueGet();
	Math::Vector< 4 > pos3Dh;
	pos3Dh( 0 ) = expect( 1 ); 
	pos3Dh( 1 ) = expect( 2 );
	pos3Dh( 2 ) = 0;
	pos3Dh( 3 ) = 1;
	Math::Vector< 3 > pos2Dh( ublas::prod( P, pos3Dh ) );
	pos2Dh /= double( pos2Dh( 2 ) );
	cvCircle( img, cvPoint( cvRound( pos2Dh( 0 ) ), cvRound( pos2Dh( 1 ) ) ), 5, CV_RGB( 64, 255, 64 ) );

	// find maximal particle weight
	double fMaxWeight = 0.0001;
	for ( std::vector< BFL::WeightedSample< MatrixWrapper::ColumnVector > >::const_iterator it = posterior->ListOfSamplesGet().begin();
		it != posterior->ListOfSamplesGet().end(); it++ )
		fMaxWeight = std::max( fMaxWeight, double( it->WeightGet() ) );

	// draw individual particles
	for ( std::vector< BFL::WeightedSample< MatrixWrapper::ColumnVector > >::const_iterator it = posterior->ListOfSamplesGet().begin();
		it != posterior->ListOfSamplesGet().end(); it++ )
	{
		LOG4CPP_TRACE( logger, "found " << Math::Vector< 4 >( it->ValueGet() ) << ", weight=" << it->WeightGet() );
		pos3Dh( 0 ) = it->ValueGet()( 1 ); 
		pos3Dh( 1 ) = it->ValueGet()( 2 );
		pos2Dh =  ublas::prod( P, pos3Dh );
		pos2Dh /= double( pos2Dh( 2 ) );
		double normWeight = it->WeightGet() / fMaxWeight;
		cvCircle( img, cvPoint( cvRound( pos2Dh( 0 ) ), cvRound( pos2Dh( 1 ) ) ), 1, 
			CV_RGB( cvRound( 255 * normWeight ), 0, cvRound( 255 - 255 * normWeight ) ) );
	}
}


} } } // namespace Ubitrack::Drivers::PeopleTracking
