/*
 * Ubitrack - Library for Ubiquitous Tracking
 * Copyright 2006, Technische Universitaet Muenchen, and individual
 * contributors as indicated by the @authors tag. See the 
 * copyright.txt in the distribution for a full listing of individual
 * contributors.
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA, or see the FSF site: http://www.fsf.org.
 */

/**
 * @ingroup drivers
 * @file
 * Declaration of the person state that contains all information about a person.
 *
 * @author Daniel Pustka <daniel.pustka@in.tum.de>
 */


#ifndef __Ubitrack_Drivers_PeopleTracker_PersonState_h_included__
#define __Ubitrack_Drivers_PeopleTracker_PersonState_h_included__

#include <list>

#include <boost/scoped_ptr.hpp>

#include <Vision/Image.h>
#include <Ubitrack/Math/Matrix.h>
#include <Ubitrack/Math/Vector.h>

#include <bfl/wrappers/matrix/matrix_wrapper.h>
#include <bfl/wrappers/matrix/vector_wrapper.h>
#include <bfl/filter/bootstrapfilter.h>

#include "Conic.h"
#include "OnScreenConic.h"
#include "VirtualScene.h"

namespace ublas = boost::numeric::ublas;

namespace Ubitrack { namespace Drivers { namespace PeopleTracking {

extern const Math::Vector< 3 > g_personSize;
extern const double g_personZ;
extern const unsigned g_samples;

/**
 * Contains all information about a tracked person. 
 * This class encapsulates the particle filter
 */
class PersonState
{
public:
	/**
	 * Construct a new person
	 * @param id id of the person
	 * @param startPos start position (in 2D
	 */
	PersonState( unsigned id, const Math::Vector< 2 >& startPos );
	
	/** destructor */
	~PersonState();

	/** 
	 * time update of the particle filter.
	 * @param deltaT time elapsed in seconds
	 */
	void advance( double deltaT );
	
	/**
	 * Make a particle filter measurement.
	 * @param scene the virtual scene containing all objects (possibly including this one)
	 * @param pSumImage image with the sum of all pixels in the difference image
	 * @param persons list of all persons to keep a minimum distance
	 * @param P 3x4 projection matrix
	 */
	void measurement( const VirtualScene& scene, int* pSumImage, const std::list< PersonState >& persons, const Math::Matrix< 3, 4 >& P );

	/**
	 * returns the occlusion conic to insert into the virtual scene.
	 * @param P 3x4 projection matrix
	 */
	const Conic& getOcclusionConic( const Math::Matrix< 3, 4 >& P );
	
	/** does the particle filter still track a person? */
	bool isValid() const;

	/** draw the particles and conics into a debug image */
	void drawDebugImage( Vision::Image& img, const VirtualScene& scene, int* pSumImage, const Math::Matrix< 3, 4 >& P );
	
	/** id of the person */
	unsigned getID() const
	{ return m_id; }

	/** currently estimated position of the person */
	const Math::Vector< 2 >& getPos() const
	{ return m_pos; }

protected:
	/** id of the person */
	unsigned m_id;

	/** last known position */
	Math::Vector< 2 > m_pos;

	/** last known occlusion conic */
	Conic m_occlusionConic;

	/** does the occlusion conic need updating */
	bool m_bUpdateOcclusionConic;

	/** the particle filter */
	boost::shared_ptr< BFL::BootstrapFilter< MatrixWrapper::ColumnVector, MatrixWrapper::ColumnVector > > m_pFilter;

	/** the maximum weight given to a particle */
	double m_maxParticleWeight;
};

} } } // namespace Ubitrack::Drivers::PeopleTracking

#endif