/*
 * Ubitrack - Library for Ubiquitous Tracking
 * Copyright 2006, Technische Universitaet Muenchen, and individual
 * contributors as indicated by the @authors tag. See the 
 * copyright.txt in the distribution for a full listing of individual
 * contributors.
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA, or see the FSF site: http://www.fsf.org.
 */

/**
 * @ingroup drivers
 * @file
 * Definition of a process description for BFL.
 *
 * @author Daniel Pustka <daniel.pustka@in.tum.de>
 */

#ifndef __Ubitrack_Drivers_PeopleTracker_PersonSystemPdf_h_included__
#define __Ubitrack_Drivers_PeopleTracker_PersonSystemPdf_h_included__

#include <bfl/pdf/conditionalpdf.h>
#include <bfl/pdf/gaussian.h>
#include <bfl/wrappers/matrix/matrix_wrapper.h>
#include <bfl/wrappers/matrix/vector_wrapper.h>

#include <log4cpp/Category.hh>
extern log4cpp::Category& logger;

namespace Ubitrack { namespace Drivers { namespace PeopleTracking {

/** 
 * Describes the system dynamics for the BFL library.
 * The dynamics are based on a 2D constant-velocity model.
 */
class PersonSystemPdf
	: public BFL::ConditionalPdf< MatrixWrapper::ColumnVector, MatrixWrapper::ColumnVector >
{
public:
	/** 
	 * constructor.
	 * @param q amount of variation
	 * @param deltaT time difference in seconds
	 */
	PersonSystemPdf( double q, double deltaT )
		: BFL::ConditionalPdf< MatrixWrapper::ColumnVector, MatrixWrapper::ColumnVector > ( 4, 2 )
		, m_q( q )
		, m_deltaT( deltaT )
	{}
	
	bool SampleFrom( BFL::Sample< MatrixWrapper::ColumnVector >& one_sample, int method, void * args ) const
	{
		// update state
		MatrixWrapper::ColumnVector state = ConditionalArgumentGet( 0 );
		Math::Vector< 4 > oldState( state );

		state( 1 ) += m_deltaT * state( 3 );
		state( 2 ) += m_deltaT * state( 4 );
		
		// construct system noise description
		MatrixWrapper::ColumnVector sysNoise_Mu( 4 );
		sysNoise_Mu( 1 ) = 0.0;
		sysNoise_Mu( 2 ) = 0.0;
		sysNoise_Mu( 3 ) = 0.0;
		sysNoise_Mu( 4 ) = 0.0;

		MatrixWrapper::SymmetricMatrix sysNoise_Cov( 4 );
		sysNoise_Cov( 1, 1 ) = m_q * m_deltaT * m_deltaT * m_deltaT / 3;
		sysNoise_Cov( 1, 2 ) = 0.0;
		sysNoise_Cov( 1, 3 ) = m_q * m_deltaT * m_deltaT / 2;
		sysNoise_Cov( 1, 4 ) = 0.0;
		sysNoise_Cov( 2, 1 ) = 0.0;
		sysNoise_Cov( 2, 2 ) = sysNoise_Cov( 1, 1 );
		sysNoise_Cov( 2, 3 ) = 0.0;
		sysNoise_Cov( 2, 4 ) = sysNoise_Cov( 1, 3 );
		sysNoise_Cov( 3, 1 ) = sysNoise_Cov( 1, 3 );
		sysNoise_Cov( 3, 2 ) = 0.0;
		sysNoise_Cov( 3, 3 ) = m_q * m_deltaT;
		sysNoise_Cov( 3, 4 ) = 0.0;
		sysNoise_Cov( 4, 1 ) = 0.0;
		sysNoise_Cov( 4, 2 ) = sysNoise_Cov( 2, 4 );
		sysNoise_Cov( 4, 3 ) = 0.0;
		sysNoise_Cov( 4, 4 ) = m_q * m_deltaT;

		BFL::Gaussian system_Uncertainty( sysNoise_Mu, sysNoise_Cov );
		
		// add some process noise
		BFL::Sample< MatrixWrapper::ColumnVector > noise;
		system_Uncertainty.SampleFrom( noise, method, args );
		state = state + noise.ValueGet();
		
		LOG4CPP_TRACE( logger, "Updated particle " << oldState << " to " << Math::Vector< 4 >( state ) );

		// update sample
		one_sample.ValueSet( state );
		
		return true;
	}
	
protected:
	double m_q;
	double m_deltaT;
};

} } } // namespace Ubitrack::Drivers::PeopleTracking

#endif
