/*
 * Ubitrack - Library for Ubiquitous Tracking
 * Copyright 2006, Technische Universitaet Muenchen, and individual
 * contributors as indicated by the @authors tag. See the 
 * copyright.txt in the distribution for a full listing of individual
 * contributors.
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA, or see the FSF site: http://www.fsf.org.
 */

/**
 * @ingroup drivers
 * @file
 * Implementation of the virtual scene that descibes all objects for occlusion handling.
 *
 * @author Daniel Pustka <daniel.pustka@in.tum.de>
 */

#include <algorithm>
#include "VirtualScene.h"

#include <log4cpp/Category.hh>
static log4cpp::Category& logger( log4cpp::Category::getInstance( "Ubitrack.Drivers.PeopleTracking" ) );


namespace Ubitrack { namespace Drivers { namespace PeopleTracking {


VirtualScene::VirtualScene( int nImageWidth, int nImageHeight, int nObjects )
	: m_nImageWidth( nImageWidth )
{
	m_events.resize( nImageHeight );

	// reserve maximum amount of space, here memory consumption could be slightly reduced by employing a better strategy.
	// also reserving memory for every line could be avoided by doing something more clever...
	
	for ( std::size_t i = 0; i < m_events.size(); i++ )
		m_events[ i ].reserve( 2 * nObjects  + 1 );
}


void VirtualScene::addObject( unsigned id, const OnScreenConic& conic )
{
	for ( unsigned i = 0; i < conic.m_x.size(); i++ )
	{
		m_events[ i + conic.m_y.begin ].push_back( ObjectEvent( conic.m_x[ i ].begin, int( id ) ) );
		m_events[ i + conic.m_y.begin ].push_back( ObjectEvent( conic.m_x[ i ].end, -int( id ) ) );
	}
}


void VirtualScene::normalize()
{
	// sort all lines to have ascending x values
	for ( std::size_t i = 0; i < m_events.size(); i++ )
	{
		std::sort( m_events[ i ].begin(), m_events[ i ].end(), sortEvents );
		
		// add an end marker to each line
		m_events[ i ].push_back( ObjectEvent( m_nImageWidth, 0 ) );
	}
}

// redefined later for evaluateParticleDebug
#define DRAW_DEBUG_PIXEL( x, y, c )
#define	DRAW_DEBUG_LINE( xbegin, xend, y, c )

double VirtualScene::evaluateParticle( const OnScreenConic& conic, unsigned id, const OnScreenConic& roi, const int* pIntensitySum ) const
{
	// run over lines in roi
	int orSum = 0; // or = max = ( inside ? 255 : intensity )
	int andSum = 0; // and = product = ( inside ? intensity : 0 )
	
	for ( int y = roi.m_y.begin; y != roi.m_y.end; y++ )
	{
		// shortcuts
		const OnScreenConic::BeginEnd& roiLine( roi.m_x[ y - roi.m_y.begin ] );
		const std::vector< ObjectEvent >& eventLine( m_events[ y ] );
		const int* pIntensityLine = pIntensitySum + m_nImageWidth * y;

		// variables for counting
		std::vector< ObjectEvent >::const_iterator it = eventLine.begin();
		int objectCounter = 0;
		int lastX;

		// search for beginning of roi
		for ( ; it->pixel < roiLine.begin; it++ )
			if ( it->id < 0 )
				if ( -it->id != int( id ) )
				{
					objectCounter--;
					DRAW_DEBUG_PIXEL( it->pixel, y, 2 );
				}
				else;
			else
				if ( it->id != int( id ) )
				{
					objectCounter++;
					DRAW_DEBUG_PIXEL( it->pixel, y, 0 );
				}
				else;

		lastX = roiLine.begin;

		if ( y >= conic.m_y.begin && y < conic.m_y.end )
		{
			// does this line contain the core conic?
			const OnScreenConic::BeginEnd& conicLine( conic.m_x[ y - conic.m_y.begin ] );

			// walk to beginning of conic and add non-occluded parts to orSum
			for ( ; it->pixel < conicLine.begin; it++ )
				if ( it->id < 0 )
					if ( -it->id != int( id ) )
					{
						objectCounter--;
						lastX = it->pixel;
						DRAW_DEBUG_PIXEL( it->pixel, y, 2 );
					}
					else;
				else
					if ( it->id != int( id ) )
					{
						if ( !objectCounter )
						{
							orSum += pIntensityLine[ it->pixel ] - pIntensityLine[ lastX ];
							DRAW_DEBUG_LINE( lastX, it->pixel, y, 0 );
						}
						objectCounter++;
						DRAW_DEBUG_PIXEL( it->pixel, y, 0 );
					}
					else;

			if ( !objectCounter )
			{
				orSum += pIntensityLine[ conicLine.begin ] - pIntensityLine[ lastX ];
				DRAW_DEBUG_LINE( lastX, conicLine.begin, y, 0 );
			}

			lastX = conicLine.begin;

			// walk to end of core conic and add non-occluded parts to andSum and orSum
			for ( ; it->pixel < conicLine.end; it++ )
				if ( it->id < 0 )
					if ( -it->id != int( id ) )
					{
						objectCounter--;
						lastX = it->pixel;
						DRAW_DEBUG_PIXEL( it->pixel, y, 2 );
					}
					else;
				else
					if ( it->id != int( id ) )
					{
						if ( !objectCounter )
						{
							orSum += 256 * ( it->pixel - lastX );
							andSum += pIntensityLine[ it->pixel ] - pIntensityLine[ lastX ];
							DRAW_DEBUG_LINE( lastX, it->pixel, y, 0 );
						}
						objectCounter++;
						DRAW_DEBUG_PIXEL( it->pixel, y, 0 );
					}
					else;
					
			if ( !objectCounter )
			{
				orSum += 256 * ( conicLine.end - lastX );
				andSum += pIntensityLine[ conicLine.end ] - pIntensityLine[ lastX ];
				DRAW_DEBUG_LINE( lastX, conicLine.end, y, 0 );
			}
			
			lastX = conicLine.end;
		}

		// walk to end of roi and add non-occluded parts to orSum
		for ( ; it->pixel < roiLine.end; it++ )
			if ( it->id < 0 )
				if ( -it->id != int( id ) )
				{
					objectCounter--;
					lastX = it->pixel;
					DRAW_DEBUG_PIXEL( it->pixel, y, 2 );
				}
				else;
			else
				if ( it->id != int( id ) )
				{
					if ( !objectCounter )
					{
						orSum += pIntensityLine[ it->pixel ] - pIntensityLine[ lastX ];
						DRAW_DEBUG_LINE( lastX, it->pixel, y, 0 );
					}
					objectCounter++;
					DRAW_DEBUG_PIXEL( it->pixel, y, 0 );
				}
				else;

		if ( !objectCounter )
		{
			orSum += pIntensityLine[ roiLine.end ] - pIntensityLine[ lastX ];
			DRAW_DEBUG_LINE( lastX, roiLine.end, y, 0 );
		}
	}
	
	double result = orSum > 0 ? ( double( andSum ) / orSum ) : 0; // TODO: should not return 0 if orSum == 0, but return an exception
	LOG4CPP_TRACE( logger, "VSc: orSum=" << orSum << ", andSum=" << andSum << ", weight=" << result );
	
	return result;
}


// redefine the debug image macros, so the source of the routine can just be copied
#undef DRAW_DEBUG_PIXEL
#define DRAW_DEBUG_PIXEL( x, y, c ) *((unsigned char*)debugImage.imageData + y * debugImage.widthStep + 3 * x + c ) = 255
#undef DRAW_DEBUG_LINE
#define	DRAW_DEBUG_LINE( xbegin, xend, y, c ) for ( int fgh = xbegin; fgh < xend; fgh++ ) { DRAW_DEBUG_PIXEL( fgh, y, 1 ); DRAW_DEBUG_PIXEL( fgh, y, c ); }


// this is an exact copy of evaluateParticle, just with different #defines
double VirtualScene::evaluateParticleDebug( const OnScreenConic& conic, unsigned id, const OnScreenConic& roi, 
	const int* pIntensitySum, Vision::Image& debugImage ) const
{
	// run over lines in roi
	int orSum = 0; // or = max = ( inside ? 255 : intensity )
	int andSum = 0; // and = product = ( inside ? intensity : 0 )
	
	for ( int y = roi.m_y.begin; y != roi.m_y.end; y++ )
	{
		// shortcuts
		const OnScreenConic::BeginEnd& roiLine( roi.m_x[ y - roi.m_y.begin ] );
		const std::vector< ObjectEvent >& eventLine( m_events[ y ] );
		const int* pIntensityLine = pIntensitySum + m_nImageWidth * y;

		// variables for counting
		std::vector< ObjectEvent >::const_iterator it = eventLine.begin();
		int objectCounter = 0;
		int lastX;

		// search for beginning of roi
		for ( ; it->pixel < roiLine.begin; it++ )
			if ( it->id < 0 )
				if ( -it->id != int( id ) )
				{
					objectCounter--;
					DRAW_DEBUG_PIXEL( it->pixel, y, 2 );
				}
				else;
			else
				if ( it->id != int( id ) )
				{
					objectCounter++;
					DRAW_DEBUG_PIXEL( it->pixel, y, 0 );
				}
				else;

		lastX = roiLine.begin;

		if ( y >= conic.m_y.begin && y < conic.m_y.end )
		{
			// does this line contain the core conic?
			const OnScreenConic::BeginEnd& conicLine( conic.m_x[ y - conic.m_y.begin ] );

			// walk to beginning of conic and add non-occluded parts to orSum
			for ( ; it->pixel < conicLine.begin; it++ )
				if ( it->id < 0 )
					if ( -it->id != int( id ) )
					{
						objectCounter--;
						lastX = it->pixel;
						DRAW_DEBUG_PIXEL( it->pixel, y, 2 );
					}
					else;
				else
					if ( it->id != int( id ) )
					{
						if ( !objectCounter )
						{
							orSum += pIntensityLine[ it->pixel ] - pIntensityLine[ lastX ];
							DRAW_DEBUG_LINE( lastX, it->pixel, y, 0 );
						}
						objectCounter++;
						DRAW_DEBUG_PIXEL( it->pixel, y, 0 );
					}
					else;

			if ( !objectCounter )
			{
				orSum += pIntensityLine[ conicLine.begin ] - pIntensityLine[ lastX ];
				DRAW_DEBUG_LINE( lastX, conicLine.begin, y, 0 );
			}

			lastX = conicLine.begin;

			// walk to end of core conic and add non-occluded parts to andSum and orSum
			for ( ; it->pixel < conicLine.end; it++ )
				if ( it->id < 0 )
					if ( -it->id != int( id ) )
					{
						objectCounter--;
						lastX = it->pixel;
						DRAW_DEBUG_PIXEL( it->pixel, y, 2 );
					}
					else;
				else
					if ( it->id != int( id ) )
					{
						if ( !objectCounter )
						{
							orSum += 256 * ( it->pixel - lastX );
							andSum += pIntensityLine[ it->pixel ] - pIntensityLine[ lastX ];
							DRAW_DEBUG_LINE( lastX, it->pixel, y, 2 );
						}
						objectCounter++;
						DRAW_DEBUG_PIXEL( it->pixel, y, 0 );
					}
					else;
					
			if ( !objectCounter )
			{
				orSum += 256 * ( conicLine.end - lastX );
				andSum += pIntensityLine[ conicLine.end ] - pIntensityLine[ lastX ];
				DRAW_DEBUG_LINE( lastX, conicLine.end, y, 2 );
			}
			
			lastX = conicLine.end;
		}

		// walk to end of roi and add non-occluded parts to orSum
		for ( ; it->pixel < roiLine.end; it++ )
			if ( it->id < 0 )
				if ( -it->id != int( id ) )
				{
					objectCounter--;
					lastX = it->pixel;
					DRAW_DEBUG_PIXEL( it->pixel, y, 2 );
				}
				else;
			else
				if ( it->id != int( id ) )
				{
					if ( !objectCounter )
					{
						orSum += pIntensityLine[ it->pixel ] - pIntensityLine[ lastX ];
						DRAW_DEBUG_LINE( lastX, it->pixel, y, 0 );
					}
					objectCounter++;
					DRAW_DEBUG_PIXEL( it->pixel, y, 0 );
				}
				else;

		if ( !objectCounter )
		{
			orSum += pIntensityLine[ roiLine.end ] - pIntensityLine[ lastX ];
			DRAW_DEBUG_LINE( lastX, roiLine.end, y, 0 );
		}
	}
	
	double result = orSum > 0 ? ( double( andSum ) / orSum ) : 0; // TODO: should not return 0 if orSum == 0, but return an exception
	LOG4CPP_DEBUG( logger, "VSc: orSum=" << orSum << ", andSum=" << andSum << ", weight=" << result );
	
	return result;
}

} } } // namespace Ubitrack::Drivers::PeopleTracking
