/*
 * Ubitrack - Library for Ubiquitous Tracking
 * Copyright 2006, Technische Universitaet Muenchen, and individual
 * contributors as indicated by the @authors tag. See the 
 * copyright.txt in the distribution for a full listing of individual
 * contributors.
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA, or see the FSF site: http://www.fsf.org.
 */

/**
 * @ingroup drivers
 * @file
 * Declaration of the virtual scene that descibes all objects for occlusion handling.
 *
 * @author Daniel Pustka <daniel.pustka@in.tum.de>
 */

#ifndef __Ubitrack_Drivers_PeopleTracker_VirtualScene_h_included__
#define __Ubitrack_Drivers_PeopleTracker_VirtualScene_h_included__

#include <vector>
#include <Vision/Image.h>
#include "OnScreenConic.h"

namespace Ubitrack { namespace Drivers { namespace PeopleTracking {

/**
 * A virtual scene, consisting of convex objects. 
 * Each object has an id and is described by its beginning and end pixels in each image line. 
 * This scene is used to describe inter-object occlusions in the people tracking.
 */
class VirtualScene
{
public:
	/** 
	 * Construct a virtual scene.
	 * @param nImageWidth width in pixels of the scene
	 * @param nImageHeight height in pixels of the scene
	 * @param nObjects number of objects to add to the scene
	 */
	VirtualScene( int nImageWidth, int nImageHeight, int nObjects );
	
	/** adds a conic object to the virtual scene. Scene must be normalized before usage. id >= 1! */
	void addObject( unsigned id, const OnScreenConic& conic );

	/** normalizes the scene */
	void normalize();
	
	/** evaluates the particle weight. */
	double evaluateParticle( const OnScreenConic& conic, unsigned id, const OnScreenConic& roi, const int* pIntensitySum ) const;
	
	/** evaluates particle weight and simultaneously draws a debug image (more expensive) */
	double evaluateParticleDebug( const OnScreenConic& conic, unsigned id, const OnScreenConic& roi, 
		const int* pIntensitySum, Vision::Image& debugImage ) const;

	/** returns the widht in pixels of the scene */
	int getWidth() const
	{ return m_nImageWidth; }
	
	/** returns the height in pixels of the scene */
	int getHeight() const
	{ return m_events.size(); }
	
protected:
	/** width of the scene. */
	int m_nImageWidth;
	
	/** describes an event (object start/object end) in an image line */
	struct ObjectEvent
	{
		ObjectEvent( int _pixel, int _id )
			: pixel( _pixel )
			, id( _id )
		{}
	
		/** the index of the pixel where the object starts or ends */
		int pixel;
		
		/** abs(id) = id of the objects. Start if positive, end if negative */
		int id;
	};

	/** for each image line a list of events */
	std::vector< std::vector< ObjectEvent > > m_events;

	/** helper predicate for sorting object events by their position. If the position is the same, starting events (positive ids) come before ending events (negative id) */
	static bool sortEvents( const VirtualScene::ObjectEvent& a, const VirtualScene::ObjectEvent& b )
	{ return a.pixel != b.pixel ? a.pixel < b.pixel : a.id > b.id; }
};

} } } // namespace Ubitrack::Drivers::PeopleTracking

#endif
