/*
 * Ubitrack - Library for Ubiquitous Tracking
 * Copyright 2006, Technische Universitaet Muenchen, and individual
 * contributors as indicated by the @authors tag. See the
 * copyright.txt in the distribution for a full listing of individual
 * contributors.
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA, or see the FSF site: http://www.fsf.org.
 */


/**
 * @ingroup driver_components
 * @file
 * RFID Stuff
 *
 * @author M.H. & B.B.
 */

#include <cstring>
#include "RFIDDataBase.h"

namespace Ubitrack { namespace Drivers { namespace RFID {

RFIDDataBase::RFIDDataBase(Graph::UTQLSubgraph::NodePtr rfidTagListNode)
{
	unsigned long testID = 1000;

	if ( !rfidTagListNode->hasAttribute( "TagCloud" ) )
		UBITRACK_THROW( "RFID tag cloud list without \"TagCloud\" attribute" );

	const TiXmlElement* pElement = rfidTagListNode->getAttribute( "TagCloud" ).getXML();
	if ( !pElement )
		UBITRACK_THROW( "Edge does not have a TagCloud attribute element" );

	const TiXmlNode* pNode = pElement->FirstChild("Value");
	if ((pNode == NULL) || (pNode->ToElement() == NULL))
		UBITRACK_THROW( "TagCloud does not have a 'Value' tag." );

	if(std::strcmp( pNode->ToElement()->Attribute( "xsi:type" ), "utql:ListOfStructValueType" ) )
		UBITRACK_THROW( "TagCloud does not have a 'Value' of type 'utql:ListOfStructValueType'" );
				
	pNode = pNode->FirstChild("Attribute");
	if ((pNode == NULL) || (pNode->ToElement() == NULL))
		UBITRACK_THROW( "TagCloud::Value does not have a Attribute tag." );

	if(std::strcmp( pNode->ToElement()->Attribute( "name" ), "Tag" ) )
		UBITRACK_THROW( "TagCloud::Value does not have a Attribute with name 'Tag'" );


	const TiXmlNode* pSubNode;

	for( ; pNode; pNode = pNode->NextSibling( "Attribute" ) )
	{
		if (pNode->ToElement() == NULL)
			UBITRACK_THROW( "TagCloud::Value::Attribute does not have an element." );


		pSubNode = pNode->FirstChild("Value");
		if ((pSubNode == NULL) || (pSubNode->ToElement() == NULL))
			UBITRACK_THROW( "Tag does not have a 'Value' child." );

		pSubNode = pSubNode->FirstChild("Attribute");
		if ((pSubNode == NULL) || (pSubNode->ToElement() == NULL))
			UBITRACK_THROW( "Tag::Value does not have a 'Attribute' child." );

		if(std::strcmp( pSubNode->ToElement()->Attribute( "name" ), "TagID" ) )
			UBITRACK_THROW( "Tag::Value::Attribute does not have name 'TagID'" );

		// get data:
//		const char* sValue = pSubNode->ToElement()->Attribute( "value" );
//		testID = strtoul(sValue,NULL,16);

		std::string stringID = std::string(pSubNode->ToElement()->Attribute( "value" ));
		unsigned long tagID = 0;
		char ptr[2];
		int base = 8;

		for(int j = stringID.size()-base; j < stringID.size(); j+=2)
		{
			ptr[0] = stringID[j];
			ptr[1] = stringID[j+1];
			unsigned char temp = (unsigned char)(strtoul( ptr, NULL, 16));
			tagID |= ( temp ) << ( (11-base) * 8 );
			base++;
		}

		pSubNode = pSubNode->NextSibling( "Attribute" );
		if ((pSubNode == NULL) || (pSubNode->ToElement() == NULL))
			UBITRACK_THROW( "Tag::Value does not have a second 'Attribute' child." );

		if(std::strcmp( pSubNode->ToElement()->Attribute( "name" ), "staticPosition" ) )
			UBITRACK_THROW( "Tag::Value::Attribute does not have name 'staticPosition'" );

		const char* sValue = pSubNode->ToElement()->Attribute( "value" );

		double p[3];
		std::istringstream positionString( sValue );
		for ( unsigned i = 0; i < 3; i++ )
			positionString >> p[i];
		
		m_db.insert( std::pair< Math::Scalar< unsigned long > , Math::Vector< 3> > (tagID,Math::Vector< 3 >( p )));

	}

}

RFIDDataBase::~RFIDDataBase()
{
}

Measurement::PositionList RFIDDataBase::GetPositionsFromIDs(const Measurement::IDList& ids)
{
	std::vector < Math::Vector< 3 > > positions;
	std::vector < Math::Scalar< unsigned long > >::iterator it = ids->begin();
	for(it++; it != ids->end(); it++)
	{
		int count = m_db.count(*it);
		if( count != 1)
		{
			if(count == 0)
				UBITRACK_THROW( "RFID Tag not known to Database" );
			else
				UBITRACK_THROW( "Tag::Value does not have a second 'Attribute' child." );
		}
		positions.push_back( m_db[*it] );
	}

	return Measurement::PositionList(ids.time(), positions);
}


} } } // namespace Ubitrack::Drivers::RFID
