/*
 * Ubitrack - Library for Ubiquitous Tracking
 * Copyright 2006, Technische Universitaet Muenchen, and individual
 * contributors as indicated by the @authors tag. See the
 * copyright.txt in the distribution for a full listing of individual
 * contributors.
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA, or see the FSF site: http://www.fsf.org.
 */


/**
 * @ingroup driver_components
 * @file
 * NoMark driver
 * This file contains the driver component to
 * the nomark edge tracker
 *
 * @author Benjamin Becker <benjamin.becker@eads.net>
 */

#include <string>
#include <cstdlib>
#include <log4cpp/Category.hh>

#include <Ubitrack/Dataflow/PushSupplier.h>
#include <Ubitrack/Dataflow/PushConsumer.h>
#include <Ubitrack/Dataflow/PullConsumer.h>
#include <Ubitrack/Dataflow/PullSupplier.h>
#include <Ubitrack/Dataflow/ComponentFactory.h>
#include <Ubitrack/Dataflow/Component.h>
#include <Ubitrack/Measurement/Measurement.h>
#include <Ubitrack/Measurement/TimestampSync.h>
#include <Ubitrack/Util/Exception.h>
#include <Ubitrack/Util/OS.h>

#include "RFIDTracker.h"
#include "RFIDDataBase.h"

static log4cpp::Category& logger( log4cpp::Category::getInstance( "Drivers.RFIDTrackerModule" ) );

namespace Ubitrack { namespace Drivers {

using namespace Dataflow;


class RFIDTrackerModule
	: public Component
{
public:

	/** constructor */
	RFIDTrackerModule( const std::string& sName, boost::shared_ptr< Graph::UTQLSubgraph > subgraph )
		: Component( sName )
		, m_in_port_idlist( "DetectedTags", *this, boost::bind( &RFIDTrackerModule::eventInIDList, this, _1 ) )
		, m_in_port_rotation( "InputRotation", *this, boost::bind( &RFIDTrackerModule::eventInRotation, this, _1 ) )
		, m_out_port_pose( "Output", *this )
		, m_out_port_errorpose( "OutputError", *this )
	{
		m_rfid_database = new RFID::RFIDDataBase(subgraph->getNode( "TagCloud" ));
		m_rfid_tracker = new RFID::RFIDTracker( /* m_rfid_database */);
	}

	/** destructor */
	~RFIDTrackerModule()
	{	
		delete m_rfid_tracker;
		delete m_rfid_database;
		LOG4CPP_DEBUG( logger, "RFIDTrackerModule deleted" );
	}

	void eventInIDList( const Ubitrack::Measurement::IDList& p_IDList )
	{	
		int i = 0; 

		std::vector < Math::Scalar< unsigned long > >::iterator it = p_IDList->begin();
		std::cout<<"RFID Reader at Power "<<*(it++)<<" detected "<< p_IDList->size()-1<< " tags:   "<<std::endl;

		if(p_IDList->size() == 1)
			return;

		Measurement::PositionList positions = m_rfid_database->GetPositionsFromIDs(p_IDList);
		std::vector < Math::Vector< 3 > >::iterator it2 = positions->begin();

		while((it != p_IDList->end()) && (it2 != positions->end()))
		{
			std::cout<<"Tag #"<<++i<<" ( "<<*it<<" )  |  ( "<<*it2<<" )"<<std::endl;
			it++;
			it2++;
		}
		std::cout<<std::endl;


		if(m_rfid_tracker->EstimatePosition( m_rot, m_rfid_database->GetPositionsFromIDs(p_IDList) ))
			m_out_port_pose.send( Measurement::Pose( p_IDList.time(), Math::Pose(*m_rot,m_rfid_tracker->GetPosition() )));
	}

	void eventInRotation( const Ubitrack::Measurement::Rotation& p_rot )
	{
		m_rot = p_rot;
	}

protected:
	PushConsumer< Ubitrack::Measurement::IDList > m_in_port_idlist;
	PushConsumer< Ubitrack::Measurement::Rotation > m_in_port_rotation;
	PushSupplier< Ubitrack::Measurement::Pose > m_out_port_pose;
	PushSupplier< Ubitrack::Measurement::ErrorPose > m_out_port_errorpose; // TODO: possible extension...

private:
	RFID::RFIDTracker* m_rfid_tracker;
	RFID::RFIDDataBase* m_rfid_database;
	Math::Pose m_pose;
	Ubitrack::Measurement::Rotation m_rot;
};

UBITRACK_REGISTER_COMPONENT( ComponentFactory* const cf ) {
	cf->registerComponent< RFIDTrackerModule > ( "RFIDTracker" );
}


} } // namespace Ubitrack::Drivers
