/*
 * File: SerialDriver.cpp
 * Author: SPAIK
 *
 * Implements a very dumbed down serial driver.  Thus, this driver is not meant
 *  to be general purpose or complete.
 *
 * 22 Apr 2004: Created.
 *
 * Copyright ThingMagic LLC 2004
 */


#include <stdio.h>
#include <string.h>
#include <time.h>

// My include files
#include "SerialDriver.h"


/** @fn SerialDriver::SerialDriver(void)
  * @brief Constructor for Serial Driver object.
  *
  * @return None.
  */
SerialDriver::SerialDriver(void)
{
  // Initialize the member variables
  m_hSerialPort     = NULL;
  m_portNum         = 0;
  m_baudRate        = 0;
  m_portOpened      = FALSE;
}


/** @fn SerialDriver::~SerialDriver(void)
  * @brief Destructor for Serial Driver object.
  *
  * @return None.
  */
SerialDriver::~SerialDriver(void)
{
  // Make sure the port is closed before destroying the object.
  ClosePort();
}


/** @fn bool SerialDriver::OpenPort(u8 portNum, u32 baudRate)
  * @brief Opens the serial port and configures it
  *
  * @param portNum  COM port number to open.  I.e. 1 for COM1
  * @param baudRate Baud rate to use.  I.e. 9600
  * @return true if serial port opened properlyl.
  */
bool SerialDriver::OpenPort(u8 portNum, u32 baudRate)
{
  char  portName[8];
  char  portSetting[64];
  DCB   dcb;


  if(m_portOpened == TRUE)
  {
    // The port is already being used.  No need to open another one!
    //  Return FALSE to alert the caller that something is not right.
    return(FALSE);
  }

  // Open the serial port
  {
    sprintf(portName, "COM%d", portNum);
	printf("foo: %s\n",portName);

    m_hSerialPort = CreateFile(portName,                      // Filename.
                               GENERIC_READ | GENERIC_WRITE,  // Desired access.
                               0,                             // Device isn't shared.
                               NULL,                          // Get default security.
                               OPEN_EXISTING,                 // Specify action to take.
                               0,                             // Flags and Atrributes.
                               NULL);                         // Template file.

    if(m_hSerialPort == NULL)
    {
      // Unable to open this serial port.
      return(FALSE);
    }
  }

  // Setup the COMMTIMEOUTS
  {
    COMMTIMEOUTS  timeOuts;

    timeOuts.ReadIntervalTimeout          = 0xFFFFFFFF;
    timeOuts.ReadTotalTimeoutConstant     = 0;
    timeOuts.ReadTotalTimeoutMultiplier   = 0;
    timeOuts.WriteTotalTimeoutConstant    = 5000;
    timeOuts.WriteTotalTimeoutMultiplier  = 0;

    SetCommTimeouts(m_hSerialPort, &timeOuts);
  }

  // Configure the serial port
  {
    sprintf(portSetting, "COM%d:%d,n,8,1", portNum, baudRate);
    
    dcb.DCBlength = sizeof(DCB);
    GetCommState(m_hSerialPort, &dcb);
    dcb.BaudRate  = baudRate;
    dcb.ByteSize  = 8;
    dcb.fRtsControl = RTS_CONTROL_DISABLE;

    if(!SetCommState(m_hSerialPort, &dcb) ||
       !SetupComm(m_hSerialPort, 10000, 10000))
    {
      CloseHandle(m_hSerialPort);
      return(FALSE);
    }
  }

  // The serial port was opened successfully!
  m_portNum         = portNum;
  m_baudRate        = baudRate;
  m_portOpened      = TRUE;

  return(TRUE);
}


/** @fn bool SerialDriver::ClosePort(void)
  * @brief Closes the serial port.
  * 
  * @return Always returns true.
  */
bool SerialDriver::ClosePort(void)
{
  if(m_hSerialPort != NULL)
  {
    CloseHandle(m_hSerialPort);
  }

  // Set member values to benign state.
  m_hSerialPort     = NULL;
  m_portNum         = 0;
  m_baudRate        = 0;
  m_portOpened      = FALSE;

	return(TRUE);
}


/** @fn s32 SerialDriver::ReadData(void *readBuf, u32 maxNumBytesToRead)
  * @brief Reads a stream of data from the serial port.
  *
  * @param readBuf            Copies data from serial port into this buffer.
  * @param maxNumBytesToRead  The size of readBuf buffer in bytes.
  * @return Number of bytes read.
  */
s32 SerialDriver::ReadData(void *readBuf, u32 maxNumBytesToRead)
{
  BOOL    readStatus;
  DWORD   bytesRead;
  DWORD   errorFlags;
  COMSTAT comStat;


  if((m_portOpened == FALSE) || (m_hSerialPort == NULL))
  {
    // The port is not open, exit.
    return(0);
  }

  ClearCommError(m_hSerialPort, &errorFlags, &comStat);
  if(!comStat.cbInQue)
  {
    // No data to read!
    return(0);
  }

  if(comStat.cbInQue < maxNumBytesToRead)
  {
    bytesRead = comStat.cbInQue;
  }
  else
  {
    bytesRead = maxNumBytesToRead;
  }

  readStatus = ReadFile(m_hSerialPort, readBuf, bytesRead, &bytesRead, NULL);

  return(bytesRead);
}


/** @fn s32 SerialDriver::SendData(const char *sendBuf, u32 numBytesToSend)
  * @brief Sends data to the serial port.
  * 
  * @param sendBuf          Buffer to send to serial port.
  * @param numBytesToSend   Size of sendBuf to send.
  * @return Number of bytes actually sent to serial port.
  */
s32 SerialDriver::SendData(const char *sendBuf, u32 numBytesToSend)
{
  DWORD numBytesWritten;

  // Check to see if the port is open
  if((m_portOpened != TRUE) || (m_hSerialPort == NULL))
  {
    // The port is not open, exit.
    printf("Serial port was not open, could not SendData().");
    return(0);
  }

  WriteFile(m_hSerialPort, sendBuf, numBytesToSend, &numBytesWritten, NULL);
//  printf("sent: ");
//  for (int i=0; i<numBytesWritten; ++i) {
//	  printf("%02X ", sendBuf[i]);
//  }
//  printf("\n");

  return((s32)numBytesWritten);
}


/** @fn s32 SerialDriver::NumBytesInReadFifo(void)
  * @brief Checks the COM port read FIFO for received characters.
  *
  * @return Number of characters in the read FIFO.
  */
s32 SerialDriver::NumBytesInReadFifo(void)
{
  DWORD   errorFlags;
  COMSTAT comStat;

  // Check to see if the port is open
  if((m_portOpened != TRUE) || (m_hSerialPort == NULL))
  {
    // The port is not open, exit.
    return(0);
  }

  ClearCommError(m_hSerialPort, &errorFlags, &comStat);

  return(comStat.cbInQue);
}


/** @fn void SerialDriver::ClearSerialPort(void)
  * @brief Clears the serial port FIFO.  
  *
  * @param void
  * @return none
  */
void SerialDriver::ClearSerialPort(void)
{
  char buf[1000];

  while(NumBytesInReadFifo() != 0)
  {
    // Read serial port data
    ReadData(buf, 1000);
  }
}


