/*
 * Ubitrack - Library for Ubiquitous Tracking
 * Copyright 2006, Technische Universitaet Muenchen, and individual
 * contributors as indicated by the @authors tag. See the
 * copyright.txt in the distribution for a full listing of individual
 * contributors.
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA, or see the FSF site: http://www.fsf.org.
 */


/**
 * @ingroup driver_components
 * @file
 * Xsens driver
 * This file contains the driver component to
 * get rotation data from an Xsens MT9-B.
 * On Linux, it uses the MotionTracker library provided by Xsens.
 *
 * The received data is sent via a push interface.
 *
 * @author Florian Echtler <echtler@in.tum.de>
 */

#include <string>
#include <cstdlib>
#include <log4cpp/Category.hh>

#include <Ubitrack/Dataflow/PushSupplier.h>
#include <Ubitrack/Dataflow/ComponentFactory.h>
#include <Ubitrack/Dataflow/Component.h>
#include <Ubitrack/Measurement/Measurement.h>
#include <Ubitrack/Measurement/TimestampSync.h>
#include <Ubitrack/Util/Exception.h>
#include <Ubitrack/Util/OS.h>


#include "m5eReader.h"

#include <iostream>

static log4cpp::Category& logger( log4cpp::Category::getInstance( "Drivers.m5eModule" ) );

namespace Ubitrack { namespace Drivers {

using namespace Dataflow;

/**
 * Module for XSens tracker, should work on both linux and windows.
 * Requires the XSens ActiveX plugin to be installed on Windows.
 */
class m5eModule
	: public Component
{
public:

	/** constructor */
	m5eModule( const std::string& sName, boost::shared_ptr< Graph::UTQLSubgraph > subgraph )
		: Component( sName )
		, m_port( "Output", *this )
		, m_triggerPort( "Trigger", *this )
		, m_keepRunning(1)
	{
		power[0] = 2650; //450mw
		power[1] = 2769; //587.5mw
		power[2] = 2860; //725mw
		power[3] = 2936; //862.5mw
		power[4] = 3000; //1W

		int port;
		subgraph->m_DataflowAttributes.getAttributeData( "port", port );

		unsigned int mode;
		subgraph->m_DataflowAttributes.getAttributeData( "mode", mode );
		if(mode > 3)
			UBITRACK_THROW( "RFID reader does not support such a mode" );

		switch (mode)
		{
		case 1:
			m_m5e = new M5eReader(port,Europe ,baud_9600,port_1,port_1); // RX, TX
			break;
		case 2:
			m_m5e = new M5eReader(port,Europe ,baud_9600,port_2,port_2); // RX, TX
			break;
		case 3:
			m_m5e = new M5eReader(port,Europe ,baud_9600,port_2,port_1); // RX, TX
			break;
		}

		
		// m5e.SetReadTXPower(3000); // Max Powers :>

		// start our own polling thread
		m_thread.reset( new boost::thread( boost::bind( &m5eModule::threadHandler, this ) ) );
	}

	/** destructor */
	~m5eModule()
	{
		m_keepRunning = 0;
		m_thread->join();
		Util::sleep(100);
		delete m_m5e;
		
		LOG4CPP_DEBUG( logger, "m5eModule stopped" );
	}

	/** thread method */
	void threadHandler()
	{
		LOG4CPP_DEBUG( logger, "m5eModule thread started" );
		std::vector < Math::Scalar< unsigned long > > tagIDs;

		//Measurement::TimestampSync syncer( 1.0 ); // timestamps are in seconds
		
		float fData[11] = {0};
		short nNew = 0;
		std::cout<<std::endl<<"START:"<<std::endl<<std::endl;


		for(int i = 0; m_keepRunning; i++)
		{
			tagIDs.clear();
			Util::sleep( 100 ); //			std::cerr << "Using " << power[i%5] << " cdBm" << std::endl;
			m_m5e->SetReadTXPower( power[i%5] ); // Max Powers :>
			tagIDs.push_back( power[i%5] );
			//m_m5e->PrintVisibleTags( 250 );
			m_m5e->GetVisibleTags( tagIDs, 150);

//			std::cerr << power[i%5] << " | "
//				;
//			for(int k = 1; k < tagIDs.size(); k++)
//				std::cerr << "****** ";
//			std::cerr << std::endl;
			
			Measurement::IDList package(Measurement::now() , tagIDs );
			m_port.send(package);
			m_triggerPort.send( Measurement::Button(Measurement::now(), int('a')) ); 
		}
		LOG4CPP_DEBUG( logger, "m5eModule thread started" );
	}
protected:

	boost::shared_ptr< boost::thread > m_thread;
	PushSupplier< Ubitrack::Measurement::IDList > m_port;
	PushSupplier< Ubitrack::Measurement::Button > m_triggerPort;

	M5eReader* m_m5e;
	int m_keepRunning;
	int power[5];
};


UBITRACK_REGISTER_COMPONENT( ComponentFactory* const cf ) {
	cf->registerComponent< m5eModule > ( "m5eModule" );
}


} } // namespace Ubitrack::Drivers
