#include "m5eReader.h"
#include "SerialDriver.h"


#include <iostream>
#include "m5eCom.h"

// OPCODES:

#define MSG_OPCODE_START_BOOTLOADER 0x09
#define MSG_OPCODE_BOOT_FIRMWARE 0x04
#define MSG_OPCODE_SET_BAUDRATE 0x06

#define MSG_OPCODE_GET_VERSION 0x03
#define MSG_OPCODE_GET_SILICON_ID 0x0B
#define MSG_OPCODE_GET_AVAILABLE_REGIONS 0x71
#define MSG_OPCODE_GET_GPIO_INPUTS 0x66

#define MSG_OPCODE_GET_AVAILABLE_PROTOCOLS 0x70
#define MSG_OPCODE_SET_TAG_PROTOCOL 0x93
#define MSG_OPCODE_GET_TAG_PROTOCOL 0x63

#define MSG_OPCODE_GET_ANTENNA_PORT 0x61
#define MSG_OPCODE_SET_ANTENNA_PORT 0x91

#define MSG_OPCODE_GET_READ_TX_POWER 0x62
#define MSG_OPCODE_SET_READ_TX_POWER 0x92

#define MSG_OPCODE_GET_WRITE_TX_POWER 0x64

#define MSG_OPCODE_SET_REGION 0x97
#define MSG_OPCODE_GET_REGION 0x67


#define MSG_OPCODE_CLEAR_TAG_ID_BUFFER 0x2A
#define MSG_OPCODE_READ_MULTIPLE_TAGS 0x22
#define MSG_OPCODE_GET_TAG_ID_BUFFER 0x29



M5eReader::M5eReader(int comPort, Region region, BaudRate baud, AntennaPort antennaRXPort, AntennaPort antennaTXPort)
{
	_m5eCom = 0;

	_readerParameter.region = region;
	_readerParameter.baudrate = baud;
	_readerParameter.antennaRXPort = antennaRXPort;
	_readerParameter.antennaTXPort = antennaTXPort;


	_msg.rx = new Message();
	_msg.tx = new Message();
	_msg.tx->dataLen = 0;

	bool init = true;
	switch (baud)	// Open the serial port
	{
	case baud_19200:
		_m5eCom = new M5eCom(comPort, 19200);
		break;
	case baud_38400:
		_m5eCom = new M5eCom(comPort, 38400);
		break;
	case baud_57600:
		_m5eCom = new M5eCom(comPort, 57600);
		break;
	case baud_115200:
		_m5eCom = new M5eCom(comPort, 115200);
		break;
	case baud_230400:
		_m5eCom = new M5eCom(comPort, 230400);
		break;
	case baud_460800:
		_m5eCom = new M5eCom(comPort, 460800);
		break;
	default:
		baud = baud_9600;
		_m5eCom = new M5eCom(comPort, 9600);
	}
		
	if(init)
		init = this->StartBootloader();
	if(init)
		init = this->BootFirmware();
	if(init)
		init = this->SetBaudrate(baud);
	if(init)
		init = this->SetAntennaPort(_readerParameter.antennaRXPort,_readerParameter.antennaTXPort);
	if(init)
		init = this->SetRegion(_readerParameter.region);
	if(init)
		init = this->SetTagProtocol(EPC_GEN2);
	if(init)
		init = this->SetReadTXPower(2500);
	
	if(!init)
	{
		std::cout<<"ERROR:  m5e reader initialization failed!"<<std::endl;
		delete _m5eCom;
		_m5eCom = 0;
		return;
	}
}

M5eReader::~M5eReader()
{
	delete _msg.rx;
	delete _msg.tx;
	delete _m5eCom;
}

bool M5eReader::StartBootloader()
{
	// start bootloader
	_msg.tx->opCode   = MSG_OPCODE_START_BOOTLOADER;
	_m5eCom->sendMessage(_msg.tx);
	Sleep(100);		// give bootloader time to initialize
	_m5eCom->receiveMessage(_msg.rx);
	if(!_m5eCom->checkOpcode(_msg.tx,_msg.rx))
	{
		std::cout << "Error! Starting m5e bootloader failed" << std::endl;
		return false;
	}
	if((_msg.rx->status[0] == 0x00) && (_msg.rx->status[1] == 0x00))
	{
		std::cout << "m5e bootloader started" << std::endl;
	}
	else if	((_msg.rx->status[0] == 0x01) && (_msg.rx->status[1] == 0x01))
	{
		std::cout << "m5e bootloader was already started" << std::endl;
	}
	else
	{
		std::cout << "Error! Starting m5e bootloader failed" << std::endl;
		return false;
	}
	
	return true;
}

bool M5eReader::BootFirmware()
{
	// boot firmware
	_msg.tx->opCode   = MSG_OPCODE_BOOT_FIRMWARE;
	if(!_m5eCom->secureCom(_msg.tx,_msg.rx,1000))
		return false;

	std::cout << "m5e firmware started" << std::endl;
	Sleep(100);		// give firmware time to initialize
	return true;
}

bool M5eReader::SetBaudrate(BaudRate baud)
{
	// set baudrate
	_msg.tx->opCode  = MSG_OPCODE_SET_BAUDRATE;
	_msg.tx->data[0] = baud / 256;
	_msg.tx->data[1] = baud;
	_msg.tx->dataLen = 2;

	if(!_m5eCom->secureCom(_msg.tx,_msg.rx))
		return false;

	_readerParameter.baudrate = baud;

	return true;
}

bool M5eReader::SetAntennaPort(AntennaPort antennaRXPort, AntennaPort antennaTXPort)
{
	// set antenna ports
	_msg.tx->opCode   = MSG_OPCODE_SET_ANTENNA_PORT;
	_msg.tx->data[0] = antennaTXPort;
	_msg.tx->data[1] = antennaRXPort;
	_msg.tx->dataLen  = 2;

	if(!_m5eCom->secureCom(_msg.tx,_msg.rx))
		return false;

	// get antenna ports
	_msg.tx->opCode   = MSG_OPCODE_GET_ANTENNA_PORT;
	_msg.tx->dataLen  = 0;
    
	if(!_m5eCom->secureCom(_msg.tx,_msg.rx))
		return false;

	_readerParameter.antennaTXPort = (AntennaPort)_msg.rx->data[0];
	_readerParameter.antennaRXPort = (AntennaPort)_msg.rx->data[1];

	return true;
}

bool M5eReader::SetRegion(Region region)
{
	// set region
	_msg.tx->opCode   = MSG_OPCODE_SET_REGION;
	_msg.tx->data[0]  = region;
	_msg.tx->dataLen  = 1;

	if(!_m5eCom->secureCom(_msg.tx,_msg.rx))
		return false;

	// get region
	_msg.tx->opCode   = MSG_OPCODE_GET_REGION;
	_msg.tx->dataLen  = 0;
    
	if(!_m5eCom->secureCom(_msg.tx,_msg.rx))
		return false;

	_readerParameter.region = (Region)_msg.rx->data[0];

	return true;
}

bool M5eReader::SetTagProtocol(int protocol)
{
	// set protocol
	_msg.tx->opCode   = MSG_OPCODE_SET_TAG_PROTOCOL;
	_msg.tx->data[0]  = 0;
	_msg.tx->data[1]  = protocol;
	_msg.tx->dataLen  = 2;

	if(!_m5eCom->secureCom(_msg.tx,_msg.rx))
		return false;

	// get region
	_msg.tx->opCode   = MSG_OPCODE_GET_TAG_PROTOCOL;
	_msg.tx->dataLen  = 0;
    
	if(!_m5eCom->secureCom(_msg.tx,_msg.rx))
		return false;

	_readerParameter.protocol = _msg.rx->data[0];

	return true;
}


bool M5eReader::SetReadTXPower(int centi_dBm)
{
	// set protocol
	_msg.tx->opCode   = MSG_OPCODE_SET_READ_TX_POWER;
	_msg.tx->data[0]  = centi_dBm/256;
	_msg.tx->data[1]  = centi_dBm;
	_msg.tx->dataLen  = 2;

	if(!_m5eCom->secureCom(_msg.tx,_msg.rx))
		return false;

	// get region
	_msg.tx->opCode   = MSG_OPCODE_GET_READ_TX_POWER;
	_msg.tx->dataLen  = 0;
    
	if(!_m5eCom->secureCom(_msg.tx,_msg.rx))
		return false;

	_readerParameter.power = _msg.rx->data[0]*256 + _msg.rx->data[1];

	return true;
}




bool M5eReader::PrintVisibleTags(int timeout)
{
	if(!_m5eCom)
		return false;

	// clear buffer
	_msg.tx->opCode   = MSG_OPCODE_CLEAR_TAG_ID_BUFFER;
	_msg.tx->dataLen  = 0;

	if(!_m5eCom->secureCom(_msg.tx,_msg.rx))
		return false;
	Sleep(10);
	
	// read multiple tags
	_msg.tx->opCode   = MSG_OPCODE_READ_MULTIPLE_TAGS;
	_msg.tx->data[0]  = timeout/256;
	_msg.tx->data[1]  = timeout;
	_msg.tx->dataLen  = 2;

	if(!_m5eCom->secureCom(_msg.tx,_msg.rx,timeout+32,60))
	{
		if	((_msg.rx->status[0] == 0x04) && (_msg.rx->status[1] == 0x00))
		{
			std::cout << "No tags were found!" << std::endl;
			return false;
		}
		std::cout<<"ERROR reading tags!"<<std::endl;
		return false;
	}

	int numberOfTags = _msg.rx->data[0];

	// get tags from the buffer:
	_msg.tx->opCode   = MSG_OPCODE_GET_TAG_ID_BUFFER;
	_msg.tx->data[0]  = 0;
	_msg.tx->data[1]  = 0;
	_msg.tx->data[2]  = numberOfTags/256;
	_msg.tx->data[3]  = numberOfTags;
	_msg.tx->dataLen  = 4;

	if(!_m5eCom->secureCom(_msg.tx,_msg.rx))
		return false;

	std::cout<<" <"<<numberOfTags<<"> ";
	for(int i = 0; i < numberOfTags; i++)
	{
		for(int j = 11; j < 12; j++)
			printf("%02X ",_msg.rx->data[i*(   2		//bitsize of tag id
											+  2		//junk
											+ 12		//tag id
											+  2)		//crc of tag id
											+  4		//bitsize+junk
											+ j]);
		std::cout<<" | ";
	}	
	std::cout<<std::endl;
	return true;
}

bool M5eReader::GetVisibleTags(std::vector<Ubitrack::Math::Scalar<unsigned long> >& tagIDs,int timeout)
{
	if(!_m5eCom)
		return false;

	// clear buffer
	_msg.tx->opCode   = MSG_OPCODE_CLEAR_TAG_ID_BUFFER;
	_msg.tx->dataLen  = 0;

	if(!_m5eCom->secureCom(_msg.tx,_msg.rx))
		return false;
	Sleep(10);
	
	// read multiple tags
	_msg.tx->opCode   = MSG_OPCODE_READ_MULTIPLE_TAGS;
	_msg.tx->data[0]  = timeout/256;
	_msg.tx->data[1]  = timeout;
	_msg.tx->dataLen  = 2;

	if(!_m5eCom->secureCom(_msg.tx,_msg.rx,timeout+32,60))
	{
		if	((_msg.rx->status[0] == 0x04) && (_msg.rx->status[1] == 0x00))
		{
		//	std::cout << "No tags were found!" << std::endl;
			return false;
		}
		std::cout<<"ERROR reading tags!"<<std::endl;
		return false;
	}

	int numberOfTags = _msg.rx->data[0];

	// get tags from the buffer:
	_msg.tx->opCode   = MSG_OPCODE_GET_TAG_ID_BUFFER;
	_msg.tx->data[0]  = 0;
	_msg.tx->data[1]  = 0;
	_msg.tx->data[2]  = numberOfTags/256;
	_msg.tx->data[3]  = numberOfTags;
	_msg.tx->dataLen  = 4;

	if(!_m5eCom->secureCom(_msg.tx,_msg.rx))
		return false;

	for(int i = 0; i < numberOfTags; i++)
	{
		int index = i*(		2		//bitsize of tag id
						+	2		//junk
						+	12		//tag id
						+	2)		//crc of tag id
						+	4;		//bitsize+junk

		unsigned long tagID = 0;
		for(int j = 8; j < 12; j++)
		{
			tagID |= ( _msg.rx->data[index+j] ) << ( (11-j) * 8 );
		}
		tagIDs.push_back(tagID);
	}	
	return true;
}
