/*
 * Ubitrack - Library for Ubiquitous Tracking
 * Copyright 2006, Technische Universitaet Muenchen, and individual
 * contributors as indicated by the @authors tag. See the
 * copyright.txt in the distribution for a full listing of individual
 * contributors.
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA, or see the FSF site: http://www.fsf.org.
 */


/**
 * @ingroup driver_components
 * @file
 * RBCalibration driver
 * This file contains the driver component to
 * get tracking data from an RBCalibrationer.
 *
 * The received data is sent via a push interface.
 *
 * @author Daniel Muhra <muhra@in.tum.de>
 */

#ifndef __RBCALIBRATION_H_INCLUDED__
#define __RBCALIBRATION_H_INCLUDED__

#include <string>
#include <boost/bind.hpp>
#include <log4cpp/Category.hh>
#include <Ubitrack/Util/Exception.h>
#include <Ubitrack/Util/Logging.h>

#include <Ubitrack/Dataflow/PushConsumer.h>
#include <Ubitrack/Dataflow/PushSupplier.h>
#include <Ubitrack/Dataflow/ComponentFactory.h>
#include <Ubitrack/Dataflow/Module.h>
#include <Ubitrack/Dataflow/Component.h>

#include <Ubitrack/Util/Exception.h>
#include <Ubitrack/Util/OS.h>

#include <Ubitrack/Calibration/FundamentalMatrix.h>
#include <Ubitrack/Math/MatrixOperations.h>
#include <Ubitrack/Calibration/3DPointReconstruction.h>
#include "ShowTrace.h"

namespace Ubitrack { namespace Drivers {

// forward declarations
class RBCalibration;
class RBCalibrationModule;

typedef Dataflow::Module< Dataflow::SingleModuleKey, Dataflow::SubgraphIdKey, RBCalibrationModule, RBCalibration > ModuleBaseC;

class RBCalibrationModule
	: public ModuleBaseC
{
public:
	/** constructor, creates thread */
	RBCalibrationModule( const Dataflow::SingleModuleKey& key, boost::shared_ptr< Graph::UTQLSubgraph >, FactoryHelper* pFactory );
		
	/** creates a cam */
	void createCam( double eMax, unsigned int nMin, long long tMax );	

	/** destroys a cam */
	void destroyCam();

	/** destructor, stops thread */
	~RBCalibrationModule();

	void startCalibration();

	double validateFundamentalMatrix( const Math::Matrix< 3, 3 > & fM, const std::vector< Math::Vector< 2 > > & fromPoints, const std::vector< Math::Vector< 2 > > & toPoints, const unsigned int step );

	void findCorrespondences( const TraceContainer & from, const TraceContainer & to, std::vector< Math::Vector< 2 > > & fromPoints, std::vector< Math::Vector< 2 > > & toPoints );

	Math::Matrix< 3, 3 > findFM( const std::map< int, TraceContainer > & firstCam, const std::map< int, TraceContainer > & secondCam );

protected:
	bool m_ok;

	//matching points
	std::vector< Math::Vector< 2 > > m_fromPoints;
	std::vector< Math::Vector< 2 > > m_toPoints;

	//number of cameras in use
	int m_nCams;

	//maximum accepted error
	double m_eMax;
	
	//minimum number of trace points
	unsigned int m_nMin;

	//maximum timestamp difference
	long long m_tMax;

	//list of Componenst
	ComponentList comps;
};

/**
 * @ingroup driver_components
 * A Component for RBCalibration.
 *
 * @par Input Ports
 * PushConsumer< Ubitrack::Measurement::PositionList2 > port with name "Input".
 *
 * @par Configuration
 * \c movement: minimum number of pixels the point has to move to be used.
 * \c number: minimum number of trace points. It has to be greater than 8.
 * \c time: maximum time difference between two views in ms.
 * \c button: button to start the calibration.
 *
 * @par Example Configuration
 * \verbatim
<Pattern name="RBCalibration" id="RBCalibration1">
	<Input>
		<Node name="Camera" id="Camera"/>
		<Node name="ImagePlane" id="ImagePlane"/>
		<Edge name="Input" source="Camera" destination="ImagePlane" pattern-ref="..." edge-ref="..."/>
	</Input>

	<DataflowConfiguration>
		<UbitrackLib class="RBCalibration"/>
		<Attribute name="movement" value="15"/>
		<Attribute name="number" value="30"/>
		<Attribute name="time" value="18"/>
		<Attribute name="error" value="10"/>
		<Attribute name="button" value="a"/>
	</DataflowConfiguration>
</Pattern>
\endverbatim
 */

class RBCalibration
	: public RBCalibrationModule::Component
{
public:
	/** constructor, creates camera */
	RBCalibration( const std::string& sName, boost::shared_ptr< Graph::UTQLSubgraph > pCfg, const Dataflow::SubgraphIdKey& componentKey, RBCalibrationModule* pModule );

	/** destructor, destroys window */
	~RBCalibration();

	void loop( const Measurement::PositionList2 & data );

	//delete closed up traces with no movement
	void cleanUp();
	
	/** method that receives coordinates of 2D Points found in the image */
	void enlist(  Measurement::PositionList2 & data );

	/** start calibration of the module */
	void calibrate( const Measurement::Button& b );

	std::map< int, TraceContainer > & getList();

	Math::Matrix< 3, 3 > & getIntrinsics();

	void setPose( Math::Pose & pose );

	/** returns the projection matrix */
	Math::Matrix< 3, 4 > & getPM();

protected:
	//current data
	Measurement::PositionList2 m_data;

	//class for trace tracking
	Calibration::ShowTrace sT;

	//active flag
	bool m_active;

	//Pose
	Math::Pose m_pose;

	//Intrinsic matrix
	Math::Matrix< 3, 3 > m_intrinsics;

	//minimum distance
	double m_dMin;

	//maximum distance
	double m_dMax;

	//minimum number of trace points
	unsigned int m_nMin;

	//the button to start the calibration
	Math::Scalar< int > m_button;

	//the point traces
	std::map< int, TraceContainer > m_traceLists;

	// the ports
	Dataflow::PushConsumer< Measurement::PositionList2 > m_inPort;
	Dataflow::PushSupplier< Measurement::Pose > m_outPort;
	Dataflow::PushSupplier< Vision::ImageMeasurement > m_debugPort;

	//the signal port
	Dataflow::PushConsumer< Measurement::Button > m_SignalPort;
};	

} } // namespace Ubitrack::Driver

#endif
