/*
 * Ubitrack - Library for Ubiquitous Tracking
 * Copyright 2006, Technische Universitaet Muenchen, and individual
 * contributors as indicated by the @authors tag. See the
 * copyright.txt in the distribution for a full listing of individual
 * contributors.
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA, or see the FSF site: http://www.fsf.org.
 */

#include "RBCalibration.h"
#include <iostream>

static log4cpp::Category& logger( log4cpp::Category::getInstance( "Drivers.RigidBodyTracking" ) );

using namespace Ubitrack;
using namespace Ubitrack::Dataflow;

// shortcuts to namespaces
namespace ublas = boost::numeric::ublas;

//
namespace Ubitrack { namespace Drivers {

/**
 * @ingroup driver_components
 * A Component for RBCalibration.
 *
 * @par Input Ports
 * PushConsumer< Ubitrack::Measurement::PositionList > port with name "Input", 2D Points with ID as z coordinate.
 *
 * @par Configuration
 * \c movement: minimum number of pixels the point has to move to be used.
 * \c number: minimum number of trace points. It has to be greater than 8.
 * \c time: maximum time difference between two views in ms.
 * \c button: button to start the calibration.
 *
 * @par Example Configuration
 * \verbatim
<Pattern name="RBCalibration" id="RBCalibration1">
	<Input>
		<Node name="Camera" id="Camera"/>
		<Node name="ImagePlane" id="ImagePlane"/>
		<Edge name="Input" source="Camera" destination="ImagePlane" pattern-ref="..." edge-ref="..."/>
	</Input>

	<DataflowConfiguration>
		<UbitrackLib class="RBCalibration"/>
		<Attribute name="min_dist" value="15"/>
		<Attribute name="max_dist" value="35"/>
		<Attribute name="movement" value="15"/>
		<Attribute name="number" value="30"/>
		<Attribute name="time" value="18"/>
		<Attribute name="error" value="10"/>
		<Attribute name="button" value="a"/>
	</DataflowConfiguration>
</Pattern>
\endverbatim
 */

RBCalibration::RBCalibration( const std::string& sName, boost::shared_ptr< Graph::UTQLSubgraph > pCfg, const Dataflow::SubgraphIdKey& componentKey, RBCalibrationModule* pModule )
	: RBCalibrationModule::Component( sName, componentKey, pModule )
	, m_inPort( "Input", *this, boost::bind( &RBCalibration::loop, this, _1 ) )
	, m_SignalPort( "Signal", *this, boost::bind( &RBCalibration::calibrate, this, _1 ) )
	, m_outPort( "Output", *this )
	, m_debugPort( "Image", *this )
	, m_active( true )
	, m_dMin( 0 )
	, m_dMax( 0 )
	, m_nMin( 0 )
	, m_button( ' ' )
{
	double eMax;	
	long long tMax;

	char button = ' ';
	pCfg->m_DataflowAttributes.getAttributeData( "min_distance", m_dMin );
	pCfg->m_DataflowAttributes.getAttributeData( "max_distance", m_dMax );
	pCfg->m_DataflowAttributes.getAttributeData( "number", m_nMin );
	pCfg->m_DataflowAttributes.getAttributeData( "error", eMax );
	pCfg->m_DataflowAttributes.getAttributeData( "time", tMax );
	pCfg->m_DataflowAttributes.getAttributeData( "button", button );

	sT = Calibration::ShowTrace( m_dMin, m_dMax );

	m_button = Math::Scalar< int >( button );
	tMax *= 1000000;
	getModule().createCam( eMax, m_nMin, tMax );

	int m_width = 356;	// 356
	int m_height = 291; //291

	m_intrinsics( 0, 0 ) = 1.25 * double( m_width );
	m_intrinsics( 0, 1 ) = 0.0;
	m_intrinsics( 0, 2 ) = double( m_width ) / -2.0;
	m_intrinsics( 1, 0 ) = 0.0;
	m_intrinsics( 1, 1 ) = 1.25 * double( m_width );
	m_intrinsics( 1, 2 ) = double( m_height )  / -2.0;
	m_intrinsics( 2, 0 ) = 0.0;
	m_intrinsics( 2, 1 ) = 0.0;
	m_intrinsics( 2, 2 ) = -1.0;
}

/** destructor, destroys window */
RBCalibration::~RBCalibration()
{}

void RBCalibration::loop( const Measurement::PositionList2 & data )
{
	m_data = Measurement::PositionList2( data.time(), Measurement::PositionList2::value_type() );
	
	for( unsigned int i=0; i< data->size(); i++ )
	{
		m_data->push_back( Math::Vector< 2 >( data->at( i ) ) ); // /100
	}
		
	if( m_active )
		enlist( m_data );
}

//delete closed up traces with no movement
void RBCalibration::cleanUp()
{
	for( std::map< int, TraceContainer >::iterator it = m_traceLists.begin(); it != m_traceLists.end(); )
	{
		std::map< int, TraceContainer >::iterator save = it;
		save++;
		if( it->second.getMovement() < m_dMin || it->second.size() < m_nMin )
		{
			m_traceLists.erase( it );
		}
		it = save;
	}
}

/** method that receives coordinates of 2D Points found in the image */
void RBCalibration::enlist(  Measurement::PositionList2 & data )
{
	sT.enlist( data, m_traceLists );
	m_debugPort.send( sT.sendImage( data.time() ) );
}

/** start calibration of the module */
void RBCalibration::calibrate( const Measurement::Button& b )
{
	if( *b == m_button && m_active )
	{
		cleanUp();
		getModule().startCalibration();
		m_active = false;
	}
	else if( *b == m_button && !m_active )
	{
		m_active = true;
	}
}

std::map< int, TraceContainer > & RBCalibration::getList()
{
	return m_traceLists;
}

Math::Matrix< 3, 3 > & RBCalibration::getIntrinsics()
{
	return m_intrinsics;
}

void RBCalibration::setPose( Math::Pose & pose )
{
	m_pose = pose;
	m_outPort.send( Measurement::Pose( Ubitrack::Measurement::now(), m_pose ) );
}

} } // namespace Ubitrack::Driver