/*
 * Ubitrack - Library for Ubiquitous Tracking
 * Copyright 2006, Technische Universitaet Muenchen, and individual
 * contributors as indicated by the @authors tag. See the
 * copyright.txt in the distribution for a full listing of individual
 * contributors.
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA, or see the FSF site: http://www.fsf.org.
 */


/**
 * @ingroup driver_components
 * @file
 * RBTracking driver
 * This file contains the driver component to
 * get tracking data from an RBTracker.
 *
 * The received data is sent via a push interface.
 *
 * @author Daniel Muhra <muhra@in.tum.de>
 */

#ifndef __RBTRACKING_H_INCLUDED__
#define __RBTRACKING_H_INCLUDED__

#include <string>
#include <boost/bind.hpp>
#include <log4cpp/Category.hh>
#include <Ubitrack/Util/Exception.h>
#include <Ubitrack/Util/Logging.h>

#include <Ubitrack/Dataflow/PushConsumer.h>
#include <Ubitrack/Dataflow/PushSupplier.h>
#include <Ubitrack/Dataflow/ComponentFactory.h>
#include <Ubitrack/Dataflow/Module.h>
#include <Ubitrack/Dataflow/Component.h>

#include <Ubitrack/Util/Exception.h>
#include <Ubitrack/Util/OS.h>

#include <Ubitrack/Calibration/FundamentalMatrix.h>
#include <Ubitrack/Math/MatrixOperations.h>
#include <Ubitrack/Calibration/3DPointReconstruction.h>
#include "Additions.h"

namespace Ubitrack { namespace Drivers {

// forward declarations
class RBTracking;
class RBTrackingModule;

typedef Dataflow::Module< Dataflow::SingleModuleKey, Dataflow::SubgraphIdKey, RBTrackingModule, RBTracking > ModuleBaseT;

class RBTrackingModule
	: public ModuleBaseT
{
public:
	/** constructor, creates thread */
	RBTrackingModule( const Dataflow::SingleModuleKey& key, boost::shared_ptr< Graph::UTQLSubgraph >, FactoryHelper* pFactory );
		
	/** creates a cam */
	void createCam( double e, long long t );	

	/** destroys a cam */
	void destroyCam();

	std::vector< std::pair< unsigned int, unsigned int > > findBody( const std::vector< Ubitrack::Math::Vector< 3 > > & points, const std::vector< Ubitrack::Math::Vector< 3 > > & body, const double & error );

	void compute();

	void sendPoints();

	/** destructor, stops thread */
	~RBTrackingModule();

protected:
	//maximum accepted error
	double m_eMax;
	
	//maximum timestamp difference
	long long m_tMax;

	//number of cameras in use
	int m_nCams;

	//list of Componenst
	ComponentList comps;
};

/**
 * @ingroup driver_components
 * A Component for RBTracking.
 *
 * @par Input Ports
 * PushConsumer< Ubitrack::Measurement::PositionList2 > port with name "Input".
 *
 * @par Configuration
 * \c movement: minimum number of pixels the point has to move to be used.
 * \c number: minimum number of trace points. It has to be greater than 8.
 * \c time: maximum time difference between two views in ms.
 * \c button: button to start the calibration.
 *
 * @par Example Configuration
 * \verbatim
<Pattern name="RBTracking" id="RBTracking1">
	<Input>
		<Node name="Camera" id="Camera"/>
		<Node name="ImagePlane" id="ImagePlane"/>
		<Edge name="Input" source="Camera" destination="ImagePlane" pattern-ref="..." edge-ref="..."/>
	</Input>

	<DataflowConfiguration>
		<UbitrackLib class="RBTracking"/>
		<Attribute name="movement" value="15"/>
		<Attribute name="number" value="30"/>
		<Attribute name="time" value="18"/>
		<Attribute name="error" value="10"/>
		<Attribute name="button" value="a"/>
	</DataflowConfiguration>
</Pattern>
\endverbatim
 */

class RBTracking
	: public RBTrackingModule::Component
{
public:
	/** constructor, creates camera */
	RBTracking( const std::string& sName, boost::shared_ptr< Graph::UTQLSubgraph > pCfg, const Dataflow::SubgraphIdKey& componentKey, RBTrackingModule* pModule );

	/** destructor, destroys window */
	~RBTracking();

	void receive( const Measurement::PositionList2 & data );

	void sendPoints( Measurement::PositionList & list );

	Measurement::PositionList2 & getData();

	void setIntrinsics( const Measurement::Matrix3x3 intrinsics );

	void setPose( const Measurement::Pose pose );

	Math::Matrix< 3, 4 > & getPM();

	Math::Matrix< 3, 3 > & getFM();

protected:
	bool m_active;

	//current data
	Measurement::PositionList2 m_data;

	//fundamental matrix
	Math::Matrix< 3, 3 > m_fM;

	//camera matrix
	Math::Matrix< 3, 3 > m_intrinsics;
	
	//camera pose
	Math::Pose m_pose;
	
	//projection matrix
	Math::Matrix< 3, 4 > m_pM;

	// the ports
	Dataflow::PushConsumer< Measurement::PositionList2 > m_inPort;
	Dataflow::PushSupplier< Measurement::PositionList > m_outPort;
	
	//ports for camera parameters
	Dataflow::PushConsumer< Measurement::Matrix3x3 > m_intrinsicsPort;
	Dataflow::PushConsumer< Measurement::Pose > m_posePort;
};	

} } // namespace Ubitrack::Driver

#endif
