/*
 * Ubitrack - Library for Ubiquitous Tracking
 * Copyright 2006, Technische Universitaet Muenchen, and individual
 * contributors as indicated by the @authors tag. See the
 * copyright.txt in the distribution for a full listing of individual
 * contributors.
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA, or see the FSF site: http://www.fsf.org.
 */

#include "RBTracking.h"
#include <iostream>

static log4cpp::Category& logger( log4cpp::Category::getInstance( "Drivers.RigidBodyTracking" ) );

using namespace Ubitrack;
using namespace Ubitrack::Dataflow;

// shortcuts to namespaces
namespace ublas = boost::numeric::ublas;

//
namespace Ubitrack { namespace Drivers {

/**
 * @ingroup driver_components
 * A Component for RBTracking.
 *
 * @par Input Ports
 * PushConsumer< Ubitrack::Measurement::PositionList2 > port with name "Input".
 *
 * @par Configuration
 * \c movement: minimum number of pixels the point has to move to be used.
 * \c number: minimum number of trace points. It has to be greater than 8.
 * \c time: maximum time difference between two views in ms.
 * \c button: button to start the calibration.
 *
 * @par Example Configuration
 * \verbatim
<Pattern name="RBTracking" id="RBTracking1">
	<Input>
		<Node name="Camera" id="Camera"/>
		<Node name="ImagePlane" id="ImagePlane"/>
		<Edge name="Input" source="Camera" destination="ImagePlane" pattern-ref="..." edge-ref="..."/>
	</Input>

	<DataflowConfiguration>
		<UbitrackLib class="RBTracking"/>
		<Attribute name="min_dist" value="15"/>
		<Attribute name="max_dist" value="35"/>
		<Attribute name="movement" value="15"/>
		<Attribute name="number" value="30"/>
		<Attribute name="time" value="18"/>
		<Attribute name="error" value="10"/>
		<Attribute name="button" value="a"/>
	</DataflowConfiguration>
</Pattern>
\endverbatim
 */

RBTracking::RBTracking( const std::string& sName, boost::shared_ptr< Graph::UTQLSubgraph > pCfg, const Dataflow::SubgraphIdKey& componentKey, RBTrackingModule* pModule )
	: RBTrackingModule::Component( sName, componentKey, pModule )
	, m_inPort( "Input", *this, boost::bind( &RBTracking::receive, this, _1 ) )
	, m_outPort( "Output", *this )
	, m_intrinsicsPort( "Intrinsics", *this, boost::bind( &RBTracking::setIntrinsics, this, _1 ) )
	, m_posePort( "Pose", *this, boost::bind( &RBTracking::setPose, this, _1 ) )
	, m_active( false )
{
	double eMax;
	long long tMax;
	pCfg->m_DataflowAttributes.getAttributeData( "time", tMax );
	pCfg->m_DataflowAttributes.getAttributeData( "error", eMax );

	tMax *= 1000000;
	getModule().createCam( eMax, tMax );
}

/** destructor, destroys window */
RBTracking::~RBTracking()
{}

void RBTracking::receive( const Measurement::PositionList2 & data )
{
	if( m_active )
	{
		m_data = Measurement::PositionList2( data.time(), Measurement::PositionList2::value_type() );
		
		for( unsigned int i=0; i< data->size(); i++ )
		{
			m_data->push_back( Math::Vector< 2 >( data->at( i ) ) );
		}

		getModule().compute();
	}
}

void RBTracking::sendPoints( Measurement::PositionList & list )
{
	m_outPort.send( list );
}

Measurement::PositionList2 & RBTracking::getData()
{
	return m_data;
}

void RBTracking::setIntrinsics( const Measurement::Matrix3x3 intrinsics )
{
	m_intrinsics = *intrinsics ;
	Math::Matrix< 3, 4 > pos( m_pose );
	m_pM = ublas::prod( m_intrinsics, pos );
	m_active = true;
}

void RBTracking::setPose( const Measurement::Pose pose )
{
	m_pose = *pose ;
	Math::Matrix< 3, 4 > pos( m_pose );
	m_pM = ublas::prod( m_intrinsics, pos );
	m_active = true;
}

/** returns the projection matrix */
Math::Matrix< 3, 4 > & RBTracking::getPM()
{
	return m_pM;
}

Math::Matrix< 3, 3 > & RBTracking::getFM()
{
	return m_fM;
}

} } // namespace Ubitrack::Driver