/*
 * Ubitrack - Library for Ubiquitous Tracking
 * Copyright 2006, Technische Universitaet Muenchen, and individual
 * contributors as indicated by the @authors tag. See the 
 * copyright.txt in the distribution for a full listing of individual
 * contributors.
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA, or see the FSF site: http://www.fsf.org.
 */

#include <Ubitrack/Calibration/Munkres.h>

#include "ShowTrace.h"

namespace ublas = boost::numeric::ublas;

namespace Ubitrack { namespace Calibration {

ShowTrace::ShowTrace()
{
	m_first = true;
	m_count = 0;
	m_dMin = 15.0;
	m_dMax = 35.0;

	cvInitFont(&font, CV_FONT_VECTOR0, .4f, .4f, 0, 1);
	blackImage = new Vision::Image( 356, 291, 1);
	char* c;

	for(int i=0; i<356*291; i++)
	{
		c = blackImage->imageData + i;
		*c = 0;
	}	
}

ShowTrace::ShowTrace( double dMin, double dMax )
{
	m_first = true;
	m_count = 0;
	m_dMin = dMin;
	m_dMax = dMax;

	cvInitFont(&font, CV_FONT_VECTOR0, .4f, .4f, 0, 1);
	blackImage = new Vision::Image( 356, 291, 1);
	char* c;

	for(int i=0; i<356*291; i++)
	{
		c = blackImage->imageData + i;
		*c = 0;
	}	
}

/** Method that computes the result. */
void ShowTrace::enlist( Measurement::PositionList2 & data, std::map< int, TraceContainer > & traceList )
{
	if ( m_first )
	{
		m_first = false;
		m_prev = *data;

		for( unsigned int i=0; i < m_prev.size(); i++ )
		{
			m_count++;
			m_matchList.push_back( i );
			m_id.push_back( i );
			m_wasMatched.push_back( true );
		}
	}
		
	unsigned int cSize = ( *data ).size();
	unsigned int pSize = m_prev.size();
	
	ublas::matrix< double > matrix( pSize, cSize );

	for( unsigned int row=0; row<pSize; row++ )
	{
		for( unsigned int col=0; col<cSize; col++ )
		{
			matrix( row, col ) = ublas::norm_2( Math::Vector< 2 >( m_prev.at(row) ) - ( *data ).at(col) );
			if( matrix( row, col ) > m_dMax )
				matrix( row, col ) = m_dMax;
		}
	}
	
	Ubitrack::Calibration::Munkres< double > m( matrix );
	m.solve();
	m_matchList = m.getColMatchList();
			
	Math::Vector< 2 > temp;
	std::vector< Math::Vector< 2 > > curr;
	std::vector< int > id;
	std::vector< bool > wasMatched;

	for( unsigned int i=0; i < m_matchList.size(); i++ )
	{
		unsigned int match = m_matchList.at( i );

		if( i < cSize )
		{
			temp = ( *data ).at( i );
			
			if ( match < pSize )		//match to existing point
			{
				
				
				if( matrix( match, i ) < m_dMax )		//match to possible point
				{
					curr.push_back( temp );
					id.push_back( m_id.at( match ) );
					wasMatched.push_back( true );
				}
				else									//match to point which is too far away -> so, it has to be a new point
				{
					curr.push_back( temp );
					id.push_back( m_count );
					wasMatched.push_back( true );
					m_count++;
					
					if( m_wasMatched.at( match ) )				//give old point a second chance
					{
						curr.push_back( m_prev.at( match ) );
						id.push_back( m_id.at( match ) );
						wasMatched.push_back( true );
					}
				}					
			}
			else						//totaly new point
			{
				temp = ( *data ).at( i );
				curr.push_back( temp );
				id.push_back( m_count );
				wasMatched.push_back( true );
				m_count++;		
			}
		}
		else			//check if point falls out because another point was too close
		{
			double dist = 0.0;
			temp = m_prev.at( m_matchList.at( i ) );
			bool tooClose = false;

			for( unsigned int j=0; j<cSize; j++ )
			{
				dist = ublas::norm_2( temp - ( *data ).at( j ) );
				if( dist < m_dMin )
				{
					unsigned int old = m_matchList.at( j );

					if( i < old )
						m_id[old] = m_count;				//if other point was not already edited change id of old point
					else id[j] = m_count;
					
					tooClose = true;
					m_count++;
				}
			}

			if( !tooClose )
			{
				if( m_wasMatched.at( match ) )				//give a second chance
				{
					curr.push_back( m_prev.at( match ) );
					id.push_back( m_id.at( match ) );
					wasMatched.push_back( false );
				}
			}
		}
	}
	m_id = id;
	m_wasMatched = wasMatched;
	m_prev = curr;

	for( unsigned int i=0; i < cSize; i++)
		traceList[ m_id[i] ].addPoint( Measurement::Position2D( data.time(), ( *data ).at( i ) ) );
}

Vision::ImageMeasurement ShowTrace::sendImage( Measurement::Timestamp t )
{
	boost::shared_ptr< Vision::Image > img = boost::shared_ptr< Vision::Image >( new Vision::Image( cvCloneImage( blackImage ) ) );
	img->origin = 1;
	std::stringstream status;

	for( unsigned int i=0; i < m_id.size(); i++ )
	{
		status.str("");

		status <<  m_id.at( i );
		cvPutText( *img, status.str().c_str(), cvPoint( int( m_prev.at( i )( 0 ) ), int( m_prev.at( i )( 1 ) ) ), &font, CV_RGB( 255, 255, 255 ) );	
	}

	return Vision::ImageMeasurement( t, img );
}
	
} } // namespace Ubitrack::Calibration
