/*
 * Ubitrack - Library for Ubiquitous Tracking
 * Copyright 2006, Technische Universitaet Muenchen, and individual
 * contributors as indicated by the @authors tag. See the
 * copyright.txt in the distribution for a full listing of individual
 * contributors.
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA, or see the FSF site: http://www.fsf.org.
 */

#include "RBTracking.h"

static log4cpp::Category& logger( log4cpp::Category::getInstance( "Drivers.RigidBodyTracking" ) );

using namespace Ubitrack;
using namespace Ubitrack::Dataflow;

// shortcuts to namespaces
namespace ublas = boost::numeric::ublas;

namespace Ubitrack { namespace Drivers {	
	
void clearArray( bool* arr, unsigned int size)
{
	for( unsigned int i=0; i<size; i++ )
		arr[i] = false;
}

std::vector< std::pair< unsigned int, unsigned int > > permutate( std::vector< std::pair< unsigned int, unsigned int > > & match, const boost::numeric::ublas::vector< double > & vP, const boost::numeric::ublas::vector< double > & vB, bool* points, bool* body, const double & error, double & val, const unsigned int count )
{
	std::vector< std::pair< unsigned int, unsigned int > > save;

	if( count > 0 )
	{
		std::vector< std::pair< unsigned int, unsigned int > > temp;		

		//the body entry we're working with
		unsigned int begin = 0;
		while( body[ begin ] )
			begin++;

		double val2 = val;
		body[ begin ] = true;

		save = permutate( match, vP, vB, points, body, error, val2, count - 1 );
		double saveVal = val2;
		
		for( unsigned int i=0; i<vP.size(); i++ )
		{
			if( !points[i] && std::abs( vB( begin ) - vP( i ) ) < error )
			{
				val2 = 	val + std::abs( vB( begin ) - vP( i ) );
				match.push_back( std::pair< unsigned int, unsigned int >( begin, i ) );
				points[ i ] = true;

				temp = permutate( match, vP, vB, points, body, error, val2, count - 1 );
				if( val2 < saveVal )
				{
					save = temp;
					saveVal = val2;
				}

				points[ i ] = false;
				match.pop_back();
			}
		}

		body[ begin ] = false;
		val = saveVal;
	}
	else
	{
		save = match;
		unsigned int sum = 0;
		for( unsigned int i=0; i<vP.size(); i++ )
		{
			if( points[i] )
				sum++;
		}

		val = ( val + error ) / sum;

		//std::cout << "val: "<< val << " size: " << sum << std::endl; 
	}

	return save;
}

bool findEntries( const boost::numeric::ublas::matrix< double > & mP, const boost::numeric::ublas::vector< double > & vB, const unsigned int & bRow, const unsigned int & bCol, std::vector< std::pair< unsigned int, unsigned int > > & bestMatch ,const double & error, double & bestVal )
{	
	//the value we're looking for
	double val = vB( bCol );

	bool found = false;
	
	//to check what was already matched
	bool* points = new bool[ mP.size1() ];
	bool* body = new bool[ vB.size() ];

	clearArray( points, mP.size1() );
	clearArray( body, vB.size() );

	body[ bRow ] = true;
	body[ bCol ] = true;

	if( bRow == bCol )
	{
		int r = 0;
	}

	//to save the temporary matches
		
	for( unsigned int row = 0; row < mP.size1() - 1; row++ )
	{
		for( unsigned int col = row + 1; col < mP.size2(); col++ )
		{
			if( std::abs( val - mP( row, col ) ) < error )
			{
				std::vector< std::pair< unsigned int, unsigned int > > match;
				double err = std::abs( val - mP( row, col ) );
				
				match.push_back( std::pair< unsigned int, unsigned int >( bRow, row ) );
				match.push_back( std::pair< unsigned int, unsigned int >( bCol, col ) );
				
				points[row] = true;				
				points[col] = true;			

				match = permutate( match, ublas::row( mP, row ), vB, points, body, error, err, vB.size() - 2 );
				if( !found )
				{
					bestVal = err;
					bestMatch = match;
					found = true;
				}
				else
					if( err < bestVal )
					{
						bestVal = err;
						bestMatch = match;
					}

				//get back to normal for next values
				points[row] = false;				
				points[col] = false;					
			}			
		}
	}

	return found;
}

std::vector< std::pair< unsigned int, unsigned int > > RBTrackingModule::findBody( const std::vector< Ubitrack::Math::Vector< 3 > > & points, const std::vector< Ubitrack::Math::Vector< 3 > > & body, const double & error )
{
	//the container for the actual matching list
	std::vector< std::pair< unsigned int, unsigned int > > matchList;

	unsigned int pSize = points.size();
	unsigned int bSize = body.size();

	//unsigned int pSize = 5;
	//unsigned int bSize = 4;

	if( pSize < 4 || bSize < 4 )
		return matchList;	

	ublas::matrix< double > mP( pSize, pSize );
	ublas::matrix< double > mB( bSize, bSize );
	
	/*mB( 0, 0 ) = 0;
	mB( 0, 1 ) = 1;
	mB( 0, 2 ) = 1;
	mB( 0, 3 ) = 4;
	mB( 1, 0 ) = 1;
	mB( 1, 1 ) = 0;
	mB( 1, 2 ) = 2;
	mB( 1, 3 ) = 5;
	mB( 2, 0 ) = 1;
	mB( 2, 1 ) = 2;
	mB( 2, 2 ) = 0;
	mB( 2, 3 ) = 3;
	mB( 3, 0 ) = 4;
	mB( 3, 1 ) = 5;
	mB( 3, 2 ) = 3;
	mB( 3, 3 ) = 0;

	mP( 0, 0 ) = 0;
	mP( 0, 1 ) = 1;
	mP( 0, 2 ) = 9;
	mP( 0, 3 ) = 3;
	mP( 0, 4 ) = 2;
	mP( 1, 0 ) = 1;
	mP( 1, 1 ) = 0;
	mP( 1, 2 ) = 6;
	mP( 1, 3 ) = 4;
	mP( 1, 4 ) = 1;
	mP( 2, 0 ) = 9;
	mP( 2, 1 ) = 6;
	mP( 2, 2 ) = 0;
	mP( 2, 3 ) = 8;
	mP( 2, 4 ) = 7;
	mP( 3, 0 ) = 3;
	mP( 3, 1 ) = 4;
	mP( 3, 2 ) = 8;
	mP( 3, 3 ) = 0;
	mP( 3, 4 ) = 5;
	mP( 4, 0 ) = 2;
	mP( 4, 1 ) = 1;
	mP( 4, 2 ) = 7;
	mP( 4, 3 ) = 5;
	mP( 4, 4 ) = 0;	
	
	*/	
	//compute distance matrices
	
	for( unsigned int row = 0; row < pSize; row++ )
	{
		for( unsigned int col = row; col < pSize; col++ )
		{
			mP( row, col ) = ublas::norm_2( points.at( row ) - points.at( col ) );
			if( row != col )
				mP( col, row ) = mP( row, col );
		}
	}

	for( unsigned int row = 0; row < bSize; row++ )
	{
		for( unsigned int col = row; col < bSize; col++ )
		{
			mB( row, col ) = ublas::norm_2( body.at( row ) - body.at( col ) );
			if( row != col )
				mB( col, row ) = mB( row, col );
		}
	}

	bool found = false;
	double val = 0;
	double bestVal;


	//std::cout << " --------------------------------" << std::endl;
	//look for all entries of body matrix
	for( unsigned int row = 0; row < bSize; row++ )
	{
		for( unsigned int col = 0; col < bSize; col++ )
		{
			//for returning the list
			std::vector< std::pair< unsigned int, unsigned int > > match;

			if( row != col )
				if( findEntries( mP, ublas::row( mB, row ), row, col, match , error, val ) )	//if a matching list was found
				{
					if( !found )											//if it is the first list
					{
						bestVal = val;
						matchList = match;
						found = true;
					}
					else													//all others should be checked if they did better
						if( val < bestVal )
						{
							bestVal = val;
							matchList = match;
						}
				}
		}
	}

	//for debug, will be deleted
	/*if( debug )
	{
		std::cout << "points: ";
		for( unsigned int i=0; i< points.size(); i++ )
			std::cout << points.at( i ) << " | ";

		std::cout << std::endl << "body: ";
		for( unsigned int i=0; i< body.size(); i++ )
			std::cout << body.at( i ) << " | ";

		std::cout << std::endl << "match: ";
		for( unsigned int i=0; i< matchList.size(); i++ )
			std::cout << matchList.at( i ).first << ", " << matchList.at( i ).second << " | ";

		std::cout << "pointsDist: " << mP << std::endl;
		std::cout << "bodyDist: " << mB << std::endl;

	}*/
	

	return matchList;
}

} } // namespace Ubitrack::Driver