/*
 * Ubitrack - Library for Ubiquitous Tracking
 * Copyright 2006, Technische Universitaet Muenchen, and individual
 * contributors as indicated by the @authors tag. See the
 * copyright.txt in the distribution for a full listing of individual
 * contributors.
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA, or see the FSF site: http://www.fsf.org.
 */

#include "RBCalibration.h"

static log4cpp::Category& logger( log4cpp::Category::getInstance( "Drivers.RigidBodyTracking" ) );

using namespace Ubitrack;
using namespace Ubitrack::Dataflow;

// shortcuts to namespaces
namespace ublas = boost::numeric::ublas;

namespace Ubitrack { namespace Drivers {	
	
//copy corresponding points to the vectors
void RBCalibrationModule::findCorrespondences( const TraceContainer & from, const TraceContainer & to, std::vector< Math::Vector< 2 > > & fromPoints, std::vector< Math::Vector< 2 > > & toPoints )
{
	long long fStart = from.getTime( 0 );
	long long tStart = to.getTime( 0 );

	// do both traces overleap?
	if ( fStart < to.getTime( to.size() - 1 ) && tStart < from.getTime( from.size() - 1 ) )
	{
		unsigned int k=0;
		unsigned int l=0;

		//find starting points as follows: if fstart < tStart then go to the point of 'from' wich is directly behind tStart
		if( fStart < tStart )
		{
			while( from.getTime( k ) < tStart && k < from.size() - 1 )
			{
				k++;
			}
			if( k > 0 )
				k--;
		}
		else
		{
			while( to.getTime( l ) < fStart && l < to.size() - 1 )
			{
				l++;
			}
			if( l > 0 )
				l--;
		}

		//find nearest corresponding point
		while( k < from.size() - 1 && l < to.size() - 1 )
		{
			unsigned int k_ = k;
			unsigned int l_ = l;

			if( abs( from.getTime( k + 1 ) - to.getTime( l ) ) < abs( from.getTime( k ) - to.getTime( l ) ) )
				k_ = k + 1;
			if( abs( from.getTime( k ) - to.getTime( l + 1 ) ) < abs( from.getTime( k_) - to.getTime( l ) ) )
			{
				k_ = k;
				l_ = l + 1;
			}

			k = k_;
			l = l_;

			if( abs( from.getTime( k ) - to.getTime( l ) ) < m_tMax )
			{
				fromPoints.push_back( from.at( k ) );
				toPoints.push_back( to.at( l ) );
			}

			k++;
			l++;
		}

		//if the two last points are left
		if( k < from.size() && l < to.size() )
			if( abs( from.getTime( k ) - to.getTime( l ) ) < m_tMax )
			{
				fromPoints.push_back( from.at( k ) );
				toPoints.push_back( to.at( l ) );
			}
	}
}


/** check integrity of FM */
double RBCalibrationModule::validateFundamentalMatrix( const Math::Matrix< 3, 3 > & fM, const std::vector< Math::Vector< 2 > > & fromPoints, const std::vector< Math::Vector< 2 > > & toPoints, const unsigned int step )
{
	double sum = 0.0;
	double term = 0.0;
	Math::Vector< 3 > x;
	Math::Vector< 2 > x_;

	for( unsigned int i=0; i < fromPoints.size(); i++ )
	{
		sum += Calibration::pointToPointDist( fromPoints.at( i ), toPoints.at( i ), fM );
	}

	return sum / fromPoints.size();
}

/** finds the fundamental matrix of two point traces*/
Math::Matrix< 3, 3 > RBCalibrationModule::findFM( const std::map< int, TraceContainer > & firstCam, const std::map< int, TraceContainer > & secondCam )
{
	Math::Matrix< 3, 3 > fundamentalMatrix;
	double errorValue = 0.0;
	unsigned int ntrace = 0;

	bool first = true;

	std::vector< Math::Vector< 2 > > fromPointsAll;
	std::vector< Math::Vector< 2 > > toPointsAll;
	std::vector< Math::Matrix< 3, 3 > > allMatrices;

	//try for all traces of firstCam
	for( std::map< int, TraceContainer >::const_iterator it1 = firstCam.begin(); it1 != firstCam.end(); it1++ )
	{
		for( std::map< int, TraceContainer >::const_iterator it2 = secondCam.begin();it2 != secondCam.end() ; it2++ )
		{
			std::vector< Math::Vector< 2 > > fromPoints;
			std::vector< Math::Vector< 2 > > toPoints;
			Math::Matrix< 3, 3 > fM;

			findCorrespondences( it1->second, it2->second, fromPoints, toPoints );

			unsigned int step = fromPoints.size() / m_nMin;

			if( step / 2)
			{
				fM = Calibration::getFundamentalMatrix( fromPoints, toPoints, step );

				double v = validateFundamentalMatrix( fM,  fromPoints, toPoints, step );

				if( errorValue > v || first)
				{
					first = false;
					errorValue = v;
					fundamentalMatrix = fM;
					ntrace = fromPoints.size();
					m_fromPoints = fromPoints;
					m_toPoints = toPoints;
				}

				if( v < m_eMax )
				{
					for( unsigned int a=0; a < fromPoints.size(); a++ )
					{
						fromPointsAll.push_back( fromPoints.at( a ) );
						toPointsAll.push_back( toPoints.at( a ) );
					}

					allMatrices.push_back( fM );
				}
			}
		}
	}

	if( first )
	{
		LOG4CPP_ERROR ( logger, "not enough trace points for calibration");
		//UBITRACK_THROW ( "not enough trace points for calibration" );
		m_ok = false;
	}
	else
	{
		m_ok = true;
		first = true;

		if( errorValue < m_eMax && allMatrices.size() > 1 )
		{
			for( unsigned int i=0; i<allMatrices.size(); i++ )
			{
				double v = validateFundamentalMatrix( allMatrices.at( i ) , fromPointsAll, toPointsAll, fromPointsAll.size() );

				if( errorValue > v || first)
				{
					first = false;
					errorValue = v;
					fundamentalMatrix = allMatrices.at( i );
					ntrace = fromPointsAll.size();
				}
			}

			Math::Matrix< 3, 3 > fM = Calibration::getFundamentalMatrix( fromPointsAll, toPointsAll, fromPointsAll.size() / ( m_nMin * allMatrices.size() ) );

			double v = validateFundamentalMatrix( fM, fromPointsAll, toPointsAll, fromPointsAll.size() / ( m_nMin * allMatrices.size() ) );

			if( errorValue > v )
				fundamentalMatrix = fM;

			m_fromPoints = fromPointsAll;
			m_toPoints = toPointsAll;
		}

		std::cout << "error value of matrix: " << errorValue << std::endl;
		std::cout << Math::Matrix< 3, 3 >( fundamentalMatrix / double( boost::numeric::ublas::norm_frobenius( fundamentalMatrix ) ) ) << std::endl;
	}
	return fundamentalMatrix;
}

} } // namespace Ubitrack::Driver