/*
 * Ubitrack - Library for Ubiquitous Tracking
 * Copyright 2006, Technische Universitaet Muenchen, and individual
 * contributors as indicated by the @authors tag. See the 
 * copyright.txt in the distribution for a full listing of individual
 * contributors.
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA, or see the FSF site: http://www.fsf.org.
 */
 
 /**
 * @ingroup drivers
 * @file
 * Implementation of the SpaceNavigator driver component.
 *
 * @author Peter Keitler <keitler@in.tum.de>
 */

#include "SpaceNavigator.h"

#include "stdio.h"
#include <Ubitrack/Util/OS.h>

#define _USE_MATH_DEFINES
#include <math.h>
#include <boost/math/quaternion.hpp>

#include <log4cpp/Category.hh>

static log4cpp::Category& logger( log4cpp::Category::getInstance( "Drivers.SpaceNavigator" ) );


namespace Ubitrack { namespace Driver {


/// helper function to check and parse position attribute
Math::Vector< 3 > readPosition( const std::string& s )
{
	double position[3];
	sscanf( s.c_str(), "%lf %lf %lf", &position[0], &position[1], &position[2] );
	return Math::Vector<3>( position[0], position[1], position[2] );
}

/// helper function to check and parse rotation attribute
Math::Quaternion readRotation( const std::string& s )
{
	double rotation[4];
	sscanf( s.c_str(), "%lf %lf %lf %lf", &rotation[0], &rotation[1], &rotation[2], &rotation[3] );
	return Math::Quaternion( rotation[0], rotation[1], rotation[2], rotation[3] );
}



SpaceNavigatorDriver::SpaceNavigatorDriver( const std::string& sName, boost::shared_ptr< Graph::UTQLSubgraph > subgraph )
	: Dataflow::Component( sName )
	, poseOutPort( "Movement", *this )
	, buttonOutPort( "Button", *this )
	, m_bStop( true )
	, posThreshold( 50 )
	, rotThreshold( 50 )
	, posSensitivity( 0.2 )
	, rotSensitivity( 0.01 )
	, axisMapping( UBITRACK_FRAME )
	, strongestEffectOnly( false )
	, resetEnabled( true )
{
	Math::Quaternion rot( 0, 0, 0, 1 );
	Math::Vector< 3 > pos( 0, 0, 0 );

	std::string sAttr = subgraph->m_DataflowAttributes.getAttributeString( "position" );
	if ( !sAttr.empty() )
		pos = readPosition( sAttr );

	sAttr = subgraph->m_DataflowAttributes.getAttributeString( "rotation" );
	if ( !sAttr.empty() )
		rot = readRotation( sAttr );

	currentPose = Math::Pose( rot, pos );
	initialPose = currentPose;

	LOG4CPP_DEBUG( logger, "Initialize pose to " << currentPose );

	subgraph->m_DataflowAttributes.getAttributeData( "posThreshold", posThreshold );
	subgraph->m_DataflowAttributes.getAttributeData( "rotThreshold", rotThreshold );
	subgraph->m_DataflowAttributes.getAttributeData( "posSensitivity", posSensitivity );
	subgraph->m_DataflowAttributes.getAttributeData( "rotSensitivity", rotSensitivity );
	std::string mapping;
	subgraph->m_DataflowAttributes.getAttributeData( "axisMapping", mapping );
	if ( mapping.compare("default") == 0 )
		axisMapping = DEFAULT;
	if ( mapping.compare("frame") == 0 )
		axisMapping = UBITRACK_FRAME;
	if ( mapping.compare("worldCam") == 0 )
		axisMapping = WORLD_CAMERA;
	subgraph->m_DataflowAttributes.getAttributeData( "suppressMinorMovements", strongestEffectOnly );
	subgraph->m_DataflowAttributes.getAttributeData( "enableReset", resetEnabled );
	std::string button( " " );
	if ( subgraph->m_DataflowAttributes.hasAttribute( "button" ) )
		button = subgraph->m_DataflowAttributes.getAttributeString( "button" );
	if ( button.empty() )
		buttonEvent = Math::Scalar< int >( -1 );
	else
		buttonEvent = Math::Scalar< int >( button[ 0 ] );

	LOG4CPP_DEBUG( logger, "Button event is [" << buttonEvent << "]" );
}


void SpaceNavigatorDriver::threadMethod()
{
	int status;
	spnav_event event;

	LOG4CPP_DEBUG( logger, "Driver thread started" );
	
	status = spnav_open();
	LOG4CPP_TRACE( logger, "Library initialization status: " << status );
	if ( status == -1 )
		LOG4CPP_INFO( logger, "SpaceNavigatorDriver: Could not initialize SpaceNavigator device. In case of a restart of this component, it might have been initialized already." );

	while ( !m_bStop )
	{
		LOG4CPP_TRACE( logger, "Wait for next event" );

		Util::sleep ( 100 );

		status = spnav_poll_event( &event );
		Measurement::Timestamp timestamp = Measurement::now();
		if ( status == 0 ) 
		{
			LOG4CPP_TRACE( logger, "Error " << status << ", timestamp: " << timestamp );
			continue;
		}
			
		LOG4CPP_TRACE( logger, "Received event, timestamp: " << timestamp );
		switch( event.type )
		{
		case SPNAV_EVENT_MOTION:
			LOG4CPP_TRACE( logger, "Received motion event, type: " << event.motion.type << ", position: [" << event.motion.x << " " << event.motion.y << " " << event.motion.z << "], orientation: [" << event.motion.rx << " " << event.motion.ry << " " << event.motion.rz << "], period: " << event.motion.period << ", timestamp: " << timestamp );

			sendPose( timestamp, &(event.motion) );
			break;
				
		case SPNAV_EVENT_BUTTON:
			LOG4CPP_TRACE( logger, "Received button event, type: " << event.button.type << ", press: " << event.button.press << ", button number: " << event.button.bnum << ", timestamp: " << timestamp );
			if ( event.button.press == 1 ) 
			{
				LOG4CPP_TRACE( logger, "button pressed" );
				if ( event.button.bnum == 0 ) 
				{
					LOG4CPP_TRACE( logger, "send button event" );
					sendButton ( timestamp );
				}
				
				else if ( resetEnabled && event.button.bnum == 1 ) 
				{
					currentPose = initialPose;
					LOG4CPP_TRACE( logger, "reset to initial pose" );
				}
			}
			break;
		}
	
		// Delete superfluous motion events
		status = spnav_remove_events( SPNAV_EVENT_MOTION );
		LOG4CPP_TRACE( logger, "Removed " << status << " motion events from queue" );
	}
	LOG4CPP_INFO (logger, "Closing..." );
}
	
Math::Quaternion SpaceNavigatorDriver::euler2Quat( double x, double y, double z )
{
	Math::Quaternion q;

	// XYZ fixed angles (see Craig, p. 46)
	Math::Quaternion rotX ( sin( x/2.), 0, 0, cos( x/2.) );
	Math::Quaternion rotY ( 0, sin( y/2.) ,0, cos( y/2.) );
	Math::Quaternion rotZ ( 0, 0, sin( z/2.), cos( z/2.) );

	q = rotZ * rotY * rotX;

	return q;
}

void SpaceNavigatorDriver::sendPose( const Measurement::Timestamp timestamp, motion_type * motion )
{
	Math::Quaternion quatInc;
	double x, y, z, rx, ry, rz;

	// Ubitrack frame mapping
	if ( axisMapping == UBITRACK_FRAME ) 
	{
		LOG4CPP_TRACE( logger, "Map axis to 'ubitrack frame' " );
		x = motion->x;
		y = motion->z;
		z = motion->y;
		rx = motion->rx;
		ry = motion->rz;
		rz = motion->ry;
	}
	// Egocentric world camera mapping
	else if ( axisMapping == WORLD_CAMERA ) 
	{
		LOG4CPP_TRACE( logger, "World camera mapping" );
		x = motion->x;
		y = motion->y;
		z = - motion->z;
		rx = motion->rx;
		ry = motion->ry;
		rz = motion->rz;
	}
	else 
	{
		LOG4CPP_TRACE( logger, "No axis mapping" );
		x = motion->x;
		y = motion->y;
		z = motion->z;
		rx = motion->rx;
		ry = motion->ry;
		rz = motion->rz;
	}

	if ( strongestEffectOnly ) 
	{
		LOG4CPP_TRACE( logger, "Apply strongest motion effect only" );

		if ( abs(x)>abs(y) && abs(x)>abs(z) && abs(x)>abs(rx) && abs(x)>abs(ry) && abs(x)>abs(rz) ) 
		{
			y = 0;
			z = 0;
			rx = 0;
			ry = 0;
			rz = 0;
		}
		else if ( abs(y)>abs(x) && abs(y)>abs(z) && abs(y)>abs(rx) && abs(y)>abs(ry) && abs(y)>abs(rz) ) 
		{
			x = 0;
			z = 0;
			rx = 0;
			ry = 0;
			rz = 0;
		}
		else if ( abs(z)>abs(x) && abs(z)>abs(y) && abs(z)>abs(rx) && abs(z)>abs(ry) && abs(z)>abs(rz) ) 
		{
			x = 0;
			y = 0;
			rx = 0;
			ry = 0;
			rz = 0;
		}
		else if ( abs(rx)>abs(x) && abs(rx)>abs(y) && abs(rx)>abs(z) && abs(rx)>abs(ry) && abs(rx)>abs(rz) ) 
		{
			x = 0;
			y = 0;
			z = 0;
			ry = 0;
			rz = 0;
		}
		else if ( abs(ry)>abs(x) && abs(ry)>abs(y) && abs(ry)>abs(z) && abs(ry)>abs(rx) && abs(ry)>abs(rz) ) 
		{
			x = 0;
			y = 0;
			z = 0;
			rx = 0;
			rz = 0;
		}
		else if ( abs(rz)>abs(x) && abs(rz)>abs(y) && abs(rz)>abs(z) && abs(rz)>abs(rx) && abs(rz)>abs(ry) ) 
		{
			x = 0;
			y = 0;
			z = 0;
			rx = 0;
			ry = 0;
		}
	}
	
	// Compute position increment
	if ( x < posThreshold && x > -posThreshold )
		x = 0;
	if ( y < posThreshold && y > -posThreshold )
		y = 0;
	if ( z < posThreshold && z > -posThreshold )
		z = 0;
	Math::Vector<3> posInc(posSensitivity * x / 1000, posSensitivity * y / 1000, posSensitivity * z / 1000 );

	// Compute orientation increment
	if ( rx < rotThreshold && rx > -rotThreshold )
		rx = 0;
	if ( ry < rotThreshold && ry > -rotThreshold )
		ry = 0;
	if ( rz < rotThreshold && rz > -rotThreshold )
		rz = 0;
	Math::Quaternion rotInc = euler2Quat( rotSensitivity * rx * M_PI/180 , rotSensitivity * ry * M_PI/180, rotSensitivity * rz * M_PI/180 );

	// Compose pose increment
	Math::Pose poseInc( rotInc, posInc );
	LOG4CPP_TRACE( logger, "Incremental pose: " << poseInc );

	// Update current pose
	currentPose = currentPose * poseInc;
	LOG4CPP_TRACE( logger, "Updated pose: " << currentPose );

 	Measurement::Pose measurement( timestamp, currentPose );
 	poseOutPort.send( measurement );
}

void SpaceNavigatorDriver::sendButton( const Measurement::Timestamp timestamp ) 
{
	buttonOutPort.send( Measurement::Button( timestamp, buttonEvent ) );
}

	
void SpaceNavigatorDriver::stop()
{
	LOG4CPP_DEBUG( logger, "Stop" );

	if ( m_running )
	{
		LOG4CPP_DEBUG( logger, "Thread is running, so stop it" );

		m_running = false;
		m_bStop = true;
		if ( m_pThread )
		{
			m_pThread->join();
		}
	}
	LOG4CPP_DEBUG( logger, "Thread stopped successfully" );
}
	
void SpaceNavigatorDriver::start()
{
	LOG4CPP_DEBUG( logger, "Start" );

	if ( !m_running )
	{
		LOG4CPP_DEBUG( logger, "Thread is not running, so start it" );

		m_running = true;
		m_bStop = false;
		m_pThread.reset( new boost::thread( boost::bind( &SpaceNavigatorDriver::threadMethod, this ) ) );
	}
}

UBITRACK_REGISTER_COMPONENT( Ubitrack::Dataflow::ComponentFactory* const cf ) {
	cf->registerComponent< Ubitrack::Driver::SpaceNavigatorDriver > ( "SpaceNavigator" );
}

} } // namespace Ubitrack::Driver
