/*
 * Ubitrack - Library for Ubiquitous Tracking
 * Copyright 2006, Technische Universitaet Muenchen, and individual
 * contributors as indicated by the @authors tag. See the 
 * copyright.txt in the distribution for a full listing of individual
 * contributors.
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA, or see the FSF site: http://www.fsf.org.
 */
 
 /**
 * @ingroup drivers @file Definition of the SpaceNavigator driver
 * class.
 *
 * It integrates the free SpaceNavigator device driver daemon * into
 * Ubitrack using an appropriate conversion of pose information.
 *
 * @author Peter Keitler <keitler@in.tum.de>
 */

#ifndef __Ubitrack_Drivers_SpaceNavigator_SpaceNavigator_h_included__
#define __Ubitrack_Drivers_SpaceNavigator_SpaceNavigator_h_included__

#include <Ubitrack/Dataflow/Component.h>
#include <Ubitrack/Dataflow/ComponentFactory.h>
#include <Ubitrack/Dataflow/PushSupplier.h>
#include <Ubitrack/Measurement/Measurement.h>
#include <Ubitrack/Util/Exception.h>

#include <boost/thread.hpp>

#include <string>

#include <spnav.h>

using namespace Ubitrack;

typedef struct spnav_event_button button_type;
typedef struct spnav_event_motion motion_type;

namespace Ubitrack { namespace Driver {

		enum cameraStyle{
		DEFAULT,
		UBITRACK_FRAME,
		WORLD_CAMERA
		};
		
		class SpaceNavigatorDriver : public Dataflow::Component
		{
		public:
			/** constructor */
			SpaceNavigatorDriver( const std::string& sName, boost::shared_ptr< Graph::UTQLSubgraph > subgraph );
			
			/** destructor */
			~SpaceNavigatorDriver( ) { stop(); };
	
			virtual void start();
			virtual void stop();

		protected:
			// Output ports of the component
			Dataflow::PushSupplier< Measurement::Pose > poseOutPort;
			Dataflow::PushSupplier< Measurement::Button > buttonOutPort;

			/** Converts motion event from daemon to Ubitrack pose and pushes it onwards */ 
			void sendPose( const Measurement::Timestamp timestamp, motion_type * motion );

			/** Converts button event from daemon to Ubitrack button event and pushes it onwards */
			void sendButton ( const Measurement::Timestamp timestamp );

			/** transforms the SpaceNavigator rotation angles to Ubitrack rotation quaternions */
			Math::Quaternion euler2Quat( double heading, double attitude, double bank ) ;

			/** thread method */
			void threadMethod( );
	
			// pointer to the thread
			boost::shared_ptr< boost::thread > m_pThread;
	
			/** thread is running? */
			bool m_bStop;

			/** Initial configured pose */
			Math::Pose initialPose;
			/** Current pose, constantly being updated by incremental motion events */
			Math::Pose currentPose;
			/** Position threshold, has to be exceeded in order to have an effect on the pose */
			int posThreshold;
			/** Orientation threshold, has to be exceeded in order to have an effect on the pose */
			int rotThreshold;
			/** Position sensitivity */
			double posSensitivity;
			/** Orientation sensitivity */
			double rotSensitivity;
			/** Axis mapping */
			cameraStyle axisMapping;
			/** If set, only the most important translation/rotation will have an effect on the pose */
			bool strongestEffectOnly;
			/** Reset to initial pose enabled */
			bool resetEnabled;
			/** Button event character */
			Math::Scalar< int > buttonEvent;
		};

	} } // namespace Ubitrack::Drivers

#endif
