/*
 * Ubitrack - Library for Ubiquitous Tracking
 * Copyright 2006, Technische Universitaet Muenchen, and individual
 * contributors as indicated by the @authors tag. See the
 * copyright.txt in the distribution for a full listing of individual
 * contributors.
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA, or see the FSF site: http://www.fsf.org.
 */


/**
 * @ingroup dataflow_components
 * @file
 * VectorfieldMapper component
 *
 * @author Florian Echtler <echtler@in.tum.de>
 */

// WARNING: all boost/serialization headers should be
//          included AFTER all boost/archive headers
#include <boost/archive/text_oarchive.hpp>
#include <boost/serialization/vector.hpp>
#include <boost/serialization/utility.hpp>

#include <string>
#include <iostream>
#include <fstream>

#include <boost/bind.hpp>

#include <Ubitrack/Math/Quaternion.h>
#include <Ubitrack/Dataflow/ComponentFactory.h>
#include <Ubitrack/Dataflow/Component.h>
#include <Ubitrack/Dataflow/PushConsumer.h>
#include <Ubitrack/Measurement/Measurement.h>

namespace Ubitrack { namespace Components {

using namespace Measurement;

class VectorfieldMapper: public Dataflow::Component
{
public:
	/**
	 * Standard component constructor.
	 *
	 * @param nm Unique name of the component.
	 * @param cfg ComponentConfiguration containing all configuration.
	 */
	VectorfieldMapper( const std::string& nm, boost::shared_ptr< Graph::UTQLSubgraph > )
    : Ubitrack::Dataflow::Component( nm )
    , m_vectorPort( "VectorInput", *this, boost::bind( &VectorfieldMapper::vectorFunc, this, _1 ) )
    , m_posePort( "PoseInput", *this, boost::bind( &VectorfieldMapper::poseFunc, this, _1 ) )
		, field("vfield.dump",std::ios_base::app)
		//, field( outfile )
	{
	}

protected:

	void vectorFunc( const Vector3D& vector ) {
		//std::cout << "Received Vector: " << vector << std::endl;
		// store most recent vector - ugly hack, but should work for the moment.
		m_vector = *vector;
	}

	void poseFunc( const Pose& pose ) {
		//std::cout << "Received Pose: " << *pose << std::endl;
		//std::cout << "Last Vector: " << m_vector << std::endl;

		Math::Quaternion rot = pose->rotation();
		Math::Vector<3> pos = pose->translation();
		// rotate measurement vector by pose
		Math::Vector<3> tmp = (~rot)*m_vector;
		field << pos << " ";
		field << tmp << std::endl;
		//std::cout << "Vector: " << tmp << std::endl;
	}

	Math::Vector<3> m_vector;

	/** Input ports of the component. */
	Dataflow::PushConsumer< Vector3D > m_vectorPort;
	Dataflow::PushConsumer< Pose > m_posePort;

	std::ofstream field;
	//boost::archive::text_oarchive field;
};


UBITRACK_REGISTER_COMPONENT( Dataflow::ComponentFactory* const cf ) {
	cf->registerComponent< VectorfieldMapper > ( "VectorfieldMapper" );
}

} } // namespace Ubitrack::Components

