/*
 * Ubitrack - Library for Ubiquitous Tracking
 * Copyright 2006, Technische Universitaet Muenchen, and individual
 * contributors as indicated by the @authors tag. See the
 * copyright.txt in the distribution for a full listing of individual
 * contributors.
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA, or see the FSF site: http://www.fsf.org.
 */


#include "Art2D.h"
#include <iostream>
#include <iomanip>
#include <sstream>
#include <Ubitrack/Dataflow/ComponentFactory.h>
#include <log4cpp/Category.hh>


/* The pattern:
<Pattern name="Art2D" id="Art2D1">

	<Output>
		<Node name="Art" id="Art">
			<Attribute name="camNetDev" value="eth0"/>
		</Node>
		<Node name="Camera" id="Cam1">
			<Attribute name="camNr" value="1"/>
		</Node>
		<Edge name="Output" source="Art" destination="Camera">
			<Attribute name="type" value="PositionList2"/>
			<Attribute name="mode" value="push"/>
		</Edge>
	</Output>

	<DataflowConfiguration>
		<UbitrackLib class="Art2D"/>
	</DataflowConfiguration>

</Pattern>
*/

namespace {
// anonymous stuff for ethernet packet parsing

/** 
 * Represents a range of bytes in a buffer
 */
class ByteRange
{
public:
	ByteRange( const u_char* pStart, int nSize )
		: m_pStart( pStart )
		, m_nSize( nSize )
	{}

	int getSize() const
	{ return m_nSize; }

	u_char operator[]( int i ) const
	{
		if ( i < 0 || i >= m_nSize ) 
			UBITRACK_THROW( "Range error" );
		return m_pStart[ i ];
	}

	ByteRange slice( int nStart, int nSize )
	{
		checkRange( nStart, nSize );
		return ByteRange( m_pStart + nStart, nSize );
	}

	unsigned getNetUShort( int i ) const
	{
		checkRange( i, 2 );
		return ( (unsigned)m_pStart[ i ] << 8 ) | (unsigned)m_pStart[ i + 1 ];
	}

	unsigned getUShort( int i ) const
	{
		checkRange( i, 2 );
		return (unsigned)m_pStart[ i ] | ( (unsigned)m_pStart[ i + 1 ] << 8 );
	}

protected:
	const u_char* m_pStart;
	int m_nSize;

	void checkRange( int nStart, int nSize ) const
	{ 
		if ( nStart < 0 || nSize < 0 || nStart + nSize > m_nSize ) 
			UBITRACK_THROW( "Range error" );
	}
};

/** gets the IP part out of an ethernetII frame */
ByteRange decodeEthernetII( ByteRange in )
{ 
	// protocol must be IP
	if ( in.getNetUShort( 12 ) != 0x0800 ) 
		UBITRACK_THROW( "Eth frame not IP" );
	return in.slice( 14, in.getSize() - 14 );
}

/** extracts the UDP part out of an IP packet */	
ByteRange decodeIP( ByteRange in )
{
	// check version
	if ( in[ 0 ] >> 4 != 4 )
		UBITRACK_THROW( "invalid IP version" );
		
	// check fragment offset
	if ( in.getNetUShort( 6 ) & 0x1FFF )
		UBITRACK_THROW( "Fragmented IP packet" );

	// check UDP protocol
	if ( in[ 9 ] != 0x11 )
		UBITRACK_THROW( "Not UDP" );
		
	int nHead = ( in[ 0 ] & 0xF ) * 4;
	int nSize = in.getNetUShort( 2 );
	return in.slice( nHead,  nSize - nHead );
}

/** decodes the payload of a udp packet */
ByteRange decodeUDP( ByteRange in )
{
	return in.slice( 8, in.getNetUShort( 4 ) - 8 );
}

/** decodes 2d art packets */
void decodeArt( ByteRange in, unsigned& frameId, int& cameraId, std::vector< Ubitrack::Math::Vector< 2 > >& points )
{
	if ( in[ 0 ] != 0x31 )
		UBITRACK_THROW( "packet type not implemented" );
		
	// frame id
	frameId = in.getUShort( 6 );
		
	// camera id
	cameraId = (unsigned)in[ 1 ];

	// number of markers
	unsigned nMarkers = in[ 2 ];
	
	// x and y coordinate for all markers
	for ( unsigned i = 0; i < nMarkers; i++ )	
		points.push_back( Ubitrack::Math::Vector< 2 >( in.getUShort( 8 + i * 6 ) / 100.0, in.getUShort( 10 + i * 6 ) / 100.0 ) );
}

} // anonymous namespace


namespace Ubitrack { namespace Drivers {

static log4cpp::Category& logger( log4cpp::Category::getInstance( "Drivers.Art2D" ) );
Art2DModule::Art2DModule( const Art2DModuleKey& moduleKey, boost::shared_ptr< Graph::UTQLSubgraph >, FactoryHelper* pFactory )
	: Module< Art2DModuleKey, Art2DComponentKey, Art2DModule, Art2DComponent >( moduleKey, pFactory )
	, m_synchronizer( 59.5 ) // assume 60 Hz framerate
	, m_bStop( false )
{
	LOG4CPP_NOTICE( logger, "Creating Art2D network service" );
	LOG4CPP_INFO( logger, "Using device " << m_moduleKey.get() );

	// open pcap device
	char sErr[ PCAP_ERRBUF_SIZE ];
	int promiscuous = 0; // set to 1 if not listening on the DTrack computer
	m_pPcap = pcap_open_live( moduleKey.get().c_str(), 2048, promiscuous, 5, sErr );
	if ( !m_pPcap )
	{
		std::string errStr( sErr );
		
		// get list of all interfaces and print it so people aren't completely left alone with finding
		// the interface name
		std::string ifList;
		pcap_if_t* pif = 0;
		pcap_findalldevs( &pif, sErr );
		for ( ; pif; pif = pif->next )
		{
			ifList += pif->name;
			if ( pif->description )
				ifList += " (" + std::string( pif->description ) + ")";
			if ( pif->next )
				ifList += "\n";
		}
		LOG4CPP_ERROR( logger, "possible network devices:\n" << ifList );
		
		UBITRACK_THROW( "Error opening pcap device '" + moduleKey.get() + "': " + errStr );
	}
	
	// network thread runs until io_service is interrupted
	m_networkThread.reset( new boost::thread( boost::bind( &Art2DModule::threadProc, this ) ) );
}


Art2DModule::~Art2DModule()
{
	LOG4CPP_NOTICE( logger, "Stopping Art2D network service" );

	m_bStop = true;
	m_networkThread->join();
	pcap_close( m_pPcap );
}


void Art2DModule::threadProc()
{
	// read packets
	pcap_pkthdr* pHead;
	const u_char* pData;
		
	while ( !m_bStop )
	{
		if ( pcap_next_ex( m_pPcap, &pHead, &pData ) == 1 && m_running )
		{
			Measurement::Timestamp arrivalTime = Measurement::now();
			LOG4CPP_TRACE( logger, "Received frame at " << Measurement::timestampToShortString( arrivalTime )
				<< ", PCAP timestamp: " << pHead->ts.tv_sec << "." << std::setw( 6 ) << std::setfill( '0' ) << pHead->ts.tv_usec
				<< ", size: " << pHead->caplen );
			try
			{
				// parse packet
				unsigned frame;
				int camId;
				boost::shared_ptr< std::vector< Math::Vector< 2 > > > pPoints( new std::vector< Math::Vector< 2 > > );
				decodeArt( 
					decodeUDP( decodeIP( decodeEthernetII( ByteRange( pData, pHead->caplen ) ) ) ), 
					frame, camId, *pPoints
				);
				
				// send measurement
				LOG4CPP_DEBUG( logger, "Measurement: frame=" << frame << ", cam=" << camId << ", pointnum=" << pPoints->size() );

				// synchronize timestamps
				// Measurement::Timestamp correctedTS = m_synchronizer.convertNativeToLocal( frame, arrivalTime );

				// as pcap timestamps seem to be accurate and frame numbers seem to be unequal for all cameras (!),
				// we'll just use the pcap timestamps
				Measurement::Timestamp correctedTS = 1000000000LL * pHead->ts.tv_sec + 1000LL * pHead->ts.tv_usec;
				
				// subtract the approximate processing time of the cameras
				correctedTS -= 12000000;

				getComponent( camId )->send( Measurement::PositionList2( correctedTS, pPoints ) );
			}
			catch( Util::Exception& e )
			{
				LOG4CPP_DEBUG( logger, "exception: " << e );
			}
		}
	}
}


// register module at factory
UBITRACK_REGISTER_COMPONENT( ComponentFactory* const cf ) 
{
	cf->registerModule< Art2DModule > ( "Art2D" );
}

} } // namespace Ubitrack::Drivers
