/*
 * Ubitrack - Library for Ubiquitous Tracking
 * Copyright 2006, Technische Universitaet Muenchen, and individual
 * contributors as indicated by the @authors tag. See the
 * copyright.txt in the distribution for a full listing of individual
 * contributors.
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA, or see the FSF site: http://www.fsf.org.
 */


/**
 * @ingroup driver_components
 * @file
 * Art2D driver.
 * 
 * This file contains the driver component to
 * talk to the Art2D infrared tracking system.
 *
 * @author Daniel Pustka <pustka@in.tum.de>
 */
#ifndef __Art2DModule_h_INCLUDED__
#define __Art2DModule_h_INCLUDED__

#include <string>
#include <pcap.h>

#include <Ubitrack/Dataflow/PushSupplier.h>
#include <Ubitrack/Dataflow/Component.h>
#include <Ubitrack/Dataflow/Module.h>
#include <Ubitrack/Measurement/Measurement.h>
#include <Ubitrack/Measurement/TimestampSync.h>

namespace Ubitrack { namespace Drivers {

using namespace Dataflow;

// forward declaration
class Art2DComponent;

// module and 
#ifdef _WIN32
	MAKE_NODEATTRIBUTEKEY( Art2DModuleKey, std::string, "Art", "camNetDev" );
#else
	MAKE_NODEATTRIBUTEKEY( Art2DModuleKey, std::string, "Art", "camNetDev" );
#endif
MAKE_NODEATTRIBUTEKEY( Art2DComponentKey, int, "Camera", "camNr" );


/**
 * Module for ART tracker.
 * Does all the work
 */
class Art2DModule
	: public Module< Art2DModuleKey, Art2DComponentKey, Art2DModule, Art2DComponent >
{
public:
	/** UTQL constructor */
	Art2DModule( const Art2DModuleKey& key, boost::shared_ptr< Graph::UTQLSubgraph >, FactoryHelper* pFactory );

	/** destruktor */
	~Art2DModule();

protected:
	/** thread method */
    void threadProc();
	
	/** the thread to handle all the stuff */
    boost::shared_ptr< boost::thread > m_networkThread;

	/** timestamp synchronization */
	Measurement::TimestampSync m_synchronizer;
	
	/** pointer to pcap */
	pcap_t* m_pPcap;
	
	/** stop the module? */
	bool m_bStop;
};


/**
 * Component for ART tracker.
 * Does nothing but provide a push port
 
 * @TODO: make this two separate components for 6d/3dlist
 */
class Art2DComponent
	: public Art2DModule::Component
{
public:
	/** constructor */
	Art2DComponent( const std::string& name, boost::shared_ptr< Graph::UTQLSubgraph > subgraph, 
		const Art2DComponentKey& componentKey, Art2DModule* pModule )
		: Art2DModule::Component( name, componentKey, pModule )
		, m_port( "Output", *this )
	{}

	/** returns the port for usage by the module */
	void send( const Measurement::PositionList2& m )
	{ m_port.send( m ); }

protected:
	// the port is the only member
	PushSupplier< Ubitrack::Measurement::PositionList2 > m_port;
};

} } // namespace Ubitrack::Drivers

#endif
