/*
 * Ubitrack - Library for Ubiquitous Tracking
 * Copyright 2006, Technische Universitaet Muenchen, and individual
 * contributors as indicated by the @authors tag. See the
 * copyright.txt in the distribution for a full listing of individual
 * contributors.
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA, or see the FSF site: http://www.fsf.org.
 */


#include <iostream>
#include <sstream>
#include <algorithm>
#include <Ubitrack/Dataflow/ComponentFactory.h>
#include <log4cpp/Category.hh>
#include <Ubitrack/Dataflow/PushConsumer.h>
#include <Ubitrack/Dataflow/PullConsumer.h>
#include <Ubitrack/Dataflow/PushSupplier.h>
#include <Ubitrack/Dataflow/Component.h>
#include <Ubitrack/Dataflow/ComponentFactory.h>
#include <Ubitrack/Measurement/Measurement.h>
#include <Ubitrack/Calibration/2D3DPoseEstimation.h>
#include <Ubitrack/Calibration/LensDistortion.h>
#include <log4cpp/Category.hh>

#include <Ubitrack/Calibration/Function/MultipleCameraProjection.h>
#include <Ubitrack/Math/UnscentedTransform.h>

static log4cpp::Category& logger( log4cpp::Category::getInstance( "Drivers.ArtErrorEstimation" ) );

/* The pattern:
<Pattern name="ArtErrorEstimator" id="ArtEE1">

	<Input>
		<!-- body information -->
		<Node name="Art"/>
		<Node name="Body"/>
		<Node name="Points"/>

		<Edge name="BodyPose" source="Art" destination="Body">
			<Predicate>type==6D&amp;&amp;mode==push</Predicate>
		</Edge>
		<Edge name="BodyPoints" source="Body" destination="Points">
			<Predicate>type==PositionList&amp;&amp;mode==pull</Predicate>
		</Edge>

		<!-- camera 1 information (repeat for n cameras) -->
		<Node name="Camera1"/>
		<Node name="ImagePlane1"/>

		<Edge name="CameraPose1" source="Camera1" destination="Art">
			<Predicate>type==6D&amp;&amp;mode==pull</Predicate>
		</Edge>
		<Edge name="Intrinsics1" source="Camera1" destination="ImagePlane1">
			<Predicate>type==Matrix3x3&amp;&amp;mode==pull</Predicate>
		</Edge>
		<Edge name="Distortion1" source="Camera1" destination="ImagePlane1">
			<Predicate>type==Vector4D&amp;&amp;mode==pull</Predicate>
		</Edge>
		<Edge name="ImagePoints1" source="ImagePlane1" destination="Points">
			<Predicate>type==PositionList2&amp;&amp;mode==push</Predicate>
		</Edge>
	</Input>
	<Output>
		<Node name="BodyCentroid"/>
		<Edge name="ErrorPose" source="Art" destination="BodyCentroid">
			<Attribute name="type" value="6DError"/>
			<Attribute name="mode" value="push"/>
		</Edge>
		<Edge name="MeanVisibleTarget" source="Body" destination="BodyCentroid">
			<Attribute name="type" value="6D"/>
			<Attribute name="mode" value="push"/>
		</Edge>
	</Output>
	<Constraints>
		<Correspondence>
			<Node node-ref="Camera"/>
			<Node node-ref="ImagePlane"/>
			<Edge edge-ref="CameraPose"/>
			<Edge edge-ref="Intrinsics"/>
			<Edge edge-ref="Distortion"/>
			<Edge edge-ref="ImagePoints"/>
		</Correspondence>
	</Constraints>

	<DataflowConfiguration>
		<UbitrackLib class="ArtErrorEstimator"/>
	</DataflowConfiguration>

</Pattern>
*/

namespace Ubitrack { namespace Drivers {

class ArtErrorEstimator
	: public Dataflow::Component
{
public:
	/**
	 * UTQL component constructor.
	 *
	 * @param nm Unique name of the component.
	 * @param cfg UTQL subgraph
	 */
	ArtErrorEstimator( const std::string& sName, boost::shared_ptr< Graph::UTQLSubgraph > subgraph )
      : Dataflow::Component( sName )
	  , m_inBodyPose( "BodyPose", *this, boost::bind( &ArtErrorEstimator::receiveBodyPose, this, _1 ) )
	  , m_inBodyPoints( "BodyPoints", *this )
      , m_outErrorPose( "ErrorPose", *this )
      , m_outMeanVisibleTarget( "MeanVisibleTarget", *this )
    {
		// dynamically create per-camera ports
		for ( Graph::UTQLSubgraph::EdgeMap::iterator it = subgraph->m_Edges.begin(); it != subgraph->m_Edges.end(); it++ )
			if ( it->second->isInput() )
				if ( 0 == it->first.compare( 0, 10, "CameraPose" ) )
					m_inCameraPose.push_back( boost::shared_ptr< Dataflow::PullConsumer< Measurement::Pose > >( 
						new Dataflow::PullConsumer< Measurement::Pose >( it->first, *this ) ) );
				else if ( 0 == it->first.compare( 0, 10, "Intrinsics" ) )
					m_inIntrinsics.push_back( boost::shared_ptr< Dataflow::PullConsumer< Measurement::Matrix3x3 > >( 
						new Dataflow::PullConsumer< Measurement::Matrix3x3 >( it->first, *this ) ) );
				else if ( 0 == it->first.compare( 0, 10, "Distortion" ) )
					m_inDistortion.push_back( boost::shared_ptr< Dataflow::PullConsumer< Measurement::Vector4D > >( 
						new Dataflow::PullConsumer< Measurement::Vector4D >( it->first, *this ) ) );
				else if ( 0 == it->first.compare( 0, 11, "ImagePoints" ) )
					m_inImagePoints.push_back( boost::shared_ptr< Dataflow::PushConsumer< Measurement::PositionList2 > >( 
						new Dataflow::PushConsumer< Measurement::PositionList2 >( it->first, *this,
							boost::bind( &ArtErrorEstimator::receiveImagePoints, this, _1, m_inImagePoints.size() ) ) ) );

		// check if there are an equal number of ports
		if ( m_inCameraPose.size() != m_inIntrinsics.size() ||
			m_inCameraPose.size() != m_inDistortion.size() ||
			m_inCameraPose.size() != m_inImagePoints.size() )
			UBITRACK_THROW( "Invalid number of ports on correspondence" );

		// sort ports alphabetically to assure same order (trackman screws it up)
		std::sort( m_inCameraPose.begin(), m_inCameraPose.end(), PortsCompare() );
		std::sort( m_inIntrinsics.begin(), m_inIntrinsics.end(), PortsCompare() );
		std::sort( m_inDistortion.begin(), m_inDistortion.end(), PortsCompare() );
		std::sort( m_inImagePoints.begin(), m_inImagePoints.end(), PortsCompare() );

		// vector of last image points
		m_imagePoints.resize( m_inImagePoints.size() );
	}

protected:

	// needed to sort the ports alphabetically
	struct PortsCompare
	{
		template< class T >
		bool operator()( const boost::shared_ptr< T >& a, const boost::shared_ptr< T >& b )
		{ return a->getName() < b->getName(); }
	};

	/** receives body pose measurements from dtrack */
	void receiveBodyPose( const Measurement::Pose& m )
	{
		m_bodyPose = m;

		// skip the checkCompute function as long as the timestamp synchronization does not fully 
		// work and compute for every 6D pose.
		for ( unsigned i = 0; i < m_imagePoints.size(); i++ )
			if ( !m_imagePoints[ i ] )
				return;
		compute( m.time() );

		//checkCompute();
	}

	/** receives image point measurements */
	void receiveImagePoints( const Measurement::PositionList2& m, std::size_t id )
	{
		m_imagePoints.at( id ) = m;
		//checkCompute();
	}

	/** checks if all ports have valid measurements */
	void checkCompute();

	/** compute the result */
	void compute( const Measurement::Timestamp );

	// data received
	Measurement::Pose m_bodyPose;
	std::vector< Measurement::PositionList2 > m_imagePoints;
	Measurement::Timestamp m_lastComputedTime;

	// Fixed input ports
	Dataflow::PushConsumer< Measurement::Pose > m_inBodyPose;
	Dataflow::PullConsumer< Measurement::PositionList > m_inBodyPoints;

	// per-camera input ports
	std::vector< boost::shared_ptr< Dataflow::PullConsumer< Measurement::Pose > > > m_inCameraPose;
	std::vector< boost::shared_ptr< Dataflow::PullConsumer< Measurement::Matrix3x3 > > > m_inIntrinsics;
	std::vector< boost::shared_ptr< Dataflow::PullConsumer< Measurement::Vector4D > > > m_inDistortion;
	std::vector< boost::shared_ptr< Dataflow::PushConsumer< Measurement::PositionList2 > > > m_inImagePoints;
	
	/** Output port of the component. */
	Dataflow::PushSupplier< Measurement::ErrorPose > m_outErrorPose;
	Dataflow::PushSupplier< Measurement::Pose > m_outMeanVisibleTarget;
};


void ArtErrorEstimator::checkCompute()
{
	// check for synchronized push, but with a small difference in the timestamps allowed
	if ( !m_bodyPose )
		return;
	Measurement::Timestamp t = m_bodyPose.time();
	
	// only send measurement once for a body
	if ( t == m_lastComputedTime )
		return;

	// check image points
	Measurement::Timestamp eps = 10000000; // 10ms, until I find a better way for synchronizing camera with dtrack measurements
	for ( std::size_t i = 0; i < m_imagePoints.size(); i++ )
	{
		if ( m_imagePoints[ i ] )
			LOG4CPP_TRACE( logger, "Camera " << i << " time difference (ms): " << double( static_cast< long long >( m_imagePoints[ i ].time() - t ) ) * 1e-6 );

		if ( !m_imagePoints[ i ] || m_imagePoints[ i ].time() + eps < t || t + eps < m_imagePoints[ i ].time() )
			return;
	}

	m_lastComputedTime = t;
	compute( t );
}


void ArtErrorEstimator::compute( const Measurement::Timestamp t )
{
	LOG4CPP_TRACE( logger, "compute()" );

	// Correction factor for estimates
	const double fCorrFactor = 2.0;

	// magic constants
	const double fImgErr = 0.005; // * fCorrFactor;
	const double fMaxImgDist = 3.0;

	// get body points
	Measurement::PositionList bodyPoints = m_inBodyPoints.get( t );

	// build observation list
	std::vector< Math::Matrix< 3, 4 > > cameraMatrices;
	std::vector< std::pair< unsigned, unsigned > > observations;
	std::vector< Math::Vector< 2 > > measurements;
	
	// get camera setup vectors
	std::vector< Math::Pose > cameraPoses;
	std::vector< Math::Matrix< 3, 3 > > cameraIntrinsics;
	std::vector< Math::Vector< 4 > > cameraDistortions;
	
	// loop over cameras
	for ( unsigned i = 0; i < m_imagePoints.size(); i++ )
	{
		// get camera projection
		Measurement::Pose camPose = m_inCameraPose[ i ]->get( t );
		Measurement::Matrix3x3 intrinsics = m_inIntrinsics[ i ]->get( t );
		Measurement::Vector4D distortion = m_inDistortion[ i ]->get( t );
		
		// store camera parameters
		cameraPoses.push_back( *camPose );
		cameraIntrinsics.push_back( *intrinsics );
		cameraDistortions.push_back( *distortion );
		
		// build projection matrix
		cameraMatrices.push_back( Math::Matrix< 3, 4 >( 
			boost::numeric::ublas::prod( *intrinsics, Math::Matrix< 3, 4 >( *camPose ) ) ) );

		std::string sVisMatrixRow;

		// check for point visibility
		Math::Pose bodyToCam( *camPose * *m_bodyPose );
		// loop over body points / markers
		for ( unsigned j = 0; j < bodyPoints->size(); j++ )
		{
			// project point
			Math::Vector< 2 > p2d = Calibration::projectWithDistortion( bodyToCam * (*bodyPoints)[ j ], *distortion, *intrinsics );

			// check if among the visible points
			bool bVisible = false;
			// loop over image points from this camera
			for ( std::vector< Math::Vector< 2 > >::iterator it = m_imagePoints[ i ]->begin(); it != m_imagePoints[ i ]->end(); it++ )
			{
				Math::Vector< 2 > diff( *it - p2d );
				if ( diff( 0 ) * diff( 0 ) + diff( 1 ) * diff( 1 ) < fMaxImgDist )
				{
					// Store visibility
					observations.push_back( std::pair< unsigned, unsigned >( j, i ) );
					// Store corresponding measurement
					measurements.push_back( *it );
					bVisible = true;
					// body point found for this camera
					break;
				}
			}
			
			sVisMatrixRow += bVisible ? " * " : " . ";
		}

		LOG4CPP_DEBUG( logger, "Camera " << i << " visibility: " << sVisMatrixRow );
	}

	if ( observations.size() < 4 )
	{
		LOG4CPP_INFO( logger, "Not enough observations" );
	}
	else
	{
		// Get mean position of visible markers
		Math::Vector< 3 > visibleTargetMean( Math::Vector< 3 >( 0, 0, 0 ) );
		unsigned visibleBodyPoints = 0;

		// Get rest of visible markers (note: observations are not in order)
		std::string sVisMatrixRow;
		for ( unsigned j = 0; j < bodyPoints->size(); j++ )
		{
			bool bVisible = false;
			for ( unsigned k = 0; k < observations.size(); k++ )
			{
				if (observations[k].first == j)
				{
					//LOG4CPP_TRACE( logger, "Body point visible: " << j );
					visibleTargetMean += (*bodyPoints)[ j ];
					visibleBodyPoints++;
					bVisible = true;
					break;
				}
			}
			sVisMatrixRow += bVisible ? " * " : " . ";
		}
		LOG4CPP_DEBUG( logger, "Marker visibility:   " << sVisMatrixRow );

		// Get visible target mean and store as pose
		visibleTargetMean /= visibleBodyPoints;
		Math::Pose meanToBody( Math::Quaternion(), -visibleTargetMean );

		// magic constant
		const double fMaxDeviation = 0.001; // 1mms

		// Output deviation
		double deviation = 0;
		for ( unsigned m = 0; m < 3; m++)
			deviation += visibleTargetMean( m ) * visibleTargetMean( m );
		deviation = sqrt(deviation);
		if ( deviation < fMaxDeviation )
			LOG4CPP_DEBUG( logger, "Visible target mean: (no deviation)")
		else
			LOG4CPP_DEBUG( logger, "Visible target mean: " << visibleTargetMean );
		
		// get pose for computation (= at visible target mean position)
		Math::Pose compPose;
		compPose = (*m_bodyPose) * meanToBody;
		
		// get new marker positions for computation
		std::vector< Math::Vector< 3 > > compBodyPoints;
		for ( unsigned j = 0; j < bodyPoints->size(); j++ )
		{
			Math::Vector< 3 > oldPosition = (*bodyPoints)[ j ];
			Math::Vector< 3 > newPosition = meanToBody * oldPosition;
			compBodyPoints.push_back( newPosition );
		}

		// Create multiple camera projection container
		Calibration::Function::MultipleCameraProjection<> mcp(
			compBodyPoints, cameraPoses, cameraIntrinsics, cameraDistortions, observations );

		// compute covariance (using Unscented Transform)
		Math::Matrix< 6, 6 > cov = Ubitrack::Math::unscentedTransform(
			measurements, fImgErr * fImgErr, mcp );

		// compute covariance (wrong way: use full target centroid)
		// Math::Matrix< 6, 6 > cov = Calibration::multipleCameraPoseError( *m_bodyPose, *bodyPoints, cameraMatrices, 
			// observations, fImgErr * fImgErr );

		// compute covariance (correct way: use visible target centroid)
		// Math::Matrix< 6, 6 > cov = Calibration::multipleCameraPoseError( compPose, compBodyPoints, cameraMatrices, 
			// observations, fImgErr * fImgErr );
		
		// push resulting ErrorPose
		m_outErrorPose.send( Measurement::ErrorPose( t, Math::ErrorPose( *m_bodyPose, cov ) ) );

		// push resulting MeanVisibleTarget
		m_outMeanVisibleTarget.send( Measurement::Pose( t, meanToBody ) );
	}
}


// register module at factory
UBITRACK_REGISTER_COMPONENT( Dataflow::ComponentFactory* const cf ) 
{
	cf->registerComponent< ArtErrorEstimator > ( "ArtErrorEstimator" );
}

} } // namespace Ubitrack::Drivers
