/*
 * Ubitrack - Library for Ubiquitous Tracking
 * Copyright 2006, Technische Universitaet Muenchen, and individual
 * contributors as indicated by the @authors tag. See the
 * copyright.txt in the distribution for a full listing of individual
 * contributors.
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA, or see the FSF site: http://www.fsf.org.
 */

/**
 * @ingroup driver_components
 * @file
 * Components that read an ART body file.
 *
 * @author Daniel Pustka <daniel.pustka@in.tum.de>
 */

#include <fstream>
#include <boost/bind.hpp>
#include <Ubitrack/Measurement/Measurement.h>
#include <Ubitrack/Dataflow/Component.h>
#include <Ubitrack/Dataflow/ComponentFactory.h>
#include <Ubitrack/Dataflow/PullSupplier.h>

using namespace Ubitrack::Dataflow;

namespace Ubitrack { namespace Drivers {

/**
 * @ingroup driver_components
 * Reads an ART body file and provides the list of 3D points as pull.
 *
 * @par Input Ports
 * None.
 *
 * @par Output Ports
 * PullSupplier< PositionList > with name "Output".
 *
 * @par Configuration
@verbatim
<DataflowConfiguration>
	<Attribute name="file" value="<filename>"/>
</DataflowConfiguration>
@endverbatim
 */
class ArtBodyFileReader
	: public Component
{
public:
	/** open the file */
	ArtBodyFileReader( const std::string& name, boost::shared_ptr< Graph::UTQLSubgraph > subgraph )
		: Component( name )
		, m_outPort( "Output", *this, boost::bind( &ArtBodyFileReader::request, this, _1 ) )
		, m_Measurement( Measurement::PositionList::value_type() )
	{
		// read filename from configuration
		std::string m_sFilename = subgraph->m_DataflowAttributes.getAttributeString( "file" );
		if ( m_sFilename.empty() )
			UBITRACK_THROW( "No \"file\" attribute for ArtBodyFileReader component " + name );

		// create ifstream
		std::ifstream stream( m_sFilename.c_str() );
		if ( !stream.good() )
			UBITRACK_THROW( "Could not open file " + m_sFilename + " for reading" );

		// read number of points
		unsigned nPoints;
		stream >> nPoints;

		// read points
		for ( unsigned i = 0; i < nPoints; i++ )
		{
			Math::Vector< 3 > pos;
			stream >> pos( 0 ) >> pos( 1 ) >> pos ( 2 );
			pos /= 1000.0;
			m_Measurement->push_back( pos );
		}
	}

protected:
	/** output port */
	Dataflow::PullSupplier< Measurement::PositionList > m_outPort;

	/** mutex to protect file from simultaneous writes */
	Measurement::PositionList m_Measurement;

	/** handler method for incoming pull requests */
	Measurement::PositionList request( Measurement::Timestamp t )
	{ return Measurement::PositionList( t, *m_Measurement ); }
};


} } // namespace Ubitrack::Drivers


UBITRACK_REGISTER_COMPONENT( Ubitrack::Dataflow::ComponentFactory* const cf )
{
	cf->registerComponent< Ubitrack::Drivers::ArtBodyFileReader > ( "ArtBodyFileReader" );
}
