/*
 * Ubitrack - Library for Ubiquitous Tracking
 * Copyright 2006, Technische Universitaet Muenchen, and individual
 * contributors as indicated by the @authors tag. See the
 * copyright.txt in the distribution for a full listing of individual
 * contributors.
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA, or see the FSF site: http://www.fsf.org.
 */


#include "ArtSenderModule.h"
#include <iostream>
#include <sstream>
#include <iomanip>
#include <Ubitrack/Dataflow/ComponentFactory.h>
#include <Ubitrack/Math/Vector.h>
#include <Ubitrack/Math/Quaternion.h>
#include <Ubitrack/Math/Pose.h>
#include <Ubitrack/Math/Matrix.h>

#include <boost/array.hpp>
#include <boost/thread.hpp>

namespace Ubitrack { namespace Drivers {

static log4cpp::Category& logger( log4cpp::Category::getInstance( "Drivers.ArtSender" ) );

ArtSenderModule::ArtSenderModule( const ArtSenderModuleKey& moduleKey, boost::shared_ptr< Graph::UTQLSubgraph >, FactoryHelper* pFactory )
	: Module< ArtSenderModuleKey, ArtSenderComponentKey, ArtSenderModule, ArtSenderComponent >( moduleKey, pFactory )
{
	stopModule();
}


ArtSenderModule::~ArtSenderModule()
{
	stopModule();
}

void ArtSenderModule::startModule()
{
    using asio::ip::udp;

	if ( !m_running )
	{
		m_running = true;
		using asio::ip::udp;

		LOG4CPP_NOTICE( logger, "Starting ART Sender network service" );

		m_IoService.reset( new asio::io_service );

		LOG4CPP_NOTICE( logger, "Opening receive socket" );
		// open new receive socket on wich we listen for new commands
		m_ReceiveSocket.reset( new udp::socket( *m_IoService ) );
		m_ReceiveSocket->open( udp::v4() );
		asio::socket_base::reuse_address option( true );
		m_ReceiveSocket->set_option( option );
		m_ReceiveSocket->bind( udp::endpoint( udp::v4(), m_moduleKey.m_ReceivePort ) );

		LOG4CPP_NOTICE( logger, "Opening sender socket" );
		// open new socket which we use for sending stuff
		m_SendSocket.reset( new udp::socket( *m_IoService ) );
		m_SendSocket->open( udp::v4() );

		LOG4CPP_NOTICE( logger, "Get Resolver" );
		// resolve destination pair and store the remote endpoint
		udp::resolver resolver( *m_IoService );

		LOG4CPP_NOTICE( logger, "Parse module key" );
		std::ostringstream portString;
		portString << m_moduleKey.m_Port;

		LOG4CPP_NOTICE( logger, "Resolving send endpoint" );
		udp::resolver::query query( udp::v4(), m_moduleKey.m_Destination, portString.str() );
		m_SendEndpoint.reset( new udp::endpoint( *resolver.resolve( query ) ) );


		LOG4CPP_NOTICE( logger, "Receive from udp receive port" );
		// recieve udp packet from recive port
		m_ReceiveSocket->async_receive_from (
			asio::buffer ( receive_data, max_receive_length ),
			sender_endpoint,
			boost::bind (&ArtSenderModule::HandleReceive, this, asio::placeholders::error, asio::placeholders::bytes_transferred));

		LOG4CPP_NOTICE( logger, "Start thread" );
		// network thread runs until io_service is interrupted
		m_NetworkThread = boost::shared_ptr< boost::thread >( new boost::thread ( boost::bind ( &asio::io_service::run, m_IoService.get() ) ) );
	}
}

void ArtSenderModule::stopModule()
{
	if ( m_running )
	{
		m_running = false;
		LOG4CPP_NOTICE( logger, "Stopping ART Sender network service" );

		if ( m_IoService )
		{
			m_IoService->stop();

			if ( m_NetworkThread )
			{
				m_NetworkThread->join();
				m_NetworkThread.reset();
			}

			if (m_ReceiveSocket)
			{
				m_ReceiveSocket->close();
				m_ReceiveSocket.reset();
			}

			if (m_SendSocket)
			{
				m_SendSocket->close();
				m_SendSocket.reset();
			}

			if (m_SendEndpoint)
			{
				m_SendEndpoint.reset();
			}
			m_IoService.reset();
		}
	}
}

void ArtSenderModule::HandleReceive (const asio::error_code err, size_t length)
{
    static unsigned int frame_counter = 0;

    if ( err && err != asio::error::message_size )
    {
        std::ostringstream msg;
        msg << "Error receiving from socket: \"" << err << "\"";
        UBITRACK_THROW( msg.str() );
    }

    if (length >= max_receive_length)
    {
        UBITRACK_THROW( "FIXME: received more than max_receive_length bytes." );
    }

    m_ReceiveSocket->async_receive_from (
        asio::buffer ( receive_data, max_receive_length ),
        sender_endpoint,
        boost::bind (&ArtSenderModule::HandleReceive, this, asio::placeholders::error, asio::placeholders::bytes_transferred));

    // make receive data null terminated
    receive_data[length] = 0;

    // parse dtrack command
    const char* dtrack_magic = "dtrack ";
    if ( strncmp( receive_data, dtrack_magic, strlen( dtrack_magic ) ) != 0 )
        return; // udp packet does not start with magic

    std::string data ( receive_data+strlen( dtrack_magic ) );
    std::vector<int> codes;

    {
        unsigned int lastpos = 0;
        unsigned int findpos = 0;
        while ( ( findpos = data.find ( " ", lastpos ) ) != std::string::npos )
        {
            std::string s = data.substr ( lastpos, ( findpos-lastpos ) );
            codes.push_back( atol( s.c_str() ) );
            lastpos = findpos+1;
        }
        std::string s = data.substr ( lastpos );
        codes.push_back( atol ( s.c_str() ) );
    }

    // codes:
    // 10 0: stop measurements
    // 10 1: start measurements without tracking calculations
    // 10 3: start measurements with tracking calculations
    // 31 : start continous output
    // 32 : stop continous output
    // 33 x: send x data packets

    // start: 10,3 + 31
    // stop: 32 + 10,0

    // for now we only implement "33 1" and ignore everything else
    if ((codes[0] == 33) && (codes[1] == 1))
    {
        std::ostringstream packet;
        packet.precision(3);
        packet.flags(std::ios::fixed);

        packet << "fr " << frame_counter++ << "\r\n";

        std::vector< boost::shared_ptr< ArtSenderComponent > > components_6d;
        std::vector< boost::shared_ptr< ArtSenderComponent > > components_6d_flystick;
        std::vector< boost::shared_ptr< ArtSenderComponent > > components_6d_mt;

		ComponentList clist = getAllComponents();
		for (ComponentList::iterator it = clist.begin(); it != clist.end(); ++it )
		{
			switch( (*it)->getKey().getTargetType() )
			{
			case ArtSenderComponentKey::target_6d:
				components_6d.push_back( *it );
				break;
			case ArtSenderComponentKey::target_6d_flystick:
				components_6d_flystick.push_back( *it );
				break;
			case ArtSenderComponentKey::target_6d_measurement_tool:
				components_6d_mt.push_back( *it );
				break;
			default:
				break;
			}
		}

        if (components_6d.size() != 0) {
            packet << "6d " << components_6d.size() << " ";
            for (std::vector< boost::shared_ptr< ArtSenderComponent > >::iterator it = components_6d.begin();
                 it != components_6d.end(); ++it) {
                boost::shared_ptr< ArtSenderComponent > comp = *it;

                Ubitrack::Measurement::Pose p = comp->getPose();
                Ubitrack::Math::Vector<3> pos = p->translation();
                Ubitrack::Math::Quaternion rot = p->rotation();
                Ubitrack::Math::Matrix<4,4> rotMat (rot, pos);
                packet << "[" << (comp->getKey().getBody())-1 << " 1.000][";
                packet << pos[0]*1000.0 << " " << pos[1]*1000.0 << " " << pos[2]*1000.0 << " "
                       << "0.000 0.000 0.000][";
                packet << rotMat( 0,0 ) << " " << rotMat( 1,0 ) << " " << rotMat( 2,0 ) << " "
                       << rotMat( 0,1 ) << " " << rotMat( 1,1 ) << " " << rotMat( 2,1 ) << " "
                       << rotMat( 0,2 ) << " " << rotMat( 1,2 ) << " " << rotMat( 2,2 ) << "] ";
            }
            packet << "\r\n";
        }

        if (components_6d_flystick.size() != 0) {
            packet << "6df " << components_6d_flystick.size() << " ";
            for (std::vector< boost::shared_ptr< ArtSenderComponent > >::iterator it = components_6d_flystick.begin();
                 it != components_6d_flystick.end(); ++it) {
                boost::shared_ptr< ArtSenderComponent > comp = *it;

                Ubitrack::Measurement::Pose p = comp->getPose();
                Ubitrack::Math::Vector<3> pos = p->translation();
                Ubitrack::Math::Quaternion rot = p->rotation();
                Ubitrack::Math::Matrix<4,4> rotMat (rot, pos);
                packet << "[" << (comp->getKey().getBody())-1 << " 1.000 0][";
                packet << pos[0]*1000.0 << " " << pos[1]*1000.0 << " " << pos[2]*1000.0 << " "
                       << "0.000 0.000 0.000][";
                packet << rotMat( 0,0 ) << " " << rotMat( 1,0 ) << " " << rotMat( 2,0 ) << " "
                       << rotMat( 0,1 ) << " " << rotMat( 1,1 ) << " " << rotMat( 2,1 ) << " "
                       << rotMat( 0,2 ) << " " << rotMat( 1,2 ) << " " << rotMat( 2,2 ) << "] ";
            }
            packet << "\r\n";
        }

        if (components_6d_mt.size() != 0) {
            packet << "6dmt " << components_6d_mt.size() << " ";
            for (std::vector< boost::shared_ptr< ArtSenderComponent > >::iterator it = components_6d_mt.begin();
                 it != components_6d_mt.end(); ++it) {
                boost::shared_ptr< ArtSenderComponent > comp = *it;

                Ubitrack::Measurement::Pose p = comp->getPose();
				Ubitrack::Math::Vector<3> pos = p->translation();
				Ubitrack::Math::Quaternion rot = p->rotation();
				Ubitrack::Math::Matrix<4,4> rotMat (rot, pos);
				packet << "[" << (comp->getKey().getBody())-1 << " 1.000 0][";
				packet << pos[0]*1000.0 << " " << pos[1]*1000.0 << " " << pos[2]*1000.0 << "][";
				packet << rotMat( 0,0 ) << " " << rotMat( 1,0 ) << " " << rotMat( 2,0 ) << " "
					   << rotMat( 0,1 ) << " " << rotMat( 1,1 ) << " " << rotMat( 2,1 ) << " "
					   << rotMat( 0,2 ) << " " << rotMat( 1,2 ) << " " << rotMat( 2,2 ) << "] ";
            }
            packet << "\r\n";
        }

        std::string outputString = packet.str();
        m_SendSocket->send_to( asio::buffer(outputString.c_str(), outputString.size()), *m_SendEndpoint );
    }
}

// register module at factory
UBITRACK_REGISTER_COMPONENT( ComponentFactory* const cf ) {
	cf->registerModule< ArtSenderModule > ( "ArtSender" );
}

} } // namespace Ubitrack::Drivers
