/*
 * Ubitrack - Library for Ubiquitous Tracking
 * Copyright 2006, Technische Universitaet Muenchen, and individual
 * contributors as indicated by the @authors tag. See the
 * copyright.txt in the distribution for a full listing of individual
 * contributors.
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA, or see the FSF site: http://www.fsf.org.
 */


#ifndef __ArtSenderModule_h_INCLUDED__
#define __ArtSenderModule_h_INCLUDED__

#include <string>
#include <cstdlib>

// on windows, asio must be included before anything that possible includes windows.h
// don't ask why.
#include <asio.hpp>

#include <boost/bind.hpp>
#include <boost/thread.hpp>

#include <Ubitrack/Dataflow/PushConsumer.h>
#include <Ubitrack/Dataflow/Component.h>
#include <Ubitrack/Dataflow/Module.h>
#include <Ubitrack/Measurement/Measurement.h>

namespace Ubitrack { namespace Drivers {

using namespace Dataflow;

// forward declaration
class ArtSenderComponent;

/**
 * Module key for art sender.
 * Represents the port number to which to send UDP packtes, the
 * port on which to listen for commands and the destion where to send
 * the tracking data.
 */
class ArtSenderModuleKey
{
public:
	ArtSenderModuleKey( boost::shared_ptr< Graph::UTQLSubgraph > subgraph )
        : m_Port( 5000 ) // default port is 5000
        , m_ReceivePort( 6000 ) // default receive port is 6000
        , m_Destination( "127.0.0.1" )
    {
		Graph::UTQLSubgraph::NodePtr senderNode = subgraph->getNode( "ArtSender" );
		senderNode->getAttributeData( "ArtPort", m_Port );
		senderNode->getAttributeData( "ArtReceivePort", m_ReceivePort );
		if ( senderNode->hasAttribute( "ArtDestinationIP" ) )
		{
			m_Destination = senderNode->getAttribute( "ArtDestinationIP" ).getText();
		}
	}

	// less than operator for map
	bool operator<( const ArtSenderModuleKey& b ) const
	{
        if ( m_Destination != b.m_Destination )
            return m_Destination < b.m_Destination;
        else if ( m_Port != b.m_Port )
            return m_Port < b.m_Port;
        else
            return m_ReceivePort < b.m_ReceivePort;
    }

    int m_Port;
    int m_ReceivePort;
    std::string m_Destination;
};


/**
 * Component key for art sender.
 * Represents the body number and target type
 */
class ArtSenderComponentKey
{
public:
    enum TargetType { target_6d, target_6d_flystick, target_6d_measurement_tool };

	// construct from configuration
	ArtSenderComponentKey( boost::shared_ptr< Graph::UTQLSubgraph > subgraph)
	{
        if ((subgraph->m_DataflowConfiguration.getXML() == NULL))
            UBITRACK_THROW( "Art target without configuration" );

        const TiXmlNode* node = subgraph->m_DataflowConfiguration.getXML()->FirstChild( "ArtTarget" );
        if ((node == NULL) || (node->ToElement() == NULL))
            UBITRACK_THROW( "Art target without xml element \"ArtTarget\"" );

        const char* idString = node->ToElement()->Attribute("id");
        if (idString == NULL)
            UBITRACK_THROW( "Art target without id attribute in ArtTarget element " );

        m_body = atol(idString);
        if (m_body <= 0)
            UBITRACK_THROW( "Art target with invalid body id (must be > 0)");

        const char* typeString = node->ToElement()->Attribute("type");
        if (typeString == NULL)
        {
            // no explicit art target type information. so we assume 6D
            m_targetType = target_6d;
        }
        else
        {
            std::string cmpString (typeString);
            if ( cmpString == "6d" )
                m_targetType = target_6d;
            else if ( cmpString == "6df" )
                m_targetType = target_6d_flystick;
            else if ( cmpString == "6dmt" )
                m_targetType = target_6d_measurement_tool;
            else
                UBITRACK_THROW( "Art target with unknown target type: " + cmpString );
        }
	}

	// construct from body number
	ArtSenderComponentKey( int a )
		: m_body( a )
        , m_targetType( target_6d )
	{}

    // construct from body number and target type
    ArtSenderComponentKey( int a, TargetType t )
        : m_body( a )
        , m_targetType( t )
    {}

	int getBody() const
	{ return m_body; }

    TargetType getTargetType() const
    {
        return m_targetType;
    }

	// less than operator for map
	bool operator<( const ArtSenderComponentKey& b ) const
    {
        if ( m_targetType == b.m_targetType )
            return m_body < b.m_body;
        else
            return m_targetType < b.m_targetType;
    }

protected:
	int m_body;
    TargetType m_targetType;
};


/**
 * Module for ART tracker.
 * Does all the work
 */
class ArtSenderModule
	: public Module< ArtSenderModuleKey, ArtSenderComponentKey, ArtSenderModule, ArtSenderComponent >
{
public:
	/** constructor */
	ArtSenderModule( const ArtSenderModuleKey& key, boost::shared_ptr< Graph::UTQLSubgraph >, FactoryHelper* pFactory );

	/** destruktor */
	~ArtSenderModule();

	/** module stop method  */
	virtual void stopModule();

	/** module start method */
	virtual void startModule();

// 	/** thread method */
    void HandleReceive (const asio::error_code err, size_t length);

protected:
	boost::shared_ptr< asio::io_service > m_IoService;

    boost::shared_ptr< asio::ip::udp::socket > m_ReceiveSocket;

    boost::shared_ptr< asio::ip::udp::socket > m_SendSocket;
    boost::shared_ptr< asio::ip::udp::endpoint > m_SendEndpoint;

    // Recive data. Do not touch from outside of async network thread
    enum { max_receive_length = 10240, receive_buffer_size = 10242 };
    char receive_data[receive_buffer_size];
    asio::ip::udp::endpoint sender_endpoint;

    boost::shared_ptr< boost::thread > m_NetworkThread;
};


/**
 * Component for ART tracker.
 * Does nothing but provide a push port
 */
class ArtSenderComponent
	: public ArtSenderModule::Component
{
public:
	/** constructor */
	ArtSenderComponent( const std::string& name, boost::shared_ptr< Graph::UTQLSubgraph >, const ArtSenderComponentKey& componentKey, ArtSenderModule* pModule )
		: ArtSenderModule::Component( name, componentKey, pModule )
		, m_port( "Input", *this, boost::bind( &ArtSenderComponent::poseIn, this, _1 ) )
		, m_pose( Math::Pose() )
	{}

	/** returns the port for usage by the module */
	PushConsumer< Ubitrack::Measurement::Pose >& getPort()
	{ return m_port; }

    // return the last pose for use in the sender module
    const Ubitrack::Measurement::Pose getPose()
    {
        boost::mutex::scoped_lock l( m_PoseMutex );
        return m_pose;
    }

protected:
    // Pose mutex to syncronize reading and writing
    boost::mutex m_PoseMutex;

    // receive a new pose from the dataflow
    void poseIn( const Ubitrack::Measurement::Pose& m )
    {
        boost::mutex::scoped_lock l( m_PoseMutex );
        m_pose = m;
    }


    // consumer port
    PushConsumer< Ubitrack::Measurement::Pose > m_port;

    // store the last pose pushed to the component
    Ubitrack::Measurement::Pose m_pose;
};

} } // namespace Ubitrack::Drivers

#endif
