/*
 * Ubitrack - Library for Ubiquitous Tracking
 * Copyright 2006, Technische Universitaet Muenchen, and individual
 * contributors as indicated by the @authors tag. See the 
 * copyright.txt in the distribution for a full listing of individual
 * contributors.
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA, or see the FSF site: http://www.fsf.org.
 */


/**
 * @ingroup driver_components
 * @file
 * GPS driver
 * This file contains the driver component to
 * get NMEA location data from a GPS receiver.
 *
 * The received data is sent via a push interface.
 *
 * @author Florian Echtler <echtler@in.tum.de>
 */

#include "GPSDriver.h"
#include <Ubitrack/Util/Exception.h>
#include <sstream>
#include <algorithm>
#include <iomanip>


// M_PI is not part of ISO C/C++
static const double g_pi = 3.14159265358979323846;


static log4cpp::Category& logger( log4cpp::Category::getInstance( "Drivers.GPS" ) );

namespace Ubitrack { namespace Drivers {


using namespace Dataflow;


/** constructor */
GPSDriver::GPSDriver( const std::string& sName, boost::shared_ptr< Graph::UTQLSubgraph > subgraph )
	: Component( sName )
	, m_output( "Output", *this )
	, m_keepRunning(1)
{
	#ifdef _WIN32
		std::string portName("COM1");
	#else // *nix
		std::string portName("/dev/ttyS0");
	#endif
	int baudRate = 9600;

	if (subgraph->m_DataflowConfiguration.getXML() == NULL) 
		UBITRACK_THROW( "No configuration for GPS component." );

	const TiXmlNode* ref = subgraph->m_DataflowConfiguration.getXML()->FirstChild("ReferencePoint");
	if ((ref == NULL) || (ref->ToElement() == NULL))
		UBITRACK_THROW( "No ReferencePoint element in GPS component configuration." );

	if ( ref->ToElement()->QueryDoubleAttribute( "latitude",  &m_offset[0] ) != TIXML_SUCCESS )
		UBITRACK_THROW( "No valid 'latitude' attribute in GPS component configuration. Specify in degrees, e.g. 41° 26.57' as \"41.4428333\"." );
	if ( ref->ToElement()->QueryDoubleAttribute( "longitude", &m_offset[1] ) != TIXML_SUCCESS )
		UBITRACK_THROW( "No valid 'longitude' attribute in GPS component configuration. Specify in degrees, e.g. 41° 26.57' as \"41.4428333\"." );
	if ( ref->ToElement()->QueryDoubleAttribute( "altitude",  &m_offset[2] ) != TIXML_SUCCESS )
		UBITRACK_THROW( "No valid 'altitude' attribute in GPS component configuration. Specify in meters above sea level." );

	const TiXmlNode* port = subgraph->m_DataflowConfiguration.getXML()->FirstChild("Port");
	if ((port != NULL) && (port->ToElement() != NULL)) {
		char* pName = (char*)(port->ToElement()->Attribute("name"));
		if (pName != NULL) portName = pName;
		char* pRate = (char*)(port->ToElement()->Attribute("rate"));
		if (pRate != NULL) baudRate = atoi(pRate);
	}

	LOG4CPP_INFO( logger, "accessing " << portName << " @ " << baudRate << " BPS" );

	m_port = new Util::SerialPort( portName, baudRate );
	m_port->open(0,1); // blocking mode

	m_thread.reset( new boost::thread( boost::bind( &GPSDriver::threadHandler, this ) ) );
}


/** destructor */
GPSDriver::~GPSDriver()
{
	m_keepRunning = 0;
	m_thread->join();

	m_port->close();
	delete m_port;
}


void GPSDriver::threadHandler()
{
	// length constants for conversion from degrees to meters
	const double m1 = 111132.92   ;
	const double m2 =   -559.82   ;
	const double m3 =      1.175  ;
	const double m4 =     -0.0023 ;
	const double p1 = 111412.84   ;
	const double p2 =    -93.5    ;
	const double p3 =      0.118  ;

	// buffer for reading serial data
	unsigned char buffer[1024];
	// workspace and tmp string to extract lines from stream (maybe use a deque instead?)
	std::string workspace,tmp;
	unsigned int count;

	// parsing variables
	std::string prefix;
	double time,lat,lon,alt,hdop;
	unsigned char ns,ew,unit;
	int quality,nos;

	// calculation space
	double min,rad,latlen,lonlen,h,m,s;
	double x,y,z;

	Measurement::TimestampSync syncer( 1.0 ); // timestamps are in seconds
		
	LOG4CPP_DEBUG( logger, "GPS thread started" );

	while (m_keepRunning)
	{
		// get data from the serial port and push it on the workspace
		count = m_port->read( buffer, sizeof(buffer) );
		workspace.append( (const char*)buffer, count );

		// seek the first line in the workspace
		while ( (count = workspace.find( '\n', 0 )) != std::string::npos ) {

			// move it to a string with spaces as separator (instead of commas)
			tmp = workspace.substr( 0, count+1 );
			workspace.erase( 0, count+1 );
			replace( tmp.begin(), tmp.end(), ',', ' ' );

			// parse NMEA data
			std::istringstream nmea( tmp );
			nmea >> prefix >> time >> lat >> ns >> lon >> ew >> quality >> nos >> hdop >> alt >> unit;

			// some sanity checks
			if (!nmea) continue;
			if (prefix.compare( "$GPGGA" )) continue;
			if (quality == 0) continue;
			if (unit != 'M') continue;

			// convert HHMMSS.SS time format to absolute seconds
			h = floor(time/10000.0); time -= h*10000.0;
			m = floor(time/100.0);   time -= m*100.0;
			s = time + 60.0*m + 3600.0*h;

			// convert DDMM.MMMMM format to absolute degrees
			lat /= 100.0; min = lat - floor(lat); lat = lat - min; min *= (100.0/60.0); lat += min;
			lon /= 100.0; min = lon - floor(lon); lon = lon - min; min *= (100.0/60.0); lon += min;

			// calculate length of a degree in meters
			rad = (lat/180.0) * g_pi;
			latlen = m1 + (m2 * cos(2 * rad)) + (m3 * cos(4 * rad)) + (m4 * cos(6 * rad));
			lonlen =      (p1 * cos(1 * rad)) + (p2 * cos(3 * rad)) + (p3 * cos(5 * rad));

			LOG4CPP_DEBUG( logger, "number of satelites: " << nos << " position error: " << hdop );
			LOG4CPP_DEBUG( logger, std::setprecision(10)<< "latitude=\"" << lat << "\" longitude=\"" << lon << "\" altitude=\"" << alt << "\"" );

			// subtract offset position and calculate final position
			x = lat - m_offset[0]; x *= latlen;
			y = lon - m_offset[1]; y *= lonlen;
			z = alt - m_offset[2];

			// at long last, send it (swap latitude/longitude to conform to usual map orientation)
			Measurement::Position pos( syncer.convertNativeToLocal( s ), Math::Vector<3>( y, x, z ) );
			m_output.send( pos );
		}
	}
}


UBITRACK_REGISTER_COMPONENT( ComponentFactory* const cf ) {
	cf->registerComponent< GPSDriver > ( "GPSDriver" );
}

} } // namespace Ubitrack::Drivers

