/*
 * Ubitrack - Library for Ubiquitous Tracking
 * Copyright 2006, Technische Universitaet Muenchen, and individual
 * contributors as indicated by the @authors tag. See the 
 * copyright.txt in the distribution for a full listing of individual
 * contributors.
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA, or see the FSF site: http://www.fsf.org.
 */


/**
 * @ingroup driver_components
 * @author Florian Echtler <echtler@in.tum.de>
 */

#ifndef WIN32
 
#include "AudioStream.h"

#include <linux/soundcard.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include <sys/ioctl.h>
#include <stdexcept>
#include <errno.h>


#define SAMPLERATE 44100
#define CHANNELS       2


/** constructor: add an audio output stream */
AudioStream::AudioStream( AVFormatContext* fcon, int codec_id )
{
	AVCodecContext *c;

	oc = fcon;

	st = av_new_stream( oc, 1 );
	if (!st) throw std::runtime_error( "Could not allocate audio stream." );

	dev_handle = ::open( "/dev/dsp", O_RDONLY, 0 );

	try_ioctl( SNDCTL_DSP_SETFMT,   AFMT_S16_NE );
	try_ioctl( SNDCTL_DSP_CHANNELS,    CHANNELS );
	try_ioctl( SNDCTL_DSP_SPEED,     SAMPLERATE );

	c = st->codec;
	c->codec_id = (CodecID)codec_id;
	c->codec_type = CODEC_TYPE_AUDIO;

	/* put sample parameters */
	c->bit_rate = 64000;
	c->sample_rate = SAMPLERATE;
	c->channels = CHANNELS;
}


void AudioStream::open() {

	AVCodecContext* c = st->codec;

	/* find the audio encoder */
	AVCodec* codec = avcodec_find_encoder( c->codec_id );
	if (!codec) throw std::runtime_error( "Audio codec not found." );

	/* open it */
	if (avcodec_open( c, codec ) < 0) throw std::runtime_error( "Could not open audio codec." );

	audio_outbuf_size = 10000;
	audio_outbuf = (uint8_t*)av_malloc( audio_outbuf_size );

	/* ugly hack for PCM codecs (will be removed ASAP with new PCM
		 support to compute the input frame size in samples */
	if (c->frame_size <= 1) {
		audio_input_frame_size = audio_outbuf_size / c->channels;
		switch(st->codec->codec_id) {
			case CODEC_ID_PCM_S16LE:
			case CODEC_ID_PCM_S16BE:
			case CODEC_ID_PCM_U16LE:
			case CODEC_ID_PCM_U16BE:
				audio_input_frame_size >>= 1;
				break;
			default:
				break;
		}
	} else {
		audio_input_frame_size = c->frame_size;
	}

	samples = (int16_t*)av_malloc(audio_input_frame_size * 2 * c->channels);
}


AudioStream::~AudioStream()
{
	avcodec_close(st->codec);

	close( dev_handle );

	av_free(samples);
	av_free(audio_outbuf);
}


void AudioStream::try_ioctl( int cmd, int val ) {
	int tmp = val;
	if (ioctl( dev_handle, cmd, &tmp ) == -1)
		throw std::runtime_error( std::string("ioctl: ") + strerror(errno) );
}


/* prepare a 16 bit dummy audio frame of 'frame_size' samples and 'nb_channels' channels */
void AudioStream::get_audio_frame(int16_t *samples, int frame_size, int nb_channels)
{
	read( dev_handle, samples, frame_size*nb_channels*2 );
}


void AudioStream::write_audio_frame( )
{
	AVCodecContext *c;
	AVPacket pkt;

	av_init_packet( &pkt );

	c = st->codec;

	get_audio_frame(samples, audio_input_frame_size, c->channels);

	pkt.size = avcodec_encode_audio(c, audio_outbuf, audio_outbuf_size, samples);

	pkt.pts = av_rescale_q(c->coded_frame->pts, c->time_base, st->time_base);
	pkt.flags |= PKT_FLAG_KEY;
	pkt.stream_index = st->index;
	pkt.data = audio_outbuf;

	/* write the compressed frame in the media file */
	if (av_write_frame( oc, &pkt ) != 0) throw std::runtime_error( "Error while writing audio frame." );
}

double AudioStream::get_pts()
{
	return (double)st->pts.val * st->time_base.num / st->time_base.den;
}

#endif // !WIN32

