/*
 * Ubitrack - Library for Ubiquitous Tracking
 * Copyright 2006, Technische Universitaet Muenchen, and individual
 * contributors as indicated by the @authors tag. See the 
 * copyright.txt in the distribution for a full listing of individual
 * contributors.
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA, or see the FSF site: http://www.fsf.org.
 */


/**
 * @ingroup driver_components
 * @file
 * Image Sink
 * This file contains the driver component to
 * write ImageMeasurements to a video file.
 *
 * @author Florian Echtler <echtler@in.tum.de>
 *
 * @par Example Configuration
\verbatim
<Pattern name="ImageSink" id="SinkID">
	<Input>
		<Node name="A"/>
		<Node name="B"/>
		<Edge name="Input" source="A" destination="B" pattern-ref="..." edge-ref="..."/>
	</Input>
	<DataflowConfiguration>
		<UbitrackLib class="ImageSink"/>
		<Attribute name="filename" value="foobar.avi"/>
		<Attribute name="width" value="640"/>
		<Attribute name="height" value="480"/>
		<Attribute name="bitrate" value="600000"/>
		<Attribute name="FPS" value="15"/>
	</DataflowConfiguration>
</Pattern>
\endverbatim
 */

#include <map>
#include <string>
#include <log4cpp/Category.hh>
#include <boost/bind.hpp>

#include <Ubitrack/Dataflow/PushConsumer.h>
#include <Ubitrack/Dataflow/ComponentFactory.h>
#include <Ubitrack/Dataflow/Component.h>
#include <Ubitrack/Measurement/Measurement.h>
#include <Ubitrack/Util/Exception.h>
#include <Ubitrack/Util/OS.h>
#include <Vision/Image.h>

#include "Stream.h"


static log4cpp::Category& logger( log4cpp::Category::getInstance( "Drivers.ImageSink" ) );


namespace Ubitrack { namespace Drivers {

class ImageSink
	: public Dataflow::Component
{

public:

	/** constructor */
	ImageSink( const std::string& sName, boost::shared_ptr< Graph::UTQLSubgraph > pConfig )
		: Component( sName )
		, m_input( "Input", *this, boost::bind( &ImageSink::eventIn, this, _1 ) )
	{
		LOG4CPP_INFO( logger, "Creating ImageSink" );
		
		m_imgcount = 0;
		m_run = 1;
		
		std::string filename( "ubitrack_imagesink_output.avi" );
		int fps     = 10;
		int width   = 640;
		int height  = 480;
		int bitrate = 400000;

		pConfig->m_DataflowAttributes.getAttributeData( "FPS", fps );
		pConfig->m_DataflowAttributes.getAttributeData( "width", width );
		pConfig->m_DataflowAttributes.getAttributeData( "height", height );
		pConfig->m_DataflowAttributes.getAttributeData( "bitrate", bitrate );
		pConfig->m_DataflowAttributes.getAttributeData( "filename", filename );

		m_stream = new StreamOutput( filename, width, height, fps, bitrate );
		LOG4CPP_INFO( logger, "created stream: " << width << "x" << height << "@" << fps << 
		                      "FPS, writing to " << filename << " @ " << bitrate << "bps" );

		m_thread.reset( new boost::thread( boost::bind( &ImageSink::threadHandler, this ) ) );
	}

	/** destructor */
	~ImageSink()
	{
		LOG4CPP_INFO( logger, "time to quit: waiting for encoding thread.." );
		m_run = 0;
		m_thread->join();

		// encode the last image again to flush the queue
		m_stream->process( 0 );
		delete m_stream;

		LOG4CPP_INFO( logger, "encoded " << m_imgcount << " images" );
	}

protected:

	void threadHandler() {
		LOG4CPP_DEBUG( logger, "Thread started" );
		while (m_run) {
			LOG4CPP_TRACE( logger, "waiting for image" );
			// TODO: change to timed_wait
			while (!m_image && m_run) Util::sleep(1); if (!m_run) break;
			boost::mutex::scoped_lock l( m_mutex );
			LOG4CPP_DEBUG( logger, "encoding image #" << m_imgcount );
			m_stream->process( m_image.get() );
			m_image.reset();
			m_imgcount++;
		}
		LOG4CPP_DEBUG( logger, "Thread stopped" );
	}

	void eventIn( const Vision::ImageMeasurement& image ) {
		boost::mutex::scoped_lock l( m_mutex );
		m_image = image;
		LOG4CPP_DEBUG( logger, "received image #" << m_imgcount );
	}

	Vision::ImageMeasurement m_image;
	Dataflow::PushConsumer< Vision::ImageMeasurement > m_input;

	boost::shared_ptr< boost::thread > m_thread;
	StreamOutput* m_stream;
	int m_imgcount;
	int m_run;
	boost::mutex m_mutex;

};

} } // namespace Ubitrack::Drivers


UBITRACK_REGISTER_COMPONENT( Ubitrack::Dataflow::ComponentFactory* const cf )
{
	cf->registerComponent< Ubitrack::Drivers::ImageSink > ( "ImageSink" );
}

