/*
 * Ubitrack - Library for Ubiquitous Tracking
 * Copyright 2006, Technische Universitaet Muenchen, and individual
 * contributors as indicated by the @authors tag. See the 
 * copyright.txt in the distribution for a full listing of individual
 * contributors.
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA, or see the FSF site: http://www.fsf.org.
 */


/**
 * @ingroup driver_components
 * @author Florian Echtler <echtler@in.tum.de>
 */

#include "Stream.h"

#include <stdexcept>
#include <iostream>
#include <string>

StreamOutput::StreamOutput( std::string filename, int width, int height, int fps, int bitrate ) {

	audio_st = NULL;
	video_st = NULL;
	data_st  = NULL;

	audio_pts = 0.0;
	video_pts = 0.0;
	data_pts  = 0.0;

	// initialize libavcodec, and register all codecs and formats 
	av_register_all();

	// auto detect the output format from the name. default is mpeg.
	fmt = guess_format( NULL, filename.c_str(), NULL );
	if (!fmt) fmt = guess_format( "mpeg", NULL, NULL );
	if (!fmt) throw std::runtime_error( "Could not find suitable output format.");

	// no audio for the moment
	fmt->audio_codec = CODEC_ID_NONE;

	// allocate the output media context
	oc = av_alloc_format_context();
	if (!oc) throw std::runtime_error( "Could not allocate output context." );
	oc->oformat = fmt; strncpy( oc->filename, filename.c_str(), sizeof(oc->filename) );

	// add the audio and video streams using the default format codecs and initialize the codecs
	if (fmt->video_codec != CODEC_ID_NONE) video_st = new VideoStream( oc, fmt->video_codec, width, height, fps, bitrate );
	if (fmt->audio_codec != CODEC_ID_NONE) audio_st = new AudioStream( oc, fmt->audio_codec );

	//if (fmt->subtitle_codec != CODEC_ID_NONE)
	data_st = new DataStream( oc, CODEC_ID_RAWVIDEO );

	// set the output parameters (must be done even if no parameters).
	if (av_set_parameters(oc, NULL) < 0) throw std::runtime_error( "Invalid output format parameters." );
	dump_format(oc, 0, filename.c_str(), 1);

	// now that all the parameters are set, we can open the audio
	// and video codecs and allocate the necessary encode buffers 
	if (video_st) video_st->open();
	if (audio_st) audio_st->open();
	if (data_st)  data_st->open();

	// open the output file, if needed
	if (!(fmt->flags & AVFMT_NOFILE))
		if (url_fopen(&oc->pb, filename.c_str(), URL_WRONLY) < 0)
			throw std::runtime_error( std::string("Could not open ") + filename );

	// write the stream header, if any 
	av_write_header(oc);
}


void StreamOutput::process( IplImage* data, unsigned char* text ) {

	// compute current audio and video time
	if (audio_st) audio_pts = audio_st->get_pts();
	if (video_st) video_pts = video_st->get_pts();

	// write interleaved audio and video frames (if available)
	if (!video_st || (video_st && audio_st && audio_pts < video_pts))
		audio_st->write_audio_frame();
	else
		video_st->write_video_frame( data );

	if (text && data_st) data_st->write_data_frame( text );
}


StreamOutput::~StreamOutput() {

	// close each codec
	if (video_st) delete video_st;
	if (audio_st) delete audio_st;
	if (data_st)  delete data_st;

	// write the trailer, if any
	av_write_trailer(oc);

	// free the streams 
	for (unsigned int i = 0; i < oc->nb_streams; i++) {
		av_freep(&oc->streams[i]->codec);
		av_freep(&oc->streams[i]);
	}

	// close the output file
	if (!(fmt->flags & AVFMT_NOFILE))
		url_fclose(oc->pb);

	// free the output context
	av_free(oc);
}

