/*
 * Ubitrack - Library for Ubiquitous Tracking
 * Copyright 2006, Technische Universitaet Muenchen, and individual
 * contributors as indicated by the @authors tag. See the 
 * copyright.txt in the distribution for a full listing of individual
 * contributors.
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA, or see the FSF site: http://www.fsf.org.
 */


/**
 * @ingroup driver_components
 * @author Florian Echtler <echtler@in.tum.de>
 */

#include "VideoStream.h"

#include <stdexcept>
#include <iostream>


AVFrame* alloc_picture( int pix_fmt, int width, int height )
{
	AVFrame *picture;
	uint8_t *picture_buf;
	int size;

	picture = avcodec_alloc_frame();
	if (!picture) return NULL;

	size = avpicture_get_size( pix_fmt, width, height );
	picture_buf = (uint8_t*)av_malloc( size );
	if (!picture_buf) {
		av_free(picture);
		return NULL;
	}

	avpicture_fill( (AVPicture*)picture, picture_buf, pix_fmt, width, height );
	return picture;
}


/* add a video output stream */
VideoStream::VideoStream( AVFormatContext* fcon, int codec_id, int width, int height, int framerate, int bitrate )
{
	AVCodecContext* c;

	oc = fcon;

	m_width = width;
	m_height = height;

	m_oldwidth = m_oldheight = 0;

	st = av_new_stream( oc, 0 );
	if (!st) throw std::runtime_error( "Could not allocate video stream." );

	c = st->codec;
	c->codec_id = (CodecID)codec_id;
	c->codec_type = CODEC_TYPE_VIDEO;

	// put sample parameters 
	c->bit_rate = bitrate;
	c->pix_fmt = PIX_FMT_YUV420P;

	// resolution must be a multiple of two
	c->width = m_width;
	c->height = m_height;

	// time base: this is the fundamental unit of time (in seconds) in terms
	// of which frame timestamps are represented. for fixed-fps content,
	// timebase should be 1/framerate and timestamp increments should be
	// identically 1.
	c->time_base.den = framerate;
	c->time_base.num = 1;

	// TODO: it's unclear whether the following parameters are needed at all

	// emit one intra frame every twelve frames at most
	c->gop_size = 12;

	// just for testing, we also add B frames
	if (c->codec_id == CODEC_ID_MPEG2VIDEO) c->max_b_frames = 2;

	// Needed to avoid using macroblocks in which some coeffs overflow.
	// This does not happen with normal video, it just happens here as
	// the motion of the chroma plane does not match the luma plane. 
	if (c->codec_id == CODEC_ID_MPEG1VIDEO) c->mb_decision = 2;

	// some formats want stream headers to be separate
	if(!strcmp(oc->oformat->name, "mp4") || !strcmp(oc->oformat->name, "mov") || !strcmp(oc->oformat->name, "3gp"))
		c->flags |= CODEC_FLAG_GLOBAL_HEADER;
}

void VideoStream::open() {

	AVCodecContext* c = st->codec;

	// find the video encoder 
	AVCodec* codec = avcodec_find_encoder( c->codec_id );
	if (!codec) throw std::runtime_error( "Video codec not found." );

	// open the codec
	if (avcodec_open( c, codec ) < 0) throw std::runtime_error( "Could not open video codec." );

	// allocate output buffer for encoded data
	video_outbuf_size = 200000;
	video_outbuf = (uint8_t*)av_malloc( video_outbuf_size );

	// allocate the encoded raw picture 
	picture = alloc_picture( c->pix_fmt, c->width, c->height );
	if (!picture) throw std::runtime_error( "Could not allocate video picture." );
}


VideoStream::~VideoStream( )
{
	avcodec_close( st->codec );
	av_free( picture->data[0] );
	av_free( picture );
	av_free( video_outbuf );
}



void VideoStream::write_video_frame( IplImage* img )
{
	int out_size = 0;
	int ret = 0;
	AVCodecContext *c;

	c = st->codec;

	if ((img) && (img->nChannels == 3)) {

		// create a new conversion context, if needed
		if ((img->width != m_oldwidth) || (img->height != m_oldheight)) {
			m_convert_ctx = sws_getContext( img->width, img->height, PIX_FMT_RGB24, m_width, m_height, PIX_FMT_YUV420P, SWS_FAST_BILINEAR, 0, 0, 0 );
			if (m_convert_ctx == NULL) throw std::runtime_error( "Cannot initialize the conversion context." );
		}

		int stride = 3 * img->width;
		uint8_t* data = (uint8_t*)img->imageData;

		// in order to flip the image, source pointer and stride have to be modified.
		// disclaimer: the ffmpeg scaling code is a maze of terror and madness.
		if (img->origin) {
			data = data + 3 * img->width * (img->height - 1);
			stride = -stride;
		}

		uint8_t* tmp[4] = { data, data, data, data };
		int strides[4] = { stride, stride, stride, stride };

		sws_scale( m_convert_ctx, tmp, strides, 0, c->height, picture->data, picture->linesize );

		m_oldwidth  = img->width;
		m_oldheight = img->height;
	}

	// encode the image
	out_size = avcodec_encode_video( c, video_outbuf, video_outbuf_size, picture );

	// if zero size, it means the image was buffered
	if (out_size > 0) {

		AVPacket pkt;
		av_init_packet(&pkt);

		pkt.pts = av_rescale_q( c->coded_frame->pts, c->time_base, st->time_base );
		if (c->coded_frame->key_frame) pkt.flags |= PKT_FLAG_KEY;
		pkt.stream_index = st->index;
		pkt.data = video_outbuf;
		pkt.size = out_size;

		// write the compressed frame in the media file
		ret = av_write_frame(oc, &pkt);
	}

	if (ret != 0) throw std::runtime_error( "Error while writing video frame." );
}

double VideoStream::get_pts()
{
	// TODO: pts.val is deprecated. find out what's needed instead..
	return (double)st->pts.val * st->time_base.num / st->time_base.den;
}

