/*
 * Ubitrack - Library for Ubiquitous Tracking
 * Copyright 2006, Technische Universitaet Muenchen, and individual
 * contributors as indicated by the @authors tag. See the 
 * copyright.txt in the distribution for a full listing of individual
 * contributors.
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA, or see the FSF site: http://www.fsf.org.
 */

/**
 * @ingroup driver_components
 * @file
 * Implementations of OpenTracker sources and sinks
 *
 * @author Daniel Pustka <daniel.pustka@in.tum.de>
 */ 
 
#include <algorithm>
#include <boost/thread.hpp>
#include <Ubitrack/Measurement/Measurement.h>
#include <Ubitrack/Dataflow/Module.h>
#include <Ubitrack/Dataflow/PushSupplier.h>
#include <OpenTracker/OpenTracker.h>
#include <OpenTracker/core/Configurator.h>
#include <OpenTracker/common/CallbackModule.h>
#include <OpenTracker/common/CallforwardModule.h>

static log4cpp::Category& logger( log4cpp::Category::getInstance( "Drivers.OpenTracker" ) );

namespace Ubitrack { namespace Drivers {

// forward decls
class OpenTrackerComponentBase;


/** 
 * Component key for OpenTracker components.
 */
class OpenTrackerComponentKey
	: public std::string
{
public:
	/** extract the "attributes" parameter of the edge config. */
	OpenTrackerComponentKey( const Dataflow::ComponentConfiguration& config )
	{
		const TiXmlNode* xmlConf( config.m_edgeConfiguration.getXML() );
		if ( !xmlConf )
			UBITRACK_THROW( "No configuration for OpenTracker component" );

		const char* sAttr = xmlConf->ToElement()->Attribute( "node" );			
		if ( !sAttr )
			UBITRACK_THROW( "Configuration for OpenTracker component has no 'node' attribute (must be a callback or callforward node)" );
		*static_cast< std::string* >( this ) = sAttr;
	}
};


/** 
 * Module used by OpenTracker components.
 * Owns the OT mainloop thread and puts together configurations.
 */
class OpenTrackerModule
	: public Dataflow::Module< Dataflow::SingleModuleKey, OpenTrackerComponentKey, OpenTrackerModule, OpenTrackerComponentBase >
{
public:
	/** simplifies our life afterwards */
	typedef Dataflow::Module< Dataflow::SingleModuleKey, OpenTrackerComponentKey, OpenTrackerModule, OpenTrackerComponentBase > BaseClass;
	
	OpenTrackerModule( const Dataflow::SingleModuleKey& key, const Dataflow::ComponentConfiguration&, FactoryHelper* fh )
		: BaseClass( key, fh )
		, m_pConfigurator( ot::Configurator::instance() )
	{
		LOG4CPP_INFO( logger, "created OpenTrackerModule" );
	}
	
	~OpenTrackerModule()
	{		
		// stop & wait for thread
		if ( m_pConfigurator && m_pConfigurator->getContext() )
			m_pConfigurator->getContext()->stopLoop();
		if ( m_pMainLoopThread )
			m_pMainLoopThread->join();
		
		LOG4CPP_INFO( logger, "destroyed OpenTrackerModule" );
	}

	/** add a component to the internal list */
	void addComponent( OpenTrackerComponentBase* pComponent )
	{ m_otComponents.push_back( pComponent ); }

	/** remove a component from the internal list */
	void removeComponent( OpenTrackerComponentBase* pComponent )
	{ m_otComponents.remove( pComponent ); }

	/** (re)configures OpenTracker */
	void reconfigureOT();

protected:
	
	/** method that runs the main loop */
	void mainloop()
	{ 
		LOG4CPP_INFO( logger, "Starting OpenTracker mainloop" );
		
		// TODO: make type of main loop configurable
		//m_pConfigurator->getContext()->runOnDemand();
		m_pConfigurator->getContext()->runAtRate( 100 );
		
		LOG4CPP_INFO( logger, "OpenTracker mainloop stopped" );
	}

	/** the main loop thread */
	boost::shared_ptr< boost::thread > m_pMainLoopThread;
	
	/** create new components */
	boost::shared_ptr< OpenTrackerComponentBase > createComponent( const std::string& type, const std::string& name, const Dataflow::ComponentConfiguration& config, 
		const OpenTrackerComponentKey& key, OpenTrackerModule* pModule );	
		
	/** the OpenTracker context we'll use */
	ot::Configurator* m_pConfigurator;
	
	/** another list of components -- the module already has some map of components, but not updated when the component constructor is called */
	std::list< OpenTrackerComponentBase* > m_otComponents;
};


/** 
 * Base class for all OpenTracker components.
 */
class OpenTrackerComponentBase
	: public OpenTrackerModule::Component
{
public:
	OpenTrackerComponentBase( const std::string& name, const Dataflow::ComponentConfiguration& config, 
		const OpenTrackerComponentKey& key, OpenTrackerModule* module );

	~OpenTrackerComponentBase();

	/** called when the component should (re)register the callback/forward with the OT context */
	virtual void registerWithContext( ot::Context& context )
	{}
	
	/** get the module configuration */
	const std::string& getModuleConfiguration() const
	{ return m_sModuleConfiguration; }
	
	/** get the node configuration */
	const std::string& getNodeConfiguration() const
	{ return m_sNodeConfiguration; }
	
protected:
	/** module configuration */
	std::string m_sModuleConfiguration;

	/** node configuration */
	std::string m_sNodeConfiguration;
};


OpenTrackerComponentBase::OpenTrackerComponentBase( const std::string& name, const Dataflow::ComponentConfiguration& config, const OpenTrackerComponentKey& key, OpenTrackerModule* module )
	: OpenTrackerModule::Component( name, key, module )
{
	// read node and module config from xml
	const TiXmlNode* xmlConf( config.m_edgeConfiguration.getXML() );
	if ( !xmlConf )
		UBITRACK_THROW( "No configuration for OpenTracker component" );

	const TiXmlNode* moduleConfig = xmlConf->ToElement()->FirstChild( "ModuleConfiguration" );
	if ( !moduleConfig || moduleConfig->NoChildren() )
		{ LOG4CPP_NOTICE( logger, "Component has no or empty ModuleConfiguration element" ); }
	else
		{
		// convert XML to string
		TiXmlPrinter printer;
		moduleConfig->FirstChild()->Accept( &printer );
		m_sModuleConfiguration = printer.Str();
		}

	const TiXmlNode* nodeConfig = xmlConf->ToElement()->FirstChild( "NodeConfiguration" );
	if ( !nodeConfig || nodeConfig->NoChildren() )
		{ LOG4CPP_NOTICE( logger, "Component has no or empty NodeConfiguration element" ); }
	else
		{
		// convert XML to string
		TiXmlPrinter printer;
		nodeConfig->FirstChild()->Accept( &printer );
		m_sNodeConfiguration = printer.Str();
		}
		
	module->addComponent( this );
}


OpenTrackerComponentBase::~OpenTrackerComponentBase()
{
	getModule().removeComponent( this );
	getModule().reconfigureOT();
}


/**
 * @ingroup driver_components
 * Component to receive events from an OpenTracker dataflow.
 *
 * @par Input Ports
 * none
 *
 * @par Output Ports
 * PushSupplier< Measurement::Pose > port with name "Output".
 *
 * @par Configuration
 *
 * - Edge configuration:
 * @verbatim
 * <Configuration node="<nodename>">
 *   <ModuleConfiguration>
 *     [some OT config]
 *   </ModuleConfiguration>
 *   <NodeConfiguration>
 *     <Callback name="<nodename>">
 *       [some OT nodes]
 *     </Callback>
 *   </NodeConfiguration>
 * </Configuration>
 * @endverbatim
 *   - nodename: name of the callback node in the OT NodeConfiguration. Must be unique in the whole ubitrack dataflow!
 *   - the content of the ModuleConfiguration block is the same as the content of the OT <configuration> node
 */
class OpenTrackerSource
	: public OpenTrackerComponentBase
{
public:
	/** subscribes to notification */
	OpenTrackerSource( const std::string& name, const Dataflow::ComponentConfiguration& config, const OpenTrackerComponentKey& key, OpenTrackerModule* module )
		: OpenTrackerComponentBase( name, config, key, module )
		, m_outPort( "Output", *this )
	{		
		LOG4CPP_INFO( logger, "Created OpenTrackerSource for Callback node " << key );
		module->reconfigureOT();
	}
	
	/** unsubscribes from notification */
	~OpenTrackerSource()
	{
		LOG4CPP_INFO( logger, "Removing OpenTrackerSource for Callback node " << getKey() );
	}

	/** called when the component should (re)register the callback with the OT context */
	void registerWithContext( ot::Context& context );
	
protected:
	/** output port */
	Dataflow::PushSupplier< Measurement::Pose > m_outPort;
	
	/** the callback */
	void onCallback( ot::Event& e );
	
	/** a static callback function for opentracker */
	static void staticCallback( ot::CallbackNode *, ot::Event& evnt, void* pThis )
	{ static_cast< OpenTrackerSource* >( pThis )->onCallback( evnt ); }
};


void OpenTrackerSource::registerWithContext( ot::Context& context )
{
	// set our callback function on the OT callback module
	ot::CallbackModule* pCallbackMod = dynamic_cast< ot::CallbackModule *>( context.getModule( "CallbackConfig" ) );
	if ( !pCallbackMod )
		UBITRACK_THROW( "no module CallbackConfig" );
	
	pCallbackMod->setCallback( getKey(), &OpenTrackerSource::staticCallback, static_cast< void* >( this ) );
}


void OpenTrackerSource::onCallback( ot::Event& e )
{
	LOG4CPP_DEBUG( logger, "Received event " );// << e.getPrintOut() );

	const std::vector< float >& pos( e.getPosition() );
	const std::vector< float >& ori( e.getOrientation() );

	// TODO: timestamp: e.getValueString( "timestamp" )
	m_outPort.send( Measurement::Pose( Measurement::now(), 
		Math::Pose( Math::Quaternion( ori[ 1 ], ori[ 2 ], ori[ 3 ], ori[ 0 ] ), Math::Vector< 3 >( pos[ 0 ], pos[ 1 ], pos[ 2 ] ) ) ) );
}


/**
 * @ingroup driver_components
 * Component to send events to an OpenTracker dataflow.
 *
 * @par Input Ports
 * PushConsumer< Measurement::Pose > port with name "Output".
 *
 * @par Output Ports
 * none
 *
 * @par Configuration
 *
 * @verbatim
 * <Configuration node="<nodename>">
 *   <ModuleConfiguration>
 *     [some OT config]
 *   </ModuleConfiguration>
 *   <NodeConfiguration>
 *     <Some OT node>
 *       <Callforward name="<nodename>"/>
 *     </Some OT node>
 *   </NodeConfiguration>
 * </Configuration>
 * @endverbatim
 *   - nodename: name of the callforward node in the OT NodeConfiguration. Must be unique in the whole ubitrack dataflow!
 *   - the content of the ModuleConfiguration block is the same as the content of the OT <configuration> node
 */
class OpenTrackerSink
	: public OpenTrackerComponentBase
{
public:
	/** constructor */
	OpenTrackerSink( const std::string& name, const Dataflow::ComponentConfiguration& config, const OpenTrackerComponentKey& key, OpenTrackerModule* module )
		: OpenTrackerComponentBase( name, config, key, module )
		, m_inPort( "Input", *this, boost::bind( &OpenTrackerSink::send, this, _1 ) )
		, m_pOtModule( 0 )
	{
		LOG4CPP_INFO( logger, "Created OpenTrackerSink for node=" << key );
		module->reconfigureOT();
	}

	/** called when the component should (re)register the callback with the OT context */
	void registerWithContext( ot::Context& context );

protected:
	/** input port */
	Dataflow::PushConsumer< Measurement::Pose > m_inPort;
	
	/** pointer to the ot context */
	ot::CallforwardModule* m_pOtModule;
	
	/** called by the dataflow to push stuff */
	void send( const Measurement::Pose& p );
};
 
 
void OpenTrackerSink::registerWithContext( ot::Context& context )
{
	// set our callback function on the OT callback module
	ot::CallforwardModule* pCallforwardMod = dynamic_cast< ot::CallforwardModule *>( context.getModule( "CallforwardConfig" ) );
	if ( !pCallforwardMod )
		UBITRACK_THROW( "no module CallforwardConfig" );
	
	m_pOtModule = pCallforwardMod;
}


void OpenTrackerSink::send( const Measurement::Pose& p )
{
	LOG4CPP_DEBUG( logger, "Sink: Sending event to OT" );

	// create an event
	std::vector< float > pos( 3 );
	pos[ 0 ] = float( p->translation()( 0 ) );
	pos[ 1 ] = float( p->translation()( 1 ) );
	pos[ 2 ] = float( p->translation()( 2 ) );

	std::vector< float > ori( 4 );
	ori[ 0 ] = float( p->rotation().w() );
	ori[ 1 ] = float( p->rotation().x() );
	ori[ 2 ] = float( p->rotation().y() );
	ori[ 3 ] = float( p->rotation().z() );

	ot::Event e;
	e.setPosition( pos );
	e.setOrientation( ori );
	
	m_pOtModule->callForward( getKey(), e );
}


void OpenTrackerModule::reconfigureOT()
{
	LOG4CPP_INFO( logger, "reconfiguring OT" );
	
	// stop the event thread
	if ( m_pMainLoopThread && m_pConfigurator->getContext() )
	{
		LOG4CPP_DEBUG( logger, "stopping the OT thread" );
		m_pConfigurator->getContext()->stopLoop();
		LOG4CPP_DEBUG( logger, "waiting for the OT thread" );
		m_pMainLoopThread->join();
		m_pMainLoopThread.reset();
	}

	if ( !m_otComponents.empty() )
	{
		/** build the configuration */
		std::string sConfig;
		//sConfig = "<?xml version=\"1.0\" encoding=\"UTF-8\"?><!DOCTYPE OpenTracker SYSTEM \"opentracker.dtd\"><OpenTracker><configuration>";
		sConfig = "<?xml version=\"1.0\" encoding=\"UTF-8\"?><OpenTracker><configuration>";

		// combine node and module configuration from the different components		
		for ( std::list< OpenTrackerComponentBase* >::iterator it = m_otComponents.begin(); it != m_otComponents.end(); it++ )
			sConfig += (*it)->getModuleConfiguration();
		sConfig += "</configuration>";
		for ( std::list< OpenTrackerComponentBase* >::iterator it = m_otComponents.begin(); it != m_otComponents.end(); it++ )
			sConfig += (*it)->getNodeConfiguration();

		sConfig += "</OpenTracker>";
		LOG4CPP_DEBUG( logger, "OpenTracker configuration string: " << sConfig );

		m_pConfigurator->changeConfigurationString( sConfig.c_str() );
		if ( !m_pConfigurator->getContext() )
			{ LOG4CPP_WARN( logger, "Configurator has no context" ); }
		else
		{
			// re-register all callbacks
			for ( std::list< OpenTrackerComponentBase* >::iterator it = m_otComponents.begin(); it != m_otComponents.end(); it++ )
				(*it)->registerWithContext( *m_pConfigurator->getContext() );

			// restart thread
			m_pMainLoopThread.reset( new boost::thread( boost::bind( &OpenTrackerModule::mainloop, this ) ) );
		}
	}
}


// has to be here, after all class declarations
boost::shared_ptr< OpenTrackerComponentBase > OpenTrackerModule::createComponent( const std::string& type, const std::string& name, const Dataflow::ComponentConfiguration& config, 
	const OpenTrackerModule::ComponentKey& key, OpenTrackerModule* pModule )
{
	if ( type == "OpenTrackerSource" )
		return boost::shared_ptr< OpenTrackerComponentBase >( new OpenTrackerSource( name, config, key, pModule ) );
	else if ( type == "OpenTrackerSink" )
		return boost::shared_ptr< OpenTrackerComponentBase >( new OpenTrackerSink( name, config, key, pModule ) );
		
	UBITRACK_THROW( "Class " + type + " not supported by OpenTracker module" );
}


} } // namespace Ubitrack::Drivers


UBITRACK_REGISTER_COMPONENT( Ubitrack::Dataflow::ComponentFactory* const cf ) {
	// create list of supported types
	std::vector< std::string > components;
	components.push_back( "OpenTrackerSource" );
	components.push_back( "OpenTrackerSink" );

	cf->registerModule< Ubitrack::Drivers::OpenTrackerModule > ( components );
}
