/*
 * Ubitrack - Library for Ubiquitous Tracking
 * Copyright 2006, Technische Universitaet Muenchen, and individual
 * contributors as indicated by the @authors tag. See the
 * copyright.txt in the distribution for a full listing of individual
 * contributors.
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA, or see the FSF site: http://www.fsf.org.
 */


/**
 * @ingroup driver_components
 * @file
 * Motion Model for Pose Generator driver
 * This file contains the motion model component to
 * generate rotation and position information from a waypoint list
 *
 * @author Benjamin Becker <Benjamin.Becker@eads.net>
 */

#include <string>
#include <cstdlib>

//#include <Ubitrack/Math/Vector.h>
#include <Ubitrack/Math/Pose.h>
#include <Ubitrack/Measurement/Timestamp.h>
#include <boost/numeric/ublas/vector.hpp>
#include <boost/numeric/ublas/blas.hpp>
//#include <boost/random/normal_distribution.hpp>
#include <boost/random.hpp>
#include <vector>


/**
 * Module for XSens tracker, should work on both linux and windows.
 * Requires the XSens ActiveX plugin to be installed on Windows.
 */
namespace Ubitrack { namespace Drivers { namespace MotionModel {

struct MotionParameters
{
	double noiseLevel;
	int noiseSeed;
	double maxVelocity;
	double maxAcceleration;
	double maxRotAcceleration;
	double mass;
	bool lookahead;
	Math::Vector<3> centerOfInterest;
};

struct Motion
{
	Math::Vector<3> position;
	Math::Vector<3> velocity;
	Math::Vector<3> acceleration;
	Math::Quaternion orientation;
	Measurement::Timestamp time;
};

class MotionModelComponent
{
public:

	/** constructor */
	MotionModelComponent(std::vector< Math::Vector< 3> > pwaypoints,
						 MotionParameters motionparams
						 )
	{
		waypoints = pwaypoints;
		mparams = motionparams;

		lastWP=0;
		nextWP=1;
		nextnextWP = (nextWP+1)%waypoints.size();
		path = waypoints[nextWP]-waypoints[lastWP];
		nextpath = waypoints[nextnextWP]-waypoints[nextWP];
		Normalize(path);
		Normalize(nextpath);
		double test = boost::numeric::ublas::blas_1::dot(path,nextpath);
		slowtest = (boost::numeric::ublas::blas_1::dot(path,nextpath)+1.0)/2.0;

		motion.orientation = Math::Quaternion();
		motion.position = waypoints[0];
		motion.velocity[0]=motion.velocity[1]=motion.velocity[2]=0.0;
		init = false;
		gotonextwp = false;

		boost::mt19937 rng;
	    rng.seed(static_cast<unsigned>(mparams.noiseSeed)); // seed generator

		// select desired probability distribution
		boost::normal_distribution<double> norm_dist(0, mparams.noiseLevel);

		// bind random number generator to distribution, forming a function
		normal_sampler = new boost::variate_generator<boost::mt19937&, boost::normal_distribution<double> > (rng, norm_dist);
	}

	/** destructor */
	~MotionModelComponent()
	{
	//	LOG4CPP_DEBUG( logger, "Stopping pose data generation" );
	}


	void Update()
	{
		if(!init)
		{
			motion.time = Measurement::now()-1E7;
			init = true;
		}
		Measurement::Timestamp currentTime = Measurement::now();
		double deltatime = 1.0/100.0; // (double)(currentTime-motion.time)/1E9

		Math::Vector<3> vTarget = waypoints[nextWP] - motion.position;
		Math::Vector<3> vState  = motion.velocity;
		Math::Vector<3> vDelta  = vTarget - vState;
		Normalize(vDelta);
		Math::Vector<3> aTotal =  mparams.maxAcceleration*vDelta;
		

		double vStateLength  = boost::numeric::ublas::norm_2(vState);
		if(vStateLength > mparams.maxVelocity)
		{
			
			if(true && (boost::numeric::ublas::blas_1::dot(aTotal,vState) < 0)) //should slow down && close to the next waypoint?
			{
   				//motion.acceleration = aTotal*slowtest;
				double deltaLen = boost::numeric::ublas::blas_1::dot(aTotal,vState) / (vStateLength*vStateLength);
				Math::Vector<3> aSide = aTotal - (deltaLen*vState);
				Math::Vector<3> aForward = aTotal - aSide;
				motion.acceleration = aSide + (slowtest*aForward);
			}
			else
			{
				// split acceleration into forward, and other:
				double deltaLen = boost::numeric::ublas::blas_1::dot(aTotal,vState) / (vStateLength*vStateLength);
				//Math::Vector<3> aSide = aTotal - (deltaLen*vState);
				//Math::Vector<3> aForward = aTotal - aSide;
				motion.acceleration = aTotal - (deltaLen*vState);
			}

		}
		else
		{
			motion.acceleration = aTotal;
		}


/*
		double vStateLength  = boost::numeric::ublas::norm_2(vState);
		if(vStateLength > mparams.maxVelocity)
		{
            double deltaLen = boost::numeric::ublas::blas_1::dot(motion.acceleration,vState) / vStateLength;
			if(deltaLen > 0)
				motion.acceleration -= (deltaLen*vState);
		}
		double aTargetLength = boost::numeric::ublas::norm_2(motion.acceleration);
		if(aTargetLength > mparams.maxAcceleration)
		{
			motion.acceleration *= (mparams.maxAcceleration/aTargetLength);
		}
*/

		motion.acceleration+=Noise3D();
		motion.velocity+= motion.acceleration*(deltatime);
		motion.position+= motion.velocity*(deltatime);
		motion.time = currentTime;
		
//		double currentVelocitySquare = pow(currentVelocity[0],2.0)+pow(currentVelocity[1],2.0)+pow(currentVelocity[2],2.0);
//		currentAcceleration = waypoints[nextWP] - currentPosition;
//		Normalize(currentAcceleration);

		Math::Vector<3> lookat;
		if(mparams.lookahead)
		{
			lookat = motion.velocity;
		}
		else
		{
			lookat = mparams.centerOfInterest - motion.position;
			Normalize(lookat);
				
		}
		motion.orientation = Math::Quaternion( Math::Vector< 3> (-lookat[1],lookat[0],0.0),//axis cross product lookat x (0.0,0.0,-1.0)
							acos(boost::numeric::ublas::blas_1::dot(lookat,Math::Vector<3>(0.0,0.0,-1.0))));


		gotonextwp = false;
		if((boost::numeric::ublas::norm_2(vTarget) <  (mparams.maxAcceleration/4.0)) || (boost::numeric::ublas::blas_1::dot(vTarget,path) < 0))
			gotonextwp = true;
/*
		//currentAcceleration = currentAcceleration*Noise()*maxacceleration;

		double currentVelocitySquare = pow(currentVelocity[0],2.0)+pow(currentVelocity[1],2.0)+pow(currentVelocity[2],2.0);
		Math::Vector<3> noise = Noise3D();

		if(currentVelocitySquare > maxVelocitySquare)
		{
			if(boost::numeric::ublas::blas_1::dot(currentAcceleration,currentVelocity) < 0)
				currentAcceleration = mparams.maxAcceleration*(noise+currentAcceleration);
			else
				currentAcceleration = noise;
		}
		else
			currentAcceleration = mparams.maxAcceleration*(noise+currentAcceleration);

		currentVelocity*=0.99;
		currentVelocity+= currentAcceleration*(deltatime);
		currentPosition+= currentVelocity*(deltatime);
		motion.time = currentTime;

		std::cout<<"P: "<<currentPosition<<" | V: "<<currentVelocity<<" | A: "<<currentAcceleration<<std::endl;

*/
		std::cout<<"P: "<<motion.position<<" | V: "<<motion.velocity<<" | A: "<<motion.acceleration<<std::endl;

		if(gotonextwp)
		{
			lastWP = nextWP;
			nextWP = (nextWP+1)%waypoints.size();
			nextnextWP = (nextWP+1)%waypoints.size();
			path = waypoints[nextWP]-waypoints[lastWP];
			nextpath = waypoints[nextnextWP]-waypoints[nextWP];
			Normalize(path);
			Normalize(nextpath);
			
			slowtest = (boost::numeric::ublas::blas_1::dot(path,nextpath)+1.0)/2.0;
	//		if(boost::numeric::ublas::blas_1::dot(path,nextpath) < 0.70710678118654752440084436210485)
	//			slow = true;
	//		else
	//			slow = false;
			//std::cout<<"Path: "<<path<<std::endl;
		}
	}

	Math::Pose GetPose()
	{
		return Math::Pose(motion.orientation,motion.position);
	}


protected:
	Math::Quaternion lastOrientation;

	std::vector< Math::Vector< 3> > waypoints;
	MotionParameters mparams;
	Motion motion;


	
	int lastWP;
	int nextWP;
	int nextnextWP;
	Math::Vector<3> path, nextpath;
	bool gotonextwp;
	bool slow;
	double slowtest;

	bool init;
	// select random number generator
	boost::variate_generator<boost::mt19937&, boost::normal_distribution<double> >*  normal_sampler;// (boost::mt19937,boost::normal_distribution<double>(double));
	

	Math::Vector< 3> CrossProduct(Math::Vector< 3>& a, Math::Vector< 3>& b)
	{
		return Math::Vector< 3> (a[1]*b[2]-a[2]*b[1],
								 a[2]*b[0]-a[0]*b[2],
								 a[0]*b[1]-a[1]*b[0]);
	}

	void Normalize(Math::Vector< 3>& trans)
	{
		double len = boost::numeric::ublas::norm_2(trans);
		trans=trans/len;
	}

	double Noise()
	{
		return 1.0;
	}

	Math::Vector< 3> Noise3D()
	{
		return Math::Vector< 3> ((*normal_sampler)(),(*normal_sampler)(),(*normal_sampler)());
	}
};

} } } // namespace Ubitrack::Drivers::MotionModel
