/*
 * Ubitrack - Library for Ubiquitous Tracking
 * Copyright 2006, Technische Universitaet Muenchen, and individual
 * contributors as indicated by the @authors tag. See the
 * copyright.txt in the distribution for a full listing of individual
 * contributors.
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA, or see the FSF site: http://www.fsf.org.
 */


/**
 * @ingroup driver_components
 * @file
 * PoseGenerator driver
 * This file contains the driver component to
 * generate rotation and position information from a waypoint list
  *
 * The generated data is sent via a push interface.
 *
 * @author Benjamin Becker <Benjamin.Becker@eads.net>
 */

#include <string>
#include <cstdlib>
#include <log4cpp/Category.hh>

#include <Ubitrack/Dataflow/PushSupplier.h>
#include <Ubitrack/Dataflow/ComponentFactory.h>
#include <Ubitrack/Dataflow/Component.h>
#include <Ubitrack/Measurement/Measurement.h>
#include <Ubitrack/Measurement/TimestampSync.h>
#include <Ubitrack/Util/Exception.h>
#include <Ubitrack/Util/OS.h>

	
#include "MotionModel.h"


static log4cpp::Category& logger( log4cpp::Category::getInstance( "Drivers.PoseGenerator" ) );

namespace Ubitrack { namespace Drivers {

using namespace Dataflow;

/**
 * Module for XSens tracker, should work on both linux and windows.
 * Requires the XSens ActiveX plugin to be installed on Windows.
 */
class PoseGenerator
	: public Component
{
public:

	/** constructor */
	PoseGenerator( const std::string& sName, boost::shared_ptr< Graph::UTQLSubgraph > subgraph )
		: Component( sName )
		, m_port( "Output", *this )
		, m_keepRunning(1)
	{
		if (subgraph->m_DataflowConfiguration.getXML() == NULL)
			UBITRACK_THROW( "No configuration for PoseGenerator component." );

		const TiXmlNode* confignode;
		MotionModel::MotionParameters mparams;
		mparams.noiseLevel = 0;						//no noise
		mparams.noiseSeed = Measurement::now();
		mparams.maxAcceleration=1.0;						//1.0m/sec
		mparams.maxRotAcceleration = 0.1;					//0.1/sec
		mparams.mass = 1.0;								//1.0kg
		mparams.maxVelocity = 2.0;						//2.0m/s
		mparams.lookahead = false; 

		confignode = subgraph->m_DataflowConfiguration.getXML()->FirstChild( "Settings" );
		if ((confignode == NULL) || (confignode->ToElement() == NULL))
			UBITRACK_THROW( "No Settings defined in PoseGenerator component configuration." );
		
		if ( confignode->ToElement()->QueryDoubleAttribute( "frequency", &frequency ) != TIXML_SUCCESS )
		{
			LOG4CPP_DEBUG( logger, "No frequency specified for PoseGenerator motion model, taking 60Hz" );
			frequency = 60;
		}
		if(frequency > 100)
			UBITRACK_THROW( "Frequency should be < 100 (Hz) in PoseGenerator component configuration." );
        
		if ( confignode->ToElement()->QueryDoubleAttribute( "noiselevel", &mparams.noiseLevel) != TIXML_SUCCESS )
		{
			LOG4CPP_DEBUG( logger, "No noise used in PoseGenerator motion model" );
			if ( confignode->ToElement()->QueryIntAttribute( "noiseseed", &mparams.noiseSeed) != TIXML_SUCCESS )
				LOG4CPP_DEBUG( logger, "Picking random noise seed for PoseGenerator motion model" );
		}

		if ( confignode->ToElement()->QueryDoubleAttribute( "acceleration", &mparams.maxAcceleration) != TIXML_SUCCESS )
			LOG4CPP_DEBUG( logger, "No acceleration defined for PoseGenerator motion model, using 1m/s" );

		if ( confignode->ToElement()->QueryDoubleAttribute( "rotacceleration", &mparams.maxRotAcceleration) != TIXML_SUCCESS )
			LOG4CPP_DEBUG( logger, "No rotation acceleration (rotacceleration) defined for PoseGenerator motion model, using 0.1/s" );

		if ( confignode->ToElement()->QueryDoubleAttribute( "mass", &mparams.mass) != TIXML_SUCCESS )
			LOG4CPP_DEBUG( logger, "Picking random noise seed for PoseGenerator motion model, using 1kg" );

		if ( confignode->ToElement()->QueryDoubleAttribute( "maxvelocity", &mparams.maxVelocity) != TIXML_SUCCESS )
			LOG4CPP_DEBUG( logger, "No maximum velocity (maxvelocity) defined for PoseGenerator motion model, using 2m/s" );

		confignode = subgraph->m_DataflowConfiguration.getXML()->FirstChild( "CenterOfInterest" );
		if ((confignode == NULL) || (confignode->ToElement() == NULL))
		{
			LOG4CPP_DEBUG( logger, "No CenterOfInterest defined in PoseGenerator component configuration." );
			mparams.lookahead = true; 
		}
		else
		{
			if ( confignode->ToElement()->QueryDoubleAttribute( "x", &mparams.centerOfInterest[0] ) != TIXML_SUCCESS )
				UBITRACK_THROW( "No \"x\" element in CenterOfInterest defined." );
			if ( confignode->ToElement()->QueryDoubleAttribute( "y", &mparams.centerOfInterest[1] ) != TIXML_SUCCESS )
				UBITRACK_THROW( "No \"y\" element in CenterOfInterest defined." );
			if ( confignode->ToElement()->QueryDoubleAttribute( "z", &mparams.centerOfInterest[2] ) != TIXML_SUCCESS )
				UBITRACK_THROW( "No \"z\" element in CenterOfInterest defined." );
		}

		confignode = subgraph->m_DataflowConfiguration.getXML()->FirstChild( "Waypoints" );
		if ((confignode == NULL) || (confignode->ToElement() == NULL))
			UBITRACK_THROW( "No Waypoints defined in PoseGenerator component configuration." );

		Math::Vector< 3 > waypoint;
		const TiXmlNode* configwpnode = confignode->FirstChild( "point" );
		while(configwpnode != NULL)
		{
			if ( configwpnode->ToElement()->QueryDoubleAttribute( "x", &waypoint[0] ) != TIXML_SUCCESS )
				UBITRACK_THROW( "No \"x\" element in waypoint defined." );
			if ( configwpnode->ToElement()->QueryDoubleAttribute( "y", &waypoint[1] ) != TIXML_SUCCESS )
				UBITRACK_THROW( "No \"y\" element in waypoint defined." );
			if ( configwpnode->ToElement()->QueryDoubleAttribute( "z", &waypoint[2] ) != TIXML_SUCCESS )
				UBITRACK_THROW( "No \"z\" element in waypoint defined." );
		
			waypoints.push_back(waypoint);
			configwpnode = configwpnode->NextSibling("point" );
		}
		if(waypoints.size() < 2)
			UBITRACK_THROW( "PoseGenerator component configuration should have more than one waypoint defined" );

		motionModel = new MotionModel::MotionModelComponent(waypoints,mparams);

//		if ( node->ToElement()->QueryIntAttribute( "num", &m_portNum ) != TIXML_SUCCESS )
//			UBITRACK_THROW( "No 'num' attribute in XsensLegacy component configuration." );

		// start the library thread
		LOG4CPP_DEBUG( logger, "Starting pose data generation" );

		// start our own polling thread
		m_thread.reset( new boost::thread( boost::bind( &PoseGenerator::threadHandler, this ) ) );
	}

	/** destructor */
	~PoseGenerator()
	{
		LOG4CPP_DEBUG( logger, "Stopping pose data generation" );
		m_keepRunning = 0;
		m_thread->join();	
	}

	/** thread method */
	void threadHandler()
	{
		LOG4CPP_DEBUG( logger, "Pose data generation started" );
		Util::sleep(1000);
		Measurement::Timestamp now;
		Measurement::Timestamp last = 0;
		int delta = 1E9/frequency; // send once per delta (in nanosec)
		int deltatime;
		while (m_keepRunning)
		{
			motionModel->Update();
			now = Measurement::now();
			deltatime = now - last;
			if(deltatime > delta)
			{
				Measurement::Pose pose(	Measurement::now(),motionModel->GetPose());
				last = now;
				m_port.send( pose );
				LOG4CPP_DEBUG( logger, "PoseGenerator measurement: " << *pose );
			}
			
			// wait 10ms until next poll
			Util::sleep(10);
		}

		LOG4CPP_DEBUG( logger, "PoseGenerator thread stopped" );
	}
protected:

	boost::shared_ptr< boost::thread > m_thread;
	PushSupplier< Ubitrack::Measurement::Pose > m_port;
	int m_keepRunning;
	std::vector < Math::Vector< 3 > > waypoints;

	double frequency;
	Measurement::Timestamp last;


	MotionModel::MotionModelComponent* motionModel;
};


UBITRACK_REGISTER_COMPONENT( ComponentFactory* const cf ) {
	cf->registerComponent< PoseGenerator > ( "PoseGenerator" );
}


} } // namespace Ubitrack::Drivers
