/*
 * Ubitrack - Library for Ubiquitous Tracking
 * Copyright 2006, Technische Universitaet Muenchen, and individual
 * contributors as indicated by the @authors tag. See the
 * copyright.txt in the distribution for a full listing of individual
 * contributors.
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA, or see the FSF site: http://www.fsf.org.
 */

#include <Ubitrack/config.h>

#ifdef HAVE_GLEW
	#include "GL/glew.h"
#endif

#include "GL/freeglut.h"

#include "RenderModule.h"

log4cpp::Category& logger( log4cpp::Category::getInstance( "Drivers.Render" ) );

#ifdef HAVE_OPENCV
	#include "BackgroundImage.h"
	#include "ZBufferOutput.h"
	#include "ImageOutput.h"
#endif

#ifdef HAVE_LAPACK
	#include "PoseErrorVisualization.h"
	#include "PositionErrorVisualization.h"
#endif

#ifdef HAVE_COIN
	#include "InventorObject.h"
#endif

#include "X3DObject.h"
#include "VectorfieldViewer.h"
#include "AntiMarker.h"
#include "PointCloud.h"
#include "Intrinsics.h"
#include "CameraPose.h"
#include "ButtonOutput.h"
#include "DropShadow.h"
#include "WorldFrame.h"
#include "Skybox.h"
#include "Projection.h"
#include "Projection3x4.h"
#include "StereoSeparation.h"
#include "Cross2D.h"

#include <Ubitrack/Util/Exception.h>
#include <Ubitrack/Util/OS.h>
#include <boost/thread/xtime.hpp>
#include <boost/scoped_ptr.hpp>
#include <iomanip>
#include <math.h>


namespace Ubitrack { namespace Drivers {

using namespace Dataflow;


// some unavoidable globals to manage the GLUT main loop and its handlers
std::deque< VirtualCamera* > g_setup;
std::map< std::string, int > g_names;
std::map< int, VirtualCamera* > g_modules;
boost::scoped_ptr< boost::thread > g_glutThread;
boost::mutex g_globalMutex;
boost::condition g_continue;
#ifdef __APPLE__
	bool g_glutInitialized = false;
#endif

int g_run = 1;

// fake command line for GLUT (yes, the library _is_ 10 years old)
int   g_argc   = 1;
char* g_argv[] = { "VirtualCamera", 0 };



void g_mainloop()
{
	LOG4CPP_DEBUG( logger, "Render thread started" );

	boost::mutex::scoped_lock l( g_globalMutex );

#ifdef __APPLE__
	if ( !g_glutInitialized )
	{	
		g_glutInitialized = true;
#else
	if ( !glutGet(GLUT_INIT_STATE) )
	{
#endif
		glutInit( &g_argc, g_argv );
		glutInitDisplayMode( GLUT_RGBA | GLUT_DOUBLE | GLUT_DEPTH );
		glutSetOption( GLUT_ACTION_ON_WINDOW_CLOSE, GLUT_ACTION_CONTINUE_EXECUTION );
	}

	while (g_run) {

		// are there any setup functions pending?
		if ( g_setup.size() > 0 ) {
			LOG4CPP_DEBUG( logger, "Calling setup()" );

			// try to setup the first window in the queue
			VirtualCamera* tmp = g_setup.front(); g_setup.pop_front();
			// if this failed, push it to the end of the queue
			if ( !(tmp->setup()) ) g_setup.push_back( tmp );
		}

		// check
		// - if a redraw is needed for any window
		// - if any windows have been deleted

		std::map<int,VirtualCamera*>::iterator pos = g_modules.begin();
		std::map<int,VirtualCamera*>::iterator end = g_modules.end();

		while ( pos != end )
			if (pos->second) {
				pos->second->redraw();
				pos++;
			} else {
				LOG4CPP_DEBUG( logger, "Destroying window" );
				glutDestroyWindow( pos->first );
				g_modules.erase( pos++ );

				// quit thread when last module is destroyed
				if ( g_modules.empty() )
				{
					LOG4CPP_DEBUG( logger, "Render thread stopped" );
					return;
				}
			}

		// let GLUT do its thing..
		glutMainLoopEvent();

		// Be nice to the rest of the system. g_glutThread->yield() doesn't have any noticeable effect
		// with timeslices of the size that is common today. 5 ms maxes the frame rate at 200 Hz, but 
		// that should be sufficient for most kinds of hardware.
		// FIXME: probably needs to be increased to at least 60 Hz again for frame-sequential stereo
		// Can this only be done when stereo is actually used? Otherwise this will be a waste of processing time.(DP)
		boost::xtime t;
		boost::xtime_get( &t, boost::TIME_UTC );
		t.nsec += 100000000;
		t.sec  += t.nsec / 1000000000;
		t.nsec %= 1000000000;
		g_continue.timed_wait( l, t );
		LOG4CPP_DEBUG( logger, "timed_wait finished" );
	}
}


void g_display()
{
	VirtualCamera* win = g_modules[ glutGetWindow() ];
	if ( win ) win->display();
}


void g_keyboard( unsigned char key, int x, int y )
{
	VirtualCamera* win = g_modules[ glutGetWindow() ];
	if ( win ) win->keyboard( key, x, y );
}


void g_reshape( int w, int h )
{
	VirtualCamera* win = g_modules[ glutGetWindow() ];
	if ( win ) win->reshape( w, h );
}



int VirtualCamera::setup()
{
    LOG4CPP_DEBUG( logger, "starting setup ************************");
	// is this supposed to be a subwindow?
	if (m_moduleKey.substr(0,3) == "Sub") {
		// check if the parent has already been created
		std::string parent = m_moduleKey.substr(3);
		if (g_names.find( parent ) == g_names.end()) return 0;
		// parent is there, so create the subwindow
		m_winHandle = glutCreateSubWindow( g_names[parent], 0, 0, m_width, m_height );
	} else {
		// create new top level window, 
		glutInitWindowSize( m_width, m_height );
		m_winHandle = glutCreateWindow( m_moduleKey.c_str() );
	}

	// store this-pointer and window handle
	g_modules[ m_winHandle ] = this;
	g_names[ m_moduleKey ] = m_winHandle;

	// GLEW provides access to OpenGL extensions
	#ifdef HAVE_GLEW
		glewInit();
	#endif

	// GL: enable and set colors
	glEnable( GL_COLOR_MATERIAL );
	glClearColor( 0.0, 0.0, 0.0, 1.0 ); // TODO: make this configurable (but black is best for optical see-through ar!)

	// GL: enable and set depth parameters
	glEnable( GL_DEPTH_TEST );
	glClearDepth( 1.0 );

	// GL: disable backface culling
	glPolygonMode( GL_FRONT_AND_BACK, GL_FILL );
	glDisable( GL_CULL_FACE );

	// GL: light parameters
	GLfloat light_pos[] = { 1.0f, 1.0f, 1.0f, 0.0f };
	GLfloat light_amb[] = { 0.2f, 0.2f, 0.2f, 1.0f };
	GLfloat light_dif[] = { 0.9f, 0.9f, 0.9f, 1.0f };

	// GL: enable lighting
	glLightfv( GL_LIGHT0, GL_POSITION, light_pos );
	glLightfv( GL_LIGHT0, GL_AMBIENT,  light_amb );
	glLightfv( GL_LIGHT0, GL_DIFFUSE,  light_dif );
	glEnable( GL_LIGHTING );
	glEnable( GL_LIGHT0 );

	// GL: bitmap handling
	glPixelStorei( GL_PACK_ALIGNMENT,   1 );
	glPixelStorei( GL_UNPACK_ALIGNMENT, 1 );

	// GL: alpha blending
	glBlendFunc( GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA );
	glEnable( GL_BLEND );

	// GL: misc stuff
	glShadeModel( GL_SMOOTH );
	glEnable( GL_NORMALIZE );

	// make functions known to GLUT
	glutKeyboardFunc( g_keyboard );
	glutDisplayFunc ( g_display  );
	glutReshapeFunc ( g_reshape  );
	
 
     ComponentList objects = getAllComponents();
	for ( ComponentList::iterator i = objects.begin(); i != objects.end(); i++ )
	{
        (*i)->loadTexture();
    }
     
	return 1;
}


void VirtualCamera::invalidate( VirtualObject* caller )
{
	if (m_redraw) return;
	// check if this is the last incoming update of several concurrent ones
	ComponentList objects = getAllComponents();
	for ( ComponentList::iterator i = objects.begin(); i != objects.end(); i++ ) {
		if ((*i).get() == caller) continue;
		if ((*i)->hasWaitingEvents()) return;
	}
	m_redraw = 1;
	LOG4CPP_DEBUG( logger, "waking up main thread" );
	g_continue.notify_all();
}

void VirtualCamera::redraw( )
{
	// TODO: fix stereo view handling
	// TODO: make minmum fps configureable (currently 2fps)
	if ((!m_redraw) && (m_moduleKey.m_stereo == 0.0 && m_lastRedrawTime + 500000000L < Measurement::now()  )) return; 
	glutSetWindow( m_winHandle );
	LOG4CPP_DEBUG( logger, "calling glutPostRedisplay" );
	glutPostRedisplay();
	m_redraw = 0;
}


VirtualCamera::VirtualCamera( const VirtualCameraKey& key, boost::shared_ptr< Graph::UTQLSubgraph >, FactoryHelper* pFactory )
	: Module< VirtualCameraKey, VirtualObjectKey, VirtualCamera, VirtualObject >( key, pFactory )
	, m_width(key.m_width)
	, m_height(key.m_height)
	, m_near(key.m_near)
	, m_far(key.m_far)
	, m_winHandle(0)
	, m_redraw(1)
	, m_doSync(0)
	, m_parity(0)
	, m_info(0)
	, m_lasttime(0)
	, m_lastframe(0)
	, m_fps(0)
	, m_lastRedrawTime(0)
	, m_vsync()
{
	// lock access to globals
	boost::mutex::scoped_lock l( g_globalMutex );

	// schedule the setup function for this window
	g_setup.push_back( this );

	// if there's no thread yet, init GLUT library first and create a new control thread
	// Note: this thread must do ALL OpenGL operations for all VirtualCameras!
	// Most GL implementation do not look kindly on context sharing between threads!
	if ( g_glutThread.get() == 0 )
		g_glutThread.reset( new boost::thread( boost::bind( g_mainloop ) ) );
}


VirtualCamera::~VirtualCamera()
{
	bool bKillThread = false;

	{
		// lock access to globals and remove the stored this-pointer
		boost::mutex::scoped_lock l( g_globalMutex );
		g_modules[ m_winHandle ] = 0;
		g_names.erase( m_moduleKey );

		// kill thread if this was the last window
		if ( g_names.empty() )
			bKillThread = true;
	}

	// the thread should die with the last render window
	if ( bKillThread )
	{
		LOG4CPP_DEBUG( logger, "Waiting for render thread to stop" );
		g_glutThread->join();
		boost::mutex::scoped_lock l( g_globalMutex );
		g_glutThread.reset();
	}

	/*if (m != n) UBITRACK_THROW( "assertion 'g_modules.size() == g_names.size()' failed!" );*/
}


void VirtualCamera::keyboard( unsigned char key, int x, int y )
{
	if (glutGetModifiers() & GLUT_ACTIVE_ALT)
	{
		switch ( key )
		{
			case 'f': glutFullScreen(); break;
			case 'i': m_info = !m_info; break;
			case 'v': m_doSync = !m_doSync; break;
			case 's': m_parity = !m_parity; break;
			// case 'q': delete this; break;
		}
	}
	else
	{
		m_lastKey = key;
	}
	glutPostRedisplay();
}


unsigned char VirtualCamera::getLastKey() {
	unsigned char tmp = m_lastKey;
	m_lastKey = 0;
	return tmp;
}




void VirtualCamera::display()
{

	m_lastRedrawTime = Measurement::now();

	// get frame counters and parity
	int curframe = m_vsync.getFrame();
	int retrace  = m_vsync.getRetrace();
	int parity = (retrace%2 == m_parity); // Nick: This seems to be broken on linux

	// predict a little bit (only for pull inputs)
	Measurement::Timestamp imageTime( Measurement::now() + 5000000L );

	// clear buffers
	glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );

	glMatrixMode( GL_MODELVIEW );
	glLoadIdentity();

	// calculate fps
	int curtime = glutGet( GLUT_ELAPSED_TIME );
	if ((curtime - m_lasttime) >= 1000) {
		m_fps = (1000.0*(curframe-m_lastframe))/((double)(curtime-m_lasttime));
		m_lasttime  = curtime;
		m_lastframe = curframe;
	}

	LOG4CPP_DEBUG( logger, "Redrawing.." );
	// do we want a stereo view?
	/* Nick: Disabled due to new StereoSeparation component.
	if (m_moduleKey.m_stereo == 0.0) {
		//drawBackground( 0 );1
		parity = 0;
	} else {
		//TODO: stereo separation has to happen in a separate component after the bg-img
		double sep = m_moduleKey.m_stereo/2.0;
		//if (parity) sep = -sep;
		//drawBackground( parity );
		if (parity) glTranslated( sep, 0, 0 );
	}
	*/

	// iterate over all components (already sorted by priority thanks to std::map)
	ComponentList objects = getAllComponents();
	for ( ComponentList::iterator i = objects.begin(); i != objects.end(); i++ )
	{
		try
		{
//             ComponentList::iterator i = objects.begin();
			(*i)->draw( imageTime, 0 ); // Parity = 0
		}
		catch( const Util::Exception& e )
		{
			LOG4CPP_WARN( logger, "Exception in main loop from component ");// << (*i)->getName() << ": " << e );
		}
	}

	if (m_moduleKey.m_stereo != 0.0) {

		// 2nd rendering pass for stereo separation.
		// Only makes sense with color mask stereo separation.
		glClear( GL_DEPTH_BUFFER_BIT ); // Let color buffer intact, only clear depth information.
		glMatrixMode( GL_MODELVIEW );
		glLoadIdentity(); // Reset transformation stack.

		for ( ComponentList::iterator i = objects.begin(); i != objects.end(); i++ )
		{
			try
			{        
				(*i)->draw( imageTime, 1 ); // Parity = 1
			}
			catch( const Util::Exception& e )
			{
				LOG4CPP_WARN( logger, "Exception in main loop from component " << (*i)->getName() << ": " << e );
			}
		}
	}

	// print info string
	if (m_info) {
  
		std::ostringstream text;
		text << std::fixed << std::showpoint << std::setprecision(2);
		text << "FPS: " << m_fps;
		text << " VSync: " << (m_doSync?"on":"off");
		text << " StereoSep: " << m_moduleKey.m_stereo;

		glMatrixMode( GL_MODELVIEW );
		glLoadIdentity();
		glMatrixMode( GL_PROJECTION ); 
		glPushMatrix();
		glLoadIdentity();
		gluOrtho2D( 0, m_width, 0, m_height );
		glDisable( GL_LIGHTING );

		GLfloat xzoom, yzoom;
		glGetFloatv( GL_ZOOM_X, &xzoom );
		glGetFloatv( GL_ZOOM_Y, &yzoom );

		glPixelZoom( 1.0, 1.0 );

		glColor4f( 1.0, 0.0, 0.0, 1.0 );
		glRasterPos2i( 10, m_height-23 );

		for ( unsigned int i = 0; i < text.str().length(); i++ )
			glutBitmapCharacter( GLUT_BITMAP_8_BY_13, text.str()[i] );

		glPixelZoom( xzoom, yzoom );

		glPopMatrix();
		glEnable( GL_LIGHTING );
	}

	// wait for the screen refresh
	m_vsync.wait( m_doSync );
	
	// put current buffer into display
	LOG4CPP_DEBUG( logger, "Swapping buffers.." );
	glutSwapBuffers();
}


void VirtualCamera::reshape( int w, int h )
{
	// store new window size for reference
	m_width  = w;
	m_height = h;

	// set a whole-window viewport
	glViewport( 0, 0, m_width, m_height );

	// create a perspective projection matrix
	glMatrixMode( GL_PROJECTION );
	glLoadIdentity();
	gluPerspective( m_moduleKey.m_fov, ((double)m_width/(double)m_height), m_moduleKey.m_near, m_moduleKey.m_far );

	// invalidate display
	glutPostRedisplay();
}


boost::shared_ptr< VirtualObject > VirtualCamera::createComponent( const std::string& type, const std::string& name, 
	boost::shared_ptr< Graph::UTQLSubgraph > pConfig, const VirtualObjectKey& key, VirtualCamera* pModule )
{
    LOG4CPP_DEBUG( logger, "starting createComponent ************************");
	if ( type == "X3DObject" )
		return boost::shared_ptr< VirtualObject >( new X3DObject( name, pConfig, key, pModule ) );
	else if ( type == "VectorfieldViewer" )
		return boost::shared_ptr< VirtualObject >( new VectorfieldViewer( name, pConfig, key, pModule ) );
	else if ( type == "AntiMarker" )
			return boost::shared_ptr< VirtualObject >( new AntiMarker( name, pConfig, key, pModule ) );
	else if ( type == "PointCloud" )
			return boost::shared_ptr< VirtualObject >( new PointCloud( name, pConfig, key, pModule ) );
	else if ( type == "Intrinsics" )
			return boost::shared_ptr< VirtualObject >( new Intrinsics( name, pConfig, key, pModule ) );
	else if ( type == "CameraPose" )
			return boost::shared_ptr< VirtualObject >( new CameraPose( name, pConfig, key, pModule ) );
	else if ( type == "ButtonOutput" )
			return boost::shared_ptr< VirtualObject >( new ButtonOutput( name, pConfig, key, pModule ) );
	else if ( type == "DropShadow" )
			return boost::shared_ptr< VirtualObject >( new DropShadow( name, pConfig, key, pModule ) );
	else if ( type == "WorldFrame" )
			return boost::shared_ptr< VirtualObject >( new WorldFrame( name, pConfig, key, pModule ) );	
  	else if ( type == "Skybox" ){
         return boost::shared_ptr< VirtualObject >( new Skybox( name, pConfig, key, pModule ) );	
         }
					
	else if ( type == "Projection" )
			return boost::shared_ptr< VirtualObject >( new Projection( name, pConfig, key, pModule ) );
	else if ( type == "Projection3x4" )
			return boost::shared_ptr< VirtualObject >( new Projection3x4( name, pConfig, key, pModule ) );
	else if ( type == "StereoSeparation" )
			return boost::shared_ptr< VirtualObject >( new StereoSeparation( name, pConfig, key, pModule ) );
	else if ( type == "Cross2D" )
			return boost::shared_ptr< VirtualObject >( new Cross2D( name, pConfig, key, pModule ) );


	#ifdef HAVE_OPENCV
	else if ( type == "ImageOutput" )
			return boost::shared_ptr< VirtualObject >( new ImageOutput( name, pConfig, key, pModule ) );
	else if ( type == "ZBufferOutput" )
			return boost::shared_ptr< VirtualObject >( new ZBufferOutput( name, pConfig, key, pModule ) );
	else if ( type == "BackgroundImage" )
			return boost::shared_ptr< VirtualObject >( new BackgroundImage( name, pConfig, key, pModule ) );
	#endif

	#ifdef HAVE_LAPACK
	else if ( type == "PoseErrorVisualization" )
			return boost::shared_ptr< VirtualObject >( new PoseErrorVisualization( name, pConfig, key, pModule ) );
	else if ( type == "PositionErrorVisualization" )
			return boost::shared_ptr< VirtualObject >( new PositionErrorVisualization( name, pConfig, key, pModule ) );
	#endif
	#ifdef HAVE_COIN
	else if ( type == "PoseErrorVisualization" )
			return boost::shared_ptr< VirtualObject >( new InventorObject( name, pConfig, key, pModule ) );
	#endif

	UBITRACK_THROW( "Class " + type + " not supported by render module" );
}


/** register module at factory */
UBITRACK_REGISTER_COMPONENT( ComponentFactory* const cf )
{
	std::vector< std::string > renderComponents;
	renderComponents.push_back( "X3DObject" );
	renderComponents.push_back( "VectorfieldViewer" );
	renderComponents.push_back( "AntiMarker" );
	renderComponents.push_back( "PointCloud" );
	renderComponents.push_back( "Intrinsics" );
	renderComponents.push_back( "CameraPose" );
	renderComponents.push_back( "ButtonOutput" );
	renderComponents.push_back( "DropShadow" );
    renderComponents.push_back( "WorldFrame" );
    renderComponents.push_back( "Skybox" );
	renderComponents.push_back( "Projection" );
	renderComponents.push_back( "Projection3x4" );
	renderComponents.push_back( "StereoSeparation" );
	renderComponents.push_back( "Cross2D" );
	#ifdef HAVE_OPENCV
		renderComponents.push_back( "ImageOutput" );
		renderComponents.push_back( "ZBufferOutput" );
		renderComponents.push_back( "BackgroundImage" );
	#endif

	#ifdef HAVE_LAPACK
		renderComponents.push_back( "PoseErrorVisualization" );
		renderComponents.push_back( "PositionErrorVisualization" );
	#endif
	#ifdef HAVE_COIN
		renderComponents.push_back( "InventorObject" );
	#endif
	cf->registerModule< VirtualCamera > ( renderComponents );
}


} } // namespace Ubitrack::Drivers

