/*
 * Ubitrack - Library for Ubiquitous Tracking
 * Copyright 2006, Technische Universitaet Muenchen, and individual
 * contributors as indicated by the @authors tag. See the
 * copyright.txt in the distribution for a full listing of individual
 * contributors.
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA, or see the FSF site: http://www.fsf.org.
 */


/**
 * @ingroup driver_components
 * @file
 * Spinball driver
 * This file contains the driver component to
 * get rotation data from the Spinball device.
 *
 * The received data is sent via a push interface.
 *
 * @author Florian Echtler <echtler@in.tum.de>
 */

#include <asm/types.h>

#include <sys/ioctl.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/types.h>

#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <fcntl.h>
#include <stdint.h>
#include <math.h>

#include <linux/input.h>
#include <algorithm>
#include <iostream>

#include <string>
#include <cstdlib>
#include <log4cpp/Category.hh>

#include <Ubitrack/Dataflow/PushSupplier.h>
#include <Ubitrack/Dataflow/ComponentFactory.h>
#include <Ubitrack/Dataflow/Component.h>
#include <Ubitrack/Measurement/Measurement.h>
#include <Ubitrack/Measurement/TimestampSync.h>
#include <Ubitrack/Util/Exception.h>
#include <Ubitrack/Util/OS.h>

static log4cpp::Category& logger( log4cpp::Category::getInstance( "Drivers.Spinball" ) );

namespace Ubitrack { namespace Drivers {

const float fx = 2*M_PI/6677.0;
const float fy = 2*M_PI/6598.0;
const float fz = 2*M_PI/3753.0;

class DataGrabber {

	public:

		DataGrabber(int fd1, int fd2) {
			m1 = fd1;
			m2 = fd2;
		}

		int get_relative( int fd, int& x, int &y ) {

			int retval = 0;
			struct input_event ev[1024]; // the events (up to 1024 at once)
			size_t read_bytes = read( fd, ev, sizeof(ev) );

			if (read_bytes < (int)sizeof(struct input_event)) {
				perror( "evtest: short read" );
				exit(1);
			}

			for ( int yalv = 0; yalv < (int)(read_bytes / sizeof(struct input_event)); yalv++ ) {
				if (ev[yalv].type == 0) { retval = 1; continue; } // sync
				if (ev[yalv].type != 2) continue; // relative
				switch (ev[yalv].code) {
					case 0: x += ev[yalv].value; break; // x axis
					case 1: y += ev[yalv].value; break; // y axis
				}
			}

			return retval;
		}

		Math::Quaternion grab() {

			fd_set readset;

			FD_ZERO( &readset );
			FD_SET( m1, &readset );
			FD_SET( m2, &readset );

			select( std::max(m1,m2)+1, &readset, 0, 0, 0 );

			int s1 = 0;
			int s2 = 0;

			int delta[2][2] = { {0,0}, {0,0} };

			if (FD_ISSET( m1, &readset )) s1 = get_relative( m1, delta[0][0], delta[0][1] );
			if (FD_ISSET( m2, &readset )) s2 = get_relative( m2, delta[1][0], delta[1][1] );

			/*std::cout << "delta1: " << delta[0][0] << "," << delta[0][1] << "       ";
			std::cout << "delta2: " << delta[1][0] << "," << delta[1][1] << std::endl;*/

			float dx = -(delta[0][0]+delta[1][0])/2; // average x component
			float dy = -(delta[0][1]+delta[1][1])/2; // common y component
			float dz = -(delta[0][1]-delta[1][1])/2; // opposing y component

			dx *= fx; dy *= fy; dz *= fz;

			Math::Quaternion deltax( sin(dx/2), 0, 0, cos(dx/2) );
			Math::Quaternion deltay( 0, sin(dy/2), 0, cos(dy/2) );
			Math::Quaternion deltaz( 0, 0, sin(dz/2), cos(dz/2) );

			deltaz = deltaz * deltay;
			deltaz = deltaz * deltax;

			//rotation = deltaz * rotation;
			return deltaz;
		}

	protected:

		int m1,m2;
};


using namespace Dataflow;

class Spinball
	: public Component
{
public:

	/** constructor */
	Spinball( const std::string& sName, boost::shared_ptr< Graph::UTQLSubgraph > subgraph )
		: Component( sName )
		, m_port( "Output", *this )
	{
		// read device name from configuration
		m_dev1 = subgraph->m_DataflowAttributes.getAttributeString( "device1" );
		m_dev2 = subgraph->m_DataflowAttributes.getAttributeString( "device2" );
		if (m_dev1.empty() || m_dev2.empty())
			UBITRACK_THROW( "No 'device1/2' attributes in Spinball component configuration." );

		LOG4CPP_INFO( logger, "Connecting to Spinball at ports " << m_dev1 << " / " << m_dev2 );

		int grab = 1;  // for the GRAB ioctl
		int m1 = 0;    // the file descriptor for the device
		int m2 = 0;

		// open file descriptors
		m1 = open( m_dev1.c_str(), O_RDONLY );
		m2 = open( m_dev2.c_str(), O_RDONLY );

		if (!m1 || !m2) UBITRACK_THROW( "Error opening devices." );

		// grab devices to avoid output in /dev/input/mice
		if (ioctl( m1, EVIOCGRAB, &grab )) perror("ioctl EVIOCGRAB: error ");
		if (ioctl( m2, EVIOCGRAB, &grab )) perror("ioctl EVIOCGRAB: error ");

		m_grabber = new DataGrabber(m1,m2);

		// start our own polling thread
		m_thread.reset( new boost::thread( boost::bind( &Spinball::threadHandler, this ) ) );
	}

	/** destructor */
	~Spinball()
	{
		m_keepRunning = 0;
		m_thread->join();
	}

	/** thread method */
	void threadHandler()
	{
		LOG4CPP_DEBUG( logger, "Spinball thread started" );

		while ( m_keepRunning ) {
			// get orientation and/or rotation velocity (different for windows and linux)
			Math::Quaternion res = m_grabber->grab();
			m_result = res * m_result;
			
			LOG4CPP_DEBUG( logger, "Spinball measurement: " << m_result );
			m_port.send( Measurement::Rotation( Measurement::now(), m_result ) );
			
			// wait 2ms until next poll
			//Util::sleep( 2 );
		}

		LOG4CPP_DEBUG( logger, "Spinball thread stopped" );
	}

protected:

	int m_keepRunning;
	std::string m_dev1,m_dev2;

	DataGrabber* m_grabber;
	Math::Quaternion m_result;

	boost::shared_ptr< boost::thread > m_thread;
	PushSupplier< Ubitrack::Measurement::Rotation > m_port;
	
};


UBITRACK_REGISTER_COMPONENT( ComponentFactory* const cf ) {
	cf->registerComponent< Spinball > ( "Spinball" );
}


} } // namespace Ubitrack::Drivers

