/*
 * Ubitrack - Library for Ubiquitous Tracking
 * Copyright 2006, Technische Universitaet Muenchen, and individual
 * contributors as indicated by the @authors tag. See the 
 * copyright.txt in the distribution for a full listing of individual
 * contributors.
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA, or see the FSF site: http://www.fsf.org.
 */

#include "SwissRanger.h"

#include <stdio.h>

#include <boost/shared_ptr.hpp>
#include <boost/scoped_ptr.hpp>


#include <boost/bind.hpp>

#include <log4cpp/Category.hh>
static log4cpp::Category& logger( log4cpp::Category::getInstance( "Drivers.Swissranger" ) );

namespace Ubitrack { namespace Drivers {

SwissRanger::SwissRanger ( const std::string& sName, boost::shared_ptr< Graph::UTQLSubgraph > pConfig )
		: Component( sName )
		, m_img_height( 0 )
		, m_img_width( 0 )
		, m_img_types( 0 )
		, m_bStop ( true )
		, m_portOut1 ( "Output", *this )
		, m_portOut2 ( "Amplitude", *this )

	{
	
		unsigned short version[4];
		
		if( SR_GetVersion( version ) != 0 )
			UBITRACK_THROW( "Cannot access MESA Swissranger SDK" );
		LOG4CPP_DEBUG( logger, "SDK Version " << version[0] << "." <<version[1] << "." << version[2] << "." << version[3] );

		// SR_CheckForNewDllVersion( 0 );
		// SR_CheckForNewDllVersion( 1 );
		// SR_CheckForNewDllVersion( 2 );
		
#ifdef _WIN32
		if( SR_OpenDlg( &srCam, 3, 0 ) <= 0 )
			UBITRACK_THROW( "Cannot find any MESA SwissRanger" );
#else
		if( SR_OpenUSB( &srCam, 0 ) <= 0 )
			UBITRACK_THROW( "Cannot find MESA SwissRanger on USB" );
#endif

		//SR_GetDeviceString()
		m_img_width = SR_GetCols( srCam );
		m_img_height = SR_GetRows( srCam );
};

void SwissRanger::start()
{
	if( m_bStop )
	{
		m_bStop = false;
		m_thread.reset( new boost::thread( boost::bind( &SwissRanger::threadProc, this ) ) );
	}
		
}

void SwissRanger::threadProc( )
{
	LOG4CPP_INFO( logger, "starting main loop" );
	if( m_portOut2.isConnected() )
	{
		SR_SetAutoExposure( srCam, 5, 255, 10, 45 ); //values are recommended from API-documentation
		//SR_SetAutoExposure( srCam, 255, 0, 0, 0 ); //values that are NOT recommended from API-documentation
		SR_SetMode( srCam, AM_CONV_GRAY ); //makes the amplitude image look like intensity-image
	}	
	
	boost::scoped_ptr< Vision::Image > imageIn( new Vision::Image ( m_img_width, m_img_height, 1, IPL_DEPTH_16U ) );
	ImgEntry *imgEntryArray;
	m_img_types = SR_GetImageList( srCam, &imgEntryArray );
	
	LOG4CPP_INFO( logger, "detected " << m_img_types << " different kind of images." );


	// double *x, *y, *z;
	// x = ( double* ) malloc( m_img_width * m_img_height * sizeof( double ) );
	// y = ( double* ) malloc( m_img_width * m_img_height * sizeof( double ) );
	// z = ( double* ) malloc( m_img_width * m_img_height * sizeof( double ) );

	

	while( !m_bStop )
	{

		if( 0 < SR_Acquire( srCam ) )
		{
			Measurement::Timestamp timestamp = Measurement::now();
			
			if( m_portOut1.isConnected() ) //the distance image
			{
				imageIn->imageData = ( char* ) SR_GetImage( srCam, 0 );
				boost::shared_ptr< Vision::Image > imageOut( new Vision::Image ( m_img_width, m_img_height, 1, IPL_DEPTH_8U ) );
				cvConvertScaleAbs( *imageIn, *imageOut, 255/double( 0xffff ) );
				m_portOut1.send( Vision::ImageMeasurement( timestamp,  imageOut ) );
				LOG4CPP_DEBUG( logger, "Distance image sended" );
			}

			if( m_portOut2.isConnected() && m_img_types > 1 ) //the amplitude image
			{
				boost::shared_ptr< Vision::Image > imageOut( new Vision::Image ( m_img_width, m_img_height, 1, IPL_DEPTH_8U ) );
				imageIn->imageData = ( char* ) SR_GetImage( srCam, 1 );
				cvConvertScaleAbs( *imageIn, *imageOut, 1. );
				m_portOut2.send( Vision::ImageMeasurement( timestamp,  imageOut ) );
				LOG4CPP_DEBUG( logger, "Amplitude image sended" );
			}
		}
	}
	
	// free( x );
	// free( y );
	// free( z );
}

void SwissRanger::stop( )
{
	if( !m_bStop )
	{
		m_bStop = true;
		m_thread->join( );
	}
}


SwissRanger::~SwissRanger( )
{
	SR_Close( srCam );
}

UBITRACK_REGISTER_COMPONENT( Ubitrack::Dataflow::ComponentFactory* const cf ) {
	cf->registerComponent< SwissRanger > ( "SwissRanger" );
}

} } //namespace Ubitrack::Drivers
