/*
 * Ubitrack - Library for Ubiquitous Tracking
 * Copyright 2006, Technische Universitaet Muenchen, and individual
 * contributors as indicated by the @authors tag. See the
 * copyright.txt in the distribution for a full listing of individual
 * contributors.
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA, or see the FSF site: http://www.fsf.org.
 */


#include "UnifeyeModule.h"
#include <iostream>
#include <sstream>
#include <Ubitrack/Dataflow/ComponentFactory.h>
#include <Ubitrack/Math/Vector.h>
#include <Ubitrack/Math/Quaternion.h>
#include <Ubitrack/Math/Pose.h>
#include <Ubitrack/Math/Matrix.h>

#include <boost/array.hpp>

#include <log4cpp/Category.hh>


/* Example dataflow:
<Pattern name="Unifeye" id="Unifeye1">
	<Output>
		<Node name="Unifeye" id="A">
			<Attribute name="unifeyePort" value="5000"/>
		</Node>
		<Node name="COS" id="B">
			<Attribute name="unifeyeId" value="1"/> <!-- COS index -->
		</Node>
		<Edge name="Output" source="Unifeye" destination="COS">
			<Attribute name="type" value="6D"/>
		</Edge>
	</Output>

	<DataflowConfiguration>
		<UbitrackLib class="Unifeye"/>
	</DataflowConfiguration>
</Pattern>
*/


static double getXmlDoubleValue( TiXmlNode* pParent, const char* childName )
{
	TiXmlElement* pXml = pParent->FirstChildElement( childName );
	if ( !pXml || !pXml->FirstChild() || !pXml->FirstChild()->ToText() )
		UBITRACK_THROW( "XML: invalid " + std::string( childName ) );
	char* dummy = 0;
	return strtod( pXml->FirstChild()->Value(), &dummy );
}


namespace Ubitrack { namespace Drivers {

static log4cpp::Category& logger( log4cpp::Category::getInstance( "Drivers.Unifeye" ) );

UnifeyeModule::UnifeyeModule( const UnifeyeModuleKey& moduleKey, boost::shared_ptr< Graph::UTQLSubgraph >, FactoryHelper* pFactory )
	: Module< UnifeyeModuleKey, UnifeyeComponentKey, UnifeyeModule, UnifeyeComponent >( moduleKey, pFactory )
	, m_synchronizer( 1000 ) // assume milliseconds for timestamp synchronization
{
	using asio::ip::udp;

	LOG4CPP_NOTICE( logger, "Creating Unifeye network service" );
	LOG4CPP_DEBUG( logger, "Using port " << static_cast< int >( m_moduleKey ) );

	m_ioService.reset( new asio::io_service );

	m_socket.reset( new udp::socket( *m_ioService ) );
	m_socket->open( udp::v4() );
	asio::socket_base::reuse_address option( true );
	m_socket->set_option( option );
	m_socket->bind( udp::endpoint( udp::v4(), m_moduleKey ) );

	m_socket->async_receive_from(
		asio::buffer( m_receiveData, max_receive_length ), m_senderEndpoint,
		boost::bind( &UnifeyeModule::handleReceive, this, asio::placeholders::error, asio::placeholders::bytes_transferred ) );

	// network thread runs until io_service is interrupted
	m_networkThread = boost::shared_ptr< boost::thread >( new boost::thread ( boost::bind ( &asio::io_service::run, m_ioService.get() ) ) );
}


UnifeyeModule::~UnifeyeModule()
{
	LOG4CPP_NOTICE( logger, "Stopping Unifeye network service" );

	m_ioService->stop();
	m_networkThread->join();
	m_socket->close();
}


void UnifeyeModule::handleReceive( const asio::error_code err, size_t length )
{
	try
	{
		// some error checking
		if ( err && err != asio::error::message_size )
		{
			std::ostringstream msg;
			msg << "Error receiving from socket: \"" << err << "\"";
			UBITRACK_THROW( msg.str() );
		}

		if ( length >= max_receive_length )
			UBITRACK_THROW( "FIXME: received more than max_receive_length bytes." );

		// make receive data null terminated
		m_receiveData[ length ] = 0;
		
		if ( m_running )
			parsePacket( m_receiveData, length );
	}
	catch( Util::Exception& e )
	{
		LOG4CPP_WARN( logger, "Exception: " << e );
	}

	// restart receiving new packet
	m_socket->async_receive_from(
		asio::buffer( m_receiveData, max_receive_length ),
		m_senderEndpoint,
		boost::bind( &UnifeyeModule::handleReceive, this, asio::placeholders::error, asio::placeholders::bytes_transferred ) );
}


void UnifeyeModule::parsePacket( char* buffer, size_t )
{
	// get system time
	Measurement::Timestamp arrivalTime = Measurement::now();

	// parse xml
	TiXmlDocument xmlDoc;
	xmlDoc.Parse( buffer );
	if ( xmlDoc.Error() )
		UBITRACK_THROW( "XML parse error: " + std::string( xmlDoc.ErrorDesc() ) );
		
	TiXmlElement* pXmlTrackingDatagram = xmlDoc.RootElement();
	
	// get timestamp
	double uniTimestamp = getXmlDoubleValue( pXmlTrackingDatagram, "Timestamp" );
	Measurement::Timestamp timestamp = m_synchronizer.convertNativeToLocal( uniTimestamp, arrivalTime );

	// read poses
	unsigned iCos = 0;
	TiXmlElement* pXmlCos = pXmlTrackingDatagram->FirstChildElement( "COS" );
	for ( ; pXmlCos; pXmlCos = pXmlCos->NextSiblingElement( "COS" ) )
	{
		iCos++;

		if ( !hasComponent( iCos ) )
		{
			LOG4CPP_DEBUG( logger, "No component for COS " << iCos );
			continue;
		}

		// get quality
		double quality = getXmlDoubleValue( pXmlCos, "Quality" );
		if ( quality < 0.1 )
		{
			LOG4CPP_DEBUG( logger, "COS " << iCos << " has low quality " << quality << ". skipping." );
			continue;
		}

		// get SixDOF
		TiXmlElement* pXmlSixDOF = pXmlCos->FirstChildElement( "SixDOF" );
		if ( !pXmlSixDOF )
		{
			LOG4CPP_DEBUG( logger, "COS " << iCos << " has no SixDOF. skipping." );
			continue;
		}

		// parse translation
		Math::Vector< 3 > translation;
		TiXmlElement* pXmlTranslation = pXmlSixDOF->FirstChildElement( "Translation" );
		if ( !pXmlTranslation )
		{
			LOG4CPP_DEBUG( logger, "COS " << iCos << " has no translation. skipping." );
			continue;
		}

		translation( 0 ) = getXmlDoubleValue( pXmlTranslation, "x" );
		translation( 1 ) = getXmlDoubleValue( pXmlTranslation, "y" );
		translation( 2 ) = getXmlDoubleValue( pXmlTranslation, "z" );
		translation /= 1000.0;

		// parse rotation
		Math::Vector< 4 > rotation;
		TiXmlElement* pXmlRotation = pXmlSixDOF->FirstChildElement( "Rotation" );
		if ( !pXmlTranslation )
		{
			LOG4CPP_DEBUG( logger, "COS " << iCos << " has no rotation. skipping." );
			continue;
		}

		rotation( 0 ) = getXmlDoubleValue( pXmlRotation, "x" );
		rotation( 1 ) = getXmlDoubleValue( pXmlRotation, "y" );
		rotation( 2 ) = getXmlDoubleValue( pXmlRotation, "z" );
		rotation( 3 ) = getXmlDoubleValue( pXmlRotation, "w" );

		// send measurement
		if ( hasComponent( iCos ) )
		{
			Measurement::Pose m( timestamp, Math::Pose( Math::Quaternion::fromVector( rotation ), translation ) );
			LOG4CPP_DEBUG( logger, "COS " << iCos << ": Sending measurement " << m );
			getComponent( iCos )->send( m );
		}
		else
			LOG4CPP_DEBUG( logger, "No component for COS " << iCos );
	}
}


// register module at factory
UBITRACK_REGISTER_COMPONENT( ComponentFactory* const cf ) 
{
	cf->registerModule< UnifeyeModule > ( "Unifeye" );
}

} } // namespace Ubitrack::Drivers
