/*
 * Ubitrack - Library for Ubiquitous Tracking
 * Copyright 2006, Technische Universitaet Muenchen, and individual
 * contributors as indicated by the @authors tag. See the
 * copyright.txt in the distribution for a full listing of individual
 * contributors.
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA, or see the FSF site: http://www.fsf.org.
 */


/**
 * @ingroup driver_components
 * @file
 * Unifeye driver.
 * 
 * This file contains the driver component to
 * talk to the metaio Unifeye infrared tracking system.
 *
 * @author Daniel Pustka <pustka@in.tum.de>
 */
#ifndef __UnifeyeModule_h_INCLUDED__
#define __UnifeyeModule_h_INCLUDED__

#include <string>
#include <cstdlib>

// on windows, asio must be included before anything that possible includes windows.h
// don't ask why.
#include <asio.hpp>

#include <Ubitrack/Dataflow/PushSupplier.h>
#include <Ubitrack/Dataflow/Component.h>
#include <Ubitrack/Dataflow/Module.h>
#include <Ubitrack/Measurement/Measurement.h>
#include <Ubitrack/Measurement/TimestampSync.h>

namespace Ubitrack { namespace Drivers {

using namespace Dataflow;

// forward declaration
class UnifeyeComponent;

// module and 
MAKE_NODEATTRIBUTEKEY_DEFAULT( UnifeyeModuleKey, int, "Unifeye", "unifeyePort", 5000 );
MAKE_NODEATTRIBUTEKEY( UnifeyeComponentKey, int, "COS", "unifeyeId" );


/**
 * Module for ART tracker.
 * Does all the work
 */
class UnifeyeModule
	: public Module< UnifeyeModuleKey, UnifeyeComponentKey, UnifeyeModule, UnifeyeComponent >
{
public:
	/** UTQL constructor */
	UnifeyeModule( const UnifeyeModuleKey& key, boost::shared_ptr< Graph::UTQLSubgraph >, FactoryHelper* pFactory );

	/** destruktor */
	~UnifeyeModule();

protected:
	/** thread method */
    void handleReceive( const asio::error_code err, size_t length );
	
	/** parses the packet */
	void parsePacket( char* data, size_t length );

    boost::shared_ptr< asio::io_service > m_ioService;
    boost::shared_ptr< asio::ip::udp::socket > m_socket;
    boost::shared_ptr< boost::thread > m_networkThread;

    // Recive data. Do not touch from outside of async network thread
    enum { max_receive_length = 10240, receive_buffer_size = 10242 };
    char m_receiveData[ receive_buffer_size ];
    asio::ip::udp::endpoint m_senderEndpoint;

	Measurement::TimestampSync m_synchronizer;
};


/**
 * Component for ART tracker.
 * Does nothing but provide a push port
 
 * @TODO: make this two separate components for 6d/3dlist
 */
class UnifeyeComponent
	: public UnifeyeModule::Component
{
public:
	/** constructor */
	UnifeyeComponent( const std::string& name, boost::shared_ptr< Graph::UTQLSubgraph > subgraph, 
		const UnifeyeComponentKey& componentKey, UnifeyeModule* pModule )
		: UnifeyeModule::Component( name, componentKey, pModule )
		, m_port( "Output", *this )
	{}

	/** returns the port for usage by the module */
	void send( const Measurement::Pose& pose )
	{ m_port.send( pose ); }

protected:
	// the port is the only member
	PushSupplier< Ubitrack::Measurement::Pose > m_port;
};

} } // namespace Ubitrack::Drivers

#endif
