/*
 * Ubitrack - Library for Ubiquitous Tracking
 * Copyright 2006, Technische Universitaet Muenchen, and individual
 * contributors as indicated by the @authors tag. See the 
 * copyright.txt in the distribution for a full listing of individual
 * contributors.
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA, or see the FSF site: http://www.fsf.org.
 */


/**
 * @ingroup driver_components
 * @file
 * Wiimote driver
 * This file contains the driver component to
 * get sensor and button data from a Nintendo Wiimote.
 * On Linux, it uses the libwiimote from the cwiid project.
 *
 * The received data is sent via a push interface.
 *
 * @author Florian Echtler <echtler@in.tum.de>
 */

#include "WiimoteDriver.h"

static log4cpp::Category& logger( log4cpp::Category::getInstance( "Drivers.Wiimote" ) );

namespace Ubitrack { namespace Drivers {


using namespace Dataflow;


//cwiid_mesg_callback_t g_wiimote_callback;
std::map< cwiid_wiimote_t*, WiimoteDriver* > g_driverList;

void g_wiimote_callback( cwiid_wiimote_t* wm, int mesg_count, union cwiid_mesg mesg[], struct timespec* )
{
	std::map< cwiid_wiimote_t*, WiimoteDriver* >::iterator target = g_driverList.find(wm);
	if (target != g_driverList.end()) target->second->handle( mesg_count, mesg );
}


/** constructor */
WiimoteDriver::WiimoteDriver( const std::string& sName, boost::shared_ptr< Graph::UTQLSubgraph > )
	: Component( sName )
	, m_buttonOut( "ButtonOutput", *this )
{
	/** bluetooth device address */
	bdaddr_t bdaddr;

	/** find any available Wiimote */
	//if ( (argc != 2) || str2ba( argv[1], &bdaddr ) ) {
	LOG4CPP_INFO( logger, "Put Wiimote in discoverable mode now." );
	bdaddr = *BDADDR_ANY;
	//}

	/** connect to the Wiimote */
	if (!(wiimote = cwiid_connect( &bdaddr, CWIID_FLAG_MESG_IFC )))
		UBITRACK_THROW( "Unable to connect to Wiimote." );

	/** select only button reports for now */
	if (cwiid_command( wiimote, CWIID_CMD_RPT_MODE, CWIID_RPT_BTN ))
		UBITRACK_THROW( "Error setting report mode." );

	/** select only button reports for now */
	if (cwiid_set_mesg_callback( wiimote, g_wiimote_callback ))
		UBITRACK_THROW( "Error setting callback." );

	g_driverList[wiimote] = this;
}

/** destructor */
WiimoteDriver::~WiimoteDriver()
{
	cwiid_disconnect( wiimote );
}

/** wiimote callback */
void WiimoteDriver::handle( int mesg_count, union cwiid_mesg mesg[] )
{
	for (int i = 0; i < mesg_count; i++) {

		switch (mesg[i].type) {

			case CWIID_MESG_BTN:
				Measurement::Button btn( Measurement::now(), mesg[i].btn_mesg.buttons );
				LOG4CPP_DEBUG( logger, "button bitmask: " << *btn );
				m_buttonOut.send( btn );
				break;
		}
	}
}


UBITRACK_REGISTER_COMPONENT( ComponentFactory* const cf ) {
	cf->registerComponent< WiimoteDriver > ( "WiimoteDriver" );
}


} } // namespace Ubitrack::Drivers

