/*
 * Ubitrack - Library for Ubiquitous Tracking
 * Copyright 2006, Technische Universitaet Muenchen, and individual
 * contributors as indicated by the @authors tag. See the
 * copyright.txt in the distribution for a full listing of individual
 * contributors.
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA, or see the FSF site: http://www.fsf.org.
 */


/**
 * @ingroup driver_components
 * @file
 * Xsens driver
 * This file contains the driver component to
 * get rotation data from an Xsens MT9-B.
 * On Linux, it uses the MotionTracker library provided by Xsens.
 *
 * The received data is sent via a push interface.
 *
 * @author Florian Echtler <echtler@in.tum.de>
 */

#include <string>
#include <cstdlib>
#include <log4cpp/Category.hh>

#include <Ubitrack/Dataflow/PushSupplier.h>
#include <Ubitrack/Dataflow/ComponentFactory.h>
#include <Ubitrack/Dataflow/Component.h>
#include <Ubitrack/Measurement/Measurement.h>
#include <Ubitrack/Measurement/TimestampSync.h>
#include <Ubitrack/Util/Exception.h>
#include <Ubitrack/Util/OS.h>

#ifdef _WIN32
	#include <Ubitrack/Util/CleanWindows.h>
	#ifndef _WIN32_WINNT
		#define _WIN32_WINNT WINVER
	#endif
	#include <objbase.h>
	#include "IMTObj.h"
	#include "IMTObj_i.c"
	#define MT_NEWDATA       1
	#define MT_NODATA        2
	#define MT_LOGQUATERNION 0
#else
	#include "MotionTracker.h"
#endif

static log4cpp::Category& logger( log4cpp::Category::getInstance( "Drivers.XSens" ) );

namespace Ubitrack { namespace Drivers {

using namespace Dataflow;

/**
 * Module for XSens tracker, should work on both linux and windows.
 * Requires the XSens ActiveX plugin to be installed on Windows.
 */
class XsensLegacy
	: public Component
{
public:

	/** constructor */
	XsensLegacy( const std::string& sName, boost::shared_ptr< Graph::UTQLSubgraph > subgraph )
		: Component( sName )
		, m_port( "Output", *this )
		, m_rotationVelocityPort( "RotationVelocity", *this )
		, m_keepRunning( 1 )
		, m_nCorInterval( 1 )
		, m_fGain( 1.0 )
		, m_fRho( 1.0 ) // 0.0 disables magnetometers, default is 1.0
		, m_bUseCalibratedData( false ) // if false, use differentiation to compute RotationVelocity
	{
		// create the tracker object
		#ifdef _WIN32
			// read serial port number from configuration
			m_portNum = 4;
			subgraph->m_DataflowAttributes.getAttributeData( "port", m_portNum );

			HRESULT hRes;
			hRes = CoInitializeEx( NULL, COINIT_MULTITHREADED );
			if ( FAILED( hRes ) )
				UBITRACK_THROW( "CoInitialize() failed" );

			LOG4CPP_DEBUG( logger, "Creating MotionTracker COM Object"  );
			hRes = CoCreateInstance( CLSID_MotionTracker, NULL, CLSCTX_SERVER, IID_IMotionTracker, (void**) &m_pTracker );
			if ( FAILED( hRes ) )
				UBITRACK_THROW( "Could not create ActiveX MT object" );

			LOG4CPP_INFO( logger, "Connecting to XSens at port COM" << m_portNum );
			m_pTracker->MT_SetCOMPort( m_portNum );
		#else
			// read device name from configuration
			m_portName = subgraph->m_DataflowAttributes.getAttributeString( "port" );
			if (m_portName.empty())
				UBITRACK_THROW( "No 'name' attribute in XsensLegacy component configuration." );

			m_pTracker = create();

			LOG4CPP_INFO( logger, "Connecting to XSens at port " << m_portName );
			m_pTracker->MT_SetCOMPort_DeviceName( const_cast< char* >( m_portName.c_str() ) );
		#endif

		short bLogCalibratedData = m_bUseCalibratedData ? 1 : 0;
		m_pTracker->MT_SetCalibratedOutput( bLogCalibratedData );
		
		subgraph->m_DataflowAttributes.getAttributeData( "rho", m_fRho );
		m_bMagnetometerOn = m_fRho != 0.0;
		m_pTracker->MT_SetFilterSettings( m_fGain, m_nCorInterval, m_fRho );

		short nTimeout = 3;
		m_pTracker->MT_SetTimeout( nTimeout );

		short nMode = MT_LOGQUATERNION;
		m_pTracker->MT_SetOutputMode( nMode );

		m_pTracker->MT_SetTimeStampOutput( 1 );

		short nDoAMD = 0;
		m_pTracker->MT_SetDoAMD( nDoAMD );

		// start the library thread
		LOG4CPP_DEBUG( logger, "MT_StartProcess" );
		#ifdef _WIN32
			hRes = m_pTracker->MT_StartProcess();
			if ( FAILED( hRes ) )
			{	LOG4CPP_ERROR( logger, "MT_StartProcess returned error " << hRes ); }
		#else
			m_pTracker->MT_StartProcess();
		#endif

		// start our own polling thread
		m_thread.reset( new boost::thread( boost::bind( &XsensLegacy::threadHandler, this ) ) );
	}

	/** destructor */
	~XsensLegacy()
	{
		m_keepRunning = 0;
		m_thread->join();

		LOG4CPP_DEBUG( logger, "MT_StopProcess" );
		m_pTracker->MT_StopProcess();
		Util::sleep( 1000 );

		#ifdef _WIN32
			m_pTracker->Release();
			m_pTracker = 0;
			CoUninitialize();
		#else
			destroy( m_pTracker );
		#endif
	}

	/** thread method */
	void threadHandler()
	{
		LOG4CPP_DEBUG( logger, "XSens thread started" );

		Measurement::TimestampSync syncer( 1.0 ); // timestamps are in seconds
		
		short nNew = 0;
		Measurement::Rotation lastRotation;
		while ( m_keepRunning )
		{
			// turn magnetometers off when the RotationVelocity port is used
			if ( m_running )
			{
				Component::MutexType::scoped_lock l( m_componentMutex );
				if ( m_bMagnetometerOn && m_rotationVelocityPort.isConnected() )
				{
					LOG4CPP_INFO( logger, "RotationVelocity used -> turning off magnetometers" );
					m_pTracker->MT_SetFilterSettings( m_fGain, m_nCorInterval, 0 );
					m_bMagnetometerOn = false;
				}
				else if ( !m_bMagnetometerOn && m_fRho > 0 && !m_rotationVelocityPort.isConnected() )
				{
					LOG4CPP_INFO( logger, "RotationVelocity not used -> turning on magnetometers" );
					m_pTracker->MT_SetFilterSettings( m_fGain, m_nCorInterval, m_fRho );
					m_bMagnetometerOn = true;
				}
			}
		
			// get orientation and/or rotation velocity (different for windows and linux)
			Measurement::Rotation rotation;
			Measurement::RotationVelocity rotationVelocity;
			
			#ifdef _WIN32
				// Variable of VARIANT type to retrieve from VARIANT output from activeX control
				VARIANT OrientationBuffer;
				m_pTracker->MT_GetOrientationData( &nNew, &OrientationBuffer, 1 );
				LOG4CPP_TRACE( logger, "MT_GetOrientationData returned " << nNew );

				if ( nNew == MT_NEWDATA && OrientationBuffer.vt != VT_EMPTY )
				{
					// Retrieve pointer to array data
					float* pData;
					HRESULT hr = SafeArrayAccessData( OrientationBuffer.parray, reinterpret_cast< void** >( &pData ) );

					if ( SUCCEEDED( hr ) )
					{
						// send orientation data
						rotation = Measurement::Rotation( syncer.convertNativeToLocal( pData[0] ) - 9000000, 
							Math::Quaternion( pData[2], pData[3], pData[4], pData[1] ) );
						
						// Variant must be cleared. This also destroys the SafeArray
						SafeArrayUnaccessData( OrientationBuffer.parray );	// Invalidate pointer
						VariantClear( &OrientationBuffer );
					}
				}	
				
				if ( m_bUseCalibratedData )
				{
					// also retrive calibrated data from XSens and output pure gyroscope rotation velocity
					m_pTracker->MT_GetCalibratedData( &nNew, &OrientationBuffer, 1 );
					LOG4CPP_TRACE( logger, "MT_GetCalibratedData returned " << nNew );

					if ( nNew == MT_NEWDATA && OrientationBuffer.vt != VT_EMPTY )
					{
						// Retrieve pointer to array data
						float* pData;
						HRESULT hr = SafeArrayAccessData( OrientationBuffer.parray, reinterpret_cast< void** >( &pData ) );

						if ( SUCCEEDED( hr ) )
						{
							// send rotation velocity data
							rotationVelocity = Measurement::RotationVelocity( syncer.convertNativeToLocal( pData[0] ) - 9000000, 
								Math::RotationVelocity( pData[4], pData[5], pData[6] ) );
							
							// Variant must be cleared. This also destroys the SafeArray
							SafeArrayUnaccessData( OrientationBuffer.parray );	// Invalidate pointer
							VariantClear( &OrientationBuffer );
						}
					}
				}
			#else
				float fData[11] = {0};
				
				m_pTracker->MT_GetOrientationData( &nNew, fData, 0 );
				LOG4CPP_TRACE( logger, "MT_GetOrientationData returned " << nNew );

				if ( nNew == MT_NEWDATA )
					// construct and send quaternion
					rotation = Measurement::Rotation( syncer.convertNativeToLocal( fData[0] ) - 9000000, 
						Math::Quaternion( fData[2], fData[3], fData[4], fData[1] ) );
					
				if ( m_bUseCalibratedData )
				{
					// also retrieve rotation velocity
					m_pTracker->MT_GetCalibratedData( &nNew, fData, 0 );
					LOG4CPP_TRACE( logger, "MT_GetCalibratedData returned " << nNew );
					
					if ( nNew == MT_NEWDATA )
						rotationVelocity = Measurement::RotationVelocity( syncer.convertNativeToLocal( fData[0] ) - 9000000, 
							Math::RotationVelocity( fData[4], fData[5], fData[6] ) );
				}
			#endif

			if ( m_running )
			{ 
				if ( rotation )
				{
					LOG4CPP_DEBUG( logger, "XSens measurement: " << *rotation );
					m_port.send( rotation );
					
					// compute rotation velocity by differentiation
					if ( !rotationVelocity && lastRotation )
						rotationVelocity = Measurement::RotationVelocity( rotation.time(),
							Math::RotationVelocity( *lastRotation, *rotation, 
								double( static_cast< long long int >( rotation.time() - lastRotation.time() ) ) * 1e-9 ) );
					
					lastRotation = rotation;
				}
			
				if ( rotationVelocity )
				{
					LOG4CPP_DEBUG( logger, "XSens rotation velocity measurement: " << *rotationVelocity );
					m_rotationVelocityPort.send( rotationVelocity );
				}
			}
				
			// wait 2ms until next poll
			Util::sleep( 2 );
		}

		LOG4CPP_DEBUG( logger, "XSens thread stopped" );
	}
protected:

	boost::shared_ptr< boost::thread > m_thread;
	PushSupplier< Ubitrack::Measurement::Rotation > m_port;
	PushSupplier< Ubitrack::Measurement::RotationVelocity > m_rotationVelocityPort;

	#ifdef _WIN32
		int m_portNum;
		IMotionTracker* m_pTracker;
	#else
		std::string m_portName;
		MotionTracker* m_pTracker;
	#endif

	int m_keepRunning;
	
	// configuration
	short m_nCorInterval;
	float m_fGain;
	float m_fRho;
	bool m_bUseCalibratedData;

	// internal state
	bool m_bMagnetometerOn;
};


UBITRACK_REGISTER_COMPONENT( ComponentFactory* const cf ) {
	cf->registerComponent< XsensLegacy > ( "XSens" );
}


} } // namespace Ubitrack::Drivers
