/*
 * Ubitrack - Library for Ubiquitous Tracking
 * Copyright 2006, Technische Universitaet Muenchen, and individual
 * contributors as indicated by the @authors tag. See the 
 * copyright.txt in the distribution for a full listing of individual
 * contributors.
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA, or see the FSF site: http://www.fsf.org.
 */


/**
 * @ingroup driver_components
 * @file
 * Xsens Raw Driver
 * This file contains the driver component to get raw acceleration,
 * magnetic field and rotation data from an Xsens MT9-B.
 *
 * The received data is sent via a push interface.
 *
 * @author Florian Echtler <echtler@in.tum.de>
 */

#include "XsensRaw.h"
#include "XsensMessage.h"
#include <Ubitrack/Util/Exception.h>
#include <sstream>
#include <algorithm>
#include <iomanip>


// M_PI is not part of ISO C/C++
static const double g_pi = 3.14159265358979323846;


static log4cpp::Category& logger( log4cpp::Category::getInstance( "Drivers.XsensRaw" ) );

namespace Ubitrack { namespace Drivers {


using namespace Dataflow;


/** constructor */
XsensRaw::XsensRaw( const std::string& sName, boost::shared_ptr< Graph::UTQLSubgraph > subgraph )
	: Component( sName )
	, m_magOutput( "MagneticField", *this )
	, m_accOutput( "Acceleration", *this )
	, m_gyrOutput( "RotationVelocity", *this )
	, m_keepRunning(1)
{
	#ifdef _WIN32
		std::string portName("COM1");
	#else // *nix
		std::string portName("/dev/ttyUSB0");
	#endif
	int baudRate = 115200;

	if (subgraph->m_DataflowConfiguration.getXML() != NULL) {
		const TiXmlNode* port = subgraph->m_DataflowConfiguration.getXML()->FirstChild("Port");
		if ((port != NULL) && (port->ToElement() != NULL)) {
			char* pName = (char*)(port->ToElement()->Attribute("name"));
			if (pName != NULL) portName = pName;
			char* pRate = (char*)(port->ToElement()->Attribute("rate"));
			if (pRate != NULL) baudRate = atoi(pRate);
		}
	}

	LOG4CPP_INFO( logger, "accessing " << portName << " @ " << baudRate << " BPS" );

	m_port = new Util::SerialPort( portName, baudRate );
	m_port->open(0,1); // blocking mode

	m_thread.reset( new boost::thread( boost::bind( &XsensRaw::threadHandler, this ) ) );
}


/** destructor */
XsensRaw::~XsensRaw()
{
	m_keepRunning = 0;
	m_thread->join();

	m_port->close();
	delete m_port;
}


void XsensRaw::threadHandler()
{
	int offset = 0;
	XSensMessage buffer[5];

  while (1) {

    int size = m_port->read( ((unsigned char*)&buffer)+offset, sizeof(buffer)-offset );
		//if (size < 0) { printf("error: read: %s\n",strerror(errno)); continue; }

		offset += size;

		while (offset >= (int)XSMSIZE) {

			// create a copy of the oldest message, prepare to delete it from the buffer
			XSensMessage tmp = buffer[0];
			int skipsize = XSMSIZE;		

			// calculate the checksum and byte-swap all unsigned short values
			checksum(&tmp);
			le2be(&tmp);

			// if the checksum is incorrect, this most likely means that we lost a byte
			// somewhere on the serial line, so we need to skip one byte less.
			if (tmp.checksum != 0) skipsize--;
			offset -= skipsize;
			if (offset) memmove( &buffer, ((char*)&buffer)+skipsize, offset );

			// sanity checks for the packet
			if (tmp.preamble != 0xFA) continue;
			if (tmp.address  != 0xFF) continue;
			if (tmp.msgid    != 0x32) continue;
			if (tmp.length   !=   22) continue;
			if (tmp.checksum !=    0) continue;

			#define SHIFT 32768.0

			double x = (tmp.mag[0]-SHIFT)/SHIFT;
			double y = (tmp.mag[1]-SHIFT)/SHIFT;
			double z = (tmp.mag[2]-SHIFT)/SHIFT;
			//std::cout << tmp.sample << " " << x << " " << y << " " << z << " ";
			//std::cout << " " << sqrt(x*x+y*y+z*z) << std::endl;

			/* std::cout << "acc: " << tmp.acc[0] << " " << tmp.acc[1] << " " << tmp.acc[2] << " " << std::endl;
			std::cout << "gyr: " << tmp.gyr[0] << " " << tmp.gyr[1] << " " << tmp.gyr[2] << " " << std::endl;
			std::cout << "mag: " << tmp.mag[0] << " " << tmp.mag[1] << " " << tmp.mag[2] << " " << std::endl;
			std::cout << "temp: " << tmp.temp << std::endl << std::endl; */

			Measurement::Vector3D mag( Measurement::now(), Math::Vector<3>( x, y, z ) );
			m_magOutput.send( mag );
		}
  }
}


UBITRACK_REGISTER_COMPONENT( ComponentFactory* const cf ) {
	cf->registerComponent< XsensRaw > ( "XsensRaw" );
}

} } // namespace Ubitrack::Drivers

