#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>

#include <sys/fcntl.h>
#include <sys/file.h>

#include <termio.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <math.h>

#include <string.h>
#include <errno.h>

#include <iostream>


#pragma pack(1)
struct XSensMessage {
	union {
		struct {
			unsigned char preamble;
			unsigned char address;
			unsigned char msgid;
			unsigned char length;
			union {
				struct {
					unsigned short sample;
					unsigned short acc[3];
					unsigned short gyr[3];
					unsigned short mag[3];
					unsigned short temp;
				};
				unsigned short values[11];
			};
			unsigned char checksum;
		};
		unsigned char data[27];
	};
};

#define XSMSIZE sizeof(XSensMessage)

// byte-swap an unsigned short
unsigned short le2be( unsigned short tmp ) { return ((tmp&0xFF00)>>8) + ((tmp&0x00FF)<<8); }

// byte-swap all measurement values in a message
void le2be( XSensMessage* xsm ) {
	for (int i = 0; i < 11; i++)
		xsm->values[i] = le2be(xsm->values[i]);
}

// New checksum is the single-byte sum over the packet, excluding the preamble
// but including the original checksum. Result should be 0, else discard packet.
void checksum( XSensMessage* xsm ) {
	unsigned char tmp = 0;
	for ( unsigned int i = 1; i < XSMSIZE; i++ )
		tmp += xsm->data[i];
	xsm->checksum = tmp;
}


int main(int argc, char *argv[]) {

  struct termios term;
  int port,size;
	int offset;
  long baud = B115200;
  XSensMessage buffer[5];

  // open the serial port
  port = open("/dev/ttyUSB0", O_RDWR | O_NOCTTY);
  if (port < 0) { perror("open"); exit(1); }
  
  if (tcgetattr(port,&term) < 0) { perror("tcgetattr"); exit(1); }
  
  // set baud rate, 8 bits, no flow control, receiver on
  term.c_cflag = baud | CS8 | CLOCAL | CREAD;
  term.c_oflag = 0;
  
  // ignore breaks, CR, parity 
  term.c_iflag = IGNBRK | IGNCR | IGNPAR;
  term.c_lflag = 0;

  // set read count and timer
  term.c_cc[VTIME] = 5;
  term.c_cc[VMIN]  = 5;
  
  // set output and input baud rates
  cfsetospeed(&term,baud);
  cfsetispeed(&term,baud);
  
  if (tcsetattr(port,TCSANOW,&term) < 0) { perror("tcsetattr"); exit(1); }
 
 	offset = 0;

  while (1) {

    size = read( port, ((char*)&buffer)+offset, sizeof(buffer)-offset );
		if (size < 0) { printf("error: read: %s\n",strerror(errno)); continue; }

		offset += size;

		while (offset >= (int)XSMSIZE) {

			// create a copy of the oldest message, prepare to delete it from the buffer
			XSensMessage tmp = buffer[0];
			int skipsize = XSMSIZE;		

			// calculate the checksum and byte-swap all unsigned short values
			checksum(&tmp);
			le2be(&tmp);

			// if the checksum is incorrect, this most likely means that we lost a byte
			// somewhere on the serial line, so we need to skip one byte less.
			if (tmp.checksum != 0) skipsize--;
			offset -= skipsize;
			if (offset) memmove( &buffer, ((char*)&buffer)+skipsize, offset );

			// sanity checks for the packet
			if (tmp.preamble != 0xFA) continue;
			if (tmp.address  != 0xFF) continue;
			if (tmp.msgid    != 0x32) continue;
			if (tmp.length   !=   22) continue;
			if (tmp.checksum !=    0) continue;

			double x = tmp.mag[0];
			double y = tmp.mag[1];
			double z = tmp.mag[2];
			std::cout << tmp.sample << " " << x << " " << y << " " << z << " ";
			std::cout << " " << sqrt(x*x+y*y+z*z) << std::endl;

			/* std::cout << "acc: " << tmp.acc[0] << " " << tmp.acc[1] << " " << tmp.acc[2] << " " << std::endl;
			std::cout << "gyr: " << tmp.gyr[0] << " " << tmp.gyr[1] << " " << tmp.gyr[2] << " " << std::endl;
			std::cout << "mag: " << tmp.mag[0] << " " << tmp.mag[1] << " " << tmp.mag[2] << " " << std::endl;
			std::cout << "temp: " << tmp.temp << std::endl << std::endl; */
		}

  }

  return 0;
}
 
