/*
 * Ubitrack - Library for Ubiquitous Tracking
 * Copyright 2006, Technische Universitaet Muenchen, and individual
 * contributors as indicated by the @authors tag. See the 
 * copyright.txt in the distribution for a full listing of individual
 * contributors.
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA, or see the FSF site: http://www.fsf.org.
 */

/**
 * @ingroup driver_components
 * helper class to read/decode Z800 rotation sensors
 *
 * @author Florian Echtler <echtler@in.tum.de>
 */

#include "RotationSensors.h"

using namespace Ubitrack::Math;
namespace ublas = boost::numeric::ublas;

#ifndef __linux

RotationSensors::RotationSensors() { } 
RotationSensors::~RotationSensors() { } 
Ubitrack::Measurement::Timestamp RotationSensors::receive( ) { return 0; }
void RotationSensors::decode( unsigned char* _buf ) { }

#else

#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <linux/ioctl.h>
#include <sys/ioctl.h>

// this should be part of the kernel driver
#define Z800_IOC_MAGIC 	'Z'
#define Z800_IOCTL_HT  _IOR(Z800_IOC_MAGIC, 31, char [16])
#define Z800_IOCTL_GET _IOWR(Z800_IOC_MAGIC, 32, char [8])


RotationSensors::RotationSensors()
	: m_rotSpeed()
	, m_magRot()
	, m_accel()
{
	m_fd = open( "/dev/tracker:0", O_RDWR );
	m_count = 0;
}


RotationSensors::~RotationSensors()
{
	close( m_fd );
}


void RotationSensors::decode(unsigned char* _buf )
{
	// acceleration vector
	m_accRaw[1] = ((int)(_buf[0]&0xf)<<6) | (_buf[1]>>2);	
	m_accRaw[0] = ((int)(_buf[1]&0x3)<<8) | (_buf[2]);		
	m_accRaw[2] = ((int)_buf[3]<<2) | (_buf[4]>>6);			

	// magnetic field vector
	m_magRaw[1] = ((int)(_buf[8]&0xf)<<6) | (_buf[9]>>2);
	m_magRaw[2] = ((int)(_buf[9]&0x3)<<8) | (_buf[10]);
	m_magRaw[0] = ((int)_buf[11]<<2) | (_buf[12]>>6);

	// gyro 1 vector
	m_rotRaw[1] = ((int)(_buf[4]&0x3f)<<4) | (_buf[5]>>4);		
	m_rotRaw[2] = ((int)(_buf[5]&0xf)<<6) | (_buf[6]>>2);		
	m_rotRaw[0] = ((int)(_buf[6]&0x3)<<8) | (_buf[7]);			

	// gyro 2 vector
	m_rotRaw[4] = ((int)(_buf[12]&0x3f)<<4) | (_buf[13]>>4);
	m_rotRaw[5] = ((int)(_buf[13]&0xf)<<6) | (_buf[14]>>2);
	m_rotRaw[3] = ((int)(_buf[14]&0x3)<<8) | (_buf[15]);

	// convert 2's complement numbers (if they're negative) 
	if(m_magRaw[0]&0x200) m_magRaw[0] = -((m_magRaw[0]^0x3ff)+1);
	if(m_magRaw[1]&0x200) m_magRaw[1] = -((m_magRaw[1]^0x3ff)+1);
	if(m_magRaw[2]&0x200) m_magRaw[2] = -((m_magRaw[2]^0x3ff)+1);

	// shift center by 512
	m_accRaw[0] -= 512;
	m_accRaw[1] -= 512;
	m_accRaw[2] -= 512;

	m_rotRaw[0] -= 512;
	m_rotRaw[1] -= 512;
	m_rotRaw[2] -= 512; m_rotRaw[2] = -m_rotRaw[2];

	m_rotRaw[3] -= 512;
	m_rotRaw[4] -= 512;
	m_rotRaw[5] -= 512; m_rotRaw[5] = -m_rotRaw[5];
}


Ubitrack::Measurement::Timestamp RotationSensors::receive( )
{
	unsigned char buffer[16];

	ioctl( m_fd, Z800_IOCTL_HT, buffer, sizeof(buffer) );
	Ubitrack::Measurement::Timestamp time = Ubitrack::Measurement::now();

	decode( buffer );

	// wait 10 samples for sensors to settle down,
	// store the raw values as sensor offsets and subtract them afterwards
	if (m_count++ <= 10) {
		for (int i = 0; i < 3; i++) m_magOffs[i] = m_magRaw[i];
		for (int i = 0; i < 6; i++) m_rotOffs[i] = m_rotRaw[i];
		return 0;
	} else
		for (int i = 0; i < 6; i++) m_rotRaw[i] -= m_rotOffs[i];

	// calculate rotation between initial and current magnetic field vector
	//std::cout << "magOffs: " << (double)m_magOffs[0] << " " << (double)m_magOffs[1] << " " << (double)m_magOffs[2] << std::endl;
	//std::cout << "magRaw: " << (double)m_magRaw[0] << " " << (double)m_magRaw[1] << " " << (double)m_magRaw[2] << std::endl;
	Vector<3> orig( (double)m_magOffs[0], (double)m_magOffs[1], (double)m_magOffs[2] ); orig = orig / ublas::norm_2( orig );
	Vector<3> curr( (double)m_magRaw [0], (double)m_magRaw [1], (double)m_magRaw [2] ); curr = curr / ublas::norm_2( curr );
	Vector<3> tmp = cross_prod( orig, curr );
	m_magRot = Quaternion( tmp[0], tmp[1], tmp[2], 1 + ublas::inner_prod( orig, curr ) );
	m_magRot.normalize();

	// calculate current acceleration from raw values: range is presumed to be [-512,512] for [-2g,2g]
	m_accel = Vector<3>( (double)m_accRaw[0]/256.0, (double)m_accRaw[1]/256.0, (double)m_accRaw[2]/256.0 );

	// calculate rotation speed (unit unknown, but probably degrees per second?)
	m_rotSpeed = RotationVelocity( Vector<3>( (double)m_rotRaw[0], (double)m_rotRaw[1], (double)m_rotRaw[2] ) ); 

	return time;
}

#endif

RotationVelocity& RotationSensors::getRotationSpeed()    { return m_rotSpeed; }
Quaternion&       RotationSensors::getMagneticRotation() { return m_magRot;   }
Vector<3>&        RotationSensors::getAcceleration()     { return m_accel;    }

