/*
 * Ubitrack - Library for Ubiquitous Tracking
 * Copyright 2006, Technische Universitaet Muenchen, and individual
 * contributors as indicated by the @authors tag. See the
 * copyright.txt in the distribution for a full listing of individual
 * contributors.
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA, or see the FSF site: http://www.fsf.org.
 */


/**
 * @ingroup driver_components
 * @file
 * Z800 driver
 * This file contains the driver component to
 * get rotation data from a Z800 HMD and control the HMD functions.
 *
 * The received data is sent via a push interface.
 *
 * @author Florian Echtler <echtler@in.tum.de>
 */

#include <string>
#include <cstdlib>
#include <log4cpp/Category.hh>

#include <Ubitrack/Dataflow/PushSupplier.h>
#include <Ubitrack/Dataflow/ComponentFactory.h>
#include <Ubitrack/Dataflow/Component.h>
#include <Ubitrack/Measurement/Measurement.h>
#include <Ubitrack/Measurement/TimestampSync.h>
#include <Ubitrack/Util/Exception.h>
#include <Ubitrack/Util/OS.h>

#include "HMDControl.h"
#include "RotationSensors.h"

static log4cpp::Category& logger( log4cpp::Category::getInstance( "Drivers.Z800" ) );

namespace Ubitrack { namespace Drivers {

using namespace Dataflow;

/**
 * Module for XSens tracker, should work on both linux and windows.
 * Requires the XSens ActiveX plugin to be installed on Windows.
 */
class Z800
	: public Component
{
public:

	/** constructor */
	Z800( const std::string& sName, boost::shared_ptr< Graph::UTQLSubgraph > subgraph )
		: Component( sName )
		, m_rvPort( "RotVelOutput", *this )
		, m_rotPort( "RotOutput", *this )
		, m_hmd()
		, m_sensors()
		, m_keepRunning(1)
	{
		if (subgraph->m_DataflowConfiguration.getXML() != NULL) {
			const TiXmlNode* node = subgraph->m_DataflowConfiguration.getXML()->FirstChild( "Mode3D" );
			if ((node != NULL) && (node->ToElement() != NULL)) {
				m_hmd.wakeup();
				m_hmd.mode3D();
			}	
		}

		int ver = m_hmd.getVer();
		LOG4CPP_INFO( logger, "Z800 Firmware Version: " << (ver >> 8) << "." << (ver & 0xFF) );

		// start our own polling thread
		m_thread.reset( new boost::thread( boost::bind( &Z800::threadHandler, this ) ) );
	}

	/** destructor */
	~Z800()
	{
		m_keepRunning = 0;
		m_thread->join();
		m_hmd.shdown();
	}

	/** thread method */
	void threadHandler()
	{
		LOG4CPP_DEBUG( logger, "Z800 thread started" );

		//Measurement::TimestampSync syncer( 1.0 ); // timestamps are in seconds
		while (m_keepRunning)
		{
			Measurement::Timestamp time = m_sensors.receive();
			if (!time) continue;
			Measurement::Rotation rot( time, m_sensors.getMagneticRotation() );
			m_rotPort.send( rot );
			// construct and send rotation velocity
			/*Measurement::RotationVelocity vrot( syncer.convertNativeToLocal( 0 ), Math::RotationVelocity( 0, 0, 0 ) );
			LOG4CPP_DEBUG( logger, "Z800 measurement: " << *vrot );
			m_port.send( vrot );*/
		}

		LOG4CPP_DEBUG( logger, "Z800thread stopped" );
	}

protected:

	boost::shared_ptr< boost::thread > m_thread;
	PushSupplier< Measurement::RotationVelocity > m_rvPort;
	PushSupplier< Measurement::Rotation > m_rotPort;
	//PushSupplier< Measurement::Acceleration > m_accPort;
	HMDControl m_hmd;
	RotationSensors m_sensors;

	int m_keepRunning;
};


UBITRACK_REGISTER_COMPONENT( ComponentFactory* const cf ) {
	cf->registerComponent< Z800 > ( "Z800" );
}


} } // namespace Ubitrack::Drivers

