#include <cstdlib>
#include <iostream>
#include <string>
#include <sstream>
#include <iomanip>

#define _USE_MATH_DEFINES
#include <math.h>

#include "stdio.h"
#include "conio.h"

// Normally set by windows.h, but on the other hand, boost/asio.h does not work once windows.h
// is included. Don't ask why
#define _WIN32_WINNT 0x0501
// On windows, asio must be included before anything that possible includes windows.h.
// Don't ask why.
#include <boost/asio.hpp>
#include <boost/bind.hpp>
#include <boost/date_time/local_time/local_time.hpp>
// WARNING: all boost/serialization headers should be
//          included AFTER all boost/archive headers
#include <boost/archive/text_oarchive.hpp>
#include <boost/serialization/vector.hpp>
#include <boost/serialization/utility.hpp>


#undef _WIN32_WINNT
#undef WINVER

#include "windows.h"
#include "CmmSpecific.h"

using boost::asio::ip::udp;

///////////////////////////////////////////////////////////////////////////////
BOOL FalseSpecific()
{
	return 0;
}
///////////////////////////////////////////////////////////////////////////////


class Position {
public:
	double x;
	double y;
	double z;

	Position(){};
	Position( double x, double y, double z )
	{
		this->x = x;
		this->y = y;
		this->z = z;
	};

	template< class Archive > 
	void serialize( Archive& ar, const unsigned int version )
	{
		ar & x;
		ar & y;
		ar & z;
	}
};

class Quaternion {
public:
	double x;
	double y;
	double z;
	double w;

	Quaternion(){};
	Quaternion( double x, double y, double z, double w )
	{
		this->x = x;
		this->y = y;
		this->z = z;
		this->w = w;
	}

	template< class Archive > 
	void serialize( Archive& ar, const unsigned int version )
	{
		ar & x;
		ar & y;
		ar & z;
		ar & w;
	}

	Quaternion times( Quaternion & b ) {
		double w = this->w * b.w - this->x * b.x - this->y * b.y - this->z * b.z;
		double x = this->w * b.x + this->x * b.w + this->y * b.z - this->z * b.y;
		double y = this->w * b.y - this->x * b.z + this->y * b.w + this->z * b.x;
		double z = this->w * b.z + this->x * b.y - this->y * b.x + this->z * b.w;
		return Quaternion( x, y, z, w );
	}
};

class Pose {
public:
	Position position;
	Quaternion orientation;

	Pose(){};
	Pose( Position position, Quaternion orientation )
	{
		this->position = position;
		this->orientation = orientation;
	};

	/// serialize Pose object
	template< class Archive > 
	void serialize( Archive& ar, const unsigned int version )
	{
		ar & orientation;
		ar & position;
	}
};

class Measurement {
public:
	__int64 timestamp;
	Pose pose;

	Measurement(){};
	Measurement( __int64 timestamp, Pose pose )
	{
		this->timestamp = timestamp;
		this->pose = pose;
	}

	template< class Archive > 
	void serialize( Archive& ar, const unsigned int version )
	{
		ar & timestamp;
		ar & pose;
	}
};


Quaternion euler2Quat( double heading, double attitude, double bank )
{
	// ZXZ Euler angles
    Quaternion bq1( 0, 0, sin( bank/2.), cos( bank/2.) );
    Quaternion bq2( sin( attitude/2.) ,0 ,0, cos( attitude/2.) );
    Quaternion bq3( 0, 0, sin( heading/2.), cos( heading/2.) );
    Quaternion q = bq3.times(bq2.times(bq1));

    return q;
}


// receive a new pose from the dataflow
void sendPose( boost::shared_ptr< udp::socket > socket, boost::shared_ptr< udp::endpoint > endpoint, double pose[] )
{
	// Output stream
	std::ostringstream stream;

	// Retrieve timestamp
	boost::posix_time::ptime pt = boost::posix_time::microsec_clock::universal_time();
	boost::posix_time::ptime start( boost::gregorian::date( 1970,1,1 ) );
	boost::posix_time::time_duration diff = pt - start;
	__int64 timestamp = diff.total_nanoseconds();

	// Convert euler angles (in degree) to quaternion
	Quaternion q = euler2Quat( pose[3] * M_PI/180 , pose[4] * M_PI/180, pose[5] * M_PI/180 );
	const Position pos( pose[0] / 1000, pose[1] / 1000, pose[2] / 1000 );
	Pose p;
	p.position = pos;
	p.orientation = q;
	Measurement m;
	m.timestamp = timestamp;
	m.pose = p;

	// Prepare UDP packet as boost serialization archive
	//
	// UDP packet layout:
	// [ component name, measurement [ timestamp, pose [rotation, translation] ], timestamp, suffix ]
	// We serialize the same primitive data types as Ubitrack, thus the serialization should be identical.
	// Note that the measurement itself also contains a timestamp, it might differ from the second timestamp theoretically, 
	// but for us it safe to set the identical timestamp!
	boost::archive::text_oarchive archive( stream );
	std::string compName( "faro" );
	std::string suffix("\n");
	archive & compName;
	archive & m;
	archive & timestamp;
	archive & suffix;

	// Send frame
	boost::system::error_code ec;
	try {
		socket->send_to( boost::asio::buffer( stream.str().c_str(), stream.str().size() ), *endpoint );
	} 
	catch( ... ) 
	{
		printf( "Could not send pose via UDP" );
	}
}


void close( HMODULE hArmLib, int (*pCmmSpecific)(void *) ) {
	int nClose = DLL_CLOSE_CMM;
	pCmmSpecific((void *)&nClose);

	FreeLibrary(hArmLib);
}


int main( int argc, char* argv[] )
{
	// Load wrapper DLL
	HMODULE hArmLib = LoadLibrary( "FaroArmUsbWrapper.dll" );
	if( NULL == hArmLib )
	{
		printf( "Could not load FaroArmUsbWrapper.dll!\n" );
		return 20;
	}

	// 
	// Create UDP port
	//
	std::string dest;
	if ( argc > 1 )
		dest = std::string( argv[1] );
	else
		dest = std::string( "localhost" );

	int port;
	if ( argc > 2 )
		port = atoi( std::string( argv[2] ).c_str() );
	else
		port = 0x5554;

	std::stringstream outStream;
	outStream << "Receiver address: " << dest << ", Receiver port: " << port << "\n";
	printf( outStream.str().c_str() );

	boost::asio::io_service * m_IoService = new boost::asio::io_service();
	boost::shared_ptr< udp::socket > m_SendSocket;
	boost::shared_ptr< udp::endpoint > m_SendEndpoint;

	// open new socket which we use for sending stuff
	m_SendSocket = boost::shared_ptr< udp::socket >( new udp::socket( *m_IoService ) );
	boost::system::error_code ec;
	ec = m_SendSocket->open( udp::v4(), ec );
	if ( ec ) {
		printf("Could not open socket!\n");
		return 10;
	}

	// resolve destination pair and store the remote endpoint
	udp::resolver resolver( *m_IoService );

	std::ostringstream portString;
	portString << port;

	udp::resolver::query query( udp::v4(), dest, portString.str() );
	m_SendEndpoint = boost::shared_ptr< udp::endpoint >( new udp::endpoint( *resolver.resolve( query, ec ) ) );
	if ( ec )
	{
		printf( "Destination could not be resolved!\n" );
		return 11;
	}

	printf("Opening device...\r\n");

	int (*pCmmSpecific)(void *);
	pCmmSpecific = (BOOL (*)(VOID *))GetProcAddress(hArmLib, "CmmSpecific");
	if(NULL == pCmmSpecific)
	{
		pCmmSpecific = (BOOL (*)(VOID *))FalseSpecific;
	}

	// Open device
	struct vOpenInfo theOpenStruct = {DLL_OPEN_CMM, NULL, L"\0", NULL};
	if( ! (TRUE == pCmmSpecific((void *)&theOpenStruct) ) )
	{
		printf("Could not find arm!\n");
		close( hArmLib, pCmmSpecific );
		return 1;
	}	

	// Print device information
	struct vArmInfo devInfoStruct;
	memset(&devInfoStruct, 0x0, sizeof(struct vArmInfo));
	devInfoStruct.nFuncType = DLL_ARM_INFO;
	if ( ! (TRUE == pCmmSpecific((void *)&devInfoStruct) ) )
	{
		printf("Could not read device information!\n");
		close( hArmLib, pCmmSpecific);
		return 2;
	}
	printf( "Serial number: ");
	printf( devInfoStruct.sArmData );
	printf( "\nCertification date: " );
	printf( devInfoStruct.sCertificationDate );
	printf( "\nCalibration date: " );
	printf( devInfoStruct.sCalibrationDate );
	printf( "\nFirmware version: %i", devInfoStruct.nSerindVersion );
	printf( "\n" );

	// Register update structure
	struct PositionUpdate theUpdateStruct;
	memset( &theUpdateStruct, 0x0, sizeof(struct PositionUpdate) );
	theUpdateStruct.nFuncType = DLL_POS_UPDATE;
	theUpdateStruct.unKey = GOOD_KEY;

	if( ! ( TRUE == pCmmSpecific((void *)&theUpdateStruct) ) )
	{
		printf( "Could not register update struct!\n" );
		close( hArmLib, pCmmSpecific );
		return 3;
	}
	
	printf("\nHit 'p' for probe dialog.\n\nFront button to take points, back button or any key to exit.\n");

	//### Implement event steering instead of polling

	// Enter main loop
	for(;;)
	{
		if( theUpdateStruct.cUpdateFlag == CMM_UPDATE_DETECT )
		{
			theUpdateStruct.cUpdateFlag = 0;
			printf("pos: %lf %lf %lf, rot: %lf %lf %lf, temp: %lf\r\n", 
			theUpdateStruct.dPosition[0],
			theUpdateStruct.dPosition[1],
			theUpdateStruct.dPosition[2],
			theUpdateStruct.dPosition[3],
			theUpdateStruct.dPosition[4],
			theUpdateStruct.dPosition[5],
			theUpdateStruct.dPosition[6]);
			sendPose( m_SendSocket, m_SendEndpoint, theUpdateStruct.dPosition );

			if(theUpdateStruct.cBackButton)
			{
				//### Maybe send some event to Ubitrack
			}
		}
		else if ( theUpdateStruct.cUpdateFlag == CMM_TIMEOUT_DETECT ) 
		{
			printf( "Timeout occurred in communication with arm. Abort\n" );
			break;
		}
		else if ( theUpdateStruct.cUpdateFlag == CMM_ERROR_DETECT )
		{
			printf( "Error reported by arm: %i\n" );
		}
		else if ( theUpdateStruct.cUpdateFlag == CMM_INACCURATE_DATA_DETECT )
		{
			printf( "Inaccurate data detected by arm.\n" );
		}
		
		// Key hit
		if(_kbhit())
		{
			char cHit = _getch();
			if(cHit == 'p' || cHit == 'P')
			{
				int bEnd = FALSE;
				struct vGetProbeInfo theProbe;
				memset( &theProbe, 0x0, sizeof(struct vGetProbeInfo) );
				do
				{
					// Run probe calibration
					theProbe.nFuncType = DLL_PROBE_DIALOG;
					int nReturn = pCmmSpecific((void *)&theProbe);
					switch( nReturn )
					{
						// Run hole calibration
						case DLL_CALIBRATE_HOLE:
						{
							theProbe.nFuncType = DLL_CALIBRATE_HOLE;
							nReturn = pCmmSpecific((void *)&theProbe);
						}
						break;
						// Run sphere calibration
						case DLL_CALIBRATE_SPHERE:
						{
							theProbe.nFuncType = DLL_CALIBRATE_SPHERE;
							nReturn = pCmmSpecific((void *)&theProbe);
						}
						break;
						// Run plane calibration
						case DLL_CALIBRATE_PLANE:
						{
							theProbe.nFuncType = DLL_CALIBRATE_PLANE;
							nReturn = pCmmSpecific((void *)&theProbe);
						}
						break;
						default:
							bEnd = TRUE;
						break;
					};
				} while( FALSE == bEnd );
				
				//turn mouse-move off
				int nMouseMode = DLL_MOUSE_MODE_OFF;
				pCmmSpecific((void *)&nMouseMode);
			}
			else if ( cHit == 'a' || cHit == 'A' ) 
			{
				int anglesDialog = DLL_ANGLES_DIALOG;
				pCmmSpecific( (void *)&anglesDialog );
			}
			else break;					
		}
	}
	
	printf("\n\nClosing...\n");
	close( hArmLib, pCmmSpecific );
	return 0;
}
