#include <boost/test/unit_test.hpp>
#include <boost/bind.hpp>
#include <boost/thread.hpp>
#include <Ubitrack/Dataflow/EventQueue.h>
#include <Ubitrack/Dataflow/Component.h>
#include <Ubitrack/Dataflow/Port.h>
#include <Ubitrack/Util/OS.h>

using namespace Ubitrack::Dataflow;

// counter for received events
static int nEvents = 0;


static void ReceiveEvent()
{
	nEvents++;
}


void TestEventQueue()
{
	// precondition
	BOOST_CHECK_EQUAL( nEvents, 0 );

	// create list with one event
	std::vector< EventQueue::QueueData > events;
	events.push_back( EventQueue::QueueData( 0, boost::function< void () >( &ReceiveEvent ) ) );

	// queueing without thread
	EventQueue::singleton().queue( events );
	EventQueue::singleton().queue( events );
	BOOST_CHECK_EQUAL( nEvents, 0 );

	EventQueue::singleton().dispatchNow();
	BOOST_CHECK_EQUAL( nEvents, 2 );
	nEvents = 0;

	// test queueing from thread

	// queue some initial events
	EventQueue::singleton().queue( events );
	EventQueue::singleton().queue( events );

	EventQueue::singleton().start();

	for ( int i = 0; i < 1000; i++ )
	{
		EventQueue::singleton().queue( events );

		// simulate a busier system and give some resources to other threads
		if ( i % 59 == 0 )
			boost::thread::yield();
	}

	// wait a certain time until all events are delivered
	int nWaits = 0;
	while ( nWaits < 400 && nEvents < 1002 )
	{
		Ubitrack::Util::sleep( 10 );
		nWaits++;
	}

	EventQueue::singleton().stop();
	BOOST_CHECK_EQUAL( nEvents, 1002 );
	nEvents = 0;

	// check removal of components
	Component comp( "EventQueueTest" );
	Port port( "testPort", comp );
	EventQueue::ReceiverInfo receiverInfo( &port );
	std::vector< EventQueue::QueueData > riEvents;
	riEvents.push_back( EventQueue::QueueData( &receiverInfo, boost::function< void () >( &ReceiveEvent ) ) );

	EventQueue::singleton().queue( riEvents );
	EventQueue::singleton().queue( events );
	EventQueue::singleton().queue( riEvents );
	EventQueue::singleton().queue( events );
	EventQueue::singleton().queue( events );
	BOOST_CHECK_EQUAL( receiverInfo.nQueuedEvents, 2 );
	
	EventQueue::singleton().removeComponent( &comp );
	BOOST_CHECK_EQUAL( receiverInfo.nQueuedEvents, 0 );
	EventQueue::singleton().dispatchNow();

	BOOST_CHECK_EQUAL( nEvents, 3 );
	BOOST_CHECK_EQUAL( receiverInfo.nQueuedEvents, 0 );

	// check queue length management
	receiverInfo.nMaxQueueLength = 2;
	nEvents = 0;

	EventQueue::singleton().queue( riEvents );
	EventQueue::singleton().queue( riEvents );
	EventQueue::singleton().queue( riEvents );
	EventQueue::singleton().queue( riEvents );
	BOOST_CHECK_EQUAL( receiverInfo.nQueuedEvents, 4 );

	EventQueue::singleton().dispatchNow();

	BOOST_CHECK_EQUAL( nEvents, 2 );
	BOOST_CHECK_EQUAL( receiverInfo.nQueuedEvents, 0 );
	
	// check clear()
	nEvents = 0;
	EventQueue::singleton().queue( riEvents );
	EventQueue::singleton().queue( riEvents );

	EventQueue::singleton().clear();
	EventQueue::singleton().dispatchNow();

	BOOST_CHECK_EQUAL( nEvents, 0 );
	BOOST_CHECK_EQUAL( receiverInfo.nQueuedEvents, 0 );
}
