#include <boost/test/unit_test.hpp>
#include <boost/test/floating_point_comparison.hpp>
#include <boost/bind.hpp>
#include <boost/scoped_ptr.hpp>

#include <Ubitrack/Dataflow/TriggerComponent.h>
#include <Ubitrack/Dataflow/TriggerOutPort.h>
#include <Ubitrack/Dataflow/TriggerInPort.h>
#include <Ubitrack/Dataflow/ExpansionInPort.h>
#include <Ubitrack/Graph/UTQLReader.h>

using namespace Ubitrack;
using namespace Ubitrack::Dataflow;

namespace {

template< class EventType >
class TestPortRepository
	: public Component
{
public:
	TestPortRepository()
		: Component( "testPortRepository" )
	{}

	// send to pushconsumer
	void send( Port& receiver, const typename EventType::value_type& value, Measurement::Timestamp t )
	{
		if ( !m_pushMap[ &receiver ] )
		{
			m_pushMap[ &receiver ].reset( new PushSupplier< EventType >( receiver.getName() + "-testSender", *this ) );
			m_pushMap[ &receiver ]->connect( receiver );
		}
		m_pushMap[ &receiver ]->send( EventType( t, value ) );
	}
	
	// internal callback
	EventType doPull( Measurement::Timestamp t, Port* receiver )
	{
		if ( m_eventMap.find( receiver ) == m_eventMap.end() )
			UBITRACK_THROW( "no measurement for " + receiver->fullName() );
		return EventType( t, m_eventMap[ receiver ] );
	}

	// pullsupplier shall throw
	void setPull( Port& receiver )
	{
		if ( !m_pullMap[ &receiver ] )
		{
			m_pullMap[ &receiver ].reset( new PullSupplier< EventType >( receiver.getName() + "-testPullSupplier", *this,
				boost::bind( &TestPortRepository< EventType >::doPull, this, _1, &receiver ) ) );
			receiver.connect( *m_pullMap[ &receiver ] );
		}
		if ( m_eventMap.find( &receiver ) != m_eventMap.end() )
			m_eventMap.erase( &receiver );
	}
	
	// pullsupplier shall return result
	void setPull( Port& receiver, const typename EventType::value_type& value )
	{
		if ( !m_pullMap[ &receiver ] )
			setPull( receiver );
		m_eventMap[ &receiver ] = value;
	}

	// get result from pull port
	EventType pull( Port& receiver, Measurement::Timestamp t )
	{
		if ( !m_pullConsumerMap[ &receiver ] )
		{
			m_pullConsumerMap[ &receiver ].reset( new PullConsumer< EventType >( receiver.getName() + "-testPullConsumer", *this ) );
			m_pullConsumerMap[ &receiver ]->connect( receiver );
		}
		return m_pullConsumerMap[ &receiver ]->get( t );
	}
	
	std::map< Port*, boost::shared_ptr< PushSupplier< EventType > > > m_pushMap;
	std::map< Port*, boost::shared_ptr< PullSupplier< EventType > > > m_pullMap;
	std::map< Port*, boost::shared_ptr< PullConsumer< EventType > > > m_pullConsumerMap;
	std::map< Port*, typename EventType::value_type > m_eventMap;
};


enum TestMode { timeExpansion, vectorSpaceExpansion, portSpaceExpansion };

class TriggerTestPattern
	: public Graph::UTQLSubgraph
{
public:
	TriggerTestPattern( TestMode testMode, bool pushA, bool pushB, bool pushC, bool pushOut )
	{
		using namespace Graph;
		UTQLSubgraph::NodePtr pA = addNode( "A", UTQLNode( InOutAttribute::Input ) );
		UTQLSubgraph::NodePtr pB = addNode( "B", UTQLNode( InOutAttribute::Input ) );
		addEdge( "InA", pA, pB, UTQLEdge( InOutAttribute::Input ) )->setAttribute( "mode", AttributeValue( pushA ? "push" : "pull" ) );
		addEdge( "InA1", pA, pB, UTQLEdge( InOutAttribute::Input ) )->setAttribute( "mode", AttributeValue( pushA ? "push" : "pull" ) );
		addEdge( "InA2", pA, pB, UTQLEdge( InOutAttribute::Input ) )->setAttribute( "mode", AttributeValue( pushA ? "push" : "pull" ) );
		addEdge( "InB", pA, pB, UTQLEdge( InOutAttribute::Input ) )->setAttribute( "mode", AttributeValue( pushB ? "push" : "pull" ) );
		addEdge( "InB1", pA, pB, UTQLEdge( InOutAttribute::Input ) )->setAttribute( "mode", AttributeValue( pushB ? "push" : "pull" ) );
		addEdge( "InB2", pA, pB, UTQLEdge( InOutAttribute::Input ) )->setAttribute( "mode", AttributeValue( pushB ? "push" : "pull" ) );
		addEdge( "InC", pA, pB, UTQLEdge( InOutAttribute::Input ) )->setAttribute( "mode", AttributeValue( pushC ? "push" : "pull" ) );
		addEdge( "Out", pA, pB, UTQLEdge( InOutAttribute::Output ) )->setAttribute( "mode", AttributeValue( pushOut ? "push" : "pull" ) );
		m_DataflowAttributes.setAttribute( "expansion", AttributeValue( testMode == timeExpansion ? "time" : "space" ) );
	}
};


class TriggerTestComponent
	: public TriggerComponent
{
public:
	typedef Math::Scalar< double > SingleContent;
	typedef Measurement::Measurement< SingleContent > SingleMeasurement;
	typedef std::vector< SingleContent > VectorContent;
	typedef Measurement::Measurement< VectorContent > VectorMeasurement;
	
	TriggerTestComponent( TestMode testMode, bool pushA, bool pushB, bool pushC, bool pushOut )
		: TriggerComponent( "TriggerExpansionComponent", boost::shared_ptr< Graph::UTQLSubgraph >( new TriggerTestPattern( testMode, pushA, pushB, pushC, pushOut ) ) )
		, m_inA( "InA", *this )
		, m_inB( "InB", *this )
		, m_inC( "InC", *this )
		, m_out( "Out", *this )
		, m_nOutPushed( 0 )
		, m_outValue( -1000 )
	{
		if ( pushOut )
		{
			m_pOutTestPush.reset( new PushConsumer< SingleMeasurement >( "OutTest", *this,
				boost::bind( &TriggerTestComponent::outTestPush, this, _1 ) ) );
			m_out.connect( *m_pOutTestPush );
		}
	}

	void outTestPush( const SingleMeasurement& d )
	{
		m_nOutPushed++;
		m_outValue = *d;
	}
	
	void compute( Measurement::Timestamp t )
	{ 
		// compute scalar product of a and b, multipled by c
		const VectorMeasurement& a( m_inA.get() );
		const VectorMeasurement& b( m_inB.get() );
		double c = *m_inC.get();
		if ( a->size() != b->size() )
			UBITRACK_THROW( "Inputs have different size" );
			
		double r = 0;
		for ( unsigned i = 0; i < a->size(); i++ )
			r += (*a)[ i ] * (*b)[ i ] * c;
			
		m_out.send( SingleMeasurement( t, r ) ); 
	}
	
	// real ports of the TriggerComponents
	ExpansionInPort< SingleContent > m_inA;
	ExpansionInPort< SingleContent > m_inB;
	TriggerInPort< SingleMeasurement > m_inC;
	TriggerOutPort< SingleMeasurement > m_out;
	
	// out push test data
	int m_nOutPushed;
	double m_outValue;	
	boost::scoped_ptr< PushConsumer< SingleMeasurement > > m_pOutTestPush;
};

} // anonymous namespace


void TestExpansionPorts()
{
	// clear the event queue, just to be sure it is empty
	EventQueue::singleton().clear();
	
	{
	// test space (vector) pull+pull+pull=pull case
	TriggerTestComponent c( vectorSpaceExpansion, false, false, false, false );
	
	TestPortRepository< TriggerTestComponent::SingleMeasurement > singlePorts;
	TestPortRepository< TriggerTestComponent::VectorMeasurement > vectorPorts;
	TriggerTestComponent::VectorContent vA, vB;

	// no input -> throw
	BOOST_CHECK_THROW( singlePorts.pull( c.m_out, 1 ),  Ubitrack::Util::Exception );

	// only port a has measurement -> throw
	vA.push_back( 1.0 );
	vectorPorts.setPull( c.m_inA, vA );
	vectorPorts.setPull( c.m_inB );
	BOOST_CHECK_THROW( singlePorts.pull( c.m_out, 1 ), Ubitrack::Util::Exception );

	// only port a&c has measurement -> throw
	singlePorts.setPull( c.m_inC, 2.0 );
	BOOST_CHECK_THROW( singlePorts.pull( c.m_out, 1 ), Ubitrack::Util::Exception );

	// good -> check result
	vB.push_back( 6.0 );
	vectorPorts.setPull( c.m_inB, vB );
	BOOST_CHECK_CLOSE( double( *singlePorts.pull( c.m_out, 1 ) ), 12.0, 1e-6 );

	// wrong number of inputs -> throw
	vA.push_back( -1.0 );
	vectorPorts.setPull( c.m_inA, vA );
	BOOST_CHECK_THROW( singlePorts.pull( c.m_out, 1 ), Ubitrack::Util::Exception );

	// two measurements -> check result
	vB.push_back( 6.0 );
	vectorPorts.setPull( c.m_inB, vB );
	BOOST_CHECK_CLOSE( double( *singlePorts.pull( c.m_out, 5 ) ), 0.0, 1e-6 );
	}

	{
	// test space (vector) pull+push+pull=push case
	TriggerTestComponent c( vectorSpaceExpansion, false, true, false, true );

	TestPortRepository< TriggerTestComponent::SingleMeasurement > singlePorts;
	TestPortRepository< TriggerTestComponent::VectorMeasurement > vectorPorts;
	TriggerTestComponent::VectorContent vA, vB;

	// port a&c throw -> no push
	vB.push_back( -2.0 );
	vectorPorts.send( c.m_inB, vB, 1 );
	EventQueue::singleton().dispatchNow();
	BOOST_CHECK( c.m_nOutPushed == 0 );

	// port a throws -> no push
	singlePorts.setPull( c.m_inC, 2.0 );
	vectorPorts.send( c.m_inB, vB, 1 );
	EventQueue::singleton().dispatchNow();
	BOOST_CHECK( c.m_nOutPushed == 0 );

	// good. -> one push, check result
	vA.push_back( -1.0 );
	vectorPorts.setPull( c.m_inA, vA );
	vectorPorts.send( c.m_inB, vB, 1 );
	EventQueue::singleton().dispatchNow();
	BOOST_CHECK_EQUAL( c.m_nOutPushed, 1 );
	BOOST_CHECK_CLOSE( c.m_outValue, 4.0, 1e-6 );
	
	// good. (two-vectors) -> check result
	vA.push_back( 1.0 );
	vectorPorts.setPull( c.m_inA, vA );
	vB.push_back( 1.0 );
	vectorPorts.send( c.m_inB, vB, 1 );
	EventQueue::singleton().dispatchNow();
	BOOST_CHECK_EQUAL( c.m_nOutPushed, 2 );
	BOOST_CHECK_CLOSE( c.m_outValue, 6.0, 1e-6 );
	}


	{
	// test time push+pull+pull=push case
	TriggerTestComponent c( timeExpansion, true, false, false, true );
	TestPortRepository< TriggerTestComponent::SingleMeasurement > singlePorts;

	// inB throws -> should do nothing
	singlePorts.send( c.m_inA, 1.0, 1 );
	EventQueue::singleton().dispatchNow();
	BOOST_CHECK( c.m_inA.get()->empty() );
	BOOST_CHECK( c.m_nOutPushed == 0 );

	// inB throws -> should do nothing
	singlePorts.setPull( c.m_inC, 2.0 );
	singlePorts.send( c.m_inA, 5.0, 1 );
	EventQueue::singleton().dispatchNow();
	BOOST_CHECK( c.m_inA.get()->empty() );
	BOOST_CHECK( c.m_nOutPushed == 0 );

	// -> append correspondence to list and compute result
	singlePorts.setPull( c.m_inB, 1.0 );
	singlePorts.send( c.m_inA, 1.0, 1 );
	EventQueue::singleton().dispatchNow();
	BOOST_CHECK_EQUAL( c.m_inA.get()->size(), unsigned( 1 ) );
	BOOST_CHECK_EQUAL( c.m_inB.get()->size(), unsigned( 1 ) );
	BOOST_CHECK_EQUAL( c.m_nOutPushed, 1 );
	BOOST_CHECK_CLOSE( c.m_outValue, 2.0, 1e-6 );

	// inC throws -> append correspondence to list but do not compute result
	singlePorts.setPull( c.m_inC );
	singlePorts.send( c.m_inA, 1.0, 1 );
	EventQueue::singleton().dispatchNow();
	BOOST_CHECK_EQUAL( c.m_inA.get()->size(), unsigned( 2 ) );
	BOOST_CHECK_EQUAL( c.m_inB.get()->size(), unsigned( 2 ) );
	BOOST_CHECK_EQUAL( c.m_nOutPushed, 1 );

	// -> append to list and compute result
	singlePorts.setPull( c.m_inC, 1.0 );
	singlePorts.send( c.m_inA, 2.0, 1 );
	EventQueue::singleton().dispatchNow();
	BOOST_CHECK_EQUAL( c.m_inA.get()->size(), unsigned( 3 ) );
	BOOST_CHECK_EQUAL( c.m_inB.get()->size(), unsigned( 3 ) );
	BOOST_CHECK_EQUAL( c.m_nOutPushed, 2 );
	BOOST_CHECK_CLOSE( c.m_outValue, 4.0, 1e-6 );
	}
	
	{
	// test space (port duplication) push+pull+pull=push case
	TriggerTestComponent c( portSpaceExpansion, true, false, false, true );
	TestPortRepository< TriggerTestComponent::SingleMeasurement > singlePorts;
	
	// have to build subgraph....
	std::istringstream stream(
		"<UTQLResponse>"
		"<Pattern name=\"lkj\" id=\"TriggerExpansionComponent\">"
		"  <Input>"
		"    <Node name=\"A\"/><Node name=\"B\"/>"
		"    <Edge name=\"InA1\" source=\"A\" destination=\"B\" pattern-ref=\"TriggerExpansionComponent\" edge-ref=\"Out\"><Attribute name=\"mode\" value=\"push\"/></Edge>"
		"    <Edge name=\"InA2\" source=\"A\" destination=\"B\" pattern-ref=\"TriggerExpansionComponent\" edge-ref=\"Out\"><Attribute name=\"mode\" value=\"push\"/></Edge>"
		"    <Edge name=\"InB1\" source=\"A\" destination=\"B\" pattern-ref=\"TriggerExpansionComponent\" edge-ref=\"Out\"><Attribute name=\"mode\" value=\"pull\"/></Edge>"
		"    <Edge name=\"InB2\" source=\"A\" destination=\"B\" pattern-ref=\"TriggerExpansionComponent\" edge-ref=\"Out\"><Attribute name=\"mode\" value=\"pull\"/></Edge>"
		"    <Edge name=\"InC\" source=\"A\" destination=\"B\" pattern-ref=\"TriggerExpansionComponent\" edge-ref=\"Out\"><Attribute name=\"mode\" value=\"pull\"/></Edge>"
		"  </Input>"
		"  <Output>"
		"    <Edge name=\"Out\" source=\"A\" destination=\"B\"><Attribute name=\"mode\" value=\"push\"/></Edge>"
		"  </Output>"
		"  <DataflowConfiguration><UbitrackLib class=\"blah\"/><Attribute name=\"expansion\" value=\"space\"/></DataflowConfiguration>"
		"</Pattern>"
		"</UTQLResponse>" );
		
	c.generateSpaceExpansionPorts( Graph::UTQLReader::processInput( stream )->getSubgraphById( "TriggerExpansionComponent" ) );
	BOOST_REQUIRE_NO_THROW( c.getPortByName( "InA1" ) );
	BOOST_REQUIRE_NO_THROW( c.getPortByName( "InA2" ) );
	BOOST_REQUIRE_NO_THROW( c.getPortByName( "InB1" ) );
	BOOST_REQUIRE_NO_THROW( c.getPortByName( "InB2" ) );

	// inB1 throws -> should do nothing
	singlePorts.send( c.getPortByName( "InA1" ), -1.0, 1 );
	EventQueue::singleton().dispatchNow();
	BOOST_CHECK( c.m_inA.get()->empty() );
	BOOST_CHECK( c.m_nOutPushed == 0 );
	
	// A1/B1 good, c throws -> should do nothing, but add measurement to list
	singlePorts.setPull( c.getPortByName( "InB1" ), 4.0 );
	singlePorts.send( c.getPortByName( "InA1" ), -1.0, 1 );
	EventQueue::singleton().dispatchNow();
	BOOST_CHECK_EQUAL( c.m_inA.get()->size(), unsigned( 1 ) );
	BOOST_CHECK_EQUAL( c.m_inB.get()->size(), unsigned( 1 ) );
	BOOST_CHECK_EQUAL( c.m_nOutPushed, 0 );

	// all 5 ports good, should compute 1 result
	singlePorts.setPull( c.m_inC, 0.5 );
	singlePorts.setPull( c.getPortByName( "InB2" ), 2.0 );
	singlePorts.send( c.getPortByName( "InA2" ), 3.0, 1 );
	EventQueue::singleton().dispatchNow();
	BOOST_CHECK_EQUAL( c.m_inA.get()->size(), unsigned( 2 ) );
	BOOST_CHECK_EQUAL( c.m_inB.get()->size(), unsigned( 2 ) );
	BOOST_CHECK_EQUAL( c.m_nOutPushed, 1 );
	BOOST_CHECK_CLOSE( c.m_outValue, 1.0, 1e-6 );

	// new timestamp...
	// A2/B2&C good, c throws -> result based on A2&B2
	singlePorts.send( c.getPortByName( "InA2" ), 3.0, 2 );
	EventQueue::singleton().dispatchNow();
	BOOST_CHECK_EQUAL( c.m_inA.get()->size(), unsigned( 1 ) );
	BOOST_CHECK_EQUAL( c.m_inB.get()->size(), unsigned( 1 ) );
	BOOST_CHECK_EQUAL( c.m_nOutPushed, 2 );
	BOOST_CHECK_CLOSE( c.m_outValue, 3.0, 1e-6 );

	// new timestamp...
	// push A1 and A2 simultaneously, still should only compute 1 result, because of event queue peeking
	singlePorts.send( c.getPortByName( "InA1" ), -1.0, 3 );
	singlePorts.send( c.getPortByName( "InA2" ), 3.0, 3 );
	EventQueue::singleton().dispatchNow();
	BOOST_CHECK_EQUAL( c.m_inA.get()->size(), unsigned( 2 ) );
	BOOST_CHECK_EQUAL( c.m_inB.get()->size(), unsigned( 2 ) );
	BOOST_CHECK_EQUAL( c.m_nOutPushed, 3 );
	BOOST_CHECK_CLOSE( c.m_outValue, 1.0, 1e-6 );
	}
	
}
