// tests whether it is possible to connect matching and non-matching ports

#include <boost/test/unit_test.hpp>
#include <boost/scoped_ptr.hpp>
#include <boost/bind.hpp>

#include <Ubitrack/Dataflow/PullConsumer.h>
#include <Ubitrack/Dataflow/PullSupplier.h>
#include <Ubitrack/Dataflow/PushConsumer.h>
#include <Ubitrack/Dataflow/PushSupplier.h>
#include <Ubitrack/Dataflow/Component.h>


// start an anonymous namespace for test-internal classes
namespace {

using namespace Ubitrack::Dataflow;

class PushPullTest
	: public Component
{
public:
	PushPullTest()
		: Component( "PushPullTest" )
		, m_pushSupplierInt( "PushIntOut", *this )
		, m_pushConsumerInt( "PushIntIn", *this, boost::bind( &PushPullTest::testPushInt, this, _1 ) )
		, m_pushSupplierString( "PushStringOut", *this )
		, m_pushConsumerString( "PushStringIn", *this, boost::bind( &PushPullTest::testPushString, this, _1 ) )
		, m_pullSupplierInt( "PullOutInt", *this, boost::bind( &PushPullTest::testPullInt, this, _1 ) )
		, m_pullConsumerInt( "PullInInt", *this )
		, m_pullSupplierString( "PullStringOut", *this, boost::bind( &PushPullTest::testPullString, this, _1 ) )
		, m_pullConsumerString( "PullStringIn", *this )
	{}

	void testPushInt( const int& )
	{}

	void testPushString( const std::string& )
	{}
	
	int testPullInt( Ubitrack::Measurement::Timestamp )
	{ return int(); }

	std::string testPullString( Ubitrack::Measurement::Timestamp )
	{ return std::string(); }

protected:
	PushSupplier< int > m_pushSupplierInt;
	PushConsumer< int > m_pushConsumerInt;
	PushSupplier< std::string > m_pushSupplierString;
	PushConsumer< std::string > m_pushConsumerString;
	PullSupplier< int > m_pullSupplierInt;
	PullConsumer< int > m_pullConsumerInt;
	PullSupplier< std::string > m_pullSupplierString;
	PullConsumer< std::string > m_pullConsumerString;
};


void doConnect( Port& a, Port& b )
{
	a.connect( b );
	b.connect( a );
}

void doDisconnect( Port& a, Port& b )
{
	a.disconnect( b );
	b.disconnect( a );
}

} // anonymous namespace


void TestPushPull()
{
	boost::scoped_ptr< PushPullTest > c( new PushPullTest );

	// connnect check
	BOOST_CHECK_NO_THROW( doConnect( c->getPortByName( "PushStringIn" ), c->getPortByName( "PushStringOut" ) ) );
	BOOST_CHECK_NO_THROW( doConnect( c->getPortByName( "PullStringIn" ), c->getPortByName( "PullStringOut" ) ) );
	
	// only one consumer can connect
	BOOST_CHECK_THROW( doConnect( c->getPortByName( "PullStringIn" ), c->getPortByName( "PullStringOut" ) ), Ubitrack::Util::Exception );
	BOOST_CHECK_NO_THROW( doDisconnect( c->getPortByName( "PullStringIn" ), c->getPortByName( "PullStringOut" ) ) );
	BOOST_CHECK_NO_THROW( doConnect( c->getPortByName( "PullStringIn" ), c->getPortByName( "PullStringOut" ) ) );
	BOOST_CHECK_NO_THROW( doDisconnect( c->getPortByName( "PullStringIn" ), c->getPortByName( "PullStringOut" ) ) );
	
	// try to connect non-matching ports of other data type
	BOOST_CHECK_THROW( doConnect( c->getPortByName( "PullStringIn" ), c->getPortByName( "PullIntOut" ) ), Ubitrack::Util::Exception );
	BOOST_CHECK_THROW( doConnect( c->getPortByName( "PullIntIn" ), c->getPortByName( "PullStringOut" ) ), Ubitrack::Util::Exception );
	BOOST_CHECK_THROW( doConnect( c->getPortByName( "PushStringIn" ), c->getPortByName( "PushIntOut" ) ), Ubitrack::Util::Exception );
	BOOST_CHECK_THROW( doConnect( c->getPortByName( "PushIntIn" ), c->getPortByName( "PushStringOut" ) ), Ubitrack::Util::Exception );

	// try to connect non-matching ports of other protocol
	BOOST_CHECK_THROW( doConnect( c->getPortByName( "PushIntIn" ), c->getPortByName( "PullIntOut" ) ), Ubitrack::Util::Exception );
	BOOST_CHECK_THROW( doConnect( c->getPortByName( "PullIntIn" ), c->getPortByName( "PushIntOut" ) ), Ubitrack::Util::Exception );
	
}
