#include <boost/test/unit_test.hpp>
#include <boost/bind.hpp>

#include <Ubitrack/Dataflow/PushConsumer.h>
#include <Ubitrack/Dataflow/PushSupplier.h>
#include <Ubitrack/Dataflow/Component.h>

// start an anonymous namespace for test-internal classes
namespace {

using namespace Ubitrack::Dataflow;

template< class EventType > class PushTest
	: public Component
{
public:
	PushTest()
		: Component( "PushTest" )
		, m_Supplier( "Out", *this )
		, m_Consumer1( "In1", *this, boost::bind( &PushTest< EventType >::receive1, this, _1 ) )
		, m_Consumer2( "In2", *this, boost::bind( &PushTest< EventType >::receive2, this, _1 ) )
	{
		m_Supplier.connect( m_Consumer1 );
		m_Consumer1.connect( m_Supplier );
	}

	~PushTest()
	{
		m_Supplier.disconnect( m_Consumer1 );
		m_Consumer1.disconnect( m_Supplier );
	}

	void test1( const EventType& data )
	{
		m_nEventsReceived = 0;
		m_pData = &data;
		m_Supplier.send( data );
		EventQueue::singleton().dispatchNow();
		BOOST_CHECK_EQUAL( m_nEventsReceived, 1 );
	}

	void test2( const EventType& data )
	{
		// connect additional port
		m_Supplier.connect( m_Consumer2 );
		m_Consumer2.connect( m_Supplier );

		m_nEventsReceived = 0;
		m_pData = &data;
		m_Supplier.send( data );
		EventQueue::singleton().dispatchNow();
		BOOST_CHECK_EQUAL( m_nEventsReceived, 2 );

		// disconnect additional port
		m_Supplier.disconnect( m_Consumer2 );
		m_Consumer2.disconnect( m_Supplier );
	}

protected:
	void receive1( const EventType& data )
	{
		m_nEventsReceived++;
		BOOST_CHECK_EQUAL( *m_pData, data );
	}

	void receive2( const EventType& data )
	{
		m_nEventsReceived++;
		BOOST_CHECK_EQUAL( *m_pData, data );
	}

	PushSupplier< EventType > m_Supplier;
	PushConsumer< EventType > m_Consumer1;
	PushConsumer< EventType > m_Consumer2;

	int m_nEventsReceived;
	const EventType* m_pData;
};

} // anonymous namespace


void TestPush()
{
	// test pushing ints
	boost::shared_ptr< PushTest< int > > testInt( new PushTest< int > );
	testInt->test1( 4 );
	testInt->test2( -263 );
	testInt->test1( 0 );

	// test pushing strings
	boost::shared_ptr< PushTest< std::string > > testString( new PushTest< std::string > );
	testString->test1( std::string() );
	testString->test2( "Hi" );
	testString->test1( "adfasfd" );
}
