#include <boost/test/unit_test.hpp>
#include <boost/test/floating_point_comparison.hpp>

#include <vector>
#include <string>
#include <sstream>
#include <boost/scoped_ptr.hpp>

#include <Ubitrack/Measurement/Measurement.h>
#include <Ubitrack/Dataflow/DataflowNetwork.h>
#include <Ubitrack/Dataflow/ComponentFactory.h>
#include <Ubitrack/Dataflow/EventQueue.h>
#include <Ubitrack/Graph/UTQLReader.h>
#include <Ubitrack/Components/ApplicationPushSink.h>
#include <Ubitrack/Components/ApplicationPushSource.h>
#include <Ubitrack/Util/Logging.h>

#include "../tools.h"


using namespace Ubitrack;

static const char* sTestXml =
"<?xml version=\"1.0\" encoding=\"utf-8\"?>"
""
"<UTQLResponse>"
""
"<Pattern name=\"Source1\" id=\"A001\">"
"  <Input>"
"    <Node name=\"A\" id=\"A\"/>"
"  </Input>"
""
"  <Output>"
"    <Node name=\"B\" id=\"B\"/>"
"    <Edge name=\"Output\" source=\"A\" destination=\"B\">"
"      <Attribute name=\"type\" value=\"6D\"/>"
"    </Edge>"
"  </Output>"
""
"  <DataflowConfiguration>"
"    <UbitrackLib class=\"ApplicationPushSourcePose\"/>"
"  </DataflowConfiguration>"
"</Pattern>"
""
"<Pattern name=\"Source2\" id=\"A002\">"
"  <Input>"
"    <Node name=\"A\" id=\"C\"/>"
"  </Input>"
""
"  <Output>"
"    <Node name=\"B\" id=\"B\"/>"
"    <Edge name=\"Output\" source=\"A\" destination=\"B\">"
"      <Attribute name=\"type\" value=\"6D\"/>"
"    </Edge>"
"  </Output>"
""
"  <DataflowConfiguration>"
"    <UbitrackLib class=\"ApplicationPushSourcePose\"/>"
"  </DataflowConfiguration>"
"</Pattern>"
""
"<Pattern name=\"StaticTransformation\" id=\"A003\">"
"  <Input>"
"    <Node name=\"A\" id=\"C\"/>"
"    <Node name=\"B\" id=\"D\"/>"
"    <Edge name=\"AB\" source=\"A\" destination=\"B\">"
"      <Attribute name=\"type\" value=\"6D\"/>"
"      <Attribute name=\"staticPosition\" value=\"0 0.2 0\"/>"
"      <Attribute name=\"staticRotation\" value=\"0 0 1.0 0\"/>"
"    </Edge>"
"  </Input>"
""
"  <Output>"
"    <Edge name=\"AB\" source=\"A\" destination=\"B\">"
"      <Attribute name=\"type\" value=\"6D\"/>"
"    </Edge>"
"  </Output>"
""
"  <DataflowConfiguration>"
"    <UbitrackLib class=\"StaticPose\"/>"
"  </DataflowConfiguration>"
"</Pattern>"
""
"<Pattern name=\"Inversion\" id=\"A004\">"
"  <Input>"
"    <Node name=\"A\" id=\"B\"/>"
"    <Node name=\"B\" id=\"C\"/>"
"	<Edge name=\"AB\" source=\"A\" destination=\"B\" pattern-ref=\"A002\" edge-ref=\"Output\">"
"     <Attribute name=\"mode\" value=\"push\"/>"
"   </Edge>"
"  </Input>"
""
"  <Output>"
"    <Edge name=\"BA\" source=\"B\" destination=\"A\">"
"      <Attribute name=\"type\" value=\"6D\"/>"
"      <Attribute name=\"mode\" value=\"push\"/>"
"    </Edge>"
"  </Output>"
""
"  <DataflowConfiguration>"
"    <UbitrackLib class=\"PoseInversion\"/>"
"  </DataflowConfiguration>"
"</Pattern>"
""
"<Pattern name=\"SyncMultiplication\" id=\"A005\">"
"  <Input>"
"    <Node name=\"A\" id=\"A\"/>"
"	<Node name=\"B\" id=\"B\"/>"
"	<Node name=\"C\" id=\"C\"/>"
"	<Edge name=\"AB\" source=\"A\" destination=\"B\" pattern-ref=\"A001\" edge-ref=\"Output\">"
"      <Attribute name=\"mode\" value=\"push\"/>"
"   </Edge>"
"	<Edge name=\"BC\" source=\"B\" destination=\"C\" pattern-ref=\"A004\" edge-ref=\"BA\">"
"      <Attribute name=\"mode\" value=\"push\"/>"
"   </Edge>"
"  </Input>"
""
"  <Output>"
"    <Edge name=\"AC\" source=\"A\" destination=\"C\">"
"	  <Attribute name=\"type\" value=\"6D\"/>"
"     <Attribute name=\"mode\" value=\"push\"/>"
"    </Edge>"
"  </Output>"
""
"  <DataflowConfiguration>"
"    <UbitrackLib class=\"PoseMultiplication\"/>"
"  </DataflowConfiguration>"
"</Pattern>"
""
"<Pattern name=\"PushPullMultiplication\" id=\"A006\">"
"  <Input>"
"    <Node name=\"A\" id=\"A\"/>"
"	<Node name=\"B\" id=\"C\"/>"
"	<Node name=\"C\" id=\"D\"/>"
"	<Edge name=\"AB\" source=\"A\" destination=\"B\" pattern-ref=\"A005\" edge-ref=\"AC\">"
"      <Attribute name=\"mode\" value=\"push\"/>"
"   </Edge>"
"	<Edge name=\"BC\" source=\"B\" destination=\"C\" pattern-ref=\"A003\" edge-ref=\"AB\">"
"      <Attribute name=\"mode\" value=\"pull\"/>"
"   </Edge>"
"  </Input>"
""
"  <Output>"
"    <Edge name=\"AC\" source=\"A\" destination=\"C\">"
"	  <Attribute name=\"type\" value=\"6D\"/>"
"     <Attribute name=\"mode\" value=\"push\"/>"
"    </Edge>"
"  </Output>"
""
"  <DataflowConfiguration>"
"    <UbitrackLib class=\"PoseMultiplication\"/>"
"  </DataflowConfiguration>"
"</Pattern>"
""
"<Pattern name=\"Query\" id=\"Query\">"
"  <Input>"
"    <Node name=\"A\" id=\"A\"/>"
"    <Node name=\"B\" id=\"D\"/>"
"    <Edge name=\"Input\" source=\"A\" destination=\"B\" pattern-ref=\"A006\" edge-ref=\"AC\">"
"      <Attribute name=\"type\" value=\"6D\"/>"
"    </Edge>"
"  </Input>"
""
"  <DataflowConfiguration>"
"    <UbitrackLib class=\"ApplicationPushSinkPose\"/>"
"  </DataflowConfiguration>"
"</Pattern>"
""
"</UTQLResponse>";

#ifdef _WIN32
#define COMP_TO_LIB( x ) x ".dll"
#else
#define COMP_TO_LIB( x ) "lib" x ".so"
#endif

namespace {

// helper class to receive pose events
struct PoseStorage
{
	PoseStorage()
		: nReceived( 0 )
	{}

	void setPose( const Measurement::Pose& m )
	{
		nReceived++;
		pose = m;
	}

	int nReceived;
	Measurement::Pose pose;
};

}

void SimpleXmlDataflowTest()
{
	// in case you get strange errors in this test case
	//Util::initLogging();

	// create a factory
	std::vector< std::string > components;
	components.push_back( COMP_TO_LIB( "Multiplication" ) );
	components.push_back( COMP_TO_LIB( "Inversion" ) );
	components.push_back( COMP_TO_LIB( "ApplicationPushSource" ) );
	components.push_back( COMP_TO_LIB( "ApplicationPushSink" ) );
	components.push_back( COMP_TO_LIB( "StaticMeasurement" ) );
	boost::scoped_ptr< Dataflow::ComponentFactory > pFactory;
	pFactory.reset( new Dataflow::ComponentFactory( components ) );

	// create a new data flow network

	// load utql document
	std::istringstream input ( sTestXml );
	boost::shared_ptr< Graph::UTQLDocument > doc = Graph::UTQLReader::processInput( input );

	// instantiate the df network from UTQL
	boost::scoped_ptr< Dataflow::DataflowNetwork > pDfn( new Dataflow::DataflowNetwork( *pFactory ) );
	pDfn->processUTQLResponse( doc );

	// perform some sanity tests on the instantiated network

	// register callback
	PoseStorage result;
	pDfn->componentByName< Components::ApplicationPushSinkPose >( "Query" )->
		setCallback( boost::bind( &PoseStorage::setPose, &result, _1 ) );

	// get sources
	Components::ApplicationPushSourcePose::CallbackType sendA2B =
		pDfn->componentByName< Components::ApplicationPushSourcePose >( "A001" )->getCallback();
	Components::ApplicationPushSourcePose::CallbackType sendC2B =
		pDfn->componentByName< Components::ApplicationPushSourcePose >( "A002" )->getCallback();

	// clear the event queue, just to be sure it is empty
	Dataflow::EventQueue::singleton().clear();
	Dataflow::EventQueue::singleton().dispatchNow();
	BOOST_CHECK_EQUAL( result.nReceived, 0 );

	// push some data through the network
	Math::Pose a2b( randomQuaternion(), randomVector< 3, double >() );
	Math::Pose c2b( randomQuaternion(), randomVector< 3, double >() );
	Math::Pose c2d( Math::Quaternion( 0.0, 0.0, 1.0, 0.0 ), Math::Vector< 3 >( 0.0, 0.2, 0.0 ) );
	Math::Pose a2d( a2b * (~c2b) * c2d );

	sendC2B( Measurement::Pose( 8, c2b ) );
	Dataflow::EventQueue::singleton().dispatchNow();
	BOOST_CHECK_EQUAL( result.nReceived, 0 );

	sendA2B( Measurement::Pose( 8, a2b ) );
	Dataflow::EventQueue::singleton().dispatchNow();
	BOOST_CHECK_EQUAL( result.nReceived, 1 );
	BOOST_CHECK_SMALL( vectorDiff( a2d.translation(), result.pose->translation() ), 1e-6 );
	BOOST_CHECK_SMALL( quaternionDiff( a2d.rotation(), result.pose->rotation() ), 1e-6 );

	sendA2B( Measurement::Pose( 10, a2b ) );
	Dataflow::EventQueue::singleton().dispatchNow();
	BOOST_CHECK_EQUAL( result.nReceived, 1 );

	sendC2B( Measurement::Pose( 13, c2b ) );
	Dataflow::EventQueue::singleton().dispatchNow();
	BOOST_CHECK_EQUAL( result.nReceived, 1 );
}
