#include <boost/test/unit_test.hpp>
#include <boost/test/floating_point_comparison.hpp>
#include <boost/bind.hpp>
#include <boost/scoped_ptr.hpp>

#include <Ubitrack/Dataflow/TriggerComponent.h>
#include <Ubitrack/Dataflow/TriggerOutPort.h>
#include <Ubitrack/Dataflow/TriggerInPort.h>

using namespace Ubitrack;
using namespace Ubitrack::Dataflow;

namespace {

class TriggerTestPattern
	: public Graph::UTQLSubgraph
{
public:
	TriggerTestPattern( bool pushA, bool pushB, bool pushOut )
	{
		using namespace Graph;
		UTQLSubgraph::NodePtr pA = addNode( "A", UTQLNode( InOutAttribute::Input ) );
		UTQLSubgraph::NodePtr pB = addNode( "B", UTQLNode( InOutAttribute::Input ) );
		addEdge( "InA", pA, pB, UTQLEdge( InOutAttribute::Input ) )->setAttribute( "mode", AttributeValue( pushA ? "push" : "pull" ) );
		addEdge( "InB", pA, pB, UTQLEdge( InOutAttribute::Input ) )->setAttribute( "mode", AttributeValue( pushB ? "push" : "pull" ) );
		addEdge( "Out", pA, pB, UTQLEdge( InOutAttribute::Output ) )->setAttribute( "mode", AttributeValue( pushOut ? "push" : "pull" ) );
	}
};

class TriggerTestComponent
	: public TriggerComponent
{
public:
	TriggerTestComponent( bool pushA, bool pushB, bool pushOut )
		: TriggerComponent( "TriggerComponent", boost::shared_ptr< Graph::UTQLSubgraph >( new TriggerTestPattern( pushA, pushB, pushOut ) ) )
		, m_inA( "InA", *this )
		, m_inB( "InB", *this )
		, m_out( "Out", *this )
		, m_nOutPushed( 0 )
		, m_inAValue( -1000 )
		, m_inBValue( -1000 )
		, m_outValue( -1000 )
	{
		if ( pushA )
		{
			m_pInATestPush.reset( new PushSupplier< Measurement::Distance >( "InATest", *this ) );
			m_pInATestPush->connect( m_inA );
		}
		else
		{
			m_pInATestPull.reset( new PullSupplier< Measurement::Distance >( "InATest", *this, 
				boost::bind( &TriggerTestComponent::inATestPull, this, _1 ) ) );
			m_inA.connect( *m_pInATestPull );
		}
				
		if ( pushB )
		{
			m_pInBTestPush.reset( new PushSupplier< Measurement::Distance >( "InBTest", *this ) );
			m_pInBTestPush->connect( m_inB );
		}
		else
		{
			m_pInBTestPull.reset( new PullSupplier< Measurement::Distance >( "InBTest", *this, 
				boost::bind( &TriggerTestComponent::inBTestPull, this, _1 ) ) );
			m_inB.connect( *m_pInBTestPull );
		}
				
		if ( pushOut )
		{
			m_pOutTestPush.reset( new PushConsumer< Measurement::Distance >( "OutTest", *this,
				boost::bind( &TriggerTestComponent::outTestPush, this, _1 ) ) );
			m_out.connect( *m_pOutTestPush );
		}
		else
		{
			m_pOutTestPull.reset( new PullConsumer< Measurement::Distance >( "OutTest", *this ) ); 
			m_pOutTestPull->connect( m_out );
		}
	}

	Measurement::Distance inATestPull( Measurement::Timestamp t )
	{
		if ( m_inAValue < 0 )
			UBITRACK_THROW( "inATestPull not available" ); 
		return Measurement::Distance( t, m_inAValue );
	}
	
	Measurement::Distance inBTestPull( Measurement::Timestamp t )
	{
		if ( m_inBValue < 0 )
			UBITRACK_THROW( "inBTestPull not available" ); 
		return Measurement::Distance( t, m_inBValue );
	}

	void outTestPush( const Measurement::Distance& d )
	{
		m_nOutPushed++;
		m_outValue = *d;
	}
	
	void compute( Measurement::Timestamp t )
	{ 
		// add the two input distances
		m_out.send( Measurement::Distance( t, *m_inA.get() + *m_inB.get() ) ); 
	}
	
	// real ports of the TriggerComponents
	TriggerInPort< Measurement::Distance > m_inA;
	TriggerInPort< Measurement::Distance > m_inB;
	TriggerOutPort< Measurement::Distance > m_out;
	
	// test data
	int m_nOutPushed;
	double m_inAValue;
	double m_inBValue;
	double m_outValue;	
	
	// ports for testing (connect to the ports above)
	boost::scoped_ptr< PushSupplier< Measurement::Distance > > m_pInATestPush;
	boost::scoped_ptr< PushSupplier< Measurement::Distance > > m_pInBTestPush;
	boost::scoped_ptr< PullSupplier< Measurement::Distance > > m_pInATestPull;
	boost::scoped_ptr< PullSupplier< Measurement::Distance > > m_pInBTestPull;
	boost::scoped_ptr< PushConsumer< Measurement::Distance > > m_pOutTestPush;
	boost::scoped_ptr< PullConsumer< Measurement::Distance > > m_pOutTestPull;
};

} // anonymous namespace


void TestTriggerPorts()
{
	// clear the event queue, just to be sure it is empty
	EventQueue::singleton().clear();
	
	{
	// test pull+pull=pull case
	TriggerTestComponent c( false, false, false );
	
	BOOST_CHECK_THROW( c.m_pOutTestPull->get( 1 ), Ubitrack::Util::Exception );

	c.m_inAValue = 1.0;
	BOOST_CHECK_THROW( c.m_pOutTestPull->get( 1 ), Ubitrack::Util::Exception );

	c.m_inBValue = 2.0;
	BOOST_CHECK_CLOSE( double( *c.m_pOutTestPull->get( 1 ) ), 3.0, 10e-6 );

	c.m_inAValue = -1000;
	BOOST_CHECK_THROW( c.m_pOutTestPull->get( 1 ), Ubitrack::Util::Exception );
	}
	
	{
	// test push+pull=push case
	TriggerTestComponent c( true, false, true );
	
	c.m_pInATestPush->send( Measurement::Distance( 1, 1.0 ) );
	EventQueue::singleton().dispatchNow();
	BOOST_CHECK( c.m_nOutPushed == 0 );

	c.m_inBValue = 40.0;
	c.m_pInATestPush->send( Measurement::Distance( 1, 10.0 ) );
	EventQueue::singleton().dispatchNow();
	BOOST_CHECK_EQUAL( c.m_nOutPushed, 1 );
	BOOST_CHECK_CLOSE( c.m_outValue, 50.0, 10e-6 );

	c.m_inBValue = -1000;
	c.m_pInATestPush->send( Measurement::Distance( 7, 10.0 ) );
	EventQueue::singleton().dispatchNow();
	BOOST_CHECK_EQUAL( c.m_nOutPushed, 1 );
	}

	{
	// test pull+push=push case
	TriggerTestComponent c( false, true, true );
	
	c.m_pInBTestPush->send( Measurement::Distance( 1, 1.0 ) );
	EventQueue::singleton().dispatchNow();
	BOOST_CHECK( c.m_nOutPushed == 0 );

	c.m_inAValue = 10.0;
	c.m_pInBTestPush->send( Measurement::Distance( 1, 20.0 ) );
	EventQueue::singleton().dispatchNow();
	BOOST_CHECK_EQUAL( c.m_nOutPushed, 1 );
	BOOST_CHECK_CLOSE( c.m_outValue, 30.0, 10e-6 );

	c.m_inAValue = -1000;
	c.m_pInBTestPush->send( Measurement::Distance( 7, 10.0 ) );
	EventQueue::singleton().dispatchNow();
	BOOST_CHECK_EQUAL( c.m_nOutPushed, 1 );
	}

	{
	// test push+push=push case
	TriggerTestComponent c( true, true, true );
	
	c.m_pInATestPush->send( Measurement::Distance( 1, 1.0 ) );
	c.m_pInBTestPush->send( Measurement::Distance( 2, 1.0 ) );
	EventQueue::singleton().dispatchNow();
	BOOST_CHECK( c.m_nOutPushed == 0 );

	c.m_pInATestPush->send( Measurement::Distance( 2, 5.0 ) );
	EventQueue::singleton().dispatchNow();
	BOOST_CHECK_EQUAL( c.m_nOutPushed, 1 );
	BOOST_CHECK_CLOSE( c.m_outValue, 6.0, 10e-6 );

	c.m_pInATestPush->send( Measurement::Distance( 3, 8.0 ) );
	EventQueue::singleton().dispatchNow();
	BOOST_CHECK_EQUAL( c.m_nOutPushed, 1 );
	
	c.m_pInATestPush->send( Measurement::Distance( 4, 9.0 ) );
	EventQueue::singleton().dispatchNow();
	BOOST_CHECK_EQUAL( c.m_nOutPushed, 1 );

	c.m_pInBTestPush->send( Measurement::Distance( 4, 5.0 ) );
	EventQueue::singleton().dispatchNow();
	BOOST_CHECK_EQUAL( c.m_nOutPushed, 2 );
	BOOST_CHECK_CLOSE( c.m_outValue, 14.0, 10e-6 );
	}
	
	{
	// test the somewhat exotic push+pull=pull case
	TriggerTestComponent c( true, false, false );
	
	BOOST_CHECK_THROW( c.m_pOutTestPull->get( 1 ), Ubitrack::Util::Exception );

	c.m_pInATestPush->send( Measurement::Distance( 1, 1.0 ) );
	EventQueue::singleton().dispatchNow();
	BOOST_CHECK_THROW( c.m_pOutTestPull->get( 1 ), Ubitrack::Util::Exception );

	c.m_inBValue = 2.0;
	BOOST_CHECK_CLOSE( double( *c.m_pOutTestPull->get( 1 ) ), 3.0, 10e-6 );
	BOOST_CHECK_THROW( c.m_pOutTestPull->get( 2 ), Ubitrack::Util::Exception );

	c.m_pInATestPush->send( Measurement::Distance( 2, 4.0 ) );
	EventQueue::singleton().dispatchNow();
	BOOST_CHECK_CLOSE( double( *c.m_pOutTestPull->get( 2 ) ), 6.0, 10e-6 );

	c.m_inBValue = -1000;
	BOOST_CHECK_THROW( c.m_pOutTestPull->get( 2 ), Ubitrack::Util::Exception );
	}
}
