#include <boost/test/unit_test.hpp>

#include <Ubitrack/Graph/Graph.h>
#include <Ubitrack/Graph/KeyValueAttributes.h>
#include <Ubitrack/Graph/InOutAttribute.h>
#include <Ubitrack/Util/Exception.h>

#include <iostream>

void TestBasicGraph()
{
	using namespace Ubitrack::Graph;

	KeyValueAttributes nodeArtAttrs;
	nodeArtAttrs["ArtConfigurationPort"] = AttributeValue("5000");
	BOOST_CHECK( nodeArtAttrs.getAttribute( "ArtConfigurationPort" ).getText() == "5000" );
	BOOST_CHECK_THROW( nodeArtAttrs.getAttribute( "baz" ), Ubitrack::Util::Exception );

	InOutAttribute edgeArtToArtHmdAttrs( InOutAttribute::Input );
	// edgeArtToArtHmdAttrs["type"]="6D";
	InOutAttribute edgeArtToArtSheepAttrs( InOutAttribute::Output );
	// edgeArtToArtSheepAttrs["type"]="6D";

	typedef Graph<KeyValueAttributes, InOutAttribute> TestGraph;
	TestGraph testGraph;

	TestGraph::NodePtr artNode = testGraph.addNode( "Art", nodeArtAttrs);
	TestGraph::NodePtr artHmdNode = testGraph.addNode( "ArtHmd" );
	TestGraph::NodePtr artSheepNode = testGraph.addNode( "ArtSheep" );
	BOOST_CHECK_THROW( testGraph.addNode( "ArtSheep" ), Ubitrack::Util::Exception );
	BOOST_CHECK( artNode != artHmdNode );
	BOOST_CHECK( artNode != artSheepNode );
	BOOST_CHECK( artHmdNode != artSheepNode );

	BOOST_CHECK( testGraph.getNode( "Art" ) == artNode );
	BOOST_CHECK( testGraph.getNode( "ArtHmd" ) == artHmdNode );
	BOOST_CHECK( testGraph.getNode( "ArtSheep" ) == artSheepNode );

	BOOST_CHECK( testGraph.empty() );

	TestGraph::EdgePtr artToArtHmdEdge = testGraph.addEdge( "ArtToArtHmd", "Art", "ArtHmd", edgeArtToArtHmdAttrs );
	TestGraph::EdgePtr artToArtSheepEdge = testGraph.addEdge( "ArtToArtSheep", "Art", "ArtSheep", edgeArtToArtSheepAttrs );
	TestGraph::EdgePtr artHmdToArtSheepEdge = testGraph.addEdge( "ArtHmdToArtSheep", "ArtHmd", "ArtSheep", InOutAttribute( InOutAttribute::Output ) );

	BOOST_CHECK( artNode->m_OutEdges.size() == 2 );
	BOOST_CHECK( artNode->m_InEdges.size() == 0 );
	BOOST_CHECK( artHmdNode->m_OutEdges.size() == 1 );
	BOOST_CHECK( artHmdNode->m_InEdges.size() == 1 );
	BOOST_CHECK( artSheepNode->m_OutEdges.size() == 0 );
	BOOST_CHECK( artSheepNode->m_InEdges.size() == 2 );
	BOOST_CHECK( testGraph.order() == 3 );
	BOOST_CHECK( testGraph.size() == 3 );

	TestGraph::Node::EdgeList& l = artNode->m_OutEdges;
	for ( TestGraph::Node::EdgeList::iterator it = l.begin();
		  it != l.end(); ++it )
	{
		std::string endPoint = it->lock()->m_Target.lock()->m_Name;
		BOOST_CHECK ( endPoint == "ArtHmd" || endPoint == "ArtSheep" );
	}

	BOOST_CHECK( artToArtHmdEdge != artToArtSheepEdge );
	BOOST_CHECK( artToArtHmdEdge != artHmdToArtSheepEdge );
	BOOST_CHECK( artToArtSheepEdge != artHmdToArtSheepEdge );

// 	BOOST_CHECK( testGraph.getEdge( "ArtToArtHmd" ) == artToArtHmdEdge );
// 	BOOST_CHECK( testGraph.getEdge( "ArtToArtSheep" ) == artToArtSheepEdge );
// 	BOOST_CHECK( testGraph.getEdge( "ArtHmdToArtSheep" ) == artHmdToArtSheepEdge );

	// Test Input/Output-Iterators
	InOutIterator<KeyValueAttributes, InOutAttribute> inputIterator( artNode->m_OutEdges, InOutIterator<KeyValueAttributes, InOutAttribute>::select_input );
	for ( ; !inputIterator.isEnd(); inputIterator++ )
	{
		std::string endPoint = inputIterator->lock()->m_Target.lock()->m_Name;
		BOOST_CHECK( endPoint == "ArtHmd" );
	}

	testGraph.removeEdge( artToArtSheepEdge );
	artToArtSheepEdge.reset();

	BOOST_CHECK( artNode->m_OutEdges.size() == 1 );
	BOOST_CHECK( artNode->m_InEdges.size() == 0 );
	BOOST_CHECK( artHmdNode->m_OutEdges.size() == 1 );
	BOOST_CHECK( artHmdNode->m_InEdges.size() == 1 );
	BOOST_CHECK( artSheepNode->m_OutEdges.size() == 0 );
	BOOST_CHECK( artSheepNode->m_InEdges.size() == 1 );
	BOOST_CHECK( testGraph.order() == 3 );
	BOOST_CHECK( testGraph.size() == 2 );

	testGraph.removeNode( "Art" );
	artNode.reset();

	BOOST_CHECK( artHmdNode->m_OutEdges.size() == 1 );
	BOOST_CHECK( artHmdNode->m_InEdges.size() == 0 );
	BOOST_CHECK( artSheepNode->m_OutEdges.size() == 0 );
	BOOST_CHECK( artSheepNode->m_InEdges.size() == 1 );
	BOOST_CHECK( testGraph.order() == 2 );
	BOOST_CHECK( testGraph.size() == 1 );

	testGraph.removeNode( "ArtSheep" );
	testGraph.removeNode( "ArtHmd" );

	BOOST_CHECK( testGraph.null() );
	BOOST_CHECK( testGraph.empty() );
}
