#include <boost/test/unit_test.hpp>
#include <boost/test/floating_point_comparison.hpp>

#include <Ubitrack/Graph/PredicateParser.h>
#include <Ubitrack/Graph/Predicate.h>
#include <Ubitrack/Graph/AttributeExpression.h>
#include <Ubitrack/Graph/KeyValueAttributes.h>
#include <Ubitrack/Graph/UTQLSubgraph.h>
#include <Ubitrack/Graph/SRGraph.h>
#include <Ubitrack/Graph/EvaluationContext.h>
#include <Ubitrack/Util/Exception.h>

// for debugging
namespace std {

static std::ostream& operator<<( std::ostream& s, const Ubitrack::Graph::Predicate::AttribList& l )
{
	using namespace Ubitrack::Graph;
	s << "{";
	for ( Predicate::AttribList::const_iterator it = l.begin(); it != l.end(); it++ )
	{
		if ( it != l.begin() )
			s << ", ";
		s << "(" << it->first << ", " << it->second << ")";
	}
	s << "}";
	
	return s;
}

}

void TestPredicate()
{
	using namespace Ubitrack;
	using namespace Ubitrack::Graph;

	// create test attribute map
	SRGNodeAttributes testAttrs;
	testAttrs.setAttribute( "alpha", AttributeValue( "a" ) );
	testAttrs.setAttribute( "a", AttributeValue( "a" ) );
	testAttrs.setAttribute( "b", AttributeValue( "b" ) );
	testAttrs.setAttribute( "c", AttributeValue( "hi" ) );
	testAttrs.setAttribute( "test", AttributeValue( "1" ) );
	EvaluationContext ca( testAttrs );

	BOOST_CHECK( parsePredicate( "alpha==\"a\"" )->evaluate( ca ) );
	BOOST_CHECK( !parsePredicate( "alpha!=\"a\"" )->evaluate( ca ) );
	BOOST_CHECK( parsePredicate( "alpha==a" )->evaluate( ca ) );
	BOOST_CHECK( !parsePredicate( "alpha!=a" )->evaluate( ca ) );
	BOOST_CHECK( parsePredicate( "test==1" )->evaluate( ca ) );
	BOOST_CHECK_THROW( parsePredicate( "test=1" ), Ubitrack::Util::Exception );
	BOOST_CHECK( parsePredicate( "alpha==\"a\" && \"b\"==b" )->evaluate( ca ) );
	BOOST_CHECK( !parsePredicate( "!(test==1) || !c==\"hi\"" )->evaluate( ca ) );
	BOOST_CHECK( !parsePredicate( "\"super\"==\"trouper\"" )->evaluate( ca ) );
	BOOST_CHECK_THROW( parsePredicate( "test==!3" ), Ubitrack::Util::Exception );
	
	// check conjunctive equalities
	Predicate::AttribList listTruth;
	
	listTruth.push_back( Predicate::AttribList::value_type( "alpha", "a" ) );
	BOOST_CHECK_EQUAL( parsePredicate( "alpha==\"a\"" )->getConjunctiveEqualities(), listTruth );
	BOOST_CHECK_EQUAL( parsePredicate( "alpha==\"a\" && beta!=\"epsilon\"" )->getConjunctiveEqualities(), listTruth );

	listTruth.push_back( std::pair< std::string, std::string >( "gamma", "c" ) );
	BOOST_CHECK_EQUAL( parsePredicate( "alpha==\"a\" && beta!=\"epsilon\" && (a==\"1\" || b==\"2\") && gamma==\"c\" && !x==\"y\"" )->getConjunctiveEqualities(), listTruth );
	BOOST_CHECK_EQUAL( parsePredicate( "a==\"1\" || b==\"2\" || (a==\"2\" && b==\"5\")" )->getConjunctiveEqualities(), Predicate::AttribList() );
	
	// checkpredicate evaluation on subgraphs
	KeyValueAttributes A_Attributes;
	KeyValueAttributes B_Attributes;
	KeyValueAttributes AB_Attributes;
	A_Attributes[ "x" ] = AttributeValue( "1" );
	A_Attributes[ "y" ] =  AttributeValue( "2" );
	AB_Attributes[ "x" ] =  AttributeValue( 2.0 );

	EdgeMatching< UTQLSubgraph, SRGraph > matching;
	matching.m_allInputAttributes[ "A" ] = &A_Attributes;
	matching.m_allInputAttributes[ "B" ] = &B_Attributes;
	matching.m_allInputAttributes[ "AB" ] = &AB_Attributes;
	EvaluationContext cm( matching );
	
	BOOST_CHECK( parsePredicate( "A.y==AB.x" )->evaluate( cm ) );
	BOOST_CHECK( parsePredicate( "A.y!=A.x" )->evaluate( cm ) );
	BOOST_CHECK( !parsePredicate( "A.z==\"1\"" )->evaluate( cm ) );
	BOOST_CHECK( !parsePredicate( "A.x!=\"1\"" )->evaluate( cm ) );
	BOOST_CHECK( parsePredicate( "\"ddd\"!=\"dd\"" )->evaluate( cm ) );
	BOOST_CHECK( parsePredicate( "NoEdge.noAttr==\"\"" )->evaluate( cm ) );
	BOOST_CHECK( !parsePredicate( "B.noAttr==\"xyz\"" )->evaluate( cm ) );
	
	// check attribute expressions
	BOOST_CHECK( parseAttributeExpression( "A.x" )->evaluate( cm ).getText() == "1" );
	BOOST_CHECK( parseAttributeExpression( "AB.x" )->evaluate( cm ).getText() == "2" );
	BOOST_CHECK( parseAttributeExpression( "AB.noExist" )->evaluate( cm ).getText() == "" );
	BOOST_CHECK( parseAttributeExpression( "noExist.noExist" )->evaluate( cm ).getText() == "" );
	BOOST_CHECK( parseAttributeExpression( "\"abc\"" )->evaluate( cm ).getText() == "abc" );
	BOOST_CHECK( parseAttributeExpression( "\"\"" )->evaluate( cm ).getText() == "" );
	BOOST_CHECK_THROW( parseAttributeExpression( "\"abc\"==a" ), Ubitrack::Util::Exception );
	
	// check attribute expressions with math
	BOOST_CHECK_CLOSE( parseAttributeExpression( "(1+4*2)^0.5-3" )->evaluate( cm ).getNumber(), 0.0, 0.001 );
	BOOST_CHECK_CLOSE( parseAttributeExpression( "A.y^2*AB.x/4+5" )->evaluate( cm ).getNumber(), 7.0, 0.001 );
}
