#include <sstream>

#include <boost/shared_ptr.hpp>
#include <boost/scoped_ptr.hpp>

#include <boost/test/unit_test.hpp>
#include <boost/bind.hpp>

#include <Ubitrack/Dataflow/DataflowNetwork.h>
#include <Ubitrack/Dataflow/DataflowNetwork.h>

#include <Ubitrack/Graph/UTQLReader.h>
#include <Ubitrack/Graph/UTQLDocument.h>
#include <Ubitrack/Graph/UTQLSRGManager.h>
#include <Ubitrack/Graph/UTQLServer.h>
#include <Ubitrack/Graph/EdgeMatching.h>
#include <Ubitrack/Graph/DataflowGeneration.h>

#include <Ubitrack/Util/Logging.h>

#ifdef _WIN32
#define COMP_TO_LIB( x ) x ".dll"
#else
#define COMP_TO_LIB( x ) "lib" x ".so"
#endif

static const char* sTestSRG =
"<?xml version=\"1.0\" encoding=\"utf-8\"?>\n"
"<UTQL>\n"
"\n"
"    <Pattern name=\"Art\">\n"
"        <Output>\n"
"            <Node name=\"Art\" id=\"Art\">\n"
"                <Attribute name=\"ArtPort\" value=\"5000\"/>\n"
"            </Node>\n"
"        </Output>\n"
"    </Pattern>\n"
"    \n"
"    <Pattern name=\"Hmd\">\n"
"        <Output>\n"
"            <Node name=\"Art\" id=\"Art\"/>\n"
"            <Node name=\"ArtHmd\" id=\"ArtHmd\"/>\n"
"            <Edge name=\"ArtToArtHmd\" source=\"Art\" destination=\"ArtHmd\"> \n"
"                <Attribute name=\"type\" value=\"6D\"/>\n"
"            </Edge>\n"
"        </Output>\n"
"        <DataflowConfiguration>\n"
"            <UbitrackLib class=\"ArtTracker\"/>\n"
"            <ArtTarget id='3'/>\n"
"        </DataflowConfiguration>   \n"
"    </Pattern>\n"
"    \n"
"    <Pattern name=\"Sheep\">\n"
"        <Output>\n"
"            <Node name=\"Art\" id=\"Art\"/>\n"
"            <Node name=\"ArtSheep\" id=\"ArtSheep\"/>\n"
"            <Edge name=\"ArtToArtSheep\" source=\"Art\" destination=\"ArtSheep\">\n"
"                <Attribute name=\"type\" value=\"6D\"/>\n"
"            </Edge>\n"
"        </Output>\n"
"        <DataflowConfiguration>\n"
"            <UbitrackLib class=\"ArtTracker\"/>\n"
"            <ArtTarget id='7'/>\n"
"        </DataflowConfiguration>\n"
"    </Pattern>\n"
"\n"
"    <Pattern name=\"Dummy\">\n"
"        <Output>\n"
"            <Node name=\"Art\" id=\"Art\"/>\n"
"            <Node name=\"Anonymous\"/>\n"
"            <Edge name=\"ArtToAnonymous\" source=\"Art\" destination=\"Anonymous\">\n"
"                <Attribute name=\"type\" value=\"3D\"/>\n"
"            </Edge>\n"
"        </Output>\n"
"        <DataflowConfiguration>\n"
"            <UbitrackLib class=\"ArtTracker\"/>\n"
"            <ArtTarget id='23'/>\n"
"        </DataflowConfiguration>\n"
"    </Pattern>\n"
"\n"
"</UTQL>\n";

static const char* sTestPattern =
"<?xml version=\"1.0\" encoding=\"utf-8\"?>\n"
"<UTQL>\n"
"\n"
"<Pattern name=\"Inversion\">\n"
"  <Input>\n"
"    <Node name=\"NodeA\"/>\n"
"    <Node name=\"NodeB\"/>\n"
"    <Edge name=\"AB\" source=\"NodeA\" destination=\"NodeB\">\n"
"      <Predicate>type==\"6D\"</Predicate>\n"
"    </Edge>\n"
"  </Input>\n"
"\n"
"  <Output>\n"
"    <Edge name=\"BA\" source=\"NodeB\" destination=\"NodeA\">\n"
"      <Attribute name=\"type\" value=\"6D\"/>\n"
"    </Edge>\n"
"  </Output>\n"
"\n"
"  <DataflowConfiguration>\n"
"    <UbitrackLib class=\"Inversion\"/>\n"
"  </DataflowConfiguration>\n"
"</Pattern>\n"
"<Pattern name=\"Concatenation\">\n"
"  <Input>\n"
"    <Node name=\"NodeA\"/>\n"
"    <Node name=\"NodeB\"/>\n"
"    <Node name=\"NodeC\"/>\n"
"    <Edge name=\"AB\" source=\"NodeA\" destination=\"NodeB\">\n"
"      <Predicate>type==\"6D\"</Predicate>\n"
"    </Edge>\n"
"    <Edge name=\"BC\" source=\"NodeB\" destination=\"NodeC\">\n"
"      <Predicate>type==\"6D\"</Predicate>\n"
"    </Edge>\n"
"  </Input>\n"
"\n"
"  <Output>\n"
"    <Edge name=\"AC\" source=\"NodeA\" destination=\"NodeC\">\n"
"      <Attribute name=\"type\" value=\"6D\"/>\n"
"    </Edge>\n"
"  </Output>\n"
"\n"
"  <DataflowConfiguration>\n"
"    <UbitrackLib class=\"Multiplication\"/>\n"
"  </DataflowConfiguration>\n"
"</Pattern>\n"
"</UTQL>\n";

static const char* sTestQuery =
"<?xml version=\"1.0\" encoding=\"utf-8\"?>\n"
"<UTQL>\n"
"\n"
"<Pattern name=\"Query\">\n"
"  <Input>\n"
"    <Node name=\"NodeA\">\n"
"      <Predicate>id==\"ArtSheep\"</Predicate>\n"
"    </Node>\n"
"    <Node name=\"NodeB\">\n"
"      <Predicate>id==\"ArtHmd\"</Predicate>\n"
"    </Node>\n"
"    <Edge name=\"AB\" source=\"NodeA\" destination=\"NodeB\">\n"
"      <Predicate>type==\"6D\"</Predicate>\n"
"    </Edge>\n"
 "  </Input>\n"
"  <DataflowConfiguration>\n"
"    <UbitrackLib class=\"Output\"/>\n"
"  </DataflowConfiguration>\n"
"</Pattern>\n"
"</UTQL>\n";

void TestServer()
{
	// Ubitrack::Util::initLogging();
	std::istringstream issSRG(sTestSRG);
	std::istringstream issPAT(sTestPattern);
	std::istringstream issQUE(sTestQuery);

	boost::shared_ptr< Ubitrack::Graph::UTQLDocument > doc = Ubitrack::Graph::UTQLReader::processInput(issSRG);
	boost::shared_ptr< Ubitrack::Graph::UTQLDocument > pat = Ubitrack::Graph::UTQLReader::processInput(issPAT);
	boost::shared_ptr< Ubitrack::Graph::UTQLDocument > que = Ubitrack::Graph::UTQLReader::processInput(issQUE);

	BOOST_CHECK_EQUAL( doc->m_Subgraphs.size(), (unsigned)4 );
	BOOST_CHECK_EQUAL( pat->m_Subgraphs.size(), (unsigned)2 );
	BOOST_CHECK_EQUAL( que->m_Subgraphs.size(), (unsigned)1 );

	Ubitrack::Graph::UTQLServer server;

	for ( Ubitrack::Graph::UTQLDocument::SubgraphList::iterator it = doc->m_Subgraphs.begin();
		  it != doc->m_Subgraphs.end(); ++it )
	{
		server.processAnnouncement( *it, "Test" );
	}

	for ( Ubitrack::Graph::UTQLDocument::SubgraphList::iterator it = pat->m_Subgraphs.begin();
		  it != pat->m_Subgraphs.end(); ++it )
	{
		server.processAnnouncement( *it, "Test" );
	}

	for ( Ubitrack::Graph::UTQLDocument::SubgraphList::iterator it = que->m_Subgraphs.begin();
		  it != que->m_Subgraphs.end(); ++it )
	{
		server.processAnnouncement( *it, "Test" );
	}

	Ubitrack::Graph::generateDataflow( sTestQuery );
}
